Type: Package
Package: openxlsx
Title: Read, Write and Edit xlsx Files
Version: 4.2.8.1
Date: 2025-10-30
Authors@R: 
    c(person(given = "Philipp",
             family = "Schauberger",
             role = "aut",
             email = "philipp@schauberger.co.at"),
      person(given = "Alexander",
             family = "Walker",
             role = "aut",
             email = "Alexander.Walker1989@gmail.com"),
      person(given = "Luca",
             family = "Braglia",
             role = "ctb"),
      person(given = "Joshua",
             family = "Sturm",
             role = "ctb"),
      person(given = "Jan Marvin",
             family = "Garbuszus",
             role = c("ctb", "cre"),
             email = "jan.garbuszus@ruhr-uni-bochum.de"),
      person(given = "Jordan Mark",
             family = "Barbone",
             role = "ctb",
             email = "jmbarbone@gmail.com",
             comment = c(ORCID = "0000-0001-9788-3628")),
      person(given = "David",
             family = "Zimmermann",
             role = "ctb",
             email = "david_j_zimmermann@hotmail.com"),
      person(given = "Reinhold",
             family = "Kainhofer",
             role = "ctb",
             email = "reinhold@kainhofer.com"))
Description: Simplifies the creation of Excel .xlsx files by providing a
    high level interface to writing, styling and editing worksheets.
    Through the use of 'Rcpp', read/write times are comparable to the
    'xlsx' and 'XLConnect' packages with the added benefit of removing the
    dependency on Java.
License: MIT + file LICENSE
URL: https://ycphs.github.io/openxlsx/index.html,
        https://github.com/ycphs/openxlsx
BugReports: https://github.com/ycphs/openxlsx/issues
Depends: R (>= 3.3.0)
Imports: grDevices, methods, Rcpp, stats, stringi, utils, zip
Suggests: curl, formula.tools, knitr, rmarkdown, testthat
LinkingTo: Rcpp
VignetteBuilder: knitr
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
Collate: 'CommentClass.R' 'HyperlinkClass.R' 'RcppExports.R'
        'class_definitions.R' 'StyleClass.R' 'WorkbookClass.R'
        'asserts.R' 'baseXML.R' 'borderFunctions.R' 'build_workbook.R'
        'chartsheet_class.R' 'conditional_formatting.R'
        'data-fontSizeLookupTables.R' 'helperFunctions.R'
        'loadWorkbook.R' 'onUnload.R' 'openXL.R' 'openxlsx-package.R'
        'openxlsx.R' 'openxlsxCoerce.R' 'readWorkbook.R'
        'setWindowSize.R' 'sheet_data_class.R' 'utils.R'
        'workbook_column_widths.R' 'workbook_read_workbook.R'
        'workbook_write_data.R' 'worksheet_class.R' 'wrappers.R'
        'writeData.R' 'writeDataTable.R' 'writexlsx.R' 'zzz.R'
LazyData: true
NeedsCompilation: yes
Packaged: 2025-10-30 20:28:52 UTC; janmarvingarbuszus
Author: Philipp Schauberger [aut],
  Alexander Walker [aut],
  Luca Braglia [ctb],
  Joshua Sturm [ctb],
  Jan Marvin Garbuszus [ctb, cre],
  Jordan Mark Barbone [ctb] (ORCID:
    <https://orcid.org/0000-0001-9788-3628>),
  David Zimmermann [ctb],
  Reinhold Kainhofer [ctb]
Maintainer: Jan Marvin Garbuszus <jan.garbuszus@ruhr-uni-bochum.de>
Repository: CRAN
Date/Publication: 2025-10-31 06:10:51 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-26 09:33:17 UTC; unix
Archs: openxlsx.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("names<-",Workbook)
S3method(as.character,formula)
S3method(getNamedRegions,Workbook)
S3method(getNamedRegions,default)
S3method(names,Workbook)
S3method(read.xlsx,Workbook)
S3method(read.xlsx,default)
export("activeSheet<-")
export("sheetVisibility<-")
export("sheetVisible<-")
export("worksheetOrder<-")
export(activeSheet)
export(addCreator)
export(addFilter)
export(addStyle)
export(addWorksheet)
export(auto_heights)
export(buildWorkbook)
export(cloneWorksheet)
export(col2int)
export(conditionalFormat)
export(conditionalFormatting)
export(convertFromExcelRef)
export(convertToDate)
export(convertToDateTime)
export(copyWorkbook)
export(createComment)
export(createNamedRegion)
export(createStyle)
export(createWorkbook)
export(dataValidation)
export(deleteData)
export(deleteDataColumn)
export(deleteNamedRegion)
export(freezePane)
export(getBaseFont)
export(getCellRefs)
export(getCreators)
export(getDateOrigin)
export(getNamedRegions)
export(getSheetNames)
export(getStyles)
export(getTables)
export(getWindowSize)
export(get_worksheet_entries)
export(groupColumns)
export(groupRows)
export(insertImage)
export(insertPlot)
export(int2col)
export(loadWorkbook)
export(makeHyperlinkString)
export(mergeCells)
export(modifyBaseFont)
export(op.openxlsx)
export(openXL)
export(openxlsx_getOp)
export(openxlsx_setOp)
export(pageBreak)
export(pageSetup)
export(protectWorkbook)
export(protectWorksheet)
export(read.xlsx)
export(readWorkbook)
export(removeCellMerge)
export(removeColWidths)
export(removeComment)
export(removeFilter)
export(removeRowHeights)
export(removeTable)
export(removeWorksheet)
export(renameWorksheet)
export(replaceStyle)
export(saveWorkbook)
export(setColWidths)
export(setFooter)
export(setHeader)
export(setHeaderFooter)
export(setLastModifiedBy)
export(setRowHeights)
export(setWindowSize)
export(sheetVisibility)
export(sheetVisible)
export(sheets)
export(showGridLines)
export(temp_xlsx)
export(ungroupColumns)
export(ungroupRows)
export(worksheetOrder)
export(write.xlsx)
export(writeComment)
export(writeData)
export(writeDataTable)
export(writeFormula)
import(Rcpp)
import(methods)
import(stringi)
importFrom(grDevices,bmp)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colours)
importFrom(grDevices,dev.copy)
importFrom(grDevices,dev.list)
importFrom(grDevices,dev.off)
importFrom(grDevices,jpeg)
importFrom(grDevices,png)
importFrom(grDevices,rgb)
importFrom(grDevices,tiff)
importFrom(stats,pchisq)
importFrom(utils,download.file)
importFrom(utils,head)
importFrom(utils,menu)
importFrom(utils,unzip)
importFrom(zip,zipr)
useDynLib(openxlsx, .registration=TRUE)

[openxlsx](https://ycphs.github.io/openxlsx/) <img src="man/figures/logo.png" align="right" height="139" />
========


[![codecov](https://codecov.io/gh/ycphs/openxlsx/branch/master/graph/badge.svg)](https://app.codecov.io/gh/ycphs/openxlsx)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/openxlsx)](https://cran.r-project.org/package=openxlsx)
[![CRAN RStudio mirror downloads](https://cranlogs.r-pkg.org/badges/openxlsx)](https://cran.r-project.org/package=openxlsx)
[![R-CMD-check](https://github.com/ycphs/openxlsx/actions/workflows/R-CMD-check.yaml/badge.svg?branch=master)](https://github.com/ycphs/openxlsx/actions/workflows/R-CMD-check.yaml)

This [R](https://www.R-project.org/) package simplifies the creation of `.xlsx`
files by providing a high level interface to writing, styling and editing
worksheets. Through the use of [`Rcpp`](https://CRAN.R-project.org/package=Rcpp),
read/write times are comparable to the [`xlsx`](https://CRAN.R-project.org/package=xlsx)
and [`XLConnect`](https://CRAN.R-project.org/package=XLConnect) packages with
the added benefit of removing the dependency on Java.

**Note:** `openxlsx` is no longer under active development. The package is
maintained, and CRAN warnings will be fixed, but non-critical issues will not be
addressed unless accompanied by a pull request. Packages that depend on
`openxlsx` do not need to take any action, but for new developments, users are
encouraged to use alternatives like `readxl`, `writexl`, or `openxlsx2`. The
first two packages provide support for reading and writing `.xlsx` files. The
latter package is a modern reinterpretation of `openxlsx` and provides similar
functions to modify worksheets. However, it is not a drop-in replacement, so you
may want to consult resources like the
[update vignette](https://janmarvin.github.io/openxlsx2/articles/Update-from-openxlsx.html).


## Installation

### Stable version

Current stable version is available on [CRAN](https://CRAN.R-project.org/) via

```R
install.packages("openxlsx", dependencies = TRUE)
```

### Development version
```R
install.packages(c("Rcpp", "remotes"), dependencies = TRUE)
remotes::install_github("ycphs/openxlsx")
```

## Example

Explore the package with a simple example:

```R
library(openxlsx)

# Create a new workbook and add a sheet
wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")

# Write data to the sheet
writeData(wb, "Sheet 1", mtcars)

# Save the workbook
saveWorkbook(wb, "my_mtcars.xlsx", overwrite = TRUE)
```

## Bug/feature request
Please let us know which version of `openxlsx` you are using when posting bug reports.
```R
packageVersion("openxlsx")
```

## News
You can find the NEWS file [here](https://raw.githubusercontent.com/ycphs/openxlsx/master/NEWS.md).

# nolint start


require('openxlsx')


wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")
writeData(wb, 1, head(iris))

## What we expect
# yellow fill and bold test rows 1:2 cols 1:5
addStyle(wb, sheet = 1, style = createStyle(fgFill = "yellow", textDecoration = "bold"), rows = 1:2, cols = 1:5, gridExpand = TRUE, stack = TRUE)

# yellow fill and bold test rows 1:2 cols 1:5
# red fill for row 1 and italic
addStyle(wb, sheet = 1, style = createStyle(fgFill =  "red", textDecoration = "italic"), rows = 1, cols = 1:5, gridExpand = TRUE, stack = TRUE)

# ## add a bluw line at row 5
addStyle(wb, sheet = 1, style = createStyle(fgFill =  "blue"), rows = 5, cols = 1:5, gridExpand = TRUE, stack = TRUE) ## non-intersecting

# ## Now borders and underlined around rows 1:3 for columns 1 and 5
addStyle(wb, sheet = 1, style = createStyle(border = "topbottomleftright", textDecoration = "underline"), rows = 2:3, cols = c(1, 5), gridExpand = TRUE, stack = TRUE)
# 
# ## Now blue border only on top for rows 1:3, column 1
addStyle(wb, sheet = 1, style = createStyle(border = "top", borderColour = "blue"), rows = 1:3, cols = 1, gridExpand = TRUE, stack = TRUE)

# 
# ## no stack! Wipe all formatting and put all black borders rows 1:4, col 3
# addStyle(wb, sheet = 1, style = createStyle(border = "topbottomleftright"), rows = 1:4, cols = c(3,3,3,3), stack = FALSE)
# 
# ## cell 3,3 red bottom border
# addStyle(wb, sheet = 1, style = createStyle(border = "bottom", borderColour = "red"), rows = 2:10, cols = 3, gridExpand = TRUE, stack = TRUE)


openXL(wb)

wb$addStyle







## Now not stacking
addWorksheet(wb, "Sheet 2")
writeData(wb, 2, matrix("abc", nrow = 4, ncol = 5))
addStyle(wb, 2, createStyle(halign = "center", border = "TopBottomLeftRight"), 1:5, 1:5, gridExpand = TRUE)
addStyle(wb, 2, createStyle(textDecoration = "bold", fgFill = "salmon"), 2:4, 2:4,gridExpand = F, stack = TRUE)

## STACk == TRUE
addWorksheet(wb, "Sheet 3")
writeData(wb, 3, matrix("abc", nrow = 4, ncol = 5))
addStyle(wb, 3, createStyle(halign = "center", border = "TopBottomLeftRight"), 1:5, 1:5, gridExpand = TRUE)
addStyle(wb, 3, createStyle(textDecoration = "bold", fgFill = "salmon"), 2:4, 2:4,gridExpand = F, stack = TRUE)



openXL(wb)

















## TEST NUMBER 2 - BUG REPORT #203

wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")
writeData(wb, 1, head(iris))



## Make a red block
addStyle(wb, sheet = 1, style = createStyle(fgFill =  "red", textDecoration = "italic"), rows = c(2, 3, 4), cols = 2:5, gridExpand = TRUE, stack = TRUE)

## Draw a yellow L around it
addStyle(wb, sheet = 1, style = createStyle(fgFill = "yellow", textDecoration = "bold"), rows = c(1,2,3,4,5,5,5,5,5), cols = c(1,1,1,1,1,2,3,4,5), gridExpand = FALSE, stack = TRUE)
# addStyle(wb, sheet = 1, style = createStyle(fgFill = "yellow", textDecoration = "bold"), rows = 5, cols = 1:5, gridExpand = TRUE, stack = TRUE)


## Now borders and underlined around rows 1:3 for columns 1 and 5
addStyle(wb, sheet = 1, style = createStyle(border = "topbottomleftright", textDecoration = "underline"), rows = 1:3, cols = c(1, 5), gridExpand = TRUE, stack = TRUE)

## Now blue border only on top for rows 1:3, column 1
addStyle(wb, sheet = 1, style = createStyle(border = "top", borderColour = "blue"), rows = 1:3, cols = 1, gridExpand = TRUE, stack = TRUE)


## no stack! Wipe all formatting and put all black borders rows 1:4, col 3
addStyle(wb, sheet = 1, style = createStyle(border = "topbottomleftright"), rows = 1:4, cols = c(3,3,3,3))

## cell 3,3 red bottom border
addStyle(wb, sheet = 1, style = createStyle(border = "bottom", borderColour = "red"), rows = 2:10, cols = 3, gridExpand = TRUE, stack = TRUE)




## Now not stacking
addWorksheet(wb, "Sheet 2")
writeData(wb, 2, matrix("abc", nrow = 4, ncol = 5))
addStyle(wb, 2, createStyle(halign = "center", border = "TopBottomLeftRight"), 1:5, 1:5, gridExpand = TRUE)
addStyle(wb, 2, createStyle(textDecoration = "bold", fgFill = "salmon"), 2:4, 2:4,gridExpand = F, stack = TRUE)

## STACk == TRUE
addWorksheet(wb, "Sheet 3")
writeData(wb, 3, matrix("abc", nrow = 4, ncol = 5))
addStyle(wb, 3, createStyle(halign = "center", border = "TopBottomLeftRight"), 1:5, 1:5, gridExpand = TRUE)
addStyle(wb, 3, createStyle(textDecoration = "bold", fgFill = "salmon"), 2:4, 2:4,gridExpand = F, stack = TRUE)



openXL(wb)

# nolint end

# nolint start
require('openxlsx')

unzip_xlsx <- function(fl){
  
  wd <- getwd()
  d <- file.path(tempdir(), paste(sample(LETTERS, 10), collapse = ""))
  unlink(d, recursive = TRUE, force = TRUE)
  dir.create(d)
  
  new_fl <- file.path(d, basename(fl))
  file.copy(from = fl, to = new_fl)
  
  setwd(d)
  unzip(zipfile = new_fl, junkpaths = FALSE)
  cmd <- paste("start", d)
  shell(cmd)
  
  unlink(new_fl)
  
  setwd(wd)
  
  return(d)
  
}


zip_xlsx <- function(d){
  
  wd <- getwd()
  setwd(d)
  
  zipfile = "a.xlsx"
  files <- list.files()
  flags = "-r1"
  extras = ""
  zip = Sys.getenv("R_ZIPCMD", "zip")
  args <- c(flags, shQuote(path.expand(zipfile)), shQuote(files), extras)
  res <- invisible(suppressWarnings(system2(zip, args, stdout = NULL)))
  setwd(wd)
  
  
}

## Get loading files
# devtools::install_github("awalker89/openxlsx_testing_files", force = TRUE)

## To install from CRAN
# detach("package:openxlsx", unload=TRUE)
# install.packages("openxlsx")


test_file_dir <- system.file(package = "openxlsx.testing.files")

################################################################################################################
## All Features
wb <- loadWorkbook(file.path(test_file_dir, "All_Features.xlsx"))
openXL(wb)


################################################################################################################
## Budget Template
wb <- loadWorkbook(file.path(test_file_dir, "Budget.xlsx"))
openXL(wb)
openXL(file.path(test_file_dir, "Budget.xlsx"))




################################################################################################################
## Chart Sheet
wb <- loadWorkbook(file.path(test_file_dir, "Chart_Sheet_Test.xlsx"))
openXL(wb)



################################################################################################################
## Chineses Characters
wb <- loadWorkbook(file.path(test_file_dir, "Chinese_Characters.xlsx"))
openXL(wb)


################################################################################################################
## Excel Diet Template
wb <- loadWorkbook(file.path(test_file_dir, "Diet.xlsx"))
openXL(wb)
openXL(file.path(test_file_dir, "Diet.xlsx"))

################################################################################################################
## Empty Workbook
wb <- loadWorkbook(file.path(test_file_dir, "empty.xlsx"))
openXL(wb)


################################################################################################################
## Encoding Test
wb <- loadWorkbook(file.path(test_file_dir, "Encoding_Test.xlsx"))
openXL(wb)
openXL(file.path(test_file_dir, "Encoding_Test.xlsx"))

################################################################################################################
## Libre Office Test File
wb <- loadWorkbook(file.path(test_file_dir, "libre_test.xlsx"))
openXL(wb)

wb <- loadWorkbook(file.path(test_file_dir, "libre_test2.xlsx"))
openXL(wb)


################################################################################################################
## Load Example Workbook
wb <- loadWorkbook(system.file("loadExample.xlsx", package = "openxlsx"))
openXL(wb)
openXL(system.file("loadExample.xlsx", package = "openxlsx"))

################################################################################################################
## Loading Pivot Tables
wb <- loadWorkbook(file.path(test_file_dir, "pivotTest.xlsx"))
openXL(wb)

wb <- loadWorkbook(file.path(test_file_dir, "pivotTest2.xlsx"))
openXL(wb)

wb <- loadWorkbook(file.path(test_file_dir, "pivotTest3.xlsx"))
openXL(wb)


################################################################################################################
## Excel Template (Sales call log and organizer1.xlsx)
wb <- loadWorkbook(file.path(test_file_dir, "Sales call log and organizer1.xlsx"))
openXL(wb)


################################################################################################################
## Whitespace - maintain whitespace
wb <- loadWorkbook(file.path(test_file_dir, "Whitespace_Test.xlsx"))
openXL(wb)


################################################################################################################
## Weight Tracket Excel Template
wb <- loadWorkbook(file.path(test_file_dir, "WeightTrackerTemplate.xlsx"))
openXL(wb)

################################################################################################################
## Schedule Excel Template
wb <- loadWorkbook(file = file.path(test_file_dir, "Schedule Template.xlsx"))
openXL(wb)


################################################################################################################
## package Example File
wb <- loadWorkbook(file = system.file("loadExample.xlsx", package = "openxlsx"))
openXL(wb)




## write jsuts a date
wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")
writeData(wb, 1, as.Date('2014-01-01'))
openXL(wb)

wb <- loadWorkbook(file.path(test_file_dir, "pivotTest.xlsx"))
writeData(wb, 1, iris[,1:3]*100, colNames = FALSE, startRow = 2)
openXL(wb)

openXL(file.path(test_file_dir, "pivotTest.xlsx"))





# nolint end

## ----setup, include = FALSE---------------------------------------------------
library(openxlsx)

## ----include = TRUE, eval = FALSE, highlight = TRUE---------------------------
# ## data.frame to write
# df <- data.frame("Date" = Sys.Date()-0:4,
#                  "Logical" = c(TRUE, FALSE, TRUE, TRUE, FALSE),
#                  "Currency" = paste("$",-2:2),
#                  "Accounting" = -2:2,
#                  "hLink" = "https://CRAN.R-project.org/",
#                  "Percentage" = seq(-1, 1, length.out=5),
#                  "TinyNumber" = runif(5) / 1E9, stringsAsFactors = FALSE)
# 
# class(df$Currency) <- "currency"
# class(df$Accounting) <- "accounting"
# class(df$hLink) <- "hyperlink"
# class(df$Percentage) <- "percentage"
# class(df$TinyNumber) <- "scientific"
# 
# ## Formatting can be applied simply through the write functions
# ## global options can be set to further simplify things
# options("openxlsx.borderStyle" = "thin")
# options("openxlsx.borderColour" = "#4F81BD")
# 
# ## create a workbook and add a worksheet
# wb <- createWorkbook()
# addWorksheet(wb, "writeData auto-formatting")
# 
# writeData(wb, 1, df, startRow = 2, startCol = 2)
# writeData(wb, 1, df, startRow = 9, startCol = 2, borders = "surrounding")
# writeData(wb, 1, df, startRow = 16, startCol = 2, borders = "rows")
# writeData(wb, 1, df, startRow = 23, startCol = 2, borders ="columns")
# writeData(wb, 1, df, startRow = 30, startCol = 2, borders ="all")
# 
# ## headerStyles
# hs1 <- createStyle(fgFill = "#4F81BD", halign = "CENTER", textDecoration = "Bold",
#                    border = "Bottom", fontColour = "white")
# 
# writeData(wb, 1, df, startRow = 16, startCol = 10, headerStyle = hs1,
#           borders = "rows", borderStyle = "medium")
# 
# ## to change the display text for a hyperlink column just write over those cells
# writeData(wb, sheet = 1, x = paste("Hyperlink", 1:5), startRow = 17, startCol = 14)
# 
# 
# ## writing as an Excel Table
# 
# addWorksheet(wb, "writeDataTable")
# writeDataTable(wb, 2, df, startRow = 2, startCol = 2)
# writeDataTable(wb, 2, df, startRow = 9, startCol = 2, tableStyle = "TableStyleLight9")
# writeDataTable(wb, 2, df, startRow = 16, startCol = 2, tableStyle = "TableStyleLight2")
# writeDataTable(wb, 2, df, startRow = 23, startCol = 2, tableStyle = "TableStyleMedium21")
# 
# openXL(wb) ## opens a temp version

## ----include = TRUE, eval = FALSE, highlight = TRUE---------------------------
# # data.frame of dates
# dates <- data.frame("d1" = Sys.Date() - 0:4)
# for(i in 1:3) dates <- cbind(dates, dates)
# names(dates) <- paste0("d", 1:8)
# 
# ## Date Formatting
# wb <- createWorkbook()
# addWorksheet(wb, "Date Formatting", gridLines = FALSE)
# writeData(wb, 1, dates) ## write without styling
# 
# ## openxlsx converts columns of class "Date" to Excel dates with the format given by
# getOption("openxlsx.dateFormat", "mm/dd/yyyy")
# 
# ## this can be set via (for example)
# options("openxlsx.dateFormat" = "yyyy/mm/dd")
# ## custom date formats can be made up of any combination of:
# ##   d, dd, ddd, dddd, m, mm, mmm, mmmm, mmmmm, yy, yyyy
# 
# ## numFmt == "DATE" will use the date format specified by the above
# addStyle(wb, 1, style = createStyle(numFmt = "DATE"), rows = 2:11, cols = 1, gridExpand = TRUE)
# 
# ## some custom date format examples
# sty <- createStyle(numFmt = "yyyy/mm/dd")
# addStyle(wb, 1, style = sty, rows = 2:11, cols = 2, gridExpand = TRUE)
# 
# sty <- createStyle(numFmt = "yyyy/mmm/dd")
# addStyle(wb, 1, style = sty, rows = 2:11, cols = 3, gridExpand = TRUE)
# 
# sty <- createStyle(numFmt = "yy / mmmm / dd")
# addStyle(wb, 1, style = sty, rows = 2:11, cols = 4, gridExpand = TRUE)
# 
# sty <- createStyle(numFmt = "ddddd")
# addStyle(wb, 1, style = sty, rows = 2:11, cols = 5, gridExpand = TRUE)
# 
# sty <- createStyle(numFmt = "yyyy-mmm-dd")
# addStyle(wb, 1, style = sty, rows = 2:11, cols = 6, gridExpand = TRUE)
# 
# sty <- createStyle(numFmt = "mm/ dd yyyy")
# addStyle(wb, 1, style = sty, rows = 2:11, cols = 7, gridExpand = TRUE)
# 
# sty <- createStyle(numFmt = "mm/dd/yy")
# addStyle(wb, 1, style = sty, rows = 2:11, cols = 8, gridExpand = TRUE)
# 
# setColWidths(wb, 1, cols = 1:10, widths = 23)
# 
# ## The default date format used in writeData and writeDataTable can be set with:
# options("openxlsx.dateFormat" = "dd/mm/yyyy")
# writeData(wb, "Date Formatting", dates, startRow  = 8, borders = "rows")
# options("openxlsx.dateFormat" = "yyyy-mm-dd")
# writeData(wb, "Date Formatting", dates, startRow  = 15)
# 
# saveWorkbook(wb, "Date Formatting.xlsx", overwrite = TRUE)

## ----include = TRUE, eval = FALSE, highlight = TRUE---------------------------
# Sys.setenv(TZ = "Australia/Sydney")
# 
# dateTimes <- data.frame("d1" = Sys.time() - 0:4*10000)
# for(i in 1:2) dateTimes <- cbind(dateTimes, dateTimes)
# names(dateTimes) <- paste0("d", 1:4)
# 
# ## POSIX Formatting
# wb <- createWorkbook()
# addWorksheet(wb, "DateTime Formatting", gridLines = FALSE)
# writeData(wb, 1, dateTimes) ## write without styling
# 
# ## openxlsx converts columns of class "POSIxt" to Excel datetimes with the format given by
# getOption("openxlsx.datetimeFormat", "yyyy/mm/dd hh:mm:ss")
# 
# ## this can be set via (for example)
# options("openxlsx.datetimeFormat" = "yyyy-mm-dd hh:mm:ss")
# ## custom datetime formats can be made up of any combination of:
# ## d, dd, ddd, dddd, m, mm, mmm, mmmm, mmmmm, yy, yyyy, h, hh, m, mm, s, ss, AM/PM
# 
# ## numFmt == "LONGDATE" will use the date format specified by the above
# long_date_style <- createStyle(numFmt = "LONGDATE")
# addStyle(wb, 1, style = long_date_style, rows = 2:11, cols = 1, gridExpand = TRUE)
# 
# ## some custom date format examples
# sty <- createStyle(numFmt = "yyyy/mm/dd hh:mm:ss AM/PM")
# addStyle(wb, 1, style = sty, rows = 2:11, cols = 2, gridExpand = TRUE)
# 
# sty <- createStyle(numFmt = "hh:mm:ss AM/PM")
# addStyle(wb, 1, style = sty, rows = 2:11, cols = 3, gridExpand = TRUE)
# 
# sty <- createStyle(numFmt = "hh:mm:ss")
# addStyle(wb, 1, style = sty, rows = 2:11, cols = 4, gridExpand = TRUE)
# 
# setColWidths(wb, 1, cols = 1:4, widths = 30)
# 
# ## The default date format used in writeData and writeDataTable can be set with:
# options("openxlsx.datetimeFormat" = "yyyy/mm/dd hh:mm:ss")
# writeData(wb, "DateTime Formatting", dateTimes, startRow  = 8, borders = "rows")
# 
# options("openxlsx.datetimeFormat" = "hh:mm:ss AM/PM")
# writeDataTable(wb, "DateTime Formatting", dateTimes, startRow  = 15)
# 
# saveWorkbook(wb, "DateTime Formatting.xlsx", overwrite = TRUE)
# openXL("DateTime Formatting.xlsx")

## ----include = TRUE, eval = FALSE, highlight = TRUE---------------------------
# wb <- createWorkbook()
# addWorksheet(wb, "cellIs")
# addWorksheet(wb, "Moving Row")
# addWorksheet(wb, "Moving Col")
# addWorksheet(wb, "Dependent on 1")
# addWorksheet(wb, "Duplicates")
# addWorksheet(wb, "containsText")
# addWorksheet(wb, "colourScale", zoom = 30)
# addWorksheet(wb, "databar")
# 
# negStyle <- createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
# posStyle <- createStyle(fontColour = "#006100", bgFill = "#C6EFCE")
# 
# ## rule applies to all each cell in range
# writeData(wb, "cellIs", -5:5)
# writeData(wb, "cellIs", LETTERS[1:11], startCol=2)
# conditionalFormatting(wb, "cellIs", cols=1, rows=1:11, rule="!=0", style = negStyle)
# conditionalFormatting(wb, "cellIs", cols=1, rows=1:11, rule="==0", style = posStyle)
# 
# ## highlight row dependent on first cell in row
# writeData(wb, "Moving Row", -5:5)
# writeData(wb, "Moving Row", LETTERS[1:11], startCol=2)
# conditionalFormatting(wb, "Moving Row", cols=1:2, rows=1:11, rule="$A1<0", style = negStyle)
# conditionalFormatting(wb, "Moving Row", cols=1:2, rows=1:11, rule="$A1>0", style = posStyle)
# 
# ## highlight column dependent on first cell in column
# writeData(wb, "Moving Col", -5:5)
# writeData(wb, "Moving Col", LETTERS[1:11], startCol=2)
# conditionalFormatting(wb, "Moving Col", cols=1:2, rows=1:11, rule="A$1<0", style = negStyle)
# conditionalFormatting(wb, "Moving Col", cols=1:2, rows=1:11, rule="A$1>0", style = posStyle)
# 
# ## highlight entire range cols X rows dependent only on cell A1
# writeData(wb, "Dependent on 1", -5:5)
# writeData(wb, "Dependent on 1", LETTERS[1:11], startCol=2)
# conditionalFormatting(wb, "Dependent on 1", cols=1:2, rows=1:11, rule="$A$1<0", style = negStyle)
# conditionalFormatting(wb, "Dependent on 1", cols=1:2, rows=1:11, rule="$A$1>0", style = posStyle)
# 
# ## highlight duplicates using default style
# writeData(wb, "Duplicates", sample(LETTERS[1:15], size = 10, replace = TRUE))
# conditionalFormatting(wb, "Duplicates", cols = 1, rows = 1:10, type = "duplicates")
# 
# ## cells containing text
# fn <- function(x) paste(sample(LETTERS, 10), collapse = "-")
# writeData(wb, "containsText", sapply(1:10, fn))
# conditionalFormatting(wb, "containsText", cols = 1, rows = 1:10, type = "contains", rule = "A")
# 
# ## colourscale colours cells based on cell value
# df <- read.xlsx(system.file("extdata", "readTest.xlsx", package = "openxlsx"), sheet = 5)
# writeData(wb, "colourScale", df, colNames=FALSE)  ## write data.frame
# 
# ## rule is a vector or colours of length 2 or 3 (any hex colour or any of colours())
# ## If rule is NULL, min and max of cells is used. Rule must be the same length as style or NULL.
# conditionalFormatting(wb, "colourScale", cols=seq_len(ncol(df)), rows=seq_len(nrow(df)),
#    style = c("black", "white"),
#    rule = c(0, 255),
#    type = "colourScale")
# 
# setColWidths(wb, "colourScale", cols = seq_len(ncol(df)), widths = 1.07)
# setRowHeights(wb, "colourScale", rows = seq_len(nrow(df)), heights = 7.5)
# 
# ## Databars
# writeData(wb, "databar", -5:5)
# conditionalFormatting(wb, "databar", cols = 1, rows = 1:12, type = "databar") ## Default colours
# 
# saveWorkbook(wb, "conditionalFormattingExample.xlsx", TRUE)
# 
# openXL(wb)

## ----include = TRUE, eval = FALSE, highlight = TRUE---------------------------
# options("openxlsx.numFmt" = NULL)
# wb <- createWorkbook()
# addWorksheet(wb, "Sheet 1")
# df <- data.frame(matrix(12.987654321, ncol = 7, nrow = 5)) ## data.frame to write
# df[ ,6:7] <- df[ ,6:7]*1E6
# 
# ## Set column 1 class to "comma" to get comma separated thousands
# class(df$X1) <- "comma"
# 
# writeData(wb, 1, df)
# s <- createStyle(numFmt = "0.0")
# addStyle(wb, 1, style = s, rows = 2:6, cols = 2, gridExpand = TRUE)
# 
# s <- createStyle(numFmt = "0.00")
# addStyle(wb, 1, style = s, rows = 2:6, cols = 3, gridExpand = TRUE)
# 
# s <- createStyle(numFmt = "0.000")
# addStyle(wb, 1, style = s, rows = 2:6, cols = 4, gridExpand = TRUE)
# 
# s <- createStyle(numFmt = "#,##0")
# addStyle(wb, 1, style = s, rows = 2:6, cols = 5, gridExpand = TRUE)
# 
# s <- createStyle(numFmt = "#,##0.00")
# addStyle(wb, 1, style = s, rows = 2:6, cols = 6, gridExpand = TRUE)
# 
# s <- createStyle(numFmt = "$ #,##0.00")
# addStyle(wb, 1, style = s, rows = 2:6, cols = 7, gridExpand = TRUE)
# 
# ## set a default number format for numeric columns of data.frames
# options("openxlsx.numFmt" = "$* #,#0.00")
# writeData(wb, 1, x = data.frame("Using Default Options" = rep(2345.1235, 5)), startCol = 9)
# 
# setColWidths(wb, 1, cols = 1:10, widths = 15)
# 
# ## Using default numFmt to round to 2 dp (Any numeric column will be affected)
# addWorksheet(wb, "Sheet 2")
# df <- iris
# df[, 1:4] <- df[1:4] + runif(1)
# writeDataTable(wb, sheet = 2, x = df)
# writeData(wb, sheet = 2, x = df, startCol = 7)
# writeData(wb, sheet = 2, x = df, startCol = 13, borders = "rows")
# 
# ## To stop auto-formatting numerics set
# options("openxlsx.numFmt" = NULL)
# addWorksheet(wb, "Sheet 3")
# writeDataTable(wb, sheet = 3, x = df)
# 
# openXL(wb)

## ----cleanup, eval = FALSE, include = FALSE-----------------------------------
# xlsx_files <- dir(pattern = "*.xlsx")
# unlink(xlsx_files)

# nolint start


wb <- createWorkbook()
addWorksheet(wb, "cellIs")
addWorksheet(wb, "Moving Row")
addWorksheet(wb, "Moving Col")
addWorksheet(wb, "Dependent on 1")
addWorksheet(wb, "Duplicates")
addWorksheet(wb, "containsText")
addWorksheet(wb, "colourScale", zoom = 30)
addWorksheet(wb, "databar")

negStyle <- createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
posStyle <- createStyle(fontColour = "#006100", bgFill = "#C6EFCE")

## rule applies to all each cell in range
writeData(wb, "cellIs", -5:5)
writeData(wb, "cellIs", LETTERS[1:11], startCol=2)
conditionalFormatting(wb, "cellIs", cols=1, rows=2:12, rule="!=0", style = negStyle)
conditionalFormatting(wb, "cellIs", cols=1, rows=2:12, rule="==0", style = posStyle)

## highlight row dependent on first cell in row
writeData(wb, "Moving Row", -5:5)
writeData(wb, "Moving Row", LETTERS[1:11], startCol=2)
conditionalFormatting(wb, "Moving Row", cols=1:2, rows=2:12, rule="$A1<0", style = negStyle)
conditionalFormatting(wb, "Moving Row", cols=1:2, rows=2:12, rule="$A1>0", style = posStyle)

## highlight column dependent on first cell in column
writeData(wb, "Moving Col", -5:5)
writeData(wb, "Moving Col", LETTERS[1:11], startCol=2)
conditionalFormatting(wb, "Moving Col", cols=1:2, rows=2:12, rule="A$1<0", style = negStyle)
conditionalFormatting(wb, "Moving Col", cols=1:2, rows=2:12, rule="A$1>0", style = posStyle)

## highlight entire range cols X rows dependent only on cell A1
writeData(wb, "Dependent on 1", -5:5)
writeData(wb, "Dependent on 1", LETTERS[1:11], startCol=2)
conditionalFormatting(wb, "Dependent on 1", cols=1:2, rows=2:12, rule="$A$1<0", style = negStyle)
conditionalFormatting(wb, "Dependent on 1", cols=1:2, rows=2:12, rule="$A$1>0", style = posStyle)

## highlight duplicates using default style
writeData(wb, "Duplicates", sample(LETTERS[1:15], size = 10, replace = TRUE))
conditionalFormatting(wb, "Duplicates", cols = 1, rows = 1:10, type = "duplicates")

## cells containing text
fn <- function(x) paste(sample(LETTERS, 10), collapse = "-")
writeData(wb, "containsText", sapply(1:10, fn))
conditionalFormatting(wb, "containsText", cols = 1, rows = 1:10, type = "contains", rule = "A")

## colourscale colours cells based on cell value
df <- read.xlsx(system.file("extdata","readTest.xlsx", package = "openxlsx"), sheet = 4)
writeData(wb, "colourScale", df, colNames=FALSE)  ## write data.frame

## rule is a vector or colours of length 2 or 3 (any hex colour or any of colours())
conditionalFormatting(wb, "colourScale", cols=1:ncol(df), rows=1:nrow(df),
   style = c("black", "red", "white"), 
   rule = c(0, 100, 255), #If rule is NULL, min and max are used. Rule must be the same length as style or NULL.
   type = "colourScale")

setColWidths(wb, "colourScale", cols = 1:ncol(df), widths = 1.07)
setRowHeights(wb, "colourScale", rows = 1:nrow(df), heights = 7.5) 

## Databars
writeData(wb, "databar", -5:5)
conditionalFormatting(wb, "databar", cols = 1, rows = 1:12, type = "databar") ## Default colours

fl <- tempfile(fileext = ".xlsx")
saveWorkbook(wb, file = fl, overwrite = TRUE)

openXL(wb)




wb <- loadWorkbook(fl)
conditionalFormatting(wb, "Duplicates", cols = 1, rows = 1:10, type = "duplicates", style = createStyle(textDecoration = 'BOLD'))
openXL(wb)


# nolint end

# nolint start

options("scipen" = 10000)

## loop through all fonts
fontDir <- "C:\\Users\\Alex\\Desktop\\font_workbooks"
files <- list.files(fontDir, patter = "\\.xlsx$", full.names = TRUE)
files <- files[!grepl("-bold.xlsx", files)]

files2 <- list.files(fontDir, patter = "\\.xlsx$", full.names = FALSE)
files2 <- files2[!grepl("-bold.xlsx", files2)]

font <- tolower(gsub(" ", ".", gsub("\\.xlsx", "", files2)))


strs <- "openxlsxFontSizeLookupTable <- \ndata.frame("
allWidths <- rep(8.43, 29)
names(allWidths) <- 1:29
for(i in seq_along(files)){
  
  f <- font[[i]]
  widths <- round(as.numeric(read.xlsx(files[[i]])[2,]), 6)
  strs <- c(strs, sprintf('"%s"= c(%s),\n', f, paste(widths, collapse = ", ")))
  
}

strs[length(strs)] <- gsub(",\n", ")", strs[length(strs)])


## bold ones

## loop through all fonts
fontDir <- "C:\\Users\\Alex\\Desktop\\font_workbooks"
files <- list.files(fontDir, patter = "\\.xlsx$", full.names = TRUE)
files <- files[grepl("-bold.xlsx", files)]

files2 <- list.files(fontDir, patter = "\\.xlsx$", full.names = FALSE)
files2 <- files2[grepl("-bold.xlsx", files2)]

font <- tolower(gsub(" ", ".", gsub("\\-bold.xlsx", "", files2)))


strsBold <- "openxlsxFontSizeLookupTableBold <- \ndata.frame("
allWidths <- rep(8.43, 29)
names(allWidths) <- 1:29
for(i in seq_along(files)){
  
  f <- font[[i]]
  widths <- round(as.numeric(read.xlsx(files[[i]])[2,]), 6)
  strsBold <- c(strsBold, sprintf('"%s"= c(%s),\n', f, paste(widths, collapse = ", ")))
  
}

strsBold[length(strsBold)] <- gsub(",\n", ")", strsBold[length(strsBold)])


allStrs <- c(strs, "\n\n\n", strsBold)
cat(allStrs)





# nolint end

## ----include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE--------------
# ## write to working directory
# library(openxlsx)
# write.xlsx(iris, file = "writeXLSX1.xlsx")
# write.xlsx(iris, file = "writeXLSXTable1.xlsx", asTable = TRUE)

## ----include = TRUE, tidy = TRUE, eval = FALSE ,highlight = TRUE--------------
# ## write a list of data.frames to individual worksheets using list names as worksheet names
# l <- list("IRIS" = iris, "MTCARS" = mtcars)
# write.xlsx(l, file = "writeXLSX2.xlsx")
# write.xlsx(l, file = "writeXLSXTable2.xlsx", asTable = TRUE)

## ----include = TRUE, tidy = TRUE, eval = FALSE ,highlight= TRUE---------------
# options("openxlsx.borderColour" = "#4F80BD")
# options("openxlsx.borderStyle" = "thin")
# options("openxlsx.dateFormat" = "mm/dd/yyyy")
# options("openxlsx.datetimeFormat" = "yyyy-mm-dd hh:mm:ss")
# options("openxlsx.numFmt" = NULL) ## For default style rounding of numeric columns
# 
# df <- data.frame("Date" = Sys.Date()-0:19, "LogicalT" = TRUE,
#                  "Time" = Sys.time()-0:19*60*60,
#                  "Cash" = paste("$",1:20), "Cash2" = 31:50,
#                  "hLink" = "https://CRAN.R-project.org/",
#                  "Percentage" = seq(0, 1, length.out=20),
#                  "TinyNumbers" = runif(20) / 1E9,  stringsAsFactors = FALSE)
# 
# class(df$Cash) <- "currency"
# class(df$Cash2) <- "accounting"
# class(df$hLink) <- "hyperlink"
# class(df$Percentage) <- "percentage"
# class(df$TinyNumbers) <- "scientific"
# 
# write.xlsx(df, "writeXLSX3.xlsx")
# write.xlsx(df, file = "writeXLSXTable3.xlsx", asTable = TRUE)
# 
# 

## ----include = TRUE, tidy = TRUE, eval = FALSE, highlight= TRUE---------------
# hs <- createStyle(fontColour = "#ffffff", fgFill = "#4F80BD",
#                   halign = "center", valign = "center", textDecoration = "Bold",
#                   border = "TopBottomLeftRight", textRotation = 45)
# 
# write.xlsx(iris, file = "writeXLSX4.xlsx", borders = "rows", headerStyle = hs)
# write.xlsx(iris, file = "writeXLSX5.xlsx", borders = "columns", headerStyle = hs)
# 
# write.xlsx(iris, "writeXLSXTable4.xlsx", asTable = TRUE,
# headerStyle = createStyle(textRotation = 45))

## ----include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE--------------
# l <- list("IRIS" = iris, "colClasses" = df)
# write.xlsx(l, file = "writeXLSX6.xlsx", borders = "columns", headerStyle = hs)
# write.xlsx(l, file = "writeXLSXTable5.xlsx", asTable = TRUE, tableStyle = "TableStyleLight2")
# 
# openXL("writeXLSX6.xlsx")
# openXL("writeXLSXTable5.xlsx")

## ----include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE--------------
# wb <- write.xlsx(iris, "writeXLSX6.xlsx")
# setColWidths(wb, sheet = 1, cols = 1:5, widths = 20)
# saveWorkbook(wb, "writeXLSX6.xlsx", overwrite = TRUE)

## ----include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE--------------
# require(ggplot2)
# wb <- createWorkbook()
# options("openxlsx.borderColour" = "#4F80BD")
# options("openxlsx.borderStyle" = "thin")
# modifyBaseFont(wb, fontSize = 10, fontName = "Arial Narrow")

## ----include = TRUE,tidy = TRUE, eval = FALSE, highlight = TRUE---------------
# addWorksheet(wb, sheetName = "Motor Trend Car Road Tests", gridLines = FALSE)
# addWorksheet(wb, sheetName = "Iris", gridLines = FALSE)

## ----include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE--------------
# freezePane(wb, sheet = 1, firstRow = TRUE, firstCol = TRUE) ## freeze first row and column
# writeDataTable(wb, sheet = 1, x = mtcars,
# colNames = TRUE, rowNames = TRUE,
# tableStyle = "TableStyleLight9")
# 
# setColWidths(wb, sheet = 1, cols = "A", widths = 18)

## ----include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE--------------
# writeDataTable(wb, sheet = 2, iris, startCol = "K", startRow = 2)
# 
# qplot(data=iris, x = Sepal.Length, y= Sepal.Width, colour = Species)
# insertPlot(wb, 2, xy=c("B", 16)) ## insert plot at cell B16
# 
# means <- aggregate(x = iris[,-5], by = list(iris$Species), FUN = mean)
# vars <- aggregate(x = iris[,-5], by = list(iris$Species), FUN = var)

## ----include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE--------------
# headSty <- createStyle(fgFill="#DCE6F1", halign="center", border = "TopBottomLeftRight")
# writeData(wb, 2, x = "Iris dataset group means", startCol = 2, startRow = 2)
# writeData(wb, 2, x = means, startCol = "B", startRow=3, borders="rows", headerStyle = headSty)

## ----include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE--------------
# writeData(wb, 2, x = "Iris dataset group variances", startCol = 2, startRow = 9)
# writeData(wb, 2, x= vars, startCol = "B", startRow=10, borders="columns",
# headerStyle = headSty)
# 
# setColWidths(wb, 2, cols=2:6, widths = 12) ## width is recycled for each col
# setColWidths(wb, 2, cols=11:15, widths = 15)

## ----include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE--------------
# s1 <- createStyle(fontSize=14, textDecoration=c("bold", "italic"))
# addStyle(wb, 2, style = s1, rows=c(2,9), cols=c(2,2))

## ----include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE--------------
# saveWorkbook(wb, "basics.xlsx", overwrite = TRUE) ## save to working directory

## ----eval = FALSE, include = TRUE---------------------------------------------
# ## inspired by xtable gallery
# #https://CRAN.R-project.org/package=xtable/vignettes/xtableGallery.pdf
# 
# ## Create a new workbook
# wb <- createWorkbook()
# data(tli, package = "xtable")
# 
# ## data.frame
# test.n <- "data.frame"
# my.df <- tli[1:10, ]
# addWorksheet(wb = wb, sheetName = test.n)
# writeData(wb = wb, sheet = test.n, x = my.df, borders = "n")
# 
# ## matrix
# test.n <- "matrix"
# design.matrix <- model.matrix(~ sex * grade, data = my.df)
# addWorksheet(wb = wb, sheetName = test.n)
# writeData(wb = wb, sheet = test.n, x = design.matrix)
# 
# ## aov
# test.n <- "aov"
# fm1 <- aov(tlimth ~ sex + ethnicty + grade + disadvg, data = tli)
# addWorksheet(wb = wb, sheetName = test.n)
# writeData(wb = wb, sheet = test.n, x = fm1)
# 
# ## lm
# test.n <- "lm"
# fm2 <- lm(tlimth ~ sex*ethnicty, data = tli)
# addWorksheet(wb = wb, sheetName = test.n)
# writeData(wb = wb, sheet = test.n, x = fm2)
# 
# ## anova 1
# test.n <- "anova"
# my.anova <- anova(fm2)
# addWorksheet(wb = wb, sheetName = test.n)
# writeData(wb = wb, sheet = test.n, x = my.anova)
# 
# ## anova 2
# test.n <- "anova2"
# fm2b <- lm(tlimth ~ ethnicty, data = tli)
# my.anova2 <- anova(fm2b, fm2)
# addWorksheet(wb = wb, sheetName = test.n)
# writeData(wb = wb, sheet = test.n, x = my.anova2)
# 
# ## glm
# test.n <- "glm"
# fm3 <- glm(disadvg ~ ethnicty*grade, data = tli, family = binomial())
# addWorksheet(wb = wb, sheetName = test.n)
# writeData(wb = wb, sheet = test.n, x = fm3)
# 
# ## prcomp
# test.n <- "prcomp"
# pr1 <- prcomp(USArrests)
# addWorksheet(wb = wb, sheetName = test.n)
# writeData(wb = wb, sheet = test.n, x = pr1)
# 
# ## summary.prcomp
# test.n <- "summary.prcomp"
# addWorksheet(wb = wb, sheetName = test.n)
# writeData(wb = wb, sheet = test.n, x = summary(pr1))
# 
# ## simple table
# test.n <- "table"
# data(airquality)
# airquality$OzoneG80 <- factor(airquality$Ozone > 80,
# levels = c(FALSE, TRUE),
# labels = c("Oz <= 80", "Oz > 80"))
# airquality$Month <- factor(airquality$Month,
# levels = 5:9,
# labels = month.abb[5:9])
# my.table <- with(airquality, table(OzoneG80,Month) )
# addWorksheet(wb = wb, sheetName = test.n)
# writeData(wb = wb, sheet = test.n, x = my.table)
# 
# ## survdiff 1
# library(survival)
# test.n <- "survdiff1"
# addWorksheet(wb = wb, sheetName = test.n)
# x <- survdiff(Surv(futime, fustat) ~ rx, data = ovarian)
# writeData(wb = wb, sheet = test.n, x = x)
# 
# ## survdiff 2
# test.n <- "survdiff2"
# addWorksheet(wb = wb, sheetName = test.n)
# expect <- survexp(futime ~ ratetable(age=(accept.dt - birth.dt),
#       sex=1,year=accept.dt,race="white"), jasa, cohort=FALSE,
#       ratetable=survexp.usr)
# x <- survdiff(Surv(jasa$futime, jasa$fustat) ~ offset(expect))
# writeData(wb = wb, sheet = test.n, x = x)
# 
# ## coxph 1
# test.n <- "coxph1"
# addWorksheet(wb = wb, sheetName = test.n)
# bladder$rx <- factor(bladder$rx, labels = c("Pla","Thi"))
# x <- coxph(Surv(stop,event) ~ rx, data = bladder)
# writeData(wb = wb, sheet = test.n, x = x)
# 
# ## coxph 2
# test.n <- "coxph2"
# addWorksheet(wb = wb, sheetName = test.n)
# x <- coxph(Surv(stop,event) ~ rx + cluster(id), data = bladder)
# writeData(wb = wb, sheet = test.n, x = x)
# 
# ## cox.zph
# test.n <- "cox.zph"
# addWorksheet(wb = wb, sheetName = test.n)
# x <- cox.zph(coxph(Surv(futime, fustat) ~ age + ecog.ps,  data=ovarian))
# writeData(wb = wb, sheet = test.n, x = x)
# 
# ## summary.coxph 1
# test.n <- "summary.coxph1"
# addWorksheet(wb = wb, sheetName = test.n)
# x <- summary(coxph(Surv(stop,event) ~ rx, data = bladder))
# writeData(wb = wb, sheet = test.n, x = x)
# 
# ## summary.coxph 2
# test.n <- "summary.coxph2"
# addWorksheet(wb = wb, sheetName = test.n)
# x <- summary(coxph(Surv(stop,event) ~ rx + cluster(id), data = bladder))
# writeData(wb = wb, sheet = test.n, x = x)
# 
# ## view without saving
# openXL(wb)
# 

## ----eval = FALSE, include = TRUE, tidy = TRUE, highlight= TRUE---------------
# require(ggplot2)
# 
# wb <- createWorkbook()
# 
# ## read historical prices from yahoo finance
# ticker <- "CBA.AX"
# csv.url <- paste0("https://query1.finance.yahoo.com/v7/finance/download/",
# ticker, "?period1=1597597610&period2=1629133610&interval=1d&events=history&includeAdjustedClose= TRue")
# prices <- read.csv(url(csv.url), as.is = TRUE)
# prices$Date <- as.Date(prices$Date)
# close <- prices$Close
# prices$logReturns = c(0, log(close[2:length(close)]/close[1:(length(close)-1)]))
# 
# ## Create plot of price series and add to worksheet
# ggplot(data = prices, aes(as.Date(Date), as.numeric(Close))) +
# geom_line(colour="royalblue2") +
# labs(x = "Date", y = "Price", title = ticker) +
# geom_area(fill = "royalblue1",alpha = 0.3) +
# coord_cartesian(ylim=c(min(prices$Close)-1.5, max(prices$Close)+1.5))
# 
# ## Add worksheet and write plot to sheet
# addWorksheet(wb, sheetName = "CBA")
# insertPlot(wb, sheet = 1, xy = c("J", 3))
# 
# ## Histogram of log returns
# ggplot(data = prices, aes(x = logReturns)) + geom_histogram(binwidth=0.0025) +
# labs(title = "Histogram of log returns")
# 
# ## currency
# class(prices$Close) <- "currency" ## styles as currency in workbook
# 
# ## write historical data and  histogram of returns
# writeDataTable(wb, sheet = "CBA", x = prices)
# insertPlot(wb, sheet = 1, startRow=25, startCol = "J")
# 
# ## Add conditional formatting to show where logReturn > 0.01 using default style
# conditionalFormat(wb, sheet = 1, cols = seq_len((prices)), rows = 2:(nrow(prices)+1),
# rule = "$H2 > 0.01")
# 
# ## style log return col as a percentage
# logRetStyle <- createStyle(numFmt = "percentage")
# 
# addStyle(wb, 1, style = logRetStyle, rows = 2:(nrow(prices) + 1),
# cols = "H", gridExpand = TRUE)
# 
# setColWidths(wb, sheet=1, cols = c("A", "F", "G", "H"), widths = 15)
# 
# ## save workbook to working directory
# saveWorkbook(wb, "stockPrice.xlsx", overwrite = TRUE)
# openXL("stockPrice.xlsx")

## ----eval = FALSE, include = TRUE---------------------------------------------
# require(openxlsx)
# require(jpeg)
# require(ggplot2)
# 
# plotFn <- function(x, ...) {
#   colvec <- grey(x)
#   colmat <- array(match(colvec, unique(colvec)), dim = dim(x)[1:2])
#   image(x = 0:(dim(colmat)[2]), y = 0:(dim(colmat)[1]), z = t(colmat[rev(seq_len(nrow(colmat))) , ]),
#     col = unique(colvec), xlab = "", ylab = "", axes = FALSE, asp = 1,
#     bty ="n", frame.plot=FALSE, ann=FALSE)
# }
# 
# ## Create workbook and add a worksheet, hide gridlines
# wb <- createWorkbook("Einstein")
# addWorksheet(wb, "Original Image", gridLines = FALSE)
# 
# A <- readJPEG(file.path(path.package("openxlsx"), "einstein.jpg"))
# height <- nrow(A)
# width <- ncol(A)
# 
# ## write "Original Image" to cell B2
# writeData(wb, 1, "Original Image", xy = c(2,2))
# 
# ## write Object size to cell B3
# writeData(wb, 1, sprintf("Image object size: %s bytes",
#                          format(object.size(A+0)[[1]], big.mark=',')),
#           xy = c(2,3))  ## equivalent to startCol = 2, startRow = 3
# 
# ## Plot image
# par(mar=rep(0, 4), xpd = NA)
# plotFn(A)
# 
# ## insert plot currently showing in plot window
# insertPlot(wb, 1, width, height, units="px", startRow= 5, startCol = 2)
# 
# ## SVD of covariance matrix
# rMeans <- rowMeans(A)
# rowMeans <- do.call("cbind", lapply(seq_len(ncol(A)), function(X) rMeans))
# A <- A - rowMeans
# E <- svd(A %*% t(A) / (ncol(A) - 1)) # SVD on covariance matrix of A
# pve <- data.frame("Eigenvalues" = E$d,
#                   "PVE" = E$d/sum(E$d),
#                   "Cumulative PVE" = cumsum(E$d/sum(E$d)))
# 
# ## write eigenvalues to worksheet
# addWorksheet(wb, "Principal Component Analysis")
# hs <- createStyle(fontColour = "#ffffff", fgFill = "#4F80BD",
#                   halign = "CENTER", textDecoration = "Bold",
#                   border = "TopBottomLeftRight", borderColour = "#4F81BD")
# 
# writeData(wb, 2, x="Proportions of variance explained by Eigenvector" ,startRow = 2)
# mergeCells(wb, sheet=2, cols=1:4, rows=2)
# 
# setColWidths(wb, 2, cols = 1:3, widths = c(14, 12, 15))
# writeData(wb, 2, x=pve, startRow = 3, startCol = 1, borders="rows", headerStyle=hs)
# 
# ## Plots
# pve <- cbind(pve, "Ind" = seq_len(nrow(pve)))
# ggplot(data = pve[1:20,], aes(x = Ind, y = 100*PVE)) +
#   geom_bar(stat="identity", position = "dodge") +
#   xlab("Principal Component Index") + ylab("Proportion of Variance Explained") +
#   geom_line(size = 1, col = "blue") + geom_point(size = 3, col = "blue")
# 
# ## Write plot to worksheet 2
# insertPlot(wb, 2, width = 5, height = 4, startCol = "E", startRow = 2)
# 
# ## Plot of cumulative explained variance
# ggplot(data = pve[1:50,], aes(x = Ind, y = 100*Cumulative.PVE)) +
#   geom_point(size=2.5) + geom_line(size=1) + xlab("Number of PCs") +
#   ylab("Cumulative Proportion of Variance Explained")
# insertPlot(wb, 2, width = 5, height = 4, xy= c("M", 2))
# 
# 
# ## Reconstruct image using increasing number of PCs
# nPCs <- c(5, 7, 12, 20, 50, 200)
# startRow <- rep(c(2, 24), each = 3)
# startCol <- rep(c("B", "H", "N"), 2)
# 
# ## create a worksheet to save reconstructed images to
# addWorksheet(wb, "Reconstructed Images", zoom = 90)
# 
# for(i in seq_len(length(nPCs))) {
# 
#   V <- E$v[, 1:nPCs[i]]
#   imgHat <- t(V) %*% A  ## project img data on to PCs
#   imgSize <- object.size(V) + object.size(imgHat) + object.size(rMeans)
# 
#   imgHat <- V %*% imgHat + rowMeans  ## reconstruct from PCs and add back row means
#   imgHat <- round((imgHat - min(imgHat)) / (max(imgHat) - min(imgHat))*255) # scale
#   plotFn(imgHat/255)
# 
#   ## write strings to worksheet 3
#   writeData(wb, "Reconstructed Images",
#             sprintf("Number of principal components used:  %s",
#                     nPCs[[i]]), startCol[i], startRow[i])
# 
#   writeData(wb, "Reconstructed Images",
#             sprintf("Sum of component object sizes: %s bytes",
#                     format(as.numeric(imgSize), big.mark=',')), startCol[i], startRow[i]+1)
# 
#   ## write reconstruced image
#   insertPlot(wb, "Reconstructed Images", width, height, units="px",
#              xy = c(startCol[i], startRow[i]+3))
# 
# }
# 
# # hide grid lines
# showGridLines(wb, sheet = 3, showGridLines = FALSE)
# 
# ## Make text above images BOLD
# boldStyle <- createStyle(textDecoration="BOLD")
# 
# ## only want to apply style to specified cells (not all combinations of rows & cols)
# addStyle(wb, "Reconstructed Images", style=boldStyle,
#          rows = c(startRow, startRow+1), cols = rep(startCol, 2),
#          gridExpand = FALSE)
# 
# ## save workbook to working directory
# saveWorkbook(wb, "Image dimensionality reduction.xlsx", overwrite = TRUE)
# 
# 
# 
# 
# ## remove example files for cran test
# if (identical(Sys.getenv("NOT_CRAN", unset = "true"), "false")) {
#   file_list<-list.files(pattern="\\.xlsx",recursive = TRUE)
#   file_list<-fl[!grepl("inst/extdata",file_list)&!grepl("man/",file_list)]
# 
#   if(length(file_list)>0) {
#     rm(file_list)
#   }
# }

## ----cleanup, eval = FALSE, include = FALSE-----------------------------------
# xlsx_files <- dir(pattern = "*.xlsx")
# unlink(xlsx_files)

##

#' @noRd
openxlsxCoerce <- function(x, rowNames) {
  UseMethod("openxlsxCoerce")
}

#' @noRd
openxlsxCoerce.default <- function(x, rowNames) {
  x <- as.data.frame(x, stringsAsFactors = FALSE)
  return(x)
}

#' @noRd
openxlsxCoerce.data.frame <- function(x, rowNames) {
  
  ## cbind rownames to x
  if (rowNames) {
    x <- cbind(data.frame("row names" = rownames(x), stringsAsFactors = FALSE), as.data.frame(x, stringsAsFactors = FALSE))
    names(x)[[1]] <- ""
  }
  
  return(x)
}

#' @noRd
openxlsxCoerce.data.table <- function(x, rowNames) {
  x <- as.data.frame(x, stringsAsFactors = FALSE)
  
  ## cbind rownames to x
  if (rowNames) {
    x <- cbind(data.frame("row names" = rownames(x), stringsAsFactors = FALSE), x)
    names(x)[[1]] <- ""
  }
  
  return(x)
}

#' @noRd
openxlsxCoerce.matrix <- function(x, rowNames) {
  x <- as.data.frame(x, stringsAsFactors = FALSE)
  
  if (rowNames) {
    x <- cbind(data.frame("row names" = rownames(x), stringsAsFactors = FALSE), x)
    names(x)[[1]] <- ""
  }
  
  return(x)
}

#' @noRd
openxlsxCoerce.array <- function(x, rowNames) {
  stop("array in writeData : currently not supported")
}

#' @noRd
openxlsxCoerce.aov <- function(x, rowNames) {
  x <- summary(x)
  x <- cbind(x[[1]])
  x <- cbind(data.frame("row name" = rownames(x), stringsAsFactors = FALSE), x)
  names(x)[1] <- ""
  
  return(x)
}

#' @noRd
openxlsxCoerce.lm <- function(x, rowNames) {
  x <- as.data.frame(summary(x)[["coefficients"]])
  x <- cbind(data.frame("Variable" = rownames(x), stringsAsFactors = FALSE), x)
  names(x)[1] <- ""
  
  return(x)
}

#' @noRd
openxlsxCoerce.anova <- function(x, rowNames) {
  x <- as.data.frame(x)
  
  if (rowNames) {
    x <- cbind(data.frame("row name" = rownames(x), stringsAsFactors = FALSE), x)
    names(x)[1] <- ""
  }
  
  return(x)
}

#' @noRd
openxlsxCoerce.glm <- function(x, rowNames) {
  x <- as.data.frame(summary(x)[["coefficients"]])
  x <- cbind(data.frame("row name" = rownames(x), stringsAsFactors = FALSE), x)
  names(x)[1] <- ""
  
  return(x)
}

#' @noRd
openxlsxCoerce.table <- function(x, rowNames) {
  x <- as.data.frame(unclass(x))
  x <- cbind(data.frame("Variable" = rownames(x), stringsAsFactors = FALSE), x)
  names(x)[1] <- ""
  
  return(x)
}

#' @noRd
openxlsxCoerce.prcomp <- function(x, rowNames) {
  x <- as.data.frame(x$rotation)
  x <- cbind(data.frame("Variable" = rownames(x), stringsAsFactors = FALSE), x)
  names(x)[1] <- ""
  
  return(x)
}

#' @noRd
openxlsxCoerce.summary.prcomp <- function(x, rowNames) {
  x <- as.data.frame(x$importance)
  x <- cbind(data.frame("Variable" = rownames(x), stringsAsFactors = FALSE), x)
  names(x)[1] <- ""
  
  return(x)
}



#' @name openxlsxCoerce.survdiff
#' @description like print.survdiff with some ideas from the ascii package
#' @param x data.frame for input
#' @param rowNames rownames
#' @importFrom stats pchisq
#' @keywords internal
#' @noRd
openxlsxCoerce.survdiff <- function(x, rowNames) {
  
  
  ## like print.survdiff with some ideas from the ascii package
  if (length(x$n) == 1) {
    z <- sign(x$exp - x$obs) * sqrt(x$chisq)
    temp <- c(x$obs, x$exp, z, 1 - pchisq(x$chisq, 1))
    names(temp) <- c("Observed", "Expected", "Z", "p")
    x <- as.data.frame(t(temp))
  } else {
    if (is.matrix(x$obs)) {
      otmp <- apply(x$obs, 1, sum)
      etmp <- apply(x$exp, 1, sum)
    }
    else {
      otmp <- x$obs
      etmp <- x$exp
    }
    chisq <- c(x$chisq, rep(NA, length(x$n) - 1))
    df <- c((sum(1 * (etmp > 0))) - 1, rep(NA, length(x$n) - 1))
    p <- c(1 - pchisq(x$chisq, df[!is.na(df)]), rep(NA, length(x$n) - 1))
    
    temp <- cbind(
      x$n, otmp, etmp,
      ((otmp - etmp)^2) / etmp, ((otmp - etmp)^2) / diag(x$var),
      chisq, df, p
    )
    
    
    colnames(temp) <- c(
      "N", "Observed", "Expected", "(O-E)^2/E", "(O-E)^2/V",
      "Chisq", "df", "p"
    )
    
    temp <- as.data.frame(temp, checknames = FALSE)
    x <- cbind("Group" = names(x$n), temp)
    names(x)[1] <- ""
  }
  
  return(x)
}


#' @noRd
openxlsxCoerce.coxph <- function(x, rowNames) {
  
  ## sligthly modified print.coxph
  coef <- x$coefficients
  se <- sqrt(diag(x$var))
  
  if (is.null(coef) || is.null(se)) {
    stop("Input is not valid")
  }
  
  if (is.null(x$naive.var)) {
    tmp <- cbind(coef, exp(coef), se, coef / se, pchisq((coef / se)^2, 1))
    colnames(tmp) <- c("coef", "exp(coef)", "se(coef)", "z", "p")
  } else {
    nse <- sqrt(diag(x$naive.var))
    tmp <- cbind(coef, exp(coef), nse, se, coef / se, pchisq((coef / se)^2, 1))
    colnames(tmp) <- c("coef", "exp(coef)", "se(coef)", "robust se", "z", "p")
  }
  
  x <- cbind("Variable" = names(coef), as.data.frame(tmp, checknames = FALSE))
  names(x)[1] <- ""
  
  return(x)
}



#' @noRd
openxlsxCoerce.summary.coxph <- function(x, rowNames) {
  coef <- x$coefficients
  ci <- x$conf.int
  # nvars <- nrow(coef)  variable not used
  
  tmp <- cbind(
    coef[, -ncol(coef), drop = FALSE], # p later
    ci[, (ncol(ci) - 1):ncol(ci), drop = FALSE], # confint
    coef[, ncol(coef), drop = FALSE]
  ) # p.value
  
  x <- as.data.frame(tmp, checknames = FALSE)
  
  x <- cbind(data.frame("row names" = rownames(x)), x)
  names(x)[[1]] <- ""
  
  return(x)
}

#' @noRd
openxlsxCoerce.cox.zph <- function(x, rowNames) {
  tmp <- as.data.frame(x$table)
  x <- cbind(data.frame("row names" = rownames(tmp)), tmp)
  names(x)[[1]] <- ""
  
  return(x)
}

#' @noRd
openxlsxCoerce.hyperlink <- function(x, rowNames) {
  ## vector of hyperlinks
  class(x) <- c("character", "hyperlink")
  x <- as.data.frame(x, stringsAsFactors = FALSE)
}

# Workbook ----------------------------------------------------------------

Workbook <- setRefClass("Workbook",
  fields = c(
    "sheet_names" = "character",

    "charts" = "ANY",
    "isChartSheet" = "logical",

    "colOutlineLevels" = "ANY",
    "colWidths" = "ANY",
    "connections" = "ANY",
    "Content_Types" = "character",
    "core" = "character",
    "drawings" = "ANY",
    "drawings_rels" = "ANY",
    "embeddings" = "ANY",
    "externalLinks" = "ANY",
    "externalLinksRels" = "ANY",

    "headFoot" = "ANY",
    "media" = "ANY",
    "outlineLevels" = "ANY",
    
    "persons" = "ANY",

    "pivotTables" = "ANY",
    "pivotTables.xml.rels" = "ANY",
    "pivotDefinitions" = "ANY",
    "pivotRecords" = "ANY",
    "pivotDefinitionsRels" = "ANY",

    "queryTables" = "ANY",
    "rowHeights" = "ANY",

    "slicers" = "ANY",
    "slicerCaches" = "ANY",

    "sharedStrings" = "ANY",
    "styleObjects" = "ANY",

    "styles" = "ANY",
    "tables" = "ANY",
    "tables.xml.rels" = "ANY",
    "theme" = "ANY",

    "vbaProject" = "ANY",
    "vml" = "ANY",
    "vml_rels" = "ANY",
    "comments" = "ANY",
    "threadComments" = "ANY",

    "workbook" = "ANY",
    "workbook.xml.rels" = "ANY",
    "worksheets" = "ANY",
    "worksheets_rels" = "ANY",
    "sheetOrder" = "integer",
    "ActiveSheet" = "integer"
  )
)

# Style -------------------------------------------------------------------

Style <- setRefClass("Style",
  fields = c(
    "fontName",
    "fontColour",
    "fontSize",
    "fontFamily",
    "fontScheme",
    "fontDecoration",
    "borderTop",
    "borderLeft",
    "borderRight",
    "borderBottom",
    "borderTopColour",
    "borderLeftColour",
    "borderRightColour",
    "borderBottomColour",
    "borderDiagonal",
    "borderDiagonalColour",
    "borderDiagonalUp",
    "borderDiagonalDown",
    "halign",
    "valign",
    "indent",
    "textRotation",
    "numFmt",
    "fill",
    "wrapText",
    "locked",
    "hidden",
    "xfId"
  ),
  methods = list()
)

# Sheet_Data --------------------------------------------------------------

Sheet_Data <- setRefClass("Sheet_Data",
  fields = c(
    "rows" = "integer",
    "cols" = "integer",
    "t" = "integer",
    "v" = "character",
    "f" = "character",
    "style_id" = "ANY",
    "data_count" = "integer",
    "n_elements" = "integer"
  )
)


# Worksheet ---------------------------------------------------------------

WorkSheet <- setRefClass("WorkSheet",
  fields = c(
    "sheetPr" = "character",
    "dimension" = "character",
    "sheetViews" = "character",
    "sheetFormatPr" = "character",
    "cols" = "character",

    "sheet_data" = "Sheet_Data",

    "autoFilter" = "character",
    "mergeCells" = "ANY",
    "conditionalFormatting" = "character",
    "dataValidations" = "ANY",
    "dataValidationsLst" = "character",

    "freezePane" = "character",
    "hyperlinks" = "ANY",

    "sheetProtection" = "character",
    "pageMargins" = "character",
    "pageSetup" = "character",
    "headerFooter" = "ANY",
    "rowBreaks" = "character",
    "colBreaks" = "character",
    "drawing" = "character",
    "legacyDrawing" = "character",
    "legacyDrawingHF" = "character",
    "oleObjects" = "character",
    "tableParts" = "character",
    "extLst" = "character"
  )
)

# ChartSheet --------------------------------------------------------------

ChartSheet <- setRefClass("ChartSheet",
  fields = c(
    "sheetPr" = "character",
    "sheetViews" = "character",
    "pageMargins" = "character",
    "drawing" = "character",
    "hyperlinks" = "ANY"
  )
)

#' @include class_definitions.R

Sheet_Data$methods(initialize = function() {
  rows <<- integer(0)
  cols <<- integer(0)

  t <<- integer(0)
  v <<- character(0)
  f <<- character(0)

  style_id <<- character(0)

  data_count <<- 0L
  n_elements <<- 0L
})



Sheet_Data$methods(delete = function(rows_in, cols_in, grid_expand) {
  cols_in <- convertFromExcelRef(cols_in)
  rows_in <- as.integer(rows_in)

  ## rows and cols need to be the same length
  if (grid_expand) {
    n <- length(rows_in)
    rows_in <- rep.int(rows_in, times = length(cols_in))
    cols_in <- rep(cols_in, each = n)
  }

  if (length(rows_in) != length(cols_in)) {
    stop("Length of rows and cols must be equal.")
  }

  inds <- which(pair_rc(rows, cols) %in% pair_rc(rows_in, cols_in))

  if (length(inds) > 0) { ## writing over existing data

    rows <<- rows[-inds]
    cols <<- cols[-inds]
    t <<- t[-inds]
    v <<- v[-inds]
    f <<- f[-inds]

    n_elements <<- as.integer(length(rows))

    if (n_elements == 0) {
      data_count <<- 0L
    }
  }
})



Sheet_Data$methods(write = function(rows_in, cols_in, t_in, v_in, f_in, any_functions = TRUE) {
  if (length(rows_in) == 0 | length(cols_in) == 0) {
    return(invisible(0))
  }


  possible_overlap <- FALSE
  if (n_elements > 0) {
    possible_overlap <- (min(cols_in, na.rm = TRUE) <= max(cols, na.rm = TRUE)) &
      (max(cols_in, na.rm = TRUE) >= min(cols, na.rm = TRUE)) &
      (min(rows_in, na.rm = TRUE) <= max(rows, na.rm = TRUE)) &
      (max(rows_in, na.rm = TRUE) >= min(rows, na.rm = TRUE))
  }

  n <- length(cols_in)
  cols_in <- rep.int(cols_in, times = length(rows_in))
  rows_in <- rep(rows_in, each = n)

  if (any_functions) {
    if (any(!is.na(f_in))) {
      v_in[!is.na(f_in)] <- as.character(NA)
      t_in[!is.na(f_in)] <- 3L ## "str"
    }
  }

  inds <- integer(0)
  if (possible_overlap) {
    inds <- which(pair_rc(rows, cols) %in% pair_rc(rows_in, cols_in))
  }

  if (length(inds) > 0) {
    rows <<- c(rows[-inds], rows_in)
    cols <<- c(cols[-inds], cols_in)
    t <<- c(t[-inds], t_in)
    v <<- c(v[-inds], v_in)
    f <<- c(f[-inds], f_in)
  } else {
    rows <<- c(rows, rows_in)
    cols <<- c(cols, cols_in)
    t <<- c(t, t_in)
    v <<- c(v, v_in)
    f <<- c(f, f_in)
  }

  n_elements <<- as.integer(length(rows))
  data_count <<- data_count + 1L
})

#' @include class_definitions.R
Workbook$methods(writeData = function(
  df,
  sheet,
  startRow,
  startCol,
  colNames,
  colClasses,
  hlinkNames,
  keepNA, 
  na.string,
  list_sep
) {
  sheet <- validateSheet(sheet)
  nCols <- ncol(df)
  nRows <- nrow(df)
  df_nms <- names(df)

  allColClasses <- unlist(colClasses)
  
  isPOSIXlt <- function(data) sapply(lapply(data, class), FUN = function(x) any(x == "POSIXlt"))
  to_convert <- isPOSIXlt(df)
  
  if (any(to_convert)) {
    message("Found POSIXlt. Converting to POSIXct")
    df[to_convert] <- lapply(df[to_convert], as.POSIXct)  
  }
  
  
  df <- as.list(df)

  ######################################################################
  ## standardise all column types

  ## pull out NaN values
  nans <- unlist(lapply(1:nCols, function(i) {
    tmp <- df[[i]]
    if (!"character" %in% class(tmp) & !"list" %in% class(tmp)) {
      v <- which(is.nan(tmp) | is.infinite(tmp))
      if (length(v) == 0) {
        return(v)
      }
      return(as.integer(nCols * (v - 1) + i)) ## row position
    }
  }))

  ## convert any Dates to integers and create date style object
  if (any(c("date", "posixct", "posixt") %in% allColClasses)) {
    dInds <- which(sapply(colClasses, function(x) "date" %in% x))

    origin <- 25569L
    if (grepl('date1904="1"|date1904="true"', stri_join(unlist(workbook), collapse = ""), ignore.case = TRUE)) {
      origin <- 24107L
    }
    
    for (i in dInds) {
      df[[i]] <- as.integer(df[[i]]) + origin
      if (origin == 25569L) {
        earlyDate <- which(df[[i]] < 60)
        df[[i]][earlyDate] <- df[[i]][earlyDate] - 1
      }
    }

    pInds <- which(sapply(colClasses, function(x) any(c("posixct", "posixt", "posixlt") %in% x)))
    if (length(pInds) > 0 & nRows > 0) {
      parseOffset <- function(tz) {
        suppressWarnings(
          ifelse(stri_sub(tz, 1, 1) == "+", 1L, -1L)
          * (as.integer(stri_sub(tz, 2, 3)) + as.integer(stri_sub(tz, 4, 5)) / 60) / 24
        )
      }

      t <- lapply(df[pInds], function(x) format(x, "%z"))
      offSet <- lapply(t, parseOffset)
      offSet <- lapply(offSet, function(x) ifelse(is.na(x), 0, x))

      for (i in seq_along(pInds)) {
        df[[pInds[i]]] <- as.numeric(as.POSIXct(df[[pInds[i]]])) / 86400 + origin + offSet[[i]]
      }
    }
  }

  ## convert any Dates to integers and create date style object
  if (any(c("currency", "accounting", "percentage", "3", "comma") %in% allColClasses)) {
    cInds <- which(sapply(colClasses, function(x) any(c("accounting", "currency", "percentage", "3", "comma") %in% tolower(x))))
    for (i in cInds) {
      df[[i]] <- as.numeric(gsub("[^0-9\\.-]", "", df[[i]], perl = TRUE))
    }
    class(df[[i]]) <- "numeric"
  }

  ## convert scientific
  if ("scientific" %in% allColClasses) {
    for (i in which(sapply(colClasses, function(x) "scientific" %in% x))) {
      class(df[[i]]) <- "numeric"
    }
  }

  ##
  if ("list" %in% allColClasses) {
    for (i in which(sapply(colClasses, function(x) "list" %in% x))) {
      # check for and replace NA
      df_i <- lapply(df[[i]], unlist)
      df_i <- lapply(df_i, function(x) {
        x[is.na(x)] <- na.string
        x
      })
      df[[i]] <- sapply(df_i, stri_join, collapse = list_sep)
    }
  }

  if ("hyperlink" %in% allColClasses) {
    for (i in which(sapply(colClasses, function(x) "hyperlink" %in% x))) {
      class(df[[i]]) <- "hyperlink"
    }
  }

  if (any(c("formula", "array_formula") %in% allColClasses)) {
    
    frm <- "formula"
    cls <- "openxlsx_formula"
    
    if ("array_formula" %in% allColClasses) {
      frm <- "array_formula"
      cls <- "openxlsx_array_formula"
    }
    
    for (i in which(sapply(colClasses, function(x) frm %in% x))) {
      df[[i]] <- replaceIllegalCharacters(as.character(df[[i]]))
      class(df[[i]]) <- cls
    }
  }

  colClasses <- sapply(df, function(x) tolower(class(x))[[1]]) ## by here all cols must have a single class only


  ## convert logicals (Excel stores logicals as 0 & 1)
  if ("logical" %in% allColClasses) {
    for (i in which(sapply(colClasses, function(x) "logical" %in% x))) {
      class(df[[i]]) <- "numeric"
    }
  }

  ## convert all numerics to character (this way preserves digits)
  if ("numeric" %in% colClasses) {
    for (i in which(sapply(colClasses, function(x) "numeric" %in% x))) {
      class(df[[i]]) <- "character"
    }
  }

  ## End standardise all column types
  ######################################################################


  ## cell types
  t <- build_cell_types_integer(classes = colClasses, n_rows = nRows)

  for (i in which(sapply(colClasses, function(x) !"character" %in% x & !"numeric" %in% x))) {
    df[[i]] <- as.character(df[[i]])
  }

  ## cell values
  v <- as.character(t(as.matrix(
    data.frame(df, stringsAsFactors = FALSE, check.names = FALSE, fix.empty.names = FALSE)
  )))
  
  
  if (keepNA) {
    if (is.null(na.string)) {
      t[is.na(v)] <- 4L
      v[is.na(v)] <- "#N/A"
    } else {
      t[is.na(v)] <- 1L
      v[is.na(v)] <- as.character(na.string)
    }
  } else {
    t[is.na(v)] <- as.integer(NA)
    v[is.na(v)] <- as.character(NA)
  }

  ## If any NaN values
  if (length(nans) > 0) {
    t[nans] <- 4L
    v[nans] <- "#NUM!"
  }


  # prepend column headers
  if (colNames) {
    t <- c(rep.int(1L, nCols), t)
    v <- c(df_nms, v)
    nRows <- nRows + 1L
  }


  ## Formulas
  f_in <- rep.int(as.character(NA), length(t))
  any_functions <- FALSE
  ref_cell <- paste0(int_2_cell_ref(startCol), startRow)

  if (any(c("openxlsx_formula", "openxlsx_array_formula") %in% colClasses)) {
    
    ## alter the elements of t where we have a formula to be "str"
    if ("openxlsx_formula" %in% colClasses) {
      formula_cols <- which(sapply(colClasses, function(x) "openxlsx_formula" %in% x, USE.NAMES = FALSE), useNames = FALSE)
      formula_strs <- stri_join("<f>", unlist(df[formula_cols], use.names = FALSE), "</f>")
    } else { # openxlsx_array_formula
      formula_cols <- which(sapply(colClasses, function(x) "openxlsx_array_formula" %in% x, USE.NAMES = FALSE), useNames = FALSE)
      formula_strs <- stri_join("<f t=\"array\" ref=\"", ref_cell, ":", ref_cell, "\">", unlist(df[formula_cols], use.names = FALSE), "</f>")
    }
    formula_inds <- unlist(lapply(formula_cols, function(i) i + (1:(nRows - colNames) - 1) * nCols + (colNames * nCols)), use.names = FALSE)
    f_in[formula_inds] <- formula_strs
    any_functions <- TRUE

    rm(formula_cols)
    rm(formula_strs)
    rm(formula_inds)
  }

  suppressWarnings(try(rm(df), silent = TRUE))

  ## Append hyperlinks, convert h to s in cell type
  hyperlink_cols <- which(sapply(colClasses, function(x) "hyperlink" %in% x, USE.NAMES = FALSE), useNames = FALSE)
  if (length(hyperlink_cols) > 0) {
    hyperlink_inds <- sort(unlist(lapply(hyperlink_cols, function(i) i + (1:(nRows - colNames) - 1) * nCols + (colNames * nCols)), use.names = FALSE))
    na_hyperlink <- intersect(hyperlink_inds, which(is.na(t)))

    if (length(hyperlink_inds) > 0) {
      t[t %in% 9] <- 1L ## set cell type to "s"

      hyperlink_refs <- convert_to_excel_ref_expand(cols = hyperlink_cols + startCol - 1, LETTERS = LETTERS, rows = as.character((startRow + colNames):(startRow + nRows - 1L)))

      if (length(na_hyperlink) > 0) {
        to_remove <- which(hyperlink_inds %in% na_hyperlink)
        hyperlink_refs <- hyperlink_refs[-to_remove]
        hyperlink_inds <- hyperlink_inds[-to_remove]
      }

      exHlinks <- worksheets[[sheet]]$hyperlinks
      targets <- replaceIllegalCharacters(v[hyperlink_inds])

      if (!is.null(hlinkNames) & length(hlinkNames) == length(hyperlink_inds)) {
        v[hyperlink_inds] <- hlinkNames
      } ## this is text to display instead of hyperlink

      ## create hyperlink objects
      newhl <- lapply(seq_along(hyperlink_inds), function(i) {
        Hyperlink$new(ref = hyperlink_refs[i], target = targets[i], location = NULL, display = NULL, is_external = TRUE)
      })

      worksheets[[sheet]]$hyperlinks <<- append(worksheets[[sheet]]$hyperlinks, newhl)
    }
  }


  ## convert all strings to references in sharedStrings and update values (v)
  strFlag <- which(t == 1L)
  newStrs <- v[strFlag]
  if (length(newStrs) > 0) {
    newStrs <- replaceIllegalCharacters(newStrs)
    vl <- stri_length(newStrs)
    
    for (i in which(vl > 32767)) {
      
      if (vl[i]>32768+30) {
        warning(
          paste0(
            stri_sub(newStrs[i], 32768, 32768 + 15),
            " ... " ,
            stri_sub(newStrs[i], vl[i] - 15, vl[i]),
            " is truncated. 
Number of characters exeed the limit of 32767."
          )
        )
      } else {
        warning(
          paste0(
            stri_sub(newStrs[i], 32768, -1),
            " is truncated. 
Number of characters exeed the limit of 32767."
          )
        )
        
      }
      
      # v[i] <- stri_sub(v[i], 1, 32767)
    }
    newStrs <- stri_join("<si><t xml:space=\"preserve\">", newStrs, "</t></si>")

    uNewStr <- unique(newStrs)

    .self$updateSharedStrings(uNewStr)
    v[strFlag] <- match(newStrs, sharedStrings) - 1L
  }

  # ## Create cell list of lists
  worksheets[[sheet]]$sheet_data$write(
    rows_in = startRow:(startRow + nRows - 1L),
    cols_in = startCol:(startCol + nCols - 1L),
    t_in = t,
    v_in = v,
    f_in = f_in,
    any_functions = any_functions
  )

  invisible(0)
})

#' @include class_definitions.R


ChartSheet$methods(initialize = function(tabSelected = FALSE,
                                         tabColour = character(0),
                                         zoom = 100) {
  if (length(tabColour) > 0) {
    tabColour <- sprintf("<sheetPr>%s</sheetPr>", tabColour)
  } else {
    tabColour <- character(0)
  }
  if (zoom < 10) {
    zoom <- 10
  } else if (zoom > 400) {
    zoom <- 400
  }

  sheetPr <<- tabColour
  sheetViews <<- sprintf('<sheetViews><sheetView workbookViewId="0" zoomScale="%s" tabSelected="%s"/></sheetViews>', as.integer(zoom), as.integer(tabSelected))
  pageMargins <<- '<pageMargins left="0.7" right="0.7" top="0.75" bottom="0.75" header="0.3" footer="0.3"/>'
  drawing <<- '<drawing r:id=\"rId1\"/>'
  hyperlinks <<- character(0)

  return(invisible(0))
})





ChartSheet$methods(get_prior_sheet_data = function() {
  xml <- '<chartsheet xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships">>'

  if (length(sheetPr) > 0) {
    xml <- paste(xml, sheetPr, collapse = "")
  }

  if (length(sheetViews) > 0) {
    xml <- paste(xml, sheetViews, collapse = "")
  }

  if (length(pageMargins) > 0) {
    xml <- paste(xml, pageMargins, collapse = "")
  }

  if (length(drawing) > 0) {
    xml <- paste(xml, drawing, collapse = "")
  }

  xml <- paste(xml, "</chartsheet>")

  return(xml)
})

#' @include class_definitions.R
#' @import stringi

Workbook$methods(
  initialize = function(creator = openxlsx_getOp("creator"),
                        title = NULL,
                        subject = NULL,
                        category = NULL) {
    charts <<- list()
    isChartSheet <<- logical(0)

    colWidths <<- list()
    colOutlineLevels <<- list()
    attr(colOutlineLevels, "hidden") <<- NULL
    connections <<- NULL
    Content_Types <<- genBaseContent_Type()
    core <<-
      genBaseCore(
        creator = creator,
        title = title,
        subject = subject,
        category = category
      )
    comments <<- list()
    threadComments <<- list()


    drawings <<- list()
    drawings_rels <<- list()

    embeddings <<- NULL
    externalLinks <<- NULL
    externalLinksRels <<- NULL

    headFoot <<- NULL

    media <<- list()

    persons <<- NULL

    pivotTables <<- NULL
    pivotTables.xml.rels <<- NULL
    pivotDefinitions <<- NULL
    pivotRecords <<- NULL
    pivotDefinitionsRels <<- NULL

    queryTables <<- NULL
    rowHeights <<- list()
    outlineLevels <<- list()
    attr(outlineLevels, "hidden") <<- NULL

    slicers <<- NULL
    slicerCaches <<- NULL

    sheet_names <<- character(0)
    sheetOrder <<- integer(0)

    sharedStrings <<- list()
    attr(sharedStrings, "uniqueCount") <<- 0

    styles <<- genBaseStyleSheet()
    styleObjects <<- list()


    tables <<- NULL
    tables.xml.rels <<- NULL
    theme <<- NULL


    vbaProject <<- NULL
    vml <<- list()
    vml_rels <<- list()

    workbook <<- genBaseWorkbook()
    workbook.xml.rels <<- genBaseWorkbook.xml.rels()

    worksheets <<- list()
    worksheets_rels <<- list()
    ActiveSheet <<- integer(0)
  }
)

Workbook$methods(
  addWorksheet = function(
    sheetName,
    showGridLines = openxlsx_getOp("showGridLines"),
    tabColour = openxlsx_getOp("tabColour"),
    zoom = 100,
    oddHeader = openxlsx_getOp("oddHeader"),
    oddFooter = openxlsx_getOp("oddFooter"),
    evenHeader = openxlsx_getOp("evenHeader"),
    evenFooter = openxlsx_getOp("evenFooter"),
    firstHeader = openxlsx_getOp("firstHeader"),
    firstFooter = openxlsx_getOp("firstFooter"),
    visible = TRUE,
    paperSize = openxlsx_getOp("paperSize", 9),
    orientation = openxlsx_getOp("orientation", "portrait"),
    hdpi = openxlsx_getOp("hdpi", 300),
    vdpi = openxlsx_getOp("vdpi", 300)
  ) {
    if (!missing(sheetName)) {
      if (grepl(pattern = ":", x = sheetName)) {
        stop("colon not allowed in sheet names in Excel")
      }
    }
    newSheetIndex <- length(worksheets) + 1L

    if (newSheetIndex > 1) {
      sheetId <-
        max(as.integer(regmatches(
          workbook$sheets,
          regexpr('(?<=sheetId=")[0-9]+', workbook$sheets, perl = TRUE)
        ))) + 1L
    } else {
      sheetId <- 1
      ActiveSheet <<- 1L
    }


    ## fix visible value
    visible <- tolower(visible)

    if (visible == "true") {
      visible <- "visible"
    } else if (visible == "false") {
      visible <- "hidden"
    } else if (visible == "veryhidden") {
      visible <- "veryHidden"
    }

    ##  Add sheet to workbook.xml
    workbook$sheets <<-
      c(
        workbook$sheets,
        sprintf(
          '<sheet name="%s" sheetId="%s" state="%s" r:id="rId%s"/>',
          sheetName,
          sheetId,
          visible,
          newSheetIndex
        )
      )

    ## append to worksheets list
    worksheets <<-
      append(
        worksheets,
        WorkSheet$new(
          showGridLines = showGridLines,
          tabSelected = newSheetIndex == 1,
          tabColour = tabColour,
          zoom = zoom,
          oddHeader = oddHeader,
          oddFooter = oddFooter,
          evenHeader = evenHeader,
          evenFooter = evenFooter,
          firstHeader = firstHeader,
          firstFooter = firstFooter,
          paperSize = paperSize,
          orientation = orientation,
          hdpi = hdpi,
          vdpi = vdpi
        )
      )


    ## update content_tyes
    ## add a drawing.xml for the worksheet
    Content_Types <<-
      c(
        Content_Types,
        sprintf(
          '<Override PartName="/xl/worksheets/sheet%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.worksheet+xml"/>',
          newSheetIndex
        ),
        sprintf(
          '<Override PartName="/xl/drawings/drawing%s.xml" ContentType="application/vnd.openxmlformats-officedocument.drawing+xml"/>',
          newSheetIndex
        )
      )

    ## Update xl/rels
    workbook.xml.rels <<- c(
      workbook.xml.rels,
      sprintf(
        '<Relationship Id="rId0" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/worksheet" Target="worksheets/sheet%s.xml"/>',
        newSheetIndex
      )
    )


    ## create sheet.rels to simplify id assignment
    worksheets_rels[[newSheetIndex]] <<-
      genBaseSheetRels(newSheetIndex)
    drawings_rels[[newSheetIndex]] <<- list()
    drawings[[newSheetIndex]] <<- list()

    vml_rels[[newSheetIndex]] <<- list()
    vml[[newSheetIndex]] <<- list()

    isChartSheet[[newSheetIndex]] <<- FALSE
    comments[[newSheetIndex]] <<- list()
    threadComments[[newSheetIndex]] <<- list()

    rowHeights[[newSheetIndex]] <<- list()
    colWidths[[newSheetIndex]] <<- list()
    colOutlineLevels[[newSheetIndex]] <<- vector("character")
    outlineLevels[[newSheetIndex]] <<- vector("character")

    sheetOrder <<- c(sheetOrder, as.integer(newSheetIndex))
    sheet_names <<- c(sheet_names, sheetName)

    invisible(newSheetIndex)
  }
)

Workbook$methods(
  cloneWorksheet = function(sheetName, clonedSheet) {
    clonedSheet <- validateSheet(clonedSheet)
    if (!missing(sheetName)) {
      if (grepl(pattern = ":", x = sheetName)) {
        stop("colon not allowed in sheet names in Excel")
      }
    }
    newSheetIndex <- length(worksheets) + 1L
    if (newSheetIndex > 1) {
      sheetId <-
        max(as.integer(regmatches(
          workbook$sheets,
          regexpr('(?<=sheetId=")[0-9]+', workbook$sheets, perl = TRUE)
        ))) + 1L
    } else {
      sheetId <- 1
    }


    ## copy visibility from cloned sheet!
    visible <-
      regmatches(
        workbook$sheets[[clonedSheet]],
        regexpr('(?<=state=")[^"]+', workbook$sheets[[clonedSheet]], perl = TRUE)
      )

    ##  Add sheet to workbook.xml
    workbook$sheets <<-
      c(
        workbook$sheets,
        sprintf(
          '<sheet name="%s" sheetId="%s" state="%s" r:id="rId%s"/>',
          sheetName,
          sheetId,
          visible,
          newSheetIndex
        )
      )

    ## append to worksheets list
    worksheets <<-
      append(worksheets, worksheets[[clonedSheet]]$copy())


    ## update content_tyes
    ## add a drawing.xml for the worksheet
    Content_Types <<-
      c(
        Content_Types,
        sprintf(
          '<Override PartName="/xl/worksheets/sheet%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.worksheet+xml"/>',
          newSheetIndex
        ),
        sprintf(
          '<Override PartName="/xl/drawings/drawing%s.xml" ContentType="application/vnd.openxmlformats-officedocument.drawing+xml"/>',
          newSheetIndex
        )
      )

    ## Update xl/rels
    workbook.xml.rels <<- c(
      workbook.xml.rels,
      sprintf(
        '<Relationship Id="rId0" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/worksheet" Target="worksheets/sheet%s.xml"/>',
        newSheetIndex
      )
    )

    ## create sheet.rels to simplify id assignment
    worksheets_rels[[newSheetIndex]] <<-
      genBaseSheetRels(newSheetIndex)
    drawings_rels[[newSheetIndex]] <<- drawings_rels[[clonedSheet]]

    # give each chart its own filename (images can re-use the same file, but charts can't)
    drawings_rels[[newSheetIndex]] <<-
      sapply(drawings_rels[[newSheetIndex]], function(rl) {
        chartfiles <-
          regmatches(
            rl,
            gregexpr("(?<=charts/)chart[0-9]+\\.xml", rl, perl = TRUE)
          )[[1]]
        for (cf in chartfiles) {
          chartid <- length(charts) + 1
          newname <- stri_join("chart", chartid, ".xml")
          fl <- charts[cf]

          # Read the chartfile and adjust all formulas to point to the new
          # sheet name instead of the clone source
          # The result is saved to a new chart xml file
          newfl <- file.path(dirname(fl), newname)
          charts[newname] <<- newfl
          chart <- readUTF8(fl)
          chart <-
            gsub(
              stri_join("(?<=')", sheet_names[[clonedSheet]], "(?='!)"),
              stri_join("'", sheetName, "'"),
              chart,
              perl = TRUE
            )
          chart <-
            gsub(
              stri_join("(?<=[^A-Za-z0-9])", sheet_names[[clonedSheet]], "(?=!)"),
              stri_join("'", sheetName, "'"),
              chart,
              perl = TRUE
            )
          writeLines(chart, newfl)
          # file.copy(fl, newfl)
          Content_Types <<-
            c(
              Content_Types,
              sprintf(
                '<Override PartName="/xl/charts/%s" ContentType="application/vnd.openxmlformats-officedocument.drawingml.chart+xml"/>',
                newname
              )
            )
          rl <- gsub(stri_join("(?<=charts/)", cf), newname, rl, perl = TRUE)
        }
        rl
      }, USE.NAMES = FALSE)
    # The IDs in the drawings array are sheet-specific, so within the new cloned sheet
    # the same IDs can be used => no need to modify drawings
    drawings[[newSheetIndex]] <<- drawings[[clonedSheet]]

    vml_rels[[newSheetIndex]] <<- vml_rels[[clonedSheet]]
    vml[[newSheetIndex]] <<- vml[[clonedSheet]]

    isChartSheet[[newSheetIndex]] <<- isChartSheet[[clonedSheet]]
    comments[[newSheetIndex]] <<- comments[[clonedSheet]]
    threadComments[[newSheetIndex]] <<- threadComments[[clonedSheet]]

    rowHeights[[newSheetIndex]] <<- rowHeights[[clonedSheet]]
    colWidths[[newSheetIndex]] <<- colWidths[[clonedSheet]]

    colOutlineLevels[[newSheetIndex]] <<- colOutlineLevels[[clonedSheet]]
    outlineLevels[[newSheetIndex]] <<- outlineLevels[[clonedSheet]]

    sheetOrder <<- c(sheetOrder, as.integer(newSheetIndex))
    sheet_names <<- c(sheet_names, sheetName)


    ############################
    ## STYLE
    ## ... objects are stored in a global list, so we need to get all styles
    ## assigned to the cloned sheet and duplicate them
    sheetStyles <- Filter(function(s) {
      s$sheet == sheet_names[[clonedSheet]]
    }, styleObjects)
    styleObjects <<- c(
      styleObjects,
      Map(function(s) {
        s$sheet <- sheetName
        s
      }, sheetStyles)
    )


    ############################
    ## TABLES
    ## ... are stored in the $tables list, with the name and sheet as attr
    ## and in the worksheets[]$tableParts list. We also need to adjust the
    ## worksheets_rels and set the content type for the new table

    tbls <- tables[attr(tables, "sheet") == clonedSheet]
    for (t in tbls) {
      # Extract table name, displayName and ID from the xml
      oldname <- regmatches(t, regexpr('(?<= name=")[^"]+', t, perl = TRUE))
      olddispname <- regmatches(t, regexpr('(?<= displayName=")[^"]+', t, perl = TRUE))
      oldid <- regmatches(t, regexpr('(?<= id=")[^"]+', t, perl = TRUE))
      ref <- regmatches(t, regexpr('(?<= ref=")[^"]+', t, perl = TRUE))

      # Find new, unused table names by appending _n, where n=1,2,...
      n <- 0
      while (stri_join(oldname, "_", n) %in% attr(tables, "tableName")) {
        n <- n + 1
      }
      newname <- stri_join(oldname, "_", n)
      newdispname <- stri_join(olddispname, "_", n)
      newid <- as.character(length(tables) + 3L)

      # Use the table definition from the cloned sheet and simply replace the names
      newt <- t
      newt <-
        gsub(
          stri_join(" name=\"", oldname, "\""),
          stri_join(" name=\"", newname, "\""),
          newt
        )
      newt <-
        gsub(
          stri_join(" displayName=\"", olddispname, "\""),
          stri_join(" displayName=\"", newdispname, "\""),
          newt
        )
      newt <-
        gsub(
          stri_join("(<table [^<]* id=\")", oldid, "\""),
          stri_join("\\1", newid, "\""),
          newt
        )

      oldtables <- tables
      tables <<- c(oldtables, newt)
      names(tables) <<- c(names(oldtables), ref)
      attr(tables, "sheet") <<-
        c(attr(oldtables, "sheet"), newSheetIndex)
      attr(tables, "tableName") <<-
        c(attr(oldtables, "tableName"), newname)

      oldparts <- worksheets[[newSheetIndex]]$tableParts
      worksheets[[newSheetIndex]]$tableParts <<-
        c(oldparts, sprintf('<tablePart r:id="rId%s"/>', newid))
      attr(worksheets[[newSheetIndex]]$tableParts, "tableName") <<-
        c(attr(oldparts, "tableName"), newname)
      names(attr(worksheets[[newSheetIndex]]$tableParts, "tableName")) <<-
        c(names(attr(oldparts, "tableName")), ref)

      Content_Types <<-
        c(
          Content_Types,
          sprintf(
            '<Override PartName="/xl/tables/table%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.table+xml"/>',
            newid
          )
        )
      tables.xml.rels <<- append(tables.xml.rels, "")

      worksheets_rels[[newSheetIndex]] <<-
        c(
          worksheets_rels[[newSheetIndex]],
          sprintf(
            '<Relationship Id="rId%s" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/table" Target="../tables/table%s.xml"/>',
            newid,
            newid
          )
        )
    }

    # TODO: The following items are currently NOT copied/duplicated for the cloned sheet:
    #   - Comments
    #   - Pivot tables

    invisible(newSheetIndex)
  }
)

Workbook$methods(
  addChartSheet = function(sheetName,
                           tabColour = NULL,
                           zoom = 100) {
    newSheetIndex <- length(worksheets) + 1L

    if (newSheetIndex > 1) {
      sheetId <-
        max(as.integer(regmatches(
          workbook$sheets,
          regexpr('(?<=sheetId=")[0-9]+', workbook$sheets, perl = TRUE)
        ))) + 1L
    } else {
      sheetId <- 1
    }

    ##  Add sheet to workbook.xml
    workbook$sheets <<-
      c(
        workbook$sheets,
        sprintf(
          '<sheet name="%s" sheetId="%s" r:id="rId%s"/>',
          sheetName,
          sheetId,
          newSheetIndex
        )
      )

    ## append to worksheets list
    worksheets <<-
      append(
        worksheets,
        ChartSheet$new(
          tabSelected = newSheetIndex == 1,
          tabColour = tabColour,
          zoom = zoom
        )
      )
    sheet_names <<- c(sheet_names, sheetName)

    ## update content_tyes
    Content_Types <<-
      c(
        Content_Types,
        sprintf(
          '<Override PartName="/xl/chartsheets/sheet%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.chartsheet+xml"/>',
          newSheetIndex
        )
      )

    ## Update xl/rels
    workbook.xml.rels <<- c(
      workbook.xml.rels,
      sprintf(
        '<Relationship Id="rId0" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/chartsheet" Target="chartsheets/sheet%s.xml"/>',
        newSheetIndex
      )
    )

    ## add a drawing.xml for the worksheet
    Content_Types <<-
      c(
        Content_Types,
        sprintf(
          '<Override PartName="/xl/drawings/drawing%s.xml" ContentType="application/vnd.openxmlformats-officedocument.drawing+xml"/>',
          newSheetIndex
        )
      )

    ## create sheet.rels to simplify id assignment
    worksheets_rels[[newSheetIndex]] <<-
      genBaseSheetRels(newSheetIndex)
    drawings_rels[[newSheetIndex]] <<- list()
    drawings[[newSheetIndex]] <<- list()

    isChartSheet[[newSheetIndex]] <<- TRUE

    rowHeights[[newSheetIndex]] <<- list()
    colWidths[[newSheetIndex]] <<- list()

    colOutlineLevels[[newSheetIndex]] <<- list()
    outlineLevels[[newSheetIndex]] <<- list()

    vml_rels[[newSheetIndex]] <<- list()
    vml[[newSheetIndex]] <<- list()

    sheetOrder <<- c(sheetOrder, newSheetIndex)

    invisible(newSheetIndex)
  }
)



Workbook$methods(
  saveWorkbook = function() {
    ## temp directory to save XML files prior to compressing
    tmpDir <- file.path(tempfile(pattern = "workbookTemp_"))

    if (dir.exists(tmpDir)) {
      unlink(tmpDir, recursive = TRUE, force = TRUE)
    }

    success <- dir.create(path = tmpDir, recursive = FALSE)
    if (!success) {
      stop(sprintf("Failed to create temporary directory '%s'", tmpDir))
    }

    .self$preSaveCleanUp()

    nSheets <- length(worksheets)
    nThemes <- length(theme)
    nPivots <- length(pivotDefinitions)
    nSlicers <- length(slicers)
    nComments <- sum(sapply(comments, length) > 0)
    nThreadComments <- sum(sapply(threadComments, length) > 0)
    nPersons <- length(persons)
    nVML <- sum(sapply(vml, length) > 0)

    relsDir <- file.path(tmpDir, "_rels")
    dir.create(path = relsDir, recursive = TRUE)

    docPropsDir <- file.path(tmpDir, "docProps")
    dir.create(path = docPropsDir, recursive = TRUE)

    xlDir <- file.path(tmpDir, "xl")
    dir.create(path = xlDir, recursive = TRUE)

    xlrelsDir <- file.path(tmpDir, "xl", "_rels")
    dir.create(path = xlrelsDir, recursive = TRUE)

    xlTablesDir <- file.path(tmpDir, "xl", "tables")
    dir.create(path = xlTablesDir, recursive = TRUE)

    xlTablesRelsDir <- file.path(xlTablesDir, "_rels")
    dir.create(path = xlTablesRelsDir, recursive = TRUE)

    if (length(media) > 0) {
      xlmediaDir <- file.path(tmpDir, "xl", "media")
      dir.create(path = xlmediaDir, recursive = TRUE)
    }


    ## will always have a theme
    xlthemeDir <- file.path(tmpDir, "xl", "theme")
    dir.create(path = xlthemeDir, recursive = TRUE)

    if (is.null(theme)) {
      con <- file(file.path(xlthemeDir, "theme1.xml"), open = "wb")
      writeBin(charToRaw(genBaseTheme()), con)
      close(con)
    } else {
      lapply(1:nThemes, function(i) {
        con <-
          file(file.path(xlthemeDir, stri_join("theme", i, ".xml")), open = "wb")
        writeBin(charToRaw(pxml(theme[[i]])), con)
        close(con)
      })
    }

    ## will always have drawings
    xlworksheetsDir <- file.path(tmpDir, "xl", "worksheets")
    dir.create(path = xlworksheetsDir, recursive = TRUE)

    xlworksheetsRelsDir <-
      file.path(tmpDir, "xl", "worksheets", "_rels")
    dir.create(path = xlworksheetsRelsDir, recursive = TRUE)

    xldrawingsDir <- file.path(tmpDir, "xl", "drawings")
    dir.create(path = xldrawingsDir, recursive = TRUE)

    xldrawingsRelsDir <- file.path(tmpDir, "xl", "drawings", "_rels")
    dir.create(path = xldrawingsRelsDir, recursive = TRUE)

    ## charts
    if (length(charts) > 0) {
      file.copy(
        from = dirname(charts[1]),
        to = file.path(tmpDir, "xl"),
        recursive = TRUE
      )
    }


    ## xl/comments.xml
    if (nComments > 0 | nVML > 0) {
      for (i in 1:nSheets) {
        if (length(comments[[i]]) > 0) {
          fn <- sprintf("comments%s.xml", i)

          Content_Types <<- c(
            Content_Types,
            sprintf(
              '<Override PartName="/xl/%s" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.comments+xml"/>',
              fn
            )
          )

          worksheets_rels[[i]] <<- unique(c(
            worksheets_rels[[i]],
            sprintf(
              '<Relationship Id="rIdcomment" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/comments" Target="../%s"/>',
              fn
            )
          ))

          writeCommentXML(
            comment_list = comments[[i]],
            file_name = file.path(tmpDir, "xl", fn)
          )
        }
      }

      .self$writeDrawingVML(xldrawingsDir)
    }

    ## Threaded Comments xl/threadedComments/threadedComment.xml
    if (nThreadComments > 0) {
      xlThreadComments <- file.path(tmpDir, "xl", "threadedComments")
      dir.create(path = xlThreadComments, recursive = TRUE)

      for (i in seq_len(nSheets)) {
        if (length(threadComments[[i]]) > 0) {
          fl <- threadComments[[i]]
          file.copy(
            from = fl,
            to = file.path(xlThreadComments, basename(fl)),
            overwrite = TRUE,
            copy.date = TRUE
          )

          worksheets_rels[[i]] <<- unique(c(
            worksheets_rels[[i]],
            sprintf(
              '<Relationship Id="rIdthread" Type="http://schemas.microsoft.com/office/2017/10/relationships/threadedComment" Target="../threadedComments/%s"/>',
              basename(fl)
            )
          ))
        }
      }
    }

    ## xl/persons/person.xml
    if (nPersons > 0) {
      personDir <- file.path(tmpDir, "xl", "persons")
      dir.create(path = personDir, recursive = TRUE)
      file.copy(
        from = persons,
        to = personDir,
        overwrite = TRUE
      )

    }


    if (length(embeddings) > 0) {
      embeddingsDir <- file.path(tmpDir, "xl", "embeddings")
      dir.create(path = embeddingsDir, recursive = TRUE)
      for (fl in embeddings) {
        file.copy(
          from = fl,
          to = embeddingsDir,
          overwrite = TRUE
        )
      }
    }


    if (nPivots > 0) {
      pivotTablesDir <- file.path(tmpDir, "xl", "pivotTables")
      dir.create(path = pivotTablesDir, recursive = TRUE)

      pivotTablesRelsDir <-
        file.path(tmpDir, "xl", "pivotTables", "_rels")
      dir.create(path = pivotTablesRelsDir, recursive = TRUE)

      pivotCacheDir <- file.path(tmpDir, "xl", "pivotCache")
      dir.create(path = pivotCacheDir, recursive = TRUE)

      pivotCacheRelsDir <-
        file.path(tmpDir, "xl", "pivotCache", "_rels")
      dir.create(path = pivotCacheRelsDir, recursive = TRUE)

      for (i in seq_along(pivotTables)) {
        file.copy(
          from = pivotTables[i],
          to = file.path(pivotTablesDir, sprintf("pivotTable%s.xml", i)),
          overwrite = TRUE,
          copy.date = TRUE
        )
      }

      for (i in seq_along(pivotDefinitions)) {
        file.copy(
          from = pivotDefinitions[i],
          to = file.path(pivotCacheDir, sprintf("pivotCacheDefinition%s.xml", i)),
          overwrite = TRUE,
          copy.date = TRUE
        )
      }

      for (i in seq_along(pivotRecords)) {
        file.copy(
          from = pivotRecords[i],
          to = file.path(pivotCacheDir, sprintf("pivotCacheRecords%s.xml", i)),
          overwrite = TRUE,
          copy.date = TRUE
        )
      }

      for (i in seq_along(pivotDefinitionsRels)) {
        file.copy(
          from = pivotDefinitionsRels[i],
          to = file.path(
            pivotCacheRelsDir,
            sprintf("pivotCacheDefinition%s.xml.rels", i)
          ),
          overwrite = TRUE,
          copy.date = TRUE
        )
      }

      for (i in seq_along(pivotTables.xml.rels)) {
        write_file(
          body = pivotTables.xml.rels[[i]],
          fl = file.path(pivotTablesRelsDir, sprintf("pivotTable%s.xml.rels", i))
        )
      }
    }

    ## slicers
    if (nSlicers > 0) {
      slicersDir <- file.path(tmpDir, "xl", "slicers")
      dir.create(path = slicersDir, recursive = TRUE)

      slicerCachesDir <- file.path(tmpDir, "xl", "slicerCaches")
      dir.create(path = slicerCachesDir, recursive = TRUE)

      for (i in seq_along(slicers)) {
        if (nchar(slicers[i]) > 0) {
          file.copy(from = slicers[i], to = file.path(slicersDir, sprintf("slicer%s.xml", i)))
        }
      }



      for (i in seq_along(slicerCaches)) {
        write_file(
          body = slicerCaches[[i]],
          fl = file.path(slicerCachesDir, sprintf("slicerCache%s.xml", i))
        )
      }
    }


    ## Write content

    ## write .rels
    write_file(
      head = '<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships">\n',
      body = '<Relationship Id="rId3" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/extended-properties" Target="docProps/app.xml"/><Relationship Id="rId2" Type="http://schemas.openxmlformats.org/package/2006/relationships/metadata/core-properties" Target="docProps/core.xml"/><Relationship Id="rId1" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/officeDocument" Target="xl/workbook.xml"/>',
      tail = "</Relationships>",
      fl = file.path(relsDir, ".rels")
    )

    app <- "<Application>Microsoft Excel</Application>"
    # further protect argument (might be extended with: <ScaleCrop>, <HeadingPairs>, <TitlesOfParts>, <LinksUpToDate>, <SharedDoc>, <HyperlinksChanged>, <AppVersion>)
    if (!is.null(workbook$apps)) app <- paste0(app, workbook$apps)

    ## write app.xml
    write_file(
      head = '<Properties xmlns="http://schemas.openxmlformats.org/officeDocument/2006/extended-properties" xmlns:vt="http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes">',
      body = app,
      tail = "</Properties>",
      fl = file.path(docPropsDir, "app.xml")
    )

    ## write core.xml
    write_file(
      head = "",
      body = pxml(core),
      tail = "",
      fl = file.path(docPropsDir, "core.xml")
    )

    ## write workbook.xml.rels
    write_file(
      head = '<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships">',
      body = pxml(workbook.xml.rels),
      tail = "</Relationships>",
      fl = file.path(xlrelsDir, "workbook.xml.rels")
    )

    ## write tables
    if (length(unlist(tables, use.names = FALSE)) > 0) {
      for (i in seq_along(unlist(tables, use.names = FALSE))) {
        if (!grepl("openxlsx_deleted", attr(tables, "tableName")[i], fixed = TRUE)) {
          write_file(
            body = pxml(unlist(tables, use.names = FALSE)[[i]]),
            fl = file.path(xlTablesDir, sprintf("table%s.xml", i + 2))
          )
          if (tables.xml.rels[[i]] != "") {
            write_file(
              body = tables.xml.rels[[i]],
              fl = file.path(xlTablesRelsDir, sprintf("table%s.xml.rels", i + 2))
            )
          }
        }
      }
    }


    ## write query tables
    if (length(queryTables) > 0) {
      xlqueryTablesDir <- file.path(tmpDir, "xl", "queryTables")
      dir.create(path = xlqueryTablesDir, recursive = TRUE)

      for (i in seq_along(queryTables)) {
        write_file(
          body = queryTables[[i]],
          fl = file.path(xlqueryTablesDir, sprintf("queryTable%s.xml", i))
        )
      }
    }

    ## connections
    if (length(connections) > 0) {
      write_file(body = connections, fl = file.path(xlDir, "connections.xml"))
    }

    ## externalLinks
    if (length(externalLinks)) {
      externalLinksDir <- file.path(tmpDir, "xl", "externalLinks")
      dir.create(path = externalLinksDir, recursive = TRUE)

      for (i in seq_along(externalLinks)) {
        write_file(
          body = externalLinks[[i]],
          fl = file.path(externalLinksDir, sprintf("externalLink%s.xml", i))
        )
      }
    }

    ## externalLinks rels
    if (length(externalLinksRels)) {
      externalLinksRelsDir <-
        file.path(tmpDir, "xl", "externalLinks", "_rels")
      dir.create(path = externalLinksRelsDir, recursive = TRUE)

      for (i in seq_along(externalLinksRels)) {
        write_file(
          body = externalLinksRels[[i]],
          fl = file.path(
            externalLinksRelsDir,
            sprintf("externalLink%s.xml.rels", i)
          )
        )
      }
    }

    # printerSettings
    printDir <- file.path(tmpDir, "xl", "printerSettings")
    dir.create(path = printDir, recursive = TRUE)
    for (i in 1:nSheets) {
      writeLines(genPrinterSettings(), file.path(printDir, sprintf("printerSettings%s.bin", i)))
    }

    ## media (copy file from origin to destination)
    for (x in media) {
      file.copy(x, file.path(xlmediaDir, names(media)[which(media == x)]))
    }

    ## VBA Macro
    if (!is.null(vbaProject)) {
      file.copy(vbaProject, xlDir)
    }

    ## write worksheet, worksheet_rels, drawings, drawing_rels
    .self$writeSheetDataXML(
      xldrawingsDir,
      xldrawingsRelsDir,
      xlworksheetsDir,
      xlworksheetsRelsDir
    )

    ## write sharedStrings.xml
    ct <- Content_Types
    if (length(sharedStrings) > 0) {
      write_file(
        head = sprintf(
          '<sst xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" count="%s" uniqueCount="%s">',
          length(sharedStrings),
          attr(sharedStrings, "uniqueCount")
        ),
        body = stri_join(sharedStrings, collapse = "", sep = " "),
        tail = "</sst>",
        fl = file.path(xlDir, "sharedStrings.xml")
      )
    } else {
      ## Remove relationship to sharedStrings
      ct <- ct[!grepl("sharedStrings", ct)]
    }

    if (nComments > 0) {
      ct <-
        c(
          ct,
          '<Default Extension="vml" ContentType="application/vnd.openxmlformats-officedocument.vmlDrawing"/>'
        )
    }

    ## write [Content_type]
    write_file(
      head = '<Types xmlns="http://schemas.openxmlformats.org/package/2006/content-types">',
      body = pxml(ct),
      tail = "</Types>",
      fl = file.path(tmpDir, "[Content_Types].xml")
    )


    styleXML <- styles
    styleXML$numFmts <-
      stri_join(
        sprintf('<numFmts count="%s">', length(styles$numFmts)),
        pxml(styles$numFmts),
        "</numFmts>"
      )
    styleXML$fonts <-
      stri_join(
        sprintf('<fonts count="%s">', length(styles$fonts)),
        pxml(styles$fonts),
        "</fonts>"
      )
    styleXML$fills <-
      stri_join(
        sprintf('<fills count="%s">', length(styles$fills)),
        pxml(styles$fills),
        "</fills>"
      )
    styleXML$borders <-
      stri_join(
        sprintf('<borders count="%s">', length(styles$borders)),
        pxml(styles$borders),
        "</borders>"
      )
    styleXML$cellStyleXfs <-
      c(
        sprintf('<cellStyleXfs count="%s">', length(styles$cellStyleXfs)),
        pxml(styles$cellStyleXfs),
        "</cellStyleXfs>"
      )
    styleXML$cellXfs <-
      stri_join(
        sprintf('<cellXfs count="%s">', length(styles$cellXfs)),
        pxml(styles$cellXfs),
        "</cellXfs>"
      )
    styleXML$cellStyles <-
      stri_join(
        sprintf('<cellStyles count="%s">', length(styles$cellStyles)),
        pxml(styles$cellStyles),
        "</cellStyles>"
      )
    styleXML$dxfs <-
      ifelse(
        length(styles$dxfs) == 0,
        '<dxfs count="0"/>',
        stri_join(
          sprintf('<dxfs count="%s">', length(styles$dxfs)),
          stri_join(unlist(styles$dxfs), sep = " ", collapse = ""),
          "</dxfs>"
        )
      )
    ## write styles.xml
    write_file(
      head = '<styleSheet xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" mc:Ignorable="x14ac x16r2 xr" xmlns:x14ac="http://schemas.microsoft.com/office/spreadsheetml/2009/9/ac" xmlns:x16r2="http://schemas.microsoft.com/office/spreadsheetml/2015/02/main" xmlns:xr="http://schemas.microsoft.com/office/spreadsheetml/2014/revision">',
      body = pxml(styleXML),
      tail = "</styleSheet>",
      fl = file.path(xlDir, "styles.xml")
    )

    ## write workbook.xml
    workbookXML <- workbook
    workbookXML$sheets <-
      stri_join("<sheets>", pxml(workbookXML$sheets), "</sheets>")
    if (length(workbookXML$definedNames) > 0) {
      workbookXML$definedNames <-
        stri_join(
          "<definedNames>",
          pxml(workbookXML$definedNames),
          "</definedNames>"
        )
    }

    write_file(
      head = '<workbook xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" mc:Ignorable="x15 xr xr6 xr10 xr2" xmlns:x15="http://schemas.microsoft.com/office/spreadsheetml/2010/11/main" xmlns:xr="http://schemas.microsoft.com/office/spreadsheetml/2014/revision" xmlns:xr6="http://schemas.microsoft.com/office/spreadsheetml/2016/revision6" xmlns:xr10="http://schemas.microsoft.com/office/spreadsheetml/2016/revision10" xmlns:xr2="http://schemas.microsoft.com/office/spreadsheetml/2015/revision2">',
      body = pxml(workbookXML),
      tail = "</workbook>",
      fl = file.path(xlDir, "workbook.xml")
    )
    workbook$sheets <<-
      workbook$sheets[order(sheetOrder)] ## Need to reset sheet order to allow multiple savings

    ## compress to xlsx
    wd <- getwd()
    tmpFile <-
      basename(tempfile(fileext = ifelse(is.null(vbaProject), ".xlsx", ".xlsm")))
    on.exit(expr = setwd(wd), add = TRUE)

    ## zip it
    setwd(dir = tmpDir)
    cl <-
      ifelse(
        !is.null(getOption("openxlsx.compresssionLevel")),
        getOption("openxlsx.compresssionLevel"),
        getOption("openxlsx.compresssionevel", 6)
      )
    zipr(
      zipfile = tmpFile, include_directories = FALSE,
      files = list.files(path = tmpDir, all.files = FALSE),
      recurse = TRUE,
      compression_level = cl
    )

    ## reset styles - maintain any changes to base font
    baseFont <- styles$fonts[[1]]
    styles <<-
      genBaseStyleSheet(styles$dxfs,
        tableStyles = styles$tableStyles,
        extLst = styles$extLst
      )
    styles$fonts[[1]] <<- baseFont


    return(file.path(tmpDir, tmpFile))
  }
)



Workbook$methods(
  updateSharedStrings = function(uNewStr) {
    ## Function will return named list of references to new strings
    uStr <- uNewStr[which(!uNewStr %in% sharedStrings)]
    uCount <- attr(sharedStrings, "uniqueCount")
    sharedStrings <<- append(sharedStrings, uStr)

    attr(sharedStrings, "uniqueCount") <<- uCount + length(uStr)
  }
)

Workbook$methods(
  validateSheet = function(sheetName) {
    if (!is.numeric(sheetName)) {
      if (is.null(sheet_names)) {
        stop("Workbook does not contain any worksheets.", call. = FALSE)
      }
    }

    if (is.numeric(sheetName)) {
      if (sheetName > length(sheet_names)) {
        stop("This Workbook only has ", length(sheet_names),
          " sheets, ", sheetName, " is not valid",
          call. = FALSE
        )
      }
      return(sheetName)
    } else if (!sheetName %in% replaceXMLEntities(sheet_names)) {
      stop(sprintf("Sheet '%s' does not exist.", replaceXMLEntities(sheetName)),
        call. = FALSE)
    }

    which(replaceXMLEntities(sheet_names) == sheetName)
  }
)


Workbook$methods(
  getSheetName = function(sheetIndex) {
    if (any(length(sheet_names) < sheetIndex)) {
      stop(sprintf("Workbook only contains %s sheet(s).", length(sheet_names)))
    }

    sheet_names[sheetIndex]
  }
)

Workbook$methods(
  buildTable = function(sheet,
                        colNames,
                        ref,
                        showColNames,
                        tableStyle,
                        tableName,
                        withFilter,
                        totalsRowCount = 0,
                        showFirstColumn = 0,
                        showLastColumn = 0,
                        showRowStripes = 1,
                        showColumnStripes = 0) {
    ## id will start at 3 and drawing will always be 1, printer Settings at 2 (printer settings has been removed)
    id <- as.character(length(tables) + 3L)
    sheet <- validateSheet(sheet)

    ## build table XML and save to tables field
    table <-
      sprintf(
        '<table xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" id="%s" name="%s" displayName="%s" ref="%s" totalsRowCount="%s"',
        id,
        tableName,
        tableName,
        ref,
        as.integer(totalsRowCount)
      )
    # because tableName might be native encoded non-ASCII strings, we need to ensure
    # it's UTF-8 encoded
    table <- enc2utf8(table)

    nms <- names(tables)
    tSheets <- attr(tables, "sheet")
    tNames <- attr(tables, "tableName")

    tableStyleXML <-
      sprintf(
        '<tableStyleInfo name="%s" showFirstColumn="%s" showLastColumn="%s" showRowStripes="%s" showColumnStripes="%s"/>',
        tableStyle,
        as.integer(showFirstColumn),
        as.integer(showLastColumn),
        as.integer(showRowStripes),
        as.integer(showColumnStripes)
      )


    tables <<-
      c(
        tables,
        build_table_xml(
          table = table,
          tableStyleXML = tableStyleXML,
          ref = ref,
          colNames = gsub("\n|\r", "_x000a_", colNames),
          showColNames = showColNames,
          withFilter = withFilter
        )
      )
    names(tables) <<- c(nms, ref)
    attr(tables, "sheet") <<- c(tSheets, sheet)
    attr(tables, "tableName") <<- c(tNames, tableName)

    worksheets[[sheet]]$tableParts <<-
      append(
        worksheets[[sheet]]$tableParts,
        sprintf('<tablePart r:id="rId%s"/>', id)
      )
    attr(worksheets[[sheet]]$tableParts, "tableName") <<-
      c(tNames[tSheets == sheet &
        !grepl("openxlsx_deleted", tNames, fixed = TRUE)], tableName)



    ## update Content_Types
    Content_Types <<-
      c(
        Content_Types,
        sprintf(
          '<Override PartName="/xl/tables/table%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.table+xml"/>',
          id
        )
      )

    ## create a table.xml.rels
    tables.xml.rels <<- append(tables.xml.rels, "")

    ## update worksheets_rels
    worksheets_rels[[sheet]] <<- c(
      worksheets_rels[[sheet]],
      sprintf(
        '<Relationship Id="rId%s" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/table" Target="../tables/table%s.xml"/>',
        id,
        id
      )
    )
  }
)









Workbook$methods(
  writeDrawingVML = function(dir) {
    for (i in seq_along(comments)) {
      id <- 1025

      cd <- unlist(lapply(comments[[i]], "[[", "clientData"))
      nComments <- length(cd)

      ## write head
      if (nComments > 0 | length(vml[[i]]) > 0) {
        write(
          x = stri_join(
            '<xml xmlns:v="urn:schemas-microsoft-com:vml"
                    xmlns:o="urn:schemas-microsoft-com:office:office"
                    xmlns:x="urn:schemas-microsoft-com:office:excel">
                    <o:shapelayout v:ext="edit">
                    <o:idmap v:ext="edit" data="1"/>
                    </o:shapelayout>
                    <v:shapetype id="_x0000_t202" coordsize="21600,21600" o:spt="202"
                    path="m,l,21600r21600,l21600,xe">
                    <v:stroke joinstyle="miter"/>
                    <v:path gradientshapeok="t" o:connecttype="rect"/>
                    </v:shapetype>'
          ),
          file = file.path(dir, sprintf("vmlDrawing%s.vml", i)),
          sep = " "
        )
      }

      if (nComments > 0) {
        for (j in 1:nComments) {
          id <- id + 1L
          write(
            x = genBaseShapeVML(cd[j], id),
            file = file.path(dir, sprintf("vmlDrawing%s.vml", i)),
            append = TRUE
          )
        }
      }

      if (length(vml[[i]]) > 0) {
        write(
          x = vml[[i]],
          file = file.path(dir, sprintf("vmlDrawing%s.vml", i)),
          append = TRUE
        )
      }

      if (nComments > 0 | length(vml[[i]]) > 0) {
        write(
          x = "</xml>",
          file = file.path(dir, sprintf("vmlDrawing%s.vml", i)),
          append = TRUE
        )
        worksheets[[i]]$legacyDrawing <<-
          '<legacyDrawing r:id="rIdvml"/>'
      }
    }
  }
)



Workbook$methods(
  updateStyles = function(style) {
    ## Updates styles.xml
    xfNode <- list(
      numFmtId = 0,
      fontId = 0,
      fillId = 0,
      borderId = 0,
      xfId = 0
    )


    alignmentFlag <- FALSE

    ## Font
    if (!is.null(style$fontName) |
      !is.null(style$fontSize) |
      !is.null(style$fontColour) |
      !is.null(style$fontDecoration) |
      !is.null(style$fontFamily) |
      !is.null(style$fontScheme)) {
      fontNode <- .self$createFontNode(style)
      fontId <- which(styles$fonts == fontNode) - 1L

      if (length(fontId) == 0) {
        fontId <- length(styles$fonts)
        styles$fonts <<- append(styles[["fonts"]], fontNode)
      }

      xfNode$fontId <- fontId
      xfNode <- append(xfNode, list("applyFont" = "1"))
    }


    ## numFmt
    if (!is.null(style$numFmt)) {
      if (as.integer(style$numFmt$numFmtId) > 0) {
        numFmtId <- style$numFmt$numFmtId
        if (as.integer(numFmtId) > 163L) {
          tmp <- style$numFmt$formatCode

          styles$numFmts <<- unique(c(
            styles$numFmts,
            sprintf(
              '<numFmt numFmtId="%s" formatCode="%s"/>',
              numFmtId,
              tmp
            )
          ))
        }

        xfNode$numFmtId <- numFmtId
        xfNode <- append(xfNode, list("applyNumberFormat" = "1"))
      }
    }

    ## Fill
    if (!is.null(style$fill)) {
      fillNode <- .self$createFillNode(style)
      if (!is.null(fillNode)) {
        fillId <- which(styles$fills == fillNode) - 1L

        if (length(fillId) == 0) {
          fillId <- length(styles$fills)
          styles$fills <<- c(styles$fills, fillNode)
        }
        xfNode$fillId <- fillId
        xfNode <- append(xfNode, list("applyFill" = "1"))
      }
    }

    ## Border
    if (any(!is.null(
      c(
        style$borderLeft,
        style$borderRight,
        style$borderTop,
        style$borderBottom,
        style$borderDiagonal
      )
    ))) {
      borderNode <- .self$createBorderNode(style)
      borderId <- which(styles$borders == borderNode) - 1L

      if (length(borderId) == 0) {
        borderId <- length(styles$borders)
        styles$borders <<- c(styles$borders, borderNode)
      }

      xfNode$borderId <- borderId
      xfNode <- append(xfNode, list("applyBorder" = "1"))
    }


    # if(!is.null(style$xfId))
    # xfNode$xfId <- style$xfId

    childNodes <- ""

    ## Alignment
    if (!is.null(style$halign) |
      !is.null(style$valign) |
      !is.null(style$wrapText) |
      !is.null(style$textRotation) | !is.null(style$indent)) {
      attrs <- list()
      alignNode <- "<alignment"

      if (!is.null(style$textRotation)) {
        alignNode <-
          stri_join(alignNode,
            sprintf('textRotation="%s"', style$textRotation),
            sep = " "
          )
      }

      if (!is.null(style$halign)) {
        alignNode <-
          stri_join(alignNode, sprintf('horizontal="%s"', style$halign), sep = " ")
      }

      if (!is.null(style$valign)) {
        alignNode <-
          stri_join(alignNode, sprintf('vertical="%s"', style$valign), sep = " ")
      }

      if (!is.null(style$indent)) {
        alignNode <-
          stri_join(alignNode, sprintf('indent="%s"', style$indent), sep = " ")
      }

      if (!is.null(style$wrapText)) {
        if (style$wrapText) {
          alignNode <- stri_join(alignNode, 'wrapText="1"', sep = " ")
        }
      }


      alignNode <- stri_join(alignNode, "/>")

      alignmentFlag <- TRUE
      xfNode <- append(xfNode, list("applyAlignment" = "1"))

      childNodes <- stri_join(childNodes, alignNode)
    }

    if (!is.null(style$hidden) | !is.null(style$locked)) {
      xfNode <- append(xfNode, list("applyProtection" = "1"))
      protectionNode <- "<protection"

      if (!is.null(style$hidden)) {
        protectionNode <-
          stri_join(protectionNode, sprintf('hidden="%s"', as.numeric(style$hidden)), sep = " ")
      }
      if (!is.null(style$locked)) {
        protectionNode <-
          stri_join(protectionNode, sprintf('locked="%s"', as.numeric(style$locked)), sep = " ")
      }

      protectionNode <- stri_join(protectionNode, "/>")
      childNodes <- stri_join(childNodes, protectionNode)
    }

    if (length(childNodes) > 0) {
      xfNode <-
        stri_join(
          "<xf ",
          stri_join(
            stri_join(names(xfNode), '="', xfNode, '"'),
            sep = " ",
            collapse = " "
          ),
          ">",
          childNodes,
          "</xf>"
        )
    } else {
      xfNode <-
        stri_join("<xf ", stri_join(
          stri_join(names(xfNode), '="', xfNode, '"'),
          sep = " ",
          collapse = " "
        ), "/>")
    }

    styleId <- which(styles$cellXfs == xfNode) - 1L
    if (length(styleId) == 0) {
      styleId <- length(styles$cellXfs)
      styles$cellXfs <<- c(styles$cellXfs, xfNode)
    }


    return(as.integer(styleId))
  }
)





Workbook$methods(
  updateCellStyles = function() {
    flag <- TRUE
    for (style in cellStyleObjects) {
      ## Updates styles.xml
      xfNode <- list(
        numFmtId = 0,
        fontId = 0,
        fillId = 0,
        borderId = 0
      )


      alignmentFlag <- FALSE

      ## Font
      if (!is.null(style$fontName) |
        !is.null(style$fontSize) |
        !is.null(style$fontColour) |
        !is.null(style$fontDecoration) |
        !is.null(style$fontFamily) |
        !is.null(style$fontScheme)) {
        fontNode <- .self$createFontNode(style)
        fontId <- which(styles$font == fontNode) - 1L

        if (length(fontId) == 0) {
          fontId <- length(styles$fonts)
          styles$fonts <<- append(styles[["fonts"]], fontNode)
        }

        xfNode$fontId <- fontId
        xfNode <- append(xfNode, list("applyFont" = "1"))
      }


      ## numFmt
      if (!is.null(style$numFmt)) {
        if (as.integer(style$numFmt$numFmtId) > 0) {
          numFmtId <- style$numFmt$numFmtId
          if (as.integer(numFmtId) > 163L) {
            tmp <- style$numFmt$formatCode

            styles$numFmts <<- unique(c(
              styles$numFmts,
              sprintf(
                '<numFmt numFmtId="%s" formatCode="%s"/>',
                numFmtId,
                tmp
              )
            ))
          }

          xfNode$numFmtId <- numFmtId
          xfNode <- append(xfNode, list("applyNumberFormat" = "1"))
        }
      }

      ## Fill
      if (!is.null(style$fill)) {
        fillNode <- .self$createFillNode(style)
        if (!is.null(fillNode)) {
          fillId <- which(styles$fills == fillNode) - 1L

          if (length(fillId) == 0) {
            fillId <- length(styles$fills)
            styles$fills <<- c(styles$fills, fillNode)
          }
          xfNode$fillId <- fillId
          xfNode <- append(xfNode, list("applyFill" = "1"))
        }
      }

      ## Border
      if (any(!is.null(
        c(
          style$borderLeft,
          style$borderRight,
          style$borderTop,
          style$borderBottom,
          style$borderDiagonal
        )
      ))) {
        borderNode <- .self$createBorderNode(style)
        borderId <- which(styles$borders == borderNode) - 1L

        if (length(borderId) == 0) {
          borderId <- length(styles$borders)
          styles$borders <<- c(styles$borders, borderNode)
        }

        xfNode$borderId <- borderId
        xfNode <- append(xfNode, list("applyBorder" = "1"))
      }

      xfNode <-
        stri_join("<xf ", stri_join(
          stri_join(names(xfNode), '="', xfNode, '"'),
          sep = " ",
          collapse = " "
        ), "/>")

      if (flag) {
        styles$cellStyleXfs <<- xfNode
        flag <- FALSE
      } else {
        styles$cellStyleXfs <<- c(styles$cellStyleXfs, xfNode)
      }
    }
  }
)








Workbook$methods(
  getBaseFont = function() {
    baseFont <- styles$fonts[[1]]

    sz <- getAttrs(baseFont, "sz")
    colour <- getAttrs(baseFont, "color")
    name <- getAttrs(baseFont, "name")

    if (length(sz[[1]]) == 0) {
      sz <- list("val" = "10")
    }

    if (length(colour[[1]]) == 0) {
      colour <- list("rgb" = "#000000")
    }

    if (length(name[[1]]) == 0) {
      name <- list("val" = "Calibri")
    }

    list(
      "size" = sz,
      "colour" = colour,
      "name" = name
    )
  }
)




Workbook$methods(
  createFontNode = function(style) {
    baseFont <- .self$getBaseFont()

    fontNode <- "<font>"

    ## size
    if (is.null(style$fontSize[[1]])) {
      fontNode <-
        stri_join(fontNode, sprintf('<sz %s="%s"/>', names(baseFont$size), baseFont$size))
    } else {
      fontNode <-
        stri_join(fontNode, sprintf('<sz %s="%s"/>', names(style$fontSize), style$fontSize))
    }

    ## colour
    if (is.null(style$fontColour[[1]])) {
      fontNode <-
        stri_join(
          fontNode,
          sprintf(
            '<color %s="%s"/>',
            names(baseFont$colour),
            baseFont$colour
          )
        )
    } else {
      if (length(style$fontColour) > 1) {
        fontNode <- stri_join(fontNode, sprintf(
          "<color %s/>",
          stri_join(
            sapply(seq_along(style$fontColour), function(i) {
              sprintf('%s="%s"', names(style$fontColour)[i], style$fontColour[i])
            }),
            sep = " ",
            collapse = " "
          )
        ))
      } else {
        fontNode <-
          stri_join(
            fontNode,
            sprintf(
              '<color %s="%s"/>',
              names(style$fontColour),
              style$fontColour
            )
          )
      }
    }


    ## name
    if (is.null(style$fontName[[1]])) {
      fontNode <-
        stri_join(
          fontNode,
          sprintf('<name %s="%s"/>', names(baseFont$name), baseFont$name)
        )
    } else {
      fontNode <-
        stri_join(
          fontNode,
          sprintf('<name %s="%s"/>', names(style$fontName), style$fontName)
        )
    }

    ### Create new font and return Id
    if (!is.null(style$fontFamily)) {
      fontNode <-
        stri_join(fontNode, sprintf('<family val = "%s"/>', style$fontFamily))
    }

    if (!is.null(style$fontScheme)) {
      fontNode <-
        stri_join(fontNode, sprintf('<scheme val = "%s"/>', style$fontScheme))
    }

    if ("BOLD" %in% style$fontDecoration) {
      fontNode <- stri_join(fontNode, "<b/>")
    }

    if ("ITALIC" %in% style$fontDecoration) {
      fontNode <- stri_join(fontNode, "<i/>")
    }

    if ("UNDERLINE" %in% style$fontDecoration) {
      fontNode <- stri_join(fontNode, '<u val="single"/>')
    }

    if ("UNDERLINE2" %in% style$fontDecoration) {
      fontNode <- stri_join(fontNode, '<u val="double"/>')
    }

    if ("ACCOUNTING" %in% style$fontDecoration) {
      fontNode <- stri_join(fontNode, '<u val="singleAccounting"/>')
    }

    if ("ACCOUNTING2" %in% style$fontDecoration) {
      fontNode <- stri_join(fontNode, '<u val="doubleAccounting"/>')
    }

    if ("STRIKEOUT" %in% style$fontDecoration) {
      fontNode <- stri_join(fontNode, "<strike/>")
    }

    stri_join(fontNode, "</font>")
  }
)


Workbook$methods(
  createBorderNode = function(style) {
    borderNode <- "<border"

    if (style$borderDiagonalUp) {
      borderNode <- stri_join(borderNode, 'diagonalUp="1"', sep = " ")
    }

    if (style$borderDiagonalDown) {
      borderNode <-
        stri_join(borderNode, 'diagonalDown="1"', sep = " ")
    }

    borderNode <- stri_join(borderNode, ">")

    if (!is.null(style$borderLeft)) {
      borderNode <-
        stri_join(
          borderNode,
          sprintf('<left style="%s">', style$borderLeft),
          sprintf(
            '<color %s="%s"/>',
            names(style$borderLeftColour),
            style$borderLeftColour
          ),
          "</left>"
        )
    }

    if (!is.null(style$borderRight)) {
      borderNode <-
        stri_join(
          borderNode,
          sprintf('<right style="%s">', style$borderRight),
          sprintf(
            '<color %s="%s"/>',
            names(style$borderRightColour),
            style$borderRightColour
          ),
          "</right>"
        )
    }

    if (!is.null(style$borderTop)) {
      borderNode <-
        stri_join(
          borderNode,
          sprintf('<top style="%s">', style$borderTop),
          sprintf(
            '<color %s="%s"/>',
            names(style$borderTopColour),
            style$borderTopColour
          ),
          "</top>"
        )
    }

    if (!is.null(style$borderBottom)) {
      borderNode <-
        stri_join(
          borderNode,
          sprintf('<bottom style="%s">', style$borderBottom),
          sprintf(
            '<color %s="%s"/>',
            names(style$borderBottomColour),
            style$borderBottomColour
          ),
          "</bottom>"
        )
    }

    if (!is.null(style$borderDiagonal)) {
      borderNode <-
        stri_join(
          borderNode,
          sprintf('<diagonal style="%s">', style$borderDiagonal),
          sprintf(
            '<color %s="%s"/>',
            names(style$borderDiagonalColour),
            style$borderDiagonalColour
          ),
          "</diagonal>"
        )
    }

    stri_join(borderNode, "</border>")
  }
)


Workbook$methods(
  createFillNode = function(style, patternType = "solid") {
    fill <- style$fill

    ## gradientFill
    if (any(grepl("gradientFill", fill))) {
      fillNode <- fill # stri_join("<fill>", fill, "</fill>")
    } else if (!is.null(fill$fillFg) | !is.null(fill$fillBg)) {
      fillNode <-
        stri_join(
          "<fill>",
          sprintf('<patternFill patternType="%s">', patternType)
        )

      if (!is.null(fill$fillFg)) {
        fillNode <-
          stri_join(fillNode, sprintf(
            "<fgColor %s/>",
            stri_join(
              stri_join(names(fill$fillFg), '="', fill$fillFg, '"'),
              sep = " ",
              collapse = " "
            )
          ))
      }

      if (!is.null(fill$fillBg)) {
        fillNode <-
          stri_join(fillNode, sprintf(
            "<bgColor %s/>",
            stri_join(
              stri_join(names(fill$fillBg), '="', fill$fillBg, '"'),
              sep = " ",
              collapse = " "
            )
          ))
      }

      fillNode <- stri_join(fillNode, "</patternFill></fill>")
    } else {
      return(NULL)
    }

    return(fillNode)
  }
)







Workbook$methods(
  setSheetName = function(sheet, newSheetName) {
    if (newSheetName %in% sheet_names) {
      stop(sprintf("Sheet %s already exists!", newSheetName))
    }

    sheet <- validateSheet(sheet)

    oldName <- sheet_names[[sheet]]
    sheet_names[[sheet]] <<- newSheetName

    ## Rename in workbook
    sheetId <-
      regmatches(
        workbook$sheets[[sheet]],
        regexpr('(?<=sheetId=")[0-9]+', workbook$sheets[[sheet]], perl = TRUE)
      )
    rId <-
      regmatches(
        workbook$sheets[[sheet]],
        regexpr('(?<= r:id="rId)[0-9]+', workbook$sheets[[sheet]], perl = TRUE)
      )
    workbook$sheets[[sheet]] <<-
      sprintf(
        '<sheet name="%s" sheetId="%s" r:id="rId%s"/>',
        newSheetName,
        sheetId,
        rId
      )

    ## rename styleObjects sheet component
    if (length(styleObjects) > 0) {
      styleObjects <<- lapply(styleObjects, function(x) {
        if (x$sheet == oldName) {
          x$sheet <- newSheetName
        }

        return(x)
      })
    }

    ## rename defined names
    if (length(workbook$definedNames) > 0) {
      belongTo <- getDefinedNamesSheet(workbook$definedNames)
      toChange <- belongTo == oldName
      if (any(toChange)) {
        newSheetName <- sprintf("'%s'", newSheetName)
        tmp <-
          gsub(oldName, newSheetName, workbook$definedName[toChange], fixed = TRUE)
        tmp <- gsub("'+", "'", tmp)
        workbook$definedNames[toChange] <<- tmp
      }
    }
  }
)


Workbook$methods(
  writeSheetDataXML = function(xldrawingsDir,
                               xldrawingsRelsDir,
                               xlworksheetsDir,
                               xlworksheetsRelsDir) {
    ## write worksheets
    # nSheets <- length(worksheets)

    for (i in seq_along(worksheets)) {
      ## Write drawing i (will always exist) skip those that are empty
      if (any(drawings[[i]] != "")) {
        write_file(
          head = '<xdr:wsDr xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing" xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships">',
          body = pxml(drawings[[i]]),
          tail = "</xdr:wsDr>",
          fl = file.path(xldrawingsDir, stri_join("drawing", i, ".xml"))
        )

        write_file(
          head = '<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships">',
          body = pxml(drawings_rels[[i]]),
          tail = "</Relationships>",
          fl = file.path(xldrawingsRelsDir, stri_join("drawing", i, ".xml.rels"))
        )
      } else {
        worksheets[[i]]$drawing <<- character(0)
      }

      ## vml drawing
      if (length(vml_rels[[i]]) > 0) {
        file.copy(
          from = vml_rels[[i]],
          to = file.path(
            xldrawingsRelsDir,
            stri_join("vmlDrawing", i, ".vml.rels")
          )
        )
      }

      # outlineLevelRow in SheetformatPr
      if ((length(outlineLevels[[i]]) > 0) && (!grepl("outlineLevelRow", worksheets[[i]]$sheetFormatPr))) {
        worksheets[[i]]$sheetFormatPr <<- gsub("/>", ' outlineLevelRow="1"/>', worksheets[[i]]$sheetFormatPr)
      }

      if (isChartSheet[i]) {
        chartSheetDir <- file.path(dirname(xlworksheetsDir), "chartsheets")
        chartSheetRelsDir <-
          file.path(dirname(xlworksheetsDir), "chartsheets", "_rels")

        if (!file.exists(chartSheetDir)) {
          dir.create(chartSheetDir, recursive = TRUE)
          dir.create(chartSheetRelsDir, recursive = TRUE)
        }

        write_file(
          body = worksheets[[i]]$get_prior_sheet_data(),
          fl = file.path(chartSheetDir, stri_join("sheet", i, ".xml"))
        )

        write_file(
          head = '<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships">',
          body = pxml(worksheets_rels[[i]]),
          tail = "</Relationships>",
          fl = file.path(chartSheetRelsDir, sprintf("sheet%s.xml.rels", i))
        )
      } else {
        ## Write worksheets
        ws <- worksheets[[i]]
        hasHL <-
          ifelse(length(worksheets[[i]]$hyperlinks) > 0, TRUE, FALSE)

        ## reorder sheet data
        worksheets[[i]]$order_sheetdata()

        prior <- ws$get_prior_sheet_data()
        post <- ws$get_post_sheet_data()

        worksheets[[i]]$sheet_data$style_id <<-
          as.character(worksheets[[i]]$sheet_data$style_id)

        if ((length(rowHeights[[i]]) == 0) & (length(outlineLevels[[i]]) == 0)) {
          write_worksheet_xml(
            prior = prior,
            post = post,
            sheet_data = ws$sheet_data,
            R_fileName = file.path(xlworksheetsDir, sprintf("sheet%s.xml", i))
          )
        } else if ((length(rowHeights[[i]]) == 0) & (length(outlineLevels[[i]]) > 0)) {
          write_worksheet_xml_2(
            prior = prior,
            post = post,
            sheet_data = ws$sheet_data,
            row_heights_ = NULL,
            outline_levels_ = unlist(outlineLevels[[i]]),
            R_fileName = file.path(xlworksheetsDir, sprintf("sheet%s.xml", i))
          )
        } else if ((length(rowHeights[[i]]) > 0) & (length(outlineLevels[[i]]) == 0)) {
          write_worksheet_xml_2(
            prior = prior,
            post = post,
            sheet_data = ws$sheet_data,
            row_heights_ = unlist(rowHeights[[i]]),
            outline_levels_ = NULL,
            R_fileName = file.path(xlworksheetsDir, sprintf("sheet%s.xml", i))
          )
        } else {
          ## row heights will always be in order and all row heights are given rows in preSaveCleanup
          write_worksheet_xml_2(
            prior = prior,
            post = post,
            sheet_data = ws$sheet_data,
            row_heights_ = unlist(rowHeights[[i]]),
            outline_levels_ = unlist(outlineLevels[[i]]),
            R_fileName = file.path(xlworksheetsDir, sprintf("sheet%s.xml", i))
          )
        }

        worksheets[[i]]$sheet_data$style_id <<- integer(0)


        ## write worksheet rels
        if (length(worksheets_rels[[i]]) > 0) {
          ws_rels <- worksheets_rels[[i]]
          if (hasHL) {
            h_inds <- stri_join(seq_along(worksheets[[i]]$hyperlinks), "h")
            ws_rels <-
              c(ws_rels, unlist(
                lapply(seq_along(h_inds), function(j) {
                  worksheets[[i]]$hyperlinks[[j]]$to_target_xml(h_inds[j])
                })
              ))
          }

          ## Check if any tables were deleted - remove these from rels
          if (length(tables) > 0) {
            table_inds <- grep("tables/table[0-9].xml", ws_rels)

            if (length(table_inds) > 0) {
              ids <-
                regmatches(
                  ws_rels[table_inds],
                  regexpr(
                    '(?<=Relationship Id=")[0-9A-Za-z]+',
                    ws_rels[table_inds],
                    perl = TRUE
                  )
                )
              inds <-
                as.integer(gsub("[^0-9]", "", ids, perl = TRUE)) - 2L
              table_nms <- attr(tables, "tableName")[inds]
              is_deleted <-
                grepl("openxlsx_deleted", table_nms, fixed = TRUE)
              if (any(is_deleted)) {
                ws_rels <- ws_rels[-table_inds[is_deleted]]
              }
            }
          }



          write_file(
            head = '<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships">',
            body = pxml(ws_rels),
            tail = "</Relationships>",
            fl = file.path(xlworksheetsRelsDir, sprintf("sheet%s.xml.rels", i))
          )
        }
      } ## end of isChartSheet[i]
    } ## end of loop through 1:nSheets

    invisible(0)
  }
)





Workbook$methods(
  setRowHeights = function(sheet, rows, heights) {
    sheet <- validateSheet(sheet)

    ## remove any conflicting heights
    flag <- names(rowHeights[[sheet]]) %in% rows
    if (any(flag)) {
      rowHeights[[sheet]] <<- rowHeights[[sheet]][!flag]
    }

    nms <- c(names(rowHeights[[sheet]]), rows)
    allRowHeights <- unlist(c(rowHeights[[sheet]], heights))
    names(allRowHeights) <- nms

    allRowHeights <-
      allRowHeights[order(as.integer(names(allRowHeights)))]

    rowHeights[[sheet]] <<- allRowHeights
  }
)

Workbook$methods(
  groupColumns = function(sheet) {
    sheet <- validateSheet(sheet)

    hidden <- attr(colOutlineLevels[[sheet]], "hidden", exact = TRUE)
    cols <- names(colOutlineLevels[[sheet]])
    max_outline <- max(colOutlineLevels[[sheet]])
    
    outline_attr <- paste0(' outlineLevelCol="', max_outline, '"')
    if (!grepl("outlineLevelCol", worksheets[[sheet]]$sheetFormatPr)) {
      worksheets[[sheet]]$sheetFormatPr <<- sub("/>", paste0(outline_attr, "/>"), worksheets[[sheet]]$sheetFormatPr)
    } else {
      worksheets[[sheet]]$sheetFormatPr <<- sub(' outlineLevelCol="[0-9]+"', outline_attr, worksheets[[sheet]]$sheetFormatPr)
    }

    # Check if column is already created (by `setColWidths()` or on import)
    # Note that columns are initiated by `setColWidths` first (see: order of execution in `preSaveCleanUp()`)
    if (any(cols %in% names(worksheets[[sheet]]$cols))) {

      for (i in intersect(cols, names(worksheets[[sheet]]$cols))) {
        outline_hidden <- attr(colOutlineLevels[[sheet]], "hidden")[attr(colOutlineLevels[[sheet]], "names") == i]
        outline_level <- colOutlineLevels[[sheet]][[i]]

        if (grepl("outlineLevel", worksheets[[sheet]]$cols[[i]], perl = TRUE)) {
          worksheets[[sheet]]$cols[[i]] <<- sub("((?<=hidden=\")(\\w+)\")", paste0(outline_hidden, "\""), worksheets[[sheet]]$cols[[i]], perl = TRUE)
        } else {
          worksheets[[sheet]]$cols[[i]] <<- sub("((?<=hidden=\")(\\w+)\")", paste0(outline_hidden, "\" outlineLevel=\"", outline_level, "\""), worksheets[[sheet]]$cols[[i]], perl = TRUE)
        }
      }

      cols <- cols[!cols %in% names(worksheets[[sheet]]$cols)]
      hidden <- attr(colOutlineLevels[[sheet]], "hidden")[attr(colOutlineLevels[[sheet]], "names") %in% cols]
    }

    if (length(cols) > 0) {
      colNodes <- sprintf('<col min="%s" max="%s" outlineLevel="%s" hidden="%s"/>', cols, cols, colOutlineLevels[[sheet]][cols], hidden)
      names(colNodes) <- cols
      colNodes = append(worksheets[[sheet]]$cols, colNodes)
      # Order by column name (=index)
      worksheets[[sheet]]$cols <<- colNodes[order(names(colNodes))]
    }
  }
)

Workbook$methods(
  groupRows = function(sheet, rows, hidden = 0, levels = -1) {
    # Validation, input data cleanup / preparation
    sheet <- validateSheet(sheet)
    hidden = rep(hidden, length.out = length(rows))
    levels = rep(levels, length.out = length(rows))
    
    # browser()
    
    # disassemble outlineLevels (extract rows, levels and hidden components)
    existing_rows <- names(outlineLevels[[sheet]])
    existing_levels <- outlineLevels[[sheet]]
    attributes(existing_levels) <- NULL
    existing_hidden <-  attr(outlineLevels[[sheet]], "hidden")

    # 1. existing entries: 
    #       - if levels==-1 => set outlineLevel to max of existing level + 1
    #       - Otherwise, set outlineLevel to given levels
    #       - update hidden flag correspondingly
    # 2. New entries
    #       - Append new entries to the end (hidden and levels vector), potentially out-of-order
    # 3. Reorder all entries to be in the correct order
    
    flag <- existing_rows %in% rows

    # Find indices of rows that already exist
    existing_outline_indices = which(flag)
    existing_outline = existing_rows[existing_outline_indices]
    existing_rows_indices = match(existing_outline, rows)
    
    # Auto-detect new level if required
    new_level <- "1"
    if (any(flag)) {
      new_level <- as.character(max(as.numeric(existing_levels[flag])) + 1)
    }
    levels[levels < 0] = as.character(new_level)
    
    if (any(flag)) {
      # Assign the given values to existing row definitions (indices were extracted above)
      existing_hidden[existing_outline_indices] <- hidden[existing_rows_indices]
      existing_levels[existing_outline_indices] <- levels[existing_rows_indices]
      
      # Append all remaining new entries:
      all_names <- c(existing_rows, rows[-existing_rows_indices])
      all_levels <- c(existing_levels, levels[-existing_rows_indices])
      all_hidden <- c(existing_hidden, hidden[-existing_rows_indices])
    } else {
      # only new rows were added, no existing modified
      all_names = c(existing_rows, rows)
      all_levels = c(existing_levels, levels)
      all_hidden = c(existing_hidden, hidden)
    }
    
    # re-order and then re-assamble the outlineLevels object (vector with proper attributes)
    ord <- order(as.numeric(all_names))
    all_levels <- as.character(all_levels[ord])
    names(all_levels) <- all_names[ord]
    attr(all_levels, "hidden") <- as.character(as.integer(all_hidden[ord]))
    
    outlineLevels[[sheet]] <<- all_levels
    

    # Finally, update the sheetFormatPr XML element with the maximum outline level
    max_outline = max(as.numeric(outlineLevels[[sheet]]))
    outline_attr <- paste0(' outlineLevelRow="', max_outline, '"')
    if (!grepl("outlineLevelRow", worksheets[[sheet]]$sheetFormatPr)) {
      worksheets[[sheet]]$sheetFormatPr <<- sub("/>", paste0(outline_attr, "/>"), worksheets[[sheet]]$sheetFormatPr)
    } else {
      worksheets[[sheet]]$sheetFormatPr <<- sub(' outlineLevelRow="[0-9]+"', outline_attr, worksheets[[sheet]]$sheetFormatPr)
    }
  }
)



Workbook$methods(
  deleteWorksheet = function(sheet) {
    # To delete a worksheet
    # Remove colwidths element
    # Remove drawing partname from Content_Types (drawing(sheet).xml)
    # Remove highest sheet from Content_Types
    # Remove drawings element
    # Remove drawings_rels element

    # Remove vml element
    # Remove vml_rels element

    # Remove rowHeights element
    # Remove styleObjects on sheet
    # Remove last sheet element from workbook
    # Remove last sheet element from workbook.xml.rels
    # Remove element from worksheets
    # Remove element from worksheets_rels
    # Remove hyperlinks
    # Reduce calcChain i attributes & remove calcs on sheet
    # Remove sheet from sheetOrder
    # Remove queryTable references from workbook$definedNames to worksheet
    # remove tables

    sheet <- validateSheet(sheet)
    sheetNames <- sheet_names
    nSheets <- length(unlist(sheetNames, use.names = FALSE))
    sheetName <- sheetNames[[sheet]]

    colWidths[[sheet]] <<- NULL
    sheet_names <<- sheet_names[-sheet]

    ## remove last drawings(sheet).xml from Content_Types
    Content_Types <<-
      Content_Types[!grepl(sprintf("drawing%s.xml", nSheets), Content_Types)]

    ## remove highest sheet
    Content_Types <<-
      Content_Types[!grepl(sprintf("sheet%s.xml", nSheets), Content_Types)]

    drawings[[sheet]] <<- NULL
    drawings_rels[[sheet]] <<- NULL

    vml[[sheet]] <<- NULL
    vml_rels[[sheet]] <<- NULL

    rowHeights[[sheet]] <<- NULL
    colOutlineLevels[[sheet]] <<- NULL
    outlineLevels[[sheet]] <<- NULL
    comments[[sheet]] <<- NULL
    threadComments[[sheet]] <<- NULL
    isChartSheet <<- isChartSheet[-sheet]

    ## sheetOrder
    toRemove <- which(sheetOrder == sheet)
    sheetOrder[sheetOrder > sheet] <<-
      sheetOrder[sheetOrder > sheet] - 1L
    sheetOrder <<- sheetOrder[-toRemove]


    ## remove styleObjects
    if (length(styleObjects) > 0) {
      styleObjects <<-
        styleObjects[unlist(lapply(styleObjects, "[[", "sheet"), use.names = FALSE) != sheetName]
    }

    ## Need to remove reference from workbook.xml.rels to pivotCache
    removeRels <- grep("pivotTables", worksheets_rels[[sheet]], value = TRUE)
    if (length(removeRels) > 0) {
      ## sheet rels links to a pivotTable file, the corresponding pivotTable_rels file links to the cacheDefn which is listing in workbook.xml.rels
      ## remove reference to this file from the workbook.xml.rels
      fileNo <-
        as.integer(unlist(regmatches(
          removeRels,
          gregexpr("(?<=pivotTable)[0-9]+(?=\\.xml)", removeRels, perl = TRUE)
        )))
      toRemove <-
        stri_join(
          sprintf("(pivotCacheDefinition%s\\.xml)", fileNo),
          sep = " ",
          collapse = "|"
        )

      fileNo <- grep(toRemove, pivotTables.xml.rels)
      toRemove <-
        stri_join(
          sprintf("(pivotCacheDefinition%s\\.xml)", fileNo),
          sep = " ",
          collapse = "|"
        )

      ## remove reference to file from workbook.xml.res
      workbook.xml.rels <<-
        workbook.xml.rels[!grepl(toRemove, workbook.xml.rels)]
    }

    ## As above for slicers
    ## Need to remove reference from workbook.xml.rels to pivotCache
    removeRels <- grepl("slicers", worksheets_rels[[sheet]])
    if (any(removeRels)) {
      workbook.xml.rels <<-
        workbook.xml.rels[!grepl(sprintf("(slicerCache%s\\.xml)", sheet), workbook.xml.rels)]
    }

    ## wont't remove tables and then won't need to reassign table r:id's but will rename them!
    worksheets[[sheet]] <<- NULL
    worksheets_rels[[sheet]] <<- NULL

    if (length(tables) > 0) {
      tableSheets <- attr(tables, "sheet")
      tableNames <- attr(tables, "tableName")

      inds <-
        tableSheets %in% sheet &
          !grepl("openxlsx_deleted", attr(tables, "tableName"), fixed = TRUE)
      tableSheets[tableSheets > sheet] <-
        tableSheets[tableSheets > sheet] - 1L

      ## Need to flag a table as deleted
      if (any(inds)) {
        tableSheets[inds] <- 0
        tableNames[inds] <-
          stri_join(tableNames[inds], "_openxlsx_deleted")
      }
      attr(tables, "tableName") <<- tableNames
      attr(tables, "sheet") <<- tableSheets
    }


    ## drawing will always be the first relationship and printerSettings second
    if (nSheets > 1) {
      for (i in 1:(nSheets - 1L)) {
        worksheets_rels[[i]][1:3] <<- genBaseSheetRels(i)
      }
    } else {
      worksheets_rels <<- list()
    }


    ## remove sheet
    sn <-
      unlist(lapply(workbook$sheets, function(x) {
        regmatches(
          x, regexpr('(?<= name=")[^"]+', x, perl = TRUE)
        )
      }))
    workbook$sheets <<- workbook$sheets[!sn %in% sheetName]

    ## Reset rIds
    if (nSheets > 1) {
      for (i in (sheet + 1L):nSheets) {
        workbook$sheets <<-
          gsub(stri_join("rId", i),
            stri_join("rId", i - 1L),
            workbook$sheets,
            fixed = TRUE
          )
      }
    } else {
      workbook$sheets <<- NULL
    }

    ## Can remove highest sheet
    workbook.xml.rels <<-
      workbook.xml.rels[!grepl(sprintf("sheet%s.xml", nSheets), workbook.xml.rels)]

    ## definedNames
    if (length(workbook$definedNames) > 0) {
      belongTo <- getDefinedNamesSheet(workbook$definedNames)
      workbook$definedNames <<-
        workbook$definedNames[!belongTo %in% sheetName]
    }

    invisible(1)
  }
)


Workbook$methods(
  addDXFS = function(style) {
    dxf <- "<dxf>"
    dxf <- stri_join(dxf, createFontNode(style))
    fillNode <- NULL

    if (!is.null(style$fill$fillFg) | !is.null(style$fill$fillBg)) {
      dxf <- stri_join(dxf, createFillNode(style))
    }

    if (any(!is.null(
      c(
        style$borderLeft,
        style$borderRight,
        style$borderTop,
        style$borderBottom,
        style$borderDiagonal
      )
    ))) {
      dxf <- stri_join(dxf, createBorderNode(style))
    }

    dxf <- stri_join(dxf, "</dxf>", sep = " ")
    if (dxf %in% styles$dxfs) {
      return(which(styles$dxfs == dxf) - 1L)
    }

    dxfId <- length(styles$dxfs)
    styles$dxfs <<- c(styles$dxfs, dxf)

    return(dxfId)
  }
)



Workbook$methods(
  dataValidation = function(sheet,
                            startRow,
                            endRow,
                            startCol,
                            endCol,
                            type,
                            operator,
                            value,
                            allowBlank,
                            showInputMsg,
                            showErrorMsg) {
    sheet <- validateSheet(sheet)
    sqref <-
      stri_join(getCellRefs(data.frame(
        "x" = c(startRow, endRow),
        "y" = c(startCol, endCol)
      )),
      sep = " ",
      collapse = ":"
      )

    header <-
      sprintf(
        '<dataValidation type="%s" operator="%s" allowBlank="%s" showInputMessage="%s" showErrorMessage="%s" sqref="%s">',
        type,
        operator,
        allowBlank,
        showInputMsg,
        showErrorMsg,
        sqref
      )


    if (type == "date") {
      origin <- 25569L
      if (grepl(
        'date1904="1"|date1904="true"',
        stri_join(unlist(workbook), sep = " ", collapse = ""),
        ignore.case = TRUE
      )) {
        origin <- 24107L
      }

      value <- as.integer(value) + origin
    }

    if (type == "time") {
      origin <- 25569L
      if (grepl(
        'date1904="1"|date1904="true"',
        stri_join(unlist(workbook), sep = " ", collapse = ""),
        ignore.case = TRUE
      )) {
        origin <- 24107L
      }

      t <- format(value[1], "%z")
      offSet <-
        suppressWarnings(ifelse(substr(t, 1, 1) == "+", 1L, -1L) * (as.integer(substr(t, 2, 3)) + as.integer(substr(t, 4, 5)) / 60) / 24)
      if (is.na(offSet)) {
        offSet[i] <- 0
      }

      value <- as.numeric(as.POSIXct(value)) / 86400 + origin + offSet
    }

    form <-
      sapply(seq_along(value), function(i) {
        sprintf("<formula%s>%s</formula%s>", i, value[i], i)
      })
    worksheets[[sheet]]$dataValidations <<-
      c(
        worksheets[[sheet]]$dataValidations,
        stri_join(header, stri_join(form, collapse = ""), "</dataValidation>")
      )

    invisible(0)
  }
)



Workbook$methods(
  dataValidation_list = function(sheet,
                                 startRow,
                                 endRow,
                                 startCol,
                                 endCol,
                                 value,
                                 allowBlank,
                                 showInputMsg,
                                 showErrorMsg) {
    sheet <- validateSheet(sheet)
    sqref <-
      stri_join(getCellRefs(data.frame(
        "x" = c(startRow, endRow),
        "y" = c(startCol, endCol)
      )),
      sep = " ",
      collapse = ":"
      )
    data_val <-
      sprintf(
        '<x14:dataValidation type="list" allowBlank="%s" showInputMessage="%s" showErrorMessage="%s">',
        allowBlank,
        showInputMsg,
        showErrorMsg
      )

    formula <-
      sprintf("<x14:formula1><xm:f>%s</xm:f></x14:formula1>", value)
    sqref <- sprintf("<xm:sqref>%s</xm:sqref>", sqref)

    xmlData <-
      stri_join(data_val, formula, sqref, "</x14:dataValidation>")

    worksheets[[sheet]]$dataValidationsLst <<-
      c(worksheets[[sheet]]$dataValidationsLst, xmlData)

    invisible(0)
  }
)



Workbook$methods(
  conditionalFormatting = function(sheet,
                                   startRow,
                                   endRow,
                                   startCol,
                                   endCol,
                                   dxfId,
                                   formula,
                                   type,
                                   values,
                                   params) {
    sheet <- validateSheet(sheet)
    sqref <-
      stri_join(getCellRefs(data.frame(
        "x" = c(startRow, endRow),
        "y" = c(startCol, endCol)
      )), collapse = ":")



    ## Increment priority of conditional formatting rule
    if (length(worksheets[[sheet]]$conditionalFormatting) > 0) {
      for (i in rev(seq_along(worksheets[[sheet]]$conditionalFormatting))) {
        priority <-
          regmatches(
            worksheets[[sheet]]$conditionalFormatting[[i]],
            regexpr(
              '(?<=priority=")[0-9]+',
              worksheets[[sheet]]$conditionalFormatting[[i]],
              perl = TRUE
            )
          )
        priority_new <- as.integer(priority) + 1L

        priority_pattern <- sprintf('priority="%s"', priority)
        priority_new <- sprintf('priority="%s"', priority_new)

        ## now replace
        worksheets[[sheet]]$conditionalFormatting[[i]] <<-
          gsub(priority_pattern,
               priority_new,
               worksheets[[sheet]]$conditionalFormatting[[i]],
               fixed = TRUE
          )
      }
    }

    nms <- c(names(worksheets[[sheet]]$conditionalFormatting), sqref)

    if (type == "colorScale") {
      ## formula contains the colours
      ## values contains numerics or is NULL
      ## dxfId is ignored

      if (is.null(values)) {
        if (length(formula) == 2L) {
          cfRule <-
            sprintf(
              '<cfRule type="colorScale" priority="1"><colorScale>
                             <cfvo type="min"/><cfvo type="max"/>
                             <color rgb="%s"/><color rgb="%s"/>
                           </colorScale></cfRule>',
              formula[[1]],
              formula[[2]]
            )
        } else {
          cfRule <-
            sprintf(
              '<cfRule type="colorScale" priority="1"><colorScale>
                             <cfvo type="min"/><cfvo type="percentile" val="50"/><cfvo type="max"/>
                             <color rgb="%s"/><color rgb="%s"/><color rgb="%s"/>
                           </colorScale></cfRule>',
              formula[[1]],
              formula[[2]],
              formula[[3]]
            )
        }
      } else {
        if (length(formula) == 2L) {
          cfRule <-
            sprintf(
              '<cfRule type="colorScale" priority="1"><colorScale>
                            <cfvo type="num" val="%s"/><cfvo type="num" val="%s"/>
                            <color rgb="%s"/><color rgb="%s"/>
                           </colorScale></cfRule>',
              values[[1]],
              values[[2]],
              formula[[1]],
              formula[[2]]
            )
        } else {
          cfRule <-
            sprintf(
              '<cfRule type="colorScale" priority="1"><colorScale>
                            <cfvo type="num" val="%s"/><cfvo type="num" val="%s"/><cfvo type="num" val="%s"/>
                            <color rgb="%s"/><color rgb="%s"/><color rgb="%s"/>
                           </colorScale></cfRule>',
              values[[1]],
              values[[2]],
              values[[3]],
              formula[[1]],
              formula[[2]],
              formula[[3]]
            )
        }
      }
    } else if (type == "dataBar") {
      # forumula is a vector of colours of length 1 or 2
      # values is NULL or a numeric vector of equal length as formula

      if (length(formula) == 2L) {
        negColour <- formula[[1]]
        posColour <- formula[[2]]
      } else {
        posColour <- formula
        negColour <- "FFFF0000"
      }

      guid <-
        stri_join(
          "F7189283-14F7-4DE0-9601-54DE9DB",
          40000L + length(worksheets[[sheet]]$extLst)
        )

      showValue <- 1
      if ("showValue" %in% names(params)) {
        showValue <- as.integer(params$showValue)
      }

      gradient <- 1
      if ("gradient" %in% names(params)) {
        gradient <- as.integer(params$gradient)
      }

      border <- 1
      if ("border" %in% names(params)) {
        border <- as.integer(params$border)
      }

      if (is.null(values)) {
        cfRule <-
          sprintf(
            '<cfRule type="dataBar" priority="1"><dataBar showValue="%s">
                          <cfvo type="min"/><cfvo type="max"/>
                          <color rgb="%s"/>
                          </dataBar>
                          <extLst><ext uri="{B025F937-C7B1-47D3-B67F-A62EFF666E3E}" xmlns:x14="http://schemas.microsoft.com/office/spreadsheetml/2009/9/main"><x14:id>{%s}</x14:id></ext>
                        </extLst></cfRule>',
            showValue,
            posColour,
            guid
          )
      } else {
        cfRule <-
          sprintf(
            '<cfRule type="dataBar" priority="1"><dataBar showValue="%s">
                            <cfvo type="num" val="%s"/><cfvo type="num" val="%s"/>
                            <color rgb="%s"/>
                            </dataBar>
                            <extLst><ext uri="{B025F937-C7B1-47D3-B67F-A62EFF666E3E}" xmlns:x14="http://schemas.microsoft.com/office/spreadsheetml/2009/9/main">
                            <x14:id>{%s}</x14:id></ext></extLst></cfRule>',
            showValue,
            values[[1]],
            values[[2]],
            posColour,
            guid
          )
      }

      worksheets[[sheet]]$extLst <<-
        c(
          worksheets[[sheet]]$extLst,
          gen_databar_extlst(
            guid = guid,
            sqref = sqref,
            posColour = posColour,
            negColour = negColour,
            values = values,
            border = border,
            gradient = gradient
          )
        )
    } else if (type == "expression") {
      cfRule <-
        sprintf(
          '<cfRule type="expression" dxfId="%s" priority="1"><formula>%s</formula></cfRule>',
          dxfId,
          formula
        )
    } else if (type == "duplicatedValues") {
      cfRule <-
        sprintf(
          '<cfRule type="duplicateValues" dxfId="%s" priority="1"/>',
          dxfId
        )
    } else if (type == "containsText") {
      cfRule <-
        sprintf(
          '<cfRule type="containsText" dxfId="%s" priority="1" operator="containsText" text="%s">
                        	<formula>NOT(ISERROR(SEARCH("%s", %s)))</formula>
                       </cfRule>',
          dxfId,
          values,
          values,
          unlist(strsplit(sqref, split = ":"))[[1]]
        )
    } else if (type == "notContainsText") {
      cfRule <-
        sprintf(
          '<cfRule type="notContainsText" dxfId="%s" priority="1" operator="notContains" text="%s">
                        	<formula>ISERROR(SEARCH("%s", %s))</formula>
                       </cfRule>',
          dxfId,
          values,
          values,
          unlist(strsplit(sqref, split = ":"))[[1]]
        )
    } else if (type == "beginsWith") {
      cfRule <-
        sprintf(
          '<cfRule type="beginsWith" dxfId="%s" priority="1" operator="beginsWith" text="%s">
                        	<formula>LEFT(%s,LEN("%s"))="%s"</formula>
                       </cfRule>',
          dxfId,
          values,

          unlist(strsplit(sqref, split = ":"))[[1]],
          values,
          values
        )
    } else if (type == "endsWith") {
      cfRule <-
        sprintf(
          '<cfRule type="endsWith" dxfId="%s" priority="1" operator="endsWith" text="%s">
                        	<formula>RIGHT(%s,LEN("%s"))="%s"</formula>
                       </cfRule>',
          dxfId,
          values,

          unlist(strsplit(sqref, split = ":"))[[1]],
          values,
          values
        )
    } else if (type == "between") {
      cfRule <-
        sprintf(
          '<cfRule type="cellIs" dxfId="%s" priority="1" operator="between"><formula>%s</formula><formula>%s</formula></cfRule>',
          dxfId,
          formula[1],
          formula[2]
        )
    } else if (type == "topN") {
      cfRule <-
        sprintf(
          '<cfRule type="top10" dxfId="%s" priority="1" rank="%s" percent="%s"></cfRule>',
          dxfId,
          values[1],
          values[2]
        )
    } else if (type == "bottomN") {
      cfRule <-
        sprintf(
          '<cfRule type="top10" dxfId="%s" priority="1" rank="%s" percent="%s" bottom="1"></cfRule>',
          dxfId,
          values[1],
          values[2]
        )
    } else if (type == "containsBlanks") {
      cfRule <-
        sprintf(
          '<cfRule type="containsBlanks" dxfId="%s" priority="1">
                        	<formula>LEN(TRIM(%s))=0</formula>
                       </cfRule>',
          dxfId,
          unlist(strsplit(sqref, split = ":"))[[1]]
        )
    } else if (type == "notContainsBlanks") {
      cfRule <-
        sprintf(
          '<cfRule type="notContainsBlanks" dxfId="%s" priority="1">
                        	<formula>LEN(TRIM(%s))&gt;0</formula>
                       </cfRule>',
          dxfId,
          unlist(strsplit(sqref, split = ":"))[[1]]
        )
    }

    worksheets[[sheet]]$conditionalFormatting <<-
      append(worksheets[[sheet]]$conditionalFormatting, cfRule)

    names(worksheets[[sheet]]$conditionalFormatting) <<- nms

    invisible(0)
  }
)




Workbook$methods(
  mergeCells = function(sheet, startRow, endRow, startCol, endCol) {
    sheet <- validateSheet(sheetName = sheet)

    sqref <-
      getCellRefs(data.frame(
        "x" = c(startRow, endRow),
        "y" = c(startCol, endCol)
      ))
    exMerges <-
      regmatches(
        worksheets[[sheet]]$mergeCells,
        regexpr("[A-Z0-9]+:[A-Z0-9]+", worksheets[[sheet]]$mergeCells)
      )

    if (!is.null(exMerges)) {
      comps <-
        lapply(exMerges, function(rectCoords) {
          unlist(strsplit(rectCoords, split = ":"))
        })
      exMergedCells <- build_cell_merges(comps = comps)
      newMerge <- unlist(build_cell_merges(comps = list(sqref)))

      ## Error if merge intersects
      mergeIntersections <-
        sapply(exMergedCells, function(x) {
          any(x %in% newMerge)
        })
      if (any(mergeIntersections)) {
        stop(
          sprintf(
            "Merge intersects with existing merged cells: \n\t\t%s.\nRemove existing merge first.",
            stri_join(exMerges[mergeIntersections], collapse = "\n\t\t")
          )
        )
      }
    }

    worksheets[[sheet]]$mergeCells <<-
      c(
        worksheets[[sheet]]$mergeCells,
        sprintf(
          '<mergeCell ref="%s"/>',
          stri_join(sqref,
            collapse = ":", sep =
              " "
          )
        )
      )
  }
)



Workbook$methods(
  removeCellMerge = function(sheet, startRow, endRow, startCol, endCol) {
    sheet <- validateSheet(sheet)

    sqref <-
      getCellRefs(data.frame(
        "x" = c(startRow, endRow),
        "y" = c(startCol, endCol)
      ))
    exMerges <-
      regmatches(
        worksheets[[sheet]]$mergeCells,
        regexpr("[A-Z0-9]+:[A-Z0-9]+", worksheets[[sheet]]$mergeCells)
      )

    if (!is.null(exMerges)) {
      comps <-
        lapply(exMerges, function(x) {
          unlist(strsplit(x, split = ":"))
        })
      exMergedCells <- build_cell_merges(comps = comps)
      newMerge <- unlist(build_cell_merges(comps = list(sqref)))

      ## Error if merge intersects
      mergeIntersections <-
        sapply(exMergedCells, function(x) {
          any(x %in% newMerge)
        })
    }

    ## Remove intersection
    worksheets[[sheet]]$mergeCells <<-
      worksheets[[sheet]]$mergeCells[!mergeIntersections]
  }
)





Workbook$methods(
  freezePanes = function(sheet,
                         firstActiveRow = NULL,
                         firstActiveCol = NULL,
                         firstRow = FALSE,
                         firstCol = FALSE) {
    sheet <- validateSheet(sheet)
    paneNode <- NULL

    if (firstRow) {
      paneNode <-
        '<pane ySplit="1" topLeftCell="A2" activePane="bottomLeft" state="frozen"/>'
    } else if (firstCol) {
      paneNode <-
        '<pane xSplit="1" topLeftCell="B1" activePane="topRight" state="frozen"/>'
    }


    if (is.null(paneNode)) {
      if (firstActiveRow == 1 & firstActiveCol == 1) {
        ## nothing to do
        return(NULL)
      }

      if (firstActiveRow > 1 & firstActiveCol == 1) {
        attrs <- sprintf('ySplit="%s"', firstActiveRow - 1L)
        activePane <- "bottomLeft"
      }

      if (firstActiveRow == 1 & firstActiveCol > 1) {
        attrs <- sprintf('xSplit="%s"', firstActiveCol - 1L)
        activePane <- "topRight"
      }

      if (firstActiveRow > 1 & firstActiveCol > 1) {
        attrs <-
          sprintf(
            'ySplit="%s" xSplit="%s"',
            firstActiveRow - 1L,
            firstActiveCol - 1L
          )
        activePane <- "bottomRight"
      }

      topLeftCell <-
        getCellRefs(data.frame(firstActiveRow, firstActiveCol))

      paneNode <-
        sprintf(
          '<pane %s topLeftCell="%s" activePane="%s" state="frozen"/><selection pane="%s"/>',
          stri_join(attrs, collapse = " ", sep = " "),
          topLeftCell,
          activePane,
          activePane
        )
    }

    worksheets[[sheet]]$freezePane <<- paneNode
  }
)



Workbook$methods(
  insertImage = function(sheet,
                         file,
                         startRow,
                         startCol,
                         width,
                         height,
                         rowOffset = 0,
                         colOffset = 0,
                         address) {
    ## within the sheet the drawing node's Id refernce an id in the sheetRels
    ## sheet rels reference the drawingi.xml file
    ## drawingi.xml reference drawingRels
    ## drawing rels reference an image in the media folder
    ## worksheetRels(sheet(i)) references drawings(j)

    sheet <- validateSheet(sheet)

    imageType <- regmatches(file, gregexpr("\\.[a-zA-Z]*$", file))
    imageType <- gsub("^\\.", "", imageType)

    imageNo <- length((drawings[[sheet]])) + 1L
    imageRelNo <- length((drawings_rels[[sheet]])) + 1L
    linkRelNo <- length((drawings_rels[[sheet]])) + 2L
    mediaNo <- length(media) + 1L

    startCol <- convertFromExcelRef(startCol)

    ## update Content_Types
    if (!any(grepl(stri_join("image/", imageType), Content_Types))) {
      Content_Types <<-
        unique(c(
          sprintf(
            '<Default Extension="%s" ContentType="image/%s"/>',
            imageType,
            imageType
          ),
          Content_Types
        ))
    }

    ## drawings rels (Reference from drawings.xml to image file in media folder)
    drawings_rels[[sheet]] <<- c(
      drawings_rels[[sheet]],
      sprintf(
        '<Relationship Id="rId%s" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/image" Target="../media/image%s.%s"/>',
        imageRelNo,
        mediaNo,
        imageType
      ), 
      if (!missing(address)) {
        sprintf(
          '<Relationship Id="rId%s" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/hyperlink" Target="%s" TargetMode="External"/>',
          linkRelNo,
          address
        )
      }
    )

    ## write file path to media slot to copy across on save
    tmp <- file
    names(tmp) <- stri_join("image", mediaNo, ".", imageType)
    media <<- append(media, tmp)

    ## create drawing.xml
    anchor <-
      '<xdr:oneCellAnchor xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing">'

    from <- sprintf(
      '<xdr:from xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing">
    <xdr:col xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing">%s</xdr:col>
    <xdr:colOff xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing">%s</xdr:colOff>
    <xdr:row xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing">%s</xdr:row>
    <xdr:rowOff xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing">%s</xdr:rowOff>
  </xdr:from>',
      startCol - 1L,
      colOffset,
      startRow - 1L,
      rowOffset
    )

    drawingsXML <- stri_join(
      anchor,
      from,
      sprintf(
        '<xdr:ext xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing" cx="%s" cy="%s"/>',
        width,
        height
      ),
      genBasePic(
        imageNo, 
        imageRelNo, 
        ifelse(
          missing(address),
          "/",
          sprintf(
            '>
              <a:hlinkClick xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" r:id="rId%s"/>
            </xdr:cNvPr',
            linkRelNo
          )
        )
      ),
      "<xdr:clientData/>",
      "</xdr:oneCellAnchor>"
    )


    ## append to workbook drawing
    drawings[[sheet]] <<- c(drawings[[sheet]], drawingsXML)
  }
)



Workbook$methods(
  preSaveCleanUp = function() {
    ## Steps
    # Order workbook.xml.rels:
    #   sheets -> style -> theme -> sharedStrings -> persons -> tables -> calcChain
    # Assign workbook.xml.rels children rIds, seq_along(workbook.xml.rels)
    # Assign workbook$sheets rIds 1:nSheets
    #
    ## drawings will always be r:id1 on worksheet
    ## tables will always have r:id equal to table xml file number tables/table(i).xml

    ## Every worksheet has a drawingXML as r:id 1
    ## Every worksheet has a printerSettings as r:id 2
    ## Tables from r:id 3 to nTables+3 - 1
    ## HyperLinks from nTables+3 to nTables+3+nHyperLinks-1
    ## vmlDrawing to have rId

    sheetRIds <-
      as.integer(unlist(regmatches(
        workbook$sheets,
        gregexpr('(?<=r:id="rId)[0-9]+', workbook$sheets, perl = TRUE)
      )))

    nSheets <- length(sheetRIds)
    nExtRefs <- length(externalLinks)
    nPivots <- length(pivotDefinitions)

    ## add a worksheet if none added
    if (nSheets == 0) {
      warning("Workbook does not contain any worksheets. A worksheet will be added.",
        call. = FALSE
      )
      .self$addWorksheet("Sheet 1")
      nSheets <- 1L
    }

    ## get index of each child element for ordering
    sheetInds <- grep("(worksheets|chartsheets)/sheet[0-9]+\\.xml", workbook.xml.rels)
    stylesInd <- grep("styles\\.xml", workbook.xml.rels)
    themeInd <- grep("theme/theme[0-9]+.xml", workbook.xml.rels)
    connectionsInd <- grep("connections.xml", workbook.xml.rels)
    extRefInds <- grep("externalLinks/externalLink[0-9]+.xml", workbook.xml.rels)
    sharedStringsInd <- grep("sharedStrings.xml", workbook.xml.rels)
    tableInds <- grep("table[0-9]+.xml", workbook.xml.rels)
    personInds <- grep("person.xml", workbook.xml.rels)


    ## Reordering of workbook.xml.rels
    ## don't want to re-assign rIds for pivot tables or slicer caches
    pivotNode <- grep("pivotCache/pivotCacheDefinition[0-9].xml", workbook.xml.rels, value = TRUE)
    slicerNode <- grep("slicerCache[0-9]+.xml", workbook.xml.rels, value = TRUE)

    ## Reorder children of workbook.xml.rels
    workbook.xml.rels <<-
      workbook.xml.rels[c(
        sheetInds,
        extRefInds,
        themeInd,
        connectionsInd,
        stylesInd,
        sharedStringsInd,
        tableInds,
        personInds
      )]

    ## Re assign rIds to children of workbook.xml.rels
    workbook.xml.rels <<-
      unlist(lapply(seq_along(workbook.xml.rels), function(i) {
        gsub('(?<=Relationship Id="rId)[0-9]+',
          i,
          workbook.xml.rels[[i]],
          perl = TRUE
        )
      }))

    workbook.xml.rels <<- c(workbook.xml.rels, pivotNode, slicerNode)



    if (!is.null(vbaProject)) {
      workbook.xml.rels <<-
        c(
          workbook.xml.rels,
          sprintf(
            '<Relationship Id="rId%s" Type="http://schemas.microsoft.com/office/2006/relationships/vbaProject" Target="vbaProject.bin"/>',
            1L + length(workbook.xml.rels)
          )
        )
    }

    ## Reassign rId to workbook sheet elements, (order sheets by sheetId first)
    workbook$sheets <<-
      unlist(lapply(seq_along(workbook$sheets), function(i) {
        gsub('(?<= r:id="rId)[0-9]+', i, workbook$sheets[[i]], perl = TRUE)
      }))

    ## re-order worksheets if need to
    if (any(sheetOrder != seq_len(nSheets))) {
      workbook$sheets <<- workbook$sheets[sheetOrder]
    }

    ## preserve window size and position on save:
    bookViews <-workbook$bookViews
    xWindow <- getAttrs(bookViews, "xWindow")$xWindow
    yWindow <- getAttrs(bookViews, "yWindow")$yWindow
    windowWidth <- getAttrs(bookViews, "windowWidth")$windowWidth
    windowHeight <- getAttrs(bookViews, "windowHeight")$windowHeight


    ## re-assign tabSelected
    state <- rep.int("visible", nSheets)
    state[grepl("hidden", workbook$sheets)] <- "hidden"
    visible_sheet_index <- which(state %in% "visible")[[1]]
    visible_sheets <- which(state %in% "visible")
    workbook$bookViews <<-
      sprintf(
        '<bookViews><workbookView xWindow="%s" yWindow="%s" windowWidth="%s" windowHeight="%s" firstSheet="%s" activeTab="%s"/></bookViews>',
        xWindow, yWindow, windowWidth, windowHeight,
        visible_sheet_index - 1L,
        ActiveSheet - 1L
      )

    for(i in seq_len(nSheets)) {
      worksheets[[i]]$sheetViews <<-
        sub(
          ' tabSelected="(1|true|false|0)"',
          ifelse(
            sheetOrder[ActiveSheet] == i,
            ' tabSelected="true"',
            ' tabSelected="false"'
          ),
          worksheets[[i]]$sheetViews,
          ignore.case = TRUE
        )
    }
    # worksheets[[visible_sheet_index]]$sheetViews

    # worksheets[[visible_sheet_index]]$sheetViews <<-
    #   sub(
    #     '( tabSelected="0")|( tabSelected="false")',
    #     ' tabSelected="1"',
    #     worksheets[[visible_sheet_index]]$sheetViews,
    #     ignore.case = TRUE
    #   )
    # if (nSheets > 1) {
    #   for (i in (1:nSheets)[!(1:nSheets) %in% visible_sheet_index]) {
    #     worksheets[[i]]$sheetViews <<-
    #       sub(
    #         ' tabSelected="(1|true|false|0)"',
    #         ' tabSelected="false"',
    #         worksheets[[i]]$sheetViews,
    #         ignore.case = TRUE
    #       )
    #   }
    # }





    if (length(workbook$definedNames) > 0) {
      sheetNames <- sheet_names[sheetOrder]

      belongTo <- getDefinedNamesSheet(workbook$definedNames)

      ## sheetNames is in re-ordered order (order it will be displayed)
      newId <- match(belongTo, sheetNames) - 1L
      oldId <-
        as.numeric(regmatches(
          workbook$definedNames,
          regexpr(
            '(?<= localSheetId=")[0-9]+',
            workbook$definedNames,
            perl = TRUE
          )
        ))

      for (i in seq_along(workbook$definedNames)) {
        if (!is.na(newId[i])) {
          workbook$definedNames[[i]] <<-
            gsub(
              sprintf('localSheetId=\"%s\"', oldId[i]),
              sprintf('localSheetId=\"%s\"', newId[i]),
              workbook$definedNames[[i]],
              fixed = TRUE
            )
        }
      }
    }




    ## update workbook r:id to match reordered workbook.xml.rels externalLink element
    if (length(extRefInds) > 0) {
      newInds <- as.integer(seq_along(extRefInds) + length(sheetInds))
      workbook$externalReferences <<-
        stri_join(
          "<externalReferences>",
          stri_join(
            sprintf('<externalReference r:id=\"rId%s\"/>', newInds),
            collapse = ""
          ),
          "</externalReferences>"
        )
    }

    ## styles
    numFmtIds <- 50000L
    for (i in which(!isChartSheet)) {
      worksheets[[i]]$sheet_data$style_id <<-
        rep.int(x = as.integer(NA), times = worksheets[[i]]$sheet_data$n_elements)
    }


    prev_sheet <- 0L
    for (x in styleObjects) {
      if (length(x$rows) > 0 & length(x$cols) > 0) {
        this.sty <- x$style$copy()

        if (!is.null(this.sty$numFmt)) {
          if (this.sty$numFmt$numFmtId == 9999) {
            this.sty$numFmt$numFmtId <- numFmtIds
            numFmtIds <- numFmtIds + 1L
          }
        }


        ## convert sheet name to index
        sheet <- which(sheet_names == x$sheet)
        sId <-
          .self$updateStyles(this.sty) ## this creates the XML for styles.XML

        cells_to_style <- pair_rc(x$rows, x$cols)
        
        # Avoid recreating this if we're looking at the same sheet over and over
        if (sheet != prev_sheet) {
          existing_cells <- pair_rc(worksheets[[sheet]]$sheet_data$rows,
                                    worksheets[[sheet]]$sheet_data$cols)  
        }

        ## In here we create any style_ids that don't yet exist in sheet_data
        worksheets[[sheet]]$sheet_data$style_id[existing_cells %in% cells_to_style] <<-
          sId


        new_cells_to_append <-
          which(!cells_to_style %in% existing_cells)
        if (length(new_cells_to_append) > 0) {
          worksheets[[sheet]]$sheet_data$style_id <<-
            c(
              worksheets[[sheet]]$sheet_data$style_id,
              rep.int(x = sId, times = length(new_cells_to_append))
            )

          worksheets[[sheet]]$sheet_data$rows <<-
            c(worksheets[[sheet]]$sheet_data$rows, x$rows[new_cells_to_append])
          worksheets[[sheet]]$sheet_data$cols <<-
            c(worksheets[[sheet]]$sheet_data$cols, x$cols[new_cells_to_append])
          worksheets[[sheet]]$sheet_data$t <<-
            c(worksheets[[sheet]]$sheet_data$t, rep(as.integer(NA), length(new_cells_to_append)))
          worksheets[[sheet]]$sheet_data$v <<-
            c(
              worksheets[[sheet]]$sheet_data$v,
              rep(as.character(NA), length(new_cells_to_append))
            )
          worksheets[[sheet]]$sheet_data$f <<-
            c(
              worksheets[[sheet]]$sheet_data$f,
              rep(as.character(NA), length(new_cells_to_append))
            )
          worksheets[[sheet]]$sheet_data$data_count <<-
            worksheets[[sheet]]$sheet_data$data_count + 1L

          worksheets[[sheet]]$sheet_data$n_elements <<-
            as.integer(length(worksheets[[sheet]]$sheet_data$rows))
        }
      }
    }


    ## Make sure all rowHeights have rows, if not append them!
    for (i in seq_along(worksheets)) {
      if (length(rowHeights[[i]]) > 0) {
        rh <- as.integer(names(rowHeights[[i]]))
        missing_rows <- rh[!rh %in% worksheets[[i]]$sheet_data$rows]
        n <- length(missing_rows)

        if (n > 0) {
          worksheets[[i]]$sheet_data$style_id <<-
            c(
              worksheets[[i]]$sheet_data$style_id,
              rep.int(as.integer(NA), times = n)
            )

          worksheets[[i]]$sheet_data$rows <<-
            c(worksheets[[i]]$sheet_data$rows, missing_rows)
          worksheets[[i]]$sheet_data$cols <<-
            c(
              worksheets[[i]]$sheet_data$cols,
              rep.int(as.integer(NA), times = n)
            )

          worksheets[[i]]$sheet_data$t <<-
            c(worksheets[[i]]$sheet_data$t, rep(as.integer(NA), times = n))
          worksheets[[i]]$sheet_data$v <<-
            c(
              worksheets[[i]]$sheet_data$v,
              rep(as.character(NA), times = n)
            )
          worksheets[[i]]$sheet_data$f <<-
            c(
              worksheets[[i]]$sheet_data$f,
              rep(as.character(NA), times = n)
            )
          worksheets[[i]]$sheet_data$data_count <<-
            worksheets[[i]]$sheet_data$data_count + 1L

          worksheets[[i]]$sheet_data$n_elements <<-
            as.integer(length(worksheets[[i]]$sheet_data$rows))
        }
      }

      ## write colwidth and coloutline XML
      if (length(colWidths[[i]]) > 0) {
        invisible(.self$setColWidths(i))
      }

      if (length(colOutlineLevels[[i]]) > 0) {
        invisible(.self$groupColumns(i))
      }


      if(ActiveSheet==i) {
        worksheets[[sheetOrder[i]]]$sheetViews <<-
          stri_replace_all_regex(
            worksheets[[sheetOrder[i]]]$sheetViews,
            "tabSelected=\"(1|true|false|0)\"",
            paste0("tabSelected=\"true\"")
          )
      } else {
        worksheets[[sheetOrder[i]]]$sheetViews <<-
          stri_replace_all_regex(
            worksheets[[sheetOrder[i]]]$sheetViews,
            "tabSelected=\"(1|true|false|0)\"",
            paste0("tabSelected=\"false\"")
          )
      }
    }
  }
)



Workbook$methods(
  addStyle = function(sheet, style, rows, cols, stack) {
    sheet <- sheet_names[[sheet]]

    if (length(styleObjects) == 0) {
      styleObjects <<- list(list(
        style = style,
        sheet = sheet,
        rows = rows,
        cols = cols
      ))
    } else if (stack) {
      nStyles <- length(styleObjects)

      ## ********** Assume all styleObjects cells have one a single worksheet **********
      ## Loop through existing styleObjects
      newInds <- seq_along(rows)
      keepStyle <- rep(TRUE, nStyles)
      for (i in 1:nStyles) {
        if (sheet == styleObjects[[i]]$sheet) {
          ## Now check rows and cols intersect
          ## toRemove are the elements that the new style doesn't apply to, we remove these from the style object as it
          ## is copied, merged with the new style and given the new data points

          ex_row_cols  <- pair_rc(styleObjects[[i]]$rows, 
                                  styleObjects[[i]]$cols)
          new_row_cols <- pair_rc(rows, cols)


          ## mergeInds are the intersection of the two styles that will need to merge
          mergeInds <- which(new_row_cols %in% ex_row_cols)

          ## newInds are inds that don't exist in the current - this cumulates until the end to see if any are new
          newInds <- newInds[!newInds %in% mergeInds]


          ## If the new style does not merge
          if (length(mergeInds) > 0) {
            to_remove_from_this_style_object <-
              which(ex_row_cols %in% new_row_cols)

            ## the new style intersects with this styleObjects[[i]], we need to remove the intersecting rows and
            ## columns from styleObjects[[i]]
            if (length(to_remove_from_this_style_object) > 0) {
              ## remove these from style object
              styleObjects[[i]]$rows <<-
                styleObjects[[i]]$rows[-to_remove_from_this_style_object]
              styleObjects[[i]]$cols <<-
                styleObjects[[i]]$cols[-to_remove_from_this_style_object]

              if (length(styleObjects[[i]]$rows) == 0 |
                length(styleObjects[[i]]$cols) == 0) {
                keepStyle[i] <-
                  FALSE
              } ## this style applies to no rows or columns anymore
            }

            ## append style object for intersecting cells

            ## we are appending a new style
            keepStyle <-
              c(keepStyle, TRUE) ## keepStyle is used to remove styles that apply to 0 rows OR 0 columns

            ## Merge Style and append to styleObjects
            styleObjects <<-
              append(styleObjects, list(
                list(
                  style = mergeStyle(styleObjects[[i]]$style, newStyle = style),
                  sheet = sheet,
                  rows = rows[mergeInds],
                  cols = cols[mergeInds]
                )
              ))
          }
        } ## if sheet == styleObjects[[i]]$sheet
      } ## End of loop through styles

      ## remove any styles that no longer have any affect
      if (!all(keepStyle)) {
        styleObjects <<- styleObjects[keepStyle]
      }

      ## append style object for non-intersecting cells
      if (length(newInds) > 0) {
        styleObjects <<- append(styleObjects, list(list(
          style = style,
          sheet = sheet,
          rows = rows[newInds],
          cols = cols[newInds]
        )))
      }
    } else {
      ## else we are not stacking

      styleObjects <<- append(styleObjects, list(list(
        style = style,
        sheet = sheet,
        rows = rows,
        cols = cols
      )))
    } ## End if(length(styleObjects) > 0) else if(stack) {}
  }
)



Workbook$methods(
  createNamedRegion = function(ref1, ref2, name, sheet, localSheetId = NULL) {
    name <- replaceIllegalCharacters(name)

    if (is.null(localSheetId)) {
      workbook$definedNames <<- c(
        workbook$definedNames,
        sprintf(
          '<definedName name="%s">\'%s\'!%s:%s</definedName>',
          name,
          sheet,
          ref1,
          ref2
        )
      )
    } else {
      workbook$definedNames <<- c(
        workbook$definedNames,
        sprintf(
          '<definedName name="%s" localSheetId="%s">\'%s\'!%s:%s</definedName>',
          name,
          localSheetId,
          sheet,
          ref1,
          ref2
        )
      )
    }
  }
)


Workbook$methods(
  validate_table_name = function(tableName) {
    tableName <-
      tolower(tableName) ## Excel forces named regions to lowercase

    if (nchar(tableName) > 255) {
      stop("tableName must be less than 255 characters.")
    }

    if (grepl("$", tableName, fixed = TRUE)) {
      stop("'$' character cannot exist in a tableName")
    }

    if (grepl(" ", tableName, fixed = TRUE)) {
      stop("spaces cannot exist in a table name")
    }

    # if(!grepl("^[A-Za-z_]", tableName, perl = TRUE))
    #   stop("tableName must begin with a letter or an underscore")

    if (grepl("R[0-9]+C[0-9]+",
      tableName,
      perl = TRUE,
      ignore.case = TRUE
    )) {
      stop("tableName cannot be the same as a cell reference, such as R1C1")
    }

    if (grepl("^[A-Z]{1,3}[0-9]+$", tableName, ignore.case = TRUE)) {
      stop("tableName cannot be the same as a cell reference")
    }

    if (tableName %in% attr(tables, "tableName")) {
      stop(sprintf("Table with name '%s' already exists!", tableName))
    }

    return(tableName)
  }
)


Workbook$methods(
  check_overwrite_tables = function(sheet,
                                    new_rows,
                                    new_cols,
                                    error_msg = "Cannot overwrite existing table with another table.",
                                    check_table_header_only = FALSE) {
    ## check not overwriting another table
    if (length(tables) > 0) {
      tableSheets <- attr(tables, "sheet")
      sheetNo <- validateSheet(sheet)

      to_check <-
        which(tableSheets %in% sheetNo &
          !grepl("openxlsx_deleted", attr(tables, "tableName"), fixed = TRUE))

      if (length(to_check) > 0) {
        ## only look at tables on this sheet

        exTable <- tables[to_check]

        rows <-
          lapply(names(exTable), function(rectCoords) {
            as.numeric(unlist(regmatches(
              rectCoords, gregexpr("[0-9]+", rectCoords)
            )))
          })
        cols <-
          lapply(names(exTable), function(rectCoords) {
            convertFromExcelRef(unlist(regmatches(
              rectCoords, gregexpr("[A-Z]+", rectCoords)
            )))
          })

        if (check_table_header_only) {
          rows <- lapply(rows, function(x) {
            c(x[1], x[1])
          })
        }


        ## loop through existing tables checking if any over lap with new table
        for (i in seq_along(exTable)) {
          existing_cols <- cols[[i]]
          existing_rows <- rows[[i]]

          if ((min(new_cols) <= max(existing_cols)) &
            (max(new_cols) >= min(existing_cols)) &
            (min(new_rows) <= max(existing_rows)) &
            (max(new_rows) >= min(existing_rows))) {
            stop(error_msg)
          }
        }
      } ## end if(sheet %in% tableSheets)
    } ## end (length(tables) > 0)

    invisible(0)
  }
)




Workbook$methods(
  show = function() {
    exSheets <- sheet_names
    nSheets <- length(exSheets)
    nImages <- length(media)
    nCharts <- length(charts)
    nStyles <- length(styleObjects)
    aSheet <- ActiveSheet
    exSheets <- replaceXMLEntities(exSheets)
    showText <- "A Workbook object.\n"

    if (length(aSheet) == 0) {
      aSheet <- 1
    }

    ## worksheets
    if (nSheets > 0) {
      showText <- c(showText, "\nWorksheets:\n")

      sheetTxt <- lapply(1:nSheets, function(i) {
        tmpTxt <- sprintf('Sheet %s: "%s"\n', i, exSheets[[i]])

        if (length(rowHeights[[i]]) > 0) {
          tmpTxt <-
            append(
              tmpTxt,
              c(
                "\n\tCustom row heights (row: height)\n\t",
                stri_join(
                  sprintf("%s: %s", names(rowHeights[[i]]), round(as.numeric(
                    rowHeights[[i]]
                  ), 2)),
                  collapse = ", ",
                  sep = " "
                )
              )
            )
        }

        if (length(outlineLevels[[i]]) > 0) {
          tmpTxt <-
            append(
              tmpTxt,
              c(
                "\n\tGrouped rows:\n\t",
                stri_join(
                  sprintf("%s", names(outlineLevels[[i]])),
                  collapse = ", ",
                  sep = " "
                )
              )
            )
        }

        if (length(colOutlineLevels[[i]]) > 0) {
          tmpTxt <-
            append(
              tmpTxt,
              c(
                "\n\tGrouped columns:\n\t",
                stri_join(
                  sprintf("%s", names(colOutlineLevels[[i]])),
                  collapse = ", ",
                  sep = " "
                )
              )
            )
        }

        if (length(colWidths[[i]]) > 0) {
          cols <- names(colWidths[[i]])
          widths <- unname(colWidths[[i]])

          widths[widths != "auto"] <-
            as.numeric(widths[widths != "auto"])
          tmpTxt <-
            append(
              tmpTxt,
              c(
                "\n\tCustom column widths (column: width)\n\t ",
                stri_join(
                  sprintf("%s: %s", cols, substr(widths, 1, 5)),
                  sep = " ",
                  collapse = ", "
                )
              )
            )
          tmpTxt <- c(tmpTxt, "\n")
        }
        c(tmpTxt, "\n\n")
      })

      showText <- c(showText, sheetTxt, "\n")
    } else {
      showText <-
        c(showText, "\nWorksheets:\n", "No worksheets attached\n")
    }

    ## images
    if (nImages > 0) {
      showText <-
        c(
          showText,
          "\nImages:\n",
          sprintf('Image %s: "%s"\n', 1:nImages, media)
        )
    }

    if (nCharts > 0) {
      showText <-
        c(
          showText,
          "\nCharts:\n",
          sprintf('Chart %s: "%s"\n', 1:nCharts, charts)
        )
    }

    if (nSheets > 0) {
      showText <-
        c(showText, sprintf(
          "Worksheet write order: %s\n",
          stri_join(sheetOrder, sep = " ", collapse = ", ")
        ))
    }



    if (aSheet >= 1 & nSheets > 0) {
      showText <-
        c(
          showText,
          sprintf(
            'Active Sheet %s: "%s" \n\tPosition: %s\n',
            sheetOrder[aSheet],
            exSheets[[sheetOrder[aSheet]]],
            aSheet
          )
        )
    }

    cat(unlist(showText))
    cat("\n")
  }
)

## TO BE DEPRECATED
Workbook$methods(
  conditionalFormatCell = function(sheet,
                                   startRow,
                                   endRow,
                                   startCol,
                                   endCol,
                                   dxfId,
                                   formula,
                                   type) {
    sheet <- validateSheet(sheet)
    sqref <-
      stri_join(getCellRefs(data.frame(
        "x" = c(startRow, endRow),
        "y" = c(startCol, endCol)
      )), collapse = ":")

    ## Increment priority of conditional formatting rule
    if (length((worksheets[[sheet]]$conditionalFormatting)) > 0) {
      for (i in rev(seq_along(worksheets[[sheet]]$conditionalFormatting))) {
        worksheets[[sheet]]$conditionalFormatting[[i]] <<-
          gsub('(?<=priority=")[0-9]+',
            i + 1L,
            worksheets[[sheet]]$conditionalFormatting[[i]],
            perl = TRUE
          )
      }
    }

    nms <- c(names(worksheets[[sheet]]$conditionalFormatting), sqref)

    if (type == "expression") {
      cfRule <-
        sprintf(
          '<cfRule type="expression" dxfId="%s" priority="1"><formula>%s</formula></cfRule>',
          dxfId,
          formula
        )
    } else if (type == "dataBar") {
      if (length(formula) == 2) {
        negColour <- formula[[1]]
        posColour <- formula[[2]]
      } else {
        posColour <- formula
        negColour <- "FFFF0000"
      }

      guid <-
        stri_join(
          "F7189283-14F7-4DE0-9601-54DE9DB",
          40000L + length(worksheets[[sheet]]$extLst)
        )
      cfRule <-
        sprintf(
          '<cfRule type="dataBar" priority="1"><dataBar><cfvo type="min"/><cfvo type="max"/><color rgb="%s"/></dataBar><extLst><ext uri="{B025F937-C7B1-47D3-B67F-A62EFF666E3E}" xmlns:x14="http://schemas.microsoft.com/office/spreadsheetml/2009/9/main"><x14:id>{%s}</x14:id></ext></extLst></cfRule>',
          posColour,
          guid
        )
    } else if (length(formula) == 2L) {
      cfRule <-
        sprintf(
          '<cfRule type="colorScale" priority="1"><colorScale><cfvo type="min"/><cfvo type="max"/><color rgb="%s"/><color rgb="%s"/></colorScale></cfRule>',
          formula[[1]],
          formula[[2]]
        )
    } else {
      cfRule <-
        sprintf(
          '<cfRule type="colorScale" priority="1"><colorScale><cfvo type="min"/><cfvo type="percentile" val="50"/><cfvo type="max"/><color rgb="%s"/><color rgb="%s"/><color rgb="%s"/></colorScale></cfRule>',
          formula[[1]],
          formula[[2]],
          formula[[3]]
        )
    }

    worksheets[[sheet]]$conditionalFormatting <<-
      append(worksheets[[sheet]]$conditionalFormatting, cfRule)

    names(worksheets[[sheet]]$conditionalFormatting) <<- nms

    invisible(0)
  }
)






Workbook$methods(
  loadStyles = function(stylesXML) {
    ## Build style objects from the styles XML
    stylesTxt <- readUTF8(stylesXML)
    stylesTxt <- removeHeadTag(stylesTxt)

    ## Indexed colours
    vals <- getNodes(xml = stylesTxt, tagIn = "<indexedColors>")
    if (length(vals) > 0) {
      styles$indexedColors <<-
        stri_join("<colors>", vals, "</colors>")
    }

    ## dxf (don't need these, I don't think)
    dxf <- getNodes(xml = stylesTxt, tagIn = "<dxfs")
    if (length(dxf) > 0) {
      dxf <- getNodes(xml = dxf[[1]], tagIn = "<dxf>")
      if (length(dxf) > 0) {
        styles$dxfs <<- dxf
      }
    }

    tableStyles <- getChildlessNode(stylesTxt, tag = "tableStyles")
    if (length(tableStyles) > 0) {
      styles$tableStyles <<- tableStyles
    }

    extLst <- getChildlessNode(stylesTxt, tag = "extLst")
    if (length(extLst) > 0) {
      styles$extLst <<- extLst
    }


    ## Number formats
    numFmts <- getChildlessNode(xml = stylesTxt, tag = "numFmt")
    numFmtFlag <- FALSE
    if (length(numFmts) > 0) {
      numFmtsIds <-
        sapply(numFmts, getAttr, tag = 'numFmtId="', USE.NAMES = FALSE)
      formatCodes <-
        sapply(numFmts, getAttr, tag = 'formatCode="', USE.NAMES = FALSE)
      numFmts <-
        lapply(seq_along(numFmts), function(i) {
          list("numFmtId" = numFmtsIds[[i]], "formatCode" = formatCodes[[i]])
        })
      numFmtFlag <- TRUE
    }

    ## fonts will maintain, sz, color, name, family scheme
    if (grepl("<font/>", stylesTxt, fixed = TRUE)) {
      ## empty font node
      fonts <- getNodes(xml = stylesTxt, tagIn = "<fonts")
      fonts <- strsplit(fonts, split = "<font/>", fixed = TRUE)[[1]]
      fonts <-
        unlist(lapply(fonts, function(xml) {
          c(getNodes(xml, tagIn = "<font>"), "")
        }))
    } else {
      fonts <- getNodes(xml = stylesTxt, tagIn = "<font>")
    }
    styles$fonts[[1]] <<- fonts[[1]]
    fonts <- buildFontList(fonts)


    fills <- getNodes(xml = stylesTxt, tagIn = "<fill>")
    fills <- buildFillList(fills)

    borders <- getOpenClosedNode(stylesTxt, "<borders ", "</borders>")
    borders <-
      substr(
        borders,
        start =  regexpr("<border>", borders)[1],
        stop = regexpr("</borders>", borders) - 1L
      )
    borders <- getNodes(xml = borders, tagIn = "<border")
    borders <- sapply(borders, buildBorder, USE.NAMES = FALSE)


    ## ------------------------------ build styleObjects ------------------------------ ##

    cellXfs <- getNodes(xml = stylesTxt, tagIn = "<cellXfs")

    xf <- getChildlessNode(xml = cellXfs, tag = "xf")
    xfAttrs <- regmatches(xf, gregexpr('[a-zA-Z]+=".*?"', xf))
    xfNames <-
      lapply(xfAttrs, function(xfAttrs) {
        regmatches(
          xfAttrs,
          regexpr('[a-zA-Z]+(?=\\=".*?")', xfAttrs, perl = TRUE)
        )
      })
    xfVals <-
      lapply(xfAttrs, function(xfAttrs) {
        regmatches(xfAttrs, regexpr('(?<=").*?(?=")', xfAttrs, perl = TRUE))
      })

    for (i in seq_along(xf)) {
      names(xfVals[[i]]) <- xfNames[[i]]
    }

    styleObjects_tmp <- list()
    flag <- FALSE
    for (s in xfVals) {
      style <- createStyle()
      if (any(s != "0")) {
        if ("fontId" %in% names(s)) {
          if (s[["fontId"]] != "0") {
            thisFont <- fonts[[(as.integer(s[["fontId"]]) + 1)]]

            if ("sz" %in% names(thisFont)) {
              style$fontSize <- thisFont$sz
            }

            if ("name" %in% names(thisFont)) {
              style$fontName <- thisFont$name
            }

            if ("family" %in% names(thisFont)) {
              style$fontFamily <- thisFont$family
            }

            if ("color" %in% names(thisFont)) {
              style$fontColour <- thisFont$color
            }

            if ("scheme" %in% names(thisFont)) {
              style$fontScheme <- thisFont$scheme
            }

            flags <-
              c("bold", "italic", "underline", "strikeout") %in% names(thisFont)
            if (any(flags)) {
              style$fontDecoration <- NULL
              if (flags[[1]]) {
                style$fontDecoration <-
                  append(style$fontDecoration, "BOLD")
              }

              if (flags[[2]]) {
                style$fontDecoration <-
                  append(style$fontDecoration, "ITALIC")
              }

              if (flags[[3]]) {
                style$fontDecoration <-
                  append(style$fontDecoration, "UNDERLINE")
              }

              if (flags[[4]]) {
                style$fontDecoration <-
                  append(style$fontDecoration, "STRIKEOUT")
              }
            }
          }
        }

        if ("numFmtId" %in% names(s)) {
          if (s[["numFmtId"]] != "0") {
            if (as.integer(s[["numFmtId"]]) < 164) {
              style$numFmt <- list(numFmtId = s[["numFmtId"]])
            } else if (numFmtFlag) {
              style$numFmt <- numFmts[[which(s[["numFmtId"]] == numFmtsIds)[1]]]
            }
          }
        }

        ## Border
        if ("borderId" %in% names(s)) {
          if (s[["borderId"]] != "0") {
            # & "applyBorder" %in% names(s)){

            border_ind <- as.integer(s[["borderId"]]) + 1L
            if (border_ind <= length(borders)) {
              thisBorder <- borders[[border_ind]]

              if ("borderLeft" %in% names(thisBorder)) {
                style$borderLeft <- thisBorder$borderLeft
                style$borderLeftColour <- thisBorder$borderLeftColour
              }

              if ("borderRight" %in% names(thisBorder)) {
                style$borderRight <- thisBorder$borderRight
                style$borderRightColour <-
                  thisBorder$borderRightColour
              }

              if ("borderTop" %in% names(thisBorder)) {
                style$borderTop <- thisBorder$borderTop
                style$borderTopColour <- thisBorder$borderTopColour
              }

              if ("borderBottom" %in% names(thisBorder)) {
                style$borderBottom <- thisBorder$borderBottom
                style$borderBottomColour <-
                  thisBorder$borderBottomColour
              }

              if ("borderDiagonal" %in% names(thisBorder)) {
                style$borderDiagonal <- thisBorder$borderDiagonal
                style$borderDiagonalColour <-
                  thisBorder$borderDiagonalColour
              }

              if ("borderDiagonalUp" %in% names(thisBorder)) {
                style$borderDiagonalUp <-
                  thisBorder$borderDiagonalUp
              }

              if ("borderDiagonalDown" %in% names(thisBorder)) {
                style$borderDiagonalDown <-
                  thisBorder$borderDiagonalDown
              }
            }
          }
        }

        ## alignment
        # applyAlignment <- "applyAlignment" %in% names(s)
        if ("horizontal" %in% names(s)) {
          # & applyAlignment)
          style$halign <- s[["horizontal"]]
        }

        if ("vertical" %in% names(s)) {
          style$valign <- s[["vertical"]]
        }

        if ("indent" %in% names(s)) {
          style$indent <- s[["indent"]]
        }

        if ("textRotation" %in% names(s)) {
          style$textRotation <- s[["textRotation"]]
        }

        ## wrap text
        if ("wrapText" %in% names(s)) {
          if (s[["wrapText"]] %in% c("1", "true")) {
            style$wrapText <- TRUE
          }
        }

        if ("fillId" %in% names(s)) {
          if (s[["fillId"]] != "0") {
            fillId <- as.integer(s[["fillId"]]) + 1L

            if ("fgColor" %in% names(fills[[fillId]])) {
              tmpFg <- fills[[fillId]]$fgColor
              tmpBg <- fills[[fillId]]$bgColor

              if (!is.null(tmpFg)) {
                style$fill$fillFg <- tmpFg
              }

              if (!is.null(tmpFg)) {
                style$fill$fillBg <- tmpBg
              }
            } else {
              style$fill <- fills[[fillId]]
            }
          }
        }


        if ("xfId" %in% names(s)) {
          if (s[["xfId"]] != "0") {
            style$xfId <- s[["xfId"]]
          }
        }
      } ## end if !all(s == "0")

      # Cell protection settings can be "0", so we cannot just skip all zeroes
      if ("locked" %in% names(s)) {
        style$locked <- (s[["locked"]] == "1")
      }

      if ("hidden" %in% names(s)) {
        style$hidden <- (s[["hidden"]] == "1")
      }

      ## we need to skip the first one as this is used as the base style
      if (flag) {
        styleObjects_tmp <- append(styleObjects_tmp, list(style))
      }

      flag <- TRUE
    } ## end of for loop through styles s in ...


    ## ------------------------------ build styleObjects Complete ------------------------------ ##


    return(styleObjects_tmp)
  }
)

Workbook$methods(
  protectWorkbook = function(protect = TRUE,
                             lockStructure = FALSE,
                             lockWindows = FALSE,
                             password = NULL,
                             type = NULL) {
    attr <- c()
    if (!is.null(password)) {
      attr["workbookPassword"] <- hashPassword(password)
    }
    if (!missing(lockStructure) && !is.null(lockStructure)) {
      attr["lockStructure"] <- toString(as.numeric(lockStructure))
    }
    if (!missing(lockWindows) && !is.null(lockWindows)) {
      attr["lockWindows"] <- toString(as.numeric(lockWindows))
    }
    # TODO: Shall we parse the existing protection settings and preserve all unchanged attributes?
    if (protect) {
      workbook$workbookProtection <<-
        sprintf(
          "<workbookProtection %s/>",
          stri_join(
            names(attr),
            '="',
            attr,
            '"',
            collapse = " ",
            sep = ""
          )
        )
      if (!is.null(type) | !is.null(password))
        workbook$apps <<- sprintf("<DocSecurity>%i</DocSecurity>", type)
    } else {
      workbook$workbookProtection <<- ""
    }
  }
)







Workbook$methods(
  addCreator = function(Creator = NULL) {
    if (!is.null(Creator)) {
      current_creator <-
        stri_match(core, regex = "<dc:creator>(.*?)</dc:creator>")[1, 2]
      core <<-
        stri_replace_all_fixed(
          core,
          pattern = current_creator,
          replacement = stri_c(current_creator, Creator, sep = ";")
        )
    }
  }
)





Workbook$methods(
  getCreators = function() {
    current_creator <-
      stri_match(core, regex = "<dc:creator>(.*?)</dc:creator>")[1, 2]

    current_creator_vec <- as.character(stri_split_fixed(
      str = current_creator,
      pattern = ";",
      simplify = TRUE
    ))

    return(current_creator_vec)
  }
)



Workbook$methods(
  changeLastModifiedBy = function(LastModifiedBy = NULL) {
    if (!is.null(LastModifiedBy)) {
      current_LastModifiedBy <-
        stri_match(core, regex = "<cp:lastModifiedBy>(.*?)</cp:lastModifiedBy>")[1, 2]
      core <<-
        stri_replace_all_fixed(
          core,
          pattern = current_LastModifiedBy,
          replacement = LastModifiedBy
        )
    }
  }
)



Workbook$methods(
  setactiveSheet = function(activeSheet = NULL) {
    if (is.character(activeSheet)) {
      if (activeSheet %in% sheet_names) {
        ActiveSheet <<- which(sheet_names[sheetOrder] == activeSheet)
      } else {
        stop(paste(activeSheet, "doesn't exist as sheet name."))
      }
    }

    if (is.integer(activeSheet)|is.numeric(activeSheet)) {
      if (activeSheet %in% seq_along(sheet_names)) {
        ActiveSheet <<- which(sheetOrder==activeSheet)
      }else {
        stop(paste(activeSheet, "doesn't exist as sheet index."))
      }
    }

    for(i in seq_along(sheet_names)){
      worksheets[[i]]$sheetViews <<- stri_replace_all_regex(worksheets[[i]]$sheetViews,
                           "tabSelected=\"(1|true|false|0)\"",
                           paste0("tabSelected=\"",
                                  ifelse(sheetOrder[ActiveSheet]  == i,"true","false")
                                  ,"\""))


    }



  }
)

test_that("Validate Table Names", {
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")

  ## case
  expect_equal(wb$validate_table_name("test"), "test")
  expect_equal(wb$validate_table_name("TEST"), "test")
  expect_equal(wb$validate_table_name("Test"), "test")

  ## length
  expect_error(wb$validate_table_name(paste(sample(LETTERS, size = 300, replace = TRUE), collapse = "")), regexp = "tableName must be less than 255 characters")

  ## look like cell ref
  expect_error(wb$validate_table_name("R1C2"), regexp = "tableName cannot be the same as a cell reference, such as R1C1", fixed = TRUE)
  expect_error(wb$validate_table_name("A1"), regexp = "tableName cannot be the same as a cell reference", fixed = TRUE)

  expect_error(wb$validate_table_name("R06821C9682"), regexp = "tableName cannot be the same as a cell reference, such as R1C1", fixed = TRUE)
  expect_error(wb$validate_table_name("ABD918751"), regexp = "tableName cannot be the same as a cell reference", fixed = TRUE)

  expect_error(wb$validate_table_name("A$100"), regexp = "'$' character cannot exist in a tableName", fixed = TRUE)
  expect_error(wb$validate_table_name("A12$100"), regexp = "'$' character cannot exist in a tableName", fixed = TRUE)

  tbl_nm <- "性別"
  expect_equal(wb$validate_table_name(tbl_nm), tbl_nm)
})





test_that("Existing Table Names", {
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")

  ## Existing names - case in-sensitive

  writeDataTable(wb, sheet = 1, x = head(iris), tableName = "Table1")
  expect_error(wb$validate_table_name("Table1"), regexp = "Table with name 'table1' already exists", fixed = TRUE)
  expect_error(writeDataTable(wb, sheet = 1, x = head(iris), tableName = "Table1", startCol = 10), regexp = "Table with name 'table1' already exists", fixed = TRUE)

  expect_error(wb$validate_table_name("TABLE1"), regexp = "Table with name 'table1' already exists", fixed = TRUE)
  expect_error(writeDataTable(wb, sheet = 1, x = head(iris), tableName = "TABLE1", startCol = 20), regexp = "Table with name 'table1' already exists", fixed = TRUE)

  expect_error(wb$validate_table_name("table1"), regexp = "Table with name 'table1' already exists", fixed = TRUE)
  expect_error(writeDataTable(wb, sheet = 1, x = head(iris), tableName = "table1", startCol = 30), regexp = "Table with name 'table1' already exists", fixed = TRUE)
})

#' @name read.xlsx
#' @title  Read from an Excel file or Workbook object
#' @description Read data from an Excel file or Workbook object into a data.frame
#' @param xlsxFile An xlsx file, Workbook object or URL to xlsx file.
#' @param sheet The name or index of the sheet to read data from.
#' @param startRow first row to begin looking for data.  Empty rows at the top of a file are always skipped,
#' regardless of the value of startRow.
#' @param colNames If `TRUE`, the first row of data will be used as column names.
#' @param skipEmptyRows If `TRUE`, empty rows are skipped else empty rows after the first row containing data
#' will return a row of NAs.
#' @param rowNames If `TRUE`, first column of data will be used as row names.
#' @param detectDates If `TRUE`, attempt to recognise dates and perform conversion.
#' @param cols A numeric vector specifying which columns in the Excel file to read.
#' If NULL, all columns are read.
#' @param rows A numeric vector specifying which rows in the Excel file to read.
#' If NULL, all rows are read.
#' @param check.names logical. If TRUE then the names of the variables in the data frame
#' are checked to ensure that they are syntactically valid variable names
#' @param sep.names One character which substitutes blanks in column names. By default, "."
#' @param namedRegion A named region in the Workbook. If not NULL startRow, rows and cols parameters are ignored.
#' @param na.strings A character vector of strings which are to be interpreted as NA. Blank cells will be returned as NA.
#' @param fillMergedCells If TRUE, the value in a merged cell is given to all cells within the merge.
#' @param skipEmptyCols If `TRUE`, empty columns are skipped.
#' @seealso [getNamedRegions()]
#' @details Formulae written using writeFormula to a Workbook object will not get picked up by read.xlsx().
#' This is because only the formula is written and left to be evaluated when the file is opened in Excel.
#' Opening, saving and closing the file with Excel will resolve this.
#' @author Alexander Walker
#' @return data.frame
#' @export
#' @examples
#'
#' xlsxFile <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
#' df1 <- read.xlsx(xlsxFile = xlsxFile, sheet = 1, skipEmptyRows = FALSE)
#' sapply(df1, class)
#'
#' df2 <- read.xlsx(xlsxFile = xlsxFile, sheet = 3, skipEmptyRows = TRUE)
#' df2$Date <- convertToDate(df2$Date)
#' sapply(df2, class)
#' head(df2)
#'
#' df2 <- read.xlsx(
#'   xlsxFile = xlsxFile, sheet = 3, skipEmptyRows = TRUE,
#'   detectDates = TRUE
#' )
#' sapply(df2, class)
#' head(df2)
#'
#' wb <- loadWorkbook(system.file("extdata", "readTest.xlsx", package = "openxlsx"))
#' df3 <- read.xlsx(wb, sheet = 2, skipEmptyRows = FALSE, colNames = TRUE)
#' df4 <- read.xlsx(xlsxFile, sheet = 2, skipEmptyRows = FALSE, colNames = TRUE)
#' all.equal(df3, df4)
#'
#' wb <- loadWorkbook(system.file("extdata", "readTest.xlsx", package = "openxlsx"))
#' df3 <- read.xlsx(wb,
#'   sheet = 2, skipEmptyRows = FALSE,
#'   cols = c(1, 4), rows = c(1, 3, 4)
#' )
#'
#' ## URL
#' ##
#' \dontrun{
#' xlsxFile <- "https://github.com/awalker89/openxlsx/raw/master/inst/readTest.xlsx"
#' head(read.xlsx(xlsxFile))
#' }
#'
#' @export
read.xlsx <- function(
  xlsxFile,
  sheet,
  startRow        = 1,
  colNames        = TRUE,
  rowNames        = FALSE,
  detectDates     = FALSE,
  skipEmptyRows   = TRUE,
  skipEmptyCols   = TRUE,
  rows            = NULL,
  cols            = NULL,
  check.names     = FALSE,
  sep.names       = ".",
  namedRegion     = NULL,
  na.strings      = "NA",
  fillMergedCells = FALSE
) {
  UseMethod("read.xlsx", xlsxFile)
}

#' @export
read.xlsx.default <- function(
  xlsxFile,
  sheet,
  startRow        = 1,
  colNames        = TRUE,
  rowNames        = FALSE,
  detectDates     = FALSE,
  skipEmptyRows   = TRUE,
  skipEmptyCols   = TRUE,
  rows            = NULL,
  cols            = NULL,
  check.names     = FALSE,
  sep.names       = ".",
  namedRegion     = NULL,
  na.strings      = "NA",
  fillMergedCells = FALSE
) {
  ## Validate inputs and get files
  xlsxFile <- getFile(xlsxFile)
  
  if (!file.exists(xlsxFile)) {
    stop("File does not exist.")
  }
  
  sheetselected <- TRUE
  if (missing(sheet)) {
    sheet <- 1
    sheetselected <- FALSE
  }
  
  if (!grepl("\\.xlsx|\\.xlsm$", xlsxFile)) {
    stop("openxlsx can only read .xlsx or .xlsm files", call. = FALSE)
  }
  
  assert_true_false1(colNames)
  assert_true_false1(rowNames)
  assert_true_false1(detectDates)
  assert_true_false1(skipEmptyRows)
  assert_true_false1(check.names)
  assert_character1(sep.names, scalar = TRUE)
  assert_length(sheet, 1L)
  assert_length(startRow, 1L)
  
  if (is.null(rows)) {
    rows <- NA
  } else if (length(rows) > 1L) {
    rows <- as.integer(sort(rows))
  }
  
  xmlDir <- paste0(tempfile(), "_excelXMLRead")
  xmlFiles <- unzip(xlsxFile, exdir = xmlDir)
  on.exit(unlink(xmlDir, recursive = TRUE), add = TRUE)
  
  sharedStringsFile <- grep("sharedStrings.xml$", xmlFiles, perl = TRUE, value = TRUE)
  workbook <- grep("workbook.xml$", xmlFiles, perl = TRUE, value = TRUE)
  workbookRelsXML <- grep("workbook.xml.rels$", xmlFiles, perl = TRUE, value = TRUE)
  
  ## get workbook names
  workbookRelsXML <- paste(readUTF8(workbookRelsXML), collapse = "")
  workbookRelsXML <- getChildlessNode(xml = workbookRelsXML, tag = "Relationship")
  
  workbook <- unlist(readUTF8(workbook))
  workbook <- removeHeadTag(workbook)
  
  sheets <- unlist(regmatches(
    workbook,
    gregexpr("(?<=<sheets>).*(?=</sheets>)", workbook, perl = TRUE)
  ))
  sheets <- unlist(regmatches(
    sheets, 
    gregexpr("<sheet[^>]*>", sheets, perl = TRUE)
  ))
  ## Some veryHidden sheets do not have a sheet content and their rId is empty.
  ## Such sheets need to be filtered out because otherwise their sheet names
  ## occur in the list of all sheet names, leading to a wrong association
  ## of sheet names with sheet indeces.
  sheets <- grep('r:id="[[:blank:]]*"', sheets, invert = TRUE, value = TRUE)
  
  
  ## make sure sheetId is 1 based
  sheetrId <- unlist(getRId(sheets))
  sheetNames <- unlist(regmatches(
    sheets, 
    gregexpr('(?<=name=")[^"]+', sheets, perl = TRUE)
  ))
  sheetNames <- replaceXMLEntities(sheetNames)
  
  
  nSheets <- length(sheetrId)
  if (nSheets == 0) {
    stop("Workbook has no worksheets")
  }
  
  ## Named region logic
  reading_named_region <- FALSE
  if (!is.null(namedRegion)) {
    dn <- getNodes(xml = workbook, tagIn = "<definedNames>")
    dn <- unlist(regmatches(dn, gregexpr("<definedName [^<]*", dn, perl = TRUE)))
    
    # search for sheetNames in list of named_region
    # sheet in between '>' and '!'
    dn_sheetNames <- gsub(".*[>]([^.]+)[!].*", "\\1", dn)
    
    # Check if there are any whitespaces in dn_sheetNames.
    # Hint: sheet names must not contain: \ / ? * [ ]
    wsp <- grepl(pattern = "'", dn_sheetNames)
    if (any(wsp)) {
      # sheetNames in between ''' and '''. If there is a whitespace in a sheet
      # name, the name will be "'sheet 1'" instead of "sheet 1.
      dn_sheetNames[wsp] <- gsub("^'+|'+$", "\\1", dn_sheetNames[wsp])
    }
    # namedRegion in between 'name="' and '"'
    dn_namedRegion <- gsub(".*name=\"([[:graph:]_]+)\".*", "\\1", dn)
    
    if (length(dn) == 0) {
      warning("Workbook has no named region.")
      return(invisible(NULL))
    }
    
    if (all(dn_namedRegion != namedRegion)) {
      warning("Workbook has no such named region.")
      return(invisible(NULL))
    }
    
    idx <- match(dn_namedRegion, namedRegion)
    
    # make sure that the length of both vectors is identical
    dn <- dn[!is.na(idx)]
    dn_namedRegion <- dn_namedRegion[!is.na(idx)]
    dn_sheetNames  <- dn_sheetNames[!is.na(idx)]
    
    # a sheet was selected
    if (sheetselected) {
      idx <- match(dn_sheetNames, sheetNames)
      if (is.numeric(sheet)) {
        idx <- which(idx == sheet)
        
      } else {
        idx <- which(dn_sheetNames == sheet)
      }
      dn <- dn[idx]
      
      if (length(dn) > 1) {
        warning("unexpectedly found more than one dn.")
        print(dn)
        return(invisible(NULL))
      }
      
      if ( identical(dn, character(0)) ) {
        warning("Workbook has no such named region on this sheet.")
        return(invisible(NULL))
      }
    }
    
    # Do not print warning if a specific sheet is requested
    if ((length(dn) > 1) && (!sheetselected)) {
      msg <- c(sprintf("Region '%s' found on multiple sheets: \n", namedRegion),
        paste(dn_sheetNames, collapse = "\n"),
        "\nUsing the first appearance.")
      message(msg)
      
      dn <- dn[1]
      dn_namedRegion <- dn_namedRegion[1]
      dn_sheetNames  <- dn_sheetNames[1]
    }
    
    # region is redefined later
    region <- regmatches(dn, regexpr("(?<=>)[^\\<]+", dn, perl = TRUE))
    sheet <- sheetNames[vapply(sheetNames, grepl, NA, dn)]
    
    if (length(sheet) > 1) {
      sheet <- sheet[which.max(nchar(sheet))]
    }
    
    region <- gsub("[^A-Z0-9:]", "", gsub(sheet, "", region, fixed = TRUE))
    
    if (grepl(":", region, fixed = TRUE)) {
      cols <- unlist(lapply(
        strsplit(region, split = ":", fixed = TRUE),
        convertFromExcelRef
      ))
      rows <- unlist(lapply(strsplit(region, split = ":", fixed = TRUE), function(x) {
        as.integer(gsub("[A-Z]", "", x, perl = TRUE))
      }))
      cols <- seq.int(min(cols), max(cols))
      rows <- seq.int(min(rows), max(rows))
    } else {
      cols <- convertFromExcelRef(region)
      rows <- as.integer(gsub("[A-Z]", "", region, perl = TRUE))
    }
    
    startRow <- 1
    reading_named_region <- TRUE
  }
  
  ## get the file_name for each sheetrId
  file_name <- sapply(sheetrId, function(rId) {
    txt <- grep(sprintf('Id="%s"', rId), workbookRelsXML, fixed = TRUE, value = TRUE)
    regmatches(txt, regexpr('(?<=Target=").+xml(?=")', txt, perl = TRUE))
  })
  
  
  ## get the correct sheets
  if (is.character(sheet)) {
    sheetNames <- replaceXMLEntities(sheetNames)
    sheetInd <- which(sheetNames == sheet)
    if (length(sheetInd) == 0) {
      stop(sprintf('Cannot find sheet named "%s"', sheet))
    }
    sheet <- file_name[sheetInd]
  } else {
    if (nSheets < sheet) {
      stop(sprintf("sheet %s does not exist.", sheet))
    }
    sheet <- file_name[sheet]
  }
  
  if (length(sheet) == 0) {
    stop("Length of sheet is 0", call. = FALSE)
  }
  
  ## get file
  worksheet <- xmlFiles[grepl(tolower(sheet), tolower(xmlFiles), fixed = TRUE)]
  if (length(worksheet) == 0) {
    stop("Length of worksheet is 0", call. = FALSE)
  }
  
  ## read in sharedStrings
  if (length(sharedStringsFile) > 0) {
    sharedStrings <-
      getSharedStringsFromFile(sharedStringsFile = sharedStringsFile, isFile = TRUE)
    if (!is.null(na.strings)) {
      sharedStrings[is.na(sharedStrings) | sharedStrings %in% na.strings] <- "openxlsx_na_vlu"
    }
  } else {
    sharedStrings <- ""
  }
  
  if (is.character(startRow)) {
    startRowStr <- startRow
    startRow <- 1
  } else {
    startRowStr <- NULL
  }
  
  ## single function get all r, s (if detect dates is TRUE), t, v
  cell_info <- getCellInfo(
    xmlFile = worksheet,
    sharedStrings = sharedStrings,
    skipEmptyRows = skipEmptyRows,
    startRow = startRow,
    rows = rows,
    getDates = detectDates
  )
  
  if (fillMergedCells && length(cell_info$cellMerge) > 0) {
    # stop("Not implemented")
    
    merge_mapping <- mergeCell2mapping(cell_info$cellMerge)
    
    ## remove any elements from  r, string_refs, b, s that existing in merge_mapping
    ## insert all missing refs into r
    
    to_remove_inds <- cell_info$r %in% merge_mapping$ref
    to_remove_elems <- cell_info$r[to_remove_inds]
    
    if (any(to_remove_inds)) {
      cell_info$r <- cell_info$r[!to_remove_inds]
      cell_info$s <- cell_info$s[!to_remove_inds]
      cell_info$v <- cell_info$v[!to_remove_inds]
      cell_info$string_refs <-
        cell_info$string_refs[!cell_info$string_refs %in% to_remove_elems]
    }
    
    ## Now insert
    inds <- match(merge_mapping$anchor_cell, cell_info$r)
    
    ## String refs (must sort)
    new_string_refs <-
      merge_mapping$ref[merge_mapping$anchor_cell %in% cell_info$string_refs]
    cell_info$string_refs <-
      c(cell_info$string_refs, new_string_refs)
    cell_info$string_refs <-
      cell_info$string_refs[order(
        as.integer(gsub("[A-Z]", "", cell_info$string_refs, perl = TRUE)),
        nchar(cell_info$string_refs),
        cell_info$string_refs
      )]
    
    ## r
    cell_info$r <- c(cell_info$r, merge_mapping$ref)
    cell_info$v <- c(cell_info$v, cell_info$v[inds])
    
    ord <- order(
      as.integer(gsub("[A-Z]", "", cell_info$r, perl = TRUE)),
      nchar(cell_info$r),
      cell_info$r
    )
    
    cell_info$r <- cell_info$r[ord]
    cell_info$v <- cell_info$v[ord]
    
    if (length(cell_info$s) > 0) {
      cell_info$s <- c(cell_info$s, cell_info$s[inds])[ord]
    }
    
    cell_info$nRows <- calc_number_rows(x = cell_info$r, skipEmptyRows = skipEmptyRows)
  }
  
  cell_rows <- as.integer(gsub("[A-Z]", "", cell_info$r, perl = TRUE))
  cell_cols <- convert_from_excel_ref(x = cell_info$r)
  
  ## subsetting ----
  ## Remove cells where cell is NA (na.strings or empty sharedString '<si><t/></si>')
  
  if (length(cell_info$v) == 0) {
    warning("No data found on worksheet.\n", call. = FALSE)
    return(NULL)
  }
  
  keep <- !is.na(cell_info$v)
  if (!is.null(cols)) {
    keep <- keep & (cell_cols %in% cols)
  }
  
  ## End of subsetting
  
  ## Subset
  cell_rows <- cell_rows[keep]
  cell_cols <- cell_cols[keep]
  
  v <- cell_info$v[keep]
  s <- cell_info$s[keep]
  
  string_refs <- match(cell_info$string_refs, cell_info$r[keep])
  string_refs <- string_refs[!is.na(string_refs)]
  
  if (skipEmptyRows) {
    nRows <- length(unique(cell_rows))
  } else if (reading_named_region) {
    ## keep region the correct size
    nRows <- max(rows) - min(rows) + 1
  } else {
    nRows <- max(cell_rows) - min(cell_rows) + 1
  }
  
  if (nRows == 0 || length(cell_rows) == 0) {
    warning("No data found on worksheet.", call. = FALSE)
    return(NULL)
  }
  
  Encoding(v) <- "UTF-8" ## only works if length(v) > 0
  
  if (!is.null(startRowStr)) {
    stop("startRowStr not implemented")
    ind <- grep(startRowStr, v, ignore.case = TRUE)
    if (length(ind) > 0) {
      startRow <- as.numeric(gsub("[A-Z]", "", r[ind[[1]]]))
      toKeep <- grep(sprintf("[A-Z]%s$", startRow), r)[[1]]
      if (toKeep > 1) {
        toRemove <- 1:(toKeep - 1)
        string_refs <- string_refs[!string_refs %in% r[toRemove]]
        v <- v[-toRemove]
        r <- r[-toRemove]
        nRows <-
          calc_number_rows(x = r, skipEmptyRows = skipEmptyRows)
      }
    }
  }
  
  ## Determine date cells (if required)
  origin <- 25569L
  if (detectDates) {
    ## get date origin
    if (grepl('date1904="1"|date1904="true"', workbook, ignore.case = TRUE)) {
      origin <- 24107L
    }
    
    stylesXML <- grep("styles.xml", xmlFiles, value = TRUE)
    styles <- readUTF8(stylesXML)
    styles <- removeHeadTag(styles)
    
    ## Number formats
    numFmts <- getChildlessNode(xml = styles, tag = "numFmt")
    
    dateIds <- NULL
    if (length(numFmts) > 0) {
      numFmtsIds <- sapply(numFmts, getAttr, tag = 'numFmtId="', USE.NAMES = FALSE)
      formatCodes <- sapply(numFmts, getAttr, tag = 'formatCode="', USE.NAMES = FALSE)
      formatCodes <- gsub(".*(?<=\\])|@", "", formatCodes, perl = TRUE)
      
      ## this regex defines what "looks" like a date
      dateIds <- numFmtsIds[!grepl("[^mdyhsapAMP[:punct:] ]", formatCodes) &
          nchar(formatCodes > 3)]
    }
    
    dateIds <- c(dateIds, 14)
    
    ## which styles are using these dateIds
    cellXfs <- getNodes(xml = styles, tagIn = "<cellXfs")
    xf <- getChildlessNode(xml = cellXfs, tag = "xf")
    lookingFor <- paste(sprintf('numFmtId="%s"', dateIds), collapse = "|")
    dateStyleIds <- which(sapply(xf, function(x) grepl(lookingFor, x), USE.NAMES = FALSE)) - 1L
    
    isDate <- s %in% dateStyleIds
    
    ## set to false if in string_refs
    isDate[seq_along(s) %in% string_refs] <- FALSE
    
    # check numbers are also integers
    not_an_integer <- numeric(length(v))
    not_an_integer[isDate] <- suppressWarnings(as.numeric(v[isDate]))
    not_an_integer <- (not_an_integer %% 1L != 0) & !is.na(not_an_integer)
    isDate[not_an_integer] <- FALSE
    if (origin == 25569L) {
      earlyDates <- suppressWarnings(as.integer(v) < 60)
      v[earlyDates & isDate] <- as.integer(v[earlyDates & isDate]) + 1L
    }
    
    ## perform int to date to character conversion (way too slow)
    v[isDate] <- format(
      as.Date(as.integer(v[isDate]) - origin, origin = "1970-01-01"),
      "%Y-%m-%d"
    )
  } else {
    isDate <- as.logical(NA)
  }
  
  ## Build data.frame
  m <- read_workbook(
    cols_in       = cell_cols,
    rows_in       = cell_rows,
    v             = v,
    string_inds   = string_refs,
    is_date       = isDate,
    hasColNames   = colNames,
    hasSepNames   = sep.names,
    skipEmptyRows = skipEmptyRows,
    skipEmptyCols = skipEmptyCols,
    nRows         = nRows,
    clean_names   = clean_names
  )
  
  if (rowNames) {
    # set the first column as the rownames
    rownames(m) <- m[[1]]
    m[[1]] <- NULL
  }
  
  if (colNames) {
    if (check.names) {
      colnames(m) <- make.names(colnames(m), unique = TRUE)
    }
  } else {
    # Need to reset column names
    if (rowNames) {
      colnames(m) <- make.names(seq_along(m))
    }
  }
  
  m
}

#' @name readWorkbook
#' @title  Read from an Excel file or Workbook object
#' @description Read data from an Excel file or Workbook object into a data.frame
#' @inheritParams read.xlsx
#' @details Creates a data.frame of all data in worksheet.
#' @author Alexander Walker
#' @return data.frame
#' @seealso [getNamedRegions()]
#' @seealso [read.xlsx()]
#' @export
#' @examples
#' xlsxFile <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
#' df1 <- readWorkbook(xlsxFile = xlsxFile, sheet = 1)
#'
#' xlsxFile <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
#' df1 <- readWorkbook(xlsxFile = xlsxFile, sheet = 1, rows = c(1, 3, 5), cols = 1:3)
readWorkbook <- function(
  xlsxFile,
  sheet           = 1,
  startRow        = 1,
  colNames        = TRUE,
  rowNames        = FALSE,
  detectDates     = FALSE,
  skipEmptyRows   = TRUE,
  skipEmptyCols   = TRUE,
  rows            = NULL,
  cols            = NULL,
  check.names     = FALSE,
  sep.names       = ".",
  namedRegion     = NULL,
  na.strings      = "NA",
  fillMergedCells = FALSE
) {
  read.xlsx(
    xlsxFile        = xlsxFile,
    sheet           = sheet,
    startRow        = startRow,
    colNames        = colNames,
    rowNames        = rowNames,
    detectDates     = detectDates,
    skipEmptyRows   = skipEmptyRows,
    skipEmptyCols   = skipEmptyCols,
    rows            = rows,
    cols            = cols,
    check.names     = check.names,
    sep.names       = sep.names,
    namedRegion     = namedRegion,
    na.strings      = na.strings,
    fillMergedCells = fillMergedCells
  )
}

test_that("read_xlsx() does not change random seed", {
  rs <- .Random.seed
  expect_identical(rs, .Random.seed)
  tf <- temp_xlsx()
  expect_identical(rs, .Random.seed)
  write.xlsx(data.frame(a = 1), tf)
  expect_identical(rs, .Random.seed)
  read.xlsx(tf)
  expect_identical(rs, .Random.seed)
  unlink(tf)
})

context("Read xlsx")


test_that("read.xlsx correct sheet", {
  fl <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
  sheet_names <- getSheetNames(file = fl)

  expected_sheet_names <- c(
    "Sheet1", "Sheet2", "Sheet 3",
    "Sheet 4", "Sheet 5", "Sheet 6",
    "1", "11", "111", "1111", "11111", "111111"
  )


  expect_equal(object = sheet_names, expected = expected_sheet_names)

  expect_equal(read.xlsx(xlsxFile = fl, sheet = 7), data.frame(x = 1))
  expect_equal(read.xlsx(xlsxFile = fl, sheet = 8), data.frame(x = 11))
  expect_equal(read.xlsx(xlsxFile = fl, sheet = 9), data.frame(x = 111))
  expect_equal(read.xlsx(xlsxFile = fl, sheet = 10), data.frame(x = 1111))
  expect_equal(read.xlsx(xlsxFile = fl, sheet = 11), data.frame(x = 11111))
  expect_equal(read.xlsx(xlsxFile = fl, sheet = 12), data.frame(x = 111111))

  expect_equal(read.xlsx(xlsxFile = fl, sheet = "1"), data.frame(x = 1))
  expect_equal(read.xlsx(xlsxFile = fl, sheet = "11"), data.frame(x = 11))
  expect_equal(read.xlsx(xlsxFile = fl, sheet = "111"), data.frame(x = 111))
  expect_equal(read.xlsx(xlsxFile = fl, sheet = "1111"), data.frame(x = 1111))
  expect_equal(read.xlsx(xlsxFile = fl, sheet = "11111"), data.frame(x = 11111))
  expect_equal(read.xlsx(xlsxFile = fl, sheet = "111111"), data.frame(x = 111111))
})

context("Skip Empty Cols")

test_that("skip empty rows", {
  xlsxfile <- temp_xlsx()
  df <- data.frame("x" = c(1, NA, NA, 2), "y" = c(1, NA, NA, 3))

  write.xlsx(df, xlsxfile)

  wb <- loadWorkbook(xlsxfile)

  df1 <- readWorkbook(xlsxfile, skipEmptyRows = FALSE)
  df2 <- readWorkbook(wb, skipEmptyRows = FALSE)

  expect_equal(df, df1)
  expect_equal(df, df2)

  v <- c("A1", "B1", "A2", "B2", "A5", "B5")
  expect_equal(calc_number_rows(x = v, skipEmptyRows = TRUE), 3)
  expect_equal(calc_number_rows(x = v, skipEmptyRows = FALSE), 5)

  ## DONT SKIP
  df1 <- readWorkbook(xlsxfile, skipEmptyRows = TRUE)
  df2 <- readWorkbook(wb, skipEmptyRows = TRUE)

  expect_equal(nrow(df1), 2)
  expect_equal(nrow(df2), 2)

  expect_equivalent(df[c(1, 4), ], df1)
  expect_equivalent(df[c(1, 4), ], df2)
})

test_that("Version 4 fixes from File", {
  fl <- system.file("extdata", "readTest.xlsx", package = "openxlsx")

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = TRUE, skipEmptyRows = TRUE, colNames = FALSE)
  expect_equal(nrow(x), 5L)
  expect_equal(ncol(x), 4L)

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = TRUE, skipEmptyRows = TRUE, colNames = TRUE)
  expect_equal(nrow(x), 5L - 1L)
  expect_equal(ncol(x), 4L)


  ##############################################################
  ## FALSE FALSE FALSE

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = FALSE, skipEmptyRows = FALSE, colNames = FALSE)
  expect_equal(nrow(x), 6L)
  expect_equal(ncol(x), 8L)

  ## NA columns
  expect_true(all(is.na(x$X1)))
  expect_true(all(is.na(x$X2)))
  expect_true(all(is.na(x$X3)))
  expect_true(all(is.na(x$X7)))

  ## NA rows
  expect_true(all(is.na(x[3, ])))



  ##############################################################
  ## FALSE FALSE TRUE

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = FALSE, skipEmptyRows = FALSE, colNames = TRUE)
  expect_equal(nrow(x), 6L - 1L)
  expect_equal(ncol(x), 8L)


  ## NA columns
  expect_true(all(is.na(x$X1)))
  expect_true(all(is.na(x$X2)))
  expect_true(all(is.na(x$X3)))
  expect_true(all(is.na(x$X7)))

  ## NA rows
  expect_true(all(is.na(x[2, ])))



  ##############################################################
  ## FALSE TRUE FALSE

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = FALSE, skipEmptyRows = TRUE, colNames = FALSE)
  expect_equal(nrow(x), 5L)
  expect_equal(ncol(x), 8L)

  ## NA columns
  expect_true(all(is.na(x$X1)))
  expect_true(all(is.na(x$X2)))
  expect_true(all(is.na(x$X3)))
  expect_true(all(is.na(x$X7)))




  ##############################################################
  ## FALSE TRUE TRUE

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = FALSE, skipEmptyRows = TRUE, colNames = TRUE)
  expect_equal(nrow(x), 5L - 1L)
  expect_equal(ncol(x), 8L)


  ## NA columns
  expect_true(all(is.na(x$X1)))
  expect_true(all(is.na(x$X2)))
  expect_true(all(is.na(x$X3)))
  expect_true(all(is.na(x$X7)))



  ##############################################################
  ## TRUE FALSE FALSE

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = TRUE, skipEmptyRows = FALSE, colNames = FALSE)
  expect_equal(nrow(x), 6L)
  expect_equal(ncol(x), 4L)

  ## NA rows
  expect_true(all(is.na(x[3, ])))



  ##############################################################
  ## TRUE FALSE TRUE

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = TRUE, skipEmptyRows = FALSE, colNames = TRUE)
  expect_equal(nrow(x), 6L - 1L)
  expect_equal(ncol(x), 4L)


  ## NA rows
  expect_true(all(is.na(x[2, ])))
})






test_that("Version 4 fixes from Workbook Objects", {
  fl <- loadWorkbook(system.file("extdata", "readTest.xlsx", package = "openxlsx"))


  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = TRUE, skipEmptyRows = TRUE, colNames = FALSE)
  expect_equal(nrow(x), 5L)
  expect_equal(ncol(x), 4L)

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = TRUE, skipEmptyRows = TRUE, colNames = TRUE)
  expect_equal(nrow(x), 5L - 1L)
  expect_equal(ncol(x), 4L)


  ##############################################################
  ## FALSE FALSE FALSE

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = FALSE, skipEmptyRows = FALSE, colNames = FALSE)
  expect_equal(nrow(x), 6L)
  expect_equal(ncol(x), 8L)

  ## NA columns
  expect_true(all(is.na(x$X1)))
  expect_true(all(is.na(x$X2)))
  expect_true(all(is.na(x$X3)))
  expect_true(all(is.na(x$X7)))

  ## NA rows
  expect_true(all(is.na(x[3, ])))



  ##############################################################
  ## FALSE FALSE TRUE

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = FALSE, skipEmptyRows = FALSE, colNames = TRUE)
  expect_equal(nrow(x), 6L - 1L)
  expect_equal(ncol(x), 8L)


  ## NA columns
  expect_true(all(is.na(x$X1)))
  expect_true(all(is.na(x$X2)))
  expect_true(all(is.na(x$X3)))
  expect_true(all(is.na(x$X7)))

  ## NA rows
  expect_true(all(is.na(x[2, ])))



  ##############################################################
  ## FALSE TRUE FALSE

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = FALSE, skipEmptyRows = TRUE, colNames = FALSE)
  expect_equal(nrow(x), 5L)
  expect_equal(ncol(x), 8L)

  ## NA columns
  expect_true(all(is.na(x$X1)))
  expect_true(all(is.na(x$X2)))
  expect_true(all(is.na(x$X3)))
  expect_true(all(is.na(x$X7)))




  ##############################################################
  ## FALSE TRUE TRUE

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = FALSE, skipEmptyRows = TRUE, colNames = TRUE)
  expect_equal(nrow(x), 5L - 1L)
  expect_equal(ncol(x), 8L)


  ## NA columns
  expect_true(all(is.na(x$X1)))
  expect_true(all(is.na(x$X2)))
  expect_true(all(is.na(x$X3)))
  expect_true(all(is.na(x$X7)))



  ##############################################################
  ## TRUE FALSE FALSE

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = TRUE, skipEmptyRows = FALSE, colNames = FALSE)
  expect_equal(nrow(x), 6L)
  expect_equal(ncol(x), 4L)

  ## NA rows
  expect_true(all(is.na(x[3, ])))



  ##############################################################
  ## TRUE FALSE TRUE

  x <- read.xlsx(xlsxFile = fl, sheet = 4, skipEmptyCols = TRUE, skipEmptyRows = FALSE, colNames = TRUE)
  expect_equal(nrow(x), 6L - 1L)
  expect_equal(ncol(x), 4L)


  ## NA rows
  expect_true(all(is.na(x[2, ])))
})

context("v3.0.0 Bug Fixes")



test_that("read.xlsx bug fixes", {
  file <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
  df <- read.xlsx(file, sheet = 1, rows = 1:2, cols = 1)
  expect_equal(df, data.frame("Var1" = TRUE))

  df <- read.xlsx(file, sheet = 1, rows = 1, cols = 1, colNames = FALSE)
  expect_equal(df, data.frame("X1" = "Var1", stringsAsFactors = FALSE))
})

context("Readind and Writing Logicals")



test_that("TRUE, FALSE, NA", {
  curr_wd <- getwd()
  fileName <- file.path(tempdir(), "T_F_NA.xlsx")

  x <- iris
  x$Species <- as.character(x$Species)
  x$all_t <- TRUE
  x$all_f <- FALSE
  x$tf <- sample(c(TRUE, FALSE), size = 150, replace = TRUE)
  x$t_na <- sample(c(TRUE, NA), size = 150, replace = TRUE)
  x$f_na <- sample(c(FALSE, NA), size = 150, replace = TRUE)
  x$tf_na <- sample(c(TRUE, FALSE, NA), size = 150, replace = TRUE)

  wb <- write.xlsx(x, file = fileName, colNames = TRUE)

  y <- read.xlsx(fileName, sheet = 1)
  expect_equal(x, y)

  ## T becomes false TRUE and NA exist in a columns
  expect_equal(x$t_na, y$t_na)
  expect_equal(x$f_na, y$f_na)

  expect_equal(is.na(x$f_na), is.na(y$f_na))
  expect_equal(is.na(x$tf_na), is.na(y$tf_na))

  ## From Workbook
  y <- read.xlsx(wb, sheet = 1)
  expect_equal(x, y)


  ## T becomes false TRUE and NA exist in a columns
  expect_equal(x$t_na, y$t_na)
  expect_equal(x$f_na, y$f_na)

  expect_equal(is.na(x$f_na), is.na(y$f_na))
  expect_equal(is.na(x$tf_na), is.na(y$tf_na))

  expect_equal(object = getwd(), curr_wd)
  unlink(fileName, recursive = TRUE, force = TRUE)
})

context("Reading from workbook is identical to reading from file readTest.xlsx")



test_that("Reading example workbook readTest.xlsx", {
  curr_wd <- getwd()
  xlsxFile <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
  wb <- loadWorkbook(xlsxFile)

  ## sheet 1
  sheet <- 1
  x <- read.xlsx(xlsxFile, sheet)
  y <- read.xlsx(wb, sheet)
  expect_equal(dim(x), c(10, 7))
  expect_equal(dim(y), c(10, 7))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, detectDates = TRUE)
  y <- read.xlsx(wb, sheet, detectDates = TRUE)
  expect_equal(dim(x), c(10, 7))
  expect_equal(dim(y), c(10, 7))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, startRow = 3, colNames = FALSE, detectDates = TRUE)
  y <- read.xlsx(wb, sheet, startRow = 3, colNames = FALSE, detectDates = TRUE)
  expect_equal(dim(x), c(9, 5))
  expect_equal(dim(y), c(9, 5))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, rows = 2:4, colNames = TRUE, detectDates = TRUE)
  y <- read.xlsx(wb, sheet, rows = 2:4, colNames = TRUE, detectDates = TRUE)
  expect_equal(dim(x), c(2, 6))
  expect_equal(dim(y), c(2, 6))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, rows = 2:4, colNames = FALSE, detectDates = TRUE)
  y <- read.xlsx(wb, sheet, rows = 2:4, colNames = FALSE, detectDates = TRUE)
  expect_equal(dim(x), c(3, 6))
  expect_equal(dim(y), c(3, 6))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, rows = 2:4, colNames = FALSE, detectDates = FALSE)
  y <- read.xlsx(wb, sheet, rows = 2:4, colNames = FALSE, detectDates = FALSE)
  expect_equal(dim(x), c(3, 6))
  expect_equal(dim(y), c(3, 6))
  expect_equal(x, y)


  x <- read.xlsx(xlsxFile, sheet, colNames = FALSE, detectDates = FALSE, cols = 2:4)
  y <- read.xlsx(wb, sheet, colNames = FALSE, detectDates = FALSE, cols = 2:4)
  expect_equal(dim(x), c(9, 2))
  expect_equal(dim(y), c(9, 2))
  expect_equal(x, y)




  ## sheet 2
  sheet <- 2
  x <- read.xlsx(xlsxFile, sheet)
  y <- read.xlsx(wb, sheet)
  expect_equal(dim(x), c(33, 9))
  expect_equal(dim(y), c(33, 9))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, startRow = 3, colNames = FALSE, detectDates = TRUE)
  y <- read.xlsx(wb, sheet, startRow = 3, colNames = FALSE, detectDates = TRUE)
  expect_equal(dim(x), c(32, 9))
  expect_equal(dim(y), c(32, 9))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, rows = 2:4, colNames = TRUE, detectDates = TRUE)
  y <- read.xlsx(wb, sheet, rows = 2:4, colNames = TRUE, detectDates = TRUE)
  expect_equal(dim(x), c(2, 9))
  expect_equal(dim(y), c(2, 9))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, rows = 2:4, colNames = FALSE, detectDates = TRUE)
  y <- read.xlsx(wb, sheet, rows = 2:4, colNames = FALSE, detectDates = TRUE)
  expect_equal(dim(x), c(3, 9))
  expect_equal(dim(y), c(3, 9))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, rows = 2:4, colNames = FALSE, detectDates = FALSE)
  y <- read.xlsx(wb, sheet, rows = 2:4, colNames = FALSE, detectDates = FALSE)
  expect_equal(dim(x), c(3, 9))
  expect_equal(dim(y), c(3, 9))
  expect_equal(x, y)


  x <- read.xlsx(xlsxFile, sheet, colNames = FALSE, detectDates = FALSE, cols = 2:4)
  y <- read.xlsx(wb, sheet, colNames = FALSE, detectDates = FALSE, cols = 2:4)
  expect_equal(dim(x), c(21, 3))
  expect_equal(dim(y), c(21, 3))
  expect_equal(x, y)



  ## sheet 3
  sheet <- 3
  x <- read.xlsx(xlsxFile, sheet)
  y <- read.xlsx(wb, sheet)
  expect_equal(dim(x), c(2083, 5))
  expect_equal(dim(y), c(2083, 5))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, startRow = 3, colNames = FALSE, detectDates = TRUE)
  y <- read.xlsx(wb, sheet, startRow = 3, colNames = FALSE, detectDates = TRUE)
  expect_equal(dim(x), c(2084, 5))
  expect_equal(dim(y), c(2084, 5))
  expect_equal(x, y)

  x <- suppressWarnings(read.xlsx(xlsxFile, sheet, rows = 2:4, colNames = TRUE, detectDates = TRUE))
  y <- suppressWarnings(read.xlsx(wb, sheet, rows = 2:4, colNames = TRUE, detectDates = TRUE))
  expect_equal(dim(x), NULL)
  expect_equal(dim(y), NULL)
  expect_equal(x, y)

  x <- suppressWarnings(read.xlsx(xlsxFile, sheet, rows = 2:4, colNames = FALSE, detectDates = TRUE))
  y <- suppressWarnings(read.xlsx(wb, sheet, rows = 2:4, colNames = FALSE, detectDates = TRUE))
  expect_equal(dim(x), NULL)
  expect_equal(dim(y), NULL)
  expect_equal(x, NULL)
  expect_equal(y, NULL)
  expect_equal(x, y)

  x <- suppressWarnings(read.xlsx(xlsxFile, sheet, rows = 2:4, colNames = FALSE, detectDates = FALSE))
  y <- suppressWarnings(read.xlsx(wb, sheet, rows = 2:4, colNames = FALSE, detectDates = FALSE))
  expect_equal(dim(x), NULL)
  expect_equal(dim(y), NULL)
  expect_equal(x, NULL)
  expect_equal(y, NULL)
  expect_equal(x, y)


  x <- read.xlsx(xlsxFile, sheet, colNames = FALSE, detectDates = FALSE, cols = 2:4)
  y <- read.xlsx(wb, sheet, colNames = FALSE, detectDates = FALSE, cols = 2:4)
  expect_equal(dim(x), c(2084, 2))
  expect_equal(dim(y), c(2084, 2))
  expect_equal(x, y)



  ## sheet 5
  sheet <- 5
  x <- read.xlsx(xlsxFile, sheet)
  y <- read.xlsx(wb, sheet)
  expect_equal(dim(x), c(271, 297))
  expect_equal(dim(y), c(271, 297))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, startRow = 3, colNames = FALSE, detectDates = TRUE)
  y <- read.xlsx(wb, sheet, startRow = 3, colNames = FALSE, detectDates = TRUE)
  expect_equal(dim(x), c(270, 297))
  expect_equal(dim(y), c(270, 297))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, rows = 2:4, colNames = TRUE, detectDates = TRUE)
  y <- read.xlsx(wb, sheet, rows = 2:4, colNames = TRUE, detectDates = TRUE)
  expect_equal(dim(x), c(2, 297))
  expect_equal(dim(y), c(2, 297))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, rows = 2:4, colNames = FALSE, detectDates = TRUE)
  y <- read.xlsx(wb, sheet, rows = 2:4, colNames = FALSE, detectDates = TRUE)
  expect_equal(dim(x), c(3, 297))
  expect_equal(dim(y), c(3, 297))
  expect_equal(x, y)

  x <- read.xlsx(xlsxFile, sheet, rows = 2:4, colNames = FALSE, detectDates = FALSE)
  y <- read.xlsx(wb, sheet, rows = 2:4, colNames = FALSE, detectDates = FALSE)
  expect_equal(dim(x), c(3, 297))
  expect_equal(dim(y), c(3, 297))
  expect_equal(x, y)


  x <- read.xlsx(xlsxFile, sheet, colNames = FALSE, detectDates = FALSE, cols = 2:4)
  y <- read.xlsx(wb, sheet, colNames = FALSE, detectDates = FALSE, cols = 2:4)
  expect_equal(dim(x), c(272, 3))
  expect_equal(dim(y), c(272, 3))
  expect_equal(x, y)

  expect_equal(object = getwd(), curr_wd)
})






test_that("Load read - Skip Empty rows/cols", {
  curr_wd <- getwd()
  xlsxFile <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
  wb <- loadWorkbook(xlsxFile)


  x <- read.xlsx(xlsxFile = xlsxFile, sheet = 4, skipEmptyCols = TRUE, skipEmptyRows = TRUE, colNames = FALSE)
  expect_equal(nrow(x), 5L)
  expect_equal(ncol(x), 4L)

  x <- read.xlsx(xlsxFile = xlsxFile, sheet = 4, skipEmptyCols = TRUE, skipEmptyRows = TRUE, colNames = TRUE)
  expect_equal(nrow(x), 5L - 1L)
  expect_equal(ncol(x), 4L)


  ##############################################################
  ## FALSE FALSE FALSE

  x <- read.xlsx(xlsxFile = xlsxFile, sheet = 4, skipEmptyCols = FALSE, skipEmptyRows = FALSE, colNames = FALSE)
  expect_equal(nrow(x), 6L)
  expect_equal(ncol(x), 8L)

  ## NA columns
  expect_true(all(is.na(x$X1)))
  expect_true(all(is.na(x$X2)))
  expect_true(all(is.na(x$X3)))
  expect_true(all(is.na(x$X7)))

  ## NA rows
  expect_true(all(is.na(x[3, ])))



  ##############################################################
  ## FALSE FALSE TRUE

  x <- read.xlsx(xlsxFile = xlsxFile, sheet = 4, skipEmptyCols = FALSE, skipEmptyRows = FALSE, colNames = TRUE)
  expect_equal(nrow(x), 6L - 1L)
  expect_equal(ncol(x), 8L)


  ## NA columns
  expect_true(all(is.na(x$X1)))
  expect_true(all(is.na(x$X2)))
  expect_true(all(is.na(x$X3)))
  expect_true(all(is.na(x$X7)))

  ## NA rows
  expect_true(all(is.na(x[2, ])))



  ##############################################################
  ## FALSE TRUE FALSE

  x <- read.xlsx(xlsxFile = xlsxFile, sheet = 4, skipEmptyCols = FALSE, skipEmptyRows = TRUE, colNames = FALSE)
  expect_equal(nrow(x), 5L)
  expect_equal(ncol(x), 8L)

  ## NA columns
  expect_true(all(is.na(x$X1)))
  expect_true(all(is.na(x$X2)))
  expect_true(all(is.na(x$X3)))
  expect_true(all(is.na(x$X7)))




  ##############################################################
  ## FALSE TRUE TRUE

  x <- read.xlsx(xlsxFile = xlsxFile, sheet = 4, skipEmptyCols = FALSE, skipEmptyRows = TRUE, colNames = TRUE)
  expect_equal(nrow(x), 5L - 1L)
  expect_equal(ncol(x), 8L)


  ## NA columns
  expect_true(all(is.na(x$X1)))
  expect_true(all(is.na(x$X2)))
  expect_true(all(is.na(x$X3)))
  expect_true(all(is.na(x$X7)))



  ##############################################################
  ## TRUE FALSE FALSE

  x <- read.xlsx(xlsxFile = xlsxFile, sheet = 4, skipEmptyCols = TRUE, skipEmptyRows = FALSE, colNames = FALSE)
  expect_equal(nrow(x), 6L)
  expect_equal(ncol(x), 4L)

  ## NA rows
  expect_true(all(is.na(x[3, ])))



  ##############################################################
  ## TRUE FALSE TRUE

  x <- read.xlsx(xlsxFile = xlsxFile, sheet = 4, skipEmptyCols = TRUE, skipEmptyRows = FALSE, colNames = TRUE)
  expect_equal(nrow(x), 6L - 1L)
  expect_equal(ncol(x), 4L)


  ## NA rows
  expect_true(all(is.na(x[2, ])))

  expect_equal(object = getwd(), curr_wd)
})




context("Reading from workbook is identical to reading from file read_failure_test.xlsx")

test_that("Reading example workbook read_failure_test.xlsx", {
  curr_wd <- getwd()
  fl <- system.file("extdata", "read_failure_test.xlsx", package = "openxlsx")
  wb <- loadWorkbook(fl)

  x <- read.xlsx(fl, sheet = 1, skipEmptyCols = TRUE)
  y <- read.xlsx(wb, sheet = 1, skipEmptyCols = TRUE)
  expect_true(all.equal(x, y))

  x <- read.xlsx(fl, sheet = 1, skipEmptyCols = FALSE)
  y <- read.xlsx(wb, sheet = 1, skipEmptyCols = FALSE)
  expect_true(all.equal(x, y))

  expect_equal(object = getwd(), curr_wd)
})

genBaseColStyle <- function(cc) {
  colStyle <- createStyle()
  specialFormat <- TRUE

  if ("date" %in% cc) {
    colStyle <- createStyle(numFmt = "date")
  } else if (any(c("posixlt", "posixct", "posixt") %in% cc)) {
    colStyle <- createStyle(numFmt = "longdate")
  } else if ("currency" %in% cc) {
    colStyle$numFmt <- list(numFmtId = "164", "formatCode" = "&quot;$&quot;#,##0.00")
  } else if ("accounting" %in% cc) {
    colStyle$numFmt <- list(numFmtId = "44")
  } else if ("hyperlink" %in% cc) {
    colStyle$fontColour <- list(theme = "10")
  } else if ("percentage" %in% cc) {
    colStyle$numFmt <- list(numFmtId = "10")
  } else if ("scientific" %in% cc) {
    colStyle$numFmt <- list(numFmtId = "11")
  } else if (any(c("3", "comma") %in% cc)) {
    colStyle$numFmt <- list(numFmtId = "3")
  } else if ("numeric" %in% cc && !grepl("[^0\\.,#\\$\\* %]", openxlsx_getOp("numFmt"))) {
    colStyle$numFmt <- list(numFmtId = 9999, formatCode = openxlsx_getOp("numFmt"))
  } else {
    colStyle$numFmt <- list(numFmtId = "0")
    specialFormat <- FALSE
  }

  list(
    style = colStyle,
    specialFormat = specialFormat
  )
}



Workbook$methods(surroundingBorders = function(
  colClasses, 
  sheet, 
  startRow, 
  startCol,
  nRow, 
  nCol, 
  borderColour, 
  borderStyle, 
  borderType
) {
  sheet <- sheet_names[[validateSheet(sheet)]]
  ## steps
  # get column class
  # get corresponding base style
  
  for (i in 1:nCol) {
    tmp <- genBaseColStyle(colClasses[[i]])
    
    colStyle <- tmp$style
    specialFormat <- tmp$specialFormat

    ## create style objects
    sTop <- colStyle$copy()
    sMid <- colStyle$copy()
    sBot <- colStyle$copy()

    ## First column
    if (i == 1) {
      if (nRow == 1 & nCol == 1) {

        ## All
        sTop$borderTop <- borderStyle
        sTop$borderTopColour <- borderColour

        sTop$borderBottom <- borderStyle
        sTop$borderBottomColour <- borderColour

        sTop$borderLeft <- borderStyle
        sTop$borderLeftColour <- borderColour

        sTop$borderRight <- borderStyle
        sTop$borderRightColour <- borderColour

        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sTop,
            "sheet" = sheet,
            "rows" = startRow,
            "cols" = startCol
          )
        ))
      } else if (nCol == 1) {

        ## Top
        sTop$borderLeft <- borderStyle
        sTop$borderLeftColour <- borderColour

        sTop$borderTop <- borderStyle
        sTop$borderTopColour <- borderColour

        sTop$borderRight <- borderStyle
        sTop$borderRightColour <- borderColour

        ## Middle
        sMid$borderLeft <- borderStyle
        sMid$borderLeftColour <- borderColour

        sMid$borderRight <- borderStyle
        sMid$borderRightColour <- borderColour

        ## Bottom
        sBot$borderBottom <- borderStyle
        sBot$borderBottomColour <- borderColour

        sBot$borderLeft <- borderStyle
        sBot$borderLeftColour <- borderColour

        sBot$borderRight <- borderStyle
        sBot$borderRightColour <- borderColour

        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sTop,
            "sheet" = sheet,
            "rows" = startRow,
            "cols" = startCol
          )
        ))

        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sMid,
            "sheet" = sheet,
            "rows" = (startRow + 1L):(startRow + nRow - 2L), # 2nd -> 2nd to last
            "cols" = rep.int(startCol, nRow - 2L)
          )
        ))

        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sBot,
            "sheet" = sheet,
            "rows" = startRow + nRow - 1L,
            "cols" = startCol
          )
        ))
      } else if (nRow == 1) {

        ## All
        sTop$borderTop <- borderStyle
        sTop$borderTopColour <- borderColour

        sTop$borderBottom <- borderStyle
        sTop$borderBottomColour <- borderColour

        sTop$borderLeft <- borderStyle
        sTop$borderLeftColour <- borderColour

        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sTop,
            "sheet" = sheet,
            "rows" = startRow,
            "cols" = startCol
          )
        ))
      } else {

        ## Top
        sTop$borderLeft <- borderStyle
        sTop$borderLeftColour <- borderColour

        sTop$borderTop <- borderStyle
        sTop$borderTopColour <- borderColour

        ## Middle
        sMid$borderLeft <- borderStyle
        sMid$borderLeftColour <- borderColour

        ## Bottom
        sBot$borderLeft <- borderStyle
        sBot$borderLeftColour <- borderColour

        sBot$borderBottom <- borderStyle
        sBot$borderBottomColour <- borderColour

        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sTop,
            "sheet" = sheet,
            "rows" = startRow,
            "cols" = startCol
          )
        ))

        if (nRow > 2) {
          styleObjects <<- append(styleObjects, list(
            list(
              "style" = sMid,
              "sheet" = sheet,
              "rows" = (startRow + 1L):(startRow + nRow - 2L), # 2nd -> 2nd to last
              "cols" = rep.int(startCol, nRow - 2L)
            )
          ))
        }

        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sBot,
            "sheet" = sheet,
            "rows" = startRow + nRow - 1L,
            "cols" = startCol
          )
        ))
      }
    } else if (i == nCol) {
      if (nRow == 1) {

        ## All
        sTop$borderTop <- borderStyle
        sTop$borderTopColour <- borderColour

        sTop$borderBottom <- borderStyle
        sTop$borderBottomColour <- borderColour

        sTop$borderRight <- borderStyle
        sTop$borderRightColour <- borderColour

        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sTop,
            "sheet" = sheet,
            "rows" = startRow,
            "cols" = startCol + nCol - 1L
          )
        ))
      } else {

        ## Top
        sTop$borderRight <- borderStyle
        sTop$borderRightColour <- borderColour

        sTop$borderTop <- borderStyle
        sTop$borderTopColour <- borderColour

        ## Middle
        sMid$borderRight <- borderStyle
        sMid$borderRightColour <- borderColour

        ## Bottom
        sBot$borderRight <- borderStyle
        sBot$borderRightColour <- borderColour

        sBot$borderBottom <- borderStyle
        sBot$borderBottomColour <- borderColour

        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sTop,
            "sheet" = sheet,
            "rows" = startRow,
            "cols" = startCol + nCol - 1L
          )
        ))

        if (nRow > 2) {
          styleObjects <<- append(styleObjects, list(
            list(
              "style" = sMid,
              "sheet" = sheet,
              "rows" = (startRow + 1L):(startRow + nRow - 2L), # 2nd -> 2nd to last
              "cols" = rep.int(startCol + nCol - 1L, nRow - 2L)
            )
          ))
        }


        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sBot,
            "sheet" = sheet,
            "rows" = startRow + nRow - 1L,
            "cols" = startCol + nCol - 1L
          )
        ))
      }
    } else { ## inside columns

      if (nRow == 1) {

        ## Top
        sTop$borderTop <- borderStyle
        sTop$borderTopColour <- borderColour

        ## Bottom
        sTop$borderBottom <- borderStyle
        sTop$borderBottomColour <- borderColour

        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sTop,
            "sheet" = sheet,
            "rows" = startRow,
            "cols" = startCol + i - 1L
          )
        ))
      } else {

        ## Top
        sTop$borderTop <- borderStyle
        sTop$borderTopColour <- borderColour

        ## Bottom
        sBot$borderBottom <- borderStyle
        sBot$borderBottomColour <- borderColour

        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sTop,
            "sheet" = sheet,
            "rows" = startRow,
            "cols" = startCol + i - 1L
          )
        ))

        ## Middle
        if (specialFormat) {
          styleObjects <<- append(styleObjects, list(
            list(
              "style" = sMid,
              "sheet" = sheet,
              "rows" = (startRow + 1L):(startRow + nRow - 2L), # 2nd -> 2nd to last
              "cols" = rep.int(startCol + i - 1L, nRow - 2L)
            )
          ))
        }

        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sBot,
            "sheet" = sheet,
            "rows" = startRow + nRow - 1L,
            "cols" = startCol + i - 1L
          )
        ))
      }
    } ## End of if(i == 1), i == NCol, else inside columns
  } ## End of loop through columns


  invisible(0)
})

Workbook$methods(rowBorders = function(
  colClasses,
  sheet,
  startRow,
  startCol, 
  nRow,
  nCol,
  borderColour,
  borderStyle, 
  borderType
) {
  sheet <- sheet_names[[validateSheet(sheet)]]
  ## steps
  # get column class
  # get corresponding base style

  for (i in 1:nCol) {
    tmp <- genBaseColStyle(colClasses[[i]])
    sTop <- tmp$style

    ## First column
    if (i == 1) {
      if (nCol == 1) {

        ## All borders (rows and surrounding)
        sTop$borderTop <- borderStyle
        sTop$borderTopColour <- borderColour

        sTop$borderBottom <- borderStyle
        sTop$borderBottomColour <- borderColour

        sTop$borderLeft <- borderStyle
        sTop$borderLeftColour <- borderColour

        sTop$borderRight <- borderStyle
        sTop$borderRightColour <- borderColour
      } else {

        ## Top, Left, Bottom
        sTop$borderTop <- borderStyle
        sTop$borderTopColour <- borderColour

        sTop$borderBottom <- borderStyle
        sTop$borderBottomColour <- borderColour

        sTop$borderLeft <- borderStyle
        sTop$borderLeftColour <- borderColour
      }
    } else if (i == nCol) {

      ## Top, Right, Bottom
      sTop$borderTop <- borderStyle
      sTop$borderTopColour <- borderColour

      sTop$borderBottom <- borderStyle
      sTop$borderBottomColour <- borderColour

      sTop$borderRight <- borderStyle
      sTop$borderRightColour <- borderColour
    } else { ## inside columns

      ## Top, Middle, Bottom
      sTop$borderTop <- borderStyle
      sTop$borderTopColour <- borderColour

      sTop$borderBottom <- borderStyle
      sTop$borderBottomColour <- borderColour
    } ## End of if(i == 1), i == NCol, else inside columns

    styleObjects <<- append(styleObjects, list(
      list(
        "style" = sTop,
        "sheet" = sheet,
        "rows" = (startRow):(startRow + nRow - 1L),
        "cols" = rep(startCol + i - 1L, nRow)
      )
    ))
  } ## End of loop through columns


  invisible(0)
})


Workbook$methods(columnBorders = function(
  colClasses,
  sheet, 
  startRow, 
  startCol, 
  nRow, 
  nCol, 
  borderColour, 
  borderStyle, 
  borderType
) {
  sheet <- sheet_names[[validateSheet(sheet)]]
  ## steps
  # get column class
  # get corresponding base style

  for (i in 1:nCol) {
    tmp <- genBaseColStyle(colClasses[[i]])
    colStyle <- tmp$style
    specialFormat <- tmp$specialFormat

    ## create style objects
    sTop <- colStyle$copy()
    sMid <- colStyle$copy()
    sBot <- colStyle$copy()

    if (nRow == 1) {

      ## Top
      sTop$borderTop <- borderStyle
      sTop$borderTopColour <- borderColour

      sTop$borderBottom <- borderStyle
      sTop$borderBottomColour <- borderColour

      sTop$borderLeft <- borderStyle
      sTop$borderLeftColour <- borderColour

      sTop$borderRight <- borderStyle
      sTop$borderRightColour <- borderColour

      styleObjects <<- append(styleObjects, list(
        list(
          "style" = sTop,
          "sheet" = sheet,
          "rows" = startRow,
          "cols" = startCol + i - 1L
        )
      ))
    } else {

      ## Top
      sTop$borderTop <- borderStyle
      sTop$borderTopColour <- borderColour

      sTop$borderLeft <- borderStyle
      sTop$borderLeftColour <- borderColour

      sTop$borderRight <- borderStyle
      sTop$borderRightColour <- borderColour

      ## Middle
      sMid$borderLeft <- borderStyle
      sMid$borderLeftColour <- borderColour

      sMid$borderRight <- borderStyle
      sMid$borderRightColour <- borderColour

      ## Bottom
      sBot$borderBottom <- borderStyle
      sBot$borderBottomColour <- borderColour

      sBot$borderLeft <- borderStyle
      sBot$borderLeftColour <- borderColour

      sBot$borderRight <- borderStyle
      sBot$borderRightColour <- borderColour

      colInd <- startCol + i - 1L

      styleObjects <<- append(styleObjects, list(
        list(
          "style" = sTop,
          "sheet" = sheet,
          "rows" = startRow,
          "cols" = colInd
        )
      ))

      if (nRow > 2) {
        styleObjects <<- append(styleObjects, list(
          list(
            "style" = sMid,
            "sheet" = sheet,
            "rows" = (startRow + 1L):(startRow + nRow - 2L),
            "cols" = rep(colInd, nRow - 2L)
          )
        ))
      }

      styleObjects <<- append(styleObjects, list(
        list(
          "style" = sBot,
          "sheet" = sheet,
          "rows" = startRow + nRow - 1L,
          "cols" = colInd
        )
      ))
    }
  } ## End of loop through columns


  invisible(0)
})


Workbook$methods(allBorders = function(
  colClasses,
  sheet,
  startRow,
  startCol,
  nRow,
  nCol,
  borderColour,
  borderStyle,
  borderType
) {
  sheet <- sheet_names[[validateSheet(sheet)]]
  ## steps
  # get column class
  # get corresponding base style

  for (i in 1:nCol) {
    tmp <- genBaseColStyle(colClasses[[i]])
    sTop <- tmp$style

    ## All borders
    sTop$borderTop <- borderStyle
    sTop$borderTopColour <- borderColour

    sTop$borderBottom <- borderStyle
    sTop$borderBottomColour <- borderColour

    sTop$borderLeft <- borderStyle
    sTop$borderLeftColour <- borderColour

    sTop$borderRight <- borderStyle
    sTop$borderRightColour <- borderColour

    styleObjects <<- append(styleObjects, list(
      list(
        "style" = sTop,
        "sheet" = sheet,
        "rows" = (startRow):(startRow + nRow - 1L),
        "cols" = rep(startCol + i - 1L, nRow)
      )
    ))
  } ## End of loop through columns


  invisible(0)
})

# Assertions for parameter validates
# These should be used at the beginning of functions to stop execution early

assert_class <- function(x, class, or_null = FALSE) {
  sx <- as.character(substitute(x))
  ok <- inherits(x, class)
  
  if (or_null) {
    ok <- ok | is.null(x)
    class <- c(class, "null")
  }
  
  if (!ok) {
    msg <- sprintf("%s must be of class %s", sx, paste(class, collapse = " or "))
    stop(msg, call. = FALSE)
  }
}

assert_length <- function(x, n) {
  stopifnot(is.integer(n))
  if (length(x) != n) {
    msg <- sprintf("%s must be of length %iL", substitute(x), n)
    stop(msg, call. = FALSE)
  }
}

assert_true_false1 <- function(x) {
  if (!is_true_false(x)) {
    stop(substitute(x), " must be TRUE or FALSE", call. = FALSE)
  }
}

assert_true_false <- function(x) {
  ok <- is.logical(x) & !is.na(x)
  if (!ok) {
    stop(substitute(x), " must be a logical vector with NAs", call. = FALSE)
  }
}

assert_character1 <- function(x, scalar = FALSE) {
  ok <- is.character(x) && length(x) == 1L
  
  if (scalar) {
    ok <- ok & nchar(x) == 1L
  }
  
  if (!ok) {
    stop(substitute(x), " must be a character vector of length 1L", call. = FALSE)
  }
}

assert_unique <- function(x, case_sensitive = TRUE) {
  msg <- paste0(substitute(x), " must be a unique vector")
  
  if (!case_sensitive) {
    x <- tolower(x)
    msg <- paste0(msg, " (case sensitive)")
  }
  
  if (anyDuplicated(x) != 0L) {
    stop(msg, call. = FALSE)
  }
}

assert_numeric1 <- function(x, scalar = FALSE) {
  msg <- paste0(substitute(x), " must be a ")
  ok <- is.numeric(x) & length(x) == 1L
  
  if (scalar) {
    ok <- ok && nchar(x) == 1L
    msg <- paste0(msg, "single number")
  } else {
    msg <- paste0(msg, "numeric vector of length 1L")
  }
  
  if (!ok) {
    stop(msg, call. = FALSE)
  }
}

# validates ---------------------------------------------------------------

validate_StyleName <- function(x) {
  m <- valid_StyleNames[match(tolower(x), valid_StyleNames_low)]
  if (anyNA(m)) {
    stop(
      "Invalid table style: ", 
      paste0(sprintf("'%s'", x[is.na(m)]), collapse = ", "),
      call. = FALSE
    )
  }
  m
}

valid_StyleNames <- c("none", paste0("TableStyleLight", 1:21), paste0("TableStyleMedium", 1:28), paste0("TableStyleDark", 1:11))
valid_StyleNames_low <- tolower(valid_StyleNames)

#' @name writeData
#' @title Write an object to a worksheet
#' @author Alexander Walker
#' @import stringi
#' @description Write an object to worksheet with optional styling.
#' @param wb A Workbook object containing a worksheet.
#' @param sheet The worksheet to write to. Can be the worksheet index or name.
#' @param x Object to be written. For classes supported look at the examples.
#' @param startCol A vector specifying the starting column to write to.
#' @param startRow A vector specifying the starting row to write to.
#' @param array A bool if the function written is of type array
#' @param xy An alternative to specifying `startCol` and
#' `startRow` individually.  A vector of the form
#' `c(startCol, startRow)`.
#' @param colNames If `TRUE`, column names of x are written.
#' @param rowNames If `TRUE`, data.frame row names of x are written.
#' @param row.names,col.names Deprecated, please use `rowNames`, `colNames` instead
#' @param headerStyle Custom style to apply to column names.
#' @param borders Either "`none`" (default), "`surrounding`",
#' "`columns`", "`rows`" or *respective abbreviations*.  If
#' "`surrounding`", a border is drawn around the data.  If "`rows`",
#' a surrounding border is drawn with a border around each row. If
#' "`columns`", a surrounding border is drawn with a border between
#' each column. If "`all`" all cell borders are drawn.
#' @param borderColour Colour of cell border.  A valid colour (belonging to `colours()` or a hex colour code, eg see [here](https://www.w3schools.com/colors/colors_picker.asp)).
#' @param borderStyle Border line style
#' \describe{
#'    \item{**none**}{ no border}
#'    \item{**thin**}{ thin border}
#'    \item{**medium**}{ medium border}
#'    \item{**dashed**}{ dashed border}
#'    \item{**dotted**}{ dotted border}
#'    \item{**thick**}{ thick border}
#'    \item{**double**}{ double line border}
#'    \item{**hair**}{ hairline border}
#'    \item{**mediumDashed**}{ medium weight dashed border}
#'    \item{**dashDot**}{ dash-dot border}
#'    \item{**mediumDashDot**}{ medium weight dash-dot border}
#'    \item{**dashDotDot**}{ dash-dot-dot border}
#'    \item{**mediumDashDotDot**}{ medium weight dash-dot-dot border}
#'    \item{**slantDashDot**}{ slanted dash-dot border}
#'   }
#' @param withFilter If `TRUE` or `NA`, add filters to the column name row. NOTE can only have one filter per worksheet.
#' @param keepNA If `TRUE`, NA values are converted to #N/A (or `na.string`, if not NULL) in Excel, else NA cells will be empty.
#' @param na.string If not NULL, and if `keepNA` is `TRUE`, NA values are converted to this string in Excel.
#' @param name If not NULL, a named region is defined.
#' @param sep Only applies to list columns. The separator used to collapse list columns to a character vector e.g. sapply(x$list_column, paste, collapse = sep).
#' @seealso [writeDataTable()]
#' @export writeData
#' @details Formulae written using writeFormula to a Workbook object will not get picked up by read.xlsx().
#' This is because only the formula is written and left to Excel to evaluate the formula when the file is opened in Excel.
#' @rdname writeData
#' @return invisible(0)
#' @examples
#'
#' ## See formatting vignette for further examples.
#'
#' ## Options for default styling (These are the defaults)
#' options("openxlsx.borderColour" = "black")
#' options("openxlsx.borderStyle" = "thin")
#' options("openxlsx.dateFormat" = "mm/dd/yyyy")
#' options("openxlsx.datetimeFormat" = "yyyy-mm-dd hh:mm:ss")
#' options("openxlsx.numFmt" = NULL)
#'
#' ## Change the default border colour to #4F81BD
#' options("openxlsx.borderColour" = "#4F81BD")
#'
#'
#' #####################################################################################
#' ## Create Workbook object and add worksheets
#' wb <- createWorkbook()
#'
#' ## Add worksheets
#' addWorksheet(wb, "Cars")
#' addWorksheet(wb, "Formula")
#'
#'
#' x <- mtcars[1:6, ]
#' writeData(wb, "Cars", x, startCol = 2, startRow = 3, rowNames = TRUE)
#'
#' #####################################################################################
#' ## Bordering
#'
#' writeData(wb, "Cars", x,
#'   rowNames = TRUE, startCol = "O", startRow = 3,
#'   borders = "surrounding", borderColour = "black"
#' ) ## black border
#'
#' writeData(wb, "Cars", x,
#'   rowNames = TRUE,
#'   startCol = 2, startRow = 12, borders = "columns"
#' )
#'
#' writeData(wb, "Cars", x,
#'   rowNames = TRUE,
#'   startCol = "O", startRow = 12, borders = "rows"
#' )
#'
#'
#' #####################################################################################
#' ## Header Styles
#'
#' hs1 <- createStyle(
#'   fgFill = "#DCE6F1", halign = "CENTER", textDecoration = "italic",
#'   border = "Bottom"
#' )
#'
#' writeData(wb, "Cars", x,
#'   colNames = TRUE, rowNames = TRUE, startCol = "B",
#'   startRow = 23, borders = "rows", headerStyle = hs1, borderStyle = "dashed"
#' )
#'
#'
#' hs2 <- createStyle(
#'   fontColour = "#ffffff", fgFill = "#4F80BD",
#'   halign = "center", valign = "center", textDecoration = "bold",
#'   border = "TopBottomLeftRight"
#' )
#'
#' writeData(wb, "Cars", x,
#'   colNames = TRUE, rowNames = TRUE,
#'   startCol = "O", startRow = 23, borders = "columns", headerStyle = hs2
#' )
#'
#'
#'
#'
#' #####################################################################################
#' ## Hyperlinks
#' ## - vectors/columns with class 'hyperlink' are written as hyperlinks'
#'
#' v <- rep("https://CRAN.R-project.org/", 4)
#' names(v) <- paste0("Hyperlink", 1:4) # Optional: names will be used as display text
#' class(v) <- "hyperlink"
#' writeData(wb, "Cars", x = v, xy = c("B", 32))
#'
#'
#' #####################################################################################
#' ## Formulas
#' ## - vectors/columns with class 'formula' are written as formulas'
#'
#' df <- data.frame(
#'   x = 1:3, y = 1:3,
#'   z = paste0(paste0("A", 1:3 + 1L), paste0("B", 1:3 + 1L), sep = " + "),
#'   stringsAsFactors = FALSE
#' )
#'
#' class(df$z) <- c(class(df$z), "formula")
#'
#' writeData(wb, sheet = "Formula", x = df)
#'
#'
#' #####################################################################################
#' ## Save workbook
#' ## Open in excel without saving file: openXL(wb)
#' \dontrun{
#' saveWorkbook(wb, "writeDataExample.xlsx", overwrite = TRUE)
#' }
writeData <- function(
  wb,
  sheet,
  x,
  startCol     = 1,
  startRow     = 1,
  array        = FALSE,
  xy           = NULL,
  colNames     = TRUE,
  rowNames     = FALSE,
  headerStyle  = openxlsx_getOp("headerStyle"),
  borders      = openxlsx_getOp("borders", "none"),
  borderColour = openxlsx_getOp("borderColour", "black"),
  borderStyle  = openxlsx_getOp("borderStyle", "thin"),
  withFilter   = openxlsx_getOp("withFilter", FALSE),
  keepNA       = openxlsx_getOp("keepNA", FALSE),
  na.string    = openxlsx_getOp("na.string"),
  name         = NULL,
  sep          = ", ",
  col.names,
  row.names
) {

  x <- force(x)

  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (!missing(row.names)) {
    warning("Please use 'rowNames' instead of 'row.names'", call. = FALSE)
    rowNames <- row.names
  }

  if (!missing(col.names)) {
    warning("Please use 'colNames' instead of 'col.names'", call. = FALSE)
    colNames <- col.names
  }

  # Set NULLs
  borders      <- borders      %||% "none"
  borderColour <- borderColour %||% "black"
  borderStyle  <- borderStyle  %||% "thin"
  withFilter   <- withFilter   %||% FALSE
  keepNA       <- keepNA       %||% FALSE

  if (is.null(x)) {
    return(invisible(0))
  }

  ## All input conversions/validations
  if (!is.null(xy)) {
    if (length(xy) != 2) {
      stop("xy parameter must have length 2")
    }
    startCol <- xy[[1]]
    startRow <- xy[[2]]
  }

  ## convert startRow and startCol
  if (!is.numeric(startCol)) {
    startCol <- convertFromExcelRef(startCol)
  }

  startRow <- as.integer(startRow)

  assert_class(wb, "Workbook")
  assert_true_false(colNames)
  assert_true_false(rowNames)
  assert_character1(sep)
  assert_class(headerStyle, "Style", or_null = TRUE)

  ## borderColours validation
  borderColour <- validateColour(borderColour, "Invalid border colour")
  borderStyle <- validateBorderStyle(borderStyle)[[1]]

  ## special case - vector of hyperlinks
  hlinkNames <- NULL
  if (inherits(x, "hyperlink")) {
    hlinkNames <- names(x)
    colNames <- FALSE
  }

  ## special case - formula
  if (inherits(x, "formula")) {
    x <- data.frame("X" = x, stringsAsFactors = FALSE)
    class(x[[1]]) <- ifelse(array, "array_formula", "formula")
    colNames <- FALSE
  }

  ## named region
  if (!is.null(name)) { ## validate name
    ex_names <- regmatches(wb$workbook$definedNames, regexpr('(?<=name=")[^"]+', wb$workbook$definedNames, perl = TRUE))
    ex_names <- replaceXMLEntities(ex_names)

    if (name %in% ex_names) {
      stop(sprintf("Named region with name '%s' already exists!", name))
    } else if (grepl("^[A-Z]{1,3}[0-9]+$", name)) {
      stop("name cannot look like a cell reference.")
    }
  }

  if (is.vector(x) || is.factor(x) || inherits(x, "Date")) {
    colNames <- FALSE
  } ## this will go to coerce.default and rowNames will be ignored

  ## Coerce to data.frame
  x <- openxlsxCoerce(x = x, rowNames = rowNames)

  nCol <- ncol(x)
  nRow <- nrow(x)

  ## If no rows and not writing column names return as nothing to write
  if (nRow == 0 && !colNames) {
    return(invisible(0))
  }

  ## If no columns and not writing row names return as nothing to write
  if (nCol == 0 && !rowNames) {
    return(invisible(0))
  }

  colClasses <- lapply(x, function(x) tolower(class(x)))
  colClasss2 <- colClasses
  colClasss2[vapply(
    colClasses,
    function(i) inherits(i, "formula") & inherits(i, "hyperlink"),
    NA
  )] <- "formula"

  if (is.numeric(sheet)) {
    sheetX <- wb$validateSheet(sheet)
  } else {
    sheetX <- wb$validateSheet(replaceXMLEntities(sheet))
    sheet <- replaceXMLEntities(sheet)
  }

  if (wb$isChartSheet[[sheetX]]) {
    stop("Cannot write to chart sheet.")
  }

  ## Check not overwriting existing table headers
  wb$check_overwrite_tables(
    sheet                   = sheet,
    new_rows                = c(startRow, startRow + nRow - 1L + colNames),
    new_cols                = c(startCol, startCol + nCol - 1L),
    check_table_header_only = TRUE,
    error_msg               = "Cannot overwrite table headers. Avoid writing over the header row or see getTables() & removeTables() to remove the table object."
  )

  ## write autoFilter, can only have a single filter per worksheet
  if (withFilter) {
    coords <- data.frame(
      x = c(startRow, startRow + nRow + colNames - 1L),
      y = c(startCol, startCol + nCol - 1L)
    )

    ref <- stri_join(getCellRefs(coords), collapse = ":")
    wb$worksheets[[sheetX]]$autoFilter <- sprintf('<autoFilter ref="%s"/>', ref)
    l <- convert_to_excel_ref(cols = unlist(coords[, 2]), LETTERS = LETTERS)
    dfn <- sprintf("'%s'!%s", names(wb)[sheetX], stri_join("$", l, "$", coords[, 1], collapse = ":"))

    dn <- sprintf('<definedName name="_xlnm._FilterDatabase" localSheetId="%s" hidden="1">%s</definedName>', sheetX - 1L, dfn)

    if (length(wb$workbook$definedNames) > 0) {
      ind <- grepl('name="_xlnm._FilterDatabase"', wb$workbook$definedNames)
      if (length(ind) > 0) {
        wb$workbook$definedNames[ind] <- dn
      }
    } else {
      wb$workbook$definedNames <- dn
    }
  }

  ## write data.frame
  wb$writeData(
    df         = x,
    colNames   = colNames,
    sheet      = sheet,
    startCol   = startCol,
    startRow   = startRow,
    colClasses = colClasss2,
    hlinkNames = hlinkNames,
    keepNA     = keepNA,
    na.string  = na.string,
    list_sep   = sep
  )

  ## header style
  if (inherits(headerStyle, "Style") && colNames) {
    addStyle(
      wb         = wb,
      sheet      = sheet,
      style      = headerStyle,
      rows       = startRow,
      cols       = 0:(nCol - 1) + startCol,
      gridExpand = TRUE,
      stack      = TRUE
    )
  }

  ## If we don't have any rows to write return
  if (nRow == 0) {
    return(invisible(0))
  }

  ## named region
  if (!is.null(name)) {
    ref1 <- stri_join("$", convert_to_excel_ref(cols = startCol, LETTERS = LETTERS), "$", startRow)
    ref2 <- stri_join("$", convert_to_excel_ref(cols = startCol + nCol - 1L, LETTERS = LETTERS), "$", startRow + nRow - 1L + colNames)
    wb$createNamedRegion(ref1 = ref1, ref2 = ref2, name = name, sheet = wb$sheet_names[wb$validateSheet(sheet)])
  }

  ## hyperlink style, if no borders
  borders <- match.arg(borders, c("none", "surrounding", "rows", "columns", "all"))

  if (borders == "none") {
    invisible(
      classStyles(
        wb,
        sheet      = sheet,
        startRow   = startRow,
        startCol   = startCol,
        colNames   = colNames,
        nRow       = nrow(x),
        colClasses = colClasses,
        stack      = TRUE
      )
    )
  } else if (borders == "surrounding") {
    wb$surroundingBorders(
      colClasses,
      sheet        = sheet,
      startRow     = startRow + colNames,
      startCol     = startCol,
      nRow         = nRow, nCol = nCol,
      borderColour = list("rgb" = borderColour),
      borderStyle  = borderStyle
    )
  } else if (borders == "rows") {
    wb$rowBorders(
      colClasses,
      sheet        = sheet,
      startRow     = startRow + colNames,
      startCol     = startCol,
      nRow         = nRow, nCol = nCol,
      borderColour = list("rgb" = borderColour),
      borderStyle  = borderStyle
    )
  } else if (borders == "columns") {
    wb$columnBorders(
      colClasses,
      sheet        = sheet,
      startRow     = startRow + colNames,
      startCol     = startCol,
      nRow         = nRow, nCol = nCol,
      borderColour = list("rgb" = borderColour),
      borderStyle  = borderStyle
    )
  } else if (borders == "all") {
    wb$allBorders(
      colClasses,
      sheet        = sheet,
      startRow     = startRow + colNames,
      startCol     = startCol,
      nRow         = nRow, nCol = nCol,
      borderColour = list("rgb" = borderColour),
      borderStyle  = borderStyle
    )
  }

  invisible(0)
}


#' @name writeFormula
#' @title Write a character vector as an Excel Formula
#' @author Alexander Walker
#' @description Write a a character vector containing Excel formula to a worksheet.
#' @details Currently only the english version of functions are supported. Please don't use the local translation.
#' The examples below show a small list of possible formulas:
#' \itemize{
#'     \item{SUM(B2:B4)}
#'     \item{AVERAGE(B2:B4)}
#'     \item{MIN(B2:B4)}
#'     \item{MAX(B2:B4)}
#'     \item{...}
#'
#' }
#' @param wb A Workbook object containing a worksheet.
#' @param sheet The worksheet to write to. Can be the worksheet index or name.
#' @param x A character vector.
#' @param startCol A vector specifying the starting column to write to.
#' @param startRow A vector specifying the starting row to write to.
#' @param array A bool if the function written is of type array
#' @param xy An alternative to specifying `startCol` and
#' `startRow` individually.  A vector of the form
#' `c(startCol, startRow)`.
#' @seealso [writeData()] [makeHyperlinkString()]
#' @export writeFormula
#' @rdname writeFormula
#' @examples
#'
#' ## There are 3 ways to write a formula
#'
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet 1")
#' writeData(wb, "Sheet 1", x = iris)
#'
#' ## SEE int2col() to convert int to Excel column label
#'
#' ## 1. -  As a character vector using writeFormula
#'
#' v <- c("SUM(A2:A151)", "AVERAGE(B2:B151)") ## skip header row
#' writeFormula(wb, sheet = 1, x = v, startCol = 10, startRow = 2)
#' writeFormula(wb, 1, x = "A2 + B2", startCol = 10, startRow = 10)
#'
#'
#' ## 2. - As a data.frame column with class "formula" using writeData
#'
#' df <- data.frame(
#'   x = 1:3,
#'   y = 1:3,
#'   z = paste(paste0("A", 1:3 + 1L), paste0("B", 1:3 + 1L), sep = " + "),
#'   z2 = sprintf("ADDRESS(1,%s)", 1:3),
#'   stringsAsFactors = FALSE
#' )
#'
#' class(df$z) <- c(class(df$z), "formula")
#' class(df$z2) <- c(class(df$z2), "formula")
#'
#' addWorksheet(wb, "Sheet 2")
#' writeData(wb, sheet = 2, x = df)
#'
#'
#'
#' ## 3. - As a vector with class "formula" using writeData
#'
#' v2 <- c("SUM(A2:A4)", "AVERAGE(B2:B4)", "MEDIAN(C2:C4)")
#' class(v2) <- c(class(v2), "formula")
#'
#' writeData(wb, sheet = 2, x = v2, startCol = 10, startRow = 2)
#'
#' ## Save workbook
#' \dontrun{
#' saveWorkbook(wb, "writeFormulaExample.xlsx", overwrite = TRUE)
#' }
#'
#'
#' ## 4. - Writing internal hyperlinks
#'
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet1")
#' addWorksheet(wb, "Sheet2")
#' writeFormula(wb, "Sheet1", x = '=HYPERLINK("#Sheet2!B3", "Text to Display - Link to Sheet2")')
#'
#' ## Save workbook
#' \dontrun{
#' saveWorkbook(wb, "writeFormulaHyperlinkExample.xlsx", overwrite = TRUE)
#' }
#'
writeFormula <- function(
  wb,
  sheet,
  x,
  startCol = 1,
  startRow = 1,
  array = FALSE,
  xy = NULL
) {

  if (!is.character(x)) {
    stop("x must be a character vector.")
  }

  dfx <- data.frame("X" = x, stringsAsFactors = FALSE)
  class(dfx$X) <- c("character", ifelse(array, "array_formula", "formula"))

  if (any(grepl("^(=|)HYPERLINK\\(", x, ignore.case = TRUE))) {
    class(dfx$X) <- c("character", "formula", "hyperlink")
  }

  writeData(
    wb       = wb,
    sheet    = sheet,
    x        = dfx,
    startCol = startCol,
    startRow = startRow,
    array    = array,
    xy       = xy,
    colNames = FALSE,
    rowNames = FALSE
  )

  invisible(0)
}

#' `as.character.formula()`
#'
#' This function exists to prevent conflicts with `as.character.formula` methods
#' from other packages
#'
#' @inheritParams base::as.character
#' @param ... Not implemented
#' @returns `base::as.character.default(x)`
#' @export
as.character.formula <- function(x, ...) {
  base::as.character.default(x)
}

#' If NULL then ...
#'
#' Replace NULL
#'
#' @param x A value to check
#' @param y A value to substitute if x is null
#' @examples
#' \dontrun{
#' x <- NULL
#' x <- x %||% "none"
#' x <- x %||% NA
#' }
#'
#' @name if_null_then
`%||%` <- function(x, y) if (is.null(x)) y else x

is_not_class <- function(x, class) {
  !(inherits(x, class) | is.null(x))
}

is_true_false <- function(x) {
  is.logical(x) && length(x) == 1L && !is.na(x)
}

pair_rc <- function(r, c) {
  # Assumes that there can't be more than 10M columns
  # Excel allows 16,384
  # Google Sheets allows 18,278
  r+c*1e-7
}

do_call_params <- function(fun, params, ..., .map = FALSE) {
  fun <- match.fun(fun)
  call_params <- c(list(...), params[names(params) %in% names(formals(fun))])
  call_params <- lapply(call_params, function(x) if (is.object(x)) list(x) else x)

  call_fun <- if (.map) {
    function(...) mapply(fun, ..., MoreArgs = NULL, SIMPLIFY = FALSE, USE.NAMES = FALSE)
  } else {
    fun
  }

  do.call(call_fun, call_params)
}

# sets temporary options
# option() returns the original values
get_set_options <- function() {
  options(
    # increase scipen to avoid writing in scientific
    scipen = 200,
    OutDec = ".",
    digits = 22
  )
}


#' helper function to create tempory directory for testing purpose
#' @param name for the temp file
#' @export
temp_xlsx <- function(name = "temp_xlsx") {
  tempfile(pattern = paste0(name, "_"), fileext = ".xlsx")
}

test_that("buildWorkbook() accepts tableName [187]", {
  x <- data.frame(a = 1, b = 2)
  
  # default name
  wb <- buildWorkbook(x, asTable = TRUE)
  expect_equal(attr(wb$tables, "tableName"), "Table3")
  
  # define 1/2 table name
  wb <- buildWorkbook(x, asTable = TRUE, tableName = "table_x")
  expect_equal(attr(wb$tables, "tableName"), "table_x")
  
  # define 2/2 table names
  wb <- buildWorkbook(list(x, x), asTable = TRUE, tableName = c("table_x", "table_y"))
  expect_equal(attr(wb$tables, "tableName"), c("table_x", "table_y"))
  
  # try to define 1/2 table names
  expect_error(buildWorkbook(list(x, x), asTable = TRUE, tableName = "table_x"))
})

test_that("row.name and col.name are deprecated", {
  x <- data.frame(a = 1)
  
  expect_warning(
    buildWorkbook(x, file = temp_xlsx(), row.names = TRUE, overwrite = TRUE),
    "Please use 'rowNames' instead of 'row.names'"
  )
  
  expect_warning(
    buildWorkbook(x, file = temp_xlsx(), row.names = TRUE, overwrite = TRUE, asTable = TRUE),
    "Please use 'rowNames' instead of 'row.names'"
  )
  
  expect_warning(
    buildWorkbook(x, file = temp_xlsx(), col.names = TRUE, overwrite = TRUE),
    "Please use 'colNames' instead of 'col.names'"
  )
  
  expect_warning(
    buildWorkbook(x, file = temp_xlsx(), col.names = TRUE, overwrite = TRUE, asTable = TRUE),
    "Please use 'colNames' instead of 'col.names'"
  )
})

#' @name createWorkbook
#' @title Create a new Workbook object
#' @description Create a new Workbook object
#' @param creator Creator of the workbook (your name). Defaults to login username
#' @param title Workbook properties title
#' @param subject Workbook properties subject
#' @param category Workbook properties category
#' @author Alexander Walker
#' @return Workbook object
#' @export
#' @seealso [loadWorkbook()]
#' @seealso [saveWorkbook()]
#' @import methods
#' @examples
#' ## Create a new workbook
#' wb <- createWorkbook()
#'
#' ## Save workbook to working directory
#' \dontrun{
#' saveWorkbook(wb, file = "createWorkbookExample.xlsx", overwrite = TRUE)
#' }
#'
#' ## Set Workbook properties
#' wb <- createWorkbook(
#'   creator = "Me",
#'   title = "title here",
#'   subject = "this & that",
#'   category = "something"
#' )
createWorkbook <- function(creator = ifelse(.Platform$OS.type == "windows", Sys.getenv("USERNAME"), Sys.getenv("USER")),
                           title = NULL,
                           subject = NULL,
                           category = NULL) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  ## check all inputs are valid
  if (length(creator) > 1) creator <- creator[[1]]
  if (length(creator) == 0) creator <- ""
  if (!"character" %in% class(creator)) creator <- ""

  if (length(title) > 1) title <- title[[1]]
  if (length(subject) > 1) subject <- subject[[1]]
  if (length(category) > 1) category <- category[[1]]

  if (!is.null(title)) {
    if (!"character" %in% class(title)) {
      stop("title must be a string")
    }
  }

  if (!is.null(subject)) {
    if (!"character" %in% class(subject)) {
      stop("subject must be a string")
    }
  }

  if (!is.null(category)) {
    if (!"character" %in% class(category)) {
      stop("category must be a string")
    }
  }

  invisible(Workbook$new(creator = creator, title = title, subject = subject, category = category))
}


#' @name saveWorkbook
#' @title save Workbook to file
#' @description save a Workbook object to file
#' @author Alexander Walker, Philipp Schauberger
#' @param wb A Workbook object to write to file
#' @param file A character string naming an xlsx file
#' @param overwrite If `TRUE`, overwrite any existing file.
#' @param returnValue  If `TRUE`, returns  `TRUE` in case of a success, else `FALSE`.
#' If flag is `FALSE`, then no return value is returned.
#' @seealso [createWorkbook()]
#' @seealso [addWorksheet()]
#' @seealso [loadWorkbook()]
#' @seealso [writeData()]
#' @seealso [writeDataTable()]
#' @export
#' @examples
#' ## Create a new workbook and add a worksheet
#' wb <- createWorkbook("Creator of workbook")
#' addWorksheet(wb, sheetName = "My first worksheet")
#'
#' ## Save workbook to working directory
#' \dontrun{
#' saveWorkbook(wb, file = "saveWorkbookExample.xlsx", overwrite = TRUE)
#' }
saveWorkbook <- function(wb, file, overwrite = FALSE, returnValue = FALSE) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  if (!is.logical(overwrite)) {
    overwrite <- FALSE
  }

  if (!is.logical(returnValue)) {
    returnValue <- FALSE
  }

  if (file.exists(file) && !overwrite) {
    stop("File already exists!")
  }

  xlsx_file <- wb$saveWorkbook()

  result <- file.copy(from = xlsx_file, to = file, overwrite = overwrite, copy.mode = FALSE)

  ## delete temporary dir
  unlink(dirname(xlsx_file), force = TRUE, recursive = TRUE)
  if (returnValue == FALSE) {
    invisible(1)
  } else {
    return(result)
  }
}


#' @name mergeCells
#' @title Merge cells within a worksheet
#' @description Merge cells within a worksheet
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param cols Columns to merge
#' @param rows corresponding rows to merge
#' @details As merged region must be rectangular, only min and max of cols and rows are used.
#' @author Alexander Walker
#' @seealso [removeCellMerge()]
#' @export
#' @examples
#' ## Create a new workbook
#' wb <- createWorkbook()
#'
#' ## Add a worksheet
#' addWorksheet(wb, "Sheet 1")
#' addWorksheet(wb, "Sheet 2")
#'
#' ## Merge cells: Row 2 column C to F (3:6)
#' mergeCells(wb, "Sheet 1", cols = 2, rows = 3:6)
#'
#' ## Merge cells:Rows 10 to 20 columns A to J (1:10)
#' mergeCells(wb, 1, cols = 1:10, rows = 10:20)
#'
#' ## Intersecting merges
#' mergeCells(wb, 2, cols = 1:10, rows = 1)
#' mergeCells(wb, 2, cols = 5:10, rows = 2)
#' mergeCells(wb, 2, cols = c(1, 10), rows = 12) ## equivalent to 1:10 as only min/max are used
#' # mergeCells(wb, 2, cols = 1, rows = c(1,10)) # Throws error because intersects existing merge
#'
#' ## remove merged cells
#' removeCellMerge(wb, 2, cols = 1, rows = 1) # removes any intersecting merges
#' mergeCells(wb, 2, cols = 1, rows = 1:10) # Now this works
#'
#' ## Save workbook
#' \dontrun{
#' saveWorkbook(wb, "mergeCellsExample.xlsx", overwrite = TRUE)
#' }
mergeCells <- function(wb, sheet, cols, rows) {
  od <- getOption("OutDec")
  options("OutDec" = ".")
  on.exit(expr = options("OutDec" = od), add = TRUE)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  if (!is.numeric(cols)) {
    cols <- convertFromExcelRef(cols)
  }

  wb$mergeCells(sheet, startRow = min(rows), endRow = max(rows), startCol = min(cols), endCol = max(cols))
}



#' @name int2col
#' @title Convert integer to Excel column
#' @description Converts an integer to an Excel column label.
#' @param x A numeric vector
#' @export
#' @examples
#' int2col(1:10)
int2col <- function(x) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (!is.numeric(x)) {
    stop("x must be numeric.")
  }

  convert_to_excel_ref(cols = x, LETTERS = LETTERS)
}

#' @name col2int
#' @title Convert Excel column to integer
#' @description Converts an Excel column label to an integer.
#' @param x A character vector
#' @export
#' @examples
#' col2int(LETTERS)
col2int <- function(x) {

  if (!is.character(x)) {
    stop("x must be character")
  }

  as.integer(sapply(x, cell_ref_to_col))
}


#' @name removeCellMerge
#' @title Create a new Workbook object
#' @description Unmerges any merged cells that intersect
#' with the region specified by, min(cols):max(cols) X min(rows):max(rows)
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param cols vector of column indices
#' @param rows vector of row indices
#' @author Alexander Walker
#' @export
#' @seealso [mergeCells()]
removeCellMerge <- function(wb, sheet, cols, rows) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  cols <- convertFromExcelRef(cols)
  rows <- as.integer(rows)

  wb$removeCellMerge(sheet, startRow = min(rows), endRow = max(rows), startCol = min(cols), endCol = max(cols))
}


#' @name sheets
#' @title Returns names of worksheets.
#' @description DEPRECATED. Use names().
#' @param wb A workbook object
#' @return Name of worksheet(s) for a given index
#' @author Alexander Walker
#' @seealso [names()] to rename a worksheet in a Workbook
#' @details DEPRECATED. Use [names()]
#' @export
#' @examples
#'
#' ## Create a new workbook
#' wb <- createWorkbook()
#'
#' ## Add some worksheets
#' addWorksheet(wb, "Worksheet Name")
#' addWorksheet(wb, "This is worksheet 2")
#' addWorksheet(wb, "The third worksheet")
#'
#' ## Return names of sheets, can not be used for assignment.
#' names(wb)
#' # openXL(wb)
#'
#' names(wb) <- c("A", "B", "C")
#' names(wb)
#' # openXL(wb)
sheets <- function(wb) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  nms <- wb$sheet_names
  nms <- replaceXMLEntities(nms)

  return(nms)
}



#' @name addWorksheet
#' @title Add a worksheet to a workbook
#' @description Add a worksheet to a Workbook object
#' @author Alexander Walker
#' @param wb A Workbook object to attach the new worksheet
#' @param sheetName A name for the new worksheet
#' @param gridLines A logical. If `FALSE`, the worksheet grid lines will be hidden.
#' @param tabColour Colour of the worksheet tab. A valid colour (belonging to colours()) or a valid hex colour beginning with "#"
#' @param zoom A numeric between 10 and 400. Worksheet zoom level as a percentage.
#' @param header document header. Character vector of length 3 corresponding to positions left, center, right. Use NA to skip a position.
#' @param footer document footer. Character vector of length 3 corresponding to positions left, center, right. Use NA to skip a position.
#' @param evenHeader document header for even pages.
#' @param evenFooter document footer for even pages.
#' @param firstHeader document header for first page only.
#' @param firstFooter document footer for first page only.
#' @param visible If FALSE, sheet is hidden else visible.
#' @param paperSize An integer corresponding to a paper size. See ?pageSetup for details.
#' @param orientation One of "portrait" or "landscape"
#' @param hdpi Horizontal DPI. Can be set with options("openxlsx.dpi" = X) or options("openxlsx.hdpi" = X)
#' @param vdpi Vertical DPI. Can be set with options("openxlsx.dpi" = X) or options("openxlsx.vdpi" = X)
#' @details Headers and footers can contain special tags
#' \describe{
#'   \item{**&\[Page\]**}{ Page number}
#'   \item{**&\[Pages\]**}{ Number of pages}
#'   \item{**&\[Date\]**}{ Current date}
#'   \item{**&\[Time\]**}{ Current time}
#'   \item{**&\[Path\]**}{ File path}
#'   \item{**&\[File\]**}{ File name}
#'   \item{**&\[Tab\]**}{ Worksheet name}
#' }
#' @return XML tree
#' @export
#' @examples
#' ## Create a new workbook
#' wb <- createWorkbook("Fred")
#'
#' ## Add 3 worksheets
#' addWorksheet(wb, "Sheet 1")
#' addWorksheet(wb, "Sheet 2", gridLines = FALSE)
#' addWorksheet(wb, "Sheet 3", tabColour = "red")
#' addWorksheet(wb, "Sheet 4", gridLines = FALSE, tabColour = "#4F81BD")
#'
#' ## Headers and Footers
#' addWorksheet(wb, "Sheet 5",
#'   header = c("ODD HEAD LEFT", "ODD HEAD CENTER", "ODD HEAD RIGHT"),
#'   footer = c("ODD FOOT RIGHT", "ODD FOOT CENTER", "ODD FOOT RIGHT"),
#'   evenHeader = c("EVEN HEAD LEFT", "EVEN HEAD CENTER", "EVEN HEAD RIGHT"),
#'   evenFooter = c("EVEN FOOT RIGHT", "EVEN FOOT CENTER", "EVEN FOOT RIGHT"),
#'   firstHeader = c("TOP", "OF FIRST", "PAGE"),
#'   firstFooter = c("BOTTOM", "OF FIRST", "PAGE")
#' )
#'
#' addWorksheet(wb, "Sheet 6",
#'   header = c("&[Date]", "ALL HEAD CENTER 2", "&[Page] / &[Pages]"),
#'   footer = c("&[Path]&[File]", NA, "&[Tab]"),
#'   firstHeader = c(NA, "Center Header of First Page", NA),
#'   firstFooter = c(NA, "Center Footer of First Page", NA)
#' )
#'
#' addWorksheet(wb, "Sheet 7",
#'   header = c("ALL HEAD LEFT 2", "ALL HEAD CENTER 2", "ALL HEAD RIGHT 2"),
#'   footer = c("ALL FOOT RIGHT 2", "ALL FOOT CENTER 2", "ALL FOOT RIGHT 2")
#' )
#'
#' addWorksheet(wb, "Sheet 8",
#'   firstHeader = c("FIRST ONLY L", NA, "FIRST ONLY R"),
#'   firstFooter = c("FIRST ONLY L", NA, "FIRST ONLY R")
#' )
#'
#' ## Need data on worksheet to see all headers and footers
#' writeData(wb, sheet = 5, 1:400)
#' writeData(wb, sheet = 6, 1:400)
#' writeData(wb, sheet = 7, 1:400)
#' writeData(wb, sheet = 8, 1:400)
#'
#' ## Save workbook
#' \dontrun{
#' saveWorkbook(wb, "addWorksheetExample.xlsx", overwrite = TRUE)
#' }
addWorksheet <- function(wb,
                         sheetName,
                         gridLines = openxlsx_getOp("gridLines", TRUE),
                         tabColour = NULL,
                         zoom = 100,
                         header = openxlsx_getOp("header"),
                         footer = openxlsx_getOp("footer"),
                         evenHeader = openxlsx_getOp("evenHeader"),
                         evenFooter = openxlsx_getOp("evenFooter"),
                         firstHeader = openxlsx_getOp("firstHeader"),
                         firstFooter = openxlsx_getOp("firstFooter"),
                         visible = TRUE,
                         paperSize = openxlsx_getOp("paperSize", 9),
                         orientation = openxlsx_getOp("orientation", "portrait"),
                         vdpi = openxlsx_getOp("vdpi", 300),
                         hdpi = openxlsx_getOp("hdpi", 300)) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (inherits(wb, "list")) {
    wb <- wb[[1]]
  }

  if (!inherits(wb, "Workbook")) {
    stop("wb must be a Workbook", call. = FALSE)
  }

  # Set NULL defaults
  gridLines <- gridLines %||% TRUE
  paperSize <- paperSize %||% 9
  orientation <- orientation %||% "portrait"
  vdpi <- vdpi %||% 300
  hdpi <- hdpi %||% 300

  if (tolower(sheetName) %in% tolower(wb$sheet_names)) {
    stop(paste0("A worksheet by the name '", sheetName, "' already exists! Sheet names must be unique case-insensitive."))
  }

  if (!is.logical(gridLines) || length(gridLines) > 1) {
    stop("gridLines must be a logical of length 1.")
  }

  if (nchar(sheetName) > 31) {
    stop(paste0("sheetName '", sheetName, "' too long! Max length is 31 characters."))
  }

  if (!is.null(tabColour)) {
    tabColour <- validateColour(tabColour, "Invalid tabColour in addWorksheet.")
  }

  if (!is.numeric(zoom)) {
    stop("zoom must be numeric")
  }

  if (!is.character(sheetName)) {
    sheetName <- as.character(sheetName)
  }

  if (!is.null(header) && length(header) != 3) {
    stop("header must have length 3 where elements correspond to positions: left, center, right.")
  }

  if (!is.null(footer) && length(footer) != 3) {
    stop("footer must have length 3 where elements correspond to positions: left, center, right.")
  }

  if (!is.null(evenHeader) && length(evenHeader) != 3) {
    stop("evenHeader must have length 3 where elements correspond to positions: left, center, right.")
  }

  if (!is.null(evenFooter) && length(evenFooter) != 3) {
    stop("evenFooter must have length 3 where elements correspond to positions: left, center, right.")
  }

  if (!is.null(firstHeader) && length(firstHeader) != 3) {
    stop("firstHeader must have length 3 where elements correspond to positions: left, center, right.")
  }

  if (!is.null(firstFooter) && length(firstFooter) != 3) {
    stop("firstFooter must have length 3 where elements correspond to positions: left, center, right.")
  }

  visible <- tolower(visible[1])
  if (!visible %in% c("true", "false", "hidden", "visible", "veryhidden")) {
    stop("visible must be one of: TRUE, FALSE, 'hidden', 'visible', 'veryHidden'")
  }

  orientation <- tolower(orientation)
  if (!orientation %in% c("portrait", "landscape")) {
    stop("orientation must be 'portrait' or 'landscape'.")
  }

  vdpi <- as.integer(vdpi)
  if (is.na(vdpi)) {
    stop("vdpi must be numeric")
  }

  hdpi <- as.integer(hdpi)
  if (is.na(hdpi)) {
    stop("hdpi must be numeric")
  }



  ## Invalid XML characters
  sheetName <- replaceIllegalCharacters(sheetName)

  invisible(wb$addWorksheet(
    sheetName = sheetName,
    showGridLines = gridLines,
    tabColour = tabColour,
    zoom = zoom[1],
    oddHeader = headerFooterSub(header),
    oddFooter = headerFooterSub(footer),
    evenHeader = headerFooterSub(evenHeader),
    evenFooter = headerFooterSub(evenFooter),
    firstHeader = headerFooterSub(firstHeader),
    firstFooter = headerFooterSub(firstFooter),
    visible = visible,
    paperSize = paperSize,
    orientation = orientation,
    vdpi = vdpi,
    hdpi = hdpi
  ))
}

#' @name cloneWorksheet
#' @title Clone a worksheet to a workbook
#' @description Clone a worksheet to a Workbook object
#' @author Reinhold Kainhofer
#' @param wb A Workbook object to attach the new worksheet
#' @param sheetName A name for the new worksheet
#' @param clonedSheet The name of the existing worksheet to be cloned.
#' @return XML tree
#' @export
#' @examples
#' ## Create a new workbook
#' wb <- createWorkbook("Fred")
#'
#' ## Add 3 worksheets
#' addWorksheet(wb, "Sheet 1")
#' cloneWorksheet(wb, "Sheet 2", clonedSheet = "Sheet 1")
#'
#' ## Save workbook
#' \dontrun{
#' saveWorkbook(wb, "cloneWorksheetExample.xlsx", overwrite = TRUE)
#' }
cloneWorksheet <- function(wb, sheetName, clonedSheet) {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  if (tolower(sheetName) %in% tolower(wb$sheet_names)) {
    stop("A worksheet by that name already exists! Sheet names must be unique case-insensitive.")
  }

  if (nchar(sheetName) > 31) {
    stop("sheetName too long! Max length is 31 characters.")
  }

  if (!is.character(sheetName)) {
    sheetName <- as.character(sheetName)
  }

  ## Invalid XML characters
  sheetName <- replaceIllegalCharacters(sheetName)

  invisible(wb$cloneWorksheet(sheetName = sheetName, clonedSheet = clonedSheet))
}


#' @name renameWorksheet
#' @title Rename a worksheet
#' @description Rename a worksheet
#' @author Alexander Walker
#' @param wb A Workbook object containing a worksheet
#' @param sheet The name or index of the worksheet to rename
#' @param newName The new name of the worksheet. No longer than 31 chars.
#' @details DEPRECATED. Use [names()]
#' @export
#' @examples
#'
#' ## Create a new workbook
#' wb <- createWorkbook("CREATOR")
#'
#' ## Add 3 worksheets
#' addWorksheet(wb, "Worksheet Name")
#' addWorksheet(wb, "This is worksheet 2")
#' addWorksheet(wb, "Not the best name")
#'
#' #' ## rename all worksheets
#' names(wb) <- c("A", "B", "C")
#'
#'
#' ## Rename worksheet 1 & 3
#' renameWorksheet(wb, 1, "New name for sheet 1")
#' names(wb)[[1]] <- "New name for sheet 1"
#' names(wb)[[3]] <- "A better name"
#'
#' ## Save workbook
#' \dontrun{
#' saveWorkbook(wb, "renameWorksheetExample.xlsx", overwrite = TRUE)
#' }
renameWorksheet <- function(wb, sheet, newName) {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  invisible(wb$setSheetName(sheet, newName))
}


#' @name convertFromExcelRef
#' @title Convert excel column name to integer index
#' @description Convert excel column name to integer index e.g. "J" to 10
#' @param col An excel column reference
#' @export
#' @examples
#' convertFromExcelRef("DOG")
#' convertFromExcelRef("COW")
#'
#' ## numbers will be removed
#' convertFromExcelRef("R22")
convertFromExcelRef <- function(col) {

  ## increase scipen to avoid writing in scientific
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  col <- toupper(col)
  charFlag <- grepl("[A-Z]", col)
  if (any(charFlag)) {
    col[charFlag] <- gsub("[0-9]", "", col[charFlag])
    d <- lapply(strsplit(col[charFlag], split = ""), function(x) match(rev(x), LETTERS))
    col[charFlag] <- unlist(lapply(seq_along(d), function(i) {
      sum(d[[i]] * (26^(
        seq_along(d[[i]]) - 1)))
    }))
  }

  col[!charFlag] <- as.integer(col[!charFlag])

  return(as.integer(col))
}



#' @name createStyle
#' @title Create a cell style
#' @description Create a new style to apply to worksheet cells
#' @author Alexander Walker
#' @seealso [addStyle()]
#' @param fontName A name of a font. Note the font name is not validated. If fontName is NULL,
#' the workbook base font is used. (Defaults to Calibri)
#' @param fontColour Colour of text in cell.  A valid hex colour beginning with "#"
#' or one of colours(). If fontColour is NULL, the workbook base font colours is used.
#' (Defaults to black)
#' @param fontSize Font size. A numeric greater than 0.
#' If fontSize is NULL, the workbook base font size is used. (Defaults to 11)
#' @param numFmt Cell formatting
#' \itemize{
#'   \item{**GENERAL**}
#'   \item{**NUMBER**}
#'   \item{**CURRENCY**}
#'   \item{**ACCOUNTING**}
#'   \item{**DATE**}
#'   \item{**LONGDATE**}
#'   \item{**TIME**}
#'   \item{**PERCENTAGE**}
#'   \item{**FRACTION**}
#'   \item{**SCIENTIFIC**}
#'   \item{**TEXT**}
#'   \item{**COMMA**  for comma separated thousands}
#'   \item{For date/datetime styling a combination of d, m, y and punctuation marks}
#'   \item{For numeric rounding use "0.00" with the preferred number of decimal places}
#' }
#'
#' @param border Cell border. A vector of "top", "bottom", "left", "right" or a single string).
#' \describe{
#'    \item{**"top"**}{ Top border}
#'    \item{**bottom**}{ Bottom border}
#'    \item{**left**}{ Left border}
#'    \item{**right**}{ Right border}
#'    \item{**TopBottom** or **c("top", "bottom")**}{ Top and bottom border}
#'    \item{**LeftRight** or **c("left", "right")**}{ Left and right border}
#'    \item{**TopLeftRight** or **c("top", "left", "right")**}{ Top, Left and right border}
#'    \item{**TopBottomLeftRight** or **c("top", "bottom", "left", "right")**}{ All borders}
#'   }
#'
#' @param borderColour Colour of cell border vector the same length as the number of sides specified in "border"
#' A valid colour (belonging to colours()) or a valid hex colour beginning with "#"
#'
#' @param borderStyle Border line style vector the same length as the number of sides specified in "border"
#' \describe{
#'    \item{**none**}{ No Border}
#'    \item{**thin**}{ thin border}
#'    \item{**medium**}{ medium border}
#'    \item{**dashed**}{ dashed border}
#'    \item{**dotted**}{ dotted border}
#'    \item{**thick**}{ thick border}
#'    \item{**double**}{ double line border}
#'    \item{**hair**}{ Hairline border}
#'    \item{**mediumDashed**}{ medium weight dashed border}
#'    \item{**dashDot**}{ dash-dot border}
#'    \item{**mediumDashDot**}{ medium weight dash-dot border}
#'    \item{**dashDotDot**}{ dash-dot-dot border}
#'    \item{**mediumDashDotDot**}{ medium weight dash-dot-dot border}
#'    \item{**slantDashDot**}{ slanted dash-dot border}
#'   }
#'
#' @param bgFill Cell background fill colour.
#' A valid colour (belonging to colours()) or a valid hex colour beginning with "#".
#' --  **Use for conditional formatting styles only.**
#' @param fgFill Cell foreground fill colour.
#' A valid colour (belonging to colours()) or a valid hex colour beginning with "#"
#'
#' @param halign
#' Horizontal alignment of cell contents
#' \describe{
#'    \item{**left**}{ Left horizontal align cell contents}
#'    \item{**right**}{ Right horizontal align cell contents}
#'    \item{**center**}{ Center horizontal align cell contents}
#'    \item{**justify**}{ Justify horizontal align cell contents}
#'   }
#'
#' @param valign A name
#' Vertical alignment of cell contents
#' \describe{
#'    \item{**top**}{ Top vertical align cell contents}
#'    \item{**center**}{ Center vertical align cell contents}
#'    \item{**bottom**}{ Bottom vertical align cell contents}
#'   }
#'
#' @param textDecoration
#' Text styling.
#' \describe{
#'    \item{**bold**}{ Bold cell contents}
#'    \item{**strikeout**}{ Strikeout cell contents}
#'    \item{**italic**}{ Italicise cell contents}
#'    \item{**underline**}{ Underline cell contents}
#'    \item{**underline2**}{ Double underline cell contents}
#'    \item{**accounting**}{ Single accounting underline cell contents}
#'    \item{**accounting2**}{ Double accounting underline cell contents}
#'   }
#'
#' @param wrapText Logical. If `TRUE` cell contents will wrap to fit in column.
#' @param textRotation Rotation of text in degrees. 255 for vertical text.
#' @param indent Horizontal indentation of cell contents.
#' @param hidden Whether the formula of the cell contents will be hidden (if worksheet protection is turned on)
#' @param locked Whether cell contents are locked (if worksheet protection is turned on)
#' @return A style object
#' @export
#' @examples
#' ## See package vignettes for further examples
#'
#' ## Modify default values of border colour and border line style
#' options("openxlsx.borderColour" = "#4F80BD")
#' options("openxlsx.borderStyle" = "thin")
#'
#' ## Size 18 Arial, Bold, left horz. aligned, fill colour #1A33CC, all borders,
#' style <- createStyle(
#'   fontSize = 18, fontName = "Arial",
#'   textDecoration = "bold", halign = "left", fgFill = "#1A33CC", border = "TopBottomLeftRight"
#' )
#'
#' ## Red, size 24, Bold, italic, underline, center aligned Font, bottom border
#' style <- createStyle(
#'   fontSize = 24, fontColour = rgb(1, 0, 0),
#'   textDecoration = c("bold", "italic", "underline"),
#'   halign = "center", valign = "center", border = "Bottom"
#' )
#'
#' # borderColour is recycled for each border or all colours can be supplied
#'
#' # colour is recycled 3 times for "Top", "Bottom" & "Right" sides.
#' createStyle(border = "TopBottomRight", borderColour = "red")
#'
#' # supply all colours
#' createStyle(border = "TopBottomLeft", borderColour = c("red", "yellow", "green"))
createStyle <- function(fontName = NULL,
                        fontSize = NULL,
                        fontColour = NULL,
                        numFmt = openxlsx_getOp("numFmt", "GENERAL"),
                        border = NULL,
                        borderColour = openxlsx_getOp("borderColour", "black"),
                        borderStyle = openxlsx_getOp("borderStyle", "thin"),
                        bgFill = NULL,
                        fgFill = NULL,
                        halign = NULL,
                        valign = NULL,
                        textDecoration = NULL,
                        wrapText = FALSE,
                        textRotation = NULL,
                        indent = NULL,
                        locked = NULL,
                        hidden = NULL) {

  ### Error checking
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  ## if num fmt is made up of dd, mm, yy
  numFmt_original <- numFmt[[1]]
  numFmt <- tolower(numFmt_original)
  validNumFmt <- c("general", "number", "currency", "accounting", "date", "longdate", "time", "percentage", "scientific", "text", "3", "4", "comma")

  if (numFmt == "date") {
    numFmt <- openxlsx_getOp("dateFormat", "date")
  } else if (numFmt == "longdate") {
    numFmt <- openxlsx_getOp("datetimeFormat", "longdate")
  } else if (!numFmt %in% validNumFmt) {
    numFmt <- replaceIllegalCharacters(numFmt_original)
  }

  numFmtMapping <- list(
    list(numFmtId = 0), # GENERAL
    list(numFmtId = 2), # NUMBER
    list(numFmtId = 164, formatCode = "&quot;$&quot;#,##0.00"), ## CURRENCY
    list(numFmtId = 44), # ACCOUNTING
    list(numFmtId = 14), # DATE
    list(numFmtId = 166, formatCode = "yyyy/mm/dd hh:mm:ss"), # LONGDATE
    list(numFmtId = 167), # TIME
    list(numFmtId = 10), # PERCENTAGE
    list(numFmtId = 11), # SCIENTIFIC
    list(numFmtId = 49), # TEXT
    list(numFmtId = 3),
    list(numFmtId = 4),
    list(numFmtId = 3)
  )

  names(numFmtMapping) <- validNumFmt

  ## Validate border line style
  if (!is.null(borderStyle)) {
    borderStyle <- validateBorderStyle(borderStyle)
  }

  if (!is.null(halign)) {
    halign <- tolower(halign[[1]])
    if (!halign %in% c("left", "right", "center", "justify")) {
      stop("Invalid halign argument!")
    }
  }

  if (!is.null(valign)) {
    valign <- tolower(valign[[1]])
    if (!valign %in% c("top", "bottom", "center")) {
      stop("Invalid valign argument!")
    }
  }

  if (!is.logical(wrapText)) {
    stop("Invalid wrapText")
  }

  if (!is.null(indent)) {
    if (!is.numeric(indent) && !is.integer(indent)) {
      stop("indent must be numeric")
    }
  }

  textDecoration <- tolower(textDecoration)
  if (!is.null(textDecoration)) {
    if (!all(textDecoration %in% c("bold", "strikeout", "italic", "underline", "underline2", "accounting", "accounting2", ""))) {
      stop("Invalid textDecoration!")
    }
  }

  borderColour <- validateColour(borderColour, "Invalid border colour!")

  if (!is.null(fontColour)) {
    fontColour <- validateColour(fontColour, "Invalid font colour!")
  }

  if (!is.null(fontSize)) {
    if (fontSize < 1) stop("Font size must be greater than 0!")
  }

  if (!is.null(locked)) {
    if (!is.logical(locked)) stop("Cell attribute locked must be TRUE or FALSE")
  }
  if (!is.null(hidden)) {
    if (!is.logical(hidden)) stop("Cell attribute hidden must be TRUE or FALSE")
  }





  ######################### error checking complete #############################
  style <- Style$new()

  if (!is.null(fontName)) {
    style$fontName <- list("val" = fontName)
  }

  if (!is.null(fontSize)) {
    style$fontSize <- list("val" = fontSize)
  }

  if (!is.null(fontColour)) {
    style$fontColour <- list("rgb" = fontColour)
  }

  style$fontDecoration <- toupper(textDecoration)

  ## background fill
  if (is.null(bgFill)) {
    # bgFillList <- NULL variable not used
  } else {
    bgFill <- validateColour(bgFill, "Invalid bgFill colour")
    style$fill <- append(style$fill, list(fillBg = list("rgb" = bgFill)))
  }

  ## foreground fill
  if (is.null(fgFill)) {
    # fgFillList <- NULL variable not used
  } else {
    fgFill <- validateColour(fgFill, "Invalid fgFill colour")
    style$fill <- append(style$fill, list(fillFg = list(rgb = fgFill)))
  }


  ## border
  if (!is.null(border)) {
    border <- toupper(border)
    border <- paste(border, collapse = "")

    ## find position of each side in string
    sides <- c("LEFT", "RIGHT", "TOP", "BOTTOM")
    pos <- sapply(sides, function(x) regexpr(x, border))
    pos <- pos[order(pos, decreasing = FALSE)]
    nSides <- sum(pos > 0)

    borderColour <- rep(borderColour, length.out = nSides)
    borderStyle <- rep(borderStyle, length.out = nSides)

    pos <- pos[pos > 0]

    if (length(pos) == 0) {
      stop("Unknown border argument")
    }

    names(borderColour) <- names(pos)
    names(borderStyle) <- names(pos)

    if ("LEFT" %in% names(pos)) {
      style$borderLeft <- borderStyle[["LEFT"]]
      style$borderLeftColour <- list("rgb" = borderColour[["LEFT"]])
    }

    if ("RIGHT" %in% names(pos)) {
      style$borderRight <- borderStyle[["RIGHT"]]
      style$borderRightColour <- list("rgb" = borderColour[["RIGHT"]])
    }

    if ("TOP" %in% names(pos)) {
      style$borderTop <- borderStyle[["TOP"]]
      style$borderTopColour <- list("rgb" = borderColour[["TOP"]])
    }

    if ("BOTTOM" %in% names(pos)) {
      style$borderBottom <- borderStyle[["BOTTOM"]]
      style$borderBottomColour <- list("rgb" = borderColour[["BOTTOM"]])
    }
  }

  ## other fields
  if (!is.null(halign)) {
    style$halign <- halign
  }

  if (!is.null(valign)) {
    style$valign <- valign
  }

  if (!is.null(indent)) {
    style$indent <- indent
  }

  if (wrapText) {
    style$wrapText <- TRUE
  }

  if (!is.null(textRotation)) {
    if (!is.numeric(textRotation)) {
      stop("textRotation must be numeric.")
    }

    if (textRotation < 0 && textRotation >= -90) {
      textRotation <- (textRotation * -1) + 90
    }

    style$textRotation <- round(textRotation[[1]], 0)
  }

  if (numFmt != "general") {
    if (numFmt %in% validNumFmt) {
      style$numFmt <- numFmtMapping[[numFmt[[1]]]]
    } else {
      style$numFmt <- list("numFmtId" = 165, formatCode = numFmt) ## Custom numFmt
    }
  }


  if (!is.null(locked)) {
    style$locked <- locked
  }

  if (!is.null(hidden)) {
    style$hidden <- hidden
  }

  return(style)
}



#' @name addStyle
#' @title Add a style to a set of cells
#' @description Function adds a style to a specified set of cells.
#' @author Alexander Walker
#' @param wb A Workbook object containing a worksheet.
#' @param sheet A worksheet to apply the style to.
#' @param style A style object returned from createStyle()
#' @param rows Rows to apply style to.
#' @param cols columns to apply style to.
#' @param gridExpand If `TRUE`, style will be applied to all combinations of rows and cols.
#' @param stack If `TRUE` the new style is merged with any existing cell styles.  If FALSE, any
#' existing style is replaced by the new style.
#' @seealso [createStyle()]
#' @seealso expand.grid
#' @export
#' @examples
#' ## See package vignette for more examples.
#'
#' ## Create a new workbook
#' wb <- createWorkbook("My name here")
#'
#' ## Add a worksheets
#' addWorksheet(wb, "Expenditure", gridLines = FALSE)
#'
#' ## write data to worksheet 1
#' writeData(wb, sheet = 1, USPersonalExpenditure, rowNames = TRUE)
#'
#' ## create and add a style to the column headers
#' headerStyle <- createStyle(
#'   fontSize = 14, fontColour = "#FFFFFF", halign = "center",
#'   fgFill = "#4F81BD", border = "TopBottom", borderColour = "#4F81BD"
#' )
#'
#' ## style for body
#' bodyStyle <- createStyle(border = "TopBottom", borderColour = "#4F81BD")
#' addStyle(wb, sheet = 1, bodyStyle, rows = 2:6, cols = 1:6, gridExpand = TRUE)
#' setColWidths(wb, 1, cols = 1, widths = 21) ## set column width for row names column
#' \dontrun{
#' saveWorkbook(wb, "addStyleExample.xlsx", overwrite = TRUE)
#' }
addStyle <- function(wb,
                     sheet,
                     style,
                     rows,
                     cols,
                     gridExpand = FALSE,
                     stack = FALSE) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)


  if (!is.null(style$numFmt) && length(wb$styleObjects) > 0) {
    if (style$numFmt$numFmtId == 165) {
      maxnumFmtId <- max(unlist(sapply(wb$styleObjects, function(i) {
        as.integer(
          max(c(i$style$numFmt$numFmtId, 0))
        )
      })), 165)
      style$numFmt$numFmtId <- maxnumFmtId + 1
    }
  }
  sheet <- wb$validateSheet(sheet)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  if (!"Style" %in% class(style)) {
    stop("style argument must be a Style object.")
  }

  if (!is.logical(stack)) {
    stop("stack parameter must be a logical!")
  }

  if (length(cols) == 0 || length(rows) == 0) {
    return(invisible(0))
  }

  cols <- convertFromExcelRef(cols)
  rows <- as.integer(rows)

  ## rows and cols need to be the same length
  if (gridExpand) {
    n <- length(cols)
    cols <- rep.int(cols, times = length(rows))
    rows <- rep(rows, each = n)
  } else if (length(rows) == 1 && length(cols) > 1) {
    rows <- rep.int(rows, times = length(cols))
  } else if (length(cols) == 1 && length(rows) > 1) {
    cols <- rep.int(cols, times = length(rows))
  } else if (length(rows) != length(cols)) {
    stop("Length of rows and cols must be equal.")
  }


  wb$addStyle(sheet = sheet, style = style, rows = rows, cols = cols, stack = stack)
}


#' @name getCellRefs
#' @title Return excel cell coordinates from (x,y) coordinates
#' @description Return excel cell coordinates from (x,y) coordinates
#' @author Philipp Schauberger, Alexander Walker
#' @param cellCoords A data.frame with two columns coordinate pairs.
#' @return Excel alphanumeric cell reference
#' @examples
#' getCellRefs(data.frame(1, 2))
#' # "B1"
#' getCellRefs(data.frame(1:3, 2:4))
#' # "B1" "C2" "D3"
#' @export
getCellRefs <- function(cellCoords) {
  if (!"data.frame" %in% class(cellCoords)) {
    stop("Provide a data.frame!")
  }



  if (!("numeric" %in% sapply(cellCoords[, 1], class) ||
    "integer" %in% sapply(cellCoords[, 1], class)) &&
    ("numeric" %in% sapply(cellCoords[, 2], class) ||
      "integer" %in% sapply(cellCoords[, 2], class))

  ) {
    stop("Provide a data.frame containing integers!")
  }

  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  l <- convert_to_excel_ref(cols = unlist(cellCoords[, 2]), LETTERS = LETTERS)
  paste0(l, cellCoords[, 1])
}


#' @name freezePane
#' @title Freeze a worksheet pane
#' @description Freeze a worksheet pane
#' @author Alexander Walker
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param firstActiveRow Top row of active region
#' @param firstActiveCol Furthest left column of active region
#' @param firstRow If `TRUE`, freezes the first row (equivalent to firstActiveRow = 2)
#' @param firstCol If `TRUE`, freezes the first column (equivalent to firstActiveCol = 2)
#' @export
#' @examples
#' ## Create a new workbook
#' wb <- createWorkbook("Kenshin")
#'
#' ## Add some worksheets
#' addWorksheet(wb, "Sheet 1")
#' addWorksheet(wb, "Sheet 2")
#' addWorksheet(wb, "Sheet 3")
#' addWorksheet(wb, "Sheet 4")
#'
#' ## Freeze Panes
#' freezePane(wb, "Sheet 1", firstActiveRow = 5, firstActiveCol = 3)
#' freezePane(wb, "Sheet 2", firstCol = TRUE) ## shortcut to firstActiveCol = 2
#' freezePane(wb, 3, firstRow = TRUE) ## shortcut to firstActiveRow = 2
#' freezePane(wb, 4, firstActiveRow = 1, firstActiveCol = "D")
#'
#' ## Save workbook
#' \dontrun{
#' saveWorkbook(wb, "freezePaneExample.xlsx", overwrite = TRUE)
#' }
freezePane <- function(wb, sheet, firstActiveRow = NULL, firstActiveCol = NULL, firstRow = FALSE, firstCol = FALSE) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (is.null(firstActiveRow) && is.null(firstActiveCol) && !firstRow && !firstCol) {
    return(invisible(0))
  }

  if (!is.logical(firstRow)) {
    stop("firstRow must be TRUE/FALSE")
  }

  if (!is.logical(firstCol)) {
    stop("firstCol must be TRUE/FALSE")
  }


  if (firstRow && !firstCol) {
    invisible(wb$freezePanes(sheet, firstRow = firstRow))
  } else if (firstCol && !firstRow) {
    invisible(wb$freezePanes(sheet, firstCol = firstCol))
  } else if (firstRow && firstCol) {
    invisible(wb$freezePanes(sheet, firstActiveRow = 2L, firstActiveCol = 2L))
  } else { ## else both firstRow and firstCol are FALSE

    ## Convert to numeric if column letter given
    if (!is.null(firstActiveRow)) {
      firstActiveRow <- convertFromExcelRef(firstActiveRow)
    } else {
      firstActiveRow <- 1L
    }

    if (!is.null(firstActiveCol)) {
      firstActiveCol <- convertFromExcelRef(firstActiveCol)
    } else {
      firstActiveCol <- 1L
    }

    invisible(wb$freezePanes(sheet, firstActiveRow = firstActiveRow, firstActiveCol = firstActiveCol, firstRow = firstRow, firstCol = firstCol))
  }
}


convert2EMU <- function(d, units) {
  if (grepl("in", units)) {
    d <- d * 2.54
  }

  if (grepl("mm|milli", units)) {
    d <- d / 10
  }

  return(d * 360000)
}




#' @name insertImage
#' @title Insert an image into a worksheet
#' @description Insert an image into a worksheet
#' @author Alexander Walker
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param file An image file. Valid file types are: jpeg, png, bmp
#' @param width Width of figure.
#' @param height Height of figure.
#' @param startRow Row coordinate of upper left corner of the image
#' @param startCol Column coordinate of upper left corner of the image
#' @param units Units of width and height. Can be "in", "cm" or "px"
#' @param dpi Image resolution used for conversion between units.
#' @param address An optional character string specifying an external URL, relative or absolute path to a file, or mailto string (e.g. "mailto:example@@example.com") that will be opened when the image is clicked.
#' @importFrom grDevices bmp png jpeg
#' @seealso [insertPlot()]
#' @export
#' @examples
#' ## Create a new workbook
#' wb <- createWorkbook("Ayanami")
#'
#' ## Add some worksheets
#' addWorksheet(wb, "Sheet 1")
#' addWorksheet(wb, "Sheet 2")
#' addWorksheet(wb, "Sheet 3")
#' addWorksheet(wb, "Sheet 4")
#'
#' ## Insert images
#' img <- system.file("extdata", "einstein.jpg", package = "openxlsx")
#' insertImage(wb, "Sheet 1", img, startRow = 5, startCol = 3, width = 6, height = 5)
#' insertImage(wb, 2, img, startRow = 2, startCol = 2)
#' insertImage(wb, 3, img, width = 15, height = 12, startRow = 3, startCol = "G", units = "cm")
#' insertImage(wb, 4, img, address = "https://github.com/ycphs/openxlsx")
#'
#' ## Save workbook
#' \dontrun{
#' saveWorkbook(wb, "insertImageExample.xlsx", overwrite = TRUE)
#' }
insertImage <- function(wb, sheet, file, width = 6, height = 3, startRow = 1, startCol = 1, units = "in", dpi = 300, address) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (!file.exists(file)) {
    stop("File does not exist.")
  }

  if (!grepl("\\\\|\\/", file)) {
    file <- file.path(getwd(), file, fsep = .Platform$file.sep)
  }

  units <- tolower(units)

  if (!units %in% c("cm", "in", "px")) {
    stop("Invalid units.\nunits must be one of: cm, in, px")
  }

  if (!missing(address)) {
    if (!is.character(address) || length(address) != 1 || is.na(address)) {
      stop("Invalid address. address must be a string and have a length of one.")
    } 
  }

  startCol <- convertFromExcelRef(startCol)
  startRow <- as.integer(startRow)

  ## convert to inches
  if (units == "px") {
    width <- width / dpi
    height <- height / dpi
  } else if (units == "cm") {
    width <- width / 2.54
    height <- height / 2.54
  }

  ## Convert to EMUs
  widthEMU <- as.integer(round(width * 914400L, 0)) # (EMUs per inch)
  heightEMU <- as.integer(round(height * 914400L, 0)) # (EMUs per inch)

  wb$insertImage(sheet, file = file, startRow = startRow, startCol = startCol, width = widthEMU, height = heightEMU, address = address)
}

pixels2ExcelColWidth <- function(pixels) {
  if (any(!is.numeric(pixels))) {
    stop("All elements of pixels must be numeric")
  }

  pixels[pixels == 0] <- 8.43
  pixels[pixels != 0] <- (pixels[pixels != 0] - 12) / 7 + 1

  pixels
}


#' @name setRowHeights
#' @title Set worksheet row heights
#' @description Set worksheet row heights
#' @author Alexander Walker
#' @param wb workbook object
#' @param sheet name or index of a worksheet
#' @param rows indices of rows to set height
#' @param heights heights to set rows to specified in Excel column height units
#' @param fontsize font size, optional (get base font size by default)
#' @param factor factor to manually adjust font width, e.g., for bold fonts,
#' optional
#' @param base_height basic row height, optional
#' @param extra_height additional row height per new line of text, optional
#' @param wrap wrap text of entries which exceed the column width, optional
#' @seealso [removeRowHeights()]
#' @export
#' @examples
#' ## Create a new workbook
#' wb <- createWorkbook()
#'
#' ## Add a worksheet
#' addWorksheet(wb, "Sheet")
#' sheet <- 1
#' 
#' ## Write dummy data
#' writeData(wb, sheet, "fixed w/fixed h", startCol = 1, startRow = 1)
#' writeData(wb, sheet, "fixed w/auto h ABC ABC ABC ABC ABC ABC ABC ABC ABC ABC ABC",
#'  startCol = 2, startRow = 2)
#' writeData(wb, sheet, "variable w/fixed h", startCol = 3, startRow = 3)
#' 
#' ## Set column widths and row heights
#' setColWidths(wb, sheet, cols = c(1, 2, 3, 4), widths = c(10, 20, "auto", 20))
#' setRowHeights(wb, sheet, rows = c(1, 2, 8, 4, 6), heights = c(30, "auto", 15, 15, 30))
#' 
#' ## Overwrite row 1 height
#' setRowHeights(wb, sheet, rows = 1, heights = 40)
#'
#' ## Save workbook
#' \dontrun{
#' saveWorkbook(wb, "setRowHeightsExample.xlsx", overwrite = TRUE)
#' }
setRowHeights <- function(wb, sheet, rows, heights,
                          fontsize = NULL, factor = 1.0,
                          base_height = 15, extra_height = 12, wrap = TRUE) {
  # validate sheet
  sheet <- wb$validateSheet(sheet)

  if (length(rows) > length(heights)) {
    heights <- rep(heights, length.out = length(rows))
  }

  if (length(heights) > length(rows)) {
    stop("Greater number of height values than rows.")
  }
  
  od <- getOption("OutDec")
  options(OutDec = ".")
  on.exit(expr = options(OutDec = od), add = TRUE)
  # clean duplicates
  heights <- heights[!duplicated(rows)]
  rows <- rows[!duplicated(rows)]

  # auto adjust row heights
  ida <- which(heights == "auto")
  selected <- rows[ida]
  out <- auto_heights(wb, sheet, selected, fontsize = fontsize, factor = factor,
                      base_height = base_height, extra_height = extra_height)
  cols <- out[[1]]
  new <- out[[2]]
  heights[ida] <- as.character(new)
  names(heights) <- rows
  # wrap text in cells
  if (wrap == TRUE) {
    wrap <- openxlsx::createStyle(wrapText = TRUE)
    openxlsx::addStyle(wb, sheet, wrap, rows = ida, cols = cols, gridExpand = TRUE, stack = TRUE)
  }

  wb$setRowHeights(sheet, rows, heights)
}

#' @name setColWidths
#' @title Set worksheet column widths
#' @description Set worksheet column widths to specific width or "auto".
#' @author Alexander Walker
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param cols Indices of cols to set width
#' @param widths widths to set cols to specified in Excel column width units or "auto" for automatic sizing. The widths argument is
#' recycled to the length of cols.
#' @param hidden Logical vector. If TRUE the column is hidden.
#' @param ignoreMergedCells Ignore any cells that have been merged with other cells in the calculation of "auto" column widths.
#' @details The global min and max column width for "auto" columns is set by (default values show):
#' \itemize{
#'   \item{options("openxlsx.minWidth" = 3)}
#'   \item{options("openxlsx.maxWidth" = 250)} ## This is the maximum width allowed in Excel
#' }
#'
#' NOTE: The calculation of column widths can be slow for large worksheets.
#'
#' NOTE: The `hidden` parameter may conflict with the one set in `groupColumns`; changing one will update the other.
#'
#' @seealso [removeColWidths()]
#' @export
#' @examples
#' ## Create a new workbook
#' wb <- createWorkbook()
#'
#' ## Add a worksheet
#' addWorksheet(wb, "Sheet 1")
#'
#'
#' ## set col widths
#' setColWidths(wb, 1, cols = c(1, 4, 6, 7, 9), widths = c(16, 15, 12, 18, 33))
#'
#' ## auto columns
#' addWorksheet(wb, "Sheet 2")
#' writeData(wb, sheet = 2, x = iris)
#' setColWidths(wb, sheet = 2, cols = 1:5, widths = "auto")
#'
#' ## Save workbook
#' \dontrun{
#' saveWorkbook(wb, "setColWidthsExample.xlsx", overwrite = TRUE)
#' }
#'
setColWidths <- function(wb, sheet, cols, widths = 8.43, hidden = rep(FALSE, length(cols)), ignoreMergedCells = FALSE) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  sheet <- wb$validateSheet(sheet)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  widths <- tolower(widths) ## possibly "auto"
  if (ignoreMergedCells) {
    widths[widths == "auto"] <- "auto2"
  }

  # should do nothing if the cols' length is zero
  if (length(cols) == 0L) {
    return(invisible(0))
  }

  if (length(widths) > length(cols)) {
    stop("More widths than columns supplied.")
  }

  if (length(hidden) > length(cols)) {
    stop("hidden argument is longer than cols.")
  }

  if (length(widths) < length(cols)) {
    widths <- rep(widths, length.out = length(cols))
  }

  if (length(hidden) < length(cols)) {
    hidden <- rep(hidden, length.out = length(cols))
  }

  ## Remove duplicates
  widths <- widths[!duplicated(cols)]
  hidden <- hidden[!duplicated(cols)]
  cols <- cols[!duplicated(cols)]
  cols <- convertFromExcelRef(cols)

  if (length(wb$colWidths[[sheet]]) > 0) {
    existing_cols <- names(wb$colWidths[[sheet]])
    existing_widths <- unname(wb$colWidths[[sheet]])
    existing_hidden <- attr(wb$colWidths[[sheet]], "hidden")

    ## check for existing custom widths
    flag <- existing_cols %in% cols
    if (any(flag)) {
      existing_cols <- existing_cols[!flag]
      existing_widths <- existing_widths[!flag]
      existing_hidden <- existing_hidden[!flag]
    }

    all_names <- c(existing_cols, cols)
    all_widths <- c(existing_widths, widths)
    all_hidden <- c(existing_hidden, as.character(as.integer(hidden)))

    ord <- order(as.integer(all_names))
    all_names <- all_names[ord]
    all_widths <- all_widths[ord]
    all_hidden <- all_hidden[ord]


    names(all_widths) <- all_names
    wb$colWidths[[sheet]] <- all_widths
    attr(wb$colWidths[[sheet]], "hidden") <- all_hidden
  } else {
    names(widths) <- cols
    wb$colWidths[[sheet]] <- widths
    attr(wb$colWidths[[sheet]], "hidden") <- as.character(as.integer(hidden))
  }

  # Check if any conflicting column outline levels
  if (length(wb$colOutlineLevels[[sheet]]) > 0) {
    existing_cols <- names(wb$colOutlineLevels[[sheet]])

    if (any(existing_cols %in% cols)) {
      for (i in intersect(existing_cols, cols)) {
        width_hidden <- attr(wb$colWidths[[sheet]], "hidden")[attr(wb$colWidths[[sheet]], "names") == i]
        outline_hidden <- attr(wb$colOutlineLevels[[sheet]], "hidden")[attr(wb$colOutlineLevels[[sheet]], "names") == i]

        if (outline_hidden != width_hidden) {
          attr(wb$colOutlineLevels[[sheet]], "hidden")[attr(wb$colOutlineLevels[[sheet]], "names") == i] <- width_hidden
        }
      }

      cols <- cols[!cols %in% existing_cols]
      hidden <- attr(wb$colWidths[[sheet]], "hidden")[attr(wb$colWidths[[sheet]], "names") %in% cols]
    }
  }

  invisible(0)
}


#' @name removeColWidths
#' @title Remove column widths from a worksheet

#' @description Remove column widths from a worksheet
#' @author Alexander Walker
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param cols Indices of columns to remove custom width (if any) from.
#' @seealso [setColWidths()]
#' @export
#' @examples
#' ## Create a new workbook
#' wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))
#'
#' ## remove column widths in columns 1 to 20
#' removeColWidths(wb, 1, cols = 1:20)
#' \dontrun{
#' saveWorkbook(wb, "removeColWidthsExample.xlsx", overwrite = TRUE)
#' }
removeColWidths <- function(wb, sheet, cols) {
  sheet <- wb$validateSheet(sheet)

  if (!is.numeric(cols)) {
    cols <- convertFromExcelRef(cols)
  }

  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  customCols <- as.integer(names(wb$colWidths[[sheet]]))
  removeInds <- which(customCols %in% cols)
  if (length(removeInds) > 0) {
    remainingCols <- customCols[-removeInds]
    if (length(remainingCols) == 0) {
      wb$colWidths[[sheet]] <- list()
    } else {
      rem_widths <- wb$colWidths[[sheet]][-removeInds]
      names(rem_widths) <- as.character(remainingCols)
      wb$colWidths[[sheet]] <- rem_widths
    }
  }
}



#' @name removeRowHeights
#' @title Remove custom row heights from a worksheet
#' @description Remove row heights from a worksheet
#' @author Alexander Walker
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param rows Indices of rows to remove custom height (if any) from.
#' @seealso [setRowHeights()]
#' @export
#' @examples
#' ## Create a new workbook
#' wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))
#'
#' ## remove any custom row heights in rows 1 to 10
#' removeRowHeights(wb, 1, rows = 1:10)
#' \dontrun{
#' saveWorkbook(wb, "removeRowHeightsExample.xlsx", overwrite = TRUE)
#' }
removeRowHeights <- function(wb, sheet, rows) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  sheet <- wb$validateSheet(sheet)

  customRows <- as.integer(names(wb$rowHeights[[sheet]]))
  removeInds <- which(customRows %in% rows)
  if (length(removeInds) > 0) {
    wb$rowHeights[[sheet]] <- wb$rowHeights[[sheet]][-removeInds]
  }
}


#' @name insertPlot
#' @title Insert the current plot into a worksheet
#' @author Alexander Walker
#' @description The current plot is saved to a temporary image file using dev.copy.
#' This file is then written to the workbook using insertImage.
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param startRow Row coordinate of upper left corner of figure.` xy[[2]]` when xy is given.
#' @param startCol Column coordinate of upper left corner of figure. `xy[[1]]` when xy is given.
#' @param xy Alternate way to specify startRow and startCol.  A vector of length 2 of form (startcol, startRow)
#' @param width Width of figure. Defaults to 6in.
#' @param height Height of figure . Defaults to 4in.
#' @param fileType File type of image
#' @param units Units of width and height. Can be "in", "cm" or "px"
#' @param dpi Image resolution
#' @seealso [insertImage()]
#' @export
#' @importFrom grDevices bmp png jpeg tiff dev.copy dev.list dev.off
#' @examples
#' \dontrun{
#' ## Create a new workbook
#' wb <- createWorkbook()
#'
#' ## Add a worksheet
#' addWorksheet(wb, "Sheet 1", gridLines = FALSE)
#'
#' ## create plot objects
#' require(ggplot2)
#' p1 <- qplot(mpg,
#'   data = mtcars, geom = "density",
#'   fill = as.factor(gear), alpha = I(.5), main = "Distribution of Gas Mileage"
#' )
#' p2 <- qplot(age, circumference,
#'   data = Orange, geom = c("point", "line"), colour = Tree
#' )
#'
#' ## Insert currently displayed plot to sheet 1, row 1, column 1
#' print(p1) # plot needs to be showing
#' insertPlot(wb, 1, width = 5, height = 3.5, fileType = "png", units = "in")
#'
#' ## Insert plot 2
#' print(p2)
#' insertPlot(wb, 1, xy = c("J", 2), width = 16, height = 10, fileType = "png", units = "cm")
#'
#' ## Save workbook
#' saveWorkbook(wb, "insertPlotExample.xlsx", overwrite = TRUE)
#' }
insertPlot <- function(wb, sheet, width = 6, height = 4, xy = NULL,
                       startRow = 1, startCol = 1, fileType = "png", units = "in", dpi = 300) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (is.null(dev.list()[[1]])) {
    warning("No plot to insert.")
    return()
  }

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  if (!is.null(xy)) {
    startCol <- xy[[1]]
    startRow <- xy[[2]]
  }

  fileType <- tolower(fileType)
  units <- tolower(units)

  if (fileType == "jpg") {
    fileType <- "jpeg"
  }

  if (!fileType %in% c("png", "jpeg", "tiff", "bmp")) {
    stop("Invalid file type.\nfileType must be one of: png, jpeg, tiff, bmp")
  }

  if (!units %in% c("cm", "in", "px")) {
    stop("Invalid units.\nunits must be one of: cm, in, px")
  }

  fileName <- tempfile(pattern = "figureImage", fileext = paste0(".", fileType))

  if (fileType == "bmp") {
    dev.copy(bmp, filename = fileName, width = width, height = height, units = units, res = dpi)
  } else if (fileType == "jpeg") {
    dev.copy(jpeg, filename = fileName, width = width, height = height, units = units, quality = 100, res = dpi)
  } else if (fileType == "png") {
    dev.copy(png, filename = fileName, width = width, height = height, units = units, res = dpi)
  } else if (fileType == "tiff") {
    dev.copy(tiff, filename = fileName, width = width, height = height, units = units, compression = "none", res = dpi)
  }

  ## write image
  invisible(dev.off())

  insertImage(wb = wb, sheet = sheet, file = fileName, width = width, height = height, startRow = startRow, startCol = startCol, units = units, dpi = dpi)
}



#' @name replaceStyle
#' @title Replace an existing cell style
#' @description Replace an existing cell style
#' @author Alexander Walker
#' @param wb A workbook object
#' @param index Index of style object to replace
#' @param newStyle A style to replace the existing style as position index
#' @description Replace a style object
#' @export
#' @seealso [getStyles()]
#' @examples
#'
#' ## load a workbook
#' wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))
#'
#' ## create a new style and replace style 2
#'
#' newStyle <- createStyle(fgFill = "#00FF00")
#'
#' ## replace style 2
#' getStyles(wb)[1:3] ## prints styles
#' replaceStyle(wb, 2, newStyle = newStyle)
#'
#' ## Save workbook
#' \dontrun{
#' saveWorkbook(wb, "replaceStyleExample.xlsx", overwrite = TRUE)
#' }
replaceStyle <- function(wb, index, newStyle) {
  nStyles <- length(wb$styleObjects)

  if (nStyles == 0) {
    stop("Workbook has no existing styles.")
  }

  if (index > nStyles) {
    stop(sprintf("Invalid index. Workbook only has %s styles.", nStyles))
  }

  if (!all("Style" %in% class(newStyle))) {
    stop("Invalid style object.")
  }

  wb$styleObjects[[index]]$style <- newStyle
}


#' @name getStyles
#' @title Returns a list of all styles in the workbook
#' @description Returns list of style objects in the workbook
#' @param wb A workbook object
#' @export
#' @seealso [replaceStyle()]
#' @examples
#' ## load a workbook
#' wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))
#' getStyles(wb)[1:3]
getStyles <- function(wb) {
  nStyles <- length(wb$styleObjects)

  if (nStyles == 0) {
    stop("Workbook has no existing styles.")
  }

  styles <- lapply(wb$styleObjects, "[[", "style")

  return(styles)
}



#' @name removeWorksheet
#' @title Remove a worksheet from a workbook
#' @description Remove a worksheet from a Workbook object
#' @author Alexander Walker
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @description Remove a worksheet from a workbook
#' @export
#' @examples
#' ## load a workbook
#' wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))
#'
#' ## Remove sheet 2
#' removeWorksheet(wb, 2)
#'
#' ## save the modified workbook
#' \dontrun{
#' saveWorkbook(wb, "removeWorksheetExample.xlsx", overwrite = TRUE)
#' }
removeWorksheet <- function(wb, sheet) {
  if (!inherits(wb, "Workbook")) {
    stop("wb must be a Workbook object!")
  }

  if (length(sheet) != 1) {
    stop("sheet must have length 1.")
  }

  wb$deleteWorksheet(sheet)

  invisible(0)
}


#' @name deleteData
#' @title Delete cell data
#' @description Delete contents and styling from a cell.
#' @author Alexander Walker
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param rows Rows to delete data from.
#' @param cols columns to delete data from.
#' @param gridExpand If `TRUE`, all data in rectangle min(rows):max(rows) X min(cols):max(cols)
#' will be removed.
#' @export
#' @examples
#' ## write some data
#' wb <- createWorkbook()
#' addWorksheet(wb, "Worksheet 1")
#' x <- data.frame(matrix(runif(200), ncol = 10))
#' writeData(wb, sheet = 1, x = x, startCol = 2, startRow = 3, colNames = FALSE)
#'
#' ## delete some data
#' deleteData(wb, sheet = 1, cols = 3:5, rows = 5:7, gridExpand = TRUE)
#' deleteData(wb, sheet = 1, cols = 7:9, rows = 5:7, gridExpand = TRUE)
#' deleteData(wb, sheet = 1, cols = LETTERS, rows = 18, gridExpand = TRUE)
#' \dontrun{
#' saveWorkbook(wb, "deleteDataExample.xlsx", overwrite = TRUE)
#' }
deleteData <- function(wb, sheet, cols, rows, gridExpand = FALSE) {
  sheet <- wb$validateSheet(sheet)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }


  wb$worksheets[[sheet]]$sheet_data$delete(rows_in = rows, cols_in = cols, grid_expand = gridExpand)


  invisible(0)
}


#' @name deleteDataColumn
#' @title Deletes a whole column from a workbook
#' @description Deletes the whole column from a workbook, shifting the remaining columns to the left
#' @author David Zimmermann
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param col A column to delete
#' @export
#' @examples
#' ## write some data
#' wb <- createWorkbook()
#' addWorksheet(wb, "tester")
#'
#' for (i in seq(5)) {
#'   mat <- data.frame(x = rep(paste0(int2col(i), i), 10))
#'   writeData(wb, sheet = 1, startRow = 1, startCol = i, mat)
#'   writeFormula(wb, sheet = 1, startRow = 12, startCol = i,
#'                x = sprintf("=COUNTA(%s2:%s11)", int2col(i), int2col(i)))
#' }
#' deleteDataColumn(wb, 1, col = 3)
#' \dontrun{
#' saveWorkbook(wb, "deleteDataColumnExample.xlsx", overwrite = TRUE)
#' }
deleteDataColumn <- function(wb, sheet, col) {
  sheet <- wb$validateSheet(sheet)
  if (is.character(col)) col <- col2int(col)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }
  ## internal helper function which corrects columns > col (reduces their col ref by 1)
  updateFormula <- function(x, col) {
    has_formula <- !is.na(x) & stringi::stri_detect(x, regex = "[A-Z]+\\d")
    if (!any(has_formula)) return(x)

    xx <- x[has_formula]

    forms <- stringi::stri_split(xx, regex = "\\b(?=[A-Z]+\\d+)")

    x[has_formula] <- sapply(forms, function(form) {
      cols <- openxlsx::col2int(stringi::stri_extract(form[-1], regex = "^[A-Z]+"))
      repl <- ifelse(cols == col, "#REF!",
                     ifelse(cols > col,
                            openxlsx::int2col(cols - 1),
                            openxlsx::int2col(cols)))

      paste(c(form[[1]],
              stringi::stri_replace(form[-1], regex = "^[A-Z]+", repl)),
            collapse = "")
    })

    x
  }

  a <- wb$worksheets[[sheet]]$sheet_data
  
  # t: if a shared string is used or if the string is a value itself
  # v: the shared string index or the string itself (if t == 0)
  # in wb$sharedStrings we find the values of the shared strings by index + 1 (0 indexed!)
  

  # check which elements to delete
  keep <- a$cols != col
  # if there is no column to delete, exit early
  if (all(keep)) return(invisible(0))

  # delete cols in cols "col", move higher cols one down
  a$cols <- as.integer(a$cols[keep] - 1 * (a$cols[keep] > col))
  a$rows <- a$rows[keep]

  # reduce the shared strings pointers if they are not used anymore
  ss <- data.frame(old = numeric(0), new = numeric(0), string = character(0))
  if (length(wb$sharedStrings) > 0) {
    ss <- data.frame(
      # the old index 0 indexed, as used in a$v
      old = as.numeric(seq(length(wb$sharedStrings)) - 1),
      # will hold the new index 0 indexed, as used in a$v
      new = NA,
      # the actual strings
      string = wb$sharedStrings
    )
  }
  
  # 1. remove the values from sheet_data (a)
  a$v <- a$v[keep]
  a$t <- a$t[keep]
  
  # update the shared strings map (ss) with the new indices
  
  # v_this_sheet etc are the indices that are still used
  v_this_sheet <- as.numeric(a$v[!is.na(a$t) & a$t == 1])
  # get all string indices from other sheets, so that strings used in other sheets are not deleted!
  v_other_sheets <- unlist(lapply(setdiff(seq_along(wb$worksheets), sheet), function(sh) {
    a <- wb$worksheets[[sh]]$sheet_data
    as.numeric(a$v[!is.na(a$t) & a$t == 1])
  }))
  
  idx <- sort(unique(c(v_this_sheet, v_other_sheets)))
  ss$new[ss$old %in% idx] <- seq_along(idx) - 1
  
  # 2. remove the values from the sharedStrings object
  wb$sharedStrings <- wb$sharedStrings[idx + 1]
  attr(wb$sharedStrings, "uniqueCount") <- length(idx)
  
  # 3. reindex the values from the sheet_data to use new shared strings indices
  a$v[a$t == 1] <- as.character(ss$new[as.numeric(a$v[a$t == 1]) + 1])
  
  # update the shared strings for all other sheets
  for (s in setdiff(seq_along(wb$worksheets), sheet)) {
    a <- wb$worksheets[[s]]$sheet_data
    a$v[a$t == 1] <- as.character(ss$new[as.numeric(a$v[a$t == 1]) + 1])
  }
  
  a$f <- updateFormula(a$f[keep], col = col)
  a$n_elements <- sum(keep)

  if ("data_count" %in% names(a)) a$data_count <- length(unique(a$v))

  # adjust styles
  sheet_name <- wb$sheet_names[[sheet]]
  this_sheet <- sapply(wb$styleObjects, function(o) {
    if (!"sheet" %in% names(o)) return(FALSE)
    o$sheet == sheet_name
  })
  if (!is.null(this_sheet) && any(this_sheet)) {
    wb$styleObjects[this_sheet] <- lapply(
      wb$styleObjects[this_sheet],
      function(style) {
        if (all(style$cols == col)) return(NULL) # only in this col
        if (!any(style$cols > col)) return(style)
        take <- style$cols != col
        style$cols <- style$cols[take]
        style$rows <- style$rows[take]
        style$cols[style$cols > col] <- style$cols[style$cols > col] - 1L
        style
      })
  }

  invisible(0)
}


#' @name modifyBaseFont
#' @title Modify the default font
#' @description Modify the default font for this workbook
#' @author Alexander Walker
#' @param wb A workbook object
#' @param fontSize font size
#' @param fontColour font colour
#' @param fontName Name of a font
#' @details The font name is not validated in anyway.  Excel replaces unknown font names
#' with Arial. Base font is black, size 11, Calibri.
#' @export
#' @examples
#' ## create a workbook
#' wb <- createWorkbook()
#' addWorksheet(wb, "S1")
#' ## modify base font to size 10 Arial Narrow in red
#' modifyBaseFont(wb, fontSize = 10, fontColour = "#FF0000", fontName = "Arial Narrow")
#'
#' writeData(wb, "S1", iris)
#' writeDataTable(wb, "S1", x = iris, startCol = 10) ## font colour does not affect tables
#' \dontrun{
#' saveWorkbook(wb, "modifyBaseFontExample.xlsx", overwrite = TRUE)
#' }
modifyBaseFont <- function(wb, fontSize = 11, fontColour = "black", fontName = "Calibri") {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (fontSize < 0) stop("Invalid fontSize")
  fontColour <- validateColour(fontColour)

  wb$styles$fonts[[1]] <- sprintf('<font><sz val="%s"/><color rgb="%s"/><name val="%s"/></font>', fontSize, fontColour, fontName)
}


#' @name getBaseFont
#' @title Return the workbook default font
#' @description Return the workbook default font
#' @author Alexander Walker
#' @param wb A workbook object
#' @description Returns the base font used in the workbook.
#' @export
#' @examples
#' ## create a workbook
#' wb <- createWorkbook()
#' getBaseFont(wb)
#'
#' ## modify base font to size 10 Arial Narrow in red
#' modifyBaseFont(wb, fontSize = 10, fontColour = "#FF0000", fontName = "Arial Narrow")
#'
#' getBaseFont(wb)
getBaseFont <- function(wb) {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  wb$getBaseFont()
}


#' @name setHeaderFooter
#' @title Set document headers and footers
#' @description Set document headers and footers
#' @author Alexander Walker
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param header document header. Character vector of length 3 corresponding to positions left, center, right. Use NA to skip a position.
#' @param footer document footer. Character vector of length 3 corresponding to positions left, center, right. Use NA to skip a position.
#' @param evenHeader document header for even pages.
#' @param evenFooter document footer for even pages.
#' @param firstHeader document header for first page only.
#' @param firstFooter document footer for first page only.
#' @details Headers and footers can contain special tags
#' \describe{
#'   \item{**&\[Page\]**}{ Page number}
#'   \item{**&\[Pages\]**}{ Number of pages}
#'   \item{**&\[Date\]**}{ Current date}
#'   \item{**&\[Time\]**}{ Current time}
#'   \item{**&\[Path\]**}{ File path}
#'   \item{**&\[File\]**}{ File name}
#'   \item{**&\[Tab\]**}{ Worksheet name}
#' }
#' @export
#' @seealso [addWorksheet()] to set headers and footers when adding a worksheet
#' @examples
#' wb <- createWorkbook()
#'
#' addWorksheet(wb, "S1")
#' addWorksheet(wb, "S2")
#' addWorksheet(wb, "S3")
#' addWorksheet(wb, "S4")
#'
#' writeData(wb, 1, 1:400)
#' writeData(wb, 2, 1:400)
#' writeData(wb, 3, 3:400)
#' writeData(wb, 4, 3:400)
#'
#' setHeaderFooter(wb,
#'   sheet = "S1",
#'   header = c("ODD HEAD LEFT", "ODD HEAD CENTER", "ODD HEAD RIGHT"),
#'   footer = c("ODD FOOT RIGHT", "ODD FOOT CENTER", "ODD FOOT RIGHT"),
#'   evenHeader = c("EVEN HEAD LEFT", "EVEN HEAD CENTER", "EVEN HEAD RIGHT"),
#'   evenFooter = c("EVEN FOOT RIGHT", "EVEN FOOT CENTER", "EVEN FOOT RIGHT"),
#'   firstHeader = c("TOP", "OF FIRST", "PAGE"),
#'   firstFooter = c("BOTTOM", "OF FIRST", "PAGE")
#' )
#'
#' setHeaderFooter(wb,
#'   sheet = 2,
#'   header = c("&[Date]", "ALL HEAD CENTER 2", "&[Page] / &[Pages]"),
#'   footer = c("&[Path]&[File]", NA, "&[Tab]"),
#'   firstHeader = c(NA, "Center Header of First Page", NA),
#'   firstFooter = c(NA, "Center Footer of First Page", NA)
#' )
#'
#' setHeaderFooter(wb,
#'   sheet = 3,
#'   header = c("ALL HEAD LEFT 2", "ALL HEAD CENTER 2", "ALL HEAD RIGHT 2"),
#'   footer = c("ALL FOOT RIGHT 2", "ALL FOOT CENTER 2", "ALL FOOT RIGHT 2")
#' )
#'
#' setHeaderFooter(wb,
#'   sheet = 4,
#'   firstHeader = c("FIRST ONLY L", NA, "FIRST ONLY R"),
#'   firstFooter = c("FIRST ONLY L", NA, "FIRST ONLY R")
#' )
#' \dontrun{
#' saveWorkbook(wb, "setHeaderFooterExample.xlsx", overwrite = TRUE)
#' }
setHeaderFooter <- function(wb, sheet,
                            header = NULL,
                            footer = NULL,
                            evenHeader = NULL,
                            evenFooter = NULL,
                            firstHeader = NULL,
                            firstFooter = NULL) {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  sheet <- wb$validateSheet(sheet)

  if (!is.null(header) && length(header) != 3) {
    stop("header must have length 3 where elements correspond to positions: left, center, right.")
  }

  if (!is.null(footer) && length(footer) != 3) {
    stop("footer must have length 3 where elements correspond to positions: left, center, right.")
  }

  if (!is.null(evenHeader) && length(evenHeader) != 3) {
    stop("evenHeader must have length 3 where elements correspond to positions: left, center, right.")
  }

  if (!is.null(evenFooter) && length(evenFooter) != 3) {
    stop("evenFooter must have length 3 where elements correspond to positions: left, center, right.")
  }

  if (!is.null(firstHeader) && length(firstHeader) != 3) {
    stop("firstHeader must have length 3 where elements correspond to positions: left, center, right.")
  }

  if (!is.null(firstFooter) && length(firstFooter) != 3) {
    stop("firstFooter must have length 3 where elements correspond to positions: left, center, right.")
  }

  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  oddHeader <- headerFooterSub(header)
  oddFooter <- headerFooterSub(footer)
  evenHeader <- headerFooterSub(evenHeader)
  evenFooter <- headerFooterSub(evenFooter)
  firstHeader <- headerFooterSub(firstHeader)
  firstFooter <- headerFooterSub(firstFooter)

  naToNULLList <- function(x) {
    lapply(x, function(x) {
      if (is.na(x)) {
        return(NULL)
      }
      x
    })
  }

  hf <- list(
    oddHeader = naToNULLList(oddHeader),
    oddFooter = naToNULLList(oddFooter),
    evenHeader = naToNULLList(evenHeader),
    evenFooter = naToNULLList(evenFooter),
    firstHeader = naToNULLList(firstHeader),
    firstFooter = naToNULLList(firstFooter)
  )

  if (all(sapply(hf, length) == 0)) {
    hf <- NULL
  }


  wb$worksheets[[sheet]]$headerFooter <- hf
}




#' @name pageSetup
#' @title Set page margins, orientation and print scaling
#' @description Set page margins, orientation and print scaling
#' @author Alexander Walker, Joshua Sturm
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param orientation Page orientation. One of "portrait" or "landscape"
#' @param scale Print scaling. Numeric value between 10 and 400
#' @param left left page margin in inches
#' @param right right page margin in inches
#' @param top top page margin in inches
#' @param bottom bottom page margin in inches
#' @param header header margin in inches
#' @param footer footer margin in inches
#' @param fitToWidth If `TRUE`, worksheet is scaled to fit to page width on printing.
#' @param fitToHeight If `TRUE`, worksheet is scaled to fit to page height on printing.
#' @param paperSize See details. Default value is 9 (A4 paper).
#' @param printTitleRows Rows to repeat at top of page when printing. Integer vector.
#' @param printTitleCols Columns to repeat at left when printing. Integer vector.
#' @param summaryRow Location of summary rows in groupings. One of "Above" or "Below".
#' @param summaryCol Location of summary columns in groupings. One of "Right" or "Left".
#' @export
#' @details
#' paperSize is an integer corresponding to:
#' \describe{
#' \item{**1**}{ Letter paper (8.5 in. by 11 in.)}
#' \item{**2**}{ Letter small paper (8.5 in. by 11 in.)}
#' \item{**3**}{ Tabloid paper (11 in. by 17 in.)}
#' \item{**4**}{ Ledger paper (17 in. by 11 in.)}
#' \item{**5**}{ Legal paper (8.5 in. by 14 in.)}
#' \item{**6**}{ Statement paper (5.5 in. by 8.5 in.)}
#' \item{**7**}{ Executive paper (7.25 in. by 10.5 in.)}
#' \item{**8**}{ A3 paper (297 mm by 420 mm)}
#' \item{**9**}{ A4 paper (210 mm by 297 mm)}
#' \item{**10**}{ A4 small paper (210 mm by 297 mm)}
#' \item{**11**}{ A5 paper (148 mm by 210 mm)}
#' \item{**12**}{ B4 paper (250 mm by 353 mm)}
#' \item{**13**}{ B5 paper (176 mm by 250 mm)}
#' \item{**14**}{ Folio paper (8.5 in. by 13 in.)}
#' \item{**15**}{ Quarto paper (215 mm by 275 mm)}
#' \item{**16**}{ Standard paper (10 in. by 14 in.)}
#' \item{**17**}{ Standard paper (11 in. by 17 in.)}
#' \item{**18**}{ Note paper (8.5 in. by 11 in.)}
#' \item{**19**}{ #9 envelope (3.875 in. by 8.875 in.)}
#' \item{**20**}{ #10 envelope (4.125 in. by 9.5 in.)}
#' \item{**21**}{ #11 envelope (4.5 in. by 10.375 in.)}
#' \item{**22**}{ #12 envelope (4.75 in. by 11 in.)}
#' \item{**23**}{ #14 envelope (5 in. by 11.5 in.)}
#' \item{**24**}{ C paper (17 in. by 22 in.)}
#' \item{**25**}{ D paper (22 in. by 34 in.)}
#' \item{**26**}{ E paper (34 in. by 44 in.)}
#' \item{**27**}{ DL envelope (110 mm by 220 mm)}
#' \item{**28**}{ C5 envelope (162 mm by 229 mm)}
#' \item{**29**}{ C3 envelope (324 mm by 458 mm)}
#' \item{**30**}{ C4 envelope (229 mm by 324 mm)}
#' \item{**31**}{ C6 envelope (114 mm by 162 mm)}
#' \item{**32**}{ C65 envelope (114 mm by 229 mm)}
#' \item{**33**}{ B4 envelope (250 mm by 353 mm)}
#' \item{**34**}{ B5 envelope (176 mm by 250 mm)}
#' \item{**35**}{ B6 envelope (176 mm by 125 mm)}
#' \item{**36**}{ Italy envelope (110 mm by 230 mm)}
#' \item{**37**}{ Monarch envelope (3.875 in. by 7.5 in.).}
#' \item{**38**}{ 6 3/4 envelope (3.625 in. by 6.5 in.)}
#' \item{**39**}{ US standard fanfold (14.875 in. by 11 in.)}
#' \item{**40**}{ German standard fanfold (8.5 in. by 12 in.)}
#' \item{**41**}{ German legal fanfold (8.5 in. by 13 in.)}
#' \item{**42**}{ ISO B4 (250 mm by 353 mm)}
#' \item{**43**}{ Japanese double postcard (200 mm by 148 mm)}
#' \item{**44**}{ Standard paper (9 in. by 11 in.)}
#' \item{**45**}{ Standard paper (10 in. by 11 in.)}
#' \item{**46**}{ Standard paper (15 in. by 11 in.)}
#' \item{**47**}{ Invite envelope (220 mm by 220 mm)}
#' \item{**50**}{ Letter extra paper (9.275 in. by 12 in.)}
#' \item{**51**}{ Legal extra paper (9.275 in. by 15 in.)}
#' \item{**52**}{ Tabloid extra paper (11.69 in. by 18 in.)}
#' \item{**53**}{ A4 extra paper (236 mm by 322 mm)}
#' \item{**54**}{ Letter transverse paper (8.275 in. by 11 in.)}
#' \item{**55**}{ A4 transverse paper (210 mm by 297 mm)}
#' \item{**56**}{ Letter extra transverse paper (9.275 in. by 12 in.)}
#' \item{**57**}{ SuperA/SuperA/A4 paper (227 mm by 356 mm)}
#' \item{**58**}{ SuperB/SuperB/A3 paper (305 mm by 487 mm)}
#' \item{**59**}{ Letter plus paper (8.5 in. by 12.69 in.)}
#' \item{**60**}{ A4 plus paper (210 mm by 330 mm)}
#' \item{**61**}{ A5 transverse paper (148 mm by 210 mm)}
#' \item{**62**}{ JIS B5 transverse paper (182 mm by 257 mm)}
#' \item{**63**}{ A3 extra paper (322 mm by 445 mm)}
#' \item{**64**}{ A5 extra paper (174 mm by 235 mm)}
#' \item{**65**}{ ISO B5 extra paper (201 mm by 276 mm)}
#' \item{**66**}{ A2 paper (420 mm by 594 mm)}
#' \item{**67**}{ A3 transverse paper (297 mm by 420 mm)}
#' \item{**68**}{ A3 extra transverse paper (322 mm by 445 mm)}
#' }
#' @examples
#' wb <- createWorkbook()
#' addWorksheet(wb, "S1")
#' addWorksheet(wb, "S2")
#' writeDataTable(wb, 1, x = iris[1:30, ])
#' writeDataTable(wb, 2, x = iris[1:30, ], xy = c("C", 5))
#'
#' ## landscape page scaled to 50%
#' pageSetup(wb, sheet = 1, orientation = "landscape", scale = 50)
#'
#' ## portrait page scales to 300% with 0.5in left and right margins
#' pageSetup(wb, sheet = 2, orientation = "portrait", scale = 300, left = 0.5, right = 0.5)
#'
#'
#' ## print titles
#' addWorksheet(wb, "print_title_rows")
#' addWorksheet(wb, "print_title_cols")
#'
#' writeData(wb, "print_title_rows", rbind(iris, iris, iris, iris))
#' writeData(wb, "print_title_cols", x = rbind(mtcars, mtcars, mtcars), rowNames = TRUE)
#'
#' pageSetup(wb, sheet = "print_title_rows", printTitleRows = 1) ## first row
#' pageSetup(wb, sheet = "print_title_cols", printTitleCols = 1, printTitleRows = 1)
#' \dontrun{
#' saveWorkbook(wb, "pageSetupExample.xlsx", overwrite = TRUE)
#' }
pageSetup <- function(wb, sheet, orientation = NULL, scale = 100,
                      left = 0.7, right = 0.7, top = 0.75, bottom = 0.75,
                      header = 0.3, footer = 0.3,
                      fitToWidth = FALSE, fitToHeight = FALSE, paperSize = NULL,
                      printTitleRows = NULL, printTitleCols = NULL,
                      summaryRow = NULL, summaryCol = NULL) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  sheet <- wb$validateSheet(sheet)
  xml <- wb$worksheets[[sheet]]$pageSetup

  if (!is.null(orientation)) {
    orientation <- tolower(orientation)
    if (!orientation %in% c("portrait", "landscape")) stop("Invalid page orientation.")
  } else {
    orientation <- ifelse(grepl("landscape", xml), "landscape", "portrait") ## get existing
  }

  if (scale < 10 || scale > 400) {
    stop("Scale must be between 10 and 400.")
  }

  if (!is.null(paperSize)) {
    paperSizes <- 1:68
    paperSizes <- paperSizes[!paperSizes %in% 48:49]
    if (!paperSize %in% paperSizes) {
      stop("paperSize must be an integer in range [1, 68]. See ?pageSetup details.")
    }
    paperSize <- as.integer(paperSize)
  } else {
    paperSize <- regmatches(xml, regexpr('(?<=paperSize=")[0-9]+', xml, perl = TRUE)) ## get existing
  }


  ##############################
  ## Keep defaults on orientation, hdpi, vdpi, paperSize
  hdpi <- regmatches(xml, regexpr('(?<=horizontalDpi=")[0-9]+', xml, perl = TRUE))
  vdpi <- regmatches(xml, regexpr('(?<=verticalDpi=")[0-9]+', xml, perl = TRUE))


  ##############################
  ## Update
  wb$worksheets[[sheet]]$pageSetup <- sprintf(
    '<pageSetup paperSize="%s" orientation="%s" scale = "%s" fitToWidth="%s" fitToHeight="%s" horizontalDpi="%s" verticalDpi="%s" r:id="rId2"/>',
    paperSize, orientation, scale, as.integer(fitToWidth), as.integer(fitToHeight), hdpi, vdpi
  )

  if (fitToHeight || fitToWidth) {
    wb$worksheets[[sheet]]$sheetPr <- unique(c(wb$worksheets[[sheet]]$sheetPr, '<pageSetUpPr fitToPage="1"/>'))
  }

  wb$worksheets[[sheet]]$pageMargins <-
    sprintf('<pageMargins left="%s" right="%s" top="%s" bottom="%s" header="%s" footer="%s"/>', left, right, top, bottom, header, footer)

  validRow <- function(summaryRow) {
    return(tolower(summaryRow) %in% c("above", "below"))
  }
  validCol <- function(summaryCol) {
    return(tolower(summaryCol) %in% c("left", "right"))
  }

  outlinepr <- ""

  if (!is.null(summaryRow)) {
    if (!validRow(summaryRow)) {
      stop("Invalid \`summaryRow\` option. Must be one of \"Above\" or \"Below\".")
    } else if (tolower(summaryRow) == "above") {
      outlinepr <- ' summaryBelow=\"0\"'
    } else {
      outlinepr <- ' summaryBelow=\"1\"'
    }
  }

  if (!is.null(summaryCol)) {
    if (!validCol(summaryCol)) {
      stop("Invalid \`summaryCol\` option. Must be one of \"Left\" or \"Right\".")
    } else if (tolower(summaryCol) == "left") {
      outlinepr <- paste0(outlinepr, ' summaryRight=\"0\"')
    } else {
      outlinepr <- paste0(outlinepr, ' summaryRight=\"1\"')
    }
  }

  if (!stri_isempty(outlinepr)) {
    wb$worksheets[[sheet]]$sheetPr <- unique(c(wb$worksheets[[sheet]]$sheetPr, paste0("<outlinePr", outlinepr, "/>")))
  }

  ## print Titles
  if (!is.null(printTitleRows) && is.null(printTitleCols)) {
    if (!is.numeric(printTitleRows)) {
      stop("printTitleRows must be numeric.")
    }

    wb$createNamedRegion(
      ref1 = paste0("$", min(printTitleRows)),
      ref2 = paste0("$", max(printTitleRows)),
      name = "_xlnm.Print_Titles",
      sheet = names(wb)[[sheet]],
      localSheetId = sheet - 1L
    )
  } else if (!is.null(printTitleCols) && is.null(printTitleRows)) {
    if (!is.numeric(printTitleCols)) {
      stop("printTitleCols must be numeric.")
    }

    cols <- convert_to_excel_ref(cols = range(printTitleCols), LETTERS = LETTERS)
    wb$createNamedRegion(
      ref1 = paste0("$", cols[1]),
      ref2 = paste0("$", cols[2]),
      name = "_xlnm.Print_Titles",
      sheet = names(wb)[[sheet]],
      localSheetId = sheet - 1L
    )
  } else if (!is.null(printTitleCols) && !is.null(printTitleRows)) {
    if (!is.numeric(printTitleRows)) {
      stop("printTitleRows must be numeric.")
    }

    if (!is.numeric(printTitleCols)) {
      stop("printTitleCols must be numeric.")
    }

    cols <- convert_to_excel_ref(cols = range(printTitleCols), LETTERS = LETTERS)
    rows <- range(printTitleRows)

    cols <- paste(paste0("$", cols[1]), paste0("$", cols[2]), sep = ":")
    rows <- paste(paste0("$", rows[1]), paste0("$", rows[2]), sep = ":")
    localSheetId <- sheet - 1L
    sheet <- names(wb)[[sheet]]

    wb$workbook$definedNames <- c(
      wb$workbook$definedNames,
      sprintf('<definedName name="_xlnm.Print_Titles" localSheetId="%s">\'%s\'!%s,\'%s\'!%s</definedName>', localSheetId, sheet, cols, sheet, rows)
    )
  }
}


#' @name protectWorksheet
#' @title Protect a worksheet from modifications
#' @description Protect or unprotect a worksheet from modifications by the user in the graphical user interface. Replaces an existing protection.
#' @author Reinhold Kainhofer
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param protect Whether to protect or unprotect the sheet (default=TRUE)
#' @param password (optional) password required to unprotect the worksheet
#' @param lockSelectingLockedCells Whether selecting locked cells is locked
#' @param lockSelectingUnlockedCells Whether selecting unlocked cells is locked
#' @param lockFormattingCells Whether formatting cells is locked
#' @param lockFormattingColumns Whether formatting columns is locked
#' @param lockFormattingRows Whether formatting rows is locked
#' @param lockInsertingColumns Whether inserting columns is locked
#' @param lockInsertingRows Whether inserting rows is locked
#' @param lockInsertingHyperlinks Whether inserting hyperlinks is locked
#' @param lockDeletingColumns Whether deleting columns is locked
#' @param lockDeletingRows Whether deleting rows is locked
#' @param lockSorting Whether sorting is locked
#' @param lockAutoFilter Whether auto-filter is locked
#' @param lockPivotTables Whether pivot tables are locked
#' @param lockObjects Whether objects are locked
#' @param lockScenarios Whether scenarios are locked
#' @export
#' @examples
#' wb <- createWorkbook()
#' addWorksheet(wb, "S1")
#' writeDataTable(wb, 1, x = iris[1:30, ])
#' # Formatting cells / columns is allowed , but inserting / deleting columns is protected:
#' protectWorksheet(wb, "S1",
#'   protect = TRUE,
#'   lockFormattingCells = FALSE, lockFormattingColumns = FALSE,
#'   lockInsertingColumns = TRUE, lockDeletingColumns = TRUE
#' )
#'
#' # Remove the protection
#' protectWorksheet(wb, "S1", protect = FALSE)
#' \dontrun{
#' saveWorkbook(wb, "pageSetupExample.xlsx", overwrite = TRUE)
#' }
protectWorksheet <- function(wb, sheet, protect = TRUE, password = NULL,
                             lockSelectingLockedCells = NULL, lockSelectingUnlockedCells = NULL,
                             lockFormattingCells = NULL, lockFormattingColumns = NULL, lockFormattingRows = NULL,
                             lockInsertingColumns = NULL, lockInsertingRows = NULL, lockInsertingHyperlinks = NULL,
                             lockDeletingColumns = NULL, lockDeletingRows = NULL,
                             lockSorting = NULL, lockAutoFilter = NULL, lockPivotTables = NULL,
                             lockObjects = NULL, lockScenarios = NULL) {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  sheet <- wb$validateSheet(sheet)
  # xml <- wb$worksheets[[sheet]]$sheetProtection variable not used

  props <- c()

  if (!missing(password) && !is.null(password)) {
    props["password"] <- hashPassword(password)
  }

  if (!missing(lockSelectingLockedCells) && !is.null(lockSelectingLockedCells)) {
    props["selectLockedCells"] <- toString(as.numeric(lockSelectingLockedCells))
  }
  if (!missing(lockSelectingUnlockedCells) && !is.null(lockSelectingUnlockedCells)) {
    props["selectUnlockedCells"] <- toString(as.numeric(lockSelectingUnlockedCells))
  }
  if (!missing(lockFormattingCells) && !is.null(lockFormattingCells)) {
    props["formatCells"] <- toString(as.numeric(lockFormattingCells))
  }
  if (!missing(lockFormattingColumns) && !is.null(lockFormattingColumns)) {
    props["formatColumns"] <- toString(as.numeric(lockFormattingColumns))
  }
  if (!missing(lockFormattingRows) && !is.null(lockFormattingRows)) {
    props["formatRows"] <- toString(as.numeric(lockFormattingRows))
  }
  if (!missing(lockInsertingColumns) && !is.null(lockInsertingColumns)) {
    props["insertColumns"] <- toString(as.numeric(lockInsertingColumns))
  }
  if (!missing(lockInsertingRows) && !is.null(lockInsertingRows)) {
    props["insertRows"] <- toString(as.numeric(lockInsertingRows))
  }
  if (!missing(lockInsertingHyperlinks) && !is.null(lockInsertingHyperlinks)) {
    props["insertHyperlinks"] <- toString(as.numeric(lockInsertingHyperlinks))
  }
  if (!missing(lockDeletingColumns) && !is.null(lockDeletingColumns)) {
    props["deleteColumns"] <- toString(as.numeric(lockDeletingColumns))
  }
  if (!missing(lockDeletingRows) && !is.null(lockDeletingRows)) {
    props["deleteRows"] <- toString(as.numeric(lockDeletingRows))
  }
  if (!missing(lockSorting) && !is.null(lockSorting)) {
    props["sort"] <- toString(as.numeric(lockSorting))
  }
  if (!missing(lockAutoFilter) && !is.null(lockAutoFilter)) {
    props["autoFilter"] <- toString(as.numeric(lockAutoFilter))
  }
  if (!missing(lockPivotTables) && !is.null(lockPivotTables)) {
    props["pivotTables"] <- toString(as.numeric(lockPivotTables))
  }
  if (!missing(lockObjects) && !is.null(lockObjects)) {
    props["objects"] <- toString(as.numeric(lockObjects))
  }
  if (!missing(lockScenarios) && !is.null(lockScenarios)) {
    props["scenarios"] <- toString(as.numeric(lockScenarios))
  }

  if (protect) {
    props["sheet"] <- "1"
    wb$worksheets[[sheet]]$sheetProtection <- sprintf("<sheetProtection %s/>", paste(names(props), '="', props, '"', collapse = " ", sep = ""))
  } else {
    wb$worksheets[[sheet]]$sheetProtection <- ""
  }
}



#' @name protectWorkbook
#' @title Protect a workbook from modifications
#' @description Protect or unprotect a workbook from modifications by the user in the graphical user interface. Replaces an existing protection.
#' @author Reinhold Kainhofer
#' @param wb A workbook object
#' @param protect Whether to protect or unprotect the sheet (default=TRUE)
#' @param password (optional) password required to unprotect the workbook
#' @param lockStructure Whether the workbook structure should be locked
#' @param lockWindows Whether the window position of the spreadsheet should be locked
#' @param type Lock type, default 1. From the xml documentation: 1 - Document is password protected. 2 - Document is recommended to be opened as read-only. 4 - Document is enforced to be opened as read-only. 8 - Document is locked for annotation.
#' @export
#' @examples
#' wb <- createWorkbook()
#' addWorksheet(wb, "S1")
#' protectWorkbook(wb, protect = TRUE, password = "Password", lockStructure = TRUE)
#' \dontrun{
#' saveWorkbook(wb, "WorkBook_Protection.xlsx", overwrite = TRUE)
#' }
#' # Remove the protection
#' protectWorkbook(wb, protect = FALSE)
#' \dontrun{
#' saveWorkbook(wb, "WorkBook_Protection_unprotected.xlsx", overwrite = TRUE)
#' }
protectWorkbook <- function(wb, protect = TRUE, password = NULL, lockStructure = FALSE, lockWindows = FALSE, type = 1L) {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  invisible(wb$protectWorkbook(protect = protect, password = password, lockStructure = lockStructure, lockWindows = lockWindows, type = type))
}





#' @name showGridLines
#' @title Set worksheet gridlines to show or hide.
#' @description Set worksheet gridlines to show or hide.
#' @author Alexander Walker
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param showGridLines A logical. If `FALSE`, grid lines are hidden.
#' @export
#' @examples
#' wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))
#' names(wb) ## list worksheets in workbook
#' showGridLines(wb, 1, showGridLines = FALSE)
#' showGridLines(wb, "testing", showGridLines = FALSE)
#' \dontrun{
#' saveWorkbook(wb, "showGridLinesExample.xlsx", overwrite = TRUE)
#' }
showGridLines <- function(wb, sheet, showGridLines = FALSE) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  sheet <- wb$validateSheet(sheet)

  if (!is.logical(showGridLines)) stop("showGridLines must be a logical")


  sv <- wb$worksheets[[sheet]]$sheetViews
  showGridLines <- as.integer(showGridLines)
  ## If attribute exists gsub
  if (grepl("showGridLines", sv)) {
    sv <- gsub('showGridLines=".?[^"]', sprintf('showGridLines="%s', showGridLines), sv, perl = TRUE)
  } else {
    sv <- gsub("<sheetView ", sprintf('<sheetView showGridLines="%s" ', showGridLines), sv)
  }

  wb$worksheets[[sheet]]$sheetViews <- sv
}





#' @name worksheetOrder
#' @title Order of worksheets in xlsx file
#' @description Get/set order of worksheets in a Workbook object
#' @details This function does not reorder the worksheets within the workbook object, it simply
#' shuffles the order when writing to file.
#' @export
#' @examples
#' ## setup a workbook with 3 worksheets
#' wb <- createWorkbook()
#' addWorksheet(wb = wb, sheetName = "Sheet 1", gridLines = FALSE)
#' writeDataTable(wb = wb, sheet = 1, x = iris)
#'
#' addWorksheet(wb = wb, sheetName = "mtcars (Sheet 2)", gridLines = FALSE)
#' writeData(wb = wb, sheet = 2, x = mtcars)
#'
#' addWorksheet(wb = wb, sheetName = "Sheet 3", gridLines = FALSE)
#' writeData(wb = wb, sheet = 3, x = Formaldehyde)
#'
#' worksheetOrder(wb)
#' names(wb)
#' worksheetOrder(wb) <- c(1, 3, 2) # switch position of sheets 2 & 3
#' writeData(wb, 2, 'This is still the "mtcars" worksheet', startCol = 15)
#' worksheetOrder(wb)
#' names(wb) ## ordering within workbook is not changed
#' \dontrun{
#' saveWorkbook(wb, "worksheetOrderExample.xlsx", overwrite = TRUE)
#' }
#' worksheetOrder(wb) <- c(3, 2, 1)
#' \dontrun{
#' saveWorkbook(wb, "worksheetOrderExample2.xlsx", overwrite = TRUE)
#' }
worksheetOrder <- function(wb) {
  if (!"Workbook" %in% class(wb)) {
    stop("Argument must be a Workbook.")
  }

  wb$sheetOrder
}

#' @rdname worksheetOrder
#' @param wb A workbook object
#' @param value Vector specifying order to write worksheets to file
#' @export
`worksheetOrder<-` <- function(wb, value) {
  if (!"Workbook" %in% class(wb)) {
    stop("Argument must be a Workbook.")
  }

  if (any(value != as.integer(value))) {
    stop("values must be integers")
  }

  value <- as.integer(value)

  value <- unique(value)
  if (length(value) != length(wb$worksheets)) {
    stop(sprintf("Worksheet order must be same length as number of worksheets [%s]", length(wb$worksheets)))
  }

  if (any(value > length(wb$worksheets))) {
    stop("Elements of order are greater than the number of worksheets")
  }


  old_ActiveSheet <- wb$ActiveSheet

  wb$sheetOrder <- value
  wb$setactiveSheet(old_ActiveSheet)


  invisible(wb)
}

#' @name as_POSIXct_utc
#' @title Convert to POSIXct with timezone UTC
#' @param x something as.POSIXct can convert
#' @keywords internal
as_POSIXct_utc <- function(x) {
  z <- as.POSIXct(x, tz = "UTC")
  attr(z, "tzone") <- "UTC"
  z
}


#' @name convertToDate
#' @title Convert from excel date number to R Date type
#' @description Convert from excel date number to R Date type
#' @param x A vector of integers
#' @param origin date. Default value is for Windows Excel 2010
#' @param ... additional parameters passed to as.Date()
#' @details Excel stores dates as number of days from some origin day
#' @seealso [writeData()]
#' @export
#' @examples
#' ## 2014 April 21st to 25th
#' convertToDate(c(41750, 41751, 41752, 41753, 41754, NA))
#' convertToDate(c(41750.2, 41751.99, NA, 41753))
convertToDate <- function(x, origin = "1900-01-01", ...) {
  x <- as.numeric(x)
  notNa <- !is.na(x)
  earlyDate <- x < 60
  if (origin == "1900-01-01") {
    x[notNa] <- x[notNa] - 2
    x[earlyDate & notNa] <- x[earlyDate & notNa] + 1
  }

  return(as.Date(x, origin = origin, ...))
}


#' @name convertToDateTime
#' @title Convert from excel time number to R POSIXct type.
#' @description Convert from excel time number to R POSIXct type.
#' @param x A numeric vector
#' @param origin date. Default value is for Windows Excel 2010
#' @param ... Additional parameters passed to as.POSIXct
#' @details Excel stores dates as number of days from some origin date
#' @export
#' @examples
#' ## 2014-07-01, 2014-06-30, 2014-06-29
#' x <- c(41821.8127314815, 41820.8127314815, NA, 41819, NaN)
#' convertToDateTime(x)
#' convertToDateTime(x, tz = "Australia/Perth")
#' convertToDateTime(x, tz = "UTC")
convertToDateTime <- function(x, origin = "1900-01-01", ...) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  x <- as.numeric(x)
  date <- convertToDate(x, origin)

  x <- x * 86400
  rem <- x %% 86400

  hours <- as.integer(floor(rem / 3600))
  minutes_fraction <- rem %% 3600
  minutes_whole <- as.integer(floor(minutes_fraction / 60))
  secs <- minutes_fraction %% 60

  y <- sprintf("%02d:%02d:%06.3f", hours, minutes_whole, secs)
  notNA <- !is.na(x)
  date_time <- rep(NA, length(x))
  date_time[notNA] <- as.POSIXct(paste(date[notNA], y[notNA]), ...)

  date_time <- .POSIXct(date_time)

  return(date_time)
}



#' @name names
#' @title get or set worksheet names
#' @description get or set worksheet names
#' @aliases names.Workbook
#' @export
#' @method names Workbook
#' @param x A `Workbook` object
#' @examples
#'
#' wb <- createWorkbook()
#' addWorksheet(wb, "S1")
#' addWorksheet(wb, "S2")
#' addWorksheet(wb, "S3")
#'
#' names(wb)
#' names(wb)[[2]] <- "S2a"
#' names(wb)
#' names(wb) <- paste("Sheet", 1:3)
names.Workbook <- function(x) {
  nms <- x$sheet_names
  nms <- replaceXMLEntities(nms)
}

#' @rdname names
#' @param value a character vector the same length as wb
#' @export
`names<-.Workbook` <- function(x, value) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (any(duplicated(tolower(value)))) {
    stop("Worksheet names must be unique.")
  }

  existing_sheets <- x$sheet_names
  inds <- which(value != existing_sheets)

  if (length(inds) == 0) {
    return(invisible(x))
  }

  if (length(value) != length(x$worksheets)) {
    stop(sprintf("names vector must have length equal to number of worksheets in Workbook [%s]", length(existing_sheets)))
  }

  if (any(nchar(value) > 31)) {
    warning("Worksheet names must less than 32 characters. Truncating names...")
    value[nchar(value) > 31] <- sapply(value[nchar(value) > 31], substr, start = 1, stop = 31)
  }

  for (i in inds) {
    invisible(x$setSheetName(i, value[[i]]))
  }

  invisible(x)
}



#' @name createNamedRegion
#' @title Create / delete a named region.
#' @description Create / delete a named region
#' @author Alexander Walker
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param rows Numeric vector specifying rows to include in region
#' @param cols Numeric vector specifying columns to include in region
#' @param name Name for region. A character vector of length 1. Note region names must be case-insensitive unique.
#' @param overwrite Boolean. Overwrite if exists ? Default to FALSE
#'
#' @details Region is given by: min(cols):max(cols) X min(rows):max(rows)
#' @export
#' @seealso [getNamedRegions()]
#' @examples
#' ## create named regions
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet 1")
#'
#' ## specify region
#' writeData(wb, sheet = 1, x = iris, startCol = 1, startRow = 1)
#' createNamedRegion(
#'   wb = wb,
#'   sheet = 1,
#'   name = "iris",
#'   rows = 1:(nrow(iris) + 1),
#'   cols = 1:ncol(iris)
#' )
#'
#'
#' ## using writeData 'name' argument
#' writeData(wb, sheet = 1, x = iris, name = "iris2", startCol = 10)
#'
#' out_file <- tempfile(fileext = ".xlsx")
#' \dontrun{
#' saveWorkbook(wb, out_file, overwrite = TRUE)
#'
#' ## see named regions
#' getNamedRegions(wb) ## From Workbook object
#' getNamedRegions(out_file) ## From xlsx file
#'
#' ## delete one
#' deleteNamedRegion(wb = wb, name = "iris2")
#' getNamedRegions(wb)
#'
#' ## read named regions
#' df <- read.xlsx(wb, namedRegion = "iris")
#' head(df)
#'
#' df <- read.xlsx(out_file, namedRegion = "iris2")
#' head(df)
#' }
#'
#' @rdname NamedRegion
createNamedRegion <- function(wb, sheet, cols, rows, name, overwrite = FALSE) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  sheet <- wb$validateSheet(sheet)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  if (!is.numeric(rows)) {
    stop("rows argument must be a numeric/integer vector")
  }

  if (!is.numeric(cols)) {
    stop("cols argument must be a numeric/integer vector")
  }

  ## check name doesn't already exist
  ## named region

  ex_names <- regmatches(wb$workbook$definedNames, regexpr('(?<=name=")[^"]+', wb$workbook$definedNames, perl = TRUE))
  ex_names <- tolower(replaceXMLEntities(ex_names))

  if (tolower(name) %in% ex_names && !overwrite) {
    stop(sprintf("Named region with name '%s' already exists! Use overwrite  = TRUE if you want to replace it", name))
  } else if (tolower(name) %in% ex_names && overwrite) {
    wb$workbook$definedNames <- wb$workbook$definedNames[!ex_names %in% tolower(name)]
  }

  if (grepl("^[A-Z]{1,3}[0-9]+$", name)) {
    stop("name cannot look like a cell reference.")
  }

  cols <- round(cols)
  rows <- round(rows)

  startCol <- min(cols)
  endCol <- max(cols)

  startRow <- min(rows)
  endRow <- max(rows)

  ref1 <- paste0("$", convert_to_excel_ref(cols = startCol, LETTERS = LETTERS), "$", startRow)
  ref2 <- paste0("$", convert_to_excel_ref(cols = endCol, LETTERS = LETTERS), "$", endRow)

  invisible(
    wb$createNamedRegion(ref1 = ref1, ref2 = ref2, name = name, sheet = wb$sheet_names[sheet])
  )
}


#' @export
#' @rdname NamedRegion
deleteNamedRegion <- function(wb, name) {

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  ex_names <- regmatches(wb$workbook$definedNames, regexpr('(?<=name=")[^"]+', wb$workbook$definedNames, perl = TRUE))
  ex_names <- tolower(replaceXMLEntities(ex_names))

  if (tolower(name) %in% ex_names) {
    wb$workbook$definedNames <- wb$workbook$definedNames[!ex_names %in% tolower(name)]
  } else {
    warning(sprintf("Cannot find Named region with name '%s'", name))
  }

  invisible(0)
}




#' @name getNamedRegions
#' @title Get named regions
#' @description Return a vector of named regions in a xlsx file or
#' Workbook object
#' @param x An xlsx file or Workbook object
#' @export
#' @seealso [createNamedRegion()]
#' @examples
#' ## create named regions
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet 1")
#'
#' ## specify region
#' writeData(wb, sheet = 1, x = iris, startCol = 1, startRow = 1)
#' createNamedRegion(
#'   wb = wb,
#'   sheet = 1,
#'   name = "iris",
#'   rows = 1:(nrow(iris) + 1),
#'   cols = 1:ncol(iris)
#' )
#'
#'
#' ## using writeData 'name' argument to create a named region
#' writeData(wb, sheet = 1, x = iris, name = "iris2", startCol = 10)
#' \dontrun{
#' out_file <- tempfile(fileext = ".xlsx")
#' saveWorkbook(wb, out_file, overwrite = TRUE)
#'
#' ## see named regions
#' getNamedRegions(wb) ## From Workbook object
#' getNamedRegions(out_file) ## From xlsx file
#'
#' ## read named regions
#' df <- read.xlsx(wb, namedRegion = "iris")
#' head(df)
#'
#' df <- read.xlsx(out_file, namedRegion = "iris2")
#' head(df)
#' }
#'
getNamedRegions <- function(x) {
  UseMethod("getNamedRegions", x)
}

#' @export
getNamedRegions.default <- function(x) {
  if (!file.exists(x)) {
    stop(sprintf("File '%s' does not exist.", x))
  }

  xmlDir <- tempfile()
  xmlFiles <- unzip(x, exdir = xmlDir)

  workbook <- grep("workbook.xml$", xmlFiles, perl = TRUE, value = TRUE)
  workbook <- unlist(readUTF8(workbook))

  dn <- getChildlessNode(xml = removeHeadTag(workbook), tag = "definedName")
  if (length(dn) == 0) {
    return(NULL)
  }

  dn_names <- get_named_regions_from_string(dn = dn)

  unlink(xmlDir, recursive = TRUE, force = TRUE)

  return(dn_names)
}


#' @export
getNamedRegions.Workbook <- function(x) {
  dn <- x$workbook$definedNames
  if (length(dn) == 0) {
    return(NULL)
  }

  dn_names <- get_named_regions_from_string(dn = dn)

  return(dn_names)
}






#' @name addFilter
#' @title Add column filters
#' @description Add excel column filters to a worksheet
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param cols columns to add filter to.
#' @param rows A row number.
#' @seealso [writeData()]
#' @details adds filters to worksheet columns, same as filter parameters in writeData.
#' writeDataTable automatically adds filters to first row of a table.
#' NOTE Can only have a single filter per worksheet unless using tables.
#' @export
#' @seealso [addFilter()]
#' @examples
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet 1")
#' addWorksheet(wb, "Sheet 2")
#' addWorksheet(wb, "Sheet 3")
#'
#' writeData(wb, 1, iris)
#' addFilter(wb, 1, row = 1, cols = 1:ncol(iris))
#'
#' ## Equivalently
#' writeData(wb, 2, x = iris, withFilter = TRUE)
#'
#' ## Similarly
#' writeDataTable(wb, 3, iris)
#' \dontrun{
#' saveWorkbook(wb, file = "addFilterExample.xlsx", overwrite = TRUE)
#' }
addFilter <- function(wb, sheet, rows, cols) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  sheet <- wb$validateSheet(sheet)

  if (length(rows) != 1) {
    stop("row must be a numeric of length 1.")
  }

  if (!is.numeric(cols)) {
    cols <- convertFromExcelRef(cols)
  }

  wb$worksheets[[sheet]]$autoFilter <- sprintf('<autoFilter ref="%s"/>', paste(getCellRefs(data.frame("x" = c(rows, rows), "y" = c(min(cols), max(cols)))), collapse = ":"))

  invisible(wb)
}


#' @name removeFilter
#' @title Remove a worksheet filter
#' @description Removes filters from addFilter() and writeData()
#' @param wb A workbook object
#' @param sheet A vector of names or indices of worksheets
#' @export
#' @examples
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet 1")
#' addWorksheet(wb, "Sheet 2")
#' addWorksheet(wb, "Sheet 3")
#'
#' writeData(wb, 1, iris)
#' addFilter(wb, 1, row = 1, cols = 1:ncol(iris))
#'
#' ## Equivalently
#' writeData(wb, 2, x = iris, withFilter = TRUE)
#'
#' ## Similarly
#' writeDataTable(wb, 3, iris)
#'
#' ## remove filters
#' removeFilter(wb, 1:2) ## remove filters
#' removeFilter(wb, 3) ## Does not affect tables!
#' \dontrun{
#' saveWorkbook(wb, file = "removeFilterExample.xlsx", overwrite = TRUE)
#' }
removeFilter <- function(wb, sheet) {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  for (s in sheet) {
    s <- wb$validateSheet(s)
    wb$worksheets[[s]]$autoFilter <- character(0)
  }

  invisible(wb)
}











#' @name setHeader
#' @title Set header for all worksheets
#' @description DEPRECATED
#' @author Alexander Walker
#' @param wb A workbook object
#' @param text header text. A character vector of length 1.
#' @param position Position of text in header. One of "left", "center" or "right"
#' @export
#' @examples
#' \dontrun{
#' wb <- createWorkbook("Edgar Anderson")
#' addWorksheet(wb, "S1")
#' writeDataTable(wb, "S1", x = iris[1:30, ], xy = c("C", 5))
#'
#' ## set all headers
#' setHeader(wb, "This is a header", position = "center")
#' setHeader(wb, "To the left", position = "left")
#' setHeader(wb, "On the right", position = "right")
#'
#' ## set all footers
#' setFooter(wb, "Center Footer Here", position = "center")
#' setFooter(wb, "Bottom left", position = "left")
#' setFooter(wb, Sys.Date(), position = "right")
#'
#' saveWorkbook(wb, "headerHeaderExample.xlsx", overwrite = TRUE)
#' }
setHeader <- function(wb, text, position = "center") {
  warning("This function is deprecated. Use function 'setHeaderFooter()'")

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  position <- tolower(position)
  if (!position %in% c("left", "center", "right")) {
    stop("Invalid position.")
  }

  if (length(text) != 1) {
    stop("Text argument must be a character vector of length 1")
  }

  # sheet <- wb$validateSheet(1) variable not used
  wb$headFoot$text[wb$headFoot$pos == position & wb$headFoot$head == "head"] <-
    as.character(text)
}


#' @name setFooter
#' @title Set footer for all worksheets
#' @description DEPRECATED
#' @author Alexander Walker
#' @param wb A workbook object
#' @param text footer text. A character vector of length 1.
#' @param position Position of text in footer. One of "left", "center" or "right"
#' @export
#' @examples
#' \dontrun{
#' wb <- createWorkbook("Edgar Anderson")
#' addWorksheet(wb, "S1")
#' writeDataTable(wb, "S1", x = iris[1:30, ], xy = c("C", 5))
#'
#' ## set all headers
#' setHeader(wb, "This is a header", position = "center")
#' setHeader(wb, "To the left", position = "left")
#' setHeader(wb, "On the right", position = "right")
#'
#' ## set all footers
#' setFooter(wb, "Center Footer Here", position = "center")
#' setFooter(wb, "Bottom left", position = "left")
#' setFooter(wb, Sys.Date(), position = "right")
#'
#' saveWorkbook(wb, "headerFooterExample.xlsx", overwrite = TRUE)
#' }
setFooter <- function(wb, text, position = "center") {
  warning("This function is deprecated. Use function 'setHeaderFooter()'")

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  position <- tolower(position)
  if (!position %in% c("left", "center", "right")) {
    stop("Invalid position.")
  }

  if (length(text) != 1) {
    stop("Text argument must be a character vector of length 1")
  }

  # sheet <- wb$validateSheet(1) variable not used
  wb$headFoot$text[wb$headFoot$pos == position & wb$headFoot$head == "foot"] <- as.character(text)
}











#' @name dataValidation
#' @title Add data validation to cells
#' @description Add Excel data validation to cells
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param cols Contiguous columns to apply conditional formatting to
#' @param rows Contiguous rows to apply conditional formatting to
#' @param type One of 'whole', 'decimal', 'date', 'time', 'textLength', 'list' (see examples)
#' @param operator One of 'between', 'notBetween', 'equal',
#'  'notEqual', 'greaterThan', 'lessThan', 'greaterThanOrEqual', 'lessThanOrEqual'
#' @param value a vector of length 1 or 2 depending on operator (see examples)
#' @param allowBlank logical
#' @param showInputMsg logical
#' @param showErrorMsg logical
#' @export
#' @examples
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet 1")
#' addWorksheet(wb, "Sheet 2")
#'
#' writeDataTable(wb, 1, x = iris[1:30, ])
#'
#' dataValidation(wb, 1,
#'   col = 1:3, rows = 2:31, type = "whole",
#'   operator = "between", value = c(1, 9)
#' )
#'
#' dataValidation(wb, 1,
#'   col = 5, rows = 2:31, type = "textLength",
#'   operator = "between", value = c(4, 6)
#' )
#'
#'
#' ## Date and Time cell validation
#' df <- data.frame(
#'   "d" = as.Date("2016-01-01") + -5:5,
#'   "t" = as.POSIXct("2016-01-01") + -5:5 * 10000
#' )
#'
#' writeData(wb, 2, x = df)
#' dataValidation(wb, 2,
#'   col = 1, rows = 2:12, type = "date",
#'   operator = "greaterThanOrEqual", value = as.Date("2016-01-01")
#' )
#'
#' dataValidation(wb, 2,
#'   col = 2, rows = 2:12, type = "time",
#'   operator = "between", value = df$t[c(4, 8)]
#' )
#' \dontrun{
#' saveWorkbook(wb, "dataValidationExample.xlsx", overwrite = TRUE)
#' }
#'
#'
#' ######################################################################
#' ## If type == 'list'
#' # operator argument is ignored.
#'
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet 1")
#' addWorksheet(wb, "Sheet 2")
#'
#' writeDataTable(wb, sheet = 1, x = iris[1:30, ])
#' writeData(wb, sheet = 2, x = sample(iris$Sepal.Length, 10))
#'
#' dataValidation(wb, 1, col = 1, rows = 2:31, type = "list", value = "'Sheet 2'!$A$1:$A$10")
#'
#' # openXL(wb)
dataValidation <- function(wb, sheet, cols, rows, type, operator, value, allowBlank = TRUE, showInputMsg = TRUE, showErrorMsg = TRUE) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  ## rows and cols
  if (!is.numeric(cols)) {
    cols <- convertFromExcelRef(cols)
  }
  rows <- as.integer(rows)

  ## check length of value
  if (length(value) > 2) {
    stop("value argument must be length < 2")
  }

  valid_types <- c(
    "whole",
    "decimal",
    "date",
    "time", ## need to conv
    "textLength",
    "list"
  )

  if (!tolower(type) %in% tolower(valid_types)) {
    stop("Invalid 'type' argument!")
  }


  ## operator == 'between' we leave out
  valid_operators <- c(
    "between",
    "notBetween",
    "equal",
    "notEqual",
    "greaterThan",
    "lessThan",
    "greaterThanOrEqual",
    "lessThanOrEqual"
  )

  if (tolower(type) != "list") {
    if (!tolower(operator) %in% tolower(valid_operators)) {
      stop("Invalid 'operator' argument!")
    }

    operator <- valid_operators[tolower(valid_operators) %in% tolower(operator)][1]
  } else {
    operator <- "between" ## ignored
  }

  if (!is.logical(allowBlank)) {
    stop("Argument 'allowBlank' must be logical!")
  }

  if (!is.logical(showInputMsg)) {
    stop("Argument 'showInputMsg' must be logical!")
  }

  if (!is.logical(showErrorMsg)) {
    stop("Argument 'showErrorMsg' must be logical!")
  }

  ## All inputs validated

  type <- valid_types[tolower(valid_types) %in% tolower(type)][1]

  ## check input combinations
  if (type == "date" && !"Date" %in% class(value)) {
    stop("If type == 'date' value argument must be a Date vector.")
  }

  if (type == "time" && !any(tolower(class(value)) %in% c("posixct", "posixt"))) {
    stop("If type == 'date' value argument must be a POSIXct or POSIXlt vector.")
  }


  value <- head(value, 2)
  allowBlank <- as.integer(allowBlank[1])
  showInputMsg <- as.integer(showInputMsg[1])
  showErrorMsg <- as.integer(showErrorMsg[1])

  if (type == "list") {
    invisible(wb$dataValidation_list(
      sheet = sheet,
      startRow = min(rows),
      endRow = max(rows),
      startCol = min(cols),
      endCol = max(cols),
      value = value,
      allowBlank = allowBlank,
      showInputMsg = showInputMsg,
      showErrorMsg = showErrorMsg
    ))
  } else {
    invisible(wb$dataValidation(
      sheet = sheet,
      startRow = min(rows),
      endRow = max(rows),
      startCol = min(cols),
      endCol = max(cols),
      type = type,
      operator = operator,
      value = value,
      allowBlank = allowBlank,
      showInputMsg = showInputMsg,
      showErrorMsg = showErrorMsg
    ))
  }



  invisible(0)
}








#' @name getDateOrigin
#' @title Get the date origin an xlsx file is using
#' @description Return the date origin used internally by an xlsx or xlsm file
#' @author Alexander Walker
#' @param xlsxFile An xlsx or xlsm file.
#' @details Excel stores dates as the number of days from either 1904-01-01 or 1900-01-01. This function
#' checks the date origin being used in an Excel file and returns is so it can be used in [convertToDate()]
#' @return One of "1900-01-01" or "1904-01-01".
#' @seealso [convertToDate()]
#' @examples
#'
#' ## create a file with some dates
#' \dontrun{
#' write.xlsx(as.Date("2015-01-10") - (0:4), file = "getDateOriginExample.xlsx")
#' m <- read.xlsx("getDateOriginExample.xlsx")
#'
#' ## convert to dates
#' do <- getDateOrigin(system.file("extdata", "readTest.xlsx", package = "openxlsx"))
#' convertToDate(m[[1]], do)
#' }
#' @export
getDateOrigin <- function(xlsxFile) {
  xlsxFile <- getFile(xlsxFile)
  if (!file.exists(xlsxFile)) {
    stop("File does not exist.")
  }

  if (grepl("\\.xls$|\\.xlm$", xlsxFile)) {
    stop("openxlsx can not read .xls or .xlm files!")
  }

  ## create temp dir and unzip
  xmlDir <- tempfile()
  xmlFiles <- unzip(xlsxFile, exdir = xmlDir)

  on.exit(unlink(xmlDir, recursive = TRUE), add = TRUE)

  workbook <- grep("workbook.xml$", xmlFiles, perl = TRUE, value = TRUE)
  workbook <- paste(unlist(readUTF8(workbook)), collapse = "")

  if (grepl('date1904="1"|date1904="true"', workbook, ignore.case = TRUE)) {
    origin <- "1904-01-01"
  } else {
    origin <- "1900-01-01"
  }

  return(origin)
}








#' @name getSheetNames
#' @title Get names of worksheets
#' @description Returns the worksheet names within an xlsx file
#' @author Alexander Walker
#' @param file An xlsx or xlsm file.
#' @return Character vector of worksheet names.
#' @examples
#' getSheetNames(system.file("extdata", "readTest.xlsx", package = "openxlsx"))
#' @export
getSheetNames <- function(file) {
  if (!file.exists(file)) {
    stop("file does not exist.")
  }

  if (grepl("\\.xls$|\\.xlm$", file)) {
    stop("openxlsx can not read .xls or .xlm files!")
  }

  ## create temp dir and unzip
  xmlDir <- tempfile()
  xmlFiles <- unzip(file, exdir = xmlDir)

  on.exit(unlink(xmlDir, recursive = TRUE), add = TRUE)

  workbook <- grep("workbook.xml$", xmlFiles, perl = TRUE, value = TRUE)
  workbook <- readUTF8(workbook)
  workbook <- removeHeadTag(workbook)
  sheets <- unlist(regmatches(workbook, gregexpr("(?<=<sheets>).*(?=</sheets>)", workbook, perl = TRUE)))
  sheets <- unlist(regmatches(sheets, gregexpr("<sheet[^>]*>", sheets, perl = TRUE)))

  ## Some veryHidden sheets do not have a sheet content and their rId is empty.
  ## Such sheets need to be filtered out because otherwise their sheet names
  ## occur in the list of all sheet names, leading to a wrong association
  ## of sheet names with sheet indeces.
  sheets <- grep('r:id="[[:blank:]]*"', sheets, invert = TRUE, value = TRUE)

  sheetNames <- unlist(regmatches(sheets, gregexpr('(?<=name=")[^"]+', sheets, perl = TRUE)))
  sheetNames <- replaceXMLEntities(sheetNames)

  return(sheetNames)
}




#' @name sheetVisibility
#' @title Get/set worksheet visible state
#' @description Get and set worksheet visible state
#' @param wb A workbook object
#' @return Character vector of worksheet names.
#' @return  Vector of "hidden", "visible", "veryHidden"
#' @examples
#'
#' wb <- createWorkbook()
#' addWorksheet(wb, sheetName = "S1", visible = FALSE)
#' addWorksheet(wb, sheetName = "S2", visible = TRUE)
#' addWorksheet(wb, sheetName = "S3", visible = FALSE)
#'
#' sheetVisibility(wb)
#' sheetVisibility(wb)[1] <- TRUE ## show sheet 1
#' sheetVisibility(wb)[2] <- FALSE ## hide sheet 2
#' sheetVisibility(wb)[3] <- "hidden" ## hide sheet 3
#' sheetVisibility(wb)[3] <- "veryHidden" ## hide sheet 3 from UI
#' @export
sheetVisibility <- function(wb) {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  state <- rep("visible", length(wb$workbook$sheets))
  state[grepl("hidden", wb$workbook$sheets)] <- "hidden"
  state[grepl("veryHidden", wb$workbook$sheets, ignore.case = TRUE)] <- "veryHidden"


  return(state)
}

#' @rdname sheetVisibility
#' @param value a logical/character vector the same length as sheetVisibility(wb)
#' @export
`sheetVisibility<-` <- function(wb, value) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  value <- tolower(as.character(value))
  if (!any(value %in% c("true", "visible"))) {
    stop("A workbook must have atleast 1 visible worksheet.")
  }

  value[value %in% "true"] <- "visible"
  value[value %in% "false"] <- "hidden"
  value[value %in% "veryhidden"] <- "veryHidden"


  exState0 <- regmatches(wb$workbook$sheets, regexpr('(?<=state=")[^"]+', wb$workbook$sheets, perl = TRUE))
  exState <- tolower(exState0)
  exState[exState %in% "true"] <- "visible"
  exState[exState %in% "hidden"] <- "hidden"
  exState[exState %in% "false"] <- "hidden"
  exState[exState %in% "veryhidden"] <- "veryHidden"

  if (length(value) != length(wb$workbook$sheets)) {
    stop(sprintf("value vector must have length equal to number of worksheets in Workbook [%s]", length(exState)))
  }

  inds <- which(value != exState)
  if (length(inds) == 0) {
    return(invisible(wb))
  }

  for (i in seq_along(wb$worksheets)) {
    wb$workbook$sheets[i] <- gsub(exState0[i], value[i], wb$workbook$sheets[i], fixed = TRUE)
  }

  invisible(wb)
}





#' @name pageBreak
#' @title add a page break to a worksheet
#' @description insert page breaks into a worksheet
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param i row or column number to insert page break.
#' @param type One of "row" or "column" for a row break or column break.
#' @export
#' @seealso [addWorksheet()]
#' @examples
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet 1")
#' writeData(wb, sheet = 1, x = iris)
#'
#' pageBreak(wb, sheet = 1, i = 10, type = "row")
#' pageBreak(wb, sheet = 1, i = 20, type = "row")
#' pageBreak(wb, sheet = 1, i = 2, type = "column")
#' \dontrun{
#' saveWorkbook(wb, "pageBreakExample.xlsx", TRUE)
#' }
#' ## In Excel: View tab -> Page Break Preview
pageBreak <- function(wb, sheet, i, type = "row") {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  sheet <- wb$validateSheet(sheet)

  type <- tolower(type)[1]
  if (!type %in% c("row", "column")) {
    stop("'type' argument must be 'row' or 'column'.")
  }

  if (!is.numeric(i)) {
    stop("'i' must be numeric.")
  }
  i <- round(i)

  if (type == "row") {
    wb$worksheets[[sheet]]$rowBreaks <- c(
      wb$worksheets[[sheet]]$rowBreaks,
      sprintf('<brk id="%s" max="16383" man="1"/>', i)
    )
  } else if (type == "column") {
    wb$worksheets[[sheet]]$colBreaks <- c(
      wb$worksheets[[sheet]]$colBreaks,
      sprintf('<brk id="%s" max="1048575" man="1"/>', i)
    )
  }


  # wb$worksheets[[sheet]]$autoFilter <- sprintf('<autoFilter ref="%s"/>', paste(getCellRefs(data.frame("x" = c(rows, rows), "y" = c(min(cols), max(cols)))), collapse = ":"))

  invisible(wb)
}


















#' @name conditionalFormat
#' @title Add conditional formatting to cells
#' @description DEPRECATED! USE [conditionalFormatting()]
#' @author Alexander Walker
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param cols Columns to apply conditional formatting to
#' @param rows Rows to apply conditional formatting to
#' @param rule The condition under which to apply the formatting or a vector of colours. See examples.
#' @param style A style to apply to those cells that satisfy the rule. A Style object returned from createStyle()
#' @details DEPRECATED! USE [conditionalFormatting()]
#'
#' Valid operators are "<", "<=", ">", ">=", "==", "!=". See Examples.
#' Default style given by: createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
#' @param type Either 'expression', 'colorscale' or 'databar'. If 'expression' the formatting is determined
#' by a formula.  If colorScale cells are coloured based on cell value. See examples.
#' @seealso [createStyle()]
#' @export
conditionalFormat <- function(wb, sheet, cols, rows, rule = NULL, style = NULL, type = "expression") {
  warning("conditionalFormat() has been deprecated. Use conditionalFormatting().")
  ## Rule always applies to top left of sqref, $ determine which cells the rule depends on
  ## Rule for "databar" and colourscale are colours of length 2/3 or 1 respectively.

  type <- tolower(type)
  if (tolower(type) %in% c("colorscale", "colourscale")) {
    type <- "colorScale"
  } else if (type == "databar") {
    type <- "dataBar"
  } else if (type != "expression") {
    stop("Invalid type argument.  Type must be 'expression', 'colourScale' or 'databar'")
  }

  ## rows and cols
  if (!is.numeric(cols)) {
    cols <- convertFromExcelRef(cols)
  }
  rows <- as.integer(rows)

  ## check valid rule
  if (type == "colorScale") {
    if (!length(rule) %in% 2:3) {
      stop("rule must be a vector containing 2 or 3 colours if type is 'colorScale'")
    }

    rule <- validateColour(rule, errorMsg = "Invalid colour specified in rule.")
    dxfId <- NULL
  } else if (type == "dataBar") {

    ## If rule is NULL use default colour
    if (is.null(rule)) {
      rule <- "FF638EC6"
    } else {
      rule <- validateColour(rule, errorMsg = "Invalid colour specified in rule.")
    }

    dxfId <- NULL
  } else { ## else type == "expression"

    rule <- toupper(gsub(" ", "", rule))
    rule <- replaceIllegalCharacters(rule)
    rule <- gsub("!=", "&lt;&gt;", rule)
    rule <- gsub("==", "=", rule)

    if (!grepl("[A-Z]", substr(rule, 1, 2))) {

      ## formula looks like "operatorX" , attach top left cell to rule
      rule <- paste0(getCellRefs(data.frame("x" = min(rows), "y" = min(cols))), rule)
    } ## else, there is a letter in the formula and apply as is

    if (is.null(style)) {
      style <- createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
    }

    invisible(dxfId <- wb$addDXFS(style))
  }


  invisible(wb$conditionalFormatCell(sheet,
    startRow = min(rows),
    endRow = max(rows),
    startCol = min(cols),
    endCol = max(cols),
    dxfId,
    formula = rule,
    type = type
  ))

  invisible(0)
}




#' @name all.equal
#' @aliases all.equal.Workbook
#' @title Check equality of workbooks
#' @description Check equality of workbooks
#' @method all.equal Workbook
#' @param target A `Workbook` object
#' @param current A `Workbook` object
#' @param ... ignored
all.equal.Workbook <- function(target, current, ...) {


  # print("Comparing workbooks...")
  #   ".rels",
  #   "app",
  #   "charts",
  #   "colWidths",
  #   "Content_Types",
  #   "core",
  #   "drawings",
  #   "drawings_rels",
  #   "media",
  #   "rowHeights",
  #   "workbook",
  #   "workbook.xml.rels",
  #   "worksheets",
  #   "sheetOrder"
  #   "sharedStrings",
  #   "tables",
  #   "tables.xml.rels",
  #   "theme"


  ## TODO
  # sheet_data

  x <- target
  y <- current




  nSheets <- length(names(x))
  failures <- NULL

  flag <- all(names(x$charts) %in% names(y$charts)) & all(names(y$charts) %in% names(x$charts))
  if (!flag) {
    message("charts not equal")
    failures <- c(failures, "wb$charts")
  }

  flag <- all(sapply(1:nSheets, function(i) isTRUE(all.equal(x$colWidths[[i]], y$colWidths[[i]]))))
  if (!flag) {
    message("colWidths not equal")
    failures <- c(failures, "wb$colWidths")
  }

  flag <- all(x$Content_Types %in% y$Content_Types) & all(y$Content_Types %in% x$Content_Types)
  if (!flag) {
    message("Content_Types not equal")
    failures <- c(failures, "wb$Content_Types")
  }

  flag <- all(unlist(x$core) == unlist(y$core))
  if (!flag) {
    message("core not equal")
    failures <- c(failures, "wb$core")
  }


  flag <- all(unlist(x$drawings) %in% unlist(y$drawings)) & all(unlist(y$drawings) %in% unlist(x$drawings))
  if (!flag) {
    message("drawings not equal")
    failures <- c(failures, "wb$drawings")
  }

  flag <- all(unlist(x$drawings_rels) %in% unlist(y$drawings_rels)) & all(unlist(y$drawings_rels) %in% unlist(x$drawings_rels))
  if (!flag) {
    message("drawings_rels not equal")
    failures <- c(failures, "wb$drawings_rels")
  }

  flag <- all(sapply(1:nSheets, function(i) isTRUE(all.equal(x$drawings_rels[[i]], y$drawings_rels[[i]]))))
  if (!flag) {
    message("drawings_rels not equal")
    failures <- c(failures, "wb$drawings_rels")
  }




  flag <- all(names(x$media) %in% names(y$media) & names(y$media) %in% names(x$media))
  if (!flag) {
    message("media not equal")
    failures <- c(failures, "wb$media")
  }

  flag <- all(sapply(1:nSheets, function(i) isTRUE(all.equal(x$rowHeights[[i]], y$rowHeights[[i]]))))
  if (!flag) {
    message("rowHeights not equal")
    failures <- c(failures, "wb$rowHeights")
  }

  flag <- all(sapply(1:nSheets, function(i) isTRUE(all.equal(names(x$rowHeights[[i]]), names(y$rowHeights[[i]])))))
  if (!flag) {
    message("rowHeights not equal")
    failures <- c(failures, "wb$rowHeights")
  }

  flag <- all(x$sharedStrings %in% y$sharedStrings) & all(y$sharedStrings %in% x$sharedStrings) & (length(x$sharedStrings) == length(y$sharedStrings))
  if (!flag) {
    message("sharedStrings not equal")
    failures <- c(failures, "wb$sharedStrings")
  }



  # flag <- sapply(1:nSheets, function(i) isTRUE(all.equal(x$worksheets[[i]]$sheet_data, y$worksheets[[i]]$sheet_data)))
  # if(!all(flag)){
  #
  #   tmp_x <- x$sheet_data[[which(!flag)[[1]]]]
  #   tmp_y <- y$sheet_data[[which(!flag)[[1]]]]
  #
  #   tmp_x_e <- sapply(tmp_x, "[[", "r")
  #   tmp_y_e <- sapply(tmp_y, "[[", "r")
  #   flag <- paste0(tmp_x_e, "") != paste0(tmp_x_e, "")
  #   if(any(flag)){
  #     message(sprintf("sheet_data %s not equal", which(!flag)[[1]]))
  #     message(sprintf("r elements: %s", paste(which(flag), collapse = ", ")))
  #     return(FALSE)
  #   }
  #
  #   tmp_x_e <- sapply(tmp_x, "[[", "t")
  #   tmp_y_e <- sapply(tmp_y, "[[", "t")
  #   flag <- paste0(tmp_x_e, "") != paste0(tmp_x_e, "")
  #   if(any(flag)){
  #     message(sprintf("sheet_data %s not equal", which(!flag)[[1]]))
  #     message(sprintf("t elements: %s", paste(which(isTRUE(flag)), collapse = ", ")))
  #     return(FALSE)
  #   }
  #
  #
  #   tmp_x_e <- sapply(tmp_x, "[[", "v")
  #   tmp_y_e <- sapply(tmp_y, "[[", "v")
  #   flag <- paste0(tmp_x_e, "") != paste0(tmp_x_e, "")
  #   if(any(flag)){
  #     message(sprintf("sheet_data %s not equal", which(!flag)[[1]]))
  #     message(sprintf("v elements: %s", paste(which(flag), collapse = ", ")))
  #     return(FALSE)
  #   }
  #
  #   tmp_x_e <- sapply(tmp_x, "[[", "f")
  #   tmp_y_e <- sapply(tmp_y, "[[", "f")
  #   flag <- paste0(tmp_x_e, "") != paste0(tmp_x_e, "")
  #   if(any(flag)){
  #     message(sprintf("sheet_data %s not equal", which(!flag)[[1]]))
  #     message(sprintf("f elements: %s", paste(which(flag), collapse = ", ")))
  #     return(FALSE)
  #   }
  # }


  flag <- all(names(x$styles) %in% names(y$styles)) & all(names(y$styles) %in% names(x$styles))
  if (!flag) {
    message("names styles not equal")
    failures <- c(failures, "names of styles not equal")
  }

  flag <- all(unlist(x$styles) %in% unlist(y$styles)) & all(unlist(y$styles) %in% unlist(x$styles))
  if (!flag) {
    message("styles not equal")
    failures <- c(failures, "styles not equal")
  }


  flag <- length(x$styleObjects) == length(y$styleObjects)
  if (!flag) {
    message("styleObjects lengths not equal")
    failures <- c(failures, "styleObjects lengths not equal")
  }


  nStyles <- length(x$styleObjects)
  if (nStyles > 0) {
    for (i in 1:nStyles) {
      sx <- x$styleObjects[[i]]
      sy <- y$styleObjects[[i]]

      flag <- isTRUE(all.equal(sx$sheet, sy$sheet))
      if (!flag) {
        message(sprintf("styleObjects '%s' sheet name not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' sheet name not equal", i))
      }


      flag <- isTRUE(all.equal(sx$rows, sy$rows))
      if (!flag) {
        message(sprintf("styleObjects '%s' rows not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' rows not equal", i))
      }

      flag <- isTRUE(all.equal(sx$cols, sy$cols))
      if (!flag) {
        message(sprintf("styleObjects '%s' cols not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' cols not equal", i))
      }

      ## check style class equality
      flag <- isTRUE(all.equal(sx$style$fontName, sy$style$fontName))
      if (!flag) {
        message(sprintf("styleObjects '%s' fontName not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' fontName not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$fontColour, sy$style$fontColour))
      if (!flag) {
        message(sprintf("styleObjects '%s' fontColour not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' fontColour not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$fontSize, sy$style$fontSize))
      if (!flag) {
        message(sprintf("styleObjects '%s' fontSize not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' fontSize not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$fontFamily, sy$style$fontFamily))
      if (!flag) {
        message(sprintf("styleObjects '%s' fontFamily not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' fontFamily not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$fontDecoration, sy$style$fontDecoration))
      if (!flag) {
        message(sprintf("styleObjects '%s' fontDecoration not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' fontDecoration not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$borderTop, sy$style$borderTop))
      if (!flag) {
        message(sprintf("styleObjects '%s' borderTop not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' borderTop not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$borderLeft, sy$style$borderLeft))
      if (!flag) {
        message(sprintf("styleObjects '%s' borderLeft not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' borderLeft not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$borderRight, sy$style$borderRight))
      if (!flag) {
        message(sprintf("styleObjects '%s' borderRight not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' borderRight not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$borderBottom, sy$style$borderBottom))
      if (!flag) {
        message(sprintf("styleObjects '%s' borderBottom not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' borderBottom not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$borderTopColour, sy$style$borderTopColour))
      if (!flag) {
        message(sprintf("styleObjects '%s' borderTopColour not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' borderTopColour not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$borderLeftColour, sy$style$borderLeftColour))
      if (!flag) {
        message(sprintf("styleObjects '%s' borderLeftColour not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' borderLeftColour not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$borderRightColour, sy$style$borderRightColour))
      if (!flag) {
        message(sprintf("styleObjects '%s' borderRightColour not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' borderRightColour not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$borderBottomColour, sy$style$borderBottomColour))
      if (!flag) {
        message(sprintf("styleObjects '%s' borderBottomColour not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' borderBottomColour not equal", i))
      }


      flag <- isTRUE(all.equal(sx$style$halign, sy$style$halign))
      if (!flag) {
        message(sprintf("styleObjects '%s' halign not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' halign not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$valign, sy$style$valign))
      if (!flag) {
        message(sprintf("styleObjects '%s' valign not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' valign not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$indent, sy$style$indent))
      if (!flag) {
        message(sprintf("styleObjects '%s' indent not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' indent not equal", i))
      }


      flag <- isTRUE(all.equal(sx$style$textRotation, sy$style$textRotation))
      if (!flag) {
        message(sprintf("styleObjects '%s' textRotation not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' textRotation not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$numFmt, sy$style$numFmt))
      if (!flag) {
        message(sprintf("styleObjects '%s' numFmt not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' numFmt not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$fill, sy$style$fill))
      if (!flag) {
        message(sprintf("styleObjects '%s' fill not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' fill not equal", i))
      }

      flag <- isTRUE(all.equal(sx$style$wrapText, sy$style$wrapText))
      if (!flag) {
        message(sprintf("styleObjects '%s' wrapText not equal", i))
        failures <- c(failures, sprintf("styleObjects '%s' wrapText not equal", i))
      }
    }
  }


  flag <- all(x$sheet_names %in% y$sheet_names) & all(y$sheet_names %in% x$sheet_names)
  if (!flag) {
    message("names workbook not equal")
    failures <- c(failures, "names workbook not equal")
  }

  flag <- all(unlist(x$workbook) %in% unlist(y$workbook)) & all(unlist(y$workbook) %in% unlist(x$workbook))
  if (!flag) {
    message("workbook not equal")
    failures <- c(failures, "wb$workbook")
  }

  flag <- all(unlist(x$workbook.xml.rels) %in% unlist(y$workbook.xml.rels)) & all(unlist(y$workbook.xml.rels) %in% unlist(x$workbook.xml.rels))
  if (!flag) {
    message("workbook.xml.rels not equal")
    failures <- c(failures, "wb$workbook.xml.rels")
  }


  for (i in 1:nSheets) {
    ws_x <- x$worksheets[[i]]
    ws_y <- y$worksheets[[i]]

    flag <- all(names(ws_x) %in% names(ws_y)) & all(names(ws_y) %in% names(ws_x))
    if (!flag) {
      message(sprintf("names of worksheet elements for sheet %s not equal", i))
      failures <- c(failures, sprintf("names of worksheet elements for sheet %s not equal", i))
    }

    nms <- c(
      "sheetPr", "dataValidations", "sheetViews", "cols", "pageMargins",
      "extLst", "conditionalFormatting", "oleObjects",
      "colBreaks", "dimension", "drawing", "sheetFormatPr", "tableParts",
      "mergeCells", "hyperlinks", "headerFooter", "autoFilter",
      "rowBreaks", "pageSetup", "freezePane", "legacyDrawingHF", "legacyDrawing"
    )

    for (j in nms) {
      flag <- isTRUE(all.equal(gsub(" |\t", "", ws_x[[j]]), gsub(" |\t", "", ws_y[[j]])))
      if (!flag) {
        message(sprintf("worksheet '%s', element '%s' not equal", i, j))
        failures <- c(failures, sprintf("worksheet '%s', element '%s' not equal", i, j))
      }
    }
  }


  flag <- all(unlist(x$sheetOrder) %in% unlist(y$sheetOrder)) & all(unlist(y$sheetOrder) %in% unlist(x$sheetOrder))
  if (!flag) {
    message("sheetOrder not equal")
    failures <- c(failures, "sheetOrder not equal")
  }


  flag <- length(x$tables) == length(y$tables)
  if (!flag) {
    message("length of tables not equal")
    failures <- c(failures, "length of tables not equal")
  }

  flag <- all(names(x$tables) == names(y$tables))
  if (!flag) {
    message("names of tables not equal")
    failures <- c(failures, "names of tables not equal")
  }

  flag <- all(unlist(x$tables) == unlist(y$tables))
  if (!flag) {
    message("tables not equal")
    failures <- c(failures, "tables not equal")
  }


  flag <- isTRUE(all.equal(x$tables.xml.rels, y$tables.xml.rels))
  if (!flag) {
    message("tables.xml.rels not equal")
    failures <- c(failures, "tables.xml.rels not equal")
  }

  flag <- x$theme == y$theme
  if (!flag) {
    message("theme not equal")
    failures <- c(failures, "theme not equal")
  }

  if (!is.null(failures)) {
    return(FALSE)
  }


  #   "connections",
  #   "externalLinks",
  #   "externalLinksRels",
  #   "headFoot",
  #   "pivotTables",
  #   "pivotTables.xml.rels",
  #   "pivotDefinitions",
  #   "pivotRecords",
  #   "pivotDefinitionsRels",
  #   "queryTables",
  #   "slicers",
  #   "slicerCaches",
  #   "vbaProject",


  return(TRUE)
}



#' @name sheetVisible
#' @title Get worksheet visible state.
#' @description DEPRECATED - Use function 'sheetVisibility()
#' @author Alexander Walker
#' @param wb A workbook object
#' @return Character vector of worksheet names.
#' @return  TRUE if sheet is visible, FALSE if sheet is hidden
#' @examples
#'
#' wb <- createWorkbook()
#' addWorksheet(wb, sheetName = "S1", visible = FALSE)
#' addWorksheet(wb, sheetName = "S2", visible = TRUE)
#' addWorksheet(wb, sheetName = "S3", visible = FALSE)
#'
#' sheetVisible(wb)
#' sheetVisible(wb)[1] <- TRUE ## show sheet 1
#' sheetVisible(wb)[2] <- FALSE ## hide sheet 2
#' @export
sheetVisible <- function(wb) {
  warning("This function is deprecated. Use function 'sheetVisibility()'")

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  state <- rep(TRUE, length(wb$workbook$sheets))
  state[grepl("hidden", wb$workbook$sheets)] <- FALSE

  return(state)
}

#' @rdname sheetVisible
#' @param value a logical vector the same length as sheetVisible(wb)
#' @export
`sheetVisible<-` <- function(wb, value) {
  warning("This function is deprecated. Use function 'sheetVisibility()'")

  if (!is.logical(value)) {
    stop("value must be a logical vector.")
  }

  if (!any(value)) {
    stop("A workbook must have atleast 1 visible worksheet.")
  }

  value <- as.character(value)
  value[value %in% "TRUE"] <- "visible"
  value[value %in% "FALSE"] <- "hidden"

  exState <- rep("visible", length(wb$workbook$sheets))
  exState[grepl("hidden", wb$workbook$sheets)] <- "hidden"

  if (length(value) != length(wb$workbook$sheets)) {
    stop(sprintf("value vector must have length equal to number of worksheets in Workbook [%s]", length(exState)))
  }

  inds <- which(value != exState)
  if (length(inds) == 0) {
    return(invisible(wb))
  }

  for (i in inds) {
    wb$workbook$sheets[i] <- gsub(exState[i], value[i], wb$workbook$sheets[i])
  }

  invisible(wb)
}



#' @name copyWorkbook
#' @title Copy a Workbook object.
#' @description Just a wrapper of wb$copy()
#' @param wb A workbook object
#' @return Workbook
#' @examples
#'
#' wb <- createWorkbook()
#' wb2 <- wb ## does not create a copy
#' wb3 <- copyWorkbook(wb) ## wrapper for wb$copy()
#'
#' addWorksheet(wb, "Sheet1") ## adds worksheet to both wb and wb2 but not wb3
#'
#' names(wb)
#' names(wb2)
#' names(wb3)
#' @export
copyWorkbook <- function(wb) {
  if (!inherits(wb, "Workbook")) {
    stop("argument must be a Workbook.")
  }

  return(wb$copy())
}





#' @name getTables
#' @title List Excel tables in a workbook
#' @description List Excel tables in a workbook
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @return character vector of table names on the specified sheet
#' @examples
#'
#' wb <- createWorkbook()
#' addWorksheet(wb, sheetName = "Sheet 1")
#' writeDataTable(wb, sheet = "Sheet 1", x = iris)
#' writeDataTable(wb, sheet = 1, x = mtcars, tableName = "mtcars", startCol = 10)
#'
#' getTables(wb, sheet = "Sheet 1")
#' @export
getTables <- function(wb, sheet) {
  if (!inherits(wb, "Workbook")) {
    stop("argument must be a Workbook.")
  }

  if (length(sheet) != 1) {
    stop("sheet argument must be length 1")
  }

  if (length(wb$tables) == 0) {
    return(character(0))
  }

  sheet <- wb$validateSheet(sheetName = sheet)

  table_sheets <- attr(wb$tables, "sheet")
  tables <- attr(wb$tables, "tableName")
  refs <- names(wb$tables)

  refs <- refs[table_sheets == sheet & !grepl("openxlsx_deleted", tables, fixed = TRUE)]
  tables <- tables[table_sheets == sheet & !grepl("openxlsx_deleted", tables, fixed = TRUE)]

  if (length(tables) > 0) {
    attr(tables, "refs") <- refs
  }

  return(tables)
}





#' @name removeTable
#' @title Remove an Excel table in a workbook
#' @description List Excel tables in a workbook
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param table Name of table to remove. See [getTables()]
#' @return character vector of table names on the specified sheet
#' @examples
#'
#' wb <- createWorkbook()
#' addWorksheet(wb, sheetName = "Sheet 1")
#' addWorksheet(wb, sheetName = "Sheet 2")
#' writeDataTable(wb, sheet = "Sheet 1", x = iris, tableName = "iris")
#' writeDataTable(wb, sheet = 1, x = mtcars, tableName = "mtcars", startCol = 10)
#'
#'
#' removeWorksheet(wb, sheet = 1) ## delete worksheet removes table objects
#'
#' writeDataTable(wb, sheet = 1, x = iris, tableName = "iris")
#' writeDataTable(wb, sheet = 1, x = mtcars, tableName = "mtcars", startCol = 10)
#'
#' ## removeTable() deletes table object and all data
#' getTables(wb, sheet = 1)
#' removeTable(wb = wb, sheet = 1, table = "iris")
#' writeDataTable(wb, sheet = 1, x = iris, tableName = "iris", startCol = 1)
#'
#' getTables(wb, sheet = 1)
#' removeTable(wb = wb, sheet = 1, table = "iris")
#' writeDataTable(wb, sheet = 1, x = iris, tableName = "iris", startCol = 1)
#' \dontrun{
#' saveWorkbook(wb = wb, file = "removeTableExample.xlsx", overwrite = TRUE)
#' }
#'
#' @export
removeTable <- function(wb, sheet, table) {
  if (!inherits(wb, "Workbook")) {
    stop("argument must be a Workbook.")
  }

  if (length(sheet) != 1) {
    stop("sheet argument must be length 1")
  }

  if (length(table) != 1) {
    stop("table argument must be length 1")
  }

  ## delete table object and all data in it
  sheet <- wb$validateSheet(sheetName = sheet)

  if (!table %in% attr(wb$tables, "tableName")) {
    stop(sprintf("table '%s' does not exist.", table), call. = FALSE)
  }

  ## get existing tables
  table_sheets <- attr(wb$tables, "sheet")
  table_names <- attr(wb$tables, "tableName")
  refs <- names(wb$tables)

  ## delete table object (by flagging as deleted)
  inds <- which(table_sheets %in% sheet & table_names %in% table)
  table_name_original <- table_names[inds]

  table_names[inds] <- paste0(table_name_original, "_openxlsx_deleted")
  attr(wb$tables, "tableName") <- table_names

  ## delete reference from worksheet to table
  worksheet_table_names <- attr(wb$worksheets[[sheet]]$tableParts, "tableName")
  to_remove <- which(worksheet_table_names == table_name_original)

  wb$worksheets[[sheet]]$tableParts <- wb$worksheets[[sheet]]$tableParts[-to_remove]
  attr(wb$worksheets[[sheet]]$tableParts, "tableName") <- worksheet_table_names[-to_remove]


  ## Now delete data from the worksheet
  refs <- strsplit(refs[[inds]], split = ":")[[1]]
  rows <- as.integer(gsub("[A-Z]", "", refs))
  rows <- seq(from = rows[1], to = rows[2], by = 1)

  cols <- convertFromExcelRef(refs)
  cols <- seq(from = cols[1], to = cols[2], by = 1)

  ## now delete data
  deleteData(wb = wb, sheet = sheet, rows = rows, cols = cols, gridExpand = TRUE)

  invisible(0)
}

#' @name groupColumns
#' @title Group columns
#' @description Group a selection of columns
#' @author Joshua Sturm, Reinhold Kainhofer
#' @param wb A workbook object.
#' @param sheet A name or index of a worksheet.
#' @param cols Indices of cols to group. Can be either a vector of indices to 
#'             group at the same level or a (named) list of numeric vectors of 
#'             indices to create multiple groupings at once. The names of the 
#'             entries determine the grouping level. If no names are given, 
#'             the `level` parameter is used as default.
#' @param hidden Logical vector. If TRUE the grouped columns are hidden. Defaults to FALSE.
#' @param level Grouping level (higher value indicates multiple nestings) for the 
#'              group. A vector to assign different grouping levels to the indices. 
#'              A value of -1 indicates that the grouping level should be derived 
#'              from the existing grouping (one level added)
#' @details Group columns together, with the option to hide them.
#'
#' NOTE: [setColWidths()] has a conflicting `hidden` parameter; changing one will update the other.
#' @seealso [ungroupColumns()] to ungroup columns. [groupRows()] for grouping rows.
#' @export
#'
groupColumns <- function(wb, sheet, cols, hidden = FALSE, level = -1) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  sheet <- wb$validateSheet(sheet)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  if (any(cols < 1L)) {
    stop("Invalid columns selected (<= 0).")
  }

  if (!is.logical(hidden)) {
    stop("Hidden should be a logical value (TRUE/FALSE).")
  }

  if (length(hidden) > length(cols)) {
    stop("Hidden argument is of greater length than number of cols.")
  }

  if(is.list(cols)) {
    if (!is.null(names(cols))) {
      levels <- unlist(lapply(names(cols), function(x) rep(as.character(x), length(cols[[x]]))))
    } else {
      levels <- rep(as.character(level), length(unlist(cols)))
    }
    cols <- unlist(cols)
  } else {
    levels <- rep(level, length(cols))
  }
  
  hidden <- as.character(as.integer(rep(hidden, length.out = length(cols))))

  hidden <- hidden[!duplicated(cols)]
  levels <- levels[!duplicated(cols)]
  cols <- cols[!duplicated(cols)]
  cols <- convertFromExcelRef(cols)
  names(levels) <- cols
  
  if (length(wb$colWidths[[sheet]]) > 0) {
    existing_cols <- names(wb$colWidths[[sheet]])
    existing_hidden <- attr(wb$colWidths[[sheet]], "hidden", exact = TRUE)

    if (any(existing_cols %in% cols)) {
      for (i in intersect(existing_cols, cols)) {
        width_hidden <- attr(wb$colWidths[[sheet]], "hidden")[attr(wb$colWidths[[sheet]], "names") == i]
        outline_hidden <- as.character(as.integer(hidden))[cols == i]

        if (width_hidden != outline_hidden) {
          attr(wb$colWidths[[sheet]], "hidden")[attr(wb$colWidths[[sheet]], "names") == i] <- outline_hidden
        }
      }

      # cols <- cols[!cols %in% existing_cols]
      # hidden <- attr(wb$colOutlineLevels[[sheet]], "hidden")[attr(wb$colOutlineLevels[[sheet]], "name") %in% cols]

      # wb$colOutlineLevels[[sheet]] <- cols
      # attr(wb$colOutlineLevels[[sheet]], "hidden") <- as.character(as.integer(hidden))
    }
  }

  if (length(wb$colOutlineLevels[[sheet]]) > 0) {
    existing_cols <- names(wb$colOutlineLevels[[sheet]])
    existing_levels <- unname(wb$colOutlineLevels[[sheet]])
    existing_hidden <- attr(wb$colOutlineLevels[[sheet]], "hidden")
    
    # check if column is already grouped
    flag <- existing_cols %in% cols
    # Find indices of cols that already exist
    existing_outline_indices = which(flag)
    existing_outline = existing_cols[existing_outline_indices]
    existing_cols_indices = match(existing_outline, cols)
    
    # Auto-detect new level if required
    new_level <- "1"
    if (any(flag)) {
      new_level <- as.character(max(as.numeric(existing_levels[flag])) + 1)
    }
    levels[levels < 0] = as.character(new_level)
    
    if (any(flag)) {
      # Assign the given values to existing col definitions (indices were extracted above)
      existing_hidden[existing_outline_indices] <- hidden[existing_cols_indices]
      existing_levels[existing_outline_indices] <- levels[existing_cols_indices]
      
      # Append all remaining new entries:
      all_names <- c(existing_cols, cols[-existing_cols_indices])
      all_levels <- c(existing_levels, levels[-existing_cols_indices])
      all_hidden <- c(existing_hidden, hidden[-existing_cols_indices])
    } else {
      # only new cols were added, no existing modified
      all_names = c(existing_cols, cols)
      all_levels = c(existing_levels, levels)
      all_hidden = c(existing_hidden, hidden)
    }

    ord <- order(as.integer(all_names))
    all_names <- all_names[ord]
    all_levels <- as.character(all_levels[ord])
    all_hidden <- all_hidden[ord]

    names(all_levels) <- all_names
    wb$colOutlineLevels[[sheet]] <- all_levels
    levels <- all_levels
    attr(wb$colOutlineLevels[[sheet]], "hidden") <- as.character(as.integer(all_hidden))
    hidden <- all_hidden
  } else {
    levels[levels < 1] = "1"
    names(levels) <- cols
    wb$colOutlineLevels[[sheet]] <- levels
    attr(wb$colOutlineLevels[[sheet]], "hidden") <- as.character(as.integer(hidden))
  }
  
  # Finally, update the sheetFormatPr XML element with the maximum outline level
  max_outline = max(as.numeric(wb$colOutlineLevels[[sheet]]))
  outline_attr <- paste0(' outlineLevelCol="', max_outline, '"')
  if (!grepl("outlineLevelCol", wb$worksheets[[sheet]]$sheetFormatPr)) {
    wb$worksheets[[sheet]]$sheetFormatPr <- sub("/>", paste0(outline_attr, "/>"), wb$worksheets[[sheet]]$sheetFormatPr)
  } else {
    wb$worksheets[[sheet]]$sheetFormatPr <- sub(' outlineLevelCol="[0-9]+"', outline_attr, wb$worksheets[[sheet]]$sheetFormatPr)
  }
  
  invisible(0)
}

#' @name ungroupColumns
#' @title Ungroup Columns
#' @description Ungroup a selection of columns
#' @author Joshua Sturm
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param cols Indices of columns to ungroup
#' @details If column was previously hidden, it will now be shown
#' @seealso [ungroupRows()] To ungroup rows
#' @export

ungroupColumns <- function(wb, sheet, cols) {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  sheet <- wb$validateSheet(sheet)

  if (!is.numeric(cols)) {
    cols <- convertFromExcelRef(cols)
  }

  if (any(cols) < 1L) {
    stop("Invalid columns selected (<= 0).")
  }

  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  customCols <- as.integer(names(wb$colOutlineLevels[[sheet]]))
  ungroupInds <- which(customCols %in% cols)
  if (length(ungroupInds) > 0) {
    # decrement the outline level by 1, set to visible and remove all columns that are no longer grouped at all (i.e. have a level "0" or "-1" (just in case))
    levels <- as.character(as.integer(wb$colOutlineLevels[[sheet]][ungroupInds]) - 1)
    wb$colOutlineLevels[[sheet]][ungroupInds] <- levels
    attr(wb$colOutlineLevels[[sheet]], "hidden")[ungroupInds] <- "0"
    
    removeInds <- which(wb$colOutlineLevels[[sheet]] %in% c("-1", "0"))
    wb$colOutlineLevels[[sheet]] <- wb$colOutlineLevels[[sheet]][-removeInds]
    attr(wb$colOutlineLevels[[sheet]], "hidden") = attr(wb$colOutlineLevels[[sheet]], "hidden")[-removeInds]
  }
  
  if (length(wb$outlineLevels[[sheet]]) == 0) {
    wb$worksheets[[sheet]]$sheetFormatPr <- sub(' outlineLevelCol="[0-9]+"', "", wb$worksheets[[sheet]]$sheetFormatPr)
  } else {
    max_level = max(as.integer(wb$colOutlineLevels[[sheet]]))
    wb$worksheets[[sheet]]$sheetFormatPr <- sub(' outlineLevelCol="[0-9]+"', paste0(" outlineLevelCol=\"", max_level, "\""), wb$worksheets[[sheet]]$sheetFormatPr)
  }
  

  if (length(wb$colWidths[[sheet]]) > 0) {
    if (any(cols %in% names(wb$colWidths[[sheet]]))) {
      attr(wb$colWidths[[sheet]], "hidden")[attr(wb$colWidths[[sheet]], "names") %in% cols] <- "0"
    }
  }
}


#' @name groupRows
#' @title Group Rows
#' @description Group a selection of rows
#' @author Joshua Sturm
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param rows Indices of rows to group. Can be either a vector of indices to group at the same level or a (named) list of numeric vectors of indices to create multiple groupings at once. The names of the entries determine the grouping level. If no names are given, the parameter level is used as default.
#' @param hidden Logical vector. If TRUE the grouped columns are hidden. Defaults to FALSE
#' @param level Grouping level (higher value indicates multiple nestings) for the 
#'              group. A vector to assign different grouping levels to the indices. 
#'              A value of -1 indicates that the grouping level should be derived 
#'              from the existing grouping (one level added)
#' @seealso [ungroupRows()] to ungroup rows. [groupColumns()] for grouping columns.
#' @examples
#' wb <- createWorkbook()
#' addWorksheet(wb, 'Sheet1')
#' addWorksheet(wb, 'Sheet2')
#'
#' writeData(wb, "Sheet1", iris)
#' writeData(wb, "Sheet2", iris)
#'
#' ## create list of groups
#' # lines used for grouping (here: species)
#' grp <- list(
#'   seq(2, 51),
#'   seq(52, 101),
#'   seq(102, 151)
#' )
#' # assign group levels
#' names(grp) <- c("1","0","1")
#' groupRows(wb, "Sheet1", rows = grp)
#'
#' # different grouping
#' names(grp) <- c("1","2","3")
#' groupRows(wb, "Sheet2", rows = grp)
#' 
#' # alternatively, one can call groupRows multiple times
#' addWorksheet(wb, 'Sheet3')
#' writeData(wb, "Sheet3", iris)
#' groupRows(wb, "Sheet3", 2:51, level = 1)
#' groupRows(wb, "Sheet3", 102:151, level = 1)
#' 
#' addWorksheet(wb, 'Sheet4')
#' writeData(wb, "Sheet4", iris)
#' groupRows(wb, "Sheet4", 2:51, level = 1)
#' groupRows(wb, "Sheet4", 52:101, level = 2)
#' groupRows(wb, "Sheet4", 102:151, level = 3)
#' 
#' # Nested grouping can also be achieved without explicitly given the levels
#' addWorksheet(wb, 'Sheet5')
#' writeData(wb, "Sheet5", iris)
#' groupRows(wb, "Sheet5", 2:151)
#' groupRows(wb, "Sheet5", 52:151)
#' groupRows(wb, "Sheet5", 102:151)
#' 
#' 
#' @export
groupRows <- function(wb, sheet, rows, hidden = FALSE, level = -1) {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  if(is.list(rows)) {
    if (!is.null(names(rows))) {
      levels <- unlist(lapply(names(rows), function(x) rep(as.character(x), length(rows[[x]]))))
    } else {
      levels <- rep(as.character(level), length(unlist(rows)))
    }
    rows <- unlist(rows)
  } else {
    levels <- rep(level, length(rows))
  }

  sheet <- wb$validateSheet(sheet)

  if (length(hidden) > length(rows)) {
    stop("Hidden argument is of greater length than number of rows.")
  }

  if (!is.logical(hidden)) {
    stop("Hidden should be a logical value (TRUE/FALSE).")
  }

  if (any(rows < 1L)) {
    stop("Invalid rows entered (<= 0).")
  }

  hidden <- rep(as.character(as.integer(hidden)), length.out = length(rows))

  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  # Remove duplicates
  hidden <- hidden[!duplicated(rows)]
  levels <- levels[!duplicated(rows)]
  rows <- rows[!duplicated(rows)]

  names(levels) <- rows

  wb$groupRows(sheet = sheet, rows = rows, hidden = hidden, levels = levels)
}

#' @name ungroupRows
#' @title Ungroup Rows
#' @description Ungroup a selection of rows
#' @author Joshua Sturm
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param rows Indices of rows to ungroup
#' @details If row was previously hidden, it will now be shown
#' @seealso [ungroupColumns()]
#' @export

ungroupRows <- function(wb, sheet, rows) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  sheet <- wb$validateSheet(sheet)

  if (any(rows < 1L)) {
    stop("Invalid rows entered (<= 0).")
  }

  customRows <- as.integer(names(wb$outlineLevels[[sheet]]))
  ungroupInds <- which(customRows %in% rows)
  if (length(ungroupInds) > 0) {
    # decrement the outline level by 1, set to visible and remove all rows that are no longer grouped at all (i.e. have a level "0" or "-1" (just in case))
    levels <- as.character(as.integer(wb$outlineLevels[[sheet]][ungroupInds]) - 1)
    wb$outlineLevels[[sheet]][ungroupInds] <- levels
    attr(wb$outlineLevels[[sheet]], "hidden")[ungroupInds] <- "0"

    removeInds <- which(wb$outlineLevels[[sheet]] %in% c("-1", "0"))
    wb$outlineLevels[[sheet]] <- wb$outlineLevels[[sheet]][-removeInds]
    attr(wb$outlineLevels[[sheet]], "hidden") = attr(wb$outlineLevels[[sheet]], "hidden")[-removeInds]
  }

  if (length(wb$outlineLevels[[sheet]]) == 0) {
    wb$worksheets[[sheet]]$sheetFormatPr <- sub(' outlineLevelRow="[0-9]+"', "", wb$worksheets[[sheet]]$sheetFormatPr)
  } else {
    max_level = max(as.integer(wb$outlineLevels[[sheet]]))
    wb$worksheets[[sheet]]$sheetFormatPr <- sub(' outlineLevelRow="[0-9]+"', paste0(" outlineLevelRow=\"", max_level, "\""), wb$worksheets[[sheet]]$sheetFormatPr)
  }
}




#' @name addCreator
#' @title Add another author to the meta data of the file.
#' @author Philipp Schauberger
#' @description Just a wrapper of wb$addCreator()
#' @param wb A workbook object
#' @param Creator A string object with the name of the creator
#' @examples
#'
#' wb <- createWorkbook()
#' addCreator(wb, "test")
#' @export
addCreator <- function(wb, Creator) {
  if (!inherits(wb, "Workbook")) {
    stop("argument must be a Workbook.")
  }

  invisible(wb$addCreator(Creator))
}

#' @name setLastModifiedBy
#' @title Set the author who modified the file last.
#' @author Philipp Schauberger
#' @description Just a wrapper of wb$changeLastModifiedBy()
#' @param wb A workbook object
#' @param LastModifiedBy A string object with the name of the LastModifiedBy-User
#' @examples
#'
#' wb <- createWorkbook()
#' setLastModifiedBy(wb, "test")
#' @export
setLastModifiedBy <- function(wb, LastModifiedBy) {
  if (!inherits(wb, "Workbook")) {
    stop("argument must be a Workbook.")
  }

  invisible(wb$changeLastModifiedBy(LastModifiedBy))
}



#' @name getCreators
#' @title Get the names of the authors from the meta data of the file.
#' @description Just a wrapper of wb$getCreators()
#' @param wb A workbook object
#' @author Philipp Schauberger
#' @return vector of creators
#' @examples
#'
#' wb <- createWorkbook()
#' getCreators(wb)
#' @export
getCreators <- function(wb) {
  if (!inherits(wb, "Workbook")) {
    stop("argument must be a Workbook.")
  }

  return(wb$getCreators())
}

#' @name activeSheet
#' @title Get/set active sheet of the workbook
#' @author Philipp Schauberger
#' @description Get and set active sheet of the workbook
#' @param wb A workbook object
#' @return return the active sheet of the workbook
#' @examples
#'
#' wb <- createWorkbook()
#' addWorksheet(wb, sheetName = "S1")
#' addWorksheet(wb, sheetName = "S2")
#' addWorksheet(wb, sheetName = "S3")
#'
#' activeSheet(wb) # default value is the first sheet active
#' activeSheet(wb) <- 1 ## active sheet S1
#' activeSheet(wb)
#' activeSheet(wb) <- "S2" ## active sheet S2
#' activeSheet(wb)
#' @export
activeSheet <- function(wb) {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }


  return(wb$ActiveSheet)
}

#' @rdname activeSheet
#' @param value index of the active sheet or name of the active sheet
#' @export
`activeSheet<-` <- function(wb, value) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  invisible(wb$setactiveSheet(value))
  invisible(wb)
}

test_that("can read default sizes from new workbook", {
  wb <- createWorkbook()
  
  expect_equal(getWindowSize(wb), 
               list(xWindow = "0", yWindow = "0", windowWidth = "13125", windowHeight = "6105"))
})

test_that("can change sizes in new workbook", {
  wb <- createWorkbook()
  setWindowSize(wb, xWindow = 100, yWindow = 200, windowWidth = 20000, windowHeight = 15000)
  
  expect_equal(getWindowSize(wb), 
               list(xWindow = "100", yWindow = "200", windowWidth = "20000", windowHeight = "15000"))
})

test_that("can change only a few parameters", {
  wb <- createWorkbook()
  setWindowSize(wb, yWindow = 300, windowWidth = 12000) # others are the default
  
  expect_equal(getWindowSize(wb), 
               list(xWindow = "0", yWindow = "300", windowWidth = "12000", windowHeight = "6105"))
})

test_that("window size and position must be coercible to integers and non negative", {
  wb <- createWorkbook()
  expect_error(setWindowSize(wb, xWindow = -300))
  expect_error(setWindowSize(wb, yWindow = -300))
  expect_error(setWindowSize(wb, windowWidth = -300))
  expect_error(setWindowSize(wb, windowHeight = -300))
  expect_error(expect_warning(setWindowSize(wb, xWindow = "string")))
  expect_error(expect_warning(setWindowSize(wb, yWindow = "string")))
  expect_error(expect_warning(setWindowSize(wb, windowWidth = "string")))
  expect_error(expect_warning(setWindowSize(wb, windowHeight = "string")))
  
  setWindowSize(wb, xWindow = "500")
  expect_equal(getWindowSize(wb)$xWindow, "500")
})

test_that("saving and loading preserve window sizes", {
  wb <- createWorkbook()
  setWindowSize(wb, xWindow = 400, yWindow = 500, windowWidth = 1000, windowHeight = 5000)
  addWorksheet(wb, "sheet1")
  tempfile <- tempfile()
  saveWorkbook(wb, tempfile)
  wb2 <- loadWorkbook(tempfile)
  
  expect_equal(getWindowSize(wb2),
               list(xWindow = "400", yWindow = "500", windowWidth = "1000", windowHeight = "5000"))
  
  setWindowSize(wb2, xWindow = 700, yWindow = 900, windowWidth = 900, windowHeight = 4000)
  
  tempfile2 <- tempfile()
  saveWorkbook(wb2, tempfile2)
  wb3 <- loadWorkbook(tempfile2)
  
  expect_equal(getWindowSize(wb3),
               list(xWindow = "700", yWindow = "900", windowWidth = "900", windowHeight = "4000"))
})

context("Style Parsing")



test_that("parsing border xml", {
  wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))
  styles <- getStyles(wb = wb)


  expected_borders <- list(
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, "medium",
    "medium", "medium", "medium", NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, "thin", NULL, "thin", "thin", NULL,
    "thin", "thin", "thin", "thin", "thin", "thin", "thin", NULL,
    "thin", "thin", "medium", "medium", "medium", "medium", "thin",
    "medium", "medium", "thin", NULL, "medium", "medium", "medium",
    "thin", "thin", "medium", "medium", "thin", "thin", "thick",
    NULL, "thick", "thick", "thick", NULL, NULL, NULL, NULL,
    NULL, "medium", "medium", NULL, "medium", "mediumDashed",
    "mediumDashed", "mediumDashed", NULL, NULL, NULL, NULL, NULL,
    NULL
  )

  expect_equal(expected_borders, sapply(styles, "[[", "borderBottom"))


  expected_borders <- list(
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, "thin", "thin", "thin", NULL,
    NULL, NULL, NULL, "medium", NULL, NULL, NULL, NULL, NULL,
    "thin", NULL, "thin", "thick", NULL, "medium", "thin", "thin",
    "thin", "thin", "thick", "thick", "thin", "thin", "thin",
    "medium", "medium", "thin", "thick", "thick", "medium", "thin",
    "thick", "thick", "medium", "thin", "thin", "medium", "thin",
    "thin", "thin", "medium", "medium", "medium", NULL, NULL,
    NULL, NULL, NULL, NULL, "mediumDashed", "mediumDashed", "mediumDashed",
    NULL, NULL, NULL, NULL, NULL, NULL
  )

  expect_equal(expected_borders, sapply(styles, "[[", "borderTop"))



  expected_borders <- list(
    NULL, NULL, NULL, NULL, NULL, NULL, "medium", NULL, "medium",
    NULL, NULL, NULL, NULL, NULL, NULL, "thin", NULL, NULL, "thin",
    NULL, NULL, "thin", "medium", NULL, NULL, NULL, NULL, "thin",
    "thin", "thin", NULL, "thin", NULL, NULL, NULL, "thin", "medium",
    "thin", "thin", "thin", "thin", "medium", "thin", "thin",
    NULL, "thin", "thick", "thin", "thick", "thick", "thin",
    "thin", "thin", "thin", "thick", NULL, "thin", "thin", "thin",
    "medium", NULL, NULL, "medium", NULL, "medium", NULL, "medium",
    NULL, "mediumDashed", NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL
  )

  expect_equal(expected_borders, sapply(styles, "[[", "borderLeft"))


  expected_borders <- list(
    NULL, NULL, NULL, NULL, NULL, "medium", NULL, "medium",
    NULL, NULL, NULL, "medium", NULL, NULL, NULL, NULL, NULL,
    "thin", NULL, NULL, "thin", NULL, NULL, NULL, "thin", NULL,
    "thin", NULL, "thin", "thin", "thin", "thin", "thick", NULL,
    "thick", "medium", "thin", "thin", "thin", "thin", "medium",
    "thin", "thin", "medium", NULL, "medium", "thin", "thin",
    "medium", "medium", "thin", "thick", "medium", "medium",
    "thin", "medium", "thin", NULL, "thick", NULL, NULL, "medium",
    NULL, "medium", NULL, NULL, NULL, "medium", NULL, NULL, "mediumDashed",
    NULL, NULL, NULL, NULL, NULL, NULL
  )

  expect_equal(expected_borders, sapply(styles, "[[", "borderRight"))



  ## COLOURS
  expected_borders <- list(
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(indexed = "64"), .Names = "indexed"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, structure(list(theme = "6"), .Names = "theme"),
    NULL, structure(list(theme = "6"), .Names = "theme"), structure(list(
      theme = "6"
    ), .Names = "theme"), NULL, structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), NULL, structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      theme = "3"
    ), .Names = "theme"), structure(list(theme = "3"), .Names = "theme"),
    structure(list(theme = "3"), .Names = "theme"), structure(list(
      theme = "6"
    ), .Names = "theme"), structure(list(indexed = "64"), .Names = "indexed"),
    structure(list(theme = "6"), .Names = "theme"), structure(list(
      theme = "6"
    ), .Names = "theme"), structure(list(indexed = "64"), .Names = "indexed"),
    NULL, structure(list(theme = "6"), .Names = "theme"), structure(list(
      theme = "7\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    structure(list(theme = "7\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      theme = "9\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    structure(list(theme = "9\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      theme = "5\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    NULL, structure(list(theme = "5\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(theme = "5\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(theme = "5\" tint=\"-0.249977111117893"), .Names = "theme"),
    NULL, NULL, NULL, NULL, NULL, structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    NULL, structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    NULL, NULL, NULL, NULL, NULL, NULL
  )

  expect_equal(expected_borders, sapply(styles, "[[", "borderBottomColour"))


  expected_borders <- list(
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, structure(list(theme = "6"), .Names = "theme"),
    structure(list(theme = "6"), .Names = "theme"), structure(list(
      theme = "6"
    ), .Names = "theme"), NULL, NULL, NULL, NULL,
    structure(list(indexed = "64"), .Names = "indexed"), NULL,
    NULL, NULL, NULL, NULL, structure(list(indexed = "64"), .Names = "indexed"),
    NULL, structure(list(indexed = "64"), .Names = "indexed"),
    structure(list(theme = "5\" tint=\"-0.249977111117893"), .Names = "theme"),
    NULL, structure(list(indexed = "64"), .Names = "indexed"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      theme = "5\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    structure(list(theme = "5\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      theme = "6"
    ), .Names = "theme"), structure(list(indexed = "64"), .Names = "indexed"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      theme = "5\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    structure(list(theme = "5\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(theme = "7\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      theme = "5\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    structure(list(theme = "5\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(theme = "9\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    NULL, NULL, NULL, NULL, NULL, NULL, structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    NULL, NULL, NULL, NULL, NULL, NULL
  )

  expect_equal(expected_borders, sapply(styles, "[[", "borderTopColour"))



  expected_borders <- list(
    NULL, NULL, NULL, NULL, NULL, NULL, structure(list(indexed = "64"), .Names = "indexed"),
    NULL, structure(list(indexed = "64"), .Names = "indexed"),
    NULL, NULL, NULL, NULL, NULL, NULL, structure(list(theme = "6"), .Names = "theme"),
    NULL, NULL, structure(list(theme = "6"), .Names = "theme"),
    NULL, NULL, structure(list(theme = "6"), .Names = "theme"),
    structure(list(indexed = "64"), .Names = "indexed"), NULL,
    NULL, NULL, NULL, structure(list(indexed = "64"), .Names = "indexed"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), NULL, structure(list(
      indexed = "64"
    ), .Names = "indexed"), NULL, NULL, NULL,
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      theme = "3"
    ), .Names = "theme"), structure(list(indexed = "64"), .Names = "indexed"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      theme = "6"
    ), .Names = "theme"), structure(list(indexed = "64"), .Names = "indexed"),
    structure(list(indexed = "64"), .Names = "indexed"), NULL,
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      theme = "5\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      theme = "5\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    structure(list(theme = "5\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      theme = "5\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    NULL, structure(list(indexed = "64"), .Names = "indexed"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    NULL, NULL, structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    NULL, structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    NULL, structure(list(indexed = "64"), .Names = "indexed"),
    NULL, structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
  )

  expect_equal(expected_borders, sapply(styles, "[[", "borderLeftColour"))



  expected_borders <- list(
    NULL, NULL, NULL, NULL, NULL, structure(list(indexed = "64"), .Names = "indexed"),
    NULL, structure(list(indexed = "64"), .Names = "indexed"),
    NULL, NULL, NULL, structure(list(indexed = "64"), .Names = "indexed"),
    NULL, NULL, NULL, NULL, NULL, structure(list(theme = "6"), .Names = "theme"),
    NULL, NULL, structure(list(theme = "6"), .Names = "theme"),
    NULL, NULL, NULL, structure(list(theme = "6"), .Names = "theme"),
    NULL, structure(list(indexed = "64"), .Names = "indexed"),
    NULL, structure(list(indexed = "64"), .Names = "indexed"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      theme = "5\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    NULL, structure(list(theme = "5\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(theme = "3"), .Names = "theme"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      theme = "6"
    ), .Names = "theme"), structure(list(indexed = "64"), .Names = "indexed"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      theme = "6"
    ), .Names = "theme"), NULL, structure(list(
      theme = "6"
    ), .Names = "theme"), structure(list(indexed = "64"), .Names = "indexed"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      theme = "7\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    structure(list(theme = "7\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      theme = "5\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    structure(list(theme = "9\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(theme = "9\" tint=\"-0.249977111117893"), .Names = "theme"),
    structure(list(indexed = "64"), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), structure(list(
      indexed = "64"
    ), .Names = "indexed"), NULL, structure(list(
      theme = "5\" tint=\"-0.249977111117893"
    ), .Names = "theme"),
    NULL, NULL, structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    NULL, structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    NULL, NULL, NULL, structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    NULL, NULL, structure("9\" tint=\"-0.249977111117893", .Names = "theme"),
    NULL, NULL, NULL, NULL, NULL, NULL
  )

  expect_equal(expected_borders, sapply(styles, "[[", "borderRightColour"))
})

context("clone Worksheet")


test_that("clone Worksheet with data", {
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  writeData(wb, "Sheet 1", 1)
  cloneWorksheet(wb, "Sheet 2", clonedSheet = "Sheet 1")


  file_name <- system.file("extdata", "cloneWorksheetExample.xlsx", package = "openxlsx")
  refwb <- loadWorkbook(file = file_name)

  expect_equal(sheets(wb), sheets(refwb))
  expect_equal(worksheetOrder(wb), worksheetOrder(refwb))
})

test_that("clone empty Worksheet", {
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")

  cloneWorksheet(wb, "Sheet 2", clonedSheet = "Sheet 1")


  file_name <- system.file("extdata", "cloneEmptyWorksheetExample.xlsx", package = "openxlsx")
  refwb <- loadWorkbook(file = file_name)

  expect_equal(sheets(wb), sheets(refwb))
  expect_equal(worksheetOrder(wb), worksheetOrder(refwb))
})

context("Load Workbook Object Tables")


test_that("Tables loaded correctly", {
  wb <- loadWorkbook(system.file("extdata", "loadExample.xlsx", package = "openxlsx"))

  expect_equal(unname(attr(wb$tables, "tableName")), c("Table2", "Table3"))
  expect_equal(names(attr(wb$tables, "tableName")), c("A1:E51", "A1:K30"))
  expect_equal(attr(wb$tables, "sheet"), c(1, 3))

  expect_equal(wb$worksheets[[1]]$tableParts, "<tablePart r:id=\"rId3\"/>", check.attributes = FALSE)
  expect_equal(unname(attr(wb$worksheets[[1]]$tableParts, "tableName")), "Table2")
  expect_equal(names(attr(wb$worksheets[[1]]$tableParts, "tableName")), "A1:E51")

  expect_equal(wb$worksheets[[3]]$tableParts, "<tablePart r:id=\"rId4\"/>", check.attributes = FALSE)
  expect_equal(unname(attr(wb$worksheets[[3]]$tableParts, "tableName")), "Table3")
  expect_equal(names(attr(wb$worksheets[[3]]$tableParts, "tableName")), "A1:K30")


  ## now remove a table
  expect_equal(unname(getTables(wb, 1)), "Table2", check.attributes = FALSE)
  expect_equal(unname(getTables(wb, 3)), "Table3", check.attributes = FALSE)

  removeTable(wb, sheet = 1, table = "Table2")

  expect_equal(getTables(wb, sheet = 1), character(0), check.attributes = FALSE)
  expect_equal(length(wb$worksheets[[1]]$tableParts), 0)
  expect_equal(wb$worksheets[[1]]$tableParts, character(0), check.attributes = FALSE)

  expect_equal(wb$worksheets[[3]]$tableParts, "<tablePart r:id=\"rId4\"/>", check.attributes = FALSE)
  expect_equal(unname(attr(wb$worksheets[[3]]$tableParts, "tableName")), "Table3")
  expect_equal(names(attr(wb$worksheets[[3]]$tableParts, "tableName")), "A1:K30")


  expect_error(removeTable(wb, sheet = 1, table = "Table2"), regexp = "table 'Table2' does not exist")
})

context("Workbook groupings")

# groupRows() requires assigning to wb to global environment
# For reference, see here: https://github.com/r-lib/testthat/issues/720

test_that("group columns", {

  # Grouping then setting widths updates hidden
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  groupColumns(wb, "Sheet 1", 2:3, hidden = TRUE)
  setColWidths(wb, "Sheet 1", 2, widths = "18", hidden = FALSE)

  expect_equal(attr(wb$colOutlineLevels[[1]], "hidden")[attr(wb$colOutlineLevels[[1]], "names") == 2], "0")

  # Setting column widths then grouping
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  setColWidths(wb, "Sheet 1", 2:3, widths = "18", hidden = FALSE)
  groupColumns(wb, "Sheet 1", 1:2, hidden = TRUE)

  expect_equal(attr(wb$colWidths[[1]], "hidden")[attr(wb$colWidths[[1]], "names") == 2], "1")
})


test_that("group rows", {

  wb <- createWorkbook()
  assign("wb", wb, envir = .GlobalEnv)
  addWorksheet(wb, "Sheet 1")
  groupRows(wb, "Sheet 1", 1:4, hidden = TRUE)

  expect_equal(names(wb$outlineLevels[[1]]), c("1", "2", "3", "4"))
  expect_equal(unique(attr(wb$outlineLevels[[1]], "hidden")), "1")
  rm(wb)
})

test_that("group rows 2", {
 
  wb <- createWorkbook()
  addWorksheet(wb, 'Sheet1')
  addWorksheet(wb, 'Sheet2')
  writeData(wb, "Sheet1", iris) 
  writeData(wb, "Sheet2", iris)

  ## create list of groups
  # lines used for grouping (here: species)
  grp <- list(
    seq(2, 51),
    seq(52, 101),
    seq(102, 151)
  )
  # assign group levels
  names(grp) <- c("1","0","1") 
  groupRows(wb, "Sheet1", rows = grp)

  # different grouping
  names(grp) <- c("1","2","3")
  groupRows(wb, "Sheet2", rows = grp)

  expect_equal(unique(wb$outlineLevels[[1]]), c("1", "0"))
  expect_equal(unique(wb$outlineLevels[[2]]), c("1", "2", "3"))
})



test_that("ungroup columns", {

  # OutlineLevelCol is removed from SheetFormatPr when no
  # column groupings left
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  setColWidths(wb, "Sheet 1", 2:3, widths = "18", hidden = FALSE)
  groupColumns(wb, "Sheet 1", 1:3, hidden = TRUE)
  ungroupColumns(wb, "Sheet 1", 1:3)

  expect_equal(unique(attr(wb$colWidths[[1]], "hidden")[attr(wb$colWidths[[1]], "names") %in% c(2, 3)]), "0")
})


test_that("ungroup rows", {
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  groupRows(wb, "Sheet 1", 1:3, hidden = TRUE)
  ungroupRows(wb, "Sheet 1", 1:3)

  expect_equal(length(wb$outlineLevels[[1]]), 0L)
})

test_that("no warnings #485", {

  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")

  expect_silent(
    groupRows(
      rows = c(1),
      sheet = "Sheet 1",
      wb = wb
    )
  )

  expect_silent(
    ungroupRows(
      rows = c(1),
      sheet = "Sheet 1",
      wb = wb
    )
  )

})


test_that("loading workbook preserves outlines", {
  fl <- system.file("extdata", "groupTest.xlsx", package = "openxlsx")
  wb <- loadWorkbook(fl)

  expect_equal(names(wb$colOutlineLevels[[1]]), c("2", "3", "4"))
  expect_equal(names(wb$outlineLevels[[1]]), c("3", "4"))
  expect_equal(unique(attr(wb$outlineLevels[[1]], "hidden")[names(wb$outlineLevels[[1]]) %in% c("3", "4")]), "true")

  wbb <- createWorkbook()

  addWorksheet(wbb, sheetName = "Test", gridLines = FALSE, tabColour = "deepskyblue")

  writeData(wbb, sheet = "Test", x = c(colA = "testcol1", colB = "testcol2"))

  groupColumns(wbb, "Test", cols = 2:3, hidden = FALSE)
  setColWidths(wbb, sheet = "Test", cols=c(1:5), widths = c(9,9,9,9,9))
  groupColumns(wbb, "Test", cols = 5:10, hidden = FALSE)
  setColWidths(wbb, "Test", cols = 15:20, widths = 9)

  tf <- temp_xlsx("test")
  tf2 <- temp_xlsx("test2")

  saveWorkbook(wbb, tf, overwrite = TRUE)
  test <- wbb$worksheets[[1]]$copy()

  wb <- loadWorkbook(tf)
  saveWorkbook(wb, tf2, overwrite = TRUE)

  testthat::expect_equal(wb$worksheets[[1]], test)

  unlink(c("tf", "tf2"), recursive = TRUE, force = TRUE)
})


test_that("Grouping after setting colwidths has correct length of hidden attributes", {
  # Issue #100 - https://github.com/ycphs/openxlsx/issues/100

  wb <- createWorkbook(title = "column width and grouping error")
  addWorksheet(wb, sheetName = 1)

  setColWidths(
    wb,
    sheet = 1,
    cols = 1:100,
    widths = 8
  )

  groupColumns(wb, sheet = 1, cols = 20:100, hidden = TRUE)

  expect_equal(length(wb$colOutlineLevels[[1]]), length(attr(wb$colOutlineLevels[[1]], "hidden")))
})

test_that("Consecutive calls to saveWorkbook doesn't corrupt attributes", {

  wbb <- createWorkbook()

  addWorksheet(wbb, sheetName = "Test", gridLines = FALSE, tabColour = "deepskyblue")

  writeData(wbb, sheet = "Test", x = c(colA = "testcol1", colB = "testcol2"))

  groupColumns(wbb, "Test", cols = 2:3, hidden = FALSE)
  setColWidths(wbb, sheet = "Test", cols=c(1:5), widths = c(9,9,9,9,9))
  groupColumns(wbb, "Test", cols = 5:10, hidden = FALSE)
  setColWidths(wbb, "Test", cols = 15:20, widths = 9)

  tf <- temp_xlsx("test")
  tf2 <- temp_xlsx("test2")

  saveWorkbook(wbb, tf, overwrite = TRUE)
  test <- wbb$worksheets[[1]]$copy()

  saveWorkbook(wbb, tf2, overwrite = TRUE)

  testthat::expect_equal(wbb$worksheets[[1]], test)

  unlink(c("tf", "tf2"), recursive = TRUE, force = TRUE)
})

context("Freeze Panes")

test_that("Freeze Panes", {
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  freezePane(wb, 1, firstActiveRow = 3, firstActiveCol = 3)

  expected <- "<pane ySplit=\"2\" xSplit=\"2\" topLeftCell=\"C3\" activePane=\"bottomRight\" state=\"frozen\"/><selection pane=\"bottomRight\"/>"
  expect_equal(wb$worksheets[[1]]$freezePane, expected)




  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  freezePane(wb, 1, firstActiveRow = 1, firstActiveCol = 3)

  expected <- "<pane xSplit=\"2\" topLeftCell=\"C1\" activePane=\"topRight\" state=\"frozen\"/><selection pane=\"topRight\"/>"
  expect_equal(wb$worksheets[[1]]$freezePane, expected)




  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  freezePane(wb, 1, firstActiveRow = 2, firstActiveCol = 1)

  expected <- "<pane ySplit=\"1\" topLeftCell=\"A2\" activePane=\"bottomLeft\" state=\"frozen\"/><selection pane=\"bottomLeft\"/>"
  expect_equal(wb$worksheets[[1]]$freezePane, expected)




  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  freezePane(wb, 1, firstActiveRow = 2, firstActiveCol = 4)

  expected <- "<pane ySplit=\"1\" xSplit=\"3\" topLeftCell=\"D2\" activePane=\"bottomRight\" state=\"frozen\"/><selection pane=\"bottomRight\"/>"
  expect_equal(wb$worksheets[[1]]$freezePane, expected)




  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  freezePane(wb, 1, firstCol = TRUE)

  expected <- "<pane xSplit=\"1\" topLeftCell=\"B1\" activePane=\"topRight\" state=\"frozen\"/>"
  expect_equal(wb$worksheets[[1]]$freezePane, expected)




  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  freezePane(wb, 1, firstRow = TRUE)

  expected <- "<pane ySplit=\"1\" topLeftCell=\"A2\" activePane=\"bottomLeft\" state=\"frozen\"/>"
  expect_equal(wb$worksheets[[1]]$freezePane, expected)




  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  freezePane(wb, 1, firstRow = TRUE, firstCol = TRUE)

  expected <- "<pane ySplit=\"1\" xSplit=\"1\" topLeftCell=\"B2\" activePane=\"bottomRight\" state=\"frozen\"/><selection pane=\"bottomRight\"/>"
  expect_equal(wb$worksheets[[1]]$freezePane, expected)


  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  addWorksheet(wb, "Sheet 2")
  addWorksheet(wb, "Sheet 3")
  addWorksheet(wb, "Sheet 4")
  addWorksheet(wb, "Sheet 5")
  addWorksheet(wb, "Sheet 6")
  addWorksheet(wb, "Sheet 7")

  freezePane(wb, sheet = 1, firstActiveRow = 3, firstActiveCol = 3)
  freezePane(wb, sheet = 2, firstActiveRow = 1, firstActiveCol = 3)
  freezePane(wb, sheet = 3, firstActiveRow = 2, firstActiveCol = 1)
  freezePane(wb, sheet = 4, firstActiveRow = 2, firstActiveCol = 4)
  freezePane(wb, sheet = 5, firstCol = TRUE)
  freezePane(wb, sheet = 6, firstRow = TRUE)
  freezePane(wb, sheet = 7, firstRow = TRUE, firstCol = TRUE)


  expected <- "<pane ySplit=\"2\" xSplit=\"2\" topLeftCell=\"C3\" activePane=\"bottomRight\" state=\"frozen\"/><selection pane=\"bottomRight\"/>"
  expect_equal(wb$worksheets[[1]]$freezePane, expected)

  expected <- "<pane xSplit=\"2\" topLeftCell=\"C1\" activePane=\"topRight\" state=\"frozen\"/><selection pane=\"topRight\"/>"
  expect_equal(wb$worksheets[[2]]$freezePane, expected)

  expected <- "<pane ySplit=\"1\" topLeftCell=\"A2\" activePane=\"bottomLeft\" state=\"frozen\"/><selection pane=\"bottomLeft\"/>"
  expect_equal(wb$worksheets[[3]]$freezePane, expected)

  expected <- "<pane ySplit=\"1\" xSplit=\"3\" topLeftCell=\"D2\" activePane=\"bottomRight\" state=\"frozen\"/><selection pane=\"bottomRight\"/>"
  expect_equal(wb$worksheets[[4]]$freezePane, expected)

  expected <- "<pane xSplit=\"1\" topLeftCell=\"B1\" activePane=\"topRight\" state=\"frozen\"/>"
  expect_equal(wb$worksheets[[5]]$freezePane, expected)

  expected <- "<pane ySplit=\"1\" topLeftCell=\"A2\" activePane=\"bottomLeft\" state=\"frozen\"/>"
  expect_equal(wb$worksheets[[6]]$freezePane, expected)

  expected <- "<pane ySplit=\"1\" xSplit=\"1\" topLeftCell=\"B2\" activePane=\"bottomRight\" state=\"frozen\"/><selection pane=\"bottomRight\"/>"
  expect_equal(wb$worksheets[[7]]$freezePane, expected)
})

context("Writing NA to sheet_data")



test_that("Writing to sheet_data with keepNA = FALSE", {
  a <- head(iris)
  a[2, 2] <- NA
  a[3, 5] <- NA
  a[5, 1] <- NA
  a[5, 5] <- NA

  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  writeData(wb, 1, a, keepNA = FALSE)

  sheet_data <- wb$worksheets[[1]]$sheet_data

  sheet_v <- sheet_data$v
  sheet_t <- sheet_data$t
  sheet_f <- sheet_data$f
  sheet_row <- sheet_data$rows
  sheet_col <- sheet_data$cols

  sheet_v <- as.numeric(sheet_v)

  sheet_v <- sheet_v[!is.na(sheet_v)]
  sheet_t <- sheet_t[!is.na(sheet_t)]
  sheet_f <- sheet_f[!is.na(sheet_f)]
  sheet_row <- sheet_row[!is.na(sheet_row)]
  sheet_col <- sheet_col[!is.na(sheet_col)]

  n_values <- prod(dim(a)) + ncol(a)

  expect_length(sheet_row, n_values)
  expect_length(sheet_col, n_values)
  expect_length(sheet_t, n_values - 4)
  expect_length(sheet_v, n_values - 4)
  expect_length(sheet_f, 0)

  ## rows/cols
  expect_equal(sheet_row, rep(1:7, each = 5))
  expect_equal(sheet_col, rep(1:5, times = 7))

  ## header types
  expect_equal(sheet_t[1:5], rep(1, ncol(a)))

  ## data.frame t & v
  expected_t <- c(
    "n", "n", "n", "n", "s", "n", "n", "n", "s", "n", "n", "n",
    "n", "n", "n", "n", "n", "s", "n", "n", "n", "n", "n", "n", "n",
    "s", NA, NA, NA, NA
  )

  expected_t <- map_cell_types_to_integer(t = expected_t)
  expect_equal(sheet_t[6:n_values], expected_t)

  expect_equal(sheet_v[1:5], 0:4)

  expected_v <- c(
    5.1, 3.5, 1.4, 0.2, 5, 4.9, 1.4, 0.2, 5, 4.7, 3.2, 1.3, 0.2,
    4.6, 3.1, 1.5, 0.2, 5, 3.6, 1.4, 0.2, 5.4, 3.9, 1.7, 0.4, 5,
    NA, NA, NA, NA
  )

  expect_equal(sheet_v[6:n_values], expected_v)
})

test_that("Writing to sheet_data with keepNA = TRUE and na.string = '*'", {
  a <- head(iris)
  a[2, 2] <- NA
  a[3, 5] <- NA
  a[5, 1] <- NA
  a[5, 5] <- NA

  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  writeData(wb, 1, a, keepNA = TRUE, na.string = "*")

  sheet_data <- wb$worksheets[[1]]$sheet_data

  sheet_v <- sheet_data$v
  sheet_t <- sheet_data$t
  sheet_f <- sheet_data$f
  sheet_row <- sheet_data$rows
  sheet_col <- sheet_data$cols

  sheet_v <- as.numeric(sheet_v)

  sheet_v <- sheet_v[!is.na(sheet_v)]
  sheet_t <- sheet_t[!is.na(sheet_t)]
  sheet_f <- sheet_f[!is.na(sheet_f)]
  sheet_row <- sheet_row[!is.na(sheet_row)]
  sheet_col <- sheet_col[!is.na(sheet_col)]

  n_values <- prod(dim(a)) + ncol(a)

  expect_length(sheet_row, n_values)
  expect_length(sheet_col, n_values)
  expect_length(sheet_t, n_values)
  expect_length(sheet_v, n_values)
  expect_length(sheet_f, 0)

  ## rows/cols
  expect_equal(sheet_row, rep(1:7, each = 5))
  expect_equal(sheet_col, rep(1:5, times = 7))

  ## header types
  expect_equal(sheet_t[1:5], rep(1, ncol(a)))

  ## data.frame t & v
  expected_t <- c(
    "n", "n", "n", "n", "s",
    "n", "s", "n", "n", "s",
    "n", "n", "n", "n", "s",
    "n", "n", "n", "n", "s",
    "s", "n", "n", "n", "s",
    "n", "n", "n", "n", "s"
  )

  expected_t <- map_cell_types_to_integer(t = expected_t)
  expect_equal(sheet_t[6:n_values], expected_t)

  expect_equal(sheet_v[1:5], 0:4)

  expected_v <- c(
    5.1, 3.5, 1.4, 0.2, 5,
    4.9, 6, 1.4, 0.2, 5,
    4.7, 3.2, 1.3, 0.2, 6,
    4.6, 3.1, 1.5, 0.2, 5,
    6, 3.6, 1.4, 0.2, 6,
    5.4, 3.9, 1.7, 0.4, 5
  )

  expect_equal(sheet_v[6:n_values], expected_v)
})

context("Re-ordering worksheets.")



test_that("Worksheet ordering from new Workbook", {
  genWS <- function(wb, sheetName) {
    addWorksheet(wb, sheetName)
    writeDataTable(wb, sheetName, data.frame("X" = sprintf("This is sheet: %s", sheetName)), colNames = FALSE)
  }

  wb <- createWorkbook()
  genWS(wb, "Sheet 1")
  genWS(wb, "Sheet 2")
  genWS(wb, "Sheet 3")


  tempFile <- temp_xlsx("orderingTest")

  ## no ordering
  saveWorkbook(wb, file = tempFile, overwrite = TRUE)
  expect_equal(names(wb), sprintf("Sheet %s", 1:3))

  wb <- loadWorkbook(tempFile)
  expect_equal(names(wb), sprintf("Sheet %s", 1:3))


  ## re-order doesnt do anything
  worksheetOrder(wb) <- c(3, 2, 1)
  expect_equal(names(wb), sprintf("Sheet %s", 1:3))

  saveWorkbook(wb, file = tempFile, overwrite = TRUE)
  expect_equal(names(wb), sprintf("Sheet %s", 1:3))



  ## reloading - reordered
  wb <- loadWorkbook(file = tempFile)
  expect_equal(names(wb), sprintf("Sheet %s", 3:1))

  x <- read.xlsx(tempFile, sheet = 1)[[1]]
  expect_equal(x, "This is sheet: Sheet 3")

  x <- read.xlsx(tempFile, sheet = 2)[[1]]
  expect_equal(x, "This is sheet: Sheet 2")

  x <- read.xlsx(tempFile, sheet = 3)[[1]]
  expect_equal(x, "This is sheet: Sheet 1")


  ## reloading  - reordered - reading from the workbook object
  x <- read.xlsx(wb, sheet = 1)[[1]]
  expect_equal(x, "This is sheet: Sheet 3")

  x <- read.xlsx(wb, sheet = 2)[[1]]
  expect_equal(x, "This is sheet: Sheet 2")

  x <- read.xlsx(wb, sheet = 3)[[1]]
  expect_equal(x, "This is sheet: Sheet 1")



  ## save and re-load again
  saveWorkbook(wb, tempFile, overwrite = TRUE)
  wb <- loadWorkbook(tempFile)
  expect_equal(names(wb), sprintf("Sheet %s", 3:1))

  x <- read.xlsx(wb, sheet = 1)[[1]]
  expect_equal(x, "This is sheet: Sheet 3")

  x <- read.xlsx(wb, sheet = 2)[[1]]
  expect_equal(x, "This is sheet: Sheet 2")

  x <- read.xlsx(wb, sheet = 3)[[1]]
  expect_equal(x, "This is sheet: Sheet 1")

  x <- read.xlsx(wb, sheet = 1)[[1]]
  expect_equal(x, "This is sheet: Sheet 3")

  x <- read.xlsx(wb, sheet = 2)[[1]]
  expect_equal(x, "This is sheet: Sheet 2")

  x <- read.xlsx(wb, sheet = 3)[[1]]
  expect_equal(x, "This is sheet: Sheet 1")




  ###### re-order again
  worksheetOrder(wb) <- c(2, 3, 1)
  saveWorkbook(wb, tempFile, overwrite = TRUE)

  x <- read.xlsx(tempFile, sheet = 1)[[1]]
  expect_equal(x, "This is sheet: Sheet 2")

  x <- read.xlsx(tempFile, sheet = 2)[[1]]
  expect_equal(x, "This is sheet: Sheet 1")

  x <- read.xlsx(tempFile, sheet = 3)[[1]]
  expect_equal(x, "This is sheet: Sheet 3")


  wb <- loadWorkbook(tempFile)
  expect_equal(names(wb), sprintf("Sheet %s", c(2, 1, 3)))

  x <- read.xlsx(wb, sheet = 1)[[1]]
  expect_equal(x, "This is sheet: Sheet 2")

  x <- read.xlsx(wb, sheet = 2)[[1]]
  expect_equal(x, "This is sheet: Sheet 1")

  x <- read.xlsx(wb, sheet = 3)[[1]]
  expect_equal(x, "This is sheet: Sheet 3")




  ## add a worksheet
  genWS(wb, sheetName = "Sheet 4")

  x <- read.xlsx(wb, sheet = 4)[[1]]
  expect_equal(x, "This is sheet: Sheet 4")

  ## re-order and add worksheet then save
  worksheetOrder(wb) <- c(3, 1, 4, 2)
  names(wb)

  saveWorkbook(wb, tempFile, overwrite = TRUE)

  ## read from file
  x <- read.xlsx(tempFile, sheet = 1)[[1]]
  expect_equal(x, "This is sheet: Sheet 3")

  x <- read.xlsx(tempFile, sheet = 2)[[1]]
  expect_equal(x, "This is sheet: Sheet 2")

  x <- read.xlsx(tempFile, sheet = 3)[[1]]
  expect_equal(x, "This is sheet: Sheet 4")

  x <- read.xlsx(tempFile, sheet = 4)[[1]]
  expect_equal(x, "This is sheet: Sheet 1")

  x <- read.xlsx(tempFile, sheet = "Sheet 3")[[1]]
  expect_equal(x, "This is sheet: Sheet 3")

  x <- read.xlsx(tempFile, sheet = "Sheet 2")[[1]]
  expect_equal(x, "This is sheet: Sheet 2")

  x <- read.xlsx(tempFile, sheet = "Sheet 4")[[1]]
  expect_equal(x, "This is sheet: Sheet 4")

  x <- read.xlsx(tempFile, sheet = "Sheet 1")[[1]]
  expect_equal(x, "This is sheet: Sheet 1")







  ## read from workbook
  wb <- loadWorkbook(tempFile)
  x <- read.xlsx(wb, sheet = 1)[[1]]
  expect_equal(x, "This is sheet: Sheet 3")

  x <- read.xlsx(wb, sheet = 2)[[1]]
  expect_equal(x, "This is sheet: Sheet 2")

  x <- read.xlsx(wb, sheet = 3)[[1]]
  expect_equal(x, "This is sheet: Sheet 4")

  x <- read.xlsx(wb, sheet = 4)[[1]]
  expect_equal(x, "This is sheet: Sheet 1")




  ## read from workbook using name
  wb <- loadWorkbook(tempFile)
  x <- read.xlsx(wb, sheet = "Sheet 3")[[1]]
  expect_equal(x, "This is sheet: Sheet 3")

  x <- read.xlsx(wb, sheet = "Sheet 2")[[1]]
  expect_equal(x, "This is sheet: Sheet 2")

  x <- read.xlsx(wb, sheet = "Sheet 1")[[1]]
  expect_equal(x, "This is sheet: Sheet 1")

  x <- read.xlsx(wb, sheet = "Sheet 4")[[1]]
  expect_equal(x, "This is sheet: Sheet 4")


  writeData(wb, sheet = "Sheet 3", iris[1:10, 1:4], startRow = 5)
  x <- read.xlsx(wb, sheet = "Sheet 3", startRow = 5, colNames = TRUE)
  expect_equal(x, iris[1:10, 1:4])


  writeData(wb, sheet = 4, iris[1:20, 1:4], startRow = 5)
  x <- read.xlsx(wb, sheet = 4, startRow = 5, colNames = TRUE)
  expect_equal(x, iris[1:20, 1:4])


  writeData(wb, sheet = 2, iris[1:30, 1:4], startRow = 5)
  x <- read.xlsx(wb, sheet = 2, startRow = 5, colNames = TRUE)
  expect_equal(x, iris[1:30, 1:4])


  ## reading from saved file
  saveWorkbook(wb, tempFile, TRUE)

  x <- read.xlsx(tempFile, sheet = "Sheet 3", startRow = 5, colNames = TRUE)
  expect_equal(x, iris[1:10, 1:4])

  x <- read.xlsx(tempFile, sheet = 4, startRow = 5, colNames = TRUE)
  expect_equal(x, iris[1:20, 1:4])

  x <- read.xlsx(tempFile, sheet = 2, startRow = 5, colNames = TRUE)
  expect_equal(x, iris[1:30, 1:4])


  ## And finally load again
  wb <- loadWorkbook(tempFile)

  x <- read.xlsx(wb, sheet = "Sheet 3", startRow = 5, colNames = TRUE)
  expect_equal(x, iris[1:10, 1:4])

  x <- read.xlsx(wb, sheet = 4, startRow = 5, colNames = TRUE)
  expect_equal(x, iris[1:20, 1:4])

  x <- read.xlsx(wb, sheet = 2, startRow = 5, colNames = TRUE)
  expect_equal(x, iris[1:30, 1:4])


  unlink(tempFile, recursive = TRUE, force = TRUE)
  rm(wb)
})



test_that("Worksheet ordering from new Workbook", {
  tempFile <- temp_xlsx()

  wb <- createWorkbook()
  addWorksheet(wb = wb, sheetName = "Sheet 1", gridLines = FALSE)
  writeDataTable(wb = wb, sheet = 1, x = iris)

  addWorksheet(wb = wb, sheetName = "mtcars (Sheet 2)", gridLines = FALSE)
  writeData(wb = wb, sheet = 2, x = mtcars)

  addWorksheet(wb = wb, sheetName = "Sheet 3", gridLines = FALSE)
  writeData(wb = wb, sheet = 3, x = Formaldehyde)

  worksheetOrder(wb)
  names(wb)
  worksheetOrder(wb) <- c(1, 3, 2) # switch position of sheets 2 & 3

  names(wb)
  writeData(wb, 2, 'This is still the "mtcars" worksheet', startCol = 15)

  names(wb)
  writeData(wb, "Sheet 3", "writing to sheet 3", startCol = 15)

  worksheetOrder(wb)
  names(wb) ## ordering within workbook is not changed

  saveWorkbook(wb, tempFile, overwrite = TRUE)

  worksheetOrder(wb) <- c(3, 2, 1)
  saveWorkbook(wb, tempFile, overwrite = TRUE)


  wb <- loadWorkbook(tempFile)
  worksheetOrder(wb) <- c(3, 2, 1)


  unlink(tempFile, recursive = TRUE, force = TRUE)
  rm(wb)
})

test_that("getBaseFont works", {
  wb <- createWorkbook()
  expect_equal(
    getBaseFont(wb),
    list(
      size = list(val = "11"),
      # should this be "#000000"?
      colour = list(rgb = "FF000000"),
      name = list(val = "Calibri")
    )
  )
  
  modifyBaseFont(wb, fontSize = 9, fontName = "Arial", fontColour = "red")
  expect_equal(
    getBaseFont(wb), 
    list(
      size = list(val = "9"),
      colour = list(rgb = "FFFF0000"),
      name = list(val = "Arial")
    )
  )
})

test_that("writeData() forces evaluation of x (#264)", {
  wbfile <- temp_xlsx()
  op <- options(stringsAsFactors = FALSE)
  
  x <- format(123.4)
  df <- data.frame(d = format(123.4))
  df2 <- data.frame(e = x)
  
  wb <- createWorkbook()
  addWorksheet(wb, "sheet")
  writeData(wb, "sheet", startCol = 1, data.frame(a = format(123.4)))
  writeData(wb, "sheet", startCol = 2, data.frame(b = as.character(123.4)))
  writeData(wb, "sheet", startCol = 3, data.frame(c = "123.4"))
  writeData(wb, "sheet", startCol = 4, df)
  writeData(wb, "sheet", startCol = 5, df2)
  
  saveWorkbook(wb, wbfile)
  out <- read.xlsx(wbfile)
  
  # Possibly overkill
  
  with(out, {
    expect_identical(a, b)
    expect_identical(a, c)
    expect_identical(a, d)
    expect_identical(a, e)
    expect_identical(b, c)
    expect_identical(b, d)
    expect_identical(b, e)
    expect_identical(c, d)
    expect_identical(c, e)
    expect_identical(d, e)
  })
  
  options(op)
  file.remove(wbfile)
})


test_that("as.character.formula() works [312]", {
  form <- y ~ x1 * x2 + x3
  expect_identical(
    as.character.default(form),
    openxlsx:::as.character.formula(form)
  )
  
  skip_if_not_installed("formula.tools")
  # "tests specifically for as.character.formula conflict"
  
  foo <- function() {
    wb <- openxlsx::buildWorkbook(
      data.frame(
        x = structure("A2 + B2", class = c("character", "formula")),
        stringsAsFactors = FALSE
      )
    )
    as.list(wb$worksheets[[1]]$sheet_data)
  }
  
  before <- foo()
  # don't required the "require" function for deps  check
  match.fun("require")("formula.tools", character.only = TRUE)
  middle <- foo()
  detach("package:formula.tools", character.only = TRUE, force = TRUE)
  end <- foo()
  
  expect_identical(before, middle, ignore.environment = TRUE)
  expect_identical(before, end,    ignore.environment = TRUE)
})

# context("Replacing styles")
#
#
#
# test_that("Replace styles", {
#
# #   tempFile <- file.path(tempdir(), "temp.xlsx")
# #   wb <- loadWorkbook(file = file.path(path.package("openxlsx"), "loadExample.xlsx"))
# #
# #   ## create a new style and replace style 2
# #
# #   newStyle <- createStyle(fgFill = "#FF0000")
# #
# #   ## replace style 2
# #   getStyles(wb) ## prints styles
# #   replaceStyle(wb, 87, newStyle = newStyle)
# #
# #   rm(wb)
#
# })

test_that("writeDataTable() warns about Excel limitations", {

  dt <- data.frame(var = c(1))
  header_255 <- paste(sample(tolower(LETTERS), 255, TRUE), collapse = "")
  header_256 <- paste(sample(tolower(LETTERS), 256, TRUE), collapse = "")

  wb <- createWorkbook()
  addWorksheet(wb, "sheetA")
  addWorksheet(wb, "sheetB")

  expect_warning({
    colnames(dt) <- header_255
    writeDataTable(wb, sheet = "sheetA", x = dt)
    },
    regexp = NA
  )

  expect_warning({
    colnames(dt) <- header_256
    writeDataTable(wb, sheet = "sheetB", x = dt)
    },
    regexp = "Column name exceeds 255 chars, possible incompatibility with MS Excel. Index: 1"
  )
})

context("save workbook")


test_that("test return values for saveWorkbook", {
  
  tempFile <- temp_xlsx()
  wb<-createWorkbook()
  addWorksheet(wb,"name")
  expect_true( saveWorkbook(wb,tempFile,returnValue = TRUE))
  
  expect_error( saveWorkbook(wb,tempFile,returnValue = TRUE))
  
  
  expect_invisible( saveWorkbook(wb,tempFile,returnValue = FALSE ,overwrite = TRUE))
  unlink(tempFile, recursive = TRUE, force = TRUE)
  
}
)

# regression test for a typo
test_that("regression test for #248", {
  
  # Basic data frame
  df <- data.frame(number = 1:3, percent = 4:6/100)
  tempFile <- temp_xlsx()
  
  # no formatting
  expect_silent(write.xlsx(df, tempFile, borders = "columns", overwrite = TRUE))
  
  # Change column class to percentage
  class(df$percent) <- "percentage"
  expect_silent(write.xlsx(df, tempFile, borders = "columns", overwrite = TRUE))
})


# test for hyperrefs
test_that("creating hyperlinks", {
  
  # prepare a file
  tempFile <- temp_xlsx()
  wb <- createWorkbook()
  sheet <- "test"
  addWorksheet(wb, sheet)
  img <- "D:/somepath/somepicture.png"
  
  # warning: col and row provided, but not required
  expect_warning(
    linkString <- makeHyperlinkString(col = 1, row = 4,
                                      text = "test.png", file = img))
  
  linkString2 <- makeHyperlinkString(text = "test.png", file = img)
  
  # col and row not needed
  expect_equal(linkString, linkString2)
  
  # write file without errors
  writeFormula(wb, sheet, x = linkString, startCol = 1, startRow = 1)
  expect_silent(saveWorkbook(wb, tempFile, overwrite = TRUE))
  
  # TODO: add a check that the written xlsx file contains linkString
  
})

context("Reading from wb object is identical to reading from file")

test_that("Reading from new workbook", {
  curr_wd <- getwd()

  wb <- createWorkbook()
  for (i in 1:4) {
    addWorksheet(wb, sprintf("Sheet %s", i))
  }


  ## colNames = TRUE, rowNames = TRUE
  writeData(wb, sheet = 1, x = mtcars, colNames = TRUE, rowNames = TRUE, startRow = 10, startCol = 5)
  x <- read.xlsx(wb, 1, colNames = TRUE, rowNames = TRUE)
  expect_equal(object = mtcars, expected = x, check.attributes = TRUE)


  ## colNames = TRUE, rowNames = FALSE
  writeData(wb, sheet = 2, x = mtcars, colNames = TRUE, rowNames = FALSE, startRow = 10, startCol = 5)
  x <- read.xlsx(wb, sheet = 2, colNames = TRUE, rowNames = FALSE)
  expect_equal(object = mtcars, expected = x, check.attributes = FALSE)
  expect_equal(object = colnames(mtcars), expected = colnames(x), check.attributes = FALSE)

  ## colNames = FALSE, rowNames = TRUE
  writeData(wb, sheet = 3, x = mtcars, colNames = FALSE, rowNames = TRUE, startRow = 2, startCol = 2)
  x <- read.xlsx(wb, sheet = 3, colNames = FALSE, rowNames = TRUE)
  expect_equal(object = mtcars, expected = x, check.attributes = FALSE)
  expect_equal(object = rownames(mtcars), expected = rownames(x))


  ## colNames = FALSE, rowNames = FALSE
  writeData(wb, sheet = 4, x = mtcars, colNames = FALSE, rowNames = FALSE, startRow = 12, startCol = 1)
  x <- read.xlsx(wb, sheet = 4, colNames = FALSE, rowNames = FALSE)
  expect_equal(object = mtcars, expected = x, check.attributes = FALSE)

  expect_equal(object = getwd(), curr_wd)
  rm(wb)
})

test_that("Empty workbook", {
  curr_wd <- getwd()
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")

  expect_equal(NULL, suppressWarnings(read.xlsx(wb)))

  expect_equal(NULL, suppressWarnings(read.xlsx(wb, sheet = 1, colNames = FALSE)))
  expect_equal(NULL, suppressWarnings(read.xlsx(wb, sheet = 1, colNames = TRUE)))

  expect_equal(NULL, suppressWarnings(read.xlsx(wb, sheet = 1, colNames = FALSE, skipEmptyRows = FALSE)))
  expect_equal(NULL, suppressWarnings(read.xlsx(wb, sheet = 1, colNames = TRUE, skipEmptyRows = FALSE)))

  expect_equal(NULL, suppressWarnings(read.xlsx(wb, sheet = 1, colNames = FALSE, skipEmptyRows = TRUE, detectDates = TRUE)))
  expect_equal(NULL, suppressWarnings(read.xlsx(wb, sheet = 1, colNames = TRUE, skipEmptyRows = TRUE, detectDates = FALSE)))

  expect_equal(NULL, suppressWarnings(read.xlsx(wb, sheet = 1, colNames = FALSE, skipEmptyRows = TRUE, detectDates = TRUE, rows = 4:10)))
  expect_equal(NULL, suppressWarnings(read.xlsx(wb, sheet = 1, colNames = TRUE, skipEmptyRows = TRUE, detectDates = FALSE, cols = 4:10)))

  expect_warning(read.xlsx(wb))

  expect_warning(read.xlsx(wb, sheet = 1, colNames = FALSE))
  expect_warning(read.xlsx(wb, sheet = 1, colNames = TRUE))

  expect_warning(read.xlsx(wb, sheet = 1, colNames = FALSE, skipEmptyRows = FALSE))
  expect_warning(read.xlsx(wb, sheet = 1, colNames = TRUE, skipEmptyRows = FALSE))

  expect_warning(read.xlsx(wb, sheet = 1, colNames = FALSE, skipEmptyRows = TRUE, detectDates = TRUE))
  expect_warning(read.xlsx(wb, sheet = 1, colNames = TRUE, skipEmptyRows = TRUE, detectDates = FALSE))

  expect_warning(read.xlsx(wb, sheet = 1, colNames = FALSE, skipEmptyRows = TRUE, detectDates = TRUE, rows = 4:10))
  expect_warning(read.xlsx(wb, sheet = 1, colNames = TRUE, skipEmptyRows = TRUE, detectDates = FALSE, cols = 4:10))


  ## 1 element
  writeData(wb, 1, "a")

  x <- read.xlsx(wb)
  expect_equal(nrow(x), 0)
  expect_equal(names(x), "a")

  x <- read.xlsx(wb, sheet = 1, colNames = FALSE)
  expect_equal(data.frame("X1" = "a", stringsAsFactors = FALSE), x)

  x <- read.xlsx(wb, sheet = 1, colNames = TRUE)
  expect_equal(nrow(x), 0)
  expect_equal(names(x), "a")

  x <- read.xlsx(wb, sheet = 1, colNames = FALSE, skipEmptyRows = FALSE)
  expect_equal(data.frame("X1" = "a", stringsAsFactors = FALSE), x)

  x <- read.xlsx(wb, sheet = 1, colNames = TRUE, skipEmptyRows = FALSE)
  expect_equal(nrow(x), 0)
  expect_equal(names(x), "a")

  writeData(wb, 1, Sys.Date(), startCol = 1, startRow = 1)
  x <- read.xlsx(wb)
  expect_equal(nrow(x), 0)
  expect_equal(convertToDate(as.integer(names(x)[1])), Sys.Date())


  x <- read.xlsx(wb, sheet = 1, colNames = FALSE)
  expect_equal(nrow(x), 1)

  x <- read.xlsx(wb, sheet = 1, colNames = FALSE, skipEmptyRows = TRUE, detectDates = TRUE)
  expect_equal(class(x[[1]]), "Date")

  x <- read.xlsx(wb, sheet = 1, colNames = FALSE, skipEmptyRows = TRUE, detectDates = TRUE)
  expect_equal(x[[1]], Sys.Date())

  x <- suppressWarnings(read.xlsx(wb, sheet = 1, colNames = FALSE, skipEmptyRows = TRUE, detectDates = TRUE, rows = 4:10))
  expect_equal(NULL, x)

  x <- suppressWarnings(read.xlsx(wb, sheet = 1, colNames = TRUE, skipEmptyRows = TRUE, detectDates = FALSE, cols = 4:10))
  expect_equal(NULL, x)



  addWorksheet(wb, "Sheet 2")
  removeWorksheet(wb, 1)


  ## 1 date
  writeData(wb, 1, Sys.Date(), colNames = FALSE)

  x <- read.xlsx(wb)
  expect_equal(convertToDate(names(x)), Sys.Date())

  x <- read.xlsx(wb, sheet = 1, colNames = FALSE)
  x1 <- convertToDate(x[[1]])
  expect_equal(x1, Sys.Date())

  x <- read.xlsx(wb, sheet = 1, colNames = FALSE, skipEmptyRows = TRUE, detectDates = TRUE)
  expect_equal(class(x[[1]]), "Date")
  expect_equal(x[[1]], Sys.Date())

  x <- read.xlsx(wb, sheet = 1, colNames = TRUE, skipEmptyRows = TRUE, detectDates = TRUE)
  expect_equal(as.Date(names(x)), Sys.Date())

  x <- suppressWarnings(read.xlsx(wb, sheet = 1, colNames = FALSE, skipEmptyRows = TRUE, detectDates = TRUE, rows = 4:10))
  expect_equal(NULL, x)

  x <- suppressWarnings(read.xlsx(wb, sheet = 1, colNames = TRUE, skipEmptyRows = TRUE, detectDates = FALSE, cols = 4:10))
  expect_equal(NULL, x)


  expect_equal(object = getwd(), curr_wd)
})


test_that("Reading NAs and NaN values", {
  fileName <- file.path(tempdir(), "NaN.xlsx")
  na.string <- "*"

  ## data
  a <- data.frame(
    X  = c(-pi / 0, NA, NaN),
    Y  = letters[1:3],
    Z  = c(pi / 0, 99, NaN),
    Z2 = c(1, NaN, NaN),
    stringsAsFactors = FALSE
  )

  b <- a
  b[b == -Inf] <- NaN
  b[b == Inf] <- NaN

  c <- b
  is_na <- sapply(c, is.na)
  is_nan <- sapply(c, is.nan)
  c[is_na & !is_nan] <- na.string
  is_nan_after <- sapply(c, is.nan)
  c[is_nan & !is_nan_after] <- NA

  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  writeData(wb, 1, a, keepNA = FALSE)

  addWorksheet(wb, "Sheet 2")
  writeData(wb, 2, a, keepNA = TRUE)

  addWorksheet(wb, "Sheet 3")
  writeData(wb, 3, a, keepNA = TRUE, na.string = na.string)

  saveWorkbook(wb, file = fileName, overwrite = TRUE)

  ## from file
  expected_df <- structure(list(
    X = c(NA_real_, NA_real_, NA_real_),
    Y = c("a", "b", "c"),
    Z = c(NA, 99, NA),
    Z2 = c(1, NA, NA)
  ),
  .Names = c("X", "Y", "Z", "Z2"),
  row.names = c(NA, 3L), class = "data.frame"
  )

  expect_equal(read.xlsx(fileName), expected_df)

  ## from workbook
  expected_df <- structure(list(
    X = c(NA_real_, NA_real_, NA_real_),
    Y = c("a", "b", "c"),
    Z = c(NA, 99, NA),
    Z2 = c(1, NA, NA)
  ),
  .Names = c("X", "Y", "Z", "Z2"),
  row.names = c(NA, 3L), class = "data.frame"
  )

  expect_equal(read.xlsx(wb), expected_df)

  ## keepNA = FALSE
  expect_equal(read.xlsx(wb), read.xlsx(fileName))
  expect_equal(b, read.xlsx(wb))
  expect_equal(b, read.xlsx(fileName))

  ## keepNA = TRUE

  expect_equal(read.xlsx(wb), expected_df)
  expect_equal(read.xlsx(fileName), expected_df)

  expect_equal(b, read.xlsx(wb, sheet = 2))
  expect_equal(b, read.xlsx(fileName, sheet = 2))

  ## keepNA = TRUE, na.string = "*"
  expect_equal(c, read.xlsx(wb, sheet = 3))
  expect_equal(c, read.xlsx(fileName, sheet = 3))

  unlink(fileName, recursive = TRUE, force = TRUE)
})


test_that("Reading from new workbook 2 ", {

  ## data
  genDf <- function() {
    data.frame(
      "Date" = Sys.Date() - 0:4,
      "Logical" = c(TRUE, FALSE, TRUE, TRUE, FALSE),
      "Currency" = -2:2,
      "Accounting" = -2:2,
      "hLink" = "https://CRAN.R-project.org/",
      "Percentage" = seq(-1, 1, length.out = 5),
      "TinyNumber" = runif(5) / 1E9, stringsAsFactors = FALSE
    )
  }

  df <- genDf()

  class(df$Currency) <- "currency"
  class(df$Accounting) <- "accounting"
  class(df$hLink) <- "hyperlink"
  class(df$Percentage) <- "percentage"
  class(df$TinyNumber) <- "scientific"

  options("openxlsx.dateFormat" = NULL)

  fileName <- file.path(tempdir(), "allClasses.xlsx")
  wb <- write.xlsx(df, file = fileName, overwrite = TRUE)

  x <- read.xlsx(wb, sheet = 1, detectDates = FALSE)
  x[[1]] <- convertToDate(x[[1]])
  expect_equal(object = x, expected = genDf(), check.attributes = FALSE)


  x <- read.xlsx(wb, sheet = 1, detectDates = TRUE)
  expect_equal(object = x, expected = genDf(), check.attributes = FALSE)

  unlink(fileName, recursive = TRUE, force = TRUE)
})





test_that("Reading from new workbook cols/rows", {
  wb <- createWorkbook()
  for (i in 1:4) {
    addWorksheet(wb, sprintf("Sheet %s", i))
  }

  tempFile <- temp_xlsx()

  ## 1
  writeData(wb, sheet = 1, x = mtcars, colNames = TRUE, rowNames = FALSE)
  saveWorkbook(wb, tempFile, overwrite = TRUE)

  cols <- 1:3
  rows <- 1:10
  x <- read.xlsx(wb, 1, colNames = TRUE, rowNames = FALSE, rows = rows, cols = cols)
  y <- read.xlsx(tempFile, 1, colNames = TRUE, rowNames = FALSE, rows = rows, cols = cols)

  df <- mtcars[sort((rows - 1)[(rows - 1) <= nrow(mtcars)]), sort(cols[cols <= ncol(mtcars)])]
  rownames(df) <- seq_len(nrow(df))

  expect_equal(object = x, expected = y)
  expect_equal(object = x, expected = df)



  ## 2
  writeData(wb, sheet = 2, x = mtcars, colNames = TRUE, rowNames = FALSE, startRow = 10, startCol = 5)
  saveWorkbook(wb, tempFile, overwrite = TRUE)

  cols <- 1:300
  rows <- 1:1000
  x <- read.xlsx(wb, sheet = 2, colNames = TRUE, rowNames = FALSE, rows = rows, cols = cols)
  y <- read.xlsx(tempFile, sheet = 2, colNames = TRUE, rowNames = FALSE, rows = rows, cols = cols)

  #
  expect_equal(object = mtcars, expected = x, check.attributes = FALSE)
  expect_equal(object = x, expected = y, check.attributes = TRUE)
  expect_equal(object = colnames(mtcars), expected = colnames(x), check.attributes = FALSE)


  cols <- 1:3
  rows <- 12:13
  x <- suppressWarnings(read.xlsx(wb, sheet = 2, colNames = TRUE, rowNames = FALSE, rows = rows, cols = cols))
  y <- suppressWarnings(read.xlsx(tempFile, sheet = 2, colNames = TRUE, rowNames = FALSE, rows = rows, cols = cols))


  expect_equal(object = NULL, expected = x, check.attributes = FALSE)
  expect_equal(object = NULL, expected = y, check.attributes = TRUE)



  ## 3
  writeData(wb, sheet = 3, x = mtcars, colNames = TRUE, rowNames = FALSE)
  saveWorkbook(wb, tempFile, overwrite = TRUE)

  cols <- c(2, 4, 6)
  rows <- seq(1, 31, by = 2)

  x <- read.xlsx(wb, sheet = 3, colNames = TRUE, rowNames = FALSE, rows = rows, cols = cols)
  y <- read.xlsx(tempFile, sheet = 3, colNames = TRUE, rowNames = FALSE, rows = rows, cols = cols)

  df <- mtcars[sort((rows - 1)[(rows - 1) <= nrow(mtcars)]), sort(cols[cols <= ncol(mtcars)])]
  rownames(df) <- seq_len(nrow(df))

  expect_equal(object = x, expected = y, check.attributes = FALSE)
  expect_equal(object = df, expected = x, check.attributes = FALSE)



  ## 4
  writeData(wb, sheet = 4, x = mtcars, colNames = TRUE, rowNames = TRUE)
  saveWorkbook(wb, tempFile, overwrite = TRUE)

  cols <- c(1, 6, 12)
  rows <- seq(1, 31, by = 2)
  x <- read.xlsx(wb, sheet = 4, colNames = TRUE, rowNames = TRUE, rows = rows, cols = cols)
  y <- read.xlsx(tempFile, sheet = 4, colNames = TRUE, rowNames = TRUE, rows = rows, cols = cols)

  df <- mtcars[sort((rows - 1)[(rows - 1) <= nrow(mtcars)]), cols[-1] - 1]
  expect_equal(object = x, expected = y, check.attributes = FALSE)
  expect_equal(object = df, expected = x, check.attributes = FALSE)

  rm(wb)
  unlink(tempFile, recursive = TRUE, force = TRUE)
})

context("error without write permissions")

test_that("test failed write errors for saveWorkbook", {
  
  skip_on_cran()
  skip_on_ci()
  
  temp_file <- tempfile()
  file.create(temp_file)
  Sys.chmod(path = temp_file, mode = "444")

  wb <- createWorkbook()
  addWorksheet(wb, "name")

  expect_warning(write.xlsx(
    x = cars, file = temp_file, overwrite = TRUE
  ),
  regexp = "Permission denied"
  )

  unlink(temp_file, recursive = TRUE, force = TRUE)
})

context("Encoding Tests")



test_that("Write read encoding equality", {
  tempFile <- temp_xlsx()

  wb <- createWorkbook()
  for (i in 1:4) {
    addWorksheet(wb, sprintf("Sheet %s", i))
  }

  df <- data.frame("X" = c("测试", "一下"), stringsAsFactors = FALSE)
  writeDataTable(wb, sheet = 1, x = df)

  saveWorkbook(wb, tempFile, overwrite = TRUE)

  x <- read.xlsx(tempFile)
  expect_equal(x, df)

  x <- read.xlsx(wb)
  expect_equal(x, df)

  ## reload
  wb <- loadWorkbook(tempFile)

  x <- read.xlsx(wb)
  expect_equal(x, df)

  saveWorkbook(wb, tempFile, overwrite = TRUE)
  x <- read.xlsx(tempFile)
  expect_equal(x, df)

  unlink(tempFile, recursive = TRUE, force = TRUE)
  rm(wb)
})


test_that("Support non-ASCII strings not in UTF-8 encodings", {
  non_ascii <- c("\u4f60\u597d", "\u4e2d\u6587", "\u6c49\u5b57", 
                 "\u5de5\u4f5c\u7c3f1", "\u6d4b\u8bd5\u540d\u5b571",
                 "\u6d4b2", "\u5de52", "\u5de5\u4f5c3")
  # Ideally, we should test against native encodings. However, the testing machine's
  #   locale encoding may not be able to represent the non-ascii letters, when
  #   it's the case, we use the UTF-8 encoding as it is.
  if (identical( enc2utf8(enc2native(non_ascii)), non_ascii )) {
    non_ascii <- enc2native(non_ascii)
  }
  non_ascii_df <- data.frame(
    X = non_ascii, Y = seq_along(non_ascii), stringsAsFactors = FALSE
  )
  colnames(non_ascii_df) <- non_ascii[3:4]
  file <- temp_xlsx()
  wb <- createWorkbook(creator = non_ascii[1])
  ws <- addWorksheet(wb, non_ascii[2])
  writeDataTable(wb, ws, non_ascii_df, tableName = non_ascii[3])
  writeData(wb, ws, non_ascii_df, xy = list("D", 1), name = non_ascii[4])
  writeComment(wb, ws, 1, 1, comment = createComment(non_ascii[5], non_ascii[6]))
  writeFormula(wb, ws, x = sprintf('"%s"&"%s"', non_ascii[1], non_ascii[2]), xy = list("G", 1))
  createNamedRegion(wb, ws, 7, 1, name = non_ascii[7])
  saveWorkbook(wb, file)
  
  wb2 <- loadWorkbook(file)
  expect_equal(
    getCreators(wb2), non_ascii[1]
  )
  expect_equal(
    getSheetNames(file), non_ascii[2]
  )
  expect_equivalent(
    getTables(wb2, ws), non_ascii[3]
  )
  expect_equivalent(
    getNamedRegions(wb2), non_ascii[c(4, 7)]
  )
  expect_equal(
    wb2$comments[[1]][[1]][c("comment", "author")],
    setNames(as.list(non_ascii[5:6]), c("comment", "author"))
  )
  expect_equal(
    read.xlsx(file, ws, cols = 1:2),
    non_ascii_df
  )
  expect_equal(
    read.xlsx(file, ws, cols = 4:5),
    non_ascii_df
  )
})

context("Group visibility and nested Grouping")



test_that("row and column grouping", {
  # Test file has several worksheets:
  #   1. "simple visible grouping": One simple grouping (rows 1:2, columns 1:2), visible
  #   2. "simple hidden grouping": One simple grouping (rows 1:2, columns 1:2), hidden
  #   3. "multiple non-nested groupings": three groupings (rows/cols 1:2, 4:5 and 7:8 (hidden); table contents only until cell G6, so last grouping has no actual content)
  #   4. "nested grouping": cols rows 1:4 (outline level 1) and 2:3 (outline level 2), contents to cell F7 (spanning all groupings)
  #   4. "nested grouping 2": Like previous, but outline level 2 is hidden, contents only to cell B2
  
  # Potential issues with groupings:
  #   - Outline level not preserved
  #   - Hidden flag not set (or not correctly loaded)
  #   - When nesting goes beyond the cell contents, rows/cols need to be added manually to the xml for the grouping settings
  #   - Hidden flag for visible rows/cols not properly loaded, so subsequent hidden cols set the hidden flag for the wrong row/col
  #   - Non-overlapping groupings 

  
  wb <- loadWorkbook(file = system.file("extdata", "nested_grouped_rowscols.xlsx", package = "openxlsx"))
  fileName <- file.path(tempdir(), "nested_grouped_rowscols_out.xlsx")
  oL1 = c(`1` = "1", `2` = "1")
  oL3 = c(`1` = "1", `2` = "1", `4` = "1", `5` = "1", `7` = "1", `8` = "1")
  oL4 = c(`1` = "1", `2` = "2", `3` = "2", `4` = "1")
  
  stripAttributes <- function(x, attr = NA) {
    y = x
    if (is.na(attr)) {
      attributes(y) <- NULL
    } else {
      attr(y, attr) <- NULL
    }
    y
  }
  
  ### 1. Loading of outline level
  
  # First sheet has simple, visible groupings
  expect_equal(oL1, stripAttributes(wb$outlineLevels[[1]], "hidden"))
  expect_equal(oL1, stripAttributes(wb$colOutlineLevels[[1]], "hidden"))
  # Second sheet has simple, hidden groupings
  expect_equal(oL1, stripAttributes(wb$outlineLevels[[2]], "hidden"))
  expect_equal(oL1, stripAttributes(wb$colOutlineLevels[[2]], "hidden"))
  # Third sheet has simple, non-overlapping groupings
  expect_equal(oL3, stripAttributes(wb$outlineLevels[[3]], "hidden"))
  expect_equal(oL3, stripAttributes(wb$colOutlineLevels[[3]], "hidden"))
  # Fourth tab has nested, visible groupings
  expect_equal(oL4, stripAttributes(wb$outlineLevels[[4]], "hidden"))
  expect_equal(oL4, stripAttributes(wb$colOutlineLevels[[4]], "hidden"))

  
  ### 2. Loading of row/columng visibility of groupings (hidden attribute)
  
  # First tab has visible groupings, second tab has hidden groupings, third has visible, fourth has visible and hidden rows/cols
  expect_equal(rep(NA_character_, 2), attr(wb$outlineLevels[[1]], "hidden"))
  expect_equal(c("1", "1"), attr(wb$outlineLevels[[2]], "hidden"))
  expect_equal(c(rep(NA_character_, 4), "1", "1"), attr(wb$outlineLevels[[3]], "hidden"))
  expect_equal(rep(NA_character_, 4), attr(wb$outlineLevels[[4]], "hidden"))
  expect_equal(c(NA_character_, "1", "1", NA_character_), attr(wb$outlineLevels[[5]], "hidden"))
  
  expect_equal(rep("0", 2), attr(wb$colOutlineLevels[[1]], "hidden"))
  expect_equal(c("1", "1"), attr(wb$colOutlineLevels[[2]], "hidden"))
  expect_equal(c("0", "0", "0", "0", "1", "1"), attr(wb$colOutlineLevels[[3]], "hidden"))
  expect_equal(rep("0", 4), attr(wb$colOutlineLevels[[4]], "hidden"))
  expect_equal(c("0", "1", "1", "0"), attr(wb$colOutlineLevels[[5]], "hidden"))
  
  
  #### 3. Export xlsx file and read it in again to check if outlines are preserved
  # The test file has groupings beyond the actual file contents, so not all 
  # grouped rows/cols are available in the data and need to be added manually
  # for the grouping data!
  
  openxlsx::saveWorkbook(wb, file = fileName, overwrite = TRUE)
  wbout <- loadWorkbook(file = fileName)
  
  expect_equal(wb$outlineLevels[[1]], wbout$outlineLevels[[1]])
  expect_equal(wb$outlineLevels[[2]], wbout$outlineLevels[[2]])
  expect_equal(wb$outlineLevels[[3]], wbout$outlineLevels[[3]])
  expect_equal(wb$outlineLevels[[4]], wbout$outlineLevels[[4]])
  expect_equal(wb$outlineLevels[[5]], wbout$outlineLevels[[5]])
  
  expect_equal(wb$colOutlineLevels[[1]], wbout$colOutlineLevels[[1]])
  expect_equal(wb$colOutlineLevels[[2]], wbout$colOutlineLevels[[2]])
  expect_equal(wb$colOutlineLevels[[3]], wbout$colOutlineLevels[[3]]) # BUG: Ordering of entries not preserved
  expect_equal(wb$colOutlineLevels[[4]], wbout$colOutlineLevels[[4]]) # BUG: Ordering of entries not preserved
  expect_equal(wb$colOutlineLevels[[5]], wbout$colOutlineLevels[[5]]) # BUG: Ordering of entries not preserved
  
  
  #### 4. Manually create non-overlapping and nested groupings of rows / cols
  
  wb <- openxlsx::createWorkbook()
  wb$addWorksheet(sheetName = "Manually")
  writeData(wb, "Manually", matrix(1:225, 15, 15), 1, 1)
  # non-nested grouping (visible and hidden)
  groupRows(wb, "Manually", 2:3, FALSE)
  groupRows(wb, "Manually", 5:6, TRUE)
  # nested grouping
  groupRows(wb, "Manually", 9:15) # Outline level 1, next groupings will be level 2! 
  groupRows(wb, "Manually", 10:11)
  groupRows(wb, "Manually", 13:14)
  expect_equal(c(`2` = "1", `3` = "1", `5` = "1", `6` = "1", `9` = "1", `10` = "2", `11` = "2", `12` = "1", `13` = "2", `14` = "2", `15` = "1"),
               stripAttributes(wb$outlineLevels[[1]], "hidden"))

  # non-nested grouping (visible and hidden)
  groupColumns(wb, "Manually", 2:3, FALSE)
  groupColumns(wb, "Manually", 5:6, TRUE)
  # nested grouping
  groupColumns(wb, "Manually", 9:15)
  groupColumns(wb, "Manually", 10:11)
  groupColumns(wb, "Manually", 13:14)
  expect_equal(c(`2` = "1", `3` = "1", `5` = "1", `6` = "1", `9` = "1", `10` = "2", `11` = "2", `12` = "1", `13` = "2", `14` = "2", `15` = "1"),
               stripAttributes(wb$colOutlineLevels[[1]], "hidden"))
  
  
  #### 5. Ungrouping rows/cols simply decrements the level (and removes the entry if no longer grouped)
  ungroupRows(wb, "Manually", 11:13)
  expect_equal(c(`2` = "1", `3` = "1", `5` = "1", `6` = "1", `9` = "1", `10` = "2", `11` = "1", `13` = "1", `14` = "2", `15` = "1"),
               stripAttributes(wb$outlineLevels[[1]], "hidden"))
  
  ungroupColumns(wb, "Manually", 11:13)
  expect_equal(c(`2` = "1", `3` = "1", `5` = "1", `6` = "1", `9` = "1", `10` = "2", `11` = "1", `13` = "1", `14` = "2", `15` = "1"),
               stripAttributes(wb$colOutlineLevels[[1]], "hidden"))
  

  #### CLEANUP
  
  unlink(fileName, recursive = TRUE, force = TRUE)
  
})

context("write.xlsx vector arguments")

test_that("Writing then reading returns identical data.frame 1", {
  tmp_file <- file.path(tempdir(), "xlsx_vector_args.xlsx")
  
  df1 <- data.frame(1:2)
  df2 <- data.frame(1:3)
  x <- list(df1, df2)
  
  write.xlsx(
    file = tmp_file,
    x = x,
    gridLines = c(FALSE, TRUE),
    sheetName = c("a", "b"),
    zoom = c(50, 90),
    tabColour = c("red", "blue")
  )
  
  wb <- loadWorkbook(tmp_file)
  
  expect_equal(getSheetNames(tmp_file), expected = c("a", "b"))
  expect_equal(names(wb), expected = c("a", "b"))
  
  expect_true(grepl('rgb="FFFF0000"', wb$worksheets[[1]]$sheetPr))
  expect_true(grepl('rgb="FF0000FF"', wb$worksheets[[2]]$sheetPr))
  
  expect_true(grepl('zoomScale="50"', wb$worksheets[[1]]$sheetViews))
  expect_true(grepl('zoomScale="90"', wb$worksheets[[2]]$sheetViews))
  
  expect_true(grepl('showGridLines="0"', wb$worksheets[[1]]$sheetViews))
  expect_true(grepl('showGridLines="1"', wb$worksheets[[2]]$sheetViews))
  
  expect_equal(read.xlsx(tmp_file, sheet = 1), df1)
  expect_equal(read.xlsx(tmp_file, sheet = 2), df2)
  
  unlink(tmp_file, recursive = TRUE, force = TRUE)
})

test_that("write.xlsx() passes withFilter and colWidths [151]", {
  df <- data.frame(x = 1, b = 2)
  x <- buildWorkbook(df) 
  y <- buildWorkbook(df, withFilter = TRUE, colWidths = 15)
  
  expect_equal(x$worksheets[[1]]$autoFilter, character())
  expect_equal(y$worksheets[[1]]$autoFilter, "<autoFilter ref=\"A1:B2\"/>")
  expect_equal(x$colWidths, list(list()))
  
  expect_equal(
    y$colWidths[[1]],
    structure(c(`1` = "15", `2` = "15"), hidden = c("0", "0"))
  )
}) 

test_that("write.xlsx() correctly passes default asTable and withFilters", {
  df <- data.frame(x = 1, b = 2)
  
  # asTable = TRUE  >> writeDataTable >> withFilter = TRUE
  # asTable = FALSE >> writeData      >> withFilter = FALSE
  x <- buildWorkbook(df, asTable = FALSE) 
  y <- buildWorkbook(df, asTable = TRUE)
  
  # Save the workbook 
  tf <- temp_xlsx()
  saveWorkbook(y, tf)
  y2 <- loadWorkbook(tf)
  
  expect_identical(x$worksheets[[1]]$autoFilter, character())
  
  # not autoFilter for tables -- not named in buildWorkbook
  expect_equal(
    y$worksheets[[1]]$tableParts,
    structure("<tablePart r:id=\"rId3\"/>", tableName = "Table3")
  )
  
  expect_equal(
    y2$worksheets[[1]]$tableParts,
    structure("<tablePart r:id=\"rId3\"/>", tableName = c(`A1:B2` = "Table3"))
  )
  
  file.remove(tf)
})

test_that("write.xlsx() correctly handles colWidths", {
  x <- data.frame(a = 1, b = 2, c = 3)
  zero3 <- rep("0", 3)
  
  # No warning when passing "auto"
  expect_warning(buildWorkbook(rep_len(list(x), 3), colWidths = "auto"), NA)
  
  # single value is repeated for all columns
  wb <- buildWorkbook(rep_len(list(x), 3), colWidths = 13)
  exp <- rep_len(list(structure(c(`1` = "13", `2` = "13", `3` = "13"), hidden = zero3)), 3)
  expect_equal(wb$colWidths, exp)
  
  # sets are repeated
  wb <- buildWorkbook(rep_len(list(x), 3), colWidths = list(c(10, 20, 30)))
  exp <- rep_len(list(structure(c(`1` = "10", `2` = "20", `3` = "30"), hidden = zero3)), 3)
  expect_equal(wb$colWidths, exp)
  
  # 3 distinct sets
  wb <- buildWorkbook(
    rep_len(list(x), 3),
    colWidths = list(
      c(10, 20, 30),
      c(100, 200, 300),
      c(1, 2, 3)
    ))
  
  expect_equal(
    wb$colWidths,
    list(
      structure(c(`1` = "10", `2` = "20", `3` = "30"), hidden = zero3),
      structure(c(`1` = "100", `2` = "200", `3` = "300"), hidden = zero3),
      structure(c(`1` = "1", `2` = "2", `3` = "3"), hidden = zero3)
    )
  )
})

context("Styles")

test_that("setStyle", {
  
  tmp_file <- temp_xlsx()
  
  # lorem ipsum
  txt <- paste0(
    "Lorem ipsum dolor sit amet, consectetur adipiscing elit, ",
    "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. ",
    "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris",
    "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in ",
    "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla",
    "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in ",
    "culpa qui officia deserunt mollit anim id est laborum."
  )
  
  ## create workbook
  wb <- createWorkbook()
  addWorksheet(wb, "Test")
  writeData(wb, "Test", txt)
  
  ## create a style
  s <- createStyle(
    fontSize = 12, 
    fontColour = "black", 
    valign="center", 
    wrapText = TRUE, 
    halign = "justify"
  )
  addStyle(wb, "Test", s, 1, 1)
  setColWidths(wb, "Test", 1, 50)
  setRowHeights(wb, "Test", 1, 150)
  
  ## save workbook
  saveWorkbook(wb, tmp_file)
  
  ## load it again
  wb2 <- loadWorkbook(tmp_file)
  s2 <- getStyles(wb2)[[1]]
  
  ## test that the style survived the round trip
  expect_equal(s2$fontSize, c(val="12"))
  expect_equal(s2$fontColour, c(rgb="FF000000"))
  expect_equal(s2$valign, "center")
  expect_equal(s2$wrapText, TRUE)
  expect_equal(s2$halign, "justify")
  
})

context("Page setup")


test_that("Page setup", {
  wb <- createWorkbook()
  addWorksheet(wb, "s1")
  addWorksheet(wb, "s2")


  pageSetup(wb,
    sheet = "s1", orientation = "landscape", scale = 100, left = 0.1,
    right = 0.1, top = .75, bottom = .75, header = 0.1, footer = 0.1,
    fitToWidth = TRUE, fitToHeight = TRUE, paperSize = 1,
    summaryRow = "below", summaryCol = "right"
  )


  pageSetup(wb,
    sheet = 2, orientation = "landscape", scale = 100, left = 0.1,
    right = 0.1, top = .75, bottom = .75, header = 0.1, footer = 0.1,
    fitToWidth = TRUE, fitToHeight = TRUE, paperSize = 1,
    summaryRow = "below", summaryCol = "right"
  )

  expect_equal(wb$worksheets[[1]]$pageSetup, wb$worksheets[[2]]$pageSetup)

  v <- gsub(" ", "", wb$worksheets[[1]]$pageSetup, fixed = TRUE)
  expect_true(grepl('paperSize="1"', v))
  expect_true(grepl('orientation="landscape"', v))
  expect_true(grepl('fitToWidth="1"', v))
  expect_true(grepl('fitToHeight="1"', v))

  pr <- wb$worksheets[[1]]$sheetPr

  # SheetPr will be a character vector of length 2; the first entry will
  # be for PageSetupPr, inserted by `fitToWidth`/`fitToHeight`.
  expect_true(grepl('<outlinePr summaryBelow="1" summaryRight="1"/>', pr[2], fixed = TRUE))
})

context("Writing Posixct")



test_that("Writing Posixct with writeData & writeDataTable", {
  options("openxlsx.datetimeFormat" = "dd/mm/yy hh:mm")

  tstart <- strptime("30/05/2017 08:30", "%d/%m/%Y %H:%M", tz = "CET")
  TimeDT <- c(0, 5, 10, 15, 30, 60, 120, 180, 240, 480, 720, 1440) * 60 + tstart
  df <- data.frame(TimeDT, TimeTxt = format(TimeDT, "%Y-%m-%d %H:%M"))

  wb <- createWorkbook()
  addWorksheet(wb, "writeData")
  addWorksheet(wb, "writeDataTable")

  writeData(wb, "writeData", df, startCol = 2, startRow = 3, rowNames = FALSE)
  writeDataTable(wb, "writeDataTable", df, startCol = 2, startRow = 3)

  wd <- as.numeric(wb$worksheets[[1]]$sheet_data$v)
  wdt <- as.numeric(wb$worksheets[[2]]$sheet_data$v)


  expected <- c(
    0, 1, 42885.3541666667, 2, 42885.3576388889, 3, 42885.3611111111,
    4, 42885.3645833333, 5, 42885.375, 6, 42885.3958333333, 7, 42885.4375,
    8, 42885.4791666667, 9, 42885.5208333333, 10, 42885.6875, 11,
    42885.8541666667, 12, 42886.3541666667, 13
  )

  expect_equal(object = round(wd, 12), expected = expected)
  expect_equal(object = round(wdt, 12), expected = expected)
  expect_equal(object = wd, expected = wdt)

  options("openxlsx.datetimeFormat" = "yyyy-mm-dd hh:mm:ss")
})


test_that("Writing mixed EDT/EST Posixct with writeData & writeDataTable", {
  options("openxlsx.datetimeFormat" = "dd/mm/yy hh:mm")

  tstart1 <- as.POSIXct("12/03/2018 08:30", format = "%d/%m/%Y %H:%M")
  tstart2 <- as.POSIXct("10/03/2018 08:30", format = "%d/%m/%Y %H:%M")
  TimeDT1 <- c(NA, 0, 10, 30, 60, 120, 240, 720, 1440) * 60 + tstart1
  TimeDT2 <- c(0, 10, 30, 60, 120, 240, 720, NA, 1440) * 60 + tstart2

  df <- data.frame(
    timeval = c(TimeDT1, TimeDT2),
    timetxt = format(c(TimeDT1, TimeDT2), "%Y-%m-%d %H:%M")
  )

  wb <- createWorkbook()
  addWorksheet(wb, "writeData")
  addWorksheet(wb, "writeDataTable")

  writeData(wb, "writeData", df, startCol = 2, startRow = 3, rowNames = FALSE)
  writeDataTable(wb, "writeDataTable", df, startCol = 2, startRow = 3)

  wd <- as.numeric(wb$worksheets[[1]]$sheet_data$v)
  wdt <- as.numeric(wb$worksheets[[2]]$sheet_data$v)
  wd <- wd[wb$worksheets[[1]]$sheet_data$cols == 2]
  wdt <- wdt[wb$worksheets[[2]]$sheet_data$cols == 2]

  # drop any integer indexes introduced in write
  wd <- wd[wd != 0 | is.na(wd)]
  wdt <- wdt[wdt  != 0 | is.na(wdt)]

  # sort everything
  wd <- convertToDateTime(wd[order(wd)])
  wdt <- convertToDateTime(wdt[order(wdt)])
  expected <- df$timeval[order(df$timeval)]

  # compare
  expect_equal(
    object = wd,
    expected = expected,
    tolerance = 10 ^ -10,
    check.tzone = FALSE
  )
  expect_equal(
    object = wdt,
    expected = expected,
    tolerance = 10 ^ -10,
    check.tzone = TRUE
  )
  expect_equal(
    object = wd,
    expected = wdt,
    check.tzone = TRUE
  )
  
  options("openxlsx.datetimeFormat" = "yyyy-mm-dd hh:mm:ss")
})

context("Setting column widths")

test_that("Resetting col widths", {
  # Write some data to a workbook object.
  wb <- createWorkbook()
  addWorksheet(wb, "iris") 
  writeData(wb, "iris", iris)
  
  # Set column widths and perform the pre-save operation to prepare col xml 
  # (typically called by `saveWorkbook()`).
  setColWidths(wb, "iris", cols = 1:2, 12)
  wb$setColWidths(1)
  
  # Set column widths again for a different range inclusive of the previous one, 
  # perform the pre-save operation to prepare col xml, and test for an error 
  # (reported in https://github.com/ycphs/openxlsx/issues/493).
  setColWidths(wb, "iris", cols = 1:5, 15)
  expect_error(wb$setColWidths(1), NA)
  
  # Test that the resulting col xml meets expectations.
  expected_col_xml <- c(
    "1" = "<col min=\"1\" max=\"1\" width=\"15.71\" hidden=\"0\" customWidth=\"1\"/>",
    "2" = "<col min=\"2\" max=\"2\" width=\"15.71\" hidden=\"0\" customWidth=\"1\"/>",
    "3" = "<col min=\"3\" max=\"3\" width=\"15.71\" hidden=\"0\" customWidth=\"1\"/>",
    "4" = "<col min=\"4\" max=\"4\" width=\"15.71\" hidden=\"0\" customWidth=\"1\"/>",
    "5" = "<col min=\"5\" max=\"5\" width=\"15.71\" hidden=\"0\" customWidth=\"1\"/>"
  )
  expect_equal(
    wb$worksheets[[1]]$cols, 
    expected_col_xml
  )
})

context("Testing 'topN' and 'bottomN' conditions in conditionalFormatting")
TNBN_test_data <- data.frame(col1 = 1:10,
                             col2 = 1:10,
                             col3 = seq(10, 100, 10),
                             col4 = seq(10, 100, 10),
                             col5 = 1:10,
                             col6 = 1:10)

bg_blue <- createStyle(bgFill = "skyblue")

wb <- createWorkbook()
sht <- "TopN_BottomN_TEST"
addWorksheet(wb, sht)
writeData(wb, sht, TNBN_test_data)
conditionalFormatting(wb, sht, cols = 1, rows = 2:11, type = "topN",    rank = 3,  style = bg_blue, percent = FALSE)
conditionalFormatting(wb, sht, cols = 2, rows = 2:11, type = "bottomN", rank = 3,  style = bg_blue, percent = FALSE)
conditionalFormatting(wb, sht, cols = 3, rows = 2:11, type = "topN",    rank = 50, style = bg_blue, percent = TRUE)
conditionalFormatting(wb, sht, cols = 4, rows = 2:11, type = "bottomN", rank = 50, style = bg_blue, percent = TRUE)
conditionalFormatting(wb, sht, cols = 5, rows = 2:11, type = "topN",    rank = 3,  style = bg_blue)
conditionalFormatting(wb, sht, cols = 6, rows = 2:11, type = "bottomN", rank = 3,  style = bg_blue)

test_that("Number of conditionalFormatting rules added equal to 6", {
  expect_equal(object = length(wb$worksheets[[1]]$conditionalFormatting), expected = 6)
})

test_that("topN conditions do not have the 'bottom' argument", {
  expect_false(object = grepl(paste('bottom'), wb$worksheets[[1]]$conditionalFormatting[1]))
  expect_false(object = grepl(paste('bottom'), wb$worksheets[[1]]$conditionalFormatting[3]))
})

test_that("bottomN conditions have the 'bottom' argument set to '1'", {
  expect_true(object = grepl(paste('bottom="1"'), wb$worksheets[[1]]$conditionalFormatting[2]))
  expect_true(object = grepl(paste('bottom="1"'), wb$worksheets[[1]]$conditionalFormatting[4]))
})

test_that("topN/bottomN rank conditions have the 'percent=FALSE' argument set to '0'", {
  expect_true(object = grepl(paste('percent="0"'), wb$worksheets[[1]]$conditionalFormatting[1]))
  expect_true(object = grepl(paste('percent="0"'), wb$worksheets[[1]]$conditionalFormatting[2]))
})

test_that("topN/bottomN rank conditions do not have the 'percent' argument is set to 'NULL'", {
  expect_true(object = grepl(paste('percent="NULL"'), wb$worksheets[[1]]$conditionalFormatting[5]))
  expect_true(object = grepl(paste('percent="NULL"'), wb$worksheets[[1]]$conditionalFormatting[6]))
})

test_that("topN/bottomN percent conditions have the 'percent' argument set to '1'", {
  expect_true(object = grepl(paste('percent="1"'), wb$worksheets[[1]]$conditionalFormatting[3]))
  expect_true(object = grepl(paste('percent="1"'), wb$worksheets[[1]]$conditionalFormatting[4]))
})

test_that("topN/bottomN conditions correspond to 'top10' type", {
  expect_true(object = grepl(paste('type="top10"'), wb$worksheets[[1]]$conditionalFormatting[1]))
  expect_true(object = grepl(paste('type="top10"'), wb$worksheets[[1]]$conditionalFormatting[2]))
  expect_true(object = grepl(paste('type="top10"'), wb$worksheets[[1]]$conditionalFormatting[3]))
  expect_true(object = grepl(paste('type="top10"'), wb$worksheets[[1]]$conditionalFormatting[4]))
  expect_true(object = grepl(paste('type="top10"'), wb$worksheets[[1]]$conditionalFormatting[5]))
  expect_true(object = grepl(paste('type="top10"'), wb$worksheets[[1]]$conditionalFormatting[6]))
})


context("Testing 'blanks' and 'notBlanks' conditions in conditionalFormatting")
BNB_test_data <- data.frame(col1 = sample(c("X", NA_character_), 10, replace = TRUE),
                             col2 = sample(c("Y", NA_character_), 10, replace = TRUE))

bg_blue <- createStyle(bgFill = "skyblue")
bg_red <- createStyle(bgFill = "red")

wb <- createWorkbook()
sht <- "Blanks_NonBlanks_TEST"
addWorksheet(wb, sht)
writeData(wb, sht, BNB_test_data)
conditionalFormatting(wb, sht, cols = 1, rows = 2:11, type = "blanks",    style = bg_red)
conditionalFormatting(wb, sht, cols = 2, rows = 2:11, type = "notBlanks", style = bg_blue)

test_that("Number of conditionalFormatting rules added equal to 2", {
  expect_equal(object = length(wb$worksheets[[1]]$conditionalFormatting), expected = 2)
})

test_that("type='blanks' calls type='containsBlanks'", {
  expect_true(object = grepl(paste('containsBlanks'), wb$worksheets[[1]]$conditionalFormatting[1]))
})

test_that("type='notBlanks' calls type='notContainsBlanks'", {
  expect_true(object = grepl(paste('notContainsBlanks'), wb$worksheets[[1]]$conditionalFormatting[2]))
})

context("Images and Tables.")


test_that("Images and Tables - reordering and removing", {
  if (FALSE) {
    options("stringsAsFactors" = FALSE)
    tempFile <- temp_xlsx("break")

    getPlot <- function(i) {
      n <- 5000
      plot(1:n, rnorm(n))
      title(main = sprintf("Plot for Sheet: %s", i))
    }

    df1 <- iris[1:5, 1:4]
    df2 <- mtcars


    df3 <- data.frame(
      "Date" = Sys.Date() - 0:10,
      "Logical" = sample(c(TRUE, FALSE), 1, replace = TRUE),
      "Currency" = as.numeric(-5:5) * 100,
      "Accounting" = as.numeric(-5:5),
      "hLink" = "https://CRAN.R-project.org/",
      "Percentage" = seq(-5, 5, length.out = 11),
      "TinyNumber" = runif(11) / 1E9, stringsAsFactors = FALSE
    )

    df3U <- df3

    class(df3$Currency) <- "currency"
    class(df3$Accounting) <- "accounting"
    class(df3$hLink) <- "hyperlink"
    class(df3$Percentage) <- "percentage"
    class(df3$TinyNumber) <- "scientific"


    df4 <- data.frame("X" = 1:10000, "Y" = sample(LETTERS, size = 10000, replace = TRUE))
    df5 <- USJudgeRatings

    hs <- createStyle(fontColour = "blue", textRotation = 45)


    wb <- createWorkbook()
    expect_equal(names(wb), character(0))

    addWorksheet(wb = wb, sheetName = "Sheet 1", gridLines = FALSE, tabColour = "red", zoom = 75)
    writeDataTable(wb, sheet = 1, x = df1, startCol = 7, startRow = 10, tableName = "Sheet1Table1")
    expect_equal(names(wb), "Sheet 1")


    addWorksheet(wb, sheetName = "Sheet 2", tabColour = "purple")
    writeDataTable(wb, sheet = "Sheet 2", x = df2, startCol = 2, startRow = 2, rowNames = TRUE)
    expect_equal(names(wb), c("Sheet 1", "Sheet 2"))



    addWorksheet(wb, sheetName = "Sheet 3", tabColour = "green")
    writeDataTable(wb, sheet = 3, x = df3, startCol = 1, startRow = 1)
    expect_equal(names(wb), c("Sheet 1", "Sheet 2", "Sheet 3"))

    addWorksheet(wb, sheetName = "Sheet 4", tabColour = "orange")
    writeDataTable(wb, sheet = 4, x = df4)
    expect_equal(names(wb), c("Sheet 1", "Sheet 2", "Sheet 3", "Sheet 4"))

    addWorksheet(wb, sheetName = "Sheet 5", tabColour = "yellow")
    writeData(wb, sheet = "Sheet 5", x = df5, rowNames = TRUE)
    expect_equal(names(wb), c("Sheet 1", "Sheet 2", "Sheet 3", "Sheet 4", "Sheet 5"))



    worksheetOrder(wb) <- c(1, 3, 5, 4, 2)
    expect_equal(names(wb), c("Sheet 1", "Sheet 2", "Sheet 3", "Sheet 4", "Sheet 5"))

    ## save and load 1
    saveWorkbook(wb, file = tempFile, overwrite = TRUE)

    wb <- loadWorkbook(tempFile)
    expect_equal(names(wb), c("Sheet 1", "Sheet 3", "Sheet 5", "Sheet 4", "Sheet 2"))


    expect_equal(df1, read.xlsx(wb, sheet = 1))
    expect_equal(df1, read.xlsx(wb, sheet = "Sheet 1"))
    expect_equal(df1, read.xlsx(tempFile, sheet = 1))
    expect_equal(df1, read.xlsx(tempFile, sheet = "Sheet 1"))


    expect_equal(df3U, read.xlsx(wb, sheet = 2, detectDates = TRUE))
    expect_equal(df3U, read.xlsx(wb, sheet = "Sheet 3", detectDates = TRUE))
    expect_equal(df3U, read.xlsx(tempFile, sheet = 2, detectDates = TRUE))
    expect_equal(df3U, read.xlsx(tempFile, sheet = "Sheet 3", detectDates = TRUE))


    expect_equal(df5, read.xlsx(wb, sheet = 3, rowNames = TRUE))
    expect_equal(df5, read.xlsx(wb, sheet = "Sheet 5", rowNames = TRUE))
    expect_equal(df5, read.xlsx(tempFile, sheet = 3, rowNames = TRUE))
    expect_equal(df5, read.xlsx(tempFile, sheet = "Sheet 5", rowNames = TRUE))


    expect_equal(df4, read.xlsx(wb, sheet = 4))
    expect_equal(df4, read.xlsx(wb, sheet = "Sheet 4"))
    expect_equal(df4, read.xlsx(tempFile, sheet = 4))
    expect_equal(df4, read.xlsx(tempFile, sheet = "Sheet 4"))


    expect_equal(df2, read.xlsx(wb, sheet = 5, rowNames = TRUE))
    expect_equal(df2, read.xlsx(wb, sheet = "Sheet 2", rowNames = TRUE))
    expect_equal(df2, read.xlsx(tempFile, sheet = 5, rowNames = TRUE))
    expect_equal(df2, read.xlsx(tempFile, sheet = "Sheet 2", rowNames = TRUE))



    ## remove "Sheet 5" by index (3)
    removeWorksheet(wb, sheet = 3)
    expect_equal(names(wb), c("Sheet 1", "Sheet 3", "Sheet 4", "Sheet 2"))

    ## remove sheet "Sheet 4"
    removeWorksheet(wb, sheet = "Sheet 4")
    expect_equal(names(wb), c("Sheet 1", "Sheet 3", "Sheet 2"))


    ## Introduce some images
    getPlot(1)
    insertPlot(wb = wb, sheet = "Sheet 1", startCol = 14, startRow = 3)

    getPlot(2)
    insertPlot(wb = wb, sheet = "Sheet 2", startCol = 14, startRow = 3)

    getPlot(3)
    insertPlot(wb = wb, sheet = "Sheet 3", startCol = 14, startRow = 3)


    expect_true(any(grepl("image1", wb$drawings_rels[[1]])))
    expect_true(any(grepl("image3", wb$drawings_rels[[2]])))
    expect_true(any(grepl("image2", wb$drawings_rels[[3]])))



    ## put back to original order
    worksheetOrder(wb) <- c(1, 3, 2)
    saveWorkbook(wb, file = tempFile, overwrite = TRUE)

    wb <- loadWorkbook(file = tempFile)


    ## drawings added in order
    expect_true(any(grepl("image1", wb$drawings_rels[[1]])))
    expect_true(any(grepl("image2", wb$drawings_rels[[2]])))
    expect_true(any(grepl("image3", wb$drawings_rels[[3]])))


    ## Introduce some more images
    getPlot("1_2")
    insertPlot(wb = wb, sheet = "Sheet 1", startCol = 14, startRow = 25)

    getPlot("2_2")
    insertPlot(wb = wb, sheet = "Sheet 2", startCol = 14, startRow = 25)


    getPlot("3_2")
    insertPlot(wb = wb, sheet = "Sheet 3", startCol = 14, startRow = 25)

    saveWorkbook(wb, file = tempFile, overwrite = TRUE)
    wb <- loadWorkbook(tempFile)

    worksheetOrder(wb) <- c(3, 2, 1)
    saveWorkbook(wb, file = tempFile, overwrite = TRUE)
    wb <- loadWorkbook(tempFile)


    hl <- rep("https://google.com.au", 5)
    names(hl) <- sprintf("Link to google %s", 1:5)
    class(hl) <- "hyperlink"
    writeData(wb, "Sheet 1", hl)

    ## Add in some column widths

    setColWidths(wb, sheet = 1, cols = 1:50, widths = "auto")
    worksheetOrder(wb) <- c(3, 2, 1)
    removeWorksheet(wb, sheet = "Sheet 2")

    saveWorkbook(wb, file = tempFile, overwrite = TRUE)
    wb <- loadWorkbook(tempFile)

    expect_equal(names(wb), c("Sheet 1", "Sheet 3"))
    expect_equal(df1, read.xlsx(tempFile, sheet = 1, startRow = 10))
    expect_equal(df3U, read.xlsx(tempFile, sheet = 2, detectDates = TRUE))

    expect_equal(df1, read.xlsx(wb, sheet = 1, startRow = 10))
    expect_equal(df3U, read.xlsx(wb, sheet = 2, detectDates = TRUE))


    unlink(tempFile, recursive = TRUE, force = TRUE)
    rm(wb)
  }
})

test_that("setColWidths() should support zero-length cols", {
  file <- temp_xlsx()
  on.exit(unlink(file), add = TRUE)
  wb <- createWorkbook()
  ws <- addWorksheet(wb, "empty")
  tbl <- data.frame(A = 1:3)
  writeData(wb, ws, tbl)
  setColWidths(wb, ws, integer(0L), widths = 12)
  saveWorkbook(wb, file)
  x <- readWorkbook(file)
  expect_equal(x, tbl)
})

context("Inserting images")

test_that("Inserting hyperlinked image", {
  # Create a workbook.
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  
  # Insert multiple images. Specifically one with a link before and after other 
  # images to test whether drawings reference the correct index of the drawing 
  # relationships.
  img <- system.file("extdata", "einstein.jpg", package = "openxlsx")
  insertImage(wb, "Sheet 1", img)
  insertImage(wb, "Sheet 1", img, address = "https://github.com/ycphs/openxlsx")
  insertImage(wb, "Sheet 1", img)
  
  # Declare expectations for drawings and drawings_rels xml.
  expected_drawings <- list(
    "<xdr:oneCellAnchor xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\"><xdr:from xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n    <xdr:col xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">0</xdr:col>\n    <xdr:colOff xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">0</xdr:colOff>\n    <xdr:row xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">0</xdr:row>\n    <xdr:rowOff xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">0</xdr:rowOff>\n  </xdr:from><xdr:ext xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\" cx=\"5486400\" cy=\"2743200\"/><xdr:pic xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n      <xdr:nvPicPr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n        <xdr:cNvPr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\" id=\"1\" name=\"Picture 1\"/>\n        <xdr:cNvPicPr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n          <a:picLocks xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" noChangeAspect=\"1\"/>\n        </xdr:cNvPicPr>\n      </xdr:nvPicPr>\n      <xdr:blipFill xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n        <a:blip xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" r:embed=\"rId1\">\n        </a:blip>\n        <a:stretch xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\">\n          <a:fillRect/>\n        </a:stretch>\n      </xdr:blipFill>\n      <xdr:spPr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n        <a:prstGeom xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" prst=\"rect\">\n          <a:avLst xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\"/>\n        </a:prstGeom>\n      </xdr:spPr>\n    </xdr:pic><xdr:clientData/></xdr:oneCellAnchor>",
    "<xdr:oneCellAnchor xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\"><xdr:from xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n    <xdr:col xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">0</xdr:col>\n    <xdr:colOff xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">0</xdr:colOff>\n    <xdr:row xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">0</xdr:row>\n    <xdr:rowOff xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">0</xdr:rowOff>\n  </xdr:from><xdr:ext xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\" cx=\"5486400\" cy=\"2743200\"/><xdr:pic xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n      <xdr:nvPicPr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n        <xdr:cNvPr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\" id=\"2\" name=\"Picture 2\">\n              <a:hlinkClick xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" r:id=\"rId3\"/>\n            </xdr:cNvPr>\n        <xdr:cNvPicPr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n          <a:picLocks xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" noChangeAspect=\"1\"/>\n        </xdr:cNvPicPr>\n      </xdr:nvPicPr>\n      <xdr:blipFill xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n        <a:blip xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" r:embed=\"rId2\">\n        </a:blip>\n        <a:stretch xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\">\n          <a:fillRect/>\n        </a:stretch>\n      </xdr:blipFill>\n      <xdr:spPr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n        <a:prstGeom xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" prst=\"rect\">\n          <a:avLst xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\"/>\n        </a:prstGeom>\n      </xdr:spPr>\n    </xdr:pic><xdr:clientData/></xdr:oneCellAnchor>",
    "<xdr:oneCellAnchor xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\"><xdr:from xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n    <xdr:col xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">0</xdr:col>\n    <xdr:colOff xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">0</xdr:colOff>\n    <xdr:row xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">0</xdr:row>\n    <xdr:rowOff xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">0</xdr:rowOff>\n  </xdr:from><xdr:ext xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\" cx=\"5486400\" cy=\"2743200\"/><xdr:pic xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n      <xdr:nvPicPr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n        <xdr:cNvPr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\" id=\"3\" name=\"Picture 3\"/>\n        <xdr:cNvPicPr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n          <a:picLocks xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" noChangeAspect=\"1\"/>\n        </xdr:cNvPicPr>\n      </xdr:nvPicPr>\n      <xdr:blipFill xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n        <a:blip xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" r:embed=\"rId4\">\n        </a:blip>\n        <a:stretch xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\">\n          <a:fillRect/>\n        </a:stretch>\n      </xdr:blipFill>\n      <xdr:spPr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\">\n        <a:prstGeom xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" prst=\"rect\">\n          <a:avLst xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\"/>\n        </a:prstGeom>\n      </xdr:spPr>\n    </xdr:pic><xdr:clientData/></xdr:oneCellAnchor>"
  )
  expected_drawings_rels <- list(
    "<Relationship Id=\"rId1\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/image\" Target=\"../media/image1.jpg\"/>",
    "<Relationship Id=\"rId2\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/image\" Target=\"../media/image2.jpg\"/>",
    "<Relationship Id=\"rId3\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/hyperlink\" Target=\"https://github.com/ycphs/openxlsx\" TargetMode=\"External\"/>",
    "<Relationship Id=\"rId4\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/image\" Target=\"../media/image3.jpg\"/>"
  )
  
  # Test expectations for drawings and drawings_rels xml.
  expect_equal(expected_drawings, wb$drawings[[1]])
  expect_equal(expected_drawings_rels, wb$drawings_rels[[1]])
  
  # Test errors.
  expect_error(
    insertImage(wb, "Sheet 1", img, address = NULL), 
    "Invalid address"
  )
  expect_error(
    insertImage(wb, "Sheet 1", img, address = NA), 
    "Invalid address"
  )
  expect_error(
    insertImage(
      wb, 
      "Sheet 1", 
      img, 
      address = c("https://github.com/ycphs/openxlsx", "https://github.com/ycphs/openxlsx/issues")
    ), 
    "Invalid address"
  )
})

context("Protection")


test_that("Protect Workbook", {
  wb <- createWorkbook()
  addWorksheet(wb, "s1")

  wb$protectWorkbook(password = "abcdefghij")

  expect_true(wb$workbook$workbookProtection == "<workbookProtection workbookPassword=\"FEF1\"/>")

  wb$protectWorkbook(protect = FALSE, password = "abcdefghij", lockStructure = TRUE, lockWindows = TRUE)
  expect_true(wb$workbook$workbookProtection == "")
})

test_that("Reading protected Workbook", {
  tmp_file <- file.path(tempdir(), "xlsx_read_protectedwb.xlsx")

  wb <- createWorkbook()
  addWorksheet(wb, "s1")
  protectWorkbook(wb, password = "abcdefghij")
  saveWorkbook(wb, tmp_file, overwrite = TRUE)

  wb2 <- loadWorkbook(file = tmp_file)
  # Check that the order of the sub-elements is preserved
  n1 <- names(wb2$workbook)
  n2 <- names(wb$workbook)[names(wb$workbook) != "apps"]
  expect_equal(n1, n2)

  unlink(tmp_file, recursive = TRUE, force = TRUE)
})

context("Reading from workbook is identical to reading from file")

test_that("Reading from loaded workbook", {
  wb <- createWorkbook()
  for (i in 1:4) {
    addWorksheet(wb, sprintf("Sheet %s", i))
  }

  writeData(wb, sheet = 1, x = mtcars, colNames = TRUE, rowNames = TRUE, startRow = 10, startCol = 5, borders = "all")
  writeData(wb, sheet = 2, x = mtcars, colNames = TRUE, rowNames = FALSE, startRow = 10, startCol = 5, borders = "rows")
  writeData(wb, sheet = 3, x = mtcars, colNames = FALSE, rowNames = TRUE, startRow = 2, startCol = 2, borders = "columns")
  writeData(wb, sheet = 4, x = mtcars, colNames = FALSE, rowNames = FALSE, startRow = 12, startCol = 1, borders = "surrounding")

  tempFile <- temp_xlsx()
  saveWorkbook(wb, tempFile, overwrite = TRUE)

  wb <- loadWorkbook(tempFile)

  ## colNames = TRUE, rowNames = TRUE
  x <- read.xlsx(wb, 1, colNames = TRUE, rowNames = TRUE)
  expect_equal(object = mtcars, expected = x, check.attributes = TRUE)


  ## colNames = TRUE, rowNames = FALSE
  x <- read.xlsx(wb, sheet = 2, colNames = TRUE, rowNames = FALSE)
  expect_equal(object = mtcars, expected = x, check.attributes = FALSE)
  expect_equal(object = colnames(mtcars), expected = colnames(x), check.attributes = FALSE)


  ## colNames = FALSE, rowNames = TRUE
  x <- read.xlsx(wb, sheet = 3, colNames = FALSE, rowNames = TRUE)
  expect_equal(object = mtcars, expected = x, check.attributes = FALSE)
  expect_equal(object = rownames(mtcars), expected = rownames(x))


  ## colNames = FALSE, rowNames = FALSE
  x <- read.xlsx(wb, sheet = 4, colNames = FALSE, rowNames = FALSE)
  expect_equal(object = mtcars, expected = x, check.attributes = FALSE)

  unlink(tempFile, recursive = TRUE, force = TRUE)
})


test_that("na.convert = TRUE", {
  
  fl <- system.file("extdata", "na_convert.xlsx", package = "openxlsx")
  
  
  # optional return empty cell as "<si><t/></si>"
  wb1 <- loadWorkbook(fl, na.convert = FALSE)
  tmp_no_na <- temp_xlsx("na_convert_false")
  saveWorkbook(wb1, tmp_no_na)
  
  # default return emtpy cell as "NA" string
  wb2 <- loadWorkbook(fl, na.convert = TRUE)
  tmp_has_na <- temp_xlsx("na_convert_true")
  saveWorkbook(wb2, tmp_has_na)
  
  # # FIXME both return identical output
  # read.xlsx(wb1)
  # read.xlsx(wb2)
  
  # # for visual comparison
  # openXL(tmp_has_na)
  # openXL(tmp_no_na)
  
  ## new option
  # check value of last cells: index position of shared strings
  exp <- c("1", "0", "0", "0", "0", "0")
  got <- tail(wb1$worksheets[[1]]$sheet_data$v)
  expect_equal(exp, got)
  
  # check value of shared string: index position 0
  exp <- "<si><t/></si>"
  got <- wb1$sharedStrings[[1]]
  expect_equal(exp, got)
  
  
  ## default
  # check value of last cells: index position of shared strings
  exp <- c("1", "0", "0", "0", "0", "0")
  got <- tail(wb2$worksheets[[1]]$sheet_data$v)
  expect_equal(exp, got)
  
  # check value of shared string: index position 0
  exp <- "<si><t>NA</t></si>"
  got <- wb2$sharedStrings[[1]]
  expect_equal(exp, got)
  
})

context("Fill Merged Cells")



test_that("fill merged cells", {
  wb <- createWorkbook()
  addWorksheet(wb, sheetName = "sheet1")
  writeData(wb = wb, sheet = 1, x = data.frame("A" = 1, "B" = 2))
  writeData(wb = wb, sheet = 1, x = 2, startRow = 2, startCol = 2)
  writeData(wb = wb, sheet = 1, x = 3, startRow = 2, startCol = 3)
  writeData(wb = wb, sheet = 1, x = 4, startRow = 2, startCol = 4)
  writeData(wb = wb, sheet = 1, x = t(matrix(1:4, 4, 4)), startRow = 3, startCol = 1, colNames = FALSE)

  mergeCells(wb = wb, sheet = 1, cols = 2:4, rows = 1)
  mergeCells(wb = wb, sheet = 1, cols = 2:4, rows = 3)
  mergeCells(wb = wb, sheet = 1, cols = 2:4, rows = 4)
  mergeCells(wb = wb, sheet = 1, cols = 2:4, rows = 5)

  tmp_file <- temp_xlsx()
  saveWorkbook(wb = wb, file = tmp_file, overwrite = TRUE)

  expect_equal(names(read.xlsx(tmp_file, fillMergedCells = FALSE)), c("A", "B", "X3", "X4"))
  expect_equal(names(read.xlsx(tmp_file, fillMergedCells = TRUE)), c("A", "B", "B", "B"))
  
  r1 <- data.frame("A" = rep(1, 5), "B" = rep(2, 5), "X3" = rep(3,5), "X4" = rep(4, 5))
  r2 <- data.frame("A" = rep(1, 5), "B" = rep(2, 5), "B1" = c(3,2,2,2,3), "B2" = c(4,2,2,2,4))
  names(r2) <- c("A", "B", "B", "B")
  
  r2_1 <- r2[1:5, 1:3]
  names(r2_1) <- c("A", "B", "B")
  
  expect_equal(read.xlsx(tmp_file, fillMergedCells = FALSE), r1)
  expect_equal(read.xlsx(tmp_file, fillMergedCells = TRUE), r2)
  
  expect_equal( read.xlsx(tmp_file, cols = 1:3, fillMergedCells = TRUE), r2_1)
  expect_equal( read.xlsx(tmp_file, rows = 1:3, fillMergedCells = TRUE), r2[1:2, ])
  expect_equal( read.xlsx(tmp_file, cols = 1:3, rows = 1:4, fillMergedCells = TRUE), r2_1[1:3,])
  
})

context("Workbook properties")

test_that("Workbook properties", {

  ## check creator
  wb <- createWorkbook(creator = "Alex", title = "title here", subject = "this & that", category = "some category")

  expect_true(grepl("<dc:creator>Alex</dc:creator>", wb$core))
  expect_true(grepl("<dc:title>title here</dc:title>", wb$core))
  expect_true(grepl("<dc:subject>this &amp; that</dc:subject>", wb$core))
  expect_true(grepl("<cp:category>some category</cp:category>", wb$core))

  fl <- tempfile(fileext = ".xlsx")
  wb <- write.xlsx(
    x = iris, file = fl,
    creator = "Alex 2",
    title = "title here 2",
    subject = "this & that 2",
    category = "some category 2"
  )

  expect_true(grepl("<dc:creator>Alex 2</dc:creator>", wb$core))
  expect_true(grepl("<dc:title>title here 2</dc:title>", wb$core))
  expect_true(grepl("<dc:subject>this &amp; that 2</dc:subject>", wb$core))
  expect_true(grepl("<cp:category>some category 2</cp:category>", wb$core))

  ## maintain on load
  wb_loaded <- loadWorkbook(fl)
  expect_equal(object = wb_loaded$core, expected = paste0(wb$core, collapse = ""))


  wb <- createWorkbook(creator = "Philipp", title = "title here", subject = "this & that", category = "some category")
  addCreator(wb, "test")
  expect_true(grepl("<dc:creator>Philipp;test</dc:creator>", wb$core))

  expect_equal(getCreators(wb), c("Philipp", "test"))
  setLastModifiedBy(wb, "Philipp 2")
  expect_true(grepl("<cp:lastModifiedBy>Philipp 2</cp:lastModifiedBy>", wb$core))
})


test_that("Workbook can print with 0 sheets [240]", {
  
  compare_text <- "A Workbook object.\n \nWorksheets:\n No worksheets attached\n"
  printed_text <- capture_output(x <- createWorkbook()$show())
  expect_null(x)
  expect_equal(compare_text, printed_text)
  
})

context("Renaming worksheets.")



test_that("Can rename worksheets under all conditions", {
  tempFile <- file.path(tempdir(), "renaming.xlsx")
  wb <- createWorkbook()
  addWorksheet(wb, "sheet 1")
  addWorksheet(wb, "sheet 2")
  addWorksheet(wb, "sheet 3")
  addWorksheet(wb, "sheet 4")
  addWorksheet(wb, "sheet 5")

  renameWorksheet(wb, sheet = 2, "THis is SHEET 2")
  expect_equal(names(wb), c("sheet 1", "THis is SHEET 2", "sheet 3", "sheet 4", "sheet 5"))


  renameWorksheet(wb, sheet = "THis is SHEET 2", "THis is STILL SHEET 2")
  expect_equal(names(wb), c("sheet 1", "THis is STILL SHEET 2", "sheet 3", "sheet 4", "sheet 5"))


  renameWorksheet(wb, sheet = 5, "THis is SHEET 5")
  expect_equal(names(wb), c("sheet 1", "THis is STILL SHEET 2", "sheet 3", "sheet 4", "THis is SHEET 5"))

  renameWorksheet(wb, sheet = 5, "THis is STILL SHEET 5")
  expect_equal(names(wb), c("sheet 1", "THis is STILL SHEET 2", "sheet 3", "sheet 4", "THis is STILL SHEET 5"))


  renameWorksheet(wb, sheet = 2, "Sheet 2")
  expect_equal(names(wb), c("sheet 1", "Sheet 2", "sheet 3", "sheet 4", "THis is STILL SHEET 5"))

  renameWorksheet(wb, sheet = 5, "Sheet 5")
  expect_equal(names(wb), c("sheet 1", "Sheet 2", "sheet 3", "sheet 4", "Sheet 5"))


  ## re-ordering
  worksheetOrder(wb) <- c(4, 3, 2, 5, 1)
  saveWorkbook(wb, tempFile, overwrite = TRUE)

  wb <- loadWorkbook(file = tempFile)
  renameWorksheet(wb, sheet = 2, "THIS is SHEET 3")

  wb <- loadWorkbook(tempFile)
  renameWorksheet(wb, sheet = "Sheet 5", "THIS is NOW SHEET 5")

  expect_equal(names(wb), c("sheet 4", "sheet 3", "Sheet 2", "THIS is NOW SHEET 5", "sheet 1"))

  names(wb)[[1]] <- "THIS IS NOW SHEET 4"
  expect_equal(names(wb), c("THIS IS NOW SHEET 4", "sheet 3", "Sheet 2", "THIS is NOW SHEET 5", "sheet 1"))


  unlink(tempFile, recursive = TRUE, force = TRUE)
})

context("active Sheet ")


test_that("get and set active sheet of a workbook", {
  
  tempFile1 <- temp_xlsx("temp1")
  tempFile2 <- temp_xlsx("temp2")
  tempFile3 <- temp_xlsx("temp3")
  wbook <- createWorkbook()
  addWorksheet(wbook, sheetName = "S1")
  addWorksheet(wbook, sheetName = "S2")
  addWorksheet(wbook, sheetName = "S3")
  
  
  saveWorkbook(wbook,tempFile1)
  # default value is the first sheet active
  expect_equal(activeSheet(wbook),1)
  expect_equal(activeSheet(wbook),loadWorkbook(tempFile1)$ActiveSheet)
  
  activeSheet(wbook) <- 1 ## active sheet S1
  saveWorkbook(wbook,tempFile2)
  expect_equal(activeSheet(wbook),1)
  expect_equal(activeSheet(wbook),loadWorkbook(tempFile2)$ActiveSheet)
  activeSheet(wbook) <- "S2" ## active sheet S2
  saveWorkbook(wbook,tempFile3)
  expect_equal(activeSheet(wbook),2)
  expect_equal(activeSheet(wbook),loadWorkbook(tempFile3)$ActiveSheet)
  
  unlink(tempFile1, recursive = TRUE, force = TRUE)
  unlink(tempFile2, recursive = TRUE, force = TRUE)
  unlink(tempFile3, recursive = TRUE, force = TRUE)
})

context("Worksheet naming")

test_that("Worksheet names", {
  
  ### test for names without special character
  wb <- createWorkbook()
  sheetname <- "test"
  addWorksheet(wb, sheetname)
  
  expect_equal(sheetname,names(wb))
  
  ### test for names with &
  
  wb <- createWorkbook()
  sheetname <- "S&P 500"
  addWorksheet(wb, sheetname)
  
  expect_equal(sheetname,names(wb))
  expect_equal("S&amp;P 500",wb$sheet_names)
  ### test for names with <
  
  wb <- createWorkbook()
  sheetname <- "<24 h"
  addWorksheet(wb, sheetname)
  
  expect_equal(sheetname,names(wb))
  expect_equal("&lt;24 h",wb$sheet_names)
  ### test for names with >
  
  wb <- createWorkbook()
  sheetname <- ">24 h"
  addWorksheet(wb, sheetname)
  
  expect_equal(sheetname,names(wb))
  expect_equal("&gt;24 h",wb$sheet_names)
  
  ### test for names with "
  
  wb <- createWorkbook()
  sheetname <- 'test "A"'
  addWorksheet(wb, sheetname)
  
  expect_equal(sheetname,names(wb))
  expect_equal("test &quot;A&quot;",wb$sheet_names)
})

context(desc = "Deleting tables from worksheets")

test_that("Deleting a Table Object", {
  wb <- createWorkbook()
  addWorksheet(wb, sheetName = "Sheet 1")
  addWorksheet(wb, sheetName = "Sheet 2")
  writeDataTable(wb, sheet = "Sheet 1", x = iris, tableName = "iris")
  writeDataTable(wb, sheet = 1, x = mtcars, tableName = "mtcars", startCol = 10)

  # Get table ----

  expect_equal(length(getTables(wb, sheet = 1)), 2L)
  expect_equal(length(getTables(wb, sheet = "Sheet 1")), 2L)

  expect_equal(length(getTables(wb, sheet = 2)), 0)
  expect_equal(length(getTables(wb, sheet = "Sheet 2")), 0)

  expect_error(getTables(wb, sheet = 3))
  expect_error(getTables(wb, sheet = "Sheet 3"))

  expect_equal(getTables(wb, sheet = 1), c("iris", "mtcars"), check.attributes = FALSE)
  expect_equal(getTables(wb, sheet = "Sheet 1"), c("iris", "mtcars"), check.attributes = FALSE)

  expect_equal(attr(getTables(wb, sheet = 1), "refs"), c("A1:E151", "J1:T33"))
  expect_equal(attr(getTables(wb, sheet = "Sheet 1"), "refs"), c("A1:E151", "J1:T33"))

  expect_equal(length(wb$tables), 2L)

  ## Deleting a worksheet ----

  removeWorksheet(wb, 1)
  expect_equal(length(wb$tables), 2L)
  expect_equal(length(getTables(wb, sheet = 1)), 0)

  expect_equal(attr(wb$tables, "tableName"), c("iris_openxlsx_deleted", "mtcars_openxlsx_deleted"))
  expect_equal(attr(wb$tables, "sheet"), c(0, 0))




  ###################################################################################
  ## write same tables again

  writeDataTable(wb, sheet = 1, x = iris, tableName = "iris")
  writeDataTable(wb, sheet = 1, x = mtcars, tableName = "mtcars", startCol = 10)

  expect_equal(attr(wb$tables, "tableName"), c("iris_openxlsx_deleted", "mtcars_openxlsx_deleted", "iris", "mtcars"))
  expect_equal(attr(wb$tables, "sheet"), c(0, 0, 1, 1))

  expect_equal(length(getTables(wb, sheet = 1)), 2L)
  expect_equal(length(getTables(wb, sheet = "Sheet 2")), 2L)

  expect_error(getTables(wb, sheet = 2))
  expect_error(getTables(wb, sheet = "Sheet 1"))

  expect_equal(getTables(wb, sheet = 1), c("iris", "mtcars"), check.attributes = FALSE)
  expect_equal(getTables(wb, sheet = "Sheet 2"), c("iris", "mtcars"), check.attributes = FALSE)

  expect_equal(attr(getTables(wb, sheet = 1), "refs"), c("A1:E151", "J1:T33"))
  expect_equal(attr(getTables(wb, sheet = "Sheet 2"), "refs"), c("A1:E151", "J1:T33"))

  expect_equal(length(wb$tables), 4L)


  ###################################################################################
  ## removeTable

  ## remove iris and re-write it
  removeTable(wb = wb, sheet = 1, table = "iris")

  expect_equal(length(wb$tables), 4L)
  expect_equal(wb$worksheets[[1]]$tableParts, "<tablePart r:id=\"rId6\"/>", check.attributes = FALSE)
  expect_equal(attr(wb$worksheets[[1]]$tableParts, "tableName"), "mtcars")

  expect_equal(attr(wb$tables, "tableName"), c(
    "iris_openxlsx_deleted",
    "mtcars_openxlsx_deleted",
    "iris_openxlsx_deleted",
    "mtcars"
  ))

  ## removeTable clears table object and all data
  writeDataTable(wb, sheet = 1, x = iris, tableName = "iris", startCol = 1)
  expect_equal(wb$worksheets[[1]]$tableParts, c("<tablePart r:id=\"rId6\"/>", "<tablePart r:id=\"rId7\"/>"), check.attributes = FALSE)
  expect_equal(attr(wb$worksheets[[1]]$tableParts, "tableName"), c("mtcars", "iris"))


  removeTable(wb = wb, sheet = 1, table = "iris")

  expect_equal(length(wb$tables), 5L)
  expect_equal(wb$worksheets[[1]]$tableParts, "<tablePart r:id=\"rId6\"/>", check.attributes = FALSE)
  expect_equal(attr(wb$worksheets[[1]]$tableParts, "tableName"), "mtcars")

  expect_equal(attr(wb$tables, "tableName"), c(
    "iris_openxlsx_deleted",
    "mtcars_openxlsx_deleted",
    "iris_openxlsx_deleted",
    "mtcars",
    "iris_openxlsx_deleted"
  ))


  expect_equal(getTables(wb, sheet = 1), "mtcars", check.attributes = FALSE)
})

test_that("Save and load Table Deletion", {
  temp_file <- temp_xlsx()

  wb <- createWorkbook()
  addWorksheet(wb, sheetName = "Sheet 1")
  addWorksheet(wb, sheetName = "Sheet 2")
  writeDataTable(wb, sheet = "Sheet 1", x = iris, tableName = "iris")
  writeDataTable(wb, sheet = 1, x = mtcars, tableName = "mtcars", startCol = 10)


  ###################################################################################
  ## Deleting a worksheet

  removeWorksheet(wb, 1)
  expect_equal(length(wb$tables), 2L)
  expect_equal(length(getTables(wb, sheet = 1)), 0)

  expect_equal(attr(wb$tables, "tableName"), c("iris_openxlsx_deleted", "mtcars_openxlsx_deleted"))
  expect_equal(attr(wb$tables, "sheet"), c(0, 0))


  ## both table were written to sheet 1 and are expected to not exist after load
  saveWorkbook(wb = wb, file = temp_file, overwrite = TRUE)
  wb <- loadWorkbook(file = temp_file)
  expect_null(wb$tables)
  unlink(temp_file)




  ###################################################################################
  ## Deleting a table

  wb <- createWorkbook()
  addWorksheet(wb, sheetName = "Sheet 1")
  addWorksheet(wb, sheetName = "Sheet 2")
  writeDataTable(wb, sheet = "Sheet 1", x = iris, tableName = "iris")
  writeDataTable(wb, sheet = 1, x = mtcars, tableName = "mtcars", startCol = 10)

  ## remove iris and re-write it
  removeTable(wb = wb, sheet = 1, table = "iris")
  expect_equal(attr(wb$tables, "tableName"), c("iris_openxlsx_deleted", "mtcars"))

  temp_file <- temp_xlsx()
  saveWorkbook(wb = wb, file = temp_file, overwrite = TRUE)
  wb <- loadWorkbook(file = temp_file)

  expect_equal(length(wb$tables), 1L)
  expect_equal(unname(attr(wb$tables, "tableName")), "mtcars")

  expect_equal(wb$worksheets[[1]]$tableParts, "<tablePart r:id=\"rId3\"/>", check.attributes = FALSE) ## rId reset
  expect_equal(unname(attr(wb$worksheets[[1]]$tableParts, "tableName")), "mtcars")
  unlink(temp_file)



  ## now delete the other table
  wb <- createWorkbook()
  addWorksheet(wb, sheetName = "Sheet 1")
  addWorksheet(wb, sheetName = "Sheet 2")
  writeDataTable(wb, sheet = "Sheet 1", x = iris, tableName = "iris")
  writeDataTable(wb, sheet = 1, x = mtcars, tableName = "mtcars", startCol = 10)
  writeDataTable(wb, sheet = 2, x = mtcars, tableName = "mtcars2", startCol = 3)

  removeTable(wb = wb, sheet = 1, table = "iris")
  removeTable(wb = wb, sheet = 1, table = "mtcars")
  expect_equal(attr(wb$tables, "tableName"), c("iris_openxlsx_deleted", "mtcars_openxlsx_deleted", "mtcars2"))

  temp_file <- temp_xlsx()
  saveWorkbook(wb = wb, file = temp_file, overwrite = TRUE)
  wb <- loadWorkbook(file = temp_file)


  expect_equal(length(wb$tables), 1L)
  expect_equal(unname(attr(wb$tables, "tableName")), "mtcars2")
  expect_length(wb$worksheets[[1]]$tableParts, 0)
  expect_equal(wb$worksheets[[2]]$tableParts, "<tablePart r:id=\"rId3\"/>", check.attributes = FALSE)
  expect_equal(unname(attr(wb$worksheets[[2]]$tableParts, "tableName")), "mtcars2")
  unlink(temp_file)


  ## write tables back in
  writeDataTable(wb, sheet = "Sheet 1", x = iris, tableName = "iris")
  writeDataTable(wb, sheet = 1, x = mtcars, tableName = "mtcars", startCol = 10)

  expect_equal(length(wb$tables), 3L)
  expect_equal(unname(attr(wb$tables, "tableName")), c("mtcars2", "iris", "mtcars"))

  expect_length(wb$worksheets[[1]]$tableParts, 2)
  expect_equal(wb$worksheets[[1]]$tableParts, c("<tablePart r:id=\"rId4\"/>", "<tablePart r:id=\"rId5\"/>"), check.attributes = FALSE)
  expect_equal(unname(attr(wb$worksheets[[1]]$tableParts, "tableName")), c("iris", "mtcars"))

  expect_length(wb$worksheets[[2]]$tableParts, 1)
  expect_equal(wb$worksheets[[2]]$tableParts, c("<tablePart r:id=\"rId3\"/>"), check.attributes = FALSE)
  expect_equal(unname(attr(wb$worksheets[[2]]$tableParts, "tableName")), "mtcars2")

  saveWorkbook(wb = wb, file = temp_file, overwrite = TRUE)


  ## Ids should get reset after load
  wb <- loadWorkbook(file = temp_file)

  expect_equal(length(wb$tables), 3L)
  expect_equal(unname(attr(wb$tables, "tableName")), c("iris", "mtcars", "mtcars2"))

  expect_length(wb$worksheets[[1]]$tableParts, 2)
  expect_equal(wb$worksheets[[1]]$tableParts, c("<tablePart r:id=\"rId3\"/>", "<tablePart r:id=\"rId4\"/>"), check.attributes = FALSE)
  expect_equal(unname(attr(wb$worksheets[[1]]$tableParts, "tableName")), c("iris", "mtcars"))

  expect_length(wb$worksheets[[2]]$tableParts, 1)
  expect_equal(wb$worksheets[[2]]$tableParts, c("<tablePart r:id=\"rId5\"/>"), check.attributes = FALSE)
  expect_equal(unname(attr(wb$worksheets[[2]]$tableParts, "tableName")), "mtcars2")

  unlink(temp_file)
})

context("Protection")


test_that("Protection", {
  wb <- createWorkbook()
  addWorksheet(wb, "s1")
  addWorksheet(wb, "s2")


  protectWorksheet(wb, sheet = "s1", protect = TRUE, password = "abcdefghij", lockSelectingLockedCells = FALSE, lockSelectingUnlockedCells = FALSE, lockFormattingCells = TRUE, lockFormattingColumns = TRUE, lockPivotTables = TRUE)

  expect_true(wb$worksheets[[1]]$sheetProtection == "<sheetProtection password=\"FEF1\" selectLockedCells=\"0\" selectUnlockedCells=\"0\" formatCells=\"1\" formatColumns=\"1\" pivotTables=\"1\" sheet=\"1\"/>")

  protectWorksheet(wb, sheet = "s2", protect = TRUE)
  expect_true(wb$worksheets[[2]]$sheetProtection == "<sheetProtection sheet=\"1\"/>")
  protectWorksheet(wb, sheet = "s2", protect = FALSE)
  expect_true(wb$worksheets[[2]]$sheetProtection == "")
})

context("Check Cell Ref")



test_that("Provide tests for single getCellRefs", {
  expect_equal(getCellRefs(data.frame(1, 2)), "B1")


  expect_error(getCellRefs(c(1, 2)))

  expect_error(getCellRefs(c(1, "a")))
})


test_that("Provide tests for multiple getCellRefs", {
  expect_equal(getCellRefs(data.frame(1:3, 2:4)), c("B1", "C2", "D3"))




  expect_error(getCellRefs(c(1:2, c("a", "b"))))
})

context("Test wrappers")

test_that("int2col and col2int", {

  nums <- 2:27

  chrs <- c("B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M",  "N",
            "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z", "AA")

  expect_equal(chrs, int2col(nums))
  expect_equal(nums, col2int(chrs))

})


test_that("deleteDataColumn basics", {
  wb <- createWorkbook()
  addWorksheet(wb, "tester")

  for (i in seq(5)) {
    mat <- data.frame(x = rep(paste0(int2col(i), i), 10))
    writeData(wb, sheet = 1, startRow = 1, startCol = i, mat)
    writeFormula(wb, sheet = 1, startRow = 12, startCol = i,
                 x = sprintf("=COUNTA(%s2:%s11)", int2col(i), int2col(i)))
  }
  expect_equal(
    setdiff(wb$worksheets[[1]]$sheet_data$f, NA),
    c("<f>=COUNTA(A2:A11)</f>", "<f>=COUNTA(B2:B11)</f>", "<f>=COUNTA(C2:C11)</f>", 
      "<f>=COUNTA(D2:D11)</f>", "<f>=COUNTA(E2:E11)</f>")
  )


  deleteDataColumn(wb, 1, col = 3)
  expect_equal(read.xlsx(wb),
               data.frame(x = rep("A1", 10), x = "B2", x = "D4", x = "E5", # no C3!
                          check.names = FALSE))
  expect_equal(
    setdiff(wb$worksheets[[1]]$sheet_data$f, NA),
    c("<f>=COUNTA(A2:A11)</f>", "<f>=COUNTA(B2:B11)</f>", 
      "<f>=COUNTA(C2:C11)</f>", "<f>=COUNTA(D2:D11)</f>")
  )


  deleteDataColumn(wb, 1, col = 2)
  expect_equal(read.xlsx(wb),
               data.frame(x = rep("A1", 10), x = "D4", x = "E5", # no B2!
                          check.names = FALSE))
  expect_equal(
    setdiff(wb$worksheets[[1]]$sheet_data$f, NA),
    c("<f>=COUNTA(A2:A11)</f>", "<f>=COUNTA(B2:B11)</f>", "<f>=COUNTA(C2:C11)</f>")
  )

  deleteDataColumn(wb, 1, col = 1)
  expect_equal(read.xlsx(wb),
               data.frame(x = rep("D4", 10), x = "E5", # no A1!
                          check.names = FALSE))
  expect_equal(
    setdiff(wb$worksheets[[1]]$sheet_data$f, NA),
    c("<f>=COUNTA(A2:A11)</f>", "<f>=COUNTA(B2:B11)</f>")
  )
})


test_that("deleteDataColumn with more complicated formulae", {
  # works with more complicated formula as well!
  wb <- createWorkbook()
  addWorksheet(wb, "tester")
  writeData(wb, sheet = 1, startRow = 1, startCol = 1,
            x = matrix(c(1, 1), ncol = 1), colNames = FALSE)

  for (c in 2:10)
    writeFormula(wb, 1, sprintf("%s1 + 1", int2col(c - 1)),
                 startRow = 1, startCol = c)

  for (c in 2:10)
    writeFormula(wb, 1, sprintf("%s1 + %s2", int2col(c), int2col(c - 1)),
                 startRow = 2, startCol = c)

  for (c in 2:10)
    writeFormula(wb, 1, sprintf("%s2 + %s2", int2col(c), int2col(c + 1)),
                 startRow = 3, startCol = c)

  deleteDataColumn(wb, 1, 3)
  # saveWorkbook(wb, "tester.xlsx") # and inspect by hand: expect lots of #REF!

  expect_equal(read.xlsx(wb), data.frame(`1` = 1, check.names = FALSE))
  expect_equal(
    wb$worksheets[[1]]$sheet_data$f,
    c(NA, "<f>A1 + 1</f>", "<f>#REF!1 + 1</f>", "<f>C1 + 1</f>", "<f>D1 + 1</f>", "<f>E1 + 1</f>", "<f>F1 + 1</f>", "<f>G1 + 1</f>", "<f>H1 + 1</f>",
      NA, "<f>B1 + A2</f>", "<f>C1 + #REF!2</f>", "<f>D1 + C2</f>", "<f>E1 + D2</f>", "<f>F1 + E2</f>", "<f>G1 + F2</f>", "<f>H1 + G2</f>", "<f>I1 + H2</f>",
      "<f>B2 + #REF!2</f>", "<f>C2 + D2</f>", "<f>D2 + E2</f>", "<f>E2 + F2</f>", "<f>F2 + G2</f>", "<f>G2 + H2</f>", "<f>H2 + I2</f>", "<f>I2 + J2</f>")
  )
})


test_that("deleteDataColumn with wide data", {
  wb <- createWorkbook()
  addWorksheet(wb, "tester")
  ncols <- 30
  nrows <- 100
  df <- as.data.frame(matrix(seq(ncols * nrows), ncol = ncols))
  colnames(df) <- int2col(seq(ncols))
  writeData(wb, sheet = 1, startRow = 1, startCol = 1, x = df, colNames = TRUE)
  expect_equal(read.xlsx(wb), df)

  deleteDataColumn(wb, 1, 2)
  expect_equal(read.xlsx(wb), df[, -2])

  deleteDataColumn(wb, 1, 100)
  expect_equal(read.xlsx(wb), df[, -2])

  deleteDataColumn(wb, 1, 55)
  expect_equal(read.xlsx(wb), df[, -c(2, 56)]) # 56 b.c. one col was already taken out

  deleteDataColumn(wb, 1, 1)
  expect_equal(read.xlsx(wb), df[, -c(1, 2, 56)])

  # delete all data
  for (i in seq(ncols - 2))
    deleteDataColumn(wb, 1, 1)
  expect_warning(read.xlsx(wb), "No data found on worksheet")
})

test_that("deleteDataColumn with formatting data", {
  wb <- createWorkbook()
  addWorksheet(wb, "tester")
  df <- data.frame(x = 1:10, y = letters[1:10], z = 10:1)
  writeData(wb, sheet = 1, startRow = 1, startCol = 1, x = df, colNames = TRUE)

  st <- openxlsx::createStyle(textDecoration = "Bold", fontSize = 20, fontColour = "red")
  openxlsx::addStyle(wb, 1, style = st, rows = 1, cols = seq(ncol(df)))

  sst <- wb$styleObjects[[1]]
  sst$rows <- c(1, 1)
  sst$cols <- c(1, 2)
  deleteDataColumn(wb, 1, 2)

  expect_length(wb$styleObjects, 1)
  expect_equal(wb$styleObjects[[1]],
               sst)
})

test_that("deleteDataColumn with shared strings does not crash or change inputs", {
  df <- data.frame("Col 1" = "Row 2 Col 1",
                   "Col 2" = "Row 2 Col 2",
                   "Col 3" = "Row 2 Col 3",
                   check.names = FALSE)

  wb <- createWorkbook()
  addWorksheet(wb, "tester")
  writeData(wb, sheet = 1, startRow = 1, startCol = 1, x = df, colNames = TRUE)

  deleteDataColumn(wb, sheet =  1, col =  2)

  expect_equal(
    wb$sharedStrings,
    structure(
      list(
        "<si><t xml:space=\"preserve\">Col 1</t></si>",
        "<si><t xml:space=\"preserve\">Col 3</t></si>",
        "<si><t xml:space=\"preserve\">Row 2 Col 1</t></si>",
        "<si><t xml:space=\"preserve\">Row 2 Col 3</t></si>"
      ),
      uniqueCount = 4L
    )
  )
  expect_equal(
    read.xlsx(wb),
    data.frame(
      "Col 1" = "Row 2 Col 1",
      "Col 3" = "Row 2 Col 3"
    )
  )
})


test_that("deleteDataColumn with shared strings on other sheets", {
  df <- data.frame("ABC" = "I am a shared string with sheet 2!")
  df2 <- data.frame("AB" = "I am a shared string with sheet 2!")

  wb <- createWorkbook()
  addWorksheet(wb, "tester")
  writeData(wb, sheet = 1, startRow = 1, startCol = 1, x = df, colNames = TRUE)

  simplify <- function(sd) data.frame(rows = sd$rows, cols = sd$cols, t = sd$t, v = sd$v)
  expect_equal(
    simplify(wb$worksheets[[1]]$sheet_data),
    data.frame(rows = c(1, 2), cols = 1, t = 1, v = c("0", "1"))
  )
  expect_equal(
    wb$sharedStrings,
    structure(
      list(
        "<si><t xml:space=\"preserve\">ABC</t></si>",
        "<si><t xml:space=\"preserve\">I am a shared string with sheet 2!</t></si>"
      ),
      uniqueCount = 2L
    )
  )

  addWorksheet(wb, "tester2")
  writeData(wb, sheet = 2, startRow = 1, startCol = 1, x = df2, colNames = TRUE)

  expect_equal(
    simplify(wb$worksheets[[2]]$sheet_data),
    data.frame(rows = c(1, 2), cols = 1, t = 1, v = c("2", "1"))
  )
  expect_equal(
    wb$sharedStrings,
    structure(
      list(
        "<si><t xml:space=\"preserve\">ABC</t></si>",
        "<si><t xml:space=\"preserve\">I am a shared string with sheet 2!</t></si>",
        "<si><t xml:space=\"preserve\">AB</t></si>"
      ),
      uniqueCount = 3L
    )
  )


  # deleting from sheet 1 does not delete the string from sheet 2!
  deleteDataColumn(wb, sheet =  1, col =  1)

  expect_equal(
    simplify(wb$worksheets[[1]]$sheet_data),
    data.frame(rows = numeric(0), cols = numeric(0), t = numeric(0), v = character(0))
  )

  # note on sheet 2, the indices v to the shared strings have to change as well!
  expect_equal(
    simplify(wb$worksheets[[2]]$sheet_data),
    data.frame(rows = c(1, 2), cols = 1, t = 1, v = c("1", "0"))
  )

  expect_equal(
    wb$sharedStrings,
    structure(
      list(
        "<si><t xml:space=\"preserve\">I am a shared string with sheet 2!</t></si>",
        "<si><t xml:space=\"preserve\">AB</t></si>"
      ),
      uniqueCount = 2L
    )
  )

  expect_equal(
    read.xlsx(wb, sheet = 2),
    data.frame(AB = "I am a shared string with sheet 2!")
  )
})

context("Named Regions")

test_that("Maintaining Named Regions on Load", {
  ## create named regions
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  addWorksheet(wb, "Sheet 2")

  ## specify region
  writeData(wb, sheet = 1, x = iris, startCol = 1, startRow = 1)
  createNamedRegion(
    wb = wb,
    sheet = 1,
    name = "iris",
    rows = seq_len(nrow(iris) + 1),
    cols = seq_len(ncol(iris))
  )

  ## using writeData 'name' argument
  writeData(wb, sheet = 1, x = iris, name = "iris2", startCol = 10)

  ## Named region size 1
  writeData(wb, sheet = 2, x = 99, name = "region1", startCol = 3, startRow = 3)

  ## save file for testing
  out_file <- temp_xlsx()
  saveWorkbook(wb, out_file, overwrite = TRUE)

  expect_equal(object = getNamedRegions(wb), expected = getNamedRegions(out_file))

  df1 <- read.xlsx(wb, namedRegion = "iris")
  df2 <- read.xlsx(out_file, namedRegion = "iris")
  expect_equal(df1, df2)

  df1 <- read.xlsx(wb, namedRegion = "region1")
  expect_s3_class(df1, "data.frame")
  expect_equal(nrow(df1), 0)
  expect_equal(ncol(df1), 1)

  df1 <- read.xlsx(wb, namedRegion = "region1", colNames = FALSE)
  expect_s3_class(df1, "data.frame")
  expect_equal(nrow(df1), 1)
  expect_equal(ncol(df1), 1)

  df1 <- read.xlsx(wb, namedRegion = "region1", rowNames = TRUE)
  expect_s3_class(df1, "data.frame")
  expect_equal(nrow(df1), 0)
  expect_equal(ncol(df1), 0)
})

test_that("Correctly Loading Named Regions Created in Excel", {

  # Load an excel workbook (in the repo, it's located in the /inst folder;
  # when installed on the user's system, it is located in the installation folder
  # of the package)
  filename <- system.file("extdata", "namedRegions.xlsx", package = "openxlsx")

  # Load this workbook. We will test read.xlsx by passing both the object wb and
  # the filename. Both should produce the same results.
  wb <- loadWorkbook(filename)

  # NamedTable refers to Sheet1!$C$5:$D$8
  table_f <- read.xlsx(filename,
    namedRegion = "NamedTable"
  )
  table_w <- read.xlsx(wb,
    namedRegion = "NamedTable"
  )

  expect_equal(object = table_f, expected = table_w)
  expect_equal(object = class(table_f), expected = "data.frame")
  expect_equal(object = ncol(table_f), expected = 2)
  expect_equal(object = nrow(table_f), expected = 3)

  # NamedCell refers to Sheet1!$C$2
  # This proeduced an error in an earlier version of the pacage when the object
  # wb was passed, but worked correctly when the filename was passed to read.xlsx
  cell_f <- read.xlsx(filename,
    namedRegion = "NamedCell",
    colNames = FALSE,
    rowNames = FALSE
  )

  cell_w <- read.xlsx(wb,
    namedRegion = "NamedCell",
    colNames = FALSE,
    rowNames = FALSE
  )

  expect_equal(object = cell_f, expected = cell_w)
  expect_equal(object = class(cell_f), expected = "data.frame")
  expect_equal(object = ncol(cell_f), expected = 1)
  expect_equal(object = nrow(cell_f), expected = 1)

  # NamedCell2 refers to Sheet1!$C$2:$C$2
  cell2_f <- read.xlsx(filename,
    namedRegion = "NamedCell2",
    colNames = FALSE,
    rowNames = FALSE
  )

  cell2_w <- read.xlsx(wb,
    namedRegion = "NamedCell2",
    colNames = FALSE,
    rowNames = FALSE
  )

  expect_equal(object = cell2_f, expected = cell2_w)
  expect_equal(object = class(cell2_f), expected = "data.frame")
  expect_equal(object = ncol(cell2_f), expected = 1)
  expect_equal(object = nrow(cell2_f), expected = 1)
})


test_that("Load names from an Excel file with funky non-region names", {
  filename <- system.file("extdata", "namedRegions2.xlsx", package = "openxlsx")
  wb <- loadWorkbook(filename)
  names <- getNamedRegions(wb)
  sheets <- attr(names, "sheet")
  positions <- attr(names, "position")

  expect_true(length(names) == length(sheets))
  expect_true(length(names) == length(positions))
  expect_equal(
    head(names, 5),
    c("barref", "barref", "fooref", "fooref", "IQ_CH")
  )
  expect_equal(
    sheets,
    c(
      "Sheet with space", "Sheet1", "Sheet with space", "Sheet1",
      rep("", 26)
    )
  )
  expect_equal(positions, c("B4", "B4", "B3", "B3", rep("", 26)))

  names2 <- getNamedRegions(filename)
  expect_equal(names, names2)
})


test_that("Missing rows in named regions", {
  temp_file <- temp_xlsx()

  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")

  ## create region
  writeData(wb, sheet = 1, x = iris[1:11, ], startCol = 1, startRow = 1)
  deleteData(wb, sheet = 1, cols = 1:2, rows = c(6, 6))

  createNamedRegion(
    wb = wb,
    sheet = 1,
    name = "iris",
    rows = 1:(5 + 1),
    cols = 1:2
  )
  createNamedRegion(
    wb = wb,
    sheet = 1,
    name = "iris2",
    rows = 1:(5 + 2),
    cols = 1:2
  )

  ## iris region is rows 1:6 & cols 1:2
  ## iris2 region is rows 1:7 & cols 1:2

  ## row 6 columns 1 & 2 are blank
  expect_equal(getNamedRegions(wb)[1:2], c("iris", "iris2"), ignore.attributes = TRUE)
  expect_equal(attr(getNamedRegions(wb), "sheet"), c("Sheet 1", "Sheet 1"))
  expect_equal(attr(getNamedRegions(wb), "position"), c("A1:B6", "A1:B7"))

  ######################################################################## from Workbook

  ## Skip empty rows
  x <- read.xlsx(xlsxFile = wb, namedRegion = "iris", colNames = TRUE, skipEmptyRows = TRUE)
  expect_equal(dim(x), c(4, 2))

  x <- read.xlsx(xlsxFile = wb, namedRegion = "iris2", colNames = TRUE, skipEmptyRows = TRUE)
  expect_equal(dim(x), c(5, 2))


  ## Keep empty rows
  x <- read.xlsx(xlsxFile = wb, namedRegion = "iris", colNames = TRUE, skipEmptyRows = FALSE)
  expect_equal(dim(x), c(5, 2))

  x <- read.xlsx(xlsxFile = wb, namedRegion = "iris2", colNames = TRUE, skipEmptyRows = FALSE)
  expect_equal(dim(x), c(6, 2))



  ######################################################################## from file
  saveWorkbook(wb, file = temp_file, overwrite = TRUE)

  ## Skip empty rows
  x <- read.xlsx(xlsxFile = temp_file, namedRegion = "iris", colNames = TRUE, skipEmptyRows = TRUE)
  expect_equal(dim(x), c(4, 2))

  x <- read.xlsx(xlsxFile = temp_file, namedRegion = "iris2", colNames = TRUE, skipEmptyRows = TRUE)
  expect_equal(dim(x), c(5, 2))


  ## Keep empty rows
  x <- read.xlsx(xlsxFile = temp_file, namedRegion = "iris", colNames = TRUE, skipEmptyRows = FALSE)
  expect_equal(dim(x), c(5, 2))

  x <- read.xlsx(xlsxFile = temp_file, namedRegion = "iris2", colNames = TRUE, skipEmptyRows = FALSE)
  expect_equal(dim(x), c(6, 2))

  unlink(temp_file)
})





test_that("Missing columns in named regions", {
  temp_file <- temp_xlsx()

  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")

  ## create region
  writeData(wb, sheet = 1, x = iris[1:11, ], startCol = 1, startRow = 1)
  deleteData(wb, sheet = 1, cols = 2, rows = 1:12, gridExpand = TRUE)

  createNamedRegion(
    wb = wb,
    sheet = 1,
    name = "iris",
    rows = 1:5,
    cols = 1:2
  )

  createNamedRegion(
    wb = wb,
    sheet = 1,
    name = "iris2",
    rows = 1:5,
    cols = 1:3
  )

  ## iris region is rows 1:5 & cols 1:2
  ## iris2 region is rows 1:5 & cols 1:3

  ## row 6 columns 1 & 2 are blank
  expect_equal(getNamedRegions(wb)[1:2], c("iris", "iris2"), ignore.attributes = TRUE)
  expect_equal(attr(getNamedRegions(wb), "sheet"), c("Sheet 1", "Sheet 1"))
  expect_equal(attr(getNamedRegions(wb), "position"), c("A1:B5", "A1:C5"))

  ######################################################################## from Workbook

  ## Skip empty cols
  x <- read.xlsx(xlsxFile = wb, namedRegion = "iris", colNames = TRUE, skipEmptyCols = TRUE)
  expect_equal(dim(x), c(4, 1))

  x <- read.xlsx(xlsxFile = wb, namedRegion = "iris2", colNames = TRUE, skipEmptyCols = TRUE)
  expect_equal(dim(x), c(4, 2))


  ## Keep empty cols
  x <- read.xlsx(xlsxFile = wb, namedRegion = "iris", colNames = TRUE, skipEmptyCols = FALSE)
  expect_equal(dim(x), c(4, 1))

  x <- read.xlsx(xlsxFile = wb, namedRegion = "iris2", colNames = TRUE, skipEmptyCols = FALSE)
  expect_equal(dim(x), c(4, 3))



  ######################################################################## from file
  saveWorkbook(wb, file = temp_file, overwrite = TRUE)

  ## Skip empty cols
  x <- read.xlsx(xlsxFile = temp_file, namedRegion = "iris", colNames = TRUE, skipEmptyCols = TRUE)
  expect_equal(dim(x), c(4, 1))

  x <- read.xlsx(xlsxFile = temp_file, namedRegion = "iris2", colNames = TRUE, skipEmptyCols = TRUE)
  expect_equal(dim(x), c(4, 2))


  ## Keep empty cols
  x <- read.xlsx(xlsxFile = temp_file, namedRegion = "iris", colNames = TRUE, skipEmptyCols = FALSE)
  expect_equal(dim(x), c(4, 1))

  x <- read.xlsx(xlsxFile = temp_file, namedRegion = "iris2", colNames = TRUE, skipEmptyCols = FALSE)
  expect_equal(dim(x), c(4, 3))

  unlink(temp_file)
})





test_that("Matching Substrings breaks reading named regions", {
  temp_file <- temp_xlsx()

  wb <- createWorkbook()
  addWorksheet(wb, "table")
  addWorksheet(wb, "table2")

  t1 <- head(iris)
  t1$Species <- as.character(t1$Species)
  t2 <- head(mtcars)

  writeData(wb, sheet = "table", x = t1, name = "t", startCol = 3, startRow = 12)
  writeData(wb, sheet = "table2", x = t2, name = "t2", startCol = 5, startRow = 24, rowNames = TRUE)

  writeData(wb, sheet = "table", x = head(t1, 3), name = "t1", startCol = 9, startRow = 3)
  writeData(wb, sheet = "table2", x = head(t2, 3), name = "t22", startCol = 15, startRow = 12, rowNames = TRUE)

  saveWorkbook(wb, file = temp_file, overwrite = TRUE)

  r1 <- getNamedRegions(wb)
  expect_equal(attr(r1, "sheet"), c("table", "table2", "table", "table2"))
  expect_equal(attr(r1, "position"), c("C12:G18", "E24:P30", "I3:M6", "O12:Z15"))
  expect_equal(r1, c("t", "t2", "t1", "t22"), check.attributes = FALSE)

  r2 <- getNamedRegions(temp_file)
  expect_equal(attr(r2, "sheet"), c("table", "table2", "table", "table2"))
  expect_equal(attr(r1, "position"), c("C12:G18", "E24:P30", "I3:M6", "O12:Z15"))
  expect_equal(r2, c("t", "t2", "t1", "t22"), check.attributes = FALSE)


  ## read file named region
  expect_equal(t1, read.xlsx(xlsxFile = temp_file, namedRegion = "t"))
  expect_equal(t2, read.xlsx(xlsxFile = temp_file, namedRegion = "t2", rowNames = TRUE))
  expect_equal(head(t1, 3), read.xlsx(xlsxFile = temp_file, namedRegion = "t1"))
  expect_equal(head(t2, 3), read.xlsx(xlsxFile = temp_file, namedRegion = "t22", rowNames = TRUE))

  ## read Workbook named region
  expect_equal(t1, read.xlsx(xlsxFile = wb, namedRegion = "t"))
  expect_equal(t2, read.xlsx(xlsxFile = wb, namedRegion = "t2", rowNames = TRUE))
  expect_equal(head(t1, 3), read.xlsx(xlsxFile = wb, namedRegion = "t1"))
  expect_equal(head(t2, 3), read.xlsx(xlsxFile = wb, namedRegion = "t22", rowNames = TRUE))



  unlink(temp_file)
})


test_that("Read namedRegion from specific sheet", {
  
  filename <- system.file("extdata", "namedRegions3.xlsx", package = "openxlsx")
  
  namedR <- "MyRange"
  sheets <- openxlsx::getSheetNames(filename)
  
  # read the correct sheets
  expect_equal(data.frame(X1 = "S1A1", X2 = "S1B1", stringsAsFactors = FALSE), read.xlsx(filename, sheet = "Sheet1", namedRegion = namedR, rowNames = FALSE, colNames = FALSE))
  expect_equal(data.frame(X1 = "S2A1", X2 = "S2B1", stringsAsFactors = FALSE), read.xlsx(filename, sheet = which(sheets %in% "Sheet2"), namedRegion = namedR, rowNames = FALSE, colNames = FALSE))
  expect_equal(data.frame(X1 = "S3A1", X2 = "S3B1", stringsAsFactors = FALSE), read.xlsx(filename, sheet = "Sheet3", namedRegion = namedR, rowNames = FALSE, colNames = FALSE))
  
  # Warning: Workbook has no such named region. (Wrong namedRegion selected.)
  expect_warning(read.xlsx(filename, sheet = "Sheet2", namedRegion = "MyRage", rowNames = FALSE, colNames = FALSE))
  
  # Warning: Workbook has no such named region on this sheet. (Correct namedRegion, but wrong sheet selected.)
  expect_warning(read.xlsx(filename, sheet = "Sheet4", namedRegion = namedR, rowNames = FALSE, colNames = FALSE))
})

test_that("Overwrite and delete named regions", {
  temp_file <- temp_xlsx()
  
  wb <- createWorkbook()
  addWorksheet(wb, "Sheet 1")
  
  ## create region
  writeData(wb, sheet = 1, x = iris[1:11, ], startCol = 1, 
            startRow = 1, name = "iris")

  
 
  init_nr <- getNamedRegions(wb) 
  expect_equal(attr(init_nr, "position"), "A1:E12")
  
  # no overwrite
  expect_error({
    writeData(wb, sheet = 1, x = iris[1:11, ], startCol = 1, 
            startRow = 1, name = "iris")
  })
 
  expect_error({
    createNamedRegion(
      wb = wb,
      sheet = 1,
      name = "iris",
      rows = 1:5,
      cols = 1:2
    )
  })
  
  # overwrite
  createNamedRegion(
    wb = wb,
    sheet = 1,
    name = "iris",
    rows = 1:5,
    cols = 1:2, 
    overwrite = TRUE
  )
  
  # check midification
  modify_nr <- getNamedRegions(wb) 
  expect_equal(attr(modify_nr, "position"), "A1:B5")
  expect_true("iris" %in% modify_nr)
 
  # delete name region
  deleteNamedRegion(wb, "iris")
  expect_false("iris" %in% getNamedRegions(wb))
})

context("Writing and reading returns similar objects")

test_that("Writing then reading returns identical data.frame 1", {
  curr_wd <- getwd()

  ## data
  genDf <- function() {
    set.seed(1)
    data.frame(
      "Date" = Sys.Date() - 0:4,
      "Logical" = c(TRUE, FALSE, TRUE, TRUE, FALSE),
      "Currency" = -2:2,
      "Accounting" = -2:2,
      "hLink" = "https://CRAN.R-project.org/",
      "Percentage" = seq(-1, 1, length.out = 5),
      "TinyNumber" = runif(5) / 1E9, stringsAsFactors = FALSE
    )
  }

  df <- genDf()
  df

  class(df$Currency) <- "currency"
  class(df$Accounting) <- "accounting"
  class(df$hLink) <- "hyperlink"
  class(df$Percentage) <- "percentage"
  class(df$TinyNumber) <- "scientific"

  op <- options()
  options("openxlsx.dateFormat" = "yyyy-mm-dd")

  fileName <- file.path(tempdir(), "allClasses.xlsx")
  write.xlsx(df, file = fileName, overwrite = TRUE)

  x <- read.xlsx(xlsxFile = fileName, detectDates = TRUE)
  expect_equal(object = x, expected = genDf(), check.attributes = FALSE)

  unlink(fileName, recursive = TRUE, force = TRUE)

  expect_equal(object = getwd(), curr_wd)
  options(op)
})

test_that("Writing then reading returns identical data.frame 2", {
  curr_wd <- getwd()

  ## data.frame of dates
  dates <- data.frame("d1" = Sys.Date() - 0:500)
  dates[nrow(dates)+1,] = as.Date("1900-01-02")
  for (i in 1:3) dates <- cbind(dates, dates)
  names(dates) <- paste0("d", 1:8)

  ## Date Formatting
  wb <- createWorkbook()
  addWorksheet(wb, "Date Formatting", gridLines = FALSE)
  writeData(wb, 1, dates) ## write without styling

  ## set default date format
  options("openxlsx.dateFormat" = "yyyy/mm/dd")

  ## numFmt == "DATE" will use the date format specified by the above
  addStyle(wb, 1, style = createStyle(numFmt = "DATE"), rows = 2:11, cols = 1, gridExpand = TRUE)

  ## some custom date format examples
  sty <- createStyle(numFmt = "yyyy/mm/dd")
  addStyle(wb, 1, style = sty, rows = 2:11, cols = 2, gridExpand = TRUE)

  sty <- createStyle(numFmt = "yyyy/mmm/dd")
  addStyle(wb, 1, style = sty, rows = 2:11, cols = 3, gridExpand = TRUE)

  sty <- createStyle(numFmt = "yy / mmmm / dd")
  addStyle(wb, 1, style = sty, rows = 2:11, cols = 4, gridExpand = TRUE)

  sty <- createStyle(numFmt = "ddddd")
  addStyle(wb, 1, style = sty, rows = 2:11, cols = 5, gridExpand = TRUE)

  sty <- createStyle(numFmt = "yyyy-mmm-dd")
  addStyle(wb, 1, style = sty, rows = 2:11, cols = 6, gridExpand = TRUE)

  sty <- createStyle(numFmt = "mm/ dd yyyy")
  addStyle(wb, 1, style = sty, rows = 2:11, cols = 7, gridExpand = TRUE)

  sty <- createStyle(numFmt = "mm/dd/yy")
  addStyle(wb, 1, style = sty, rows = 2:11, cols = 8, gridExpand = TRUE)

  setColWidths(wb, 1, cols = 1:10, widths = 23)


  fileName <- file.path(tempdir(), "DateFormatting.xlsx")
  write.xlsx(dates, file = fileName, overwrite = TRUE)

  x <- read.xlsx(xlsxFile = fileName, detectDates = TRUE)
  expect_equal(object = x, expected = dates, check.attributes = FALSE)


  xNoDateDetection <- read.xlsx(xlsxFile = fileName, detectDates = FALSE)
  dateOrigin <- getDateOrigin(fileName)


  expect_equal(object = dateOrigin, expected = "1900-01-01", check.attributes = FALSE)

  for (i in seq_len(ncol(x))) {
    xNoDateDetection[[i]] <- convertToDate(xNoDateDetection[[i]], origin = dateOrigin)
  }

  expect_equal(object = xNoDateDetection, expected = dates, check.attributes = FALSE)

  expect_equal(object = getwd(), curr_wd)
  unlink(fileName, recursive = TRUE, force = TRUE)
})

test_that("Writing then reading rowNames, colNames combinations", {
  op <- options()
  options(stringsAsFactors = FALSE)
  
  
  fileName <- temp_xlsx()
  curr_wd <- getwd()
  mt <- utils::head(mtcars) # don't need the whole thing

  # write the row and column names for testing
  write.xlsx(mt, file = fileName, overwrite = TRUE, rowNames = TRUE, colNames = TRUE)
  
  # rowNames = colNames = TRUE
  # Row names = first column
  # Col names = first row
  x <- read.xlsx(fileName, sheet = 1, rowNames = TRUE)
  expect_equal(x, mt)


  # rowNames = TRUE, colNames = FALSE
  # Row names = first column
  # Col names = X1, X2, etc
  
  # need to create an expected output
  y <- as.data.frame(rbind(colnames(mt), as.matrix(mt)))
  colnames(y) <- c(make.names(seq_along(mt)))
  x <- read.xlsx(fileName, sheet = 1, rowNames = TRUE, colNames = FALSE)
  expect_equal(x, y)


  # rowNames = FALSE, colNames = TRUE
  # Row names = ""
  # Cl names = first row
  y2 <- cbind(row.names(mt), mt)
  colnames(y2)[1] <- ""
  row.names(y2) <- NULL
  x <- read.xlsx(fileName, sheet = 1, rowNames = FALSE, colNames = TRUE)
  expect_equal(x, y2)
  
  # rowNames = FALSE, colNames = FALSE
  # Row names = ""
  # Col names = X1, X2, etc
  y3 <- cbind(row.names(y), y)
  colnames(y3) <- make.names(seq_along(y3))
  row.names(y3) <- NULL
  x <- read.xlsx(fileName, sheet = 1, rowNames = FALSE, colNames = FALSE)
  expect_equal(x, y3)

  
  # Check wd
  expect_equal(getwd(), curr_wd)
  
  unlink(fileName, recursive = TRUE, force = TRUE)
  options(op)
})


test_that("Writing then reading returns identical data.frame 3", {
  op <- options()
  options(openxlsx.dateFormat = "yyyy-mm-dd")

  ## data
  df <- data.frame(
    Date       = as.Date("2021-05-21") - 0:4,
    Logical    = c(TRUE, FALSE, TRUE, TRUE, FALSE),
    Currency   = -2:2,
    Accounting = -2:2,
    hLink      = "https://CRAN.R-project.org/",
    Percentage = seq.int(-1, 1, length.out = 5),
    TinyNumber = runif(5) / 1E9,
    stringsAsFactors = FALSE
  )

  class(df$Currency) <- "currency"
  class(df$Accounting) <- "accounting"
  class(df$hLink) <- "hyperlink"
  class(df$Percentage) <- "percentage"
  class(df$TinyNumber) <- "scientific"

  fileName <- tempfile("allClasses", fileext = ".xlsx")
  write.xlsx(df, file = fileName, overwrite = TRUE)

  ## rows, cols combinations
  rows <- 1:4
  cols <- c(1, 3, 5)
  x <- read.xlsx(fileName, detectDates = TRUE, rows = rows, cols = cols)
  exp <- df[sort((rows - 1)[(rows - 1) <= nrow(df)]), sort(cols[cols <= ncol(df)])]
  expect_equal(x, exp)

  rows <- 1:4
  cols <- 1:9
  x <- read.xlsx(xlsxFile = fileName, detectDates = TRUE, rows = rows, cols = cols)
  exp <- df[sort((rows - 1)[(rows - 1) <= nrow(df)]), sort(cols[cols <= ncol(df)])]
  expect_equal(x, exp)

  rows <- 1:200
  cols <- c(5, 99, 2)
  x <- read.xlsx(xlsxFile = fileName, detectDates = TRUE, rows = rows, cols = cols)
  exp <- df[sort((rows - 1)[(rows - 1) <= nrow(df)]), sort(cols[cols <= ncol(df)])]
  expect_equal(x, exp)


  rows <- 1000:900
  cols <- c(5, 99, 2)
  suppressWarnings(x <- read.xlsx(xlsxFile = fileName, detectDates = TRUE, rows = rows, cols = cols))
  expect_identical(x, NULL)

  unlink(fileName, recursive = TRUE, force = TRUE)
  options(op)
})


test_that("Writing then reading returns identical data.frame 4", {

  ## data
  df <- head(iris[, 1:4])
  df[1, 2] <- NA
  df[3, 1] <- NA
  df[6, 4] <- NA


  tf <- temp_xlsx()
  write.xlsx(x = df, file = tf, keepNA = TRUE)
  x <- read.xlsx(tf)

  expect_equal(object = x, expected = df, check.attributes = TRUE)
  unlink(tf, recursive = TRUE, force = TRUE)


  tf <- temp_xlsx()
  write.xlsx(x = df, file = tf, keepNA = FALSE)
  x <- read.xlsx(tf)

  expect_equal(object = x, expected = df, check.attributes = TRUE)
  unlink(tf, recursive = TRUE, force = TRUE)
})

test_that("Writing then reading returns identical data.frame 5", {

  ## data
  df <- head(iris[, 1:4])
  df[1, 2] <- NA
  df[3, 1] <- NA
  df[6, 4] <- NA

  na.string <- "*"
  df_expected <- df
  df_expected[1, 2] <- na.string
  df_expected[3, 1] <- na.string
  df_expected[6, 4] <- na.string


  tf <- temp_xlsx()
  write.xlsx(x = df, file = tf, keepNA = TRUE, na.string = na.string)
  x <- read.xlsx(tf)

  expect_equal(object = x, expected = df_expected, check.attributes = TRUE)
  unlink(tf, recursive = TRUE, force = TRUE)
})



test_that("Special characters in sheet names", {
  tf <- temp_xlsx()

  ## data
  sheet_name <- "A & B < D > D"

  wb <- createWorkbook()
  addWorksheet(wb, sheetName = sheet_name)
  addWorksheet(wb, sheetName = "test")
  writeData(wb, sheet = 1, x = 1:10)
  saveWorkbook(wb = wb, file = tf, overwrite = TRUE)

  expect_equal(getSheetNames(tf)[1], sheet_name)
  expect_equal(getSheetNames(tf)[2], "test")

  expect_equal(read.xlsx(tf, colNames = FALSE)[[1]], 1:10)

  unlink(tf, recursive = TRUE, force = TRUE)
})

context("Removing worksheets.")

test_that("Deleting worksheets", {
  tempFile <- temp_xlsx()
  genWS <- function(wb, sheetName) {
    addWorksheet(wb, sheetName)
    writeDataTable(wb, sheetName, data.frame("X" = sprintf("This is sheet: %s", sheetName)), colNames = FALSE)
  }

  wb <- createWorkbook()
  genWS(wb, "Sheet 1")
  genWS(wb, "Sheet 2")
  genWS(wb, "Sheet 3")

  expect_equal(names(wb), c("Sheet 1", "Sheet 2", "Sheet 3"))

  removeWorksheet(wb, sheet = 1)
  expect_equal(names(wb), c("Sheet 2", "Sheet 3"))


  removeWorksheet(wb, sheet = 1)
  expect_equal(names(wb), c("Sheet 3"))


  ## add to end
  genWS(wb, "Sheet 1")
  genWS(wb, "Sheet 2")
  expect_equal(names(wb), c("Sheet 3", "Sheet 1", "Sheet 2"))

  saveWorkbook(wb, tempFile, overwrite = TRUE)

  ## re-load & re-order worksheets

  wb <- loadWorkbook(tempFile)
  expect_equal(names(wb), c("Sheet 3", "Sheet 1", "Sheet 2"))

  writeData(wb, sheet = "Sheet 2", x = iris[1:10, 1:4], startRow = 5)
  expect_equal(iris[1:10, 1:4], read.xlsx(wb, "Sheet 2", startRow = 5))


  writeData(wb, sheet = 1, x = iris[1:20, 1:4], startRow = 5)
  expect_equal(iris[1:20, 1:4], read.xlsx(wb, "Sheet 3", startRow = 5))


  removeWorksheet(wb, sheet = 1)
  expect_equal("This is sheet: Sheet 1", read.xlsx(wb, 1, startRow = 1)[[1]])

  removeWorksheet(wb, sheet = 2)
  expect_equal("This is sheet: Sheet 1", read.xlsx(wb, 1, startRow = 1)[[1]])

  removeWorksheet(wb, sheet = 1)
  expect_equal(names(wb), character(0))



  unlink(tempFile, recursive = TRUE, force = TRUE)
  rm(wb)
})

context("Data validation")

# Basic test workbook ==========================================================

wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")
addWorksheet(wb, "Sheet 2")
writeDataTable(wb, sheet = 1, x = iris[1:30, ])
writeData(wb, sheet = 2, x = sample(iris$Sepal.Length, 10))

# Unit tests ===================================================================

test_that("Data validation for lists is performed without warnings", {
  expect_invisible(
    dataValidation(
      wb, 1, col = 1, rows = 2:31, type = "list", value = "'Sheet 2'!$A$1:$A$10"
    )
  )
})

context("Read Sources")



test_that("read.xlsx from different sources", {
  skip_if_offline()

  ## URL
  xlsxFile <- "https://github.com/ycphs/openxlsx/raw/master/inst/extdata/readTest.xlsx"
  df_url <- read.xlsx(xlsxFile)

  ## File
  xlsxFile <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
  df_file <- read.xlsx(xlsxFile)

  expect_true(all.equal(df_url, df_file), label = "Read from URL")


  ## Non-existing URL
  xlsxFile <- "https://github.com/ycphs/openxlsx/raw/master/inst/extdata/readTest2.xlsx"
  expect_error(suppressWarnings(read.xlsx(xlsxFile)))


  ## Non-existing File
  xlsxFile <- file.path(dirname(system.file("extdata", "readTest.xlsx", package = "openxlsx")), "readTest00.xlsx")
  expect_error(read.xlsx(xlsxFile), regexp = "File does not exist.")
})




test_that("loadWorkbook from different sources", {
  skip_if_offline()

  ## URL
  xlsxFile <- "https://github.com/ycphs/openxlsx/raw/master/inst/extdata/readTest.xlsx"
  wb_url <- loadWorkbook(xlsxFile)

  ## File
  xlsxFile <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
  wb_file <- loadWorkbook(xlsxFile)

  ## check
  expect_true(all.equal.Workbook(wb_url, wb_file), "Loading from URL vs local not equal")
})



test_that("getDateOrigin from different sources", {
  skip_if_offline()

  ## URL
  xlsxFile <- "https://github.com/ycphs/openxlsx/raw/master/inst/extdata/readTest.xlsx"
  origin_url <- getDateOrigin(xlsxFile)

  ## File
  xlsxFile <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
  origin_file <- getDateOrigin(xlsxFile)

  ## check
  expect_equal(origin_url, origin_file)
  expect_equal(origin_url, "1900-01-01")
})

context("Writing over tables")



test_that("writeDataTable over tables", {
  overwrite_table_error <- "Cannot overwrite existing table with another table"
  df1 <- data.frame("X" = 1:10)

  wb <- createWorkbook()
  addWorksheet(wb, "Sheet1")

  ## table covers rows 4->10 and cols 4->8
  writeDataTable(wb = wb, sheet = 1, x = head(iris), startCol = 4, startRow = 4)

  ## should all run without error
  writeDataTable(wb = wb, sheet = 1, x = df1, startCol = 3, startRow = 2)
  writeDataTable(wb = wb, sheet = 1, x = df1, startCol = 9, startRow = 2)
  writeDataTable(wb = wb, sheet = 1, x = df1, startCol = 4, startRow = 11)
  writeDataTable(wb = wb, sheet = 1, x = df1, startCol = 5, startRow = 11)
  writeDataTable(wb = wb, sheet = 1, x = df1, startCol = 6, startRow = 11)
  writeDataTable(wb = wb, sheet = 1, x = df1, startCol = 7, startRow = 11)
  writeDataTable(wb = wb, sheet = 1, x = df1, startCol = 8, startRow = 11)
  writeDataTable(wb = wb, sheet = 1, x = head(iris, 2), startCol = 4, startRow = 1)



  ## Now error
  expect_error(writeDataTable(wb = wb, sheet = 1, x = df1, startCol = "H", startRow = 21), regexp = overwrite_table_error)
  expect_error(writeDataTable(wb = wb, sheet = 1, x = df1, startCol = 3, startRow = 12), regexp = overwrite_table_error)
  expect_error(writeDataTable(wb = wb, sheet = 1, x = df1, startCol = 9, startRow = 12), regexp = overwrite_table_error)
  expect_error(writeDataTable(wb = wb, sheet = 1, x = df1, startCol = "i", startRow = 12), regexp = overwrite_table_error)


  ## more errors
  expect_error(writeDataTable(wb = wb, sheet = 1, x = head(iris)), regexp = overwrite_table_error)
  expect_error(writeDataTable(wb = wb, sheet = 1, x = head(iris), startCol = 4, startRow = 21), regexp = overwrite_table_error)

  ## should work
  writeDataTable(wb = wb, sheet = 1, x = head(iris), startCol = 4, startRow = 22)
  writeDataTable(wb = wb, sheet = 1, x = head(iris), startCol = 4, startRow = 40)


  ## more errors
  expect_error(writeDataTable(wb = wb, sheet = 1, x = head(iris, 2), startCol = 4, startRow = 38), regexp = overwrite_table_error)
  expect_error(writeDataTable(wb = wb, sheet = 1, x = head(iris, 2), startCol = 4, startRow = 38, colNames = FALSE), regexp = overwrite_table_error)

  expect_error(writeDataTable(wb = wb, sheet = 1, x = head(iris), startCol = "H", startRow = 40), regexp = overwrite_table_error)
  writeDataTable(wb = wb, sheet = 1, x = head(iris), startCol = "I", startRow = 40)
  writeDataTable(wb = wb, sheet = 1, x = head(iris)[, 1:3], startCol = "A", startRow = 40)

  expect_error(writeDataTable(wb = wb, sheet = 1, x = head(iris, 2), startCol = 4, startRow = 38, colNames = FALSE), regexp = overwrite_table_error)
  expect_error(writeDataTable(wb = wb, sheet = 1, x = head(iris, 2), startCol = 1, startRow = 46, colNames = FALSE), regexp = overwrite_table_error)
})





test_that("writeData over tables", {
  overwrite_table_error <- "Cannot overwrite table headers. Avoid writing over the header row"
  df1 <- data.frame("X" = 1:10)

  wb <- createWorkbook()
  addWorksheet(wb, "Sheet1")

  ## table covers rows 4->10 and cols 4->8
  writeDataTable(wb = wb, sheet = 1, x = head(iris), startCol = 4, startRow = 4)

  ## Anywhere on row 5 is fine
  for (i in 1:10) {
    writeData(wb = wb, sheet = 1, x = head(iris), startRow = 5, startCol = i)
  }

  ## Anywhere on col i is fine
  for (i in 1:10) {
    writeData(wb = wb, sheet = 1, x = head(iris), startRow = i, startCol = "i")
  }



  ## Now errors on headers
  expect_error(writeData(wb = wb, sheet = 1, x = head(iris), startCol = 4, startRow = 4), regexp = overwrite_table_error)
  writeData(wb = wb, sheet = 1, x = head(iris), startCol = 4, startRow = 5)
  writeData(wb = wb, sheet = 1, x = head(iris)[1:3])
  writeData(wb = wb, sheet = 1, x = head(iris, 2), startCol = 4)
  writeData(wb = wb, sheet = 1, x = head(iris, 2), startCol = 4, colNames = FALSE)


  ## Example of how this should be used
  writeDataTable(wb = wb, sheet = 1, x = head(iris), startCol = 4, startRow = 30)
  writeData(wb = wb, sheet = 1, x = head(iris), startCol = 4, startRow = 31, colNames = FALSE)

  writeDataTable(wb = wb, sheet = 1, x = head(iris), startCol = 10, startRow = 30)
  writeData(wb = wb, sheet = 1, x = tail(iris), startCol = 10, startRow = 31, colNames = FALSE)

  writeDataTable(wb = wb, sheet = 1, x = head(iris)[, 1:3], startCol = 1, startRow = 30)
  writeData(wb = wb, sheet = 1, x = tail(iris), startCol = 1, startRow = 31, colNames = FALSE)
})

context("Date/Time Conversions")

test_that("as_POSIXct_utc", {
  exp <- "2022-03-02 19:27:35"
  got <- as_POSIXct_utc("2022-03-02 19:27:35")
  expect_equal(exp, as.character(got))
})

test_that("convert to date", {
  dates <- as.Date("2015-02-07") + -10:10
  origin <- 25569L
  n <- as.integer(dates) + origin

  expect_equal(convertToDate(n), dates)

  earlyDate <- as.Date("1900-01-03")
  serialDate <- 3
  expect_equal(convertToDate(serialDate), earlyDate)

})


test_that("convert to datetime", {
  x <- 43037 + 2 / 1440
  expect_equal(object = convertToDateTime(x, tx = Sys.timezone()), expected = as.POSIXct("2017-10-29 00:02:00", tz = Sys.timezone()))

  x <- 43037 + 2 / 1440 + 1 / 86400
  expect_equal(object = convertToDateTime(x, tx = Sys.timezone()), expected = as.POSIXct("2017-10-29 00:02:01", tz = Sys.timezone()))

  x <- 43037 + 2.50 / 1440
  expect_equal(object = convertToDateTime(x, tx = Sys.timezone()), expected = as.POSIXct("2017-10-29 00:02:30", tz = Sys.timezone()))

  x <- 43037 + 2 / 1440 + 12.12 / 86400
  x_datetime <- convertToDateTime(x, tx = "UTC")
  attr(x_datetime, "tzone") <- "UTC"
})


test_that("read.xlsx detectDates", {
  
  xlsxFile <- system.file("extdata", "gh_issue_288.xlsx", package = "openxlsx")
  
  ref_dat <- data.frame(Date = c(as.Date(c("2021-10-20", "2021-11-03"))))
  ref_num <- data.frame(Date = c(44489.4, 44503.0))
  
  tst_dat <- read.xlsx(xlsxFile, detectDates = TRUE)
  tst_num <- read.xlsx(xlsxFile, detectDates = FALSE)
  
  expect_equal(ref_dat, tst_dat)
  expect_equal(ref_num, tst_num)
  
})

#' @export
read.xlsx.Workbook <- function(xlsxFile,
                               sheet = 1,
                               startRow = 1,
                               colNames = TRUE,
                               rowNames = FALSE,
                               detectDates = FALSE,
                               skipEmptyRows = TRUE,
                               skipEmptyCols = TRUE,
                               rows = NULL,
                               cols = NULL,
                               check.names = FALSE,
                               sep.names = ".",
                               namedRegion = NULL,
                               na.strings = "NA",
                               fillMergedCells = FALSE) {


  ## Validate inputs and get files
  if (!is.logical(colNames)) {
    stop("colNames must be TRUE/FALSE.")
  }

  if (!is.logical(rowNames)) {
    stop("rowNames must be TRUE/FALSE.")
  }

  if (!is.logical(detectDates)) {
    stop("detectDates must be TRUE/FALSE.")
  }

  if (!is.logical(skipEmptyRows)) {
    stop("skipEmptyRows must be TRUE/FALSE.")
  }

  if (!is.logical(check.names)) {
    stop("check.names must be TRUE/FALSE.")
  }

  if (!is.character(sep.names) || nchar(sep.names) != 1) {
    stop("sep.names must be a character and only one.")
  }

  if (length(sheet) != 1) {
    stop("sheet must be of length 1.")
  }

  ## Named region logic
  reading_named_region <- FALSE
  if (!is.null(namedRegion)) {
    dn <- xlsxFile$workbook$definedNames
    if (length(dn) == 0) {
      warning("Workbook has no named regions.")
      return(NULL)
    }

    dn_names <- replaceXMLEntities(regmatches(dn, regexpr('(?<=name=")[^"]+', dn, perl = TRUE)))
    ind <- tolower(dn_names) == tolower(namedRegion)
    if (!any(ind)) {
      stop(sprintf("Region '%s' not found!", namedRegion))
    }

    ## pull out first node value
    dn <- dn[ind]
    region <- regmatches(dn, regexpr("(?<=>)[^\\<]+", dn, perl = TRUE))
    sheet <- names(xlsxFile)[sapply(names(xlsxFile), function(x) grepl(x, dn))]
    if (length(sheet) > 1) {
      sheet <- sheet[which.max(nchar(sheet))]
    }

    region <- gsub("[^A-Z0-9:]", "", gsub(sheet, "", region, fixed = TRUE))

    if (grepl(":", region, fixed = TRUE)) {
      cols <- unlist(lapply(strsplit(region, split = ":", fixed = TRUE), convertFromExcelRef))
      rows <- unlist(lapply(strsplit(region, split = ":", fixed = TRUE), function(x) as.integer(gsub("[A-Z]", "", x))))

      cols <- seq(from = cols[1], to = cols[2], by = 1)
      rows <- seq(from = rows[1], to = rows[2], by = 1)
    } else {
      cols <- convertFromExcelRef(region)
      rows <- as.integer(gsub("[A-Z]", "", region, perl = TRUE))
    }
    startRow <- 1
    reading_named_region <- TRUE
    named_region_rows <- rows
  }


  if (is.null(rows)) {
    rows <- NA
  } else if (length(rows) > 1) {
    rows <- as.integer(sort(rows))
  }

  ## check startRow
  if (!is.null(startRow)) {
    if (length(startRow) > 1) {
      stop("startRow must have length 1.")
    }
  }



  ## create temp dir and unzip
  nSheets <- length(xlsxFile$worksheets)
  if (nSheets == 0) {
    stop("Workbook has no worksheets")
  }

  ## get workbook names
  sheetNames <- xlsxFile$sheet_names

  if ("character" %in% class(sheet)) {
    sheetNames <- replaceXMLEntities(sheetNames)
    if (!sheet %in% sheetNames) {
      stop(sprintf('Cannot find sheet named "%s"', sheet))
    }
    sheet <- which(sheetNames == sheet)
  } else {
    sheet <- sheet
    if (sheet > nSheets) {
      stop(sprintf("sheet %s does not exist.", sheet))
    }
  }


  ## read in sharedStrings
  sharedStrings <- paste(unlist(xlsxFile$sharedStrings), collapse = "\n")
  if (length(sharedStrings) > 0) {
    sharedStrings <- getSharedStringsFromFile(sharedStringsFile = sharedStrings, isFile = FALSE)
    if (!is.null(na.strings)) {
      sharedStrings[sharedStrings %in% na.strings] <- NA
    }
  }

  ## read in worksheet and get cells with a value node, skip emptyStrs cells
  xlsxFile$worksheets[[sheet]]$order_sheetdata()
  sheet_data <- xlsxFile$worksheets[[sheet]]$sheet_data



  ######################################################
  ## What data to read


  keep <- rep.int(TRUE, length(sheet_data$rows))
  if (!is.na(rows[1])) {
    keep <- keep & (sheet_data$rows %in% rows)
  }

  if (!is.null(cols[1])) {
    keep <- keep & (sheet_data$cols %in% cols)
  }

  if (startRow > 1) {
    keep <- keep & (sheet_data$rows >= startRow)
  }

  ## error cells
  keep <- keep & (sheet_data$t != 4 & !is.na(sheet_data$t) & !is.na(sheet_data$v)) ## "e" or missing
  if (any(is.na(sharedStrings))) {
    keep[(sheet_data$t %in% 1 & (sheet_data$v %in% as.character(which(is.na(sharedStrings)) - 1L)))] <- FALSE
  }

  ## End what data to read
  ######################################################



  rows <- sheet_data$rows[keep]
  cols <- sheet_data$cols[keep]
  v <- sheet_data$v[keep]
  t <- sheet_data$t[keep]

  if (length(v) == 0) {
    warning("No data found on worksheet.", call. = FALSE)
    return(NULL)
  }


  if (is.null(rows)) {
    warning("No data found on worksheet.", call. = FALSE)
    return(NULL)
  } else {
    if (skipEmptyRows) {
      nRows <- length(unique(rows))
    } else if (reading_named_region) {
      nRows <- max(named_region_rows) - min(named_region_rows) + 1
    } else {
      nRows <- max(rows) - min(rows) + 1
    }
  }

  ## get references for string cells
  string_refs <- which(t == 2 | t == 1) ## "b" or "s"
  if (length(string_refs) == 0) {
    string_refs <- -1L
  }


  ## get Refs for boolean
  bool_refs <- which(t == 2) ## "b"
  if (length(bool_refs) == 0) {
    bool_refs <- -1L
  }

  if (bool_refs[1] != -1L) {
    false_ind <- which(sharedStrings == "FALSE") - 1L
    if (length(false_ind) == 0) {
      false_ind <- length(sharedStrings)
      sharedStrings <- c(sharedStrings, "FALSE")
    }

    true_ind <- which(sharedStrings == "TRUE") - 1L
    if (length(true_ind) == 0) {
      true_ind <- length(sharedStrings)
      sharedStrings <- c(sharedStrings, "TRUE")
    }

    logical_vals <- v[bool_refs]
    logical_vals[logical_vals == "0"] <- false_ind[1]
    logical_vals[logical_vals == "1"] <- true_ind[1]
    v[bool_refs] <- logical_vals

    rm(logical_vals)
    rm(bool_refs)
  }


  ## If any t="str" exist, add v to sharedStrings and replace v with newSharedStringsInd
  str_inds <- which(t %in% c(3, 5)) ## "str" or "inlineStr"
  if (length(str_inds) > 0) {
    unique_strs <- unique(v[str_inds])
    unique_strs[unique_strs == "#N/A"] <- NA

    ## Match references of "str" cells to r
    new_shared_string_inds <- length(sharedStrings):(length(sharedStrings) + length(unique_strs) - 1L)

    ## replace strings in v with reference to sharedStrings, (now can convert v to numeric)
    v[str_inds] <- new_shared_string_inds[match(v[str_inds], unique_strs)]

    ## append new strings to sharedStrings
    sharedStrings <- c(sharedStrings, unique_strs)
    if (string_refs[1] == -1L) {
      string_refs <- str_inds
    } else {
      string_refs <- sort(c(string_refs, str_inds))
    }
  }

  ## Now safe to convert v to numeric
  vn <- as.numeric(v)

  ## Using -1 as a flag for no strings
  if (length(sharedStrings) == 0 || string_refs[1] == -1L) {
    string_refs <- as.integer(NA)
  } else {

    ## set encoding of sharedStrings &  replace values in v with string values
    Encoding(sharedStrings) <- "UTF-8"
    v[string_refs] <- sharedStrings[vn[string_refs] + 1L]

    ## any NA sharedStrings - remove
    v_na <- which(is.na(v))
    if (length(v_na) > 0) {
      string_refs <- setdiff(string_refs, v_na)
    }
  }


  ## date detection
  origin <- 25569L
  isDate <- as.logical(NA)

  if (detectDates) {

    ## get date origin
    if (length(xlsxFile$workbook$workbookPr) > 0) {
      if (grepl('date1904="1"|date1904="true"', xlsxFile$workbook$workbookPr, ignore.case = TRUE)) {
        origin <- 24107L
      }
    }

    sO <- xlsxFile$styleObjects
    sO <- sO[unlist(lapply(sO, "[[", "sheet")) == sheetNames[sheet]]


    styles <- lapply(sO, function(x) {
      fc <- x[["style"]][["numFmt"]]$formatCode
      if (is.null(fc)) {
        fc <- x[["style"]][["numFmt"]]$numFmtId
      }
      fc
    })



    sO <- sO[sapply(styles, length) > 0]
    format_codes <- unlist(lapply(sO, function(x) {
      fc <- x[["style"]][["numFmt"]]$formatCode
      if (is.null(fc)) {
        fc <- x[["style"]][["numFmt"]]$numFmtId
      }
      fc
    }))


    # dateIds <- NULL variable not used
    if (length(format_codes) > 0) {

      ## this regex defines what "looks" like a date
      format_codes <- gsub(".*(?<=\\])|@", "", format_codes, perl = TRUE)
      sO <- sO[(!grepl("[^mdyhsapAMP[:punct:] ]", format_codes) & nchar(format_codes > 3)) | format_codes == 14]
    }

    if (length(sO) > 0) {
      style_rows <- unlist(lapply(sO, "[[", "rows"))
      style_cols <- unlist(lapply(sO, "[[", "cols"))
      isDate <- pair_rc(rows, cols) %in% pair_rc(style_rows, style_cols)

      ## check numbers are also integers
      not_an_integer <- suppressWarnings(as.numeric(v[isDate]))
      not_an_integer <- (not_an_integer %% 1L != 0) | is.na(not_an_integer)
      isDate[not_an_integer] <- FALSE

      ## perform int to date to character convertsion (way too slow)
      v[isDate] <- format(as.Date(as.integer(v[isDate]) - origin, origin = "1970-01-01"), "%Y-%m-%d")
    }
  } ## end of detectDates


  ## Build data.frame
  m <- read_workbook(
    cols_in = cols,
    rows_in = rows,
    v = v,
    string_inds = string_refs,
    is_date = isDate,
    hasColNames = colNames,
    hasSepNames = sep.names,
    skipEmptyRows = skipEmptyRows,
    skipEmptyCols = skipEmptyCols,
    nRows = nRows,
    clean_names = clean_names
  )

  if (colNames && check.names) {
    colnames(m) <- make.names(colnames(m), unique = TRUE)
  }

  if (rowNames) {
    rownames(m) <- m[[1]]
    m[[1]] <- NULL
  }

  return(m)
}

#' @name makeHyperlinkString
#' @title create Excel hyperlink string
#' @description Wrapper to create internal hyperlink string to pass to writeFormula(). Either link to external urls or local files or straight to cells of local Excel sheets.
#' @param sheet Name of a worksheet
#' @param row integer row number for hyperlink to link to
#' @param col column number of letter for hyperlink to link to
#' @param text display text
#' @param file Excel file name to point to. If NULL hyperlink is internal.
#' @seealso [writeFormula()]
#' @export makeHyperlinkString
#' @examples
#'
#' ## Writing internal hyperlinks
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet1")
#' addWorksheet(wb, "Sheet2")
#' addWorksheet(wb, "Sheet 3")
#' writeData(wb, sheet = 3, x = iris)
#'
#' ## External Hyperlink
#' x <- c("https://www.google.com", "https://www.google.com.au")
#' names(x) <- c("google", "google Aus")
#' class(x) <- "hyperlink"
#'
#' writeData(wb, sheet = 1, x = x, startCol = 10)
#'
#'
#' ## Internal Hyperlink - create hyperlink formula manually
#' writeFormula(
#'   wb, "Sheet1",
#'   x = '=HYPERLINK(\"#Sheet2!B3\", "Text to Display - Link to Sheet2")',
#'   startCol = 3
#' )
#'
#' ## Internal - No text to display using makeHyperlinkString() function
#' writeFormula(
#'   wb, "Sheet1",
#'   startRow = 1,
#'   x = makeHyperlinkString(sheet = "Sheet 3", row = 1, col = 2)
#' )
#'
#' ## Internal - Text to display
#' writeFormula(
#'   wb, "Sheet1",
#'   startRow = 2,
#'   x = makeHyperlinkString(
#'     sheet = "Sheet 3", row = 1, col = 2,
#'     text = "Link to Sheet 3"
#'   )
#' )
#'
#' ## Link to file - No text to display
#' writeFormula(
#'   wb, "Sheet1",
#'   startRow = 4,
#'   x = makeHyperlinkString(
#'     sheet = "testing", row = 3, col = 10,
#'     file = system.file("extdata", "loadExample.xlsx", package = "openxlsx")
#'   )
#' )
#'
#' ## Link to file - Text to display
#' writeFormula(
#'   wb, "Sheet1",
#'   startRow = 3,
#'   x = makeHyperlinkString(
#'     sheet = "testing", row = 3, col = 10,
#'     file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"),
#'     text = "Link to File."
#'   )
#' )
#'
#' ## Link to external file - Text to display
#' writeFormula(
#'   wb, "Sheet1",
#'   startRow = 10, startCol = 1,
#'   x = '=HYPERLINK("[C:/Users]", "Link to an external file")'
#' )
#'
#' ## Link to internal file
#' x = makeHyperlinkString(text = "test.png", file = "D:/somepath/somepicture.png")
#' writeFormula(wb, "Sheet1", startRow = 11, startCol = 1, x = x)
#'
#' \dontrun{
#' saveWorkbook(wb, "internalHyperlinks.xlsx", overwrite = TRUE)
#' }
#'
makeHyperlinkString <- function(sheet, row = 1, col = 1, text = NULL, file = NULL) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)
  
  if (missing(sheet)) {
    if (!missing(row) || !missing(col)) warning("Option for col and/or row found, but no sheet was provided.")
    
    if (is.null(text))
      str <- sprintf("=HYPERLINK(\"%s\")", file)
    
    if (is.null(file))
      str <- sprintf("=HYPERLINK(\"%s\")", text)
    
    if (!is.null(text) && !is.null(file))
      str <- sprintf("=HYPERLINK(\"%s\", \"%s\")", file, text)
  } else {
    cell <- paste0(int2col(col), row)
    if (!is.null(file)) {
      dest <- sprintf('"[%s]%s!%s"', file, sheet, cell)
    } else {
      dest <- sprintf('"#\'%s\'!%s"', sheet, cell)
    }
    
    if (is.null(text)) {
      str <- sprintf('=HYPERLINK(%s)', dest)
    } else {
      str <- sprintf('=HYPERLINK(%s, \"%s\")', dest, text)
    }
  }
  
  return(str)
}


getRId <- function(x) {
  regmatches(x, gregexpr('(?<= r:id=")[0-9A-Za-z]+', x, perl = TRUE))
}

getId <- function(x) {
  regmatches(x, gregexpr('(?<= Id=")[0-9A-Za-z]+', x, perl = TRUE))
}



## creates style object based on column classes
## Used in writeData for styling when no borders and writeData table for all column-class based styling
classStyles <- function(wb, sheet, startRow, startCol, colNames, nRow, colClasses, stack = TRUE) {
  sheet <- wb$validateSheet(sheet)
  allColClasses <- unlist(colClasses, use.names = FALSE)
  rowInds <- (1 + startRow + colNames - 1L):(nRow + startRow + colNames - 1L)
  startCol <- startCol - 1L
  
  newStylesElements <- NULL
  names(colClasses) <- NULL
  
  # For custom number formats, ensure unique IDs (extract the current maximum and add 1 for each new format)
  maxnumFmtId <- max(unlist(sapply(wb$styleObjects, function(i) {
    as.integer(
      max(c(i$style$numFmt$numFmtId, 0))
    )
  })), 165)

  
  if ("hyperlink" %in% allColClasses) {
    
    ## style hyperlinks
    inds <- which(sapply(colClasses, function(x) "hyperlink" %in% x))
    
    hyperlinkstyle <- createStyle(textDecoration = "underline")
    hyperlinkstyle$fontColour <- list("theme" = "10")
    styleElements <- list(
      "style" = hyperlinkstyle,
      "sheet" = wb$sheet_names[sheet],
      "rows" = rep.int(rowInds, times = length(inds)),
      "cols" = rep(inds + startCol, each = length(rowInds))
    )
    
    newStylesElements <- append(newStylesElements, list(styleElements))
  }
  
  if ("date" %in% allColClasses) {
    
    ## style dates
    inds <- which(sapply(colClasses, function(x) "date" %in% x))
    
    # make sure the style has a unique ID:
    style = createStyle(numFmt = "date")
    if (style$numFmt$numFmtId == 165) {
      style$numFmt$numFmtId <- maxnumFmtId + 1
      maxnumFmtId <- style$numFmt$numFmtId
    }

    styleElements <- list(
      "style" = style,
      "sheet" = wb$sheet_names[sheet],
      "rows" = rep.int(rowInds, times = length(inds)),
      "cols" = rep(inds + startCol, each = length(rowInds))
    )
    
    newStylesElements <- append(newStylesElements, list(styleElements))
  }
  
  if (any(c("posixlt", "posixct", "posixt") %in% allColClasses)) {
    
    ## style POSIX
    inds <- which(sapply(colClasses, function(x) any(c("posixct", "posixt", "posixlt") %in% x)))
    
    # make sure the style has a unique ID:
    style = createStyle(numFmt = "LONGDATE")
    if (style$numFmt$numFmtId == 165) {
      style$numFmt$numFmtId <- maxnumFmtId + 1
      maxnumFmtId <- style$numFmt$numFmtId
    }

    styleElements <- list(
      "style" = style,
      "sheet" = wb$sheet_names[sheet],
      "rows" = rep.int(rowInds, times = length(inds)),
      "cols" = rep(inds + startCol, each = length(rowInds))
    )
    
    newStylesElements <- append(newStylesElements, list(styleElements))
  }
  
  
  ## style currency as CURRENCY
  if ("currency" %in% allColClasses) {
    inds <- which(sapply(colClasses, function(x) "currency" %in% x))
    
    styleElements <- list(
      "style" = createStyle(numFmt = "CURRENCY"),
      "sheet" = wb$sheet_names[sheet],
      "rows" = rep.int(rowInds, times = length(inds)),
      "cols" = rep(inds + startCol, each = length(rowInds))
    )
    
    newStylesElements <- append(newStylesElements, list(styleElements))
  }
  
  ## style accounting as ACCOUNTING
  if ("accounting" %in% allColClasses) {
    inds <- which(sapply(colClasses, function(x) "accounting" %in% x))
    
    styleElements <- list(
      "style" = createStyle(numFmt = "ACCOUNTING"),
      "sheet" = wb$sheet_names[sheet],
      "rows" = rep.int(rowInds, times = length(inds)),
      "cols" = rep(inds + startCol, each = length(rowInds))
    )
    
    newStylesElements <- append(newStylesElements, list(styleElements))
  }
  
  ## style percentages
  if ("percentage" %in% allColClasses) {
    inds <- which(sapply(colClasses, function(x) "percentage" %in% x))
    
    styleElements <- list(
      "style" = createStyle(numFmt = "percentage"),
      "sheet" = wb$sheet_names[sheet],
      "rows" = rep.int(rowInds, times = length(inds)),
      "cols" = rep(inds + startCol, each = length(rowInds))
    )
    
    newStylesElements <- append(newStylesElements, list(styleElements))
  }
  
  ## style big mark
  if ("scientific" %in% allColClasses) {
    inds <- which(sapply(colClasses, function(x) "scientific" %in% x))
    
    styleElements <- list(
      "style" = createStyle(numFmt = "scientific"),
      "sheet" = wb$sheet_names[sheet],
      "rows" = rep.int(rowInds, times = length(inds)),
      "cols" = rep(inds + startCol, each = length(rowInds))
    )
    
    newStylesElements <- append(newStylesElements, list(styleElements))
  }
  
  ## style big mark
  if ("3" %in% allColClasses || "comma" %in% allColClasses) {
    inds <- which(sapply(colClasses, function(x) "3" %in% tolower(x) | "comma" %in% tolower(x)))
    
    styleElements <- list(
      "style" = createStyle(numFmt = "3"),
      "sheet" = wb$sheet_names[sheet],
      "rows" = rep.int(rowInds, times = length(inds)),
      "cols" = rep(inds + startCol, each = length(rowInds))
    )
    
    newStylesElements <- append(newStylesElements, list(styleElements))
  }
  
  ## numeric sigfigs (Col must be numeric and numFmt options must only have 0s and \\.)
  if ("numeric" %in% allColClasses && !grepl("[^0\\.,#\\$\\* %]", getOption("openxlsx.numFmt", "GENERAL"))) {
    inds <- which(sapply(colClasses, function(x) "numeric" %in% tolower(x)))
    
    styleElements <- list(
      "style" = createStyle(numFmt = getOption("openxlsx.numFmt", "0")),
      "sheet" = wb$sheet_names[sheet],
      "rows" = rep.int(rowInds, times = length(inds)),
      "cols" = rep(inds + startCol, each = length(rowInds))
    )
    
    newStylesElements <- append(newStylesElements, list(styleElements))
  }
  
  
  if (!is.null(newStylesElements)) {
    if (stack) {
      for (i in seq_along(newStylesElements)) {
        wb$addStyle(
          sheet = sheet,
          style = newStylesElements[[i]]$style,
          rows = newStylesElements[[i]]$rows,
          cols = newStylesElements[[i]]$cols, stack = TRUE
        )
      }
    } else {
      wb$styleObjects <- append(wb$styleObjects, newStylesElements)
    }
  }
  
  
  
  invisible(1)
}


#' @name validateColour
#' @description validate the colour input
#' @param colour colour
#' @param errorMsg Error message
#' @author Philipp Schauberger
#' @importFrom grDevices colours
#' @keywords internal
#' @noRd
validateColour <- function(colour, errorMsg = "Invalid colour!") {
  
  ## check if
  if (is.null(colour)) {
    colour <- "black"
  }
  
  validColours <- colours()
  
  if (any(colour %in% validColours)) {
    colour[colour %in% validColours] <- col2hex(colour[colour %in% validColours])
  }
  
  if (any(!grepl("^#[A-Fa-f0-9]{6}$", colour))) {
    stop(errorMsg, call. = FALSE)
  }
  
  colour <- gsub("^#", "FF", toupper(colour))
  
  return(colour)
}

#' @name col2hex
#' @description convert rgb to hex
#' @param creator my.col
#' @author Philipp Schauberger
#' @importFrom grDevices col2rgb rgb
#' @keywords internal
#' @noRd
col2hex <- function(my.col) {
  rgb(t(col2rgb(my.col)), maxColorValue = 255)
}


## header and footer replacements
headerFooterSub <- function(x) {
  if (!is.null(x)) {
    x <- replaceIllegalCharacters(x)
    x <- gsub("\\[Page\\]", "P", x)
    x <- gsub("\\[Pages\\]", "N", x)
    x <- gsub("\\[Date\\]", "D", x)
    x <- gsub("\\[Time\\]", "T", x)
    x <- gsub("\\[Path\\]", "Z", x)
    x <- gsub("\\[File\\]", "F", x)
    x <- gsub("\\[Tab\\]", "A", x)
  }
  
  return(x)
}


writeCommentXML <- function(comment_list, file_name) {
  authors <- unique(sapply(comment_list, "[[", "author"))
  xml <- '<comments xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" mc:Ignorable="xr" xmlns:xr="http://schemas.microsoft.com/office/spreadsheetml/2014/revision">'
  xml <- c(xml, paste0("<authors>", paste(sprintf("<author>%s</author>", authors), collapse = ""), "</authors><commentList>"))
  
  for (i in seq_along(comment_list)) {
    authorInd <- which(authors == comment_list[[i]]$author) - 1L
    xml <- c(xml, sprintf('<comment ref="%s" authorId="%s" shapeId="0"><text>', comment_list[[i]]$ref, authorInd))
    
    if (length(comment_list[[i]]$style) != 0) {
      ## check that style information is present
      for (j in seq_along(comment_list[[i]]$comment)) {
        xml <- c(xml, sprintf('<r>%s<t xml:space="preserve">%s</t></r>', 
          comment_list[[i]]$style[[j]], 
          comment_list[[i]]$comment[[j]]))
      }
    } else {
      ## Case with no styling information.
      for (j in seq_along(comment_list[[i]]$comment)) {
        xml <- c(xml, sprintf('<t>%s</t>', 
          comment_list[[i]]$comment[[j]]))
      }
    }
    
    xml <- c(xml, "</text></comment>")
  }
  
  write_file(body = paste(xml, collapse = ""), tail = "</commentList></comments>", fl = file_name)
  
  NULL
}


illegalchars <- c("&", '"', "'", "<", ">", "\a", "\b", "\v", "\f")
illegalcharsreplace <- c("&amp;", "&quot;", "&apos;", "&lt;", "&gt;", "", "", "", "")

replaceIllegalCharacters <- function(v) {
  vEnc <- Encoding(v)
  v <- as.character(v)
  
  flg <- vEnc != "UTF-8"
  if (any(flg)) {
    v[flg] <- stri_conv(v[flg], from = "", to = "UTF-8")
  }
  
  v <- stri_replace_all_fixed(v, illegalchars, illegalcharsreplace, vectorize_all = FALSE)
  
  return(v)
}


replaceXMLEntities <- function(v) {
  v <- gsub("&amp;", "&", v, fixed = TRUE)
  v <- gsub("&quot;", '"', v, fixed = TRUE)
  v <- gsub("&apos;", "'", v, fixed = TRUE)
  v <- gsub("&lt;", "<", v, fixed = TRUE)
  v <- gsub("&gt;", ">", v, fixed = TRUE)
  
  return(v)
}


pxml <- function(x) {
  paste(unique(unlist(x)), collapse = "")
}


removeHeadTag <- function(x) {
  x <- paste(x, collapse = "")
  
  if (any(grepl("<\\?", x))) {
    x <- gsub("<\\?xml [^>]+", "", x)
  }
  
  x <- gsub("^>", "", x)
  x
}

validateBorderStyle <- function(borderStyle) {
  valid <- c(
    "none", "thin", "medium", "dashed", "dotted", "thick", "double", "hair", "mediumDashed",
    "dashDot", "mediumDashDot", "dashDotDot", "mediumDashDotDot", "slantDashDot"
  )
  
  ind <- match(tolower(borderStyle), tolower(valid))
  if (any(is.na(ind))) {
    stop("Invalid borderStyle", call. = FALSE)
  }
  
  return(valid[ind])
}

getAttrsFont <- function(xml, tag) {
  x <- lapply(xml, getChildlessNode, tag = tag)
  x[sapply(x, length) == 0] <- ""
  x <- unlist(x)
  a <- lapply(x, function(x) unlist(regmatches(x, gregexpr('[a-zA-Z]+=".*?"', x))))
  
  nms <- lapply(a, function(xml) regmatches(xml, regexpr('[a-zA-Z]+(?=\\=".*?")', xml, perl = TRUE)))
  vals <- lapply(a, function(xml) regmatches(xml, regexpr('(?<=").*?(?=")', xml, perl = TRUE)))
  vals <- lapply(vals, function(x) {
    Encoding(x) <- "UTF-8"
    x
  })
  vals <- lapply(seq_along(vals), function(i) {
    names(vals[[i]]) <- nms[[i]]
    vals[[i]]
  })
  
  return(vals)
}

getAttrs <- function(xml, tag) {
  x <- lapply(xml, getChildlessNode_ss, tag = tag)
  x[sapply(x, length) == 0] <- ""
  a <- lapply(x, function(x) regmatches(x, regexpr('[a-zA-Z]+=".*?"', x)))
  
  names <- lapply(a, function(xml) regmatches(xml, regexpr('[a-zA-Z]+(?=\\=".*?")', xml, perl = TRUE)))
  vals <- lapply(a, function(xml) regmatches(xml, regexpr('(?<=").*?(?=")', xml, perl = TRUE)))
  vals <- lapply(vals, function(x) {
    Encoding(x) <- "UTF-8"
    x
  })
  
  names(vals) <- names
  return(vals)
}


buildFontList <- function(fonts) {
  sz <- getAttrs(fonts, "sz")
  colour <- getAttrsFont(fonts, "color")
  name <- getAttrs(fonts, tag = "name")
  family <- getAttrs(fonts, "family")
  scheme <- getAttrs(fonts, "scheme")
  
  italic <- lapply(fonts, getChildlessNode, tag = "i")
  bold <- lapply(fonts, getChildlessNode, tag = "b")
  underline <- lapply(fonts, getChildlessNode, tag = "u")  
  strikeout <- lapply(fonts, getChildlessNode, tag = "strike")

  ## Build font objects
  ft <- replicate(list(), n = length(fonts))
  for (i in seq_along(fonts)) {
    f <- NULL
    nms <- NULL
    if (length(unlist(sz[i])) > 0) {
      f <- c(f, sz[i])
      nms <- c(nms, "sz")
    }
    
    if (length(unlist(colour[i])) > 0) {
      f <- c(f, colour[i])
      nms <- c(nms, "color")
    }
    
    if (length(unlist(name[i])) > 0) {
      f <- c(f, name[i])
      nms <- c(nms, "name")
    }
    
    if (length(unlist(family[i])) > 0) {
      f <- c(f, family[i])
      nms <- c(nms, "family")
    }
    
    if (length(unlist(scheme[i])) > 0) {
      f <- c(f, scheme[i])
      nms <- c(nms, "scheme")
    }
    
    if (length(italic[[i]]) > 0) {
      f <- c(f, "italic")
      nms <- c(nms, "italic")
    }
    
    if (length(bold[[i]]) > 0) {
      f <- c(f, "bold")
      nms <- c(nms, "bold")
    }
    
    if (length(underline[[i]]) > 0) {
      f <- c(f, "underline")
      nms <- c(nms, "underline")
    }

    if (length(unlist(strikeout[i])) > 0) {
      f <- c(f, strikeout[i])
      nms <- c(nms, "strikeout")
    }
    
    f <- lapply(seq_along(f), function(i) unlist(f[i]))
    names(f) <- nms
    
    ft[[i]] <- f
  }
  
  ft
}



get_named_regions_from_string <- function(dn) {
  dn <- gsub("</definedNames>", "", dn, fixed = TRUE)
  dn <- gsub("</workbook>", "", dn, fixed = TRUE)
  
  dn <- unique(unlist(strsplit(dn, split = "</definedName>", fixed = TRUE)))
  dn <- grep("<definedName", dn, fixed = TRUE, value = TRUE)
  
  dn_names <- regmatches(dn, regexpr('(?<=name=")[^"]+', dn, perl = TRUE))
  
  dn_pos <- regmatches(dn, regexpr("(?<=>).*", dn, perl = TRUE))
  dn_pos <- gsub("[$']", "", dn_pos)
  
  has_bang <- grepl("!", dn_pos, fixed = TRUE)
  dn_sheets <- ifelse(has_bang,
    gsub("^(.*)!.*$", "\\1", dn_pos),
    ""
  )
  dn_coords <- ifelse(has_bang,
    gsub("^.*!(.*)$", "\\1", dn_pos),
    ""
  )
  
  attr(dn_names, "sheet") <- dn_sheets
  attr(dn_names, "position") <- dn_coords
  
  return(dn_names)
}



nodeAttributes <- function(x) {
  x <- paste0("<", unlist(strsplit(x, split = "<")))
  x <- grep("<bgColor|<fgColor", x, value = TRUE)
  
  if (length(x) == 0) {
    return("")
  }
  
  attrs <- regmatches(x, gregexpr(' [a-zA-Z]+="[^"]*', x, perl = TRUE))
  tags <- regmatches(x, gregexpr("<[a-zA-Z]+ ", x, perl = TRUE))
  tags <- lapply(tags, gsub, pattern = "<| ", replacement = "")
  attrs <- lapply(attrs, gsub, pattern = '"', replacement = "")
  
  attrs <- lapply(attrs, strsplit, split = "=")
  for (i in seq_along(attrs)) {
    nms <- lapply(attrs[[i]], "[[", 1)
    vals <- lapply(attrs[[i]], "[[", 2)
    a <- unlist(vals)
    names(a) <- unlist(nms)
    attrs[[i]] <- a
  }
  
  names(attrs) <- unlist(tags)
  
  attrs
}


buildBorder <- function(x) {
  style <- list()
  if (grepl('diagonalup="1"', tolower(x), fixed = TRUE)) {
    style$borderDiagonalUp <- TRUE
  }
  
  if (grepl('diagonaldown="1"', tolower(x), fixed = TRUE)) {
    style$borderDiagonalDown <- TRUE
  }
  
  ## gets all borders that have children
  x <- unlist(lapply(c("<left", "<right", "<top", "<bottom", "<diagonal"), function(tag) getNodes(xml = x, tagIn = tag)))
  if (length(x) == 0) {
    return(NULL)
  }
  
  sides <- c("TOP", "BOTTOM", "LEFT", "RIGHT", "DIAGONAL")
  sideBorder <- character(length = length(x))
  for (i in seq_along(x)) {
    tmp <- sides[sapply(sides, function(s) grepl(s, x[[i]], ignore.case = TRUE))]
    if (length(tmp) > 1) tmp <- tmp[[1]]
    if (length(tmp) == 1) {
      sideBorder[[i]] <- tmp
    }
  }
  
  sideBorder <- sideBorder[sideBorder != ""]
  x <- x[sideBorder != ""]
  if (length(sideBorder) == 0) {
    return(NULL)
  }
  
  
  ## style
  weight <- gsub('style=|"', "", regmatches(x, regexpr('style="[a-z]+"', x, perl = TRUE, ignore.case = TRUE)))
  
  
  ## Colours
  cols <- replicate(n = length(sideBorder), list(rgb = "FF000000"))
  colNodes <- unlist(sapply(x, getChildlessNode, tag = "color", USE.NAMES = FALSE))
  
  if (length(colNodes) > 0) {
    attrs <- regmatches(colNodes, regexpr('(theme|indexed|rgb|auto)=".+"', colNodes))
  } else {
    attrs <- NULL
  }
  
  if (length(attrs) != length(x)) {
    return(
      list(
        "borders" = paste(sideBorder, collapse = ""),
        "colour" = cols
      )
    )
  }
  
  attrs <- strsplit(attrs, split = "=")
  cols <- sapply(attrs, function(attr) {
    if (length(attr) == 2) {
      y <- list(gsub('"', "", attr[2]))
      names(y) <- gsub(" ", "", attr[[1]])
    } else {
      tmp <- paste(attr[-1], collapse = "=")
      y <- gsub('^"|"$', "", tmp)
      names(y) <- gsub(" ", "", attr[[1]])
    }
    return(y)
  })
  
  ## sideBorder & cols
  if ("LEFT" %in% sideBorder) {
    style$borderLeft <- weight[which(sideBorder == "LEFT")]
    style$borderLeftColour <- cols[which(sideBorder == "LEFT")]
  }
  
  if ("RIGHT" %in% sideBorder) {
    style$borderRight <- weight[which(sideBorder == "RIGHT")]
    style$borderRightColour <- cols[which(sideBorder == "RIGHT")]
  }
  
  if ("TOP" %in% sideBorder) {
    style$borderTop <- weight[which(sideBorder == "TOP")]
    style$borderTopColour <- cols[which(sideBorder == "TOP")]
  }
  
  if ("BOTTOM" %in% sideBorder) {
    style$borderBottom <- weight[which(sideBorder == "BOTTOM")]
    style$borderBottomColour <- cols[which(sideBorder == "BOTTOM")]
  }
  
  if ("DIAGONAL" %in% sideBorder) {
    style$borderDiagonal <- weight[which(sideBorder == "DIAGONAL")]
    style$borderDiagonalColour <- cols[which(sideBorder == "DIAGONAL")]
  }
  
  return(style)
}


genHeaderFooterNode <- function(x) {
  
  # <headerFooter differentOddEven="1" differentFirst="1" scaleWithDoc="0" alignWithMargins="0">
  #   <oddHeader>&amp;Lfirst L&amp;CfC&amp;RfR</oddHeader>
  #   <oddFooter>&amp;LfFootL&amp;CfFootC&amp;RfFootR</oddFooter>
  #   <evenHeader>&amp;LTIS&amp;CIS&amp;REVEN H</evenHeader>
  #   <evenFooter>&amp;LEVEN L F&amp;CEVEN C F&amp;REVEN RIGHT F</evenFooter>
  #   <firstHeader>&amp;L&amp;P&amp;Cfirst C&amp;Rfirst R</firstHeader>
  #   <firstFooter>&amp;Lfirst L Foot&amp;Cfirst C Foot&amp;Rfirst R Foot</firstFooter>
  #   </headerFooter>
  
  ## ODD
  if (length(x$oddHeader) > 0) {
    oddHeader <- paste0(
      "<oddHeader>", 
      sprintf("&amp;L%s", x$oddHeader[[1]]), 
      sprintf("&amp;C%s", x$oddHeader[[2]]), 
      sprintf("&amp;R%s", x$oddHeader[[3]]),
      "</oddHeader>", 
      collapse = ""
    )
  } else {
    oddHeader <- NULL
  }
  
  if (length(x$oddFooter) > 0) {
    oddFooter <- paste0(
      "<oddFooter>", 
      sprintf("&amp;L%s", x$oddFooter[[1]]),
      sprintf("&amp;C%s", x$oddFooter[[2]]), 
      sprintf("&amp;R%s", x$oddFooter[[3]]),
      "</oddFooter>", 
      collapse = ""
    )
  } else {
    oddFooter <- NULL
  }
  
  ## EVEN
  if (length(x$evenHeader) > 0) {
    evenHeader <- paste0(
      "<evenHeader>", 
      sprintf("&amp;L%s", x$evenHeader[[1]]),
      sprintf("&amp;C%s", x$evenHeader[[2]]),
      sprintf("&amp;R%s", x$evenHeader[[3]]),
      "</evenHeader>",
      collapse = ""
    )
  } else {
    evenHeader <- NULL
  }
  
  if (length(x$evenFooter) > 0) {
    evenFooter <- paste0(
      "<evenFooter>", 
      sprintf("&amp;L%s", x$evenFooter[[1]]),
      sprintf("&amp;C%s", x$evenFooter[[2]]), 
      sprintf("&amp;R%s", x$evenFooter[[3]]),
      "</evenFooter>", 
      collapse = ""
    )
  } else {
    evenFooter <- NULL
  }
  
  ## FIRST
  if (length(x$firstHeader) > 0) {
    firstHeader <- paste0(
      "<firstHeader>", 
      sprintf("&amp;L%s", x$firstHeader[[1]]), 
      sprintf("&amp;C%s", x$firstHeader[[2]]),
      sprintf("&amp;R%s", x$firstHeader[[3]]),
      "</firstHeader>", 
      collapse = ""
    )
  } else {
    firstHeader <- NULL
  }
  
  if (length(x$firstFooter) > 0) {
    firstFooter <- paste0(
      "<firstFooter>",
      sprintf("&amp;L%s", x$firstFooter[[1]]), 
      sprintf("&amp;C%s", x$firstFooter[[2]]), 
      sprintf("&amp;R%s", x$firstFooter[[3]]), 
      "</firstFooter>",
      collapse = ""
      )
  } else {
    firstFooter <- NULL
  }
  
  headTag <- sprintf(
    '<headerFooter differentOddEven="%s" differentFirst="%s" scaleWithDoc="0" alignWithMargins="0">',
    as.integer(!(is.null(evenHeader) & is.null(evenFooter))),
    as.integer(!(is.null(firstHeader) & is.null(firstFooter)))
  )
  
  paste0(
    headTag, 
    oddHeader, 
    oddFooter, 
    evenHeader, 
    evenFooter, 
    firstHeader, 
    firstFooter,
    "</headerFooter>"
  )
}


buildFillList <- function(fills) {
  fillAttrs <- rep(list(list()), length(fills))
  
  ## patternFill
  inds <- grepl("patternFill", fills)
  fillAttrs[inds] <- lapply(fills[inds], nodeAttributes)
  
  ## gradientFill
  inds <- grepl("gradientFill", fills)
  fillAttrs[inds] <- fills[inds]
  
  return(fillAttrs)
}

# Can test with below:
# x <- "<definedName name=\"_xlnm._FilterDatabase\" localSheetId=\"0\" hidden=\"1\">'A & B < D > D'!$A$1:$A$10</definedName>"
getDefinedNamesSheet <- function(x) {
  sub("'?\\!.*", "", sub("^.*>'", "", x))
}

# Not used but kepted in case fix above isn't correct
getDefinedNamedSheet_ <- function(x) {
  belongTo <- unlist(lapply(strsplit(x, split = ">|<"), "[[", 3))
  quoted <- grepl("^'", belongTo)
  
  belongTo[quoted] <- regmatches(belongTo[quoted], regexpr("(?<=').*(?='!)", belongTo[quoted], perl = TRUE))
  belongTo[!quoted] <- gsub("!\\$[A-Z0-9].*", "", belongTo[!quoted])
  belongTo[!quoted] <- gsub("!#REF!.*", "", belongTo[!quoted])
  
  return(belongTo)
}

getSharedStringsFromFile <- function(sharedStringsFile, isFile) {
  
  ## read in, get si tags, get t tag value and  pull out all string nodes
  sharedStrings <- get_shared_strings(xmlFile = sharedStringsFile, isFile = isFile) ## read from file
  
  
  Encoding(sharedStrings) <- "UTF-8"
  z <- tolower(sharedStrings)
  sharedStrings[z == "true"] <- "TRUE"
  sharedStrings[z == "false"] <- "FALSE"
  z <- NULL ## effectivel remove z
  
  ## XML replacements
  sharedStrings <- replaceXMLEntities(sharedStrings)
  
  return(sharedStrings)
}


clean_names <- function(x, schar) {
  x <- gsub("^[[:space:]]+|[[:space:]]+$", "", x)
  x <- gsub("[[:space:]]+", schar, x)
  return(x)
}



mergeCell2mapping <- function(x) {
  refs <- regmatches(x, regexpr("(?<=ref=\")[A-Z0-9:]+", x, perl = TRUE))
  refs <- strsplit(refs, split = ":")
  rows <- lapply(refs, function(r) {
    r <- as.integer(gsub(pattern = "[A-Z]", replacement = "", r, perl = TRUE))
    seq(from = r[1], to = r[2], by = 1)
  })
  
  cols <- lapply(refs, function(r) {
    r <- convertFromExcelRef(r)
    seq(from = r[1], to = r[2], by = 1)
  })
  
  ## for each we grid.expand
  refs <- do.call("rbind", lapply(seq_along(rows), function(i) {
    tmp <- expand.grid("cols" = cols[[i]], "rows" = rows[[i]])
    tmp$ref <- paste0(convert_to_excel_ref(cols = tmp$cols, LETTERS = LETTERS), tmp$rows)
    tmp$anchor_cell <- tmp$ref[1]
    return(tmp[, c("anchor_cell", "ref", "rows")])
  }))
  
  
  refs <- refs[refs$anchor_cell != refs$ref, ]
  
  return(refs)
}



getFile <- function(xlsxFile) {
  
  ## Is this a file or URL (code taken from read.table())
  on.exit(try(close(fl), silent = TRUE), add = TRUE)
  fl <- file(description = xlsxFile)
  
  ## If URL download
  if ("url" %in% class(fl)) {
    tmpFile <- tempfile(fileext = ".xlsx")
    download.file(url = xlsxFile, destfile = tmpFile, cacheOK = FALSE, mode = "wb", quiet = TRUE)
    xlsxFile <- tmpFile
  }
  
  return(xlsxFile)
}

#' @name get_worksheet_entries
#' @title Get entries from workbook worksheet
#' @description Get all entries from workbook worksheet without xml tags
#' @param wb workbook
#' @param sheet worksheet
#' @author David Breuer
#' @return vector of strings
#' @export
#' @examples
#' ## Create new workbook
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet")
#' sheet <- 1
#'
#' ## Write dummy data
#' writeData(wb, sheet, c("A", "BB", "CCC"), startCol = 2, startRow = 3)
#' writeData(wb, sheet, c(4, 5), startCol = 4, startRow = 3)
#'
#' ## Get text entries
#' get_worksheet_entries(wb, sheet)
#'
get_worksheet_entries <- function(wb, sheet) {
  # get worksheet data
  dat <- wb$worksheets[[sheet]]$sheet_data
  # get vector of entries
  val <- dat$v
  # get boolean vector of text entries
  typ <- (dat$t == 1) & !is.na(dat$t)
  # get text entry strings
  str <- unlist(wb$sharedStrings[as.numeric(val[typ]) + 1])
  # remove xml tags
  str <- gsub("<.*?>", "", str)
  # write strings to vector of entries
  val[typ] <- str
  # return vector of entries
  val
}

#' @name auto_heights
#' @title Compute optimal row heights
#' @description Compute optimal row heights for cell with fixed with and
#' enabled automatic row heights parameter
#' @param wb workbook
#' @param sheet worksheet
#' @param selected selected rows
#' @param fontsize font size, optional (get base font size by default)
#' @param factor factor to manually adjust font width, e.g., for bold fonts,
#' optional
#' @param base_height basic row height, optional
#' @param extra_height additional row height per new line of text, optional
#' @author David Breuer
#' @return list of indices of columns with fixed widths and optimal row heights
#' @export
#' @examples
#' ## Create new workbook
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet")
#' sheet <- 1
#'
#' ## Write dummy data
#' long_string <- "ABC ABC ABC ABC ABC ABC ABC ABC ABC ABC ABC"
#' writeData(wb, sheet, c("A", long_string, "CCC"), startCol = 2, startRow = 3)
#' writeData(wb, sheet, c(4, 5), startCol = 4, startRow = 3)
#'
#' ## Set column widths and get optimal row heights
#' setColWidths(wb, sheet, c(1,2,3,4), c(10,20,10,20))
#' auto_heights(wb, sheet, 1:5)
#'
auto_heights <- function(wb, sheet, selected, fontsize = NULL, factor = 1.0,
                         base_height = 15, extra_height = 12) {
  # get base font size
  if (is.null(fontsize)) {
    fontsize <- as.integer(openxlsx::getBaseFont(wb)$size$val)
  }
  # set factor to adjust font width (empiricially found scale factor 4 here)
  factor <- 4 * factor / fontsize
  # get worksheet data
  dat <- wb$worksheets[[sheet]]$sheet_data
  # get columns widths
  colWidths <- wb$colWidths[[sheet]]
  # select fixed (non-auto) and visible (non-hidden) columns only
  specified <- (colWidths != "auto") & (attr(colWidths, "hidden") == "0")
  # return default row heights if no column widths are fixed
  if (length(specified) == 0) {
    # message("No column widths specified, returning default row heights.")
    cols <- integer(0)
    heights <- rep(base_height, length(selected))
    return(list(cols, heights))
  }
  # get fixed column indices
  cols <- as.integer(names(specified)[specified])
  # get fixed column widths
  widths <- as.numeric(colWidths[specified])
  # get all worksheet entries
  val <- get_worksheet_entries(wb, sheet)
  # compute optimal height per selected row
  heights <- sapply(selected, function(row) {
    # select entries in given row and columns of fixed widths
    index <- (dat$rows == row) & (dat$cols %in% cols)
    # remove line break characters
    chr <- gsub("\\r|\\n", "", val[index])
    # measure width of entry (in pixels)
    wdt <- graphics::strwidth(chr, unit = "in") * 20 / 1.43 # 20 px = 1.43 in
    # compute optimal height
    if (length(wdt) == 0) {
      base_height
    } else {
      base_height + extra_height * as.integer(max(wdt / widths * factor))
    }
  })
  # return list of indices of columns with fixed widths and optimal row heights
  list(cols, heights)
}


# Rotate the 15-bit integer by n bits to the
hashPassword <- function(password) {
  # password limited to 15 characters
  chars <- head(strsplit(password, "")[[1]], 15)
  # See OpenOffice's documentation of the Excel format: http://www.openoffice.org/sc/excelfileformat.pdf
  # Start from the last character and for each character
  # - XOR hash with the ASCII character code
  # - rotate hash (16 bits) one bit to the left
  # Finally, XOR hash with 0xCE4B and XOR with password length
  # Output as hex (uppercase)
  rotate16bit <- function(hash, n = 1) {
    bitwOr(bitwAnd(bitwShiftR(hash, 15 - n), 0x01), bitwAnd(bitwShiftL(hash, n), 0x7fff))
  }
  hash <- Reduce(function(char, h) {
    h <- bitwXor(h, as.integer(charToRaw(char)))
    rotate16bit(h, 1)
  }, chars, 0, right = TRUE)
  hash <- bitwXor(bitwXor(hash, length(chars)), 0xCE4B)
  format(as.hexmode(hash), upper.case = TRUE)
}

readUTF8 <- function(x) {
  readLines(x, warn = FALSE, encoding = "UTF-8")
}

context("silent get_worksheet_entries")

test_that("get_worksheet_entries is silent in reference to string in other sheet", {
  fl <- system.file("extdata", "silent_worksheet_entries.xlsx", package = "openxlsx")
  wb <- loadWorkbook(fl)
  expect_silent(get_worksheet_entries(wb, 1))
})

#' Build Workbook
#'
#' Build a workbook from a data.frame or named list
#'
#' @details
#' This function can be used as shortcut to create a workbook object from a
#'   data.frame or named list.  If names are available in the list they will be
#'   used as the worksheet names.  The parameters in `...` are collected
#'   and passed to [writeData()] or [writeDataTable()] to
#'   initially create the Workbook objects then appropriate parameters are
#'   passed to [setColWidths()].
#'
#' @param x A data.frame or a (named) list of objects that can be handled by
#'   [writeData()] or [writeDataTable()] to write to file
#' @param asTable If `TRUE` will use [writeDataTable()] rather
#'   than [writeData()] to write `x` to the file (default:
#'   `FALSE`)
#' @param ... Additional arguments passed to [writeData()],
#'   [writeDataTable()], [setColWidths()] (see Optional
#'   Parameters)
#' @author Jordan Mark Barbone
#' @returns A Workbook object
#'
#' @details
#' columns of x with class Date or POSIXt are automatically
#' styled as dates and datetimes respectively.
#'
#' @section Optional Parameters:
#'
#' **createWorkbook Parameters**
#' \describe{
#'   \item{**creator**}{ A string specifying the workbook author}
#' }
#'
#' **addWorksheet Parameters**
#' \describe{
#'   \item{**sheetName**}{ Name of the worksheet}
#'   \item{**gridLines**}{ A logical. If `FALSE`, the worksheet grid lines will be hidden.}
#'   \item{**tabColour**}{ Colour of the worksheet tab. A valid colour (belonging to colours())
#'   or a valid hex colour beginning with "#".}
#'   \item{**zoom**}{ A numeric between 10 and 400. Worksheet zoom level as a percentage.}
#' }
#'
#' **writeData/writeDataTable Parameters**
#' \describe{
#'   \item{**startCol**}{ A vector specifying the starting column(s) to write df}
#'   \item{**startRow**}{ A vector specifying the starting row(s) to write df}
#'   \item{**xy**}{ An alternative to specifying startCol and startRow individually.
#'  A vector of the form c(startCol, startRow)}
#'   \item{**colNames or col.names**}{ If `TRUE`, column names of x are written.}
#'   \item{**rowNames or row.names**}{ If `TRUE`, row names of x are written.}
#'   \item{**headerStyle**}{ Custom style to apply to column names.}
#'   \item{**borders**}{ Either "surrounding", "columns" or "rows" or NULL.  If "surrounding", a border is drawn around the
#' data.  If "rows", a surrounding border is drawn a border around each row. If "columns", a surrounding border is drawn with a border
#' between each column.  If "`all`" all cell borders are drawn.}
#'   \item{**borderColour**}{ Colour of cell border}
#'   \item{**borderStyle**}{ Border line style.}
#'   \item{**keepNA**}{If `TRUE`, NA values are converted to #N/A (or `na.string`, if not NULL) in Excel, else NA cells will be empty. Defaults to FALSE.}
#'   \item{**na.string**}{If not NULL, and if `keepNA` is `TRUE`, NA values are converted to this string in Excel. Defaults to NULL.}
#' }
#'
#' **freezePane Parameters**
#' \describe{
#'   \item{**firstActiveRow**}{Top row of active region to freeze pane.}
#'   \item{**firstActiveCol**}{Furthest left column of active region to freeze pane.}
#'   \item{**firstRow**}{If `TRUE`, freezes the first row (equivalent to firstActiveRow = 2)}
#'   \item{**firstCol**}{If `TRUE`, freezes the first column (equivalent to firstActiveCol = 2)}
#' }
#'
#' **colWidths Parameters**
#' \describe{
#'   \item{**colWidths**}{May be a single value for all columns (or "auto"), or a list of vectors that will be recycled for each sheet (see examples)}
#' }
#'
#' @examples
#' x <- data.frame(a = 1, b = 2)
#' wb <- buildWorkbook(x)
#'
#' y <- list(a = x, b = x, c = x)
#' buildWorkbook(y, asTable = TRUE)
#' buildWorkbook(y, asTable = TRUE, tableStyle = "TableStyleLight8")
#'
#' @seealso [write.xlsx()]
#'
#' @export

buildWorkbook <- function(x, asTable = FALSE, ...) {
  if (!is.logical(asTable)) {
    stop("asTable must be a logical.")
  }

  params <- list(...)
  isList <- inherits(x, "list")

  if (isList) {
    params[["sheetName"]] <- params[["sheetName"]] %||% names(x) %||% paste0("Sheet ", seq_along(x))
  }

  ## create new Workbook object
  wb <- do_call_params(createWorkbook, params)

  ## If a list is supplied write to individual worksheets using names if available
  if (isList) {
    do_call_params(addWorksheet, params, wb = list(wb), .map = TRUE)
  } else {
    params[["sheetName"]] <- params[["sheetName"]] %||% "Sheet 1"
    do_call_params(addWorksheet, params, wb = wb)
  }

  params[["sheet"]] <- params[["sheet"]] %||% params[["sheetName"]]

  # write Data
  if (asTable) {
    do_call_params(writeDataTable, params, x = x, wb = list(wb), .map = TRUE)
  } else {
    do_call_params(writeData, params, x = x, wb = wb, .map = TRUE)
  }

  do_setColWidths(wb, x, params, isList)
  do_call_params(freezePane, params, wb = list(wb), .map = TRUE)
  wb
}


do_setColWidths <- function(wb, x, params, isList) {
  if (!isList) {
    x <- list(x)
  }

  params[["startCol"]] <- params[["startCol"]] %||% 1
  params[["startCol"]] <- rep_len(list(params[["startCol"]]), length.out = length(x))
  params[["colWidths"]] <- params[["colWidths"]] %||% ""
  params[["colWidths"]] <- rep_len(as.list(params[["colWidths"]]), length.out = length(x))

  for (i in seq_along(wb[["worksheets"]])) {
    if (identical(params[["colWidths"]][[i]], "auto")) {
      setColWidths(
        wb,
        sheet = i,
        cols = seq_along(x[[i]]) + params[["startCol"]][[i]] - 1L,
        widths = "auto"
      )
    } else if (!identical(params[["colWidths"]][[i]], "")) {
      setColWidths(
        wb,
        sheet = i,
        cols = seq_along(x[[i]]) + params[["startCol"]][[i]] - 1L,
        widths = params[["colWidths"]][[i]]
      )
    }
  }
  wb
}

#' @include class_definitions.R


Style$methods(initialize = function() {
  fontName <<- NULL
  fontColour <<- NULL
  fontSize <<- NULL
  fontFamily <<- NULL
  fontScheme <<- NULL
  fontDecoration <<- NULL

  borderTop <<- NULL
  borderLeft <<- NULL
  borderRight <<- NULL
  borderBottom <<- NULL
  borderTopColour <<- NULL
  borderLeftColour <<- NULL
  borderRightColour <<- NULL
  borderBottomColour <<- NULL
  borderDiagonal <<- NULL
  borderDiagonalColour <<- NULL
  borderDiagonalUp <<- FALSE
  borderDiagonalDown <<- FALSE

  halign <<- NULL
  valign <<- NULL
  indent <<- NULL
  textRotation <<- NULL
  numFmt <<- NULL
  fill <<- NULL
  wrapText <<- NULL
  hidden <<- NULL
  locked <<- NULL
  xfId <<- NULL
})



mergeStyle <- function(oldStyle, newStyle) {

  ## This function is used to merge an existing cell style with a new style to create a stacked style.
  oldStyle <- oldStyle$copy()

  if (!is.null(newStyle$fontName)) {
    oldStyle$fontName <- newStyle$fontName
  }

  if (!is.null(newStyle$fontColour)) {
    oldStyle$fontColour <- newStyle$fontColour
  }

  if (!is.null(newStyle$fontSize)) {
    oldStyle$fontSize <- newStyle$fontSize
  }

  if (!is.null(newStyle$fontFamily)) {
    oldStyle$fontFamily <- newStyle$fontFamily
  }

  if (!is.null(newStyle$fontScheme)) {
    oldStyle$fontScheme <- newStyle$fontScheme
  }

  if (length(newStyle$fontDecoration) > 0) {
    if (length(oldStyle$fontDecoration) == 0) {
      oldStyle$fontDecoration <- newStyle$fontDecoration
    } else {
      oldStyle$fontDecoration <- c(oldStyle$fontDecoration, newStyle$fontDecoration)
    }
  }


  ## borders
  if (!is.null(newStyle$borderTop)) {
    oldStyle$borderTop <- newStyle$borderTop
  }

  if (!is.null(newStyle$borderLeft)) {
    oldStyle$borderLeft <- newStyle$borderLeft
  }

  if (!is.null(newStyle$borderRight)) {
    oldStyle$borderRight <- newStyle$borderRight
  }

  if (!is.null(newStyle$borderBottom)) {
    oldStyle$borderBottom <- newStyle$borderBottom
  }

  if (!is.null(newStyle$borderDiagonal)) {
    oldStyle$borderDiagonal <- newStyle$borderDiagonal
  }

  oldStyle$borderDiagonalUp <- newStyle$borderDiagonalUp
  oldStyle$borderDiagonalDown <- newStyle$borderDiagonalDown


  if (!is.null(newStyle$borderTopColour)) {
    oldStyle$borderTopColour <- newStyle$borderTopColour
  }

  if (!is.null(newStyle$borderLeftColour)) {
    oldStyle$borderLeftColour <- newStyle$borderLeftColour
  }

  if (!is.null(newStyle$borderRightColour)) {
    oldStyle$borderRightColour <- newStyle$borderRightColour
  }

  if (!is.null(newStyle$borderBottomColour)) {
    oldStyle$borderBottomColour <- newStyle$borderBottomColour
  }



  ## other
  if (!is.null(newStyle$halign)) {
    oldStyle$halign <- newStyle$halign
  }

  if (!is.null(newStyle$valign)) {
    oldStyle$valign <- newStyle$valign
  }

  if (!is.null(newStyle$indent)) {
    oldStyle$indent <- newStyle$indent
  }

  if (!is.null(newStyle$textRotation)) {
    oldStyle$textRotation <- newStyle$textRotation
  }

  if (!is.null(newStyle$numFmt)) {
    oldStyle$numFmt <- newStyle$numFmt
  }

  if (!is.null(newStyle$fill)) {
    oldStyle$fill <- newStyle$fill
  }

  if (!is.null(newStyle$wrapText)) {
    oldStyle$wrapText <- newStyle$wrapText
  }

  if (!is.null(newStyle$locked)) {
    oldStyle$locked <- newStyle$locked
  }

  if (!is.null(newStyle$hidden)) {
    oldStyle$hidden <- newStyle$hidden
  }

  if (!is.null(newStyle$xfId)) {
    oldStyle$xfId <- newStyle$xfId
  }

  return(oldStyle)
}





Style$methods(show = function(print = TRUE) {
  numFmtMapping <- list(
    list("numFmtId" = 0),
    list("numFmtId" = 2),
    list("numFmtId" = 164),
    list("numFmtId" = 44),
    list("numFmtId" = 14),
    list("numFmtId" = 167),
    list("numFmtId" = 10),
    list("numFmtId" = 11),
    list("numFmtId" = 49)
  )

  validNumFmt <- c("GENERAL", "NUMBER", "CURRENCY", "ACCOUNTING", "DATE", "TIME", "PERCENTAGE", "SCIENTIFIC", "TEXT")

  if (!is.null(numFmt)) {
    if (as.integer(numFmt$numFmtId) %in% unlist(numFmtMapping)) {
      numFmtStr <- validNumFmt[unlist(numFmtMapping) == as.integer(numFmt$numFmtId)]
    } else {
      numFmtStr <- sprintf('"%s"', numFmt$formatCode)
    }
  } else {
    numFmtStr <- "GENERAL"
  }

  borders <- c(sprintf("Top: %s", borderTop), sprintf("Bottom: %s", borderBottom), sprintf("Left: %s", borderLeft), sprintf("Right: %s", borderRight))
  borderColours <- gsub("^FF", "#", c(borderTopColour, borderBottomColour, borderLeftColour, borderRightColour))

  fgFill <- fill$fillFg
  bgFill <- fill$fillBg

  styleShow <- "A custom cell style. \n\n"

  styleShow <- append(styleShow, sprintf("Cell formatting: %s \n", numFmtStr)) ## numFmt
  styleShow <- append(styleShow, sprintf("Font name: %s \n", fontName[[1]])) ## Font name
  styleShow <- append(styleShow, sprintf("Font size: %s \n", fontSize[[1]])) ## Font size
  styleShow <- append(styleShow, sprintf("Font colour: %s \n", gsub("^FF", "#", fontColour[[1]]))) ## Font colour

  ## Font decoration
  if (length(fontDecoration) > 0) {
    styleShow <- append(styleShow, sprintf("Font decoration: %s \n", paste(fontDecoration, collapse = ", ")))
  }

  if (length(borders) > 0) {
    styleShow <- append(styleShow, sprintf("Cell borders: %s \n", paste(borders, collapse = ", "))) ## Cell borders
    styleShow <- append(styleShow, sprintf("Cell border colours: %s \n", paste(borderColours, collapse = ", "))) ## Cell borders
  }

  if (!is.null(halign)) {
    styleShow <- append(styleShow, sprintf("Cell horz. align: %s \n", halign))
  } ## Cell horizontal alignment

  if (!is.null(valign)) {
    styleShow <- append(styleShow, sprintf("Cell vert. align: %s \n", valign))
  } ## Cell vertical alignment

  if (!is.null(indent)) {
    styleShow <- append(styleShow, sprintf("Cell indent: %s \n", indent))
  } ## Cell indent

  if (!is.null(textRotation)) {
    styleShow <- append(styleShow, sprintf("Cell text rotation: %s \n", textRotation))
  } ## Cell text rotation

  ## Cell fill colour
  if (length(fgFill) > 0) {
    styleShow <- append(styleShow, sprintf("Cell fill foreground: %s \n", paste(paste0(names(fgFill), ": ", sub("^FF", "#", fgFill)), collapse = ", ")))
  }

  if (length(bgFill) > 0) {
    styleShow <- append(styleShow, sprintf("Cell fill background: %s \n", paste(paste0(names(bgFill), ": ", sub("^FF", "#", bgFill)), collapse = ", ")))
  }

  if (!is.null(locked)) {
    styleShow <- append(styleShow, sprintf("Cell protection: %s \n", locked))
  } ## Cell protection
  if (!is.null(hidden)) {
    styleShow <- append(styleShow, sprintf("Cell formula hidden: %s \n", hidden))
  } ## Cell formula hidden

  styleShow <- append(styleShow, sprintf("wraptext: %s", wrapText)) ## wrap text

  styleShow <- c(styleShow, "\n\n")

  if (print) {
    cat(styleShow)
  }

  return(invisible(styleShow))
})




Style$methods(as.list = function() {
  l <- list(
    "fontName" = fontName,
    "fontColour" = fontColour,
    "fontSize" = fontSize,
    "fontFamily" = fontFamily,
    "fontScheme" = fontScheme,
    "fontDecoration" = fontDecoration,

    "borderTop" = borderTop,
    "borderLeft" = borderLeft,
    "borderRight" = borderRight,
    "borderBottom" = borderBottom,
    "borderTopColour" = borderTopColour,
    "borderLeftColour" = borderLeftColour,
    "borderRightColour" = borderRightColour,
    "borderBottomColour" = borderBottomColour,

    "halign" = halign,
    "valign" = valign,
    "indent" = indent,
    "textRotation" = textRotation,
    "numFmt" = numFmt,
    "fillFg" = fill$fillFg,
    "fillBg" = fill$fillBg,
    "wrapText" = wrapText,
    "locked" = locked,
    "hidden" = hidden,
    "xfId" = xfId
  )

  l[sapply(l, length) > 0]
})

#' @name writeDataTable
#' @title Write to a worksheet as an Excel table
#' @description Write to a worksheet and format as an Excel table
#' @param wb A Workbook object containing a
#' worksheet.
#' @param sheet The worksheet to write to. Can be the worksheet index or name.
#' @param x A dataframe.
#' @param startCol A vector specifying the starting column to write df
#' @param startRow A vector specifying the starting row to write df
#' @param xy An alternative to specifying startCol and startRow individually.
#' A vector of the form c(startCol, startRow)
#' @param colNames If `TRUE`, column names of x are written.
#' @param rowNames If `TRUE`, row names of x are written.
#' @param row.names,col.names Deprecated, please use `rowNames`, `colNames` instead
#' @param tableStyle Any excel table style name or "none" (see "formatting" vignette).
#' @param tableName name of table in workbook. The table name must be unique.
#' @param headerStyle Custom style to apply to column names.
#' @param withFilter If `TRUE` or `NA`, columns with have filters in the first row.
#' @param keepNA If `TRUE`, NA values are converted to #N/A (or `na.string`, if not NULL) in Excel, else NA cells will be empty.
#' @param na.string If not NULL, and if `keepNA` is `TRUE`, NA values are converted to this string in Excel.
#' @param sep Only applies to list columns. The separator used to collapse list columns to a character vector e.g. sapply(x$list_column, paste, collapse = sep).
#' @param stack If `TRUE` the new style is merged with any existing cell styles.  If FALSE, any
#' existing style is replaced by the new style.
#' \cr\cr
#' \cr**The below options correspond to Excel table options:**
#' \cr
#' \if{html}{\figure{tableoptions.png}{options: alt="Figure: table_options.png"}}
#' \if{latex}{\figure{tableoptions.pdf}{options: width=265px}}
#'
#' @param firstColumn logical. If TRUE, the first column is bold
#' @param lastColumn logical. If TRUE, the last column is bold
#' @param bandedRows logical. If TRUE, rows are colour banded
#' @param bandedCols logical. If TRUE, the columns are colour banded
#' @details columns of x with class Date/POSIXt, currency, accounting,
#' hyperlink, percentage are automatically styled as dates, currency, accounting,
#' hyperlinks, percentages respectively.
#' @seealso [addWorksheet()]
#' @seealso [writeData()]
#' @seealso [removeTable()]
#' @seealso [getTables()]
#' @export
#' @examples
#' ## see package vignettes for further examples.
#'
#' #####################################################################################
#' ## Create Workbook object and add worksheets
#' wb <- createWorkbook()
#' addWorksheet(wb, "S1")
#' addWorksheet(wb, "S2")
#' addWorksheet(wb, "S3")
#'
#'
#' #####################################################################################
#' ## -- write data.frame as an Excel table with column filters
#' ## -- default table style is "TableStyleMedium2"
#'
#' writeDataTable(wb, "S1", x = iris)
#'
#' writeDataTable(wb, "S2",
#'   x = mtcars, xy = c("B", 3), rowNames = TRUE,
#'   tableStyle = "TableStyleLight9"
#' )
#'
#' df <- data.frame(
#'   "Date" = Sys.Date() - 0:19,
#'   "T" = TRUE, "F" = FALSE,
#'   "Time" = Sys.time() - 0:19 * 60 * 60,
#'   "Cash" = paste("$", 1:20), "Cash2" = 31:50,
#'   "hLink" = "https://CRAN.R-project.org/",
#'   "Percentage" = seq(0, 1, length.out = 20),
#'   "TinyNumbers" = runif(20) / 1E9, stringsAsFactors = FALSE
#' )
#'
#' ## openxlsx will apply default Excel styling for these classes
#' class(df$Cash) <- c(class(df$Cash), "currency")
#' class(df$Cash2) <- c(class(df$Cash2), "accounting")
#' class(df$hLink) <- "hyperlink"
#' class(df$Percentage) <- c(class(df$Percentage), "percentage")
#' class(df$TinyNumbers) <- c(class(df$TinyNumbers), "scientific")
#'
#' writeDataTable(wb, "S3", x = df, startRow = 4, rowNames = TRUE, tableStyle = "TableStyleMedium9")
#'
#' #####################################################################################
#' ## Additional Header Styling and remove column filters
#'
#' writeDataTable(wb,
#'   sheet = 1, x = iris, startCol = 7, headerStyle = createStyle(textRotation = 45),
#'   withFilter = FALSE
#' )
#'
#'
#' #####################################################################################
#' ## Save workbook
#' ## Open in excel without saving file: openXL(wb)
#' \dontrun{
#' saveWorkbook(wb, "writeDataTableExample.xlsx", overwrite = TRUE)
#' }
#'
#'
#'
#'
#'
#' #####################################################################################
#' ## Pre-defined table styles gallery
#'
#' wb <- createWorkbook(paste0("tableStylesGallery.xlsx"))
#' addWorksheet(wb, "Style Samples")
#' for (i in 1:21) {
#'   style <- paste0("TableStyleLight", i)
#'   writeDataTable(wb,
#'     x = data.frame(style), sheet = 1,
#'     tableStyle = style, startRow = 1, startCol = i * 3 - 2
#'   )
#' }
#'
#' for (i in 1:28) {
#'   style <- paste0("TableStyleMedium", i)
#'   writeDataTable(wb,
#'     x = data.frame(style), sheet = 1,
#'     tableStyle = style, startRow = 4, startCol = i * 3 - 2
#'   )
#' }
#'
#' for (i in 1:11) {
#'   style <- paste0("TableStyleDark", i)
#'   writeDataTable(wb,
#'     x = data.frame(style), sheet = 1,
#'     tableStyle = style, startRow = 7, startCol = i * 3 - 2
#'   )
#' }
#'
#' ## openXL(wb)
#' \dontrun{
#' saveWorkbook(wb, file = "tableStylesGallery.xlsx", overwrite = TRUE)
#' }
#'
writeDataTable <- function(
  wb,
  sheet,
  x,
  startCol    = 1,
  startRow    = 1,
  xy          = NULL,
  colNames    = TRUE,
  rowNames    = FALSE,
  tableStyle  = openxlsx_getOp("tableStyle", "TableStyleLight9"),
  tableName   = NULL,
  headerStyle = openxlsx_getOp("headerStyle"),
  withFilter  = openxlsx_getOp("withFilter", TRUE),
  keepNA      = openxlsx_getOp("keepNA", FALSE),
  na.string   = openxlsx_getOp("na.string"),
  sep         = ", ",
  stack       = FALSE,
  firstColumn = openxlsx_getOp("firstColumn", FALSE),
  lastColumn  = openxlsx_getOp("lastColumn", FALSE),
  bandedRows  = openxlsx_getOp("bandedRows", TRUE),
  bandedCols  = openxlsx_getOp("bandedCols", FALSE),
  col.names,
  row.names
  ) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  ## increase scipen to avoid writing in scientific

  if (!missing(row.names)) {
    warning("Please use 'rowNames' instead of 'row.names'", call. = FALSE)
    row.names <- rowNames
  }

  if (!missing(col.names)) {
    warning("Please use 'colNames' instead of 'col.names'", call. = FALSE)
    colNames <- col.names
  }

  # Set NULLs
  withFilter  <- withFilter  %||% TRUE
  keepNA      <- keepNA      %||% FALSE
  firstColumn <- firstColumn %||% FALSE
  lastColumn  <- lastColumn  %||% FALSE
  bandedRows  <- bandedRows  %||% TRUE
  bandedCols  <- bandedCols  %||% FALSE
  withFilter  <- withFilter  %||% TRUE

  if (!is.null(xy)) {
    if (length(xy) != 2) {
      stop("xy parameter must have length 2")
    }
    startCol <- xy[[1]]
    startRow <- xy[[2]]
  }

  # Assert parameters
  assert_class(wb, "Workbook")
  assert_class(x, "data.frame")
  assert_true_false(colNames)
  assert_true_false(rowNames)
  assert_class(headerStyle, "Style", or_null = TRUE)
  assert_true_false(withFilter)
  assert_character1(sep)
  assert_true_false(firstColumn)
  assert_true_false(lastColumn)
  assert_true_false(bandedRows)
  assert_true_false(bandedCols)

  if (is.null(tableName)) {
    tableName <- sprintf("Table%i", length(wb$tables) + 3L)
  } else {
    tableName <- wb$validate_table_name(tableName)
  }

  ## convert startRow and startCol
  if (!is.numeric(startCol)) {
    startCol <- convertFromExcelRef(startCol)
  }
  startRow <- as.integer(startRow)

  ## Coordinates for each section
  if (rowNames) {
    x <- cbind(data.frame("row names" = rownames(x)), as.data.frame(x))
  }

  ## If 0 rows append a blank row

  tableStyle <- validate_StyleName(tableStyle)

  ## header style
  if (inherits(headerStyle, "Style")) {
    addStyle(
      wb         = wb,
      sheet      = sheet,
      style      = headerStyle,
      rows       = startRow,
      cols       = 0:(ncol(x) - 1L) + startCol,
      gridExpand = TRUE
    )
  }

  showColNames <- colNames

  if (colNames) {
    colNames <- colnames(x)
    assert_unique(colNames, case_sensitive = FALSE)

    ## zero char names are invalid
    char0 <- nchar(colNames) == 0
    if (any(char0)) {
      colNames[char0] <- colnames(x)[char0] <- paste0("Column", which(char0))
    }

    # Compatibility with MS Excel: throw warning if a table column name exceeds
    # the length of 255 chars.
    char_over255 <- nchar(colNames) > 255
    if (any(char_over255)) {
      warning_msg <- sprintf(
        "Column name exceeds 255 chars, possible incompatibility with MS Excel. Index: %s",
        toString(which(char_over255)))
      warning(warning_msg)
    }

  } else {
    colNames <- paste0("Column", seq_along(x))
    names(x) <- colNames
  }

  ## If zero rows, append an empty row (prevent XML from corrupting)
  if (nrow(x) == 0) {
    x <- rbind(
      as.data.frame(x),
      matrix("", nrow = 1, ncol = ncol(x), dimnames = list(character(), colnames(x)))
    )
    names(x) <- colNames
  }

  ref1 <- paste0(convert_to_excel_ref(cols = startCol, LETTERS = LETTERS), startRow)
  ref2 <- paste0(convert_to_excel_ref(cols = startCol + ncol(x) - 1, LETTERS = LETTERS), startRow + nrow(x))
  ref <- paste(ref1, ref2, sep = ":")

  ## check not overwriting another table
  wb$check_overwrite_tables(
    sheet = sheet,
    new_rows = c(startRow, startRow + nrow(x) - 1L + 1L), ## + header
    new_cols = c(startCol, startCol + ncol(x) - 1L)
  )


  ## column class styling
  # consider not using lowercase and instead use inherits(x, class)
  colClasses <- lapply(x, function(x) tolower(class(x)))
  classStyles(
    wb,
    sheet      = sheet,
    startRow   = startRow,
    startCol   = startCol,
    colNames   = TRUE,
    nRow       = nrow(x),
    colClasses = colClasses,
    stack      = stack
  )

  ## write data to worksheet
  wb$writeData(
    df         = x,
    colNames   = TRUE,
    sheet      = sheet,
    startRow   = startRow,
    startCol   = startCol,
    colClasses = colClasses,
    hlinkNames = NULL,
    keepNA     = keepNA,
    na.string  = na.string,
    list_sep   = sep
  )

  ## replace invalid XML characters
  colNames <- replaceIllegalCharacters(colNames)

  ## create table.xml and assign an id to worksheet tables
  wb$buildTable(
    sheet             = sheet,
    colNames          = colNames,
    ref               = ref,
    showColNames      = showColNames,
    tableStyle        = tableStyle,
    tableName         = tableName,
    withFilter        = withFilter[1],
    totalsRowCount    = 0L,
    showFirstColumn   = firstColumn[1],
    showLastColumn    = lastColumn[1],
    showRowStripes    = bandedRows[1],
    showColumnStripes = bandedCols[1]
  )
}

#' @include class_definitions.R


Workbook$methods(setColWidths = function(sheet) {
  sheet <- validateSheet(sheet)

  widths <- colWidths[[sheet]]
  hidden <- attr(colWidths[[sheet]], "hidden", exact = TRUE)
  if (length(hidden) != length(widths)) {
    hidden <- rep("0", length(widths))
  }

  cols <- names(colWidths[[sheet]])

  autoColsInds <- widths %in% c("auto", "auto2")
  autoCols <- cols[autoColsInds]

  ## If any not auto
  if (any(!autoColsInds)) {
    widths[!autoColsInds] <- as.numeric(widths[!autoColsInds]) + 0.71
  }

  ## If any auto
  if (length(autoCols) > 0) {

    ## only run if data on worksheet
    if (worksheets[[sheet]]$sheet_data$n_elements == 0) {
      missingAuto <- autoCols
    } else if (all(is.na(worksheets[[sheet]]$sheet_data$v))) {
      missingAuto <- autoCols
    } else {

      ## First thing - get base font max character width
      baseFont <- getBaseFont()
      baseFontName <- unlist(baseFont$name, use.names = FALSE)
      if (is.null(baseFontName)) {
        baseFontName <- "calibri"
      } else {
        baseFontName <- gsub(" ", ".", tolower(baseFontName), fixed = TRUE)
        if (!baseFontName %in% names(openxlsxFontSizeLookupTable)) {
          baseFontName <- "calibri"
        }
      }

      baseFontSize <- unlist(baseFont$size, use.names = FALSE)
      if (is.null(baseFontSize)) {
        baseFontSize <- 11
      } else {
        baseFontSize <- as.numeric(baseFontSize)
        baseFontSize <- ifelse(baseFontSize < 8, 8, ifelse(baseFontSize > 36, 36, baseFontSize))
      }

      baseFontCharWidth <- openxlsxFontSizeLookupTable[[baseFontName]][baseFontSize - 7]
      allCharWidths <- rep(baseFontCharWidth, worksheets[[sheet]]$sheet_data$n_elements)
      ######### ----------------------------------------------------------------

      ## get char widths for each style object
      if (length(styleObjects) > 0 & any(!is.na(worksheets[[sheet]]$sheet_data$style_id))) {
        thisSheetName <- sheet_names[sheet]

        ## Calc font width for all styles on this worksheet
        styleIds <- worksheets[[sheet]]$sheet_data$style_id
        styObSubet <- styleObjects[sort(unique(styleIds))]
        stySubset <- lapply(styObSubet, "[[", "style")

        ## loop through stlye objects assignin a charWidth else baseFontCharWidth
        styleCharWidths <- sapply(stySubset, get_style_max_char_width, USE.NAMES = FALSE)


        ## Now assign all cells a character width
        allCharWidths <- styleCharWidths[worksheets[[sheet]]$sheet_data$style_id]
        allCharWidths[is.na(allCharWidths)] <- baseFontCharWidth
      }

      ## Now check for columns that are auto2
      auto2Inds <- which(widths %in% "auto2")
      if (length(auto2Inds) > 0 & length(worksheets[[sheet]]$mergeCells) > 0) {

        ## get cell merges
        merged_cells <- regmatches(worksheets[[sheet]]$mergeCells, regexpr("[A-Z0-9]+:[A-Z0-9]+", worksheets[[sheet]]$mergeCells))

        comps <- lapply(merged_cells, function(rectCoords) unlist(strsplit(rectCoords, split = ":")))
        merge_cols <- lapply(comps, convertFromExcelRef)
        merge_cols <- lapply(merge_cols, function(x) x[x %in% cols[auto2Inds]]) ## subset to auto2Inds

        merge_rows <- lapply(comps, function(x) as.numeric(gsub("[A-Z]", "", x, perl = TRUE)))
        merge_rows <- merge_rows[sapply(merge_cols, length) > 0]
        merge_cols <- merge_cols[sapply(merge_cols, length) > 0]

        sd <- worksheets[[sheet]]$sheet_data

        if (length(merge_cols) > 0) {
          all_merged_cells <- lapply(seq_along(merge_cols), function(i) {
            expand.grid(
              "rows" = min(merge_rows[[i]]):max(merge_rows[[i]]),
              "cols" = min(merge_cols[[i]]):max(merge_cols[[i]])
            )
          })

          all_merged_cells <- do.call("rbind", all_merged_cells)

          ## only want the sheet data in here
          refs <- pair_rc(all_merged_cells[[1]], all_merged_cells[[2]])
          existing_cells <- pair_rc(worksheets[[sheet]]$sheet_data$rows, 
                                    worksheets[[sheet]]$sheet_data$cols)
          keep <- which(!existing_cells %in% refs & !is.na(worksheets[[sheet]]$sheet_data$v))

          sd <- Sheet_Data$new()
          sd$cols <- worksheets[[sheet]]$sheet_data$cols[keep]
          sd$t <- worksheets[[sheet]]$sheet_data$t[keep]
          sd$v <- worksheets[[sheet]]$sheet_data$v[keep]
          sd$n_elements <- length(sd$cols)
          allCharWidths <- allCharWidths[keep]
        } else {
          sd <- worksheets[[sheet]]$sheet_data
        }
      } else {
        sd <- worksheets[[sheet]]$sheet_data
      }

      ## Now that we have the max character width for the largest font on the page calculate the column widths
      calculatedWidths <- calc_column_widths(
        sheet_data = sd,
        sharedStrings = unlist(sharedStrings, use.names = FALSE),
        autoColumns = as.integer(autoCols),
        widths = allCharWidths,
        baseFontCharWidth = baseFontCharWidth,
        minW = openxlsx_getOp("minWidth", 3),
        maxW = openxlsx_getOp("maxWidth", 250)
      )

      missingAuto <- autoCols[!autoCols %in% names(calculatedWidths)]
      widths[names(calculatedWidths)] <- calculatedWidths + 0.71
    }

    widths[missingAuto] <- 9.15
  }

  # Check if any conflicting existing levels
  if (any(cols %in% names(worksheets[[sheet]]$cols))) {

    for (i in intersect(cols, names(worksheets[[sheet]]$cols))) {
      
      width_hidden <- attr(colWidths[[sheet]], "hidden")[attr(colWidths[[sheet]], "names") == i]
      width_widths <- as.numeric(colWidths[[sheet]][attr(colWidths[[sheet]], "names") == i]) + 0.71

      # If column already has a custom width, just update the width and hidden attributes
      if (grepl("customWidth", worksheets[[sheet]]$cols[[i]])) {
        worksheets[[sheet]]$cols[[i]] <<- sub('(width=\\").*?(\\"\\shidden=\\").*?(\\")', paste0("\\1", width_widths, "\\2", width_hidden, "\\3"), worksheets[[sheet]]$cols[[i]], perl = TRUE)
      } else {
      # If column exists, but doesn't have a custom width
        worksheets[[sheet]]$cols[[i]] <<- sub("((?<=hidden=\")(\\w)\")", paste0(width_hidden, "\" width=\"", width_widths, "\" customWidth=\"1\"/>"), worksheets[[sheet]]$cols[[i]], perl = TRUE)
      }
    }

    remaining_cols <- !cols %in% names(worksheets[[sheet]]$cols)
    cols <- cols[remaining_cols]
    widths <- widths[remaining_cols]
    hidden <- hidden[remaining_cols]
  }

  # Add remaining columns
  if (length(cols) > 0) {
    colNodes <- sprintf('<col min="%s" max="%s" width="%s" hidden="%s" customWidth="1"/>', cols, cols, widths, hidden)
    names(colNodes) <- cols
    worksheets[[sheet]]$cols <<- append(worksheets[[sheet]]$cols, colNodes)
  }
  
})




get_style_max_char_width <- function(thisStyle) {
  fN <- unlist(thisStyle$fontName, use.names = FALSE)
  if (is.null(fN)) {
    fN <- "calibri"
  } else {
    fN <- gsub(" ", ".", tolower(fN), fixed = TRUE)
    if (!fN %in% names(openxlsxFontSizeLookupTable)) {
      fN <- "calibri"
    }
  }

  fS <- unlist(thisStyle$fontSize, use.names = FALSE)
  if (is.null(fS)) {
    fS <- 11
  } else {
    fS <- as.numeric(fS)
    fS <- ifelse(fS < 8, 8, ifelse(fS > 36, 36, fS))
  }

  if ("BOLD" %in% thisStyle$fontDecoration) {
    styleMaxCharWidth <- openxlsxFontSizeLookupTableBold[[fN]][fS - 7]
  } else {
    styleMaxCharWidth <- openxlsxFontSizeLookupTable[[fN]][fS - 7]
  }

  return(styleMaxCharWidth)
}

Hyperlink <- setRefClass("Hyperlink",
  fields = c(
    "ref",
    "target",
    "location",
    "display",
    "is_external"
  ),

  methods = list()
)


Hyperlink$methods(initialize = function(ref, target, location, display = NULL, is_external = TRUE) {
  ref <<- ref
  target <<- target
  location <<- location
  display <<- display
  is_external <<- is_external
})

Hyperlink$methods(to_xml = function(id) {
  loc <- sprintf('location="%s"', location)
  disp <- sprintf('display="%s"', display)
  rf <- sprintf('ref="%s"', ref)

  if (is_external) {
    rid <- sprintf('r:id="rId%s"', id)
  } else {
    rid <- NULL
  }

  paste("<hyperlink", rf, rid, disp, loc, "/>")
})

Hyperlink$methods(to_target_xml = function(id) {
  if (is_external) {
    return(sprintf('<Relationship Id="rId%s" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/hyperlink" Target="%s" TargetMode="External"/>', id, target))
  } else {
    return(NULL)
  }
})



xml_to_hyperlink <- function(xml) {

  # xml <- c('<hyperlink ref="A1" r:id="rId1" location="Authority"/>',
  # '<hyperlink ref="B1" r:id="rId2"/>',
  # '<hyperlink ref="A1" location="Sheet2!A1" display="Sheet2!A1"/>')

  if (length(xml) == 0) {
    return(xml)
  }

  targets <- names(xml)
  if (is.null(targets)) {
    targets <- rep(NA, length(xml))
  }

  xml <- unname(xml)

  a <- unlist(lapply(xml, function(x) regmatches(x, gregexpr('[a-zA-Z]+=".*?"', x))), recursive = FALSE)
  names <- lapply(a, function(xml) regmatches(xml, regexpr('[a-zA-Z]+(?=\\=".*?")', xml, perl = TRUE)))
  vals <- lapply(a, function(xml) regmatches(xml, regexpr('(?<=").*?(?=")', xml, perl = TRUE)))
  vals <- lapply(vals, function(x) {
    Encoding(x) <- "UTF-8"
    x
  })

  hyperlink_objects <- lapply(seq_along(xml), function(i) {
    tmp_vals <- vals[[i]]
    tmp_nms <- names[[i]]
    names(tmp_vals) <- tmp_nms

    ## ref
    ref <- tmp_vals[["ref"]]

    ## location
    if ("location" %in% tmp_nms) {
      location <- tmp_vals[["location"]]
    } else {
      location <- NULL
    }

    ## location
    if ("display" %in% tmp_nms) {
      display <- tmp_vals[["display"]]
    } else {
      display <- NULL
    }

    ## target/external
    if (is.na(targets[i])) {
      target <- NULL
      is_external <- FALSE
    } else {
      is_external <- TRUE
      target <- targets[i]
    }

    Hyperlink$new(ref = ref, target = target, location = location, display = display, is_external = is_external)
  })

  return(hyperlink_objects)
}

#' Set and Get Window Size for xlsx file
#'
#' @param wb A Workbook object
#' @param xWindow the horizontal coordinate of the top left corner of the window
#' @param yWindow the vertical coordinate of the top left corner of the window
#' @param windowWidth the width of the window
#' @param windowHeight the height of the window
#' 
#' Set the size and position of the window when you open the xlsx file.  The units are in twips.  See
#' [Microsoft's documentation for the xlsx standard](https://learn.microsoft.com/en-us/dotnet/api/documentformat.openxml.spreadsheet.workbookview?view=openxml-2.8.1)
#'
#' @export
#'
#' @examples
#' ## Create Workbook object and add worksheets
#' wb <- createWorkbook()
#' addWorksheet(wb, "S1")
#' getWindowSize(wb)
#' setWindowSize(wb, windowWidth = 10000)
setWindowSize <- function(wb, xWindow = NULL, yWindow = NULL, windowWidth = NULL, windowHeight= NULL) {

  bookViews <- wb$workbook$bookViews

  if(!is.null(xWindow)) {
    if(as.integer(xWindow) >= 0L) {
      bookViews <- sub("xWindow=\"\\d+", paste0("xWindow=\"", xWindow), bookViews)
    } else {
      stop("xWindow must be >= 0")
    }
  }

  if(!is.null(yWindow)) {
    if(as.integer(yWindow) >= 0L) {
      bookViews <- sub("yWindow=\"\\d+", paste0("yWindow=\"", yWindow), bookViews)
    } else {
      stop("yWindow must be >= 0")
    }
  }

  if(!is.null(windowWidth)) {
    if(as.integer(windowWidth) >= 100L) {
      bookViews <- sub("windowWidth=\"\\d+", paste0("windowWidth=\"", windowWidth), bookViews)
    } else {
      stop("windowWidth must be >= 100")
    }
  }

  if(!is.null(windowHeight)) {
    if(as.integer(windowHeight) >= 100L) {
      bookViews <- sub("windowHeight=\"\\d+", paste0("windowHeight=\"", windowHeight), bookViews)
    } else {
      stop("windowHeight must be >= 100")
    }
  }

  wb$workbook$bookViews <- bookViews
}

#' @rdname setWindowSize
#' @export

getWindowSize <- function(wb) {
  bookViews <- wb$workbook$bookViews

  c(getAttrs(bookViews, "xWindow"),
    getAttrs(bookViews, "yWindow"),
    getAttrs(bookViews, "windowWidth"),
    getAttrs(bookViews, "windowHeight"))

}

#' @include class_definitions.R


WorkSheet$methods(initialize = function(
  showGridLines = TRUE,
  tabSelected = FALSE,
  tabColour = NULL,
  zoom = 100,
  
  oddHeader = NULL,
  oddFooter = NULL,
  evenHeader = NULL,
  evenFooter = NULL,
  firstHeader = NULL,
  firstFooter = NULL,
  
  paperSize = 9,
  orientation = "portrait",
  hdpi = 300,
  vdpi = 300
  ) {
  if (!is.null(tabColour)) {
    tabColour <- sprintf('<sheetPr><tabColor rgb="%s"/></sheetPr>', tabColour)
  } else {
    tabColour <- character(0)
  }

  if (zoom < 10) {
    zoom <- 10
  } else if (zoom > 400) {
    zoom <- 400
  }

  naToNULLList <- function(x) {
    lapply(x, function(x) {
      if (is.na(x)) {
        return(NULL)
      }
      x
    })
  }

  # hf <- list(
  #   oddHeader = naToNULLList(oddHeader),
  #   oddFooter = naToNULLList(oddFooter),
  #   evenHeader = naToNULLList(evenHeader),
  #   evenFooter = naToNULLList(evenFooter),
  #   firstHeader = naToNULLList(firstHeader),
  #   firstFooter = naToNULLList(firstFooter)
  # )
  
  hf <- list(
    oddHeader   = oddHeader,
    oddFooter   = oddFooter,
    evenHeader  = evenHeader,
    evenFooter  = evenFooter,
    firstHeader = firstHeader,
    firstFooter = firstFooter
  )
  
  if (all(vapply(hf, is.null, NA))) {
    hf <- list()
  }

  ## list of all possible children
  sheetPr <<- tabColour
  dimension <<- '<dimension ref="A1"/>'
  sheetViews <<- sprintf('<sheetViews><sheetView workbookViewId="0" zoomScale="%s" showGridLines="%s" tabSelected="%s"/></sheetViews>', as.integer(zoom), as.integer(showGridLines), as.integer(tabSelected))
  sheetFormatPr <<- '<sheetFormatPr defaultRowHeight="15.0" baseColWidth="10"/>'
  cols <<- character(0)

  autoFilter <<- character(0)
  mergeCells <<- character(0)
  conditionalFormatting <<- character(0)
  dataValidations <<- NULL
  dataValidationsLst <<- character(0)
  hyperlinks <<- list()
  pageMargins <<- '<pageMargins left="0.7" right="0.7" top="0.75" bottom="0.75" header="0.3" footer="0.3"/>'
  pageSetup <<- sprintf('<pageSetup paperSize="%s" orientation="%s" horizontalDpi="%s" verticalDpi="%s" r:id="rId2"/>', paperSize, orientation, hdpi, vdpi) ## will always be 2
  headerFooter <<- hf
  rowBreaks <<- character(0)
  colBreaks <<- character(0)
  drawing <<- '<drawing r:id=\"rId1\"/>' ## will always be 1
  legacyDrawing <<- character(0)
  legacyDrawingHF <<- character(0)
  oleObjects <<- character(0)
  tableParts <<- character(0)
  extLst <<- character(0)

  freezePane <<- character(0)

  sheet_data <<- Sheet_Data$new()
})



WorkSheet$methods(get_prior_sheet_data = function() {
  xml <- '<worksheet xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing" xmlns:x14="http://schemas.microsoft.com/office/spreadsheetml/2009/9/main" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" mc:Ignorable="x14ac xr xr2 xr3" xmlns:x14ac="http://schemas.microsoft.com/office/spreadsheetml/2009/9/ac" xmlns:xr="http://schemas.microsoft.com/office/spreadsheetml/2014/revision" xmlns:xr2="http://schemas.microsoft.com/office/spreadsheetml/2015/revision2" xmlns:xr3="http://schemas.microsoft.com/office/spreadsheetml/2016/revision3">'

  if (length(sheetPr) > 0) {
    tmp <- sheetPr
    if (!any(grepl("<sheetPr", tmp, fixed = TRUE))) {
      tmp <- paste0("<sheetPr>", paste(tmp, collapse = ""), "</sheetPr>")
    }

    xml <- paste(xml, tmp, collapse = "")
  }

  if (length(dimension) > 0) {
    xml <- paste(xml, dimension, collapse = "")
  }

  ## sheetViews handled here
  if (length(freezePane) > 0) {
    xml <- paste(xml, gsub("/></sheetViews>", paste0(">", freezePane, "</sheetView></sheetViews>"), sheetViews, fixed = TRUE), collapse = "")
  } else if (length(sheetViews) > 0) {
    xml <- paste(xml, sheetViews, collapse = "")
  }

  if (length(sheetFormatPr) > 0) {
    xml <- paste(xml, sheetFormatPr, collapse = "")
  }

  if (length(cols) > 0) {
    xml <- paste(xml, pxml(c("<cols>", cols, "</cols>")), collapse = "")
  }


  return(xml)
})


WorkSheet$methods(get_post_sheet_data = function() {
  xml <- ""

  if (length(sheetProtection) > 0) {
    xml <- paste0(xml, sheetProtection, collapse = "")
  }

  if (length(autoFilter) > 0) {
    xml <- paste0(xml, autoFilter, collapse = "")
  }

  if (length(mergeCells) > 0) {
    xml <- paste0(xml, paste0(sprintf('<mergeCells count="%s">', length(mergeCells)), pxml(mergeCells), "</mergeCells>"), collapse = "")
  }

  if (length(conditionalFormatting) > 0) {
    nms <- names(conditionalFormatting)
    xml <- paste0(xml,
      paste(
        sapply(unique(nms), function(x) {
          paste0(
            sprintf('<conditionalFormatting sqref="%s">', x),
            pxml(conditionalFormatting[nms == x]),
            "</conditionalFormatting>"
          )
        }),
        collapse = ""
      ),
      collapse = ""
    )
  }


  if (length(dataValidations) > 0) {
    xml <- paste0(xml, paste0(sprintf('<dataValidations count="%s">', length(dataValidations)), pxml(dataValidations), "</dataValidations>"))
  }

  if (length(hyperlinks) > 0) {
    h_inds <- paste0(seq_along(hyperlinks), "h")
    xml <- paste(xml, paste("<hyperlinks>", paste(sapply(seq_along(h_inds), function(i) hyperlinks[[i]]$to_xml(h_inds[i])), collapse = ""), "</hyperlinks>"), collapse = "")
  }

  if (length(pageMargins) > 0) {
    xml <- paste0(xml, pageMargins, collapse = "")
  }

  if (length(pageSetup) > 0) {
    xml <- paste0(xml, pageSetup, collapse = "")
  }

  if (!identical(headerFooter, list()) && length(headerFooter) > 0) {
    xml <- paste0(xml, genHeaderFooterNode(headerFooter), collapse = "")
  }

  ## rowBreaks and colBreaks
  if (length(rowBreaks) > 0) {
    xml <- paste0(xml,
      paste0(sprintf('<rowBreaks count="%s" manualBreakCount="%s">', length(rowBreaks), length(rowBreaks)), paste(rowBreaks, collapse = ""), "</rowBreaks>"),
      collapse = ""
    )
  }

  if (length(colBreaks) > 0) {
    xml <- paste0(xml,
      paste0(sprintf('<colBreaks count="%s" manualBreakCount="%s">', length(colBreaks), length(colBreaks)), paste(colBreaks, collapse = ""), "</colBreaks>"),
      collapse = ""
    )
  }

  if (length(drawing) > 0) {
    xml <- paste0(xml, drawing, collapse = "")
  }

  if (length(legacyDrawing) > 0) {
    xml <- paste0(xml, legacyDrawing, collapse = "")
  }

  if (length(legacyDrawingHF) > 0) {
    xml <- paste0(xml, legacyDrawingHF, collapse = "")
  }

  if (length(oleObjects) > 0) {
    xml <- paste0(xml, oleObjects, collapse = "")
  }

  if (length(tableParts) > 0) {
    xml <- paste0(xml,
      paste0(sprintf('<tableParts count="%s">', length(tableParts)), pxml(tableParts), "</tableParts>"),
      collapse = ""
    )
  }


  if (length(dataValidationsLst) > 0) {
    dataValidationsLst_xml <- paste0(sprintf('<ext uri="{CCE6A557-97BC-4b89-ADB6-D9C93CAAB3DF}" xmlns:x14="http://schemas.microsoft.com/office/spreadsheetml/2009/9/main"><x14:dataValidations count="%s" xmlns:xm="http://schemas.microsoft.com/office/excel/2006/main">', length(dataValidationsLst)),
      paste0(pxml(dataValidationsLst), "</x14:dataValidations></ext>"),
      collapse = ""
    )
  } else {
    dataValidationsLst_xml <- character(0)
  }


  if (length(extLst) > 0 || length(dataValidationsLst) > 0) {
    xml <- paste0(xml, sprintf("<extLst>%s</extLst>", paste0(pxml(extLst), dataValidationsLst_xml)))
  }

  xml <- paste0(xml, "</worksheet>")

  return(xml)
})


WorkSheet$methods(order_sheetdata = function() {
  if (sheet_data$n_elements == 0) {
    return(invisible(0))
  }

  if (sheet_data$data_count > 1) {
    ord <- order(sheet_data$rows, sheet_data$cols, method = "radix", na.last = TRUE)
    sheet_data$rows <<- sheet_data$rows[ord]
    sheet_data$cols <<- sheet_data$cols[ord]
    sheet_data$t <<- sheet_data$t[ord]
    sheet_data$v <<- sheet_data$v[ord]
    sheet_data$f <<- sheet_data$f[ord]

    sheet_data$style_id <<- sheet_data$style_id[ord]

    sheet_data$data_count <<- 1L

    dm1 <- paste0(int_2_cell_ref(cols = sheet_data$cols[1]), sheet_data$rows[1])
    dm2 <- paste0(int_2_cell_ref(cols = sheet_data$cols[sheet_data$n_elements]), sheet_data$rows[sheet_data$n_elements])

    if (length(dm1) == 1 & length(dm2) != 1) {
      if (!is.na(dm1) & !is.na(dm2) & dm1 != "NA" & dm2 != "NA") {
        dimension <<- sprintf("<dimension ref=\"%s:%s\"/>", dm1, dm2)
      }
    }
  }


  invisible(0)
})

Comment <- setRefClass("Comment",
  fields = c(
    "text",
    "author",
    "style",
    "visible",
    "width",
    "height"
  ),

  methods = list()
)


Comment$methods(initialize = function(text, author, style, visible = TRUE, width = 2, height = 4) {
  text <<- text
  author <<- author
  style <<- style
  visible <<- visible
  width <<- width
  height <<- height
})


Comment$methods(show = function() {
  showText <- sprintf("Author: %s\n", author)
  showText <- c(showText, sprintf("Text:\n %s\n\n", paste(text, collapse = "")))
  styleShow <- "Style:\n"

  if ("list" %in% class(style)) {
    for (i in seq_along(style)) {
      styleShow <- append(styleShow, sprintf("Font name: %s\n", style[[i]]$fontName[[1]])) ## Font name
      styleShow <- append(styleShow, sprintf("Font size: %s\n", style[[i]]$fontSize[[1]])) ## Font size
      styleShow <- append(styleShow, sprintf("Font colour: %s\n", gsub("^FF", "#", style[[i]]$fontColour[[1]]))) ## Font colour

      ## Font decoration
      if (length(style[[i]]$fontDecoration) > 0) {
        styleShow <- append(styleShow, sprintf("Font decoration: %s\n", paste(style[[i]]$fontDecoration, collapse = ", ")))
      }

      styleShow <- append(styleShow, "\n\n")
    }
  } else {
    styleShow <- append(styleShow, sprintf("Font name: %s \n", style$fontName[[1]])) ## Font name
    styleShow <- append(styleShow, sprintf("Font size: %s \n", style$fontSize[[1]])) ## Font size
    styleShow <- append(styleShow, sprintf("Font colour: %s \n", gsub("^FF", "#", style$fontColour[[1]]))) ## Font colour

    ## Font decoration
    if (length(style$fontDecoration) > 0) {
      styleShow <- append(styleShow, sprintf("Font decoration: %s \n", paste(style$fontDecoration, collapse = ", ")))
    }

    styleShow <- append(styleShow, "\n\n")
  }

  showText <- paste0(paste(showText, collapse = ""), paste(styleShow, collapse = ""), collapse = "")
  cat(showText)
})



#' @name createComment
#' @title create a Comment object
#' @description Create a cell Comment object to pass to writeComment()
#' @param comment Comment text. Character vector.
#' @param author Author of comment. Character vector of length 1
#' @param style A Style object or list of style objects the same length as comment vector. See [createStyle()].
#' @param visible TRUE or FALSE. Is comment visible.
#' @param width,height Width and height of textbook (in number of cells);
#'   doubles are rounded with \code{base::round()}
#' @export
#' @seealso [writeComment()]
#' @examples
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet 1")
#'
#' c1 <- createComment(comment = "this is comment")
#' writeComment(wb, 1, col = "B", row = 10, comment = c1)
#'
#' s1 <- createStyle(fontSize = 12, fontColour = "red", textDecoration = c("BOLD"))
#' s2 <- createStyle(fontSize = 9, fontColour = "black")
#'
#' c2 <- createComment(comment = c("This Part Bold red\n\n", "This part black"), style = c(s1, s2))
#' c2
#'
#' writeComment(wb, 1, col = 6, row = 3, comment = c2)
#' \dontrun{
#' saveWorkbook(wb, file = "createCommentExample.xlsx", overwrite = TRUE)
#' }
createComment <- function(comment,
                          author = Sys.getenv("USERNAME"),
                          style = NULL,
                          visible = TRUE,
                          width = 2,
                          height = 4) {
  
  if (!is.character(comment)) {
    stop("comment argument must be a character vector")
  }

  assert_character1(author)
  assert_numeric1(width)
  assert_numeric1(height)
  assert_true_false1(visible)

  width <- round(width)
  height <- round(height)

  if (is.null(style)) {
    style <- createStyle(fontName = "Tahoma", fontSize = 9, fontColour = "black")
  }

  author <- replaceIllegalCharacters(author)
  comment <- replaceIllegalCharacters(comment)

  invisible(Comment$new(text = comment, author = author, style = style, visible = visible, width = width[1], height = height[1]))
}


#' @name writeComment
#' @title write a cell comment
#' @description Write a Comment object to a worksheet
#' @param wb A workbook object
#' @param sheet A vector of names or indices of worksheets
#' @param col Column a column number of letter
#' @param row A row number.
#' @param comment A Comment object. See [createComment()].
#' @param xy An alternative to specifying `col` and
#' `row` individually.  A vector of the form
#' `c(col, row)`.
#' @export
#' @seealso [createComment()]
#' @examples
#' wb <- createWorkbook()
#' addWorksheet(wb, "Sheet 1")
#'
#' c1 <- createComment(comment = "this is comment")
#' writeComment(wb, 1, col = "B", row = 10, comment = c1)
#'
#' s1 <- createStyle(fontSize = 12, fontColour = "red", textDecoration = c("BOLD"))
#' s2 <- createStyle(fontSize = 9, fontColour = "black")
#'
#' c2 <- createComment(comment = c("This Part Bold red\n\n", "This part black"), style = c(s1, s2))
#' c2
#'
#' writeComment(wb, 1, col = 6, row = 3, comment = c2)
#' \dontrun{
#' saveWorkbook(wb, file = "writeCommentExample.xlsx", overwrite = TRUE)
#' }
writeComment <- function(wb, sheet, col, row, comment, xy = NULL) {
  if (!"Workbook" %in% class(wb)) {
    stop("First argument must be a Workbook.")
  }

  if (!"Comment" %in% class(comment)) {
    stop("comment argument must be a Comment object")
  }


  if (length(comment$style) == 1) {
    rPr <- wb$createFontNode(comment$style)
  } else {
    rPr <- sapply(comment$style, function(x) wb$createFontNode(x))
  }

  rPr <- gsub("font>", "rPr>", rPr)
  sheet <- wb$validateSheet(sheet)

  ## All input conversions/validations
  if (!is.null(xy)) {
    if (length(xy) != 2) {
      stop("xy parameter must have length 2")
    }
    col <- xy[[1]]
    row <- xy[[2]]
  }

  if (!is.numeric(col)) {
    col <- convertFromExcelRef(col)
  }

  ref <- paste0(convert_to_excel_ref(cols = col, LETTERS = LETTERS), row)

  comment_list <- list(
    "ref" = ref,
    "author" = comment$author,
    "comment" = comment$text,
    "style" = rPr,
    "clientData" = genClientData(col, row, visible = comment$visible, height = comment$height, width = comment$width)
  )

  wb$comments[[sheet]] <- append(wb$comments[[sheet]], list(comment_list))

  invisible(wb)
}


#' @name removeComment
#' @title Remove a comment from a cell
#' @description Remove a cell comment from a worksheet
#' @param wb A workbook object
#' @param sheet A vector of names or indices of worksheets
#' @param cols Columns to delete comments from
#' @param rows Rows to delete comments from
#' @param gridExpand If `TRUE`, all data in rectangle min(rows):max(rows) X min(cols):max(cols)
#' will be removed.
#' @export
#' @seealso [createComment()]
#' @seealso [writeComment()]
removeComment <- function(wb, sheet, cols, rows, gridExpand = TRUE) {
  sheet <- wb$validateSheet(sheet)

  assert_class(wb, "Workbook")
  cols <- convertFromExcelRef(cols)
  rows <- as.integer(rows)

  ## rows and cols need to be the same length
  if (gridExpand) {
    combs <- expand.grid(rows, cols)
    rows <- combs[, 1]
    cols <- combs[, 2]
  }

  if (length(rows) != length(cols)) {
    stop("Length of rows and cols must be equal.")
  }

  comb <- paste0(convert_to_excel_ref(cols = cols, LETTERS = LETTERS), rows)
  toKeep <- !sapply(wb$comments[[sheet]], "[[", "ref") %in% comb

  wb$comments[[sheet]] <- wb$comments[[sheet]][toKeep]
}

test_that("createComment() works", {
  # error checking
  expect_error(createComment("hi", width = 1), NA)
  expect_error(createComment("hi", width = 1L), NA)
  expect_error(createComment("hi", width = 1:2), "width")
  
  expect_error(createComment("hi", height = 1), NA)
  expect_error(createComment("hi", height = 1L), NA)
  expect_error(createComment("hi", height = 1:2), "height")
  
  expect_error(createComment("hi", visible = NULL))
  expect_error(createComment("hi", visible = c(TRUE, FALSE)), "visible")
  
  expect_error(createComment("hi", author = 1))
  expect_error(createComment("hi", author = c("a", "a")), "author")
  
  expect_s4_class(createComment("hello"), "Comment")
})

#' @name loadWorkbook
#' @title Load an existing .xlsx file
#' @author Alexander Walker, Philipp Schauberger
#' @param file A path to an existing .xlsx or .xlsm file
#' @param xlsxFile alias for file
#' @param isUnzipped Set to TRUE if the xlsx file is already unzipped
#' @param na.convert Should empty/blank cells be converted to `NA_character_`. Defaults to TRUE.
#' @description  loadWorkbook returns a workbook object conserving styles and
#' formatting of the original .xlsx file.
#' @return Workbook object.
#' @export
#' @seealso [removeWorksheet()]
#' @examples
#' ## load existing workbook from package folder
#' wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))
#' names(wb) # list worksheets
#' wb ## view object
#' ## Add a worksheet
#' addWorksheet(wb, "A new worksheet")
#'
#' ## Save workbook
#' \dontrun{
#' saveWorkbook(wb, "loadExample.xlsx", overwrite = TRUE)
#' }
#'
loadWorkbook <- function(file, xlsxFile = NULL, isUnzipped = FALSE, na.convert = TRUE) {

  ## If this is a unzipped workbook, skip the temp dir stuff
  if (isUnzipped) {
    xmlDir <- file
    xmlFiles <- list.files(path = xmlDir, full.names = TRUE, recursive = TRUE, all.files = TRUE)
  } else {
    if (!is.null(xlsxFile)) {
      file <- xlsxFile
    }

    file <- getFile(file)
    if (!file.exists(file)) {
      stop("File does not exist.")
    }

    ## create temp dir
    xmlDir <- tempfile()

    ## Unzip files to temp directory
    xmlFiles <- unzip(file, exdir = xmlDir)
  }
  wb <- createWorkbook()

  ## Not used
  # .relsXML           <- xmlFiles[grepl("_rels/.rels$", xmlFiles, perl = TRUE)]
  # appXML             <- xmlFiles[grepl("app.xml$", xmlFiles, perl = TRUE)]

  drawingsXML   <- grep("drawings/drawing[0-9]+.xml$", xmlFiles, perl = TRUE, value = TRUE)
  worksheetsXML <- grep("/worksheets/sheet[0-9]+", xmlFiles, perl = TRUE, value = TRUE)

  coreXML           <- grep("core.xml$", xmlFiles, perl = TRUE, value = TRUE)
  workbookXML       <- grep("workbook.xml$", xmlFiles, perl = TRUE, value = TRUE)
  stylesXML         <- grep("styles.xml$", xmlFiles, perl = TRUE, value = TRUE)
  sharedStringsXML  <- grep("sharedStrings.xml$", xmlFiles, perl = TRUE, value = TRUE)
  themeXML          <- grep("theme[0-9]+.xml$", xmlFiles, perl = TRUE, value = TRUE)
  drawingRelsXML    <- grep("drawing[0-9]+.xml.rels$", xmlFiles, perl = TRUE, value = TRUE)
  sheetRelsXML      <- grep("sheet[0-9]+.xml.rels$", xmlFiles, perl = TRUE, value = TRUE)
  media             <- grep("image[0-9]+.[a-z]+$", xmlFiles, perl = TRUE, value = TRUE)
  vmlDrawingXML     <- grep("drawings/vmlDrawing[0-9]+\\.vml$", xmlFiles, perl = TRUE, value = TRUE)
  vmlDrawingRelsXML <- grep("vmlDrawing[0-9]+.vml.rels$", xmlFiles, perl = TRUE, value = TRUE)
  commentsXML       <- grep("xl/comments[0-9]+\\.xml", xmlFiles, perl = TRUE, value = TRUE)
  threadCommentsXML <- grep("xl/threadedComments/threadedComment[0-9]+\\.xml", xmlFiles, perl = TRUE, value = TRUE)
  personXML         <- grep("xl/persons/person.xml$", xmlFiles, perl = TRUE, value = TRUE)
  embeddings        <- grep("xl/embeddings", xmlFiles, perl = TRUE, value = TRUE)
  charts            <- grep("xl/charts/.*xml$", xmlFiles, perl = TRUE, value = TRUE)
  chartsRels        <- grep("xl/charts/_rels", xmlFiles, perl = TRUE, value = TRUE)
  chartSheetsXML    <- grep("xl/chartsheets/sheet[0-9]+\\.xml", xmlFiles, perl = TRUE, value = TRUE)
  tablesXML         <- grep("tables/table[0-9]+.xml$", xmlFiles, perl = TRUE, value = TRUE)
  tableRelsXML      <- grep("table[0-9]+.xml.rels$", xmlFiles, perl = TRUE, value = TRUE)
  queryTablesXML    <- grep("queryTable[0-9]+.xml$", xmlFiles, perl = TRUE, value = TRUE)
  connectionsXML    <- grep("connections.xml$", xmlFiles, perl = TRUE, value = TRUE)
  extLinksXML       <- grep("externalLink[0-9]+.xml$", xmlFiles, perl = TRUE, value = TRUE)
  extLinksRelsXML   <- grep("externalLink[0-9]+.xml.rels$", xmlFiles, perl = TRUE, value = TRUE)


  # pivot tables
  pivotTableXML     <- grep("pivotTable[0-9]+.xml$", xmlFiles, perl = TRUE, value = TRUE)
  pivotTableRelsXML <- grep("pivotTable[0-9]+.xml.rels$", xmlFiles, perl = TRUE, value = TRUE)
  pivotDefXML       <- grep("pivotCacheDefinition[0-9]+.xml$", xmlFiles, perl = TRUE, value = TRUE)
  pivotDefRelsXML   <- grep("pivotCacheDefinition[0-9]+.xml.rels$", xmlFiles, perl = TRUE, value = TRUE)
  pivotCacheRecords <- grep("pivotCacheRecords[0-9]+.xml$", xmlFiles, perl = TRUE, value = TRUE)

  ## slicers
  slicerXML       <- grep("slicer[0-9]+.xml$", xmlFiles, perl = TRUE, value = TRUE)
  slicerCachesXML <- grep("slicerCache[0-9]+.xml$", xmlFiles, perl = TRUE, value = TRUE)

  ## VBA Macro
  vbaProject <- grep("vbaProject\\.bin$", xmlFiles, perl = TRUE, value = TRUE)

  ## remove all EXCEPT media and charts
  if (!isUnzipped) {
    on.exit({
      paths <- grep(
        "charts|media|vmlDrawing|comment|embeddings|pivot|slicer|vbaProject|person", 
        xmlFiles, 
        ignore.case = TRUE,
        value = TRUE,
        invert = TRUE
      )
      unlink(paths, recursive = TRUE, force = TRUE)
    },
      add = TRUE
    )
}

  ## core
  if (length(coreXML) == 1) {
    coreXML <- paste(readUTF8(coreXML), collapse = "")
    wb$core <- removeHeadTag(x = coreXML)
  }

  nSheets <- length(worksheetsXML) + length(chartSheetsXML)

  ## get Rid of chartsheets, these do not have a worksheet/sheeti.xml
  worksheet_rId_mapping <- NULL
  workbookRelsXML <- grep("workbook.xml.rels$", xmlFiles, perl = TRUE, value = TRUE)
  if (length(workbookRelsXML) > 0) {
    workbookRelsXML <- paste(readUTF8(workbookRelsXML), collapse = "")
    workbookRelsXML <- getChildlessNode(xml = workbookRelsXML, tag = "Relationship")
    worksheet_rId_mapping <- grep("worksheets/sheet", workbookRelsXML, fixed = TRUE, value = TRUE)
  }

  ##
  chartSheetRIds <- NULL
  if (length(chartSheetsXML) > 0) {
    workbookRelsXML <- grep("chartsheets/sheet", workbookRelsXML, fixed = TRUE, value = TRUE)

    chartSheetRIds <- unlist(getId(workbookRelsXML))
    chartsheet_rId_mapping <- unlist(regmatches(workbookRelsXML, gregexpr("sheet[0-9]+\\.xml", workbookRelsXML, perl = TRUE, ignore.case = TRUE)))

    sheetNo <- as.integer(regmatches(chartSheetsXML, regexpr("(?<=sheet)[0-9]+(?=\\.xml)", chartSheetsXML, perl = TRUE)))
    chartSheetsXML <- chartSheetsXML[order(sheetNo)]

    chartSheetsRelsXML <- grep("xl/chartsheets/_rels", xmlFiles, perl = TRUE, value = TRUE)
    sheetNo2 <- as.integer(regmatches(chartSheetsRelsXML, regexpr("(?<=sheet)[0-9]+(?=\\.xml\\.rels)", chartSheetsRelsXML, perl = TRUE)))
    chartSheetsRelsXML <- chartSheetsRelsXML[order(sheetNo2)]

    chartSheetsRelsDir <- dirname(chartSheetsRelsXML[1])
  }


  ## xl\
  ## xl\workbook
  if (length(workbookXML) > 0) {
    workbook <- readUTF8(workbookXML)
    workbook <- removeHeadTag(workbook)

    sheets <- unlist(regmatches(workbook, gregexpr("(?<=<sheets>).*(?=</sheets>)", workbook, perl = TRUE)))
    sheets <- unlist(regmatches(sheets, gregexpr("<sheet[^>]*>", sheets, perl = TRUE)))
    
    ## Some veryHidden sheets do not have a sheet content and their rId is empty.
    ## Such sheets need to be filtered out because otherwise their sheet names
    ## occur in the list of all sheet names, leading to a wrong association
    ## of sheet names with sheet indeces.
    sheets <- grep('r:id="[[:blank:]]*"', sheets, invert = TRUE, value = TRUE)


    ## sheetId is meaningless
    ## sheet rId links to the workbook.xml.resl which links worksheets/sheet(i).xml file
    ## order they appear here gives order of worksheets in xlsx file

    sheetrId <- unlist(getRId(sheets))
    sheetId <- unlist(regmatches(sheets, gregexpr('(?<=sheetId=")[0-9]+', sheets, perl = TRUE)))
    sheetNames <- unlist(regmatches(sheets, gregexpr('(?<=name=")[^"]+', sheets, perl = TRUE)))
    sheetNames <- replaceXMLEntities(sheetNames)


    is_chart_sheet <- sheetrId %in% chartSheetRIds
    is_visible <- !grepl("hidden", sheets)
    if (length(is_visible) != length(sheetrId)) {
      is_visible <- rep(TRUE, length(sheetrId))
    }
    

# #active sheet -----------------------------------------------------------

      
   
    
    ## add worksheets to wb
    j <- 1
    for (i in seq_along(sheetrId)) {
      if (is_chart_sheet[i]) {
        # count <- 0  variable not used
        txt <- paste(readUTF8(chartSheetsXML[j]), collapse = "")

        zoom <- regmatches(txt, regexpr('(?<=zoomScale=")[0-9]+', txt, perl = TRUE))
        if (length(zoom) == 0) {
          zoom <- 100
        }

        tabColour <- getChildlessNode(xml = txt, tag = "tabColor")
        if (length(tabColour) == 0) {
          tabColour <- NULL
        }

        j <- j + 1L

        wb$addChartSheet(sheetName = sheetNames[i], tabColour = tabColour, zoom = as.numeric(zoom))
      } else {
        wb$addWorksheet(sheetNames[i], visible = is_visible[i])
      }
    }


    ## replace sheetId
    for (i in 1:nSheets) {
      wb$workbook$sheets[[i]] <- gsub(sprintf(' sheetId="%s"', i), sprintf(' sheetId="%s"', sheetId[i]), wb$workbook$sheets[[i]])
    }


    ## additional workbook attributes
    calcPr <- getChildlessNode(xml = workbook, tag = "calcPr")
    if (length(calcPr) > 0) {
      wb$workbook$calcPr <- calcPr
    }

    workbookPr <- getChildlessNode(xml = workbook, tag = "workbookPr")
    if (length(workbookPr) > 0) {
      wb$workbook$workbookPr <- workbookPr
    }

    workbookProtection <- getChildlessNode(xml = workbook, tag = "workbookProtection")
    if (length(workbookProtection) > 0) {
      wb$workbook$workbookProtection <- workbookProtection
    }


    ## defined Names
    dNames <- getNodes(xml = workbook, tagIn = "<definedNames>")
    if (length(dNames) > 0) {
      dNames <- gsub("^<definedNames>|</definedNames>$", "", dNames)
      wb$workbook$definedNames <- paste0(getNodes(xml = dNames, tagIn = "<definedName"), ">")
    }
  }





  ## xl\sharedStrings
  if (length(sharedStringsXML) > 0) {
    sharedStrings <- readUTF8(sharedStringsXML)
    sharedStrings <- paste(sharedStrings, collapse = "\n")
    sharedStrings <- removeHeadTag(sharedStrings)

    uniqueCount <- as.integer(regmatches(sharedStrings, regexpr('(?<=uniqueCount=")[0-9]+', sharedStrings, perl = TRUE)))

    ## read in and get <si> nodes
    vals <- getNodes(xml = sharedStrings, tagIn = "<si>")

    if ("<si><t/></si>" %in% vals) {
      if (na.convert) {
        vals[vals == "<si><t/></si>"] <- "<si><t>NA</t></si>"
      }
      Encoding(vals) <- "UTF-8"
      attr(vals, "uniqueCount") <- uniqueCount - 1L
    } else {
      Encoding(vals) <- "UTF-8"
      attr(vals, "uniqueCount") <- uniqueCount
    }

    wb$sharedStrings <- vals
  }

  ## xl\pivotTables & xl\pivotCache
  if (length(pivotTableXML) > 0) {

    # pivotTable cacheId links to workbook.xml which links to workbook.xml.rels via rId
    # we don't modify the cacheId, only the rId
    nPivotTables <- length(pivotDefXML)
    rIds <- 20000L + 1:nPivotTables

    ## pivot tables
    pivotTableXML <- pivotTableXML[order(nchar(pivotTableXML), pivotTableXML)]
    pivotTableRelsXML <- pivotTableRelsXML[order(nchar(pivotTableRelsXML), pivotTableRelsXML)]

    ## Cache
    pivotDefXML <- pivotDefXML[order(nchar(pivotDefXML), pivotDefXML)]
    pivotDefRelsXML <- pivotDefRelsXML[order(nchar(pivotDefRelsXML), pivotDefRelsXML)]
    pivotCacheRecords <- pivotCacheRecords[order(nchar(pivotCacheRecords), pivotCacheRecords)]


    wb$pivotDefinitionsRels <- character(nPivotTables)

    pivot_content_type <- NULL

    if (length(pivotTableRelsXML) > 0) {
      wb$pivotTables.xml.rels <- unlist(lapply(pivotTableRelsXML, function(x) removeHeadTag(cppReadFile(x))))
    }


    # ## Check what caches are used
    cache_keep <- unlist(regmatches(wb$pivotTables.xml.rels, gregexpr("(?<=pivotCache/pivotCacheDefinition)[0-9](?=\\.xml)",
      wb$pivotTables.xml.rels,
      perl = TRUE, ignore.case = TRUE
    )))

    ## pivot cache records
    tmp <- unlist(regmatches(pivotCacheRecords, gregexpr("(?<=pivotCache/pivotCacheRecords)[0-9]+(?=\\.xml)", pivotCacheRecords, perl = TRUE, ignore.case = TRUE)))
    pivotCacheRecords <- pivotCacheRecords[tmp %in% cache_keep]

    ## pivot cache definitions rels
    tmp <- unlist(regmatches(pivotDefRelsXML, gregexpr("(?<=_rels/pivotCacheDefinition)[0-9]+(?=\\.xml)", pivotDefRelsXML, perl = TRUE, ignore.case = TRUE)))
    pivotDefRelsXML <- pivotDefRelsXML[tmp %in% cache_keep]

    ## pivot cache definitions
    tmp <- unlist(regmatches(pivotDefXML, gregexpr("(?<=pivotCache/pivotCacheDefinition)[0-9]+(?=\\.xml)", pivotDefXML, perl = TRUE, ignore.case = TRUE)))
    pivotDefXML <- pivotDefXML[tmp %in% cache_keep]



    if (length(pivotTableXML) > 0) {
      wb$pivotTables[seq_along(pivotTableXML)] <- pivotTableXML
      pivot_content_type <- c(
        pivot_content_type,
        sprintf('<Override PartName="/xl/pivotTables/pivotTable%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.pivotTable+xml"/>', seq_along(pivotTableXML))
      )
    }

    if (length(pivotDefXML) > 0) {
      wb$pivotDefinitions[seq_along(pivotDefXML)] <- pivotDefXML
      pivot_content_type <- c(
        pivot_content_type,
        sprintf('<Override PartName="/xl/pivotCache/pivotCacheDefinition%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.pivotCacheDefinition+xml"/>', seq_along(pivotDefXML))
      )
    }

    if (length(pivotCacheRecords) > 0) {
      wb$pivotRecords[seq_along(pivotCacheRecords)] <- pivotCacheRecords
      pivot_content_type <- c(
        pivot_content_type,
        sprintf('<Override PartName="/xl/pivotCache/pivotCacheRecords%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.pivotCacheRecords+xml"/>', seq_along(pivotCacheRecords))
      )
    }

    if (length(pivotDefRelsXML) > 0) {
      wb$pivotDefinitionsRels[seq_along(pivotDefRelsXML)] <- pivotDefRelsXML
    }




    ## update content_types
    wb$Content_Types <- c(wb$Content_Types, pivot_content_type)


    ## workbook rels
    wb$workbook.xml.rels <- c(
      wb$workbook.xml.rels,
      sprintf('<Relationship Id="rId%s" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/pivotCacheDefinition" Target="pivotCache/pivotCacheDefinition%s.xml"/>', rIds, seq_along(pivotDefXML))
    )


    caches <- getNodes(xml = workbook, tagIn = "<pivotCaches>")
    caches <- getChildlessNode(xml = caches, tag = "pivotCache")
    for (i in seq_along(caches)) {
      caches[i] <- gsub('"rId[0-9]+"', sprintf('"rId%s"', rIds[i]), caches[i])
    }

    wb$workbook$pivotCaches <- paste0("<pivotCaches>", paste(caches, collapse = ""), "</pivotCaches>")
  }

  ## xl\vbaProject
  if (length(vbaProject) > 0) {
    wb$vbaProject <- vbaProject
    wb$Content_Types[grepl('<Override PartName="/xl/workbook.xml" ', wb$Content_Types)] <- '<Override PartName="/xl/workbook.xml" ContentType="application/vnd.ms-excel.sheet.macroEnabled.main+xml"/>'
    wb$Content_Types <- c(wb$Content_Types, '<Override PartName="/xl/vbaProject.bin" ContentType="application/vnd.ms-office.vbaProject"/>')
  }


  ## xl\styles
  if (length(stylesXML) > 0) {
    styleObjects <- wb$loadStyles(stylesXML)
  } else {
    styleObjects <- list()
  }

  ## xl\media
  if (length(media) > 0) {
    mediaNames <- regmatches(media, regexpr("image[0-9]+\\.[a-z]+$", media))
    fileTypes <- unique(gsub("image[0-9]+\\.", "", mediaNames))

    contentNodes <- sprintf('<Default Extension="%s" ContentType="image/%s"/>', fileTypes, fileTypes)
    contentNodes[fileTypes == "emf"] <- '<Default Extension="emf" ContentType="image/x-emf"/>'

    wb$Content_Types <- c(contentNodes, wb$Content_Types)
    names(media) <- mediaNames
    wb$media <- media
  }



  ## xl\chart
  if (length(charts) > 0) {
    chartNames <- basename(charts)
    nCharts <- sum(grepl("chart[0-9]+.xml", chartNames))
    nChartStyles <- sum(grepl("style[0-9]+.xml", chartNames))
    nChartCol <- sum(grepl("colors[0-9]+.xml", chartNames))

    if (nCharts > 0) {
      wb$Content_Types <- c(wb$Content_Types, sprintf('<Override PartName="/xl/charts/chart%s.xml" ContentType="application/vnd.openxmlformats-officedocument.drawingml.chart+xml"/>', seq_len(nCharts)))
    }

    if (nChartStyles > 0) {
      wb$Content_Types <- c(wb$Content_Types, sprintf('<Override PartName="/xl/charts/style%s.xml" ContentType="application/vnd.ms-office.chartstyle+xml"/>', seq_len(nChartStyles)))
    }

    if (nChartCol > 0) {
      wb$Content_Types <- c(wb$Content_Types, sprintf('<Override PartName="/xl/charts/colors%s.xml" ContentType="application/vnd.ms-office.chartcolorstyle+xml"/>', seq_len(nChartCol)))
    }

    if (length(chartsRels)) {
      charts <- c(charts, chartsRels)
      chartNames <- c(chartNames, file.path("_rels", basename(chartsRels)))
    }

    names(charts) <- chartNames
    wb$charts <- charts
  }






  ## xl\theme
  if (length(themeXML) > 0) {
    wb$theme <- removeHeadTag(paste(unlist(lapply(sort(themeXML)[[1]], readUTF8)), collapse = ""))
  }


  ## externalLinks
  if (length(extLinksXML) > 0) {
    wb$externalLinks <- lapply(sort(extLinksXML), function(x) removeHeadTag(cppReadFile(x)))

    wb$Content_Types <- c(
      wb$Content_Types,
      sprintf('<Override PartName="/xl/externalLinks/externalLink%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.externalLink+xml"/>', seq_along(extLinksXML))
    )

    wb$workbook.xml.rels <- c(wb$workbook.xml.rels, sprintf(
      '<Relationship Id="rId4" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/externalLink" Target="externalLinks/externalLink%s.xml"/>',
      seq_along(extLinksXML)
    ))
  }

  ## externalLinksRels
  if (length(extLinksRelsXML) > 0) {
    wb$externalLinksRels <- lapply(sort(extLinksRelsXML), function(x) removeHeadTag(cppReadFile(x)))
  }







  ##* ----------------------------------------------------------------------------------------------*##
  ### BEGIN READING IN WORKSHEET DATA
  ##* ----------------------------------------------------------------------------------------------*##

  ## xl\worksheets
  file_names <- regmatches(worksheet_rId_mapping, regexpr("sheet[0-9]+\\.xml", worksheet_rId_mapping, perl = TRUE))
  file_rIds <- unlist(getId(worksheet_rId_mapping))
  file_names <- file_names[match(sheetrId, file_rIds)]

  worksheetsXML <- file.path(dirname(worksheetsXML), file_names)
  wb <- loadworksheets(wb = wb, styleObjects = styleObjects, xmlFiles = worksheetsXML, is_chart_sheet = is_chart_sheet)

  ## Fix styleobject encoding
  if (length(wb$styleObjects) > 0) {
    style_names <- sapply(wb$styleObjects, "[[", "sheet")
    Encoding(style_names) <- "UTF-8"
    wb$styleObjects <- lapply(seq_along(style_names), function(i) {
      wb$styleObjects[[i]]$sheet <- style_names[[i]]
      wb$styleObjects[[i]]
    })
  }


  ## Fix headers/footers
  for (i in seq_along(worksheetsXML)) {
    if (!is_chart_sheet[i]) {
      if (length(wb$worksheets[[i]]$headerFooter) > 0) {
        
        amp_split <- function(x) {
          if (length(x) == 0) return (NULL)
          # First split the entry before &L, &C, &R => left/center/right header/footer entries
          parts <- unlist(stri_split_regex(x, "(?=&amp;[LCR])", omit_empty = TRUE))
          # Then extract from each entry the justification (LCR) and the content 
          matches <- stri_match(parts, regex = "^(?:&amp;([LRC])|)(.*)$")
          # The OOXML standard says that entries without an &[LRC] are to be understood as centered
          matches[(matches[,2]==""),2] = "C"
                   
          # => convert to named character vector
          z <- matches[,3]
          names(z) <- matches[,2]
          
          # return 3-string vector for left/center/right content:
          res <- z[c("L", "C", "R")]
          res[is.na(res)] <- ""
          unname(res)
        }
        
        head_foot <- c("oddHeader", "oddFooter",
                       "evenHeader", "evenFooter",
                       "firstHeader", "firstFooter")
        
        headerFooter <- vector("list", length = length(head_foot))
        names(headerFooter) <- head_foot
        
        headerFooterXMl <- paste0(wb$worksheets[[i]]$headerFooter,
                                  collapse = "")
        
        for (hf in head_foot) {
          node <- getChildlessNode(xml = headerFooterXMl, tag = hf)
          node <- gsub(paste0("<", hf, ">(.+)</", hf, ">"), "\\1", node)
          headerFooter[[hf]] <- amp_split(node)
        }
        
        wb$worksheets[[i]]$headerFooter <- headerFooter
      }
    }
  }


  ##* ----------------------------------------------------------------------------------------------*##
  ### READING IN WORKSHEET DATA COMPLETE
  ##* ----------------------------------------------------------------------------------------------*##


  ## Next sheetRels to see which drawings_rels belongs to which sheet
  if (length(sheetRelsXML) > 0) {

    ## sheetrId is order sheet appears in xlsx file
    ## create a 1-1 vector of rels to worksheet
    ## haveRels is boolean vector where i-the element is TRUE/FALSE if sheet has a rels sheet

    if (length(chartSheetsXML) == 0) {
      allRels <- file.path(dirname(sheetRelsXML[1]), paste0(file_names, ".rels"))
      haveRels <- allRels %in% sheetRelsXML
    } else {
      haveRels <- rep(FALSE, length(wb$worksheets))
      allRels <- rep("", length(wb$worksheets))

      for (i in 1:nSheets) {
        if (is_chart_sheet[i]) {
          ind <- which(chartSheetRIds == sheetrId[i])
          rels_file <- file.path(chartSheetsRelsDir, paste0(chartsheet_rId_mapping[ind], ".rels"))
        } else {
          ind <- sheetrId[i]
          rels_file <- file.path(xmlDir, "xl", "worksheets", "_rels", paste0(file_names[i], ".rels"))
        }
        if (file.exists(rels_file)) {
          allRels[i] <- rels_file
          haveRels[i] <- TRUE
        }
      }
    }

    ## sheet.xml have been reordered to be in the order of sheetrId
    ## not every sheet has a worksheet rels

    xml <- lapply(seq_along(allRels), function(i) {
      if (haveRels[i]) {
        xml <- readUTF8(allRels[[i]])
        xml <- removeHeadTag(xml)
        xml <- gsub("<Relationships .*?>", "", xml)
        xml <- gsub("</Relationships>", "", xml)
        xml <- getChildlessNode(xml = xml, tag = "Relationship")
      } else {
        xml <- "<Relationship >"
      }
      return(xml)
    })




    
    ## Slicers -------------------------------------------------------------------------------------

    

    if (length(slicerXML) > 0) {
      slicerXML <- slicerXML[order(nchar(slicerXML), slicerXML)]
      slicersFiles <- lapply(xml, function(x) as.integer(regmatches(x, regexpr("(?<=slicer)[0-9]+(?=\\.xml)", x, perl = TRUE))))
      inds <- sapply(slicersFiles, length) > 0


      ## worksheet_rels Id for slicer will be rId0
      k <- 1L
      wb$slicers <- rep("", nSheets)
      for (i in 1:nSheets) {

        ## read in slicer[j].XML sheets into sheet[i]
        if (inds[i]) {
          wb$slicers[[i]] <- slicerXML[k]
          k <- k + 1L

          wb$worksheets_rels[[i]] <- unlist(c(
            wb$worksheets_rels[[i]],
            sprintf('<Relationship Id="rId0" Type="http://schemas.microsoft.com/office/2007/relationships/slicer" Target="../slicers/slicer%s.xml"/>', i)
          ))
          wb$Content_Types <- c(
            wb$Content_Types,
            sprintf('<Override PartName="/xl/slicers/slicer%s.xml" ContentType="application/vnd.ms-excel.slicer+xml"/>', i)
          )

          slicer_xml_exists <- FALSE
          ## Append slicer to worksheet extLst

          if (length(wb$worksheets[[i]]$extLst) > 0) {
            if (grepl('x14:slicer r:id="rId[0-9]+"', wb$worksheets[[i]]$extLst)) {
              wb$worksheets[[i]]$extLst <- sub('x14:slicer r:id="rId[0-9]+"', 'x14:slicer r:id="rId0"', wb$worksheets[[i]]$extLst)
              slicer_xml_exists <- TRUE
            }
          }

          if (!slicer_xml_exists) {
            wb$worksheets[[i]]$extLst <- c(wb$worksheets[[i]]$extLst, genBaseSlicerXML())
          }
        }
      }
    }


    if (length(slicerCachesXML) > 0) {

      ## ---- slicerCaches
      inds <- seq_along(slicerCachesXML)
      wb$Content_Types <- c(wb$Content_Types, sprintf('<Override PartName="/xl/slicerCaches/slicerCache%s.xml" ContentType="application/vnd.ms-excel.slicerCache+xml"/>', inds))
      wb$slicerCaches <- sapply(slicerCachesXML[order(nchar(slicerCachesXML), slicerCachesXML)], function(x) removeHeadTag(cppReadFile(x)))
      wb$workbook.xml.rels <- c(wb$workbook.xml.rels, sprintf('<Relationship Id="rId%s" Type="http://schemas.microsoft.com/office/2007/relationships/slicerCache" Target="slicerCaches/slicerCache%s.xml"/>', 1E5 + inds, inds))
      
      wb$workbook$extLst <- c(wb$workbook$extLst, genSlicerCachesExtLst(1E5 + inds))
    }


    ## Tables --------------------------------------------------------------------------------------

    

    if (length(tablesXML) > 0) {
      tables <- lapply(xml, function(x) as.integer(regmatches(x, regexpr("(?<=table)[0-9]+(?=\\.xml)", x, perl = TRUE))))
      tableSheets <- unlist(lapply(seq_along(sheetrId), function(i) rep(i, length(tables[[i]]))))

      if (length(unlist(tables)) > 0) {
        ## get the tables that belong to each worksheet and create a worksheets_rels for each
        tCount <- 2L ## table r:Ids start at 3
        for (i in seq_along(tables)) {
          if (length(tables[[i]]) > 0) {
            k <- seq_along(tables[[i]]) + tCount
            wb$worksheets_rels[[i]] <- unlist(c(
              wb$worksheets_rels[[i]],
              sprintf('<Relationship Id="rId%s" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/table" Target="../tables/table%s.xml"/>', k, k)
            ))


            wb$worksheets[[i]]$tableParts <- sprintf("<tablePart r:id=\"rId%s\"/>", k)
            tCount <- tCount + length(k)
          }
        }

        ## sort the tables into the order they appear in the xml and tables variables
        names(tablesXML) <- basename(tablesXML)
        tablesXML <- tablesXML[sprintf("table%s.xml", unlist(tables))]

        ## tables are now in correct order so we can read them in as they are
        wb$tables <- sapply(tablesXML, function(x) removeHeadTag(paste(readUTF8(x), collapse = "")))

        ## pull out refs and attach names
        refs <- regmatches(wb$tables, regexpr('(?<=ref=")[0-9A-Z:]+', wb$tables, perl = TRUE))
        names(wb$tables) <- refs

        wb$Content_Types <- c(wb$Content_Types, sprintf('<Override PartName="/xl/tables/table%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.table+xml"/>', seq_along(wb$tables) + 2))

        ## relabel ids
        for (i in seq_along(wb$tables)) {
          newId <- sprintf(' id="%s" ', i + 2)
          wb$tables[[i]] <- sub(' id="[0-9]+" ', newId, wb$tables[[i]])
        }

        displayNames <- unlist(regmatches(wb$tables, regexpr('(?<=displayName=").*?[^"]+', wb$tables, perl = TRUE)))
        if (length(displayNames) != length(tablesXML)) {
          displayNames <- paste0("Table", seq_along(tablesXML))
        }

        attr(wb$tables, "sheet") <- tableSheets
        attr(wb$tables, "tableName") <- displayNames

        for (i in seq_along(tableSheets)) {
          table_sheet_i <- tableSheets[i]
          attr(wb$worksheets[[table_sheet_i]]$tableParts, "tableName") <- c(attr(wb$worksheets[[table_sheet_i]]$tableParts, "tableName"), displayNames[i])
        }
      }
    } ## if(length(tablesXML) > 0)

    ## might we have some external hyperlinks
    if (any(sapply(wb$worksheets[!is_chart_sheet], function(x) length(x$hyperlinks) > 0))) {

      ## Do we have external hyperlinks
      hlinks <- lapply(xml, function(x) x[grepl("hyperlink", x) & grepl("External", x)])
      hlinksInds <- which(sapply(hlinks, length) > 0)

      ## If it's an external hyperlink it will have a target in the sheet_rels
      if (length(hlinksInds) > 0) {
        for (i in hlinksInds) {
          ids <- unlist(lapply(hlinks[[i]], function(x) regmatches(x, gregexpr('(?<=Id=").*?"', x, perl = TRUE))[[1]]))
          ids <- gsub('"$', "", ids)

          targets <- unlist(lapply(hlinks[[i]], function(x) regmatches(x, gregexpr('(?<=Target=").*?"', x, perl = TRUE))[[1]]))
          targets <- gsub('"$', "", targets)

          ids2 <- lapply(wb$worksheets[[i]]$hyperlinks, function(x) regmatches(x, gregexpr('(?<=r:id=").*?"', x, perl = TRUE))[[1]])
          ids2[sapply(ids2, length) == 0] <- NA
          ids2 <- gsub('"$', "", unlist(ids2))

          targets <- targets[match(ids2, ids)]
          names(wb$worksheets[[i]]$hyperlinks) <- targets
        }
      }
    }



    ## Drawings ------------------------------------------------------------------------------------

    

    ## xml is in the order of the sheets, drawIngs is toes to sheet position of hasDrawing
    ## Not every sheet has a drawing.xml


    drawXMLrelationship <- lapply(xml, function(x) grep("drawings/drawing", x, value = TRUE))
    hasDrawing <- sapply(drawXMLrelationship, length) > 0 ## which sheets have a drawing

    if (length(drawingRelsXML) > 0) {
      dRels <- lapply(drawingRelsXML, readUTF8)
      dRels <- unlist(lapply(dRels, removeHeadTag))
      dRels <- gsub("<Relationships .*?>", "", dRels)
      dRels <- gsub("</Relationships>", "", dRels)
    }

    if (length(drawingsXML) > 0) {
      dXML <- lapply(drawingsXML, readUTF8)
      dXML <- unlist(lapply(dXML, removeHeadTag))
      dXML <- gsub("<xdr:wsDr .*?>", "", dXML)
      dXML <- gsub("</xdr:wsDr>", "", dXML)

      #       ptn1 <- "<(mc:AlternateContent|xdr:oneCellAnchor|xdr:twoCellAnchor|xdr:absoluteAnchor)"
      #       ptn2 <- "</(mc:AlternateContent|xdr:oneCellAnchor|xdr:twoCellAnchor|xdr:absoluteAnchor)>"

      ## split at one/two cell Anchor
      # dXML <- regmatches(dXML, gregexpr(paste0(ptn1, ".*?", ptn2), dXML))
    }


    ## loop over all worksheets and assign drawing to sheet
    if (any(hasDrawing)) {
      for (i in seq_along(xml)) {
        if (hasDrawing[i]) {
          target <- unlist(lapply(drawXMLrelationship[[i]], function(x) regmatches(x, gregexpr('(?<=Target=").*?"', x, perl = TRUE))[[1]]))
          target <- basename(gsub('"$', "", target))

          ## sheet_i has which(hasDrawing)[[i]]
          relsInd <- grepl(target, drawingRelsXML)
          if (any(relsInd)) {
            wb$drawings_rels[i] <- dRels[relsInd]
          }

          drawingInd <- grepl(target, drawingsXML)
          if (any(drawingInd)) {
            wb$drawings[i] <- dXML[drawingInd]
          }
        }
      }
    }




    ## VML Drawings --------------------------------------------------------------------------------


    if (length(vmlDrawingXML) > 0) {
      wb$Content_Types <- c(wb$Content_Types, '<Default Extension="vml" ContentType="application/vnd.openxmlformats-officedocument.vmlDrawing"/>')

      drawXMLrelationship <- lapply(xml, function(x) grep("drawings/vmlDrawing", x, value = TRUE))
      hasDrawing <- sapply(drawXMLrelationship, length) > 0 ## which sheets have a drawing

      ## loop over all worksheets and assign drawing to sheet
      if (any(hasDrawing)) {
        for (i in seq_along(xml)) {
          if (hasDrawing[i]) {
            target <- unlist(lapply(drawXMLrelationship[[i]], function(x) regmatches(x, gregexpr('(?<=Target=").*?"', x, perl = TRUE))[[1]]))
            target <- basename(gsub('"$', "", target))
            ind <- grepl(target, vmlDrawingXML)

            if (any(ind)) {
              txt <- paste(readUTF8(vmlDrawingXML[ind]), collapse = "\n")
              txt <- removeHeadTag(txt)

              i1 <- regexpr("<v:shapetype", txt, fixed = TRUE)
              i2 <- regexpr("</xml>", txt, fixed = TRUE)

              wb$vml[[i]] <- substring(text = txt, first = i1, last = (i2 - 1L))

              relsInd <- grepl(target, vmlDrawingRelsXML)
              if (any(relsInd)) {
                wb$vml_rels[i] <- vmlDrawingRelsXML[relsInd]
              }
            }
          }
        }
      }
    }







    ## vmlDrawing and comments
    if (length(commentsXML) > 0) {
      drawXMLrelationship <- lapply(xml, function(x) grep("drawings/vmlDrawing[0-9]+\\.vml", x, value = TRUE))
      hasDrawing <- sapply(drawXMLrelationship, length) > 0 ## which sheets have a drawing

      commentXMLrelationship <- lapply(xml, function(x) grep("comments[0-9]+\\.xml", x, value = TRUE))
      hasComment <- sapply(commentXMLrelationship, length) > 0 ## which sheets have a comment

      for (i in seq_along(xml)) {
        if (hasComment[i]) {
          target <- unlist(lapply(drawXMLrelationship[[i]], function(x) regmatches(x, gregexpr('(?<=Target=").*?"', x, perl = TRUE))[[1]]))
          target <- basename(gsub('"$', "", target))
          ind <- grepl(target, vmlDrawingXML)

          if (any(ind)) {
            txt <- paste(readUTF8(vmlDrawingXML[ind]), collapse = "\n")
            txt <- removeHeadTag(txt)

            cd <- unique(getNodes(xml = txt, tagIn = "<x:ClientData"))
            cd <- grep('ObjectType="Note"', cd, value = TRUE)
            cd <- paste0(cd, ">")

            ## now loada comment
            target <- unlist(lapply(commentXMLrelationship[[i]], function(x) regmatches(x, gregexpr('(?<=Target=").*?"', x, perl = TRUE))[[1]]))
            target <- basename(gsub('"$', "", target))

            txt <- paste(readUTF8(grep(target, commentsXML, value = TRUE)), collapse = "\n")
            txt <- removeHeadTag(txt)

            authors <- getNodes(xml = txt, tagIn = "<author>")
            authors <- gsub("<author>|</author>", "", authors)

            comments <- getNodes(xml = txt, tagIn = "<commentList>")
            comments <- gsub("<commentList>", "", comments)
            comments <- getNodes(xml = comments, tagIn = "<comment")

            refs <- regmatches(comments, regexpr('(?<=ref=").*?[^"]+', comments, perl = TRUE))

            authorsInds <- as.integer(regmatches(comments, regexpr('(?<=authorId=").*?[^"]+', comments, perl = TRUE))) + 1
            authors <- authors[authorsInds]

            style <- lapply(comments, getNodes, tagIn = "<rPr>")

            comments <- regmatches(comments, 
                                   gregexpr("(?<=<t( |>))[\\s\\S]+?(?=</t>)", comments, perl = TRUE))
            comments <- lapply(comments, function(x) gsub(".*?>", "", x, perl = TRUE))


            wb$comments[[i]] <- lapply(seq_along(comments), function(j) {
              list(
                "ref" = refs[j],
                "author" = authors[j],
                "comment" = comments[[j]],
                "style" = style[[j]],
                "clientData" = cd[[j]]
              )
            })
          }
        }
      }
    }

    ## Threaded comments
    if (length(threadCommentsXML) > 0) {
      threadCommentsXMLrelationship <- lapply(xml, function(x) grep("threadedComment[0-9]+\\.xml", x, value = TRUE))
      hasThreadComments<- sapply(threadCommentsXMLrelationship, length) > 0
      if(any(hasThreadComments)) {
        for (i in seq_along(xml)) {
          if (hasThreadComments[i]) {
            target <- unlist(lapply(threadCommentsXMLrelationship[[i]], function(x) regmatches(x, gregexpr('(?<=Target=").*?"', x, perl = TRUE))[[1]]))
            target <- basename(gsub('"$', "", target))

            wb$threadComments[[i]] <- grep(target, threadCommentsXML, value = TRUE)
            
          }
        }
      }
      wb$Content_Types <- c(
        wb$Content_Types, 
        sprintf('<Override PartName="/xl/threadedComments/%s" ContentType="application/vnd.ms-excel.threadedcomments+xml"/>',
                sapply(threadCommentsXML, basename))
        )
    }
    
    ## Persons (needed for Threaded Comment)
    if (length(personXML) > 0) {
      wb$persons <- personXML
      wb$Content_Types <- c(
        wb$Content_Types,
        '<Override PartName="/xl/persons/person.xml" ContentType="application/vnd.ms-excel.person+xml"/>'
      )
      wb$workbook.xml.rels <- c(
        wb$workbook.xml.rels, 
        '<Relationship Id="rId5" Type="http://schemas.microsoft.com/office/2017/10/relationships/person" Target="persons/person.xml"/>')
    }
    

    ## rels image
    drawXMLrelationship <- lapply(xml, function(x) grep("relationships/image", x, value = TRUE))
    hasDrawing <- sapply(drawXMLrelationship, length) > 0 ## which sheets have a drawing
    if (any(hasDrawing)) {
      for (i in seq_along(xml)) {
        if (hasDrawing[i]) {
          image_ids <- unlist(getId(drawXMLrelationship[[i]]))
          new_image_ids <- paste0("rId", seq_along(image_ids) + 70000)
          for (j in seq_along(image_ids)) {
            wb$worksheets[[i]]$oleObjects <- gsub(image_ids[j], new_image_ids[j], wb$worksheets[[i]]$oleObjects, fixed = TRUE)
            wb$worksheets_rels[[i]] <- c(wb$worksheets_rels[[i]], gsub(image_ids[j], new_image_ids[j], drawXMLrelationship[[i]][j], fixed = TRUE))
          }
        }
      }
    }

    ## rels image
    drawXMLrelationship <- lapply(xml, function(x) grep("relationships/package", x, value = TRUE))
    hasDrawing <- sapply(drawXMLrelationship, length) > 0 ## which sheets have a drawing
    if (any(hasDrawing)) {
      for (i in seq_along(xml)) {
        if (hasDrawing[i]) {
          image_ids <- unlist(getId(drawXMLrelationship[[i]]))
          new_image_ids <- paste0("rId", seq_along(image_ids) + 90000)
          for (j in seq_along(image_ids)) {
            wb$worksheets[[i]]$oleObjects <- gsub(image_ids[j], new_image_ids[j], wb$worksheets[[i]]$oleObjects, fixed = TRUE)
            wb$worksheets_rels[[i]] <- c(
              wb$worksheets_rels[[i]],
              sprintf("<Relationship Id=\"%s\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/package\" Target=\"../embeddings/Microsoft_Word_Document1.docx\"/>", new_image_ids[j])
            )
          }
        }
      }
    }



    ## Embedded docx
    if (length(embeddings) > 0) {
      wb$Content_Types <- c(wb$Content_Types, '<Default Extension="docx" ContentType="application/vnd.openxmlformats-officedocument.wordprocessingml.document"/>')
      wb$embeddings <- embeddings
    }



    ## pivot tables
    if (length(pivotTableXML) > 0) {
      # pivotTableJ <- lapply(xml, function(x) as.integer(regmatches(x, regexpr("(?<=pivotTable)[0-9]+(?=\\.xml)", x, perl = TRUE)))) variable not used
      # sheetWithPivot <- which(sapply(pivotTableJ, length) > 0)  variable not used

      pivotRels <- lapply(xml, function(x) {
        y <- grep("pivotTable", x, value = TRUE)
        y[order(nchar(y), y)]
      })
      hasPivot <- sapply(pivotRels, length) > 0

      ## Modify rIds
      for (i in seq_along(pivotRels)) {
        if (hasPivot[i]) {
          for (j in seq_along(pivotRels[[i]])) {
            pivotRels[[i]][j] <- gsub('"rId[0-9]+"', sprintf('"rId%s"', 20000L + j), pivotRels[[i]][j])
          }

          wb$worksheets_rels[[i]] <- c(wb$worksheets_rels[[i]], pivotRels[[i]])
        }
      }


      ## remove any workbook_res references to pivot tables that are not being used in worksheet_rels
      inds <- seq_along(wb$pivotTables.xml.rels)
      fileNo <- as.integer(unlist(regmatches(unlist(wb$worksheets_rels), gregexpr("(?<=pivotTable)[0-9]+(?=\\.xml)", unlist(wb$worksheets_rels), perl = TRUE))))
      inds <- inds[!inds %in% fileNo]

      if (length(inds) > 0) {
        toRemove <- paste(sprintf("(pivotCacheDefinition%s\\.xml)", inds), collapse = "|")
        fileNo <- grep(toRemove, wb$pivotTables.xml.rels)
        toRemove <- paste(sprintf("(pivotCacheDefinition%s\\.xml)", fileNo), collapse = "|")

        ## remove reference to file from workbook.xml.res
        wb$workbook.xml.rels <- wb$workbook.xml.rels[!grepl(toRemove, wb$workbook.xml.rels)]
      }
    }
  } ## end of worksheetRels

  ## convert hyperliks to hyperlink objects
  for (i in 1:nSheets) {
    wb$worksheets[[i]]$hyperlinks <- xml_to_hyperlink(wb$worksheets[[i]]$hyperlinks)
  }



  ## queryTables
  if (length(queryTablesXML) > 0) {
    ids <- as.numeric(regmatches(queryTablesXML, regexpr("[0-9]+(?=\\.xml)", queryTablesXML, perl = TRUE)))
    wb$queryTables <- unlist(lapply(queryTablesXML[order(ids)], function(x) removeHeadTag(cppReadFile(xmlFile = x))))
    wb$Content_Types <- c(
      wb$Content_Types,
      sprintf('<Override PartName="/xl/queryTables/queryTable%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.queryTable+xml"/>', seq_along(queryTablesXML))
    )
  }


  ## connections
  if (length(connectionsXML) > 0) {
    wb$connections <- removeHeadTag(cppReadFile(xmlFile = connectionsXML))
    wb$workbook.xml.rels <- c(wb$workbook.xml.rels, '<Relationship Id="rId3" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/connections" Target="connections.xml"/>')
    wb$Content_Types <- c(wb$Content_Types, '<Override PartName="/xl/connections.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.connections+xml"/>')
  }




  ## table rels
  if (length(tableRelsXML) > 0) {

    ## table_i_might have tableRels_i but I am re-ordering the tables to be in order of worksheets
    ## I make every table have a table_rels so i need to fill in the gaps if any table_rels are missing

    tmp <- paste0(basename(tablesXML), ".rels")
    hasRels <- tmp %in% basename(tableRelsXML)

    ## order tableRelsXML
    tableRelsXML <- tableRelsXML[match(tmp[hasRels], basename(tableRelsXML))]

    ##
    wb$tables.xml.rels <- character(length = length(tablesXML))

    ## which sheet does it belong to
    xml <- sapply(tableRelsXML, cppReadFile, USE.NAMES = FALSE)
    xml <- sapply(xml, removeHeadTag, USE.NAMES = FALSE)

    wb$tables.xml.rels[hasRels] <- xml
  } else if (length(tablesXML) > 0) {
    wb$tables.xml.rels <- rep("", length(tablesXML))
  }


  activesheet <- unlist(regmatches(workbook, gregexpr("(?<=<bookViews>).*(?=</bookViews>)", workbook, perl = TRUE)))
  activesheet <- unlist(regmatches(activesheet, gregexpr("<workbookView[^>]*>", activesheet, perl = TRUE)))
  
  wb$ActiveSheet <- as.integer(getAttrs(activesheet,"activeTab")$activeTab) + 1L
  
  if (length(wb$ActiveSheet) == 0) {
    wb$ActiveSheet <- 1L
  }

  # Preserve window size and position:
  xWindow <- getAttrs(activesheet,c("xWindow"))$xWindow
  yWindow <- getAttrs(activesheet,c("yWindow"))$yWindow
  windowWidth <- getAttrs(activesheet,c("windowWidth"))$windowWidth
  windowHeight <- getAttrs(activesheet,c("windowHeight"))$windowHeight

  bookViews <- wb$workbook$bookViews
  bookViews <- sub("xWindow=\"0", paste0("xWindow=\"", xWindow), bookViews)
  bookViews <- sub("yWindow=\"0", paste0("yWindow=\"", yWindow), bookViews)
  bookViews <- sub("windowWidth=\"13125", paste0("windowWidth=\"", windowWidth), bookViews)
  bookViews <- sub("windowHeight=\"6105", paste0("windowHeight=\"", windowHeight), bookViews)
  wb$workbook$bookViews <- bookViews
  # Done Preserve window size and position


  return(wb)
}

context("Load Unzipped Workbook Object")


test_that("Loading unzipped readTest.xlsx", {
  fl <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
  wb <- loadWorkbook(fl)

  ## make unzipped file & load
  tmp_dir <- file.path(tempdir(), paste(sample(LETTERS, 6), collapse = ""))
  if (dir.exists(tmp_dir)) unlink(tmp_dir, recursive = TRUE)
  dir.create(tmp_dir)

  unzip(zipfile = fl, exdir = tmp_dir)
  wb2 <- loadWorkbook(file = tmp_dir, isUnzipped = TRUE)

  expect_true(all.equal(wb, wb2))

  unlink(tmp_dir, recursive = TRUE)
})








""

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

calc_column_widths <- function(sheet_data, sharedStrings, autoColumns, widths, baseFontCharWidth, minW, maxW) {
    .Call(`_openxlsx_calc_column_widths`, sheet_data, sharedStrings, autoColumns, widths, baseFontCharWidth, minW, maxW)
}

convert_to_excel_ref <- function(cols, LETTERS) {
    .Call(`_openxlsx_convert_to_excel_ref`, cols, LETTERS)
}

convert_from_excel_ref <- function(x) {
    .Call(`_openxlsx_convert_from_excel_ref`, x)
}

convert_to_excel_ref_expand <- function(cols, LETTERS, rows) {
    .Call(`_openxlsx_convert_to_excel_ref_expand`, cols, LETTERS, rows)
}

isInternalHyperlink <- function(x) {
    .Call(`_openxlsx_isInternalHyperlink`, x)
}

write_file <- function(head = "", body = "", tail = "", fl = "") {
    .Call(`_openxlsx_write_file`, head, body, tail, fl)
}

cppReadFile <- function(xmlFile) {
    .Call(`_openxlsx_cppReadFile`, xmlFile)
}

read_file_newline <- function(xmlFile) {
    .Call(`_openxlsx_read_file_newline`, xmlFile)
}

get_letters <- function() {
    .Call(`_openxlsx_get_letters`)
}

markUTF8 <- function(x, clone) {
    .Call(`_openxlsx_markUTF8`, x, clone)
}

loadworksheets <- function(wb, styleObjects, xmlFiles, is_chart_sheet) {
    .Call(`_openxlsx_loadworksheets`, wb, styleObjects, xmlFiles, is_chart_sheet)
}

getNodes <- function(xml, tagIn) {
    .Call(`_openxlsx_getNodes`, xml, tagIn)
}

getOpenClosedNode <- function(xml, open_tag, close_tag) {
    .Call(`_openxlsx_getOpenClosedNode`, xml, open_tag, close_tag)
}

getAttr <- function(x, tag) {
    .Call(`_openxlsx_getAttr`, x, tag)
}

getChildlessNode_ss <- function(xml, tag) {
    .Call(`_openxlsx_getChildlessNode_ss`, xml, tag)
}

getChildlessNode <- function(xml, tag) {
    .Call(`_openxlsx_getChildlessNode`, xml, tag)
}

get_extLst_Major <- function(xml) {
    .Call(`_openxlsx_get_extLst_Major`, xml)
}

cell_ref_to_col <- function(x) {
    .Call(`_openxlsx_cell_ref_to_col`, x)
}

int_2_cell_ref <- function(cols) {
    .Call(`_openxlsx_int_2_cell_ref`, cols)
}

get_shared_strings <- function(xmlFile, isFile) {
    .Call(`_openxlsx_get_shared_strings`, xmlFile, isFile)
}

getCellInfo <- function(xmlFile, sharedStrings, skipEmptyRows, startRow, rows, getDates) {
    .Call(`_openxlsx_getCellInfo`, xmlFile, sharedStrings, skipEmptyRows, startRow, rows, getDates)
}

read_workbook <- function(cols_in, rows_in, v, string_inds, is_date, hasColNames, hasSepNames, skipEmptyRows, skipEmptyCols, nRows, clean_names) {
    .Call(`_openxlsx_read_workbook`, cols_in, rows_in, v, string_inds, is_date, hasColNames, hasSepNames, skipEmptyRows, skipEmptyCols, nRows, clean_names)
}

calc_number_rows <- function(x, skipEmptyRows) {
    .Call(`_openxlsx_calc_number_rows`, x, skipEmptyRows)
}

map_cell_types_to_integer <- function(t) {
    .Call(`_openxlsx_map_cell_types_to_integer`, t)
}

map_cell_types_to_char <- function(t) {
    .Call(`_openxlsx_map_cell_types_to_char`, t)
}

build_cell_types_integer <- function(classes, n_rows) {
    .Call(`_openxlsx_build_cell_types_integer`, classes, n_rows)
}

buildCellTypes <- function(classes, nRows) {
    .Call(`_openxlsx_buildCellTypes`, classes, nRows)
}

build_cell_merges <- function(comps) {
    .Call(`_openxlsx_build_cell_merges`, comps)
}

buildCellList <- function(r, t, v) {
    .Call(`_openxlsx_buildCellList`, r, t, v)
}

#' @import Rcpp
write_worksheet_xml <- function(prior, post, sheet_data, R_fileName) {
    .Call(`_openxlsx_write_worksheet_xml`, prior, post, sheet_data, R_fileName)
}

buildMatrixNumeric <- function(v, rowInd, colInd, colNames, nRows, nCols) {
    .Call(`_openxlsx_buildMatrixNumeric`, v, rowInd, colInd, colNames, nRows, nCols)
}

buildMatrixMixed <- function(v, rowInd, colInd, colNames, nRows, nCols, charCols, dateCols) {
    .Call(`_openxlsx_buildMatrixMixed`, v, rowInd, colInd, colNames, nRows, nCols, charCols, dateCols)
}

matrixRowInds <- function(indices) {
    .Call(`_openxlsx_matrixRowInds`, indices)
}

build_table_xml <- function(table, tableStyleXML, ref, colNames, showColNames, withFilter) {
    .Call(`_openxlsx_build_table_xml`, table, tableStyleXML, ref, colNames, showColNames, withFilter)
}

write_worksheet_xml_2 <- function(prior, post, sheet_data, row_heights_ = NULL, outline_levels_ = NULL, R_fileName = "output") {
    .Call(`_openxlsx_write_worksheet_xml_2`, prior, post, sheet_data, row_heights_, outline_levels_, R_fileName)
}

genBaseContent_Type <- function() {
  c(
    '<Default Extension="bin" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.printerSettings"/>',
    '<Default Extension="rels" ContentType="application/vnd.openxmlformats-package.relationships+xml"/>',
    '<Default Extension="xml" ContentType="application/xml"/>',
    '<Override PartName="/docProps/app.xml" ContentType="application/vnd.openxmlformats-officedocument.extended-properties+xml"/>',
    '<Override PartName="/docProps/core.xml" ContentType="application/vnd.openxmlformats-package.core-properties+xml"/>',
    '<Override PartName="/xl/sharedStrings.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.sharedStrings+xml"/>',
    '<Override PartName="/xl/styles.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.styles+xml"/>',
    '<Override PartName="/xl/workbook.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet.main+xml"/>',
    '<Override PartName="/xl/theme/theme1.xml" ContentType="application/vnd.openxmlformats-officedocument.theme+xml"/>'
  )
}


genBaseShapeVML <- function(clientData, id) {
  if (grepl("visible", clientData, ignore.case = TRUE)) {
    visible <- "visible"
  } else {
    visible <- "hidden"
  }

  paste0(
    sprintf('<v:shape id="_x0000_s%s" type="#_x0000_t202" style=\'position:absolute;', id),
    sprintf('margin-left:107.25pt;margin-top:172.5pt;width:147pt;height:96pt;z-index:1;
          visibility:%s;mso-wrap-style:tight\' fillcolor="#ffffe1" o:insetmode="auto">', visible),
    '<v:fill color2="#ffffe1"/>
            <v:shadow color="black" obscured="t"/>
            <v:path o:connecttype="none"/>
            <v:textbox style=\'mso-direction-alt:auto\'>
            <div style=\'text-align:left\'/>
            </v:textbox>', clientData, "</v:shape>"
  )
}





genClientData <- function(col, row, visible, height, width) {
  txt <- sprintf(
    '<x:ClientData ObjectType="Note"><x:MoveWithCells/><x:SizeWithCells/><x:Anchor>%s, 15, %s, 10, %s, 147, %s, 18</x:Anchor><x:AutoFill>False</x:AutoFill><x:Row>%s</x:Row><x:Column>%s</x:Column>',
    col, row - 2L, col + width - 1L, row + height - 1L, row - 1L, col - 1L
  )

  if (visible) {
    txt <- paste0(txt, "<x:Visible/>")
  }

  txt <- paste0(txt, "</x:ClientData>")

  return(txt)
}


# genBaseRels <- function(){
#
#   '<Relationship Id="rId1" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/officeDocument" Target="xl/workbook.xml"/>
#    <Relationship Id="rId2" Type="http://schemas.openxmlformats.org/package/2006/relationships/metadata/core-properties" Target="docProps/core.xml"/>
#    <Relationship Id="rId3" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/extended-properties" Target="docProps/app.xml"/>'
#
# }
#
#
# genBaseApp <- function(){
#   list('<Application>Microsoft Excel</Application>')
# }


genBaseCore <- function(creator = "", title = NULL, subject = NULL, category = NULL) {
  core <- '<coreProperties xmlns="http://schemas.openxmlformats.org/package/2006/metadata/core-properties" xmlns:cp="http://schemas.openxmlformats.org/package/2006/metadata/core-properties" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'

  core <- stri_c(core, sprintf("<dc:creator>%s</dc:creator>", replaceIllegalCharacters(creator)))
  core <- stri_c(core, sprintf("<cp:lastModifiedBy>%s</cp:lastModifiedBy>", replaceIllegalCharacters(creator)))
  core <- stri_c(core, sprintf('<dcterms:created xsi:type="dcterms:W3CDTF">%s</dcterms:created>', format(as_POSIXct_utc(Sys.time()), "%Y-%m-%dT%H:%M:%SZ")))

  if (!is.null(title)) {
    core <- stri_c(core, sprintf("<dc:title>%s</dc:title>", replaceIllegalCharacters(title)))
  }

  if (!is.null(subject)) {
    core <- stri_c(core, sprintf("<dc:subject>%s</dc:subject>", replaceIllegalCharacters(subject)))
  }

  if (!is.null(category)) {
    core <- stri_c(core, sprintf("<cp:category>%s</cp:category>", replaceIllegalCharacters(category)))
  }

  core <- stri_c(core, "</coreProperties>")

  return(core)
}

#
# addAuthor <- function(wb,Author = NULL){
#
#   if (!is.null(Author)) {
#     current_creator <-
#       stri_match(wb$core, regex = "<dc:creator>(.*?)</dc:creator>")[1, 2]
#     wb$core <-
#       stri_replace_all_fixed(
#         wb$core,
#         pattern = current_creator,
#         replacement = stri_c(current_creator, Author, sep = ";")
#       )
#   }
#
#
# }
#
#
# setAuthor <- function(wb,Author = NULL){
#
#   if (!is.null(Author)) {
#     current_creator <-
#       stri_match(wb$core, regex = "<dc:creator>(.*?)</dc:creator>")[1, 2]
#     wb$core <-
#       stri_replace_all_fixed(
#         wb$core,
#         pattern = current_creator,
#         replacement =  Author
#       )
#   }
#
#
# }
#
# setLastModifiedBy <- function(wb,ModifiedBy=NULL){
#
#   if (!is.null(addmodifier)) {
#     current_lastmodifier <-
#       stri_match(wb$core, regex = "<cp:lastModifiedBy>(.*?)</cp:lastModifiedBy>")[1, 2]
#     wb$core <-
#       stri_replace_all_fixed(
#         wb$core,
#         pattern = current_lastmodifier,
#         replacement = ModifiedBy
#       )
#   }
#
#
# }
#
#

#
#
# setBaseCore <- function(core,setcreator="",setmodifier="",
#                         title = NULL, subject = NULL, category = NULL){
#
#
#   core <- c(core, sprintf('<dc:creator>%s</dc:creator>', setcreator))
#   core <- c(core, sprintf('<dcterms:created xsi:type="dcterms:W3CDTF">%s</dcterms:created>', format(Sys.time(), "%Y-%m-%dT%H:%M:%SZ")))
#
#   if(!is.null(title))
#     core <- c(core, sprintf('<dc:title>%s</dc:title>', replaceIllegalCharacters(title)))
#
#   if(!is.null(subject))
#     core <- c(core, sprintf('<dc:subject>%s</dc:subject>', replaceIllegalCharacters(subject)))
#
#   if(!is.null(category))
#     core <- c(core, sprintf('<cp:category>%s</cp:category>', replaceIllegalCharacters(category)))
#
#   core <- c(core, '</coreProperties>')
#
#   return(core)
#
# }




genBaseWorkbook.xml.rels <- function() {
  c(
    '<Relationship Id="rId1" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/sharedStrings" Target="sharedStrings.xml"/>',
    '<Relationship Id="rId2" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/styles" Target="styles.xml"/>',
    '<Relationship Id="rId4" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/theme" Target="theme/theme1.xml"/>'
  )
}


genBaseWorkbook <- function() {
  list(
    workbookPr = '<workbookPr date1904="false"/>',
    workbookProtection = NULL,
    bookViews = '<bookViews><workbookView xWindow="0" yWindow="0" windowWidth="13125" windowHeight="6105"/></bookViews>',
    sheets = NULL,
    externalReferences = NULL,
    definedNames = NULL,
    calcPr = NULL,
    pivotCaches = NULL,
    extLst = NULL
  )
}




genBaseSheetRels <- function(sheetInd) {
  c(
    sprintf('<Relationship Id="rId1" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/drawing" Target="../drawings/drawing%s.xml"/>', sheetInd),
    sprintf('<Relationship Id="rId2" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/printerSettings" Target="../printerSettings/printerSettings%s.bin"/>', sheetInd),
    sprintf('<Relationship Id="rIdvml" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/vmlDrawing" Target="../drawings/vmlDrawing%s.vml"/>', sheetInd)
  )
}

genBaseStyleSheet <- function(dxfs = NULL, tableStyles = NULL, extLst = NULL) {
  list(
    numFmts = NULL,

    fonts = c('<font><sz val="11"/><color rgb="FF000000"/><name val="Calibri"/><family val="2"/><scheme val="minor"/></font>'),

    fills = c(
      '<fill><patternFill patternType="none"/></fill>',
      '<fill><patternFill patternType="gray125"/></fill>'
    ),

    borders = c("<border><left/><right/><top/><bottom/><diagonal/></border>"),

    cellStyleXfs = c('<xf numFmtId="0" fontId="0" fillId="0" borderId="0"/>'),

    cellXfs = c('<xf numFmtId="0" fontId="0" fillId="0" borderId="0" xfId="0"/>'),

    cellStyles = c('<cellStyle name="Normal" xfId="0" builtinId="0"/>'),

    dxfs = dxfs,

    tableStyles = tableStyles,

    indexedColors = NULL,

    extLst = extLst
  )
}


genBasePic <- function(imageNo, imageRelNo, hyperlinkXML) {
  sprintf('<xdr:pic xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing">
      <xdr:nvPicPr xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing">
        <xdr:cNvPr xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing" id="%s" name="Picture %s"%s>
        <xdr:cNvPicPr xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing">
          <a:picLocks xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main" noChangeAspect="1"/>
        </xdr:cNvPicPr>
      </xdr:nvPicPr>
      <xdr:blipFill xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing">
        <a:blip xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" r:embed="rId%s">
        </a:blip>
        <a:stretch xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main">
          <a:fillRect/>
        </a:stretch>
      </xdr:blipFill>
      <xdr:spPr xmlns:xdr="http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing">
        <a:prstGeom xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main" prst="rect">
          <a:avLst xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main"/>
        </a:prstGeom>
      </xdr:spPr>
    </xdr:pic>', imageNo, imageNo, hyperlinkXML, imageRelNo)
}








genBaseTheme <- function() {
  '<a:theme xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main" name="Office Theme">
  <a:themeElements><a:clrScheme name="Office">
  <a:dk1><a:sysClr val="windowText" lastClr="000000"/></a:dk1>
  <a:lt1><a:sysClr val="window" lastClr="FFFFFF"/></a:lt1>
  <a:dk2><a:srgbClr val="1F497D"/></a:dk2>
  <a:lt2><a:srgbClr val="EEECE1"/></a:lt2>
  <a:accent1><a:srgbClr val="4F81BD"/></a:accent1><a:accent2><a:srgbClr val="C0504D"/></a:accent2>
  <a:accent3><a:srgbClr val="9BBB59"/></a:accent3><a:accent4><a:srgbClr val="8064A2"/></a:accent4>
  <a:accent5><a:srgbClr val="4BACC6"/></a:accent5><a:accent6><a:srgbClr val="F79646"/></a:accent6>
  <a:hlink><a:srgbClr val="0000FF"/></a:hlink><a:folHlink><a:srgbClr val="800080"/></a:folHlink>
  </a:clrScheme><a:fontScheme name="Office">
  <a:majorFont>
  <a:latin typeface="Cambria"/>
  <a:ea typeface=""/>
  <a:cs typeface=""/>
  <a:font script="Arab" typeface="Times New Roman"/>
  <a:font script="Hebr" typeface="Times New Roman"/>
  <a:font script="Thai" typeface="Tahoma"/>
  <a:font script="Ethi" typeface="Nyala"/>
  <a:font script="Beng" typeface="Vrinda"/>
  <a:font script="Gujr" typeface="Shruti"/>
  <a:font script="Khmr" typeface="MoolBoran"/>
  <a:font script="Knda" typeface="Tunga"/>
  <a:font script="Guru" typeface="Raavi"/>
  <a:font script="Cans" typeface="Euphemia"/>
  <a:font script="Cher" typeface="Plantagenet Cherokee"/>
  <a:font script="Yiii" typeface="Microsoft Yi Baiti"/>
  <a:font script="Tibt" typeface="Microsoft Himalaya"/>
  <a:font script="Thaa" typeface="MV Boli"/>
  <a:font script="Deva" typeface="Mangal"/>
  <a:font script="Telu" typeface="Gautami"/>
  <a:font script="Taml" typeface="Latha"/>
  <a:font script="Syrc" typeface="Estrangelo Edessa"/>
  <a:font script="Orya" typeface="Kalinga"/>
  <a:font script="Mlym" typeface="Kartika"/>
  <a:font script="Laoo" typeface="DokChampa"/>
  <a:font script="Sinh" typeface="Iskoola Pota"/>
  <a:font script="Mong" typeface="Mongolian Baiti"/>
  <a:font script="Viet" typeface="Times New Roman"/>
  <a:font script="Uigh" typeface="Microsoft Uighur"/>
  <a:font script="Geor" typeface="Sylfaen"/>
  </a:majorFont>
  <a:minorFont>
  <a:latin typeface="Calibri"/>
  <a:ea typeface=""/>
  <a:cs typeface=""/>
  <a:font script="Arab" typeface="Arial"/>
  <a:font script="Hebr" typeface="Arial"/>
  <a:font script="Thai" typeface="Tahoma"/>
  <a:font script="Ethi" typeface="Nyala"/>
  <a:font script="Beng" typeface="Vrinda"/>
  <a:font script="Gujr" typeface="Shruti"/>
  <a:font script="Khmr" typeface="DaunPenh"/>
  <a:font script="Knda" typeface="Tunga"/>
  <a:font script="Guru" typeface="Raavi"/>
  <a:font script="Cans" typeface="Euphemia"/>
  <a:font script="Cher" typeface="Plantagenet Cherokee"/>
  <a:font script="Yiii" typeface="Microsoft Yi Baiti"/>
  <a:font script="Tibt" typeface="Microsoft Himalaya"/>
  <a:font script="Thaa" typeface="MV Boli"/>
  <a:font script="Deva" typeface="Mangal"/>
  <a:font script="Telu" typeface="Gautami"/>
  <a:font script="Taml" typeface="Latha"/>
  <a:font script="Syrc" typeface="Estrangelo Edessa"/>
  <a:font script="Orya" typeface="Kalinga"/>
  <a:font script="Mlym" typeface="Kartika"/>
  <a:font script="Laoo" typeface="DokChampa"/>
  <a:font script="Sinh" typeface="Iskoola Pota"/>
  <a:font script="Mong" typeface="Mongolian Baiti"/>
  <a:font script="Viet" typeface="Arial"/>
  <a:font script="Uigh" typeface="Microsoft Uighur"/>
  <a:font script="Geor" typeface="Sylfaen"/>
  </a:minorFont>
  </a:fontScheme>
  <a:fmtScheme name="Office">
  <a:fillStyleLst>
  <a:solidFill><a:schemeClr val="phClr"/></a:solidFill>
  <a:gradFill rotWithShape="1">
  <a:gsLst>
  <a:gs pos="0"><a:schemeClr val="phClr"><a:tint val="50000"/><a:satMod val="300000"/></a:schemeClr></a:gs>
  <a:gs pos="35000"><a:schemeClr val="phClr"><a:tint val="37000"/><a:satMod val="300000"/></a:schemeClr></a:gs>
  <a:gs pos="100000"><a:schemeClr val="phClr"><a:tint val="15000"/><a:satMod val="350000"/></a:schemeClr></a:gs>
  </a:gsLst>
  <a:lin ang="16200000" scaled="1"/>
  </a:gradFill>
  <a:gradFill rotWithShape="1">
  <a:gsLst>
  <a:gs pos="0"><a:schemeClr val="phClr"><a:shade val="51000"/><a:satMod val="130000"/></a:schemeClr></a:gs>
  <a:gs pos="80000"><a:schemeClr val="phClr"><a:shade val="93000"/><a:satMod val="130000"/></a:schemeClr></a:gs>
  <a:gs pos="100000"><a:schemeClr val="phClr"><a:shade val="94000"/><a:satMod val="135000"/></a:schemeClr></a:gs>
  </a:gsLst>
  <a:lin ang="16200000" scaled="0"/>
  </a:gradFill>
  </a:fillStyleLst>
  <a:lnStyleLst>
  <a:ln w="9525" cap="flat" cmpd="sng" algn="ctr">
  <a:solidFill><a:schemeClr val="phClr"><a:shade val="95000"/><a:satMod val="105000"/></a:schemeClr>
  </a:solidFill>
  <a:prstDash val="solid"/></a:ln><a:ln w="25400" cap="flat" cmpd="sng" algn="ctr">
  <a:solidFill><a:schemeClr val="phClr"/></a:solidFill>
  <a:prstDash val="solid"/></a:ln>
  <a:ln w="38100" cap="flat" cmpd="sng" algn="ctr"><a:solidFill><a:schemeClr val="phClr"/></a:solidFill><a:prstDash val="solid"/></a:ln>
  </a:lnStyleLst>
  <a:effectStyleLst>
  <a:effectStyle>
  <a:effectLst>
  <a:outerShdw blurRad="40000" dist="20000" dir="5400000" rotWithShape="0">
  <a:srgbClr val="000000">
  <a:alpha val="38000"/>
  </a:srgbClr>
  </a:outerShdw>
  </a:effectLst>
  </a:effectStyle>
  <a:effectStyle>
  <a:effectLst>
  <a:outerShdw blurRad="40000" dist="23000" dir="5400000" rotWithShape="0">
  <a:srgbClr val="000000">
  <a:alpha val="35000"/>
  </a:srgbClr>
  </a:outerShdw>
  </a:effectLst>
  </a:effectStyle>
  <a:effectStyle><a:effectLst><a:outerShdw blurRad="40000" dist="23000" dir="5400000" rotWithShape="0"><a:srgbClr val="000000"><a:alpha val="35000"/>
  </a:srgbClr>
  </a:outerShdw>
  </a:effectLst>
  <a:scene3d>
  <a:camera prst="orthographicFront">
  <a:rot lat="0" lon="0" rev="0"/>
  </a:camera>
  <a:lightRig rig="threePt" dir="t"><a:rot lat="0" lon="0" rev="1200000"/></a:lightRig>
  </a:scene3d>
  <a:sp3d><a:bevelT w="63500" h="25400"/></a:sp3d>
  </a:effectStyle>
  </a:effectStyleLst>
  <a:bgFillStyleLst><a:solidFill><a:schemeClr val="phClr"/></a:solidFill><a:gradFill rotWithShape="1">
  <a:gsLst>
  <a:gs pos="0"><a:schemeClr val="phClr"><a:tint val="40000"/><a:satMod val="350000"/></a:schemeClr></a:gs>
  <a:gs pos="40000"><a:schemeClr val="phClr"><a:tint val="45000"/> <a:shade val="99000"/><a:satMod val="350000"/></a:schemeClr></a:gs>
  <a:gs pos="100000"><a:schemeClr val="phClr"><a:shade val="20000"/><a:satMod val="255000"/></a:schemeClr></a:gs>
  </a:gsLst>
  <a:path path="circle"><a:fillToRect l="50000" t="-80000" r="50000" b="180000"/></a:path></a:gradFill><a:gradFill rotWithShape="1"><a:gsLst>
  <a:gs pos="0"><a:schemeClr val="phClr"><a:tint val="80000"/><a:satMod val="300000"/></a:schemeClr></a:gs>
  <a:gs pos="100000"><a:schemeClr val="phClr"><a:shade val="30000"/><a:satMod val="200000"/></a:schemeClr></a:gs>
  </a:gsLst>
  <a:path path="circle"><a:fillToRect l="50000" t="50000" r="50000" b="50000"/></a:path>
  </a:gradFill></a:bgFillStyleLst></a:fmtScheme></a:themeElements><a:objectDefaults/><a:extraClrSchemeLst/></a:theme>'
}




genPrinterSettings <- function() {
  "5c 00 5c 00 41 00 55 00 43 00 41 00 4c 00 50 00 52 00 4f 00 44 00 46 00 50 00 5c 00 4c 00 31 00 34 00 78 00 65 00 72 00 6f 00 78 00 31 00 20 00 2d 00 20 00 58 00 65 00 72 00 6f 00 00 00 00 00 01 04 00 52 dc 00 5c 05 13 ff 81 07 02 00 09 00 9a 0b 34 08 64 00 01 00 0f 00 2c 01 02 00 02 00 2c 01 03 00 01 00 41 00 34 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 01 00 00 00 01 00 00 00 01 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 52 c0 21 46 00 58 00 20 00 41 00 70 00 65 00 6f 00 73 00 50 00 6f 00 72 00 74 00 2d 00 49 00 49 00 49 00 20 00 43 00 34 00 34 00 30 00 30 00 20 00 50 00 43 00 4c 00 20 00 36 00 00 00 00 00 00 00 00 00 4e 08 a0 13 40 09 08 00 0b 01 64 00 01 00 07 00 01 00 00 00 00 00 00 00 00 00 07 00 01 00 08 01 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 08 08 08 00 08 08 08 00 08 08 08 00 08 08 08 00 00 01 03 00 02 02 00 01 02 02 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 01 00 00 00 00 00 00 00 02 02 48 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 bc 02 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 05 00 00 00 00 00 00 08 00 01 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 01 00 00 00 00 00 00 01 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 0b 96 00 00 00 c8 00 01 01 01 01 01 01 01 01 01 01 01 01 09 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 bc 02 00 00 00 00 00 00 00 00 02 00 41 00 72 00 69 00 61 00 6c 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 01 00 01 00 01 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 12 70 5f 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 01 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00"
}


gen_databar_extlst <- function(guid, sqref, posColour, negColour, values, border, gradient) {
  xml <- sprintf('<x14:cfRule type="dataBar" id="{%s}"><x14:dataBar minLength="0" maxLength="100" border="%s" gradient = "%s" negativeBarBorderColorSameAsPositive="0">', guid, border, gradient)

  if (is.null(values)) {
    xml <- sprintf('<ext uri="{78C0D931-6437-407d-A8EE-F0AAD7539E65}" xmlns:x14="http://schemas.microsoft.com/office/spreadsheetml/2009/9/main"><x14:conditionalFormattings><x14:conditionalFormatting xmlns:xm="http://schemas.microsoft.com/office/excel/2006/main">
                      %s
                      <x14:cfvo type="autoMin"/><x14:cfvo type="autoMax"/><x14:borderColor rgb="%s"/><x14:negativeFillColor rgb="%s"/><x14:negativeBorderColor rgb="%s"/><x14:axisColor rgb="FF000000"/>
                      </x14:dataBar></x14:cfRule><xm:sqref>%s</xm:sqref></x14:conditionalFormatting></x14:conditionalFormattings></ext>', xml, posColour, negColour, negColour, sqref)
  } else {
    xml <- sprintf('<ext uri="{78C0D931-6437-407d-A8EE-F0AAD7539E65}" xmlns:x14="http://schemas.microsoft.com/office/spreadsheetml/2009/9/main"><x14:conditionalFormattings><x14:conditionalFormatting xmlns:xm="http://schemas.microsoft.com/office/excel/2006/main">
                      %s
                      <x14:cfvo type="num"><xm:f>%s</xm:f></x14:cfvo><x14:cfvo type="num"><xm:f>%s</xm:f></x14:cfvo>
                      <x14:borderColor rgb="%s"/><x14:negativeFillColor rgb="%s"/><x14:negativeBorderColor rgb="%s"/><x14:axisColor rgb="FF000000"/>
                      </x14:dataBar></x14:cfRule><xm:sqref>%s</xm:sqref></x14:conditionalFormatting></x14:conditionalFormattings></ext>', xml, values[[1]], values[[2]], posColour, negColour, negColour, sqref)
  }

  return(xml)
}



contentTypePivotXML <- function(i) {
  c(
    sprintf('<Override PartName="/xl/pivotCache/pivotCacheDefinition%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.pivotCacheDefinition+xml"/>', i),
    sprintf('<Override PartName="/xl/pivotCache/pivotCacheRecords%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.pivotCacheRecords+xml"/>', i),
    sprintf('<Override PartName="/xl/pivotTables/pivotTable%s.xml" ContentType="application/vnd.openxmlformats-officedocument.spreadsheetml.pivotTable+xml"/>', i)
  )
}

contentTypeSlicerCacheXML <- function(i) {
  c(
    sprintf('<Override PartName="/xl/slicerCaches/slicerCache%s.xml" ContentType="application/vnd.ms-excel.slicerCache+xml"/>', i),
    sprintf('<Override PartName="/xl/slicers/slicer%s.xml" ContentType="application/vnd.ms-excel.slicer+xml"/>', i)
  )
}


genBaseSlicerXML <- function() {
  '<ext uri="{A8765BA9-456A-4dab-B4F3-ACF838C121DE}" xmlns:x14="http://schemas.microsoft.com/office/spreadsheetml/2009/9/main">
    <x14:slicerList>
    <x14:slicer r:id="rId0"/>
      </x14:slicerList>
      </ext>'
}


genSlicerCachesExtLst <- function(i) {
  paste0(
    '<extLst>
    <ext uri=\"{BBE1A952-AA13-448e-AADC-164F8A28A991}\" xmlns:x14=\"http://schemas.microsoft.com/office/spreadsheetml/2009/9/main\">
    <x14:slicerCaches>',

    paste(sprintf('<x14:slicerCache r:id="rId%s"/>', i), collapse = ""),

    "</x14:slicerCaches></ext></extLst>"
  )
}

#' Font Size Lookup tables
#'
#' Lookup tables for font size
#'
#' @format A data.frame with column names corresponding to font names
#' @examples
#' data(openxlsxFontSizeLookupTable)
#' data(openxlsxFontSizeLookupTableBold)
"openxlsxFontSizeLookupTable"

#' @rdname openxlsxFontSizeLookupTable
#' @format NULL
"openxlsxFontSizeLookupTableBold"

#' xlsx reading, writing and editing.
#'
#' openxlsx simplifies the the process of writing and styling Excel xlsx files from R
#' and removes the dependency on Java.
#'
#' @name openxlsx
#' @docType package
#' @useDynLib openxlsx, .registration=TRUE
#' @importFrom zip zipr
#' @importFrom utils download.file head menu unzip
#'
#' @seealso
#' \itemize{
#'    \item{`vignette("Introduction", package = "openxlsx")`}
#'    \item{`vignette("formatting", package = "openxlsx")`}
#'    \item{[writeData()]}
#'    \item{[writeDataTable()]}
#'    \item{[write.xlsx()]}
#'    \item{[read.xlsx()]}
#'    \item{[op.openxlsx()]}
#'   }
#' for examples
#'
#' @details
#' The openxlsx package uses global options, most to simplify formatting.  These
#'   are stored in the `op.openxlsx` object.
#'
#' \describe{
#'  \item{openxlsx.bandedCols}{FALSE}
#'  \item{openxlsx.bandedRows}{TRUE}
#'  \item{openxlsx.borderColour}{"black"}
#'  \item{openxlsx.borders}{"none"}
#'  \item{openxlsx.borderStyle}{"thin"}
#'  \item{openxlsx.compressionLevel}{"9"}
#'  \item{openxlsx.creator}{""}
#'  \item{openxlsx.dateFormat}{"mm/dd/yyyy"}
#'  \item{openxlsx.datetimeFormat}{"yyyy-mm-dd hh:mm:ss"}
#'  \item{openxlsx.headerStyle}{NULL}
#'  \item{openxlsx.keepNA}{FALSE}
#'  \item{openxlsx.na.string}{NULL}
#'  \item{openxlsx.numFmt}{NULL}
#'  \item{openxlsx.orientation}{"portrait"}
#'  \item{openxlsx.paperSize}{9}
#'  \item{openxlsx.tabColour}{"TableStyleLight9"}
#'  \item{openxlsx.tableStyle}{"TableStyleLight9"}
#'  \item{openxlsx.withFilter}{NA Whether to write data with or without a 
#'  filter. If NA will make filters with `writeDataTable` and will not for
#'  `writeData`}
#' }
#' 
#' See the Formatting vignette for examples.
#'
#' Additional options
#'
#'
 "_PACKAGE"

#' openxlsx Options
#' 
#' See and get the openxlsx options
#' 
#' @details
#' 
#' `openxlsx_getOp()` retrieves the `"openxlsx"` options found in
#'   `op.openxlsx`.  If none are set (currently `NULL`) retrieves the 
#'   default option from `op.openxlsx`.  This will also check that the
#'   intended option is a standard option (listed in `op.openxlsx`) and
#'    will provide a warning otherwise.
#'   
#' `openxlsx_setOp()` is a safer way to set an option as it will first
#'   check that the option is a standard option (as above) before setting.
#' 
#' @examples
#' openxlsx_getOp("borders")
#' op.openxlsx[["openxlsx.borders"]]
#' 
#' @export
#' @name openxlsx_options
op.openxlsx <- list(
  openxlsx.bandedCols       = FALSE,
  openxlsx.bandedRows       = TRUE,
  openxlsx.borderColour     = "black",
  openxlsx.borders          = NULL,
  openxlsx.borderStyle      = "thin",
  # Where is compressionLevel called?
  openxlsx.compressionLevel = 9,
  openxlsx.creator          = "",
  openxlsx.dateFormat       = "date",
  openxlsx.datetimeFormat   = "longdate",
  openxlsx.hdpi             = 300,
  openxlsx.header           = NULL,
  openxlsx.headerStyle      = NULL,
  openxlsx.firstColumn      = NULL,
  openxlsx.firstFooter      = NULL,
  openxlsx.firstHeader      = NULL,
  openxlsx.footer           = NULL,
  openxlsx.evenFooter       = NULL,
  openxlsx.evenHeader       = NULL,
  openxlsx.gridLines        = TRUE,
  openxlsx.keepNA           = FALSE,
  openxlsx.lastColumn       = NULL,
  openxlsx.na.string        = NULL,
  openxlsx.maxWidth         = 250,
  openxlsx.minWidth         = 3,
  openxlsx.numFmt           = "GENERAL",
  openxlsx.oddFooter        = NULL,
  openxlsx.oddHeader        = NULL,
  openxlsx.orientation      = "portrait",
  openxlsx.paperSize        = 9,
  openxlsx.showGridLines    = NA,
  openxlsx.tabColour        = NULL,
  openxlsx.tableStyle       = "TableStyleLight9",
  openxlsx.vdpi             = 300,
  openxlsx.withFilter       = NULL
)

#' @param x An option name (`"openxlsx."` prefix optional)
#' @param default A default value if `NULL`
#' @rdname openxlsx_options
#' @export
openxlsx_getOp <- function(x, default = NULL) {
  if (length(x) != 1L || length(default) > 1L) {
   stop("x must be length 1 and default NULL or length 1", call. = FALSE)
  } 
  
  x <- check_openxlsx_op(x)
  getOption(x, op.openxlsx[[x]]) %||% default
}

#' @param value The new value for the option (optional if x is a named list)
#' @rdname openxlsx_options
#' @export
openxlsx_setOp <- function(x, value) {
  if (is.list(x)) {
    if (is.null(names(x))) {
      stop("x cannot be an unnamed list", call. = FALSE)
    }
    
    return(invisible(mapply(openxlsx_setOp, x = names(x), value = x)))
  }
  
  value <- as.list(value)
  names(value) <- check_openxlsx_op(x)
  options(value)
}

check_openxlsx_op <- function(x) {
  if (length(x) != 1L || !is.character(x)) {
    stop("option must be a character vector of length 1", call. = FALSE)
  }
  
  if (!grepl("^openxlsx[.]", x)) {
    x <- paste0("openxlsx.", x)
  }
  
  if (!x %in% names(op.openxlsx)) {
    warning(
      x, " is not a standard openxlsx option\nCheck spelling",
      call. = FALSE
    )
  }
  
  x
}

openxlsx_resetOp <- function() {
  options(op.openxlsx)
}

test_that("option names are appropriate", {
  bad <- grep("^openxlsx[.].*", names(op.openxlsx), value = TRUE, invert = TRUE)
  expect_equal(bad, character(0))
})

test_that("changing options", {
  op <- options()
  
  # Set via options()
  options(openxlsx.borders = "whatever")
  expect_equal(openxlsx_getOp("borders"), getOption("openxlsx.borders"))
  expect_equal(openxlsx_getOp("borders"), "whatever")
  
  # Set via openxlsx_setOp()
  openxlsx_setOp("borders", "new_whatever")
  expect_equal(openxlsx_getOp("borders"), getOption("openxlsx.borders"))
  expect_equal(openxlsx_getOp("borders"), "new_whatever")
  
  # with openxlsx. prefix
  openxlsx_setOp("openxlsx.borders", "new_new_whatever")
  expect_equal(openxlsx_getOp("openxlsx.borders"), getOption("openxlsx.borders"))
  expect_equal(openxlsx_getOp("openxlsx.borders"), "new_new_whatever")
  
  options(openxlsx.tableStyle = "Cool format")
  expect_equal(openxlsx_getOp("tableStyle"), openxlsx_getOp("openxlsx.tableStyle"))
  
  # Setting to NULL will return default
  options(openxlsx.borders = NULL)
  expect_equal(openxlsx_getOp("borders"), op.openxlsx[["openxlsx.borders"]])
  
  # Bad options names will trigger warning but still be produced
  options(openxlsx.likelyNotARealOption = TRUE)
  expect_warning(
    expect_true(openxlsx_getOp("likelyNotARealOption")),
    "not a standard openxlsx option"
  )
  
  # Multiple Ops returns error
  expect_error(openxlsx_getOp(c("withFilter", "borders")), "length 1")
  
  openxlsx_resetOp()
  options(op)
})

test_that("openxlsx_setOp() works with list [#215]", {
  op <- options()
  expect_error(openxlsx_setOp(list(withFilter = TRUE, keepNA = TRUE)), NA)
  openxlsx_resetOp()
  options(op)
})

#' @name conditionalFormatting
#' @aliases databar
#' @title Add conditional formatting to cells
#' @description Add conditional formatting to cells
#' @author Alexander Walker, Philipp Schauberger
#' @param wb A workbook object
#' @param sheet A name or index of a worksheet
#' @param cols Columns to apply conditional formatting to
#' @param rows Rows to apply conditional formatting to
#' @param rule The condition under which to apply the formatting. See examples.
#' @param style A style to apply to those cells that satisfy the rule. Default is createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
#' @param type Either 'expression', 'colourScale', 'databar', 'duplicates', 'beginsWith', 
#' 'endsWith', 'topN', 'bottomN', 'blanks', 'notBlanks', 'contains' or 'notContains' (case insensitive).
#' @param ... See below
#' @details See Examples.
#'
#' If type == "expression"
#' \itemize{
#'   \item{style is a Style object. See [createStyle()]}
#'   \item{rule is an expression. Valid operators are "<", "<=", ">", ">=", "==", "!=".}
#' }
#'
#' If type == "colourScale"
#' \itemize{
#'   \item{style is a vector of colours with length 2 or 3}
#'   \item{rule can be NULL or a vector of colours of equal length to styles}
#' }
#'
#' If type == "databar"
#' \itemize{
#'   \item{style is a vector of colours with length 2 or 3}
#'   \item{rule is a numeric vector specifying the range of the databar colours. Must be equal length to style}
#'   \item{...
#'   \itemize{
#'     \item{**showvalue** If FALSE the cell value is hidden. Default TRUE.}
#'     \item{**gradient** If FALSE colour gradient is removed. Default TRUE.}
#'     \item{**border** If FALSE the border around the database is hidden. Default TRUE.}
#'      }
#'    }
#' }
#'
#' If type == "duplicates"
#' \itemize{
#'   \item{style is a Style object. See [createStyle()]}
#'   \item{rule is ignored.}
#' }
#'
#' If type == "contains"
#' \itemize{
#'   \item{style is a Style object. See [createStyle()]}
#'   \item{rule is the text to look for within cells}
#' }
#'
#' If type == "between"
#' \itemize{
#'   \item{style is a Style object. See [createStyle()]}
#'   \item{rule is a numeric vector of length 2 specifying lower and upper bound (Inclusive)}
#' }
#' If type == "blanks"
#' \itemize{
#'   \item{style is a Style object. See [createStyle()]}
#'   \item{rule is ignored.}
#' }
#'
#' If type == "notBlanks"
#' \itemize{
#'   \item{style is a Style object. See [createStyle()]}
#'   \item{rule is ignored.}
#' }

#'
#' If type == "topN"
#' \itemize{
#'   \item{style is a Style object. See [createStyle()]}
#'   \item{rule is ignored}
#'   \item{...
#'   \itemize{
#'     \item{**rank** numeric vector of length 1 indicating number of highest values.}
#'     \item{**percent** TRUE if you want top N percentage.}
#'      }
#'    }
#' }
#' 
#' If type == "bottomN"
#' \itemize{
#'   \item{style is a Style object. See [createStyle()]}
#'   \item{rule is ignored}
#'   \item{...
#'   \itemize{
#'     \item{**rank** numeric vector of length 1 indicating number of lowest values.}
#'     \item{**percent** TRUE if you want bottom N percentage.}
#'      }
#'    }
#' }
#' 
#' @seealso [createStyle()]
#' @export
#' @examples
#' wb <- createWorkbook()
#' addWorksheet(wb, "cellIs")
#' addWorksheet(wb, "Moving Row")
#' addWorksheet(wb, "Moving Col")
#' addWorksheet(wb, "Dependent on")
#' addWorksheet(wb, "Duplicates")
#' addWorksheet(wb, "containsText")
#' addWorksheet(wb, "notcontainsText")
#' addWorksheet(wb, "beginsWith")
#' addWorksheet(wb, "endsWith")
#' addWorksheet(wb, "colourScale", zoom = 30)
#' addWorksheet(wb, "databar")
#' addWorksheet(wb, "between")
#' addWorksheet(wb, "topN")
#' addWorksheet(wb, "bottomN")
#' addWorksheet(wb, "containsBlanks")
#' addWorksheet(wb, "notContainsBlanks")
#' addWorksheet(wb, "logical operators")
#'
#' negStyle <- createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
#' posStyle <- createStyle(fontColour = "#006100", bgFill = "#C6EFCE")
#'
#' ## rule applies to all each cell in range
#' writeData(wb, "cellIs", -5:5)
#' writeData(wb, "cellIs", LETTERS[1:11], startCol = 2)
#' conditionalFormatting(wb, "cellIs",
#'   cols = 1,
#'   rows = 1:11, rule = "!=0", style = negStyle
#' )
#' conditionalFormatting(wb, "cellIs",
#'   cols = 1,
#'   rows = 1:11, rule = "==0", style = posStyle
#' )
#'
#' ## highlight row dependent on first cell in row
#' writeData(wb, "Moving Row", -5:5)
#' writeData(wb, "Moving Row", LETTERS[1:11], startCol = 2)
#' conditionalFormatting(wb, "Moving Row",
#'   cols = 1:2,
#'   rows = 1:11, rule = "$A1<0", style = negStyle
#' )
#' conditionalFormatting(wb, "Moving Row",
#'   cols = 1:2,
#'   rows = 1:11, rule = "$A1>0", style = posStyle
#' )
#'
#' ## highlight column dependent on first cell in column
#' writeData(wb, "Moving Col", -5:5)
#' writeData(wb, "Moving Col", LETTERS[1:11], startCol = 2)
#' conditionalFormatting(wb, "Moving Col",
#'   cols = 1:2,
#'   rows = 1:11, rule = "A$1<0", style = negStyle
#' )
#' conditionalFormatting(wb, "Moving Col",
#'   cols = 1:2,
#'   rows = 1:11, rule = "A$1>0", style = posStyle
#' )
#'
#' ## highlight entire range cols X rows dependent only on cell A1
#' writeData(wb, "Dependent on", -5:5)
#' writeData(wb, "Dependent on", LETTERS[1:11], startCol = 2)
#' conditionalFormatting(wb, "Dependent on",
#'   cols = 1:2,
#'   rows = 1:11, rule = "$A$1<0", style = negStyle
#' )
#' conditionalFormatting(wb, "Dependent on",
#'   cols = 1:2,
#'   rows = 1:11, rule = "$A$1>0", style = posStyle
#' )
#'
#' ## highlight cells in column 1 based on value in column 2
#' writeData(wb, "Dependent on", data.frame(x = 1:10, y = runif(10)), startRow = 15)
#' conditionalFormatting(wb, "Dependent on",
#'   cols = 1,
#'   rows = 16:25, rule = "B16<0.5", style = negStyle
#' )
#' conditionalFormatting(wb, "Dependent on",
#'   cols = 1,
#'   rows = 16:25, rule = "B16>=0.5", style = posStyle
#' )
#'
#'
#' ## highlight duplicates using default style
#' writeData(wb, "Duplicates", sample(LETTERS[1:15], size = 10, replace = TRUE))
#' conditionalFormatting(wb, "Duplicates", cols = 1, rows = 1:10, type = "duplicates")
#'
#' ## cells containing text
#' fn <- function(x) paste(sample(LETTERS, 10), collapse = "-")
#' writeData(wb, "containsText", sapply(1:10, fn))
#' conditionalFormatting(wb, "containsText", cols = 1, rows = 1:10, type = "contains", rule = "A")
#' 
#' ## cells not containing text
#' fn <- function(x) paste(sample(LETTERS, 10), collapse = "-")
#' writeData(wb, "containsText", sapply(1:10, fn))
#' conditionalFormatting(wb, "notcontainsText", cols = 1, 
#'                      rows = 1:10, type = "notcontains", rule = "A")
#' 
#' 
#' ## cells begins with text
#' fn <- function(x) paste(sample(LETTERS, 10), collapse = "-")
#' writeData(wb, "beginsWith", sapply(1:100, fn))
#' conditionalFormatting(wb, "beginsWith", cols = 1, rows = 1:100, type = "beginsWith", rule = "A")
#' 
#' 
#' ## cells ends with text
#' fn <- function(x) paste(sample(LETTERS, 10), collapse = "-")
#' writeData(wb, "endsWith", sapply(1:100, fn))
#' conditionalFormatting(wb, "endsWith", cols = 1, rows = 1:100, type = "endsWith", rule = "A")
#'
#' ## colourscale colours cells based on cell value
#' df <- read.xlsx(system.file("extdata", "readTest.xlsx", package = "openxlsx"), sheet = 4)
#' writeData(wb, "colourScale", df, colNames = FALSE) ## write data.frame
#'
#' ## rule is a vector or colours of length 2 or 3 (any hex colour or any of colours())
#' ## If rule is NULL, min and max of cells is used. Rule must be the same length as style or NULL.
#' conditionalFormatting(wb, "colourScale",
#'   cols = 1:ncol(df), rows = 1:nrow(df),
#'   style = c("black", "white"),
#'   rule = c(0, 255),
#'   type = "colourScale"
#' )
#'
#' setColWidths(wb, "colourScale", cols = 1:ncol(df), widths = 1.07)
#' setRowHeights(wb, "colourScale", rows = 1:nrow(df), heights = 7.5)
#'
#' ## Databars
#' writeData(wb, "databar", -5:5)
#' conditionalFormatting(wb, "databar", cols = 1, rows = 1:11, type = "databar") ## Default colours
#'
#' ## Between
#' # Highlight cells in interval [-2, 2]
#' writeData(wb, "between", -5:5)
#' conditionalFormatting(wb, "between", cols = 1, rows = 1:11, type = "between", rule = c(-2, 2))
#'
#' ## Top N 
#' writeData(wb, "topN", data.frame(x = 1:10, y = rnorm(10)))
#' # Highlight top 5 values in column x
#' conditionalFormatting(wb, "topN", cols = 1, rows = 2:11, 
#'  style = posStyle, type = "topN", rank = 5)#'
#' # Highlight top 20 percentage in column y
#' conditionalFormatting(wb, "topN", cols = 2, rows = 2:11, 
#'  style = posStyle, type = "topN", rank = 20, percent = TRUE)
#'
#'## Bottom N 
#' writeData(wb, "bottomN", data.frame(x = 1:10, y = rnorm(10)))
#' # Highlight bottom 5 values in column x
#' conditionalFormatting(wb, "bottomN", cols = 1, rows = 2:11, 
#'  style = negStyle, type = "topN", rank = 5)
#' # Highlight bottom 20 percentage in column y
#' conditionalFormatting(wb, "bottomN", cols = 2, rows = 2:11, 
#'  style = negStyle, type = "topN", rank = 20, percent = TRUE)
#'  
#' ## cells containing blanks
#' sample_data <- sample(c("X", NA_character_), 10, replace = TRUE)
#' writeData(wb, "containsBlanks", sample_data)
#' conditionalFormatting(wb, "containsBlanks", cols = 1, rows = 1:10, 
#' type = "blanks", style = negStyle)
#' 
#' ## cells not containing blanks
#' sample_data <- sample(c("X", NA_character_), 10, replace = TRUE)
#' writeData(wb, "notContainsBlanks", sample_data)
#' conditionalFormatting(wb, "notContainsBlanks", cols = 1, rows = 1:10, 
#' type = "notBlanks", style = posStyle)
#'
#' ## Logical Operators
#' # You can use Excels logical Operators
#' writeData(wb, "logical operators", 1:10)
#' conditionalFormatting(wb, "logical operators",
#'   cols = 1, rows = 1:10,
#'   rule = "OR($A1=1,$A1=3,$A1=5,$A1=7)"
#' )
#' \dontrun{
#' saveWorkbook(wb, "conditionalFormattingExample.xlsx", TRUE)
#' }
#'
#'
#' #########################################################################
#' ## Databar Example
#'
#' wb <- createWorkbook()
#' addWorksheet(wb, "databar")
#'
#' ## Databars
#' writeData(wb, "databar", -5:5, startCol = 1)
#' conditionalFormatting(wb, "databar", cols = 1, rows = 1:11, type = "databar") ## Defaults
#'
#' writeData(wb, "databar", -5:5, startCol = 3)
#' conditionalFormatting(wb, "databar", cols = 3, rows = 1:11, type = "databar", border = FALSE)
#'
#' writeData(wb, "databar", -5:5, startCol = 5)
#' conditionalFormatting(wb, "databar",
#'   cols = 5, rows = 1:11,
#'   type = "databar", style = c("#a6a6a6"), showValue = FALSE
#' )
#'
#' writeData(wb, "databar", -5:5, startCol = 7)
#' conditionalFormatting(wb, "databar",
#'   cols = 7, rows = 1:11,
#'   type = "databar", style = c("#a6a6a6"), showValue = FALSE, gradient = FALSE
#' )
#'
#' writeData(wb, "databar", -5:5, startCol = 9)
#' conditionalFormatting(wb, "databar",
#'   cols = 9, rows = 1:11,
#'   type = "databar", style = c("#a6a6a6", "#a6a6a6"), showValue = FALSE, gradient = FALSE
#' )
#' \dontrun{
#' saveWorkbook(wb, file = "databarExample.xlsx", overwrite = TRUE)
#' }
#'
conditionalFormatting <-
  function(wb,
           sheet,
           cols,
           rows,
           rule = NULL,
           style = NULL,
           type = "expression",
           ...) {
    op <- get_set_options()
    on.exit(options(op), add = TRUE)

    type <- tolower(type)
    params <- list(...)

    if (type %in% c("colorscale", "colourscale")) {
      type <- "colorScale"
    } else if (type == "databar") {
      type <- "dataBar"
    } else if (type == "duplicates") {
      type <- "duplicatedValues"
    } else if (type == "contains") {
      type <- "containsText"
    } else if (type == "notcontains") {
      type <- "notContainsText"
    } else if (type == "beginswith") {
      type <- "beginsWith"
    } else if (type == "endswith") {
      type <- "endsWith"
    } else if (type == "between") {
      type <- "between"
    } else if (type == "topn") {
      type <- "topN"
    } else if (type == "bottomn") {
      type <- "bottomN"
    } else if (type == "blanks") {
      type <- "containsBlanks"
    } else if (type == "notblanks") {
      type <- "notContainsBlanks"
    }else if (type != "expression") {
      stop(
        "Invalid type argument.  Type must be one of 'expression', 'colourScale', 'databar', 'duplicates', 'beginsWith', 'endsWith', 'blanks', 'notBlanks', 'contains' or 'notContains'"
      )
    }

    ## rows and cols
    if (!is.numeric(cols)) {
      cols <- convertFromExcelRef(cols)
    }
    rows <- as.integer(rows)


    ## check valid rule
    values <- NULL
    dxfId <- NULL

    if (type == "colorScale") {
      # type == "colourScale"
      # - style is a vector of colours with length 2 or 3
      # - rule specifies the quantiles (numeric vector of length 2 or 3), if NULL min and max are used

      if (is.null(style)) {
        stop("If type == 'colourScale', style must be a vector of colours of length 2 or 3.")
      }

      if (!inherits(style, "character")) {
        stop("If type == 'colourScale', style must be a vector of colours of length 2 or 3.")
      }

      if (!length(style) %in% 2:3) {
        stop("If type == 'colourScale', style must be a vector of length 2 or 3.")
      }

      if (!is.null(rule)) {
        if (length(rule) != length(style)) {
          stop("If type == 'colourScale', rule and style must have equal lengths.")
        }
      }

      style <-
        validateColour(style, errorMsg = "Invalid colour specified in style.")

      values <- rule
      rule <- style
    } else if (type == "dataBar") {
      # type == "databar"
      # - style is a vector of colours of length 2 or 3
      # - rule specifies the quantiles (numeric vector of length 2 or 3), if NULL min and max are used

      if (is.null(style)) {
        style <- "#638EC6"
      }

      if (!inherits(style, "character")) {
        stop("If type == 'dataBar', style must be a vector of colours of length 1 or 2.")
      }

      if (!length(style) %in% 1:2) {
        stop("If type == 'dataBar', style must be a vector of length 1 or 2.")
      }

      if (!is.null(rule)) {
        if (length(rule) != length(style)) {
          stop("If type == 'dataBar', rule and style must have equal lengths.")
        }
      }


      ## Additional parameters passed by ...
      if ("showValue" %in% names(params)) {
        params$showValue <- as.integer(params$showValue)
        if (is.na(params$showValue)) {
          stop("showValue must be 0/1 or TRUE/FALSE")
        }
      }

      if ("gradient" %in% names(params)) {
        params$gradient <- as.integer(params$gradient)
        if (is.na(params$gradient)) {
          stop("gradient must be 0/1 or TRUE/FALSE")
        }
      }

      if ("border" %in% names(params)) {
        params$border <- as.integer(params$border)
        if (is.na(params$border)) {
          stop("border must be 0/1 or TRUE/FALSE")
        }
      }

      style <-
        validateColour(style, errorMsg = "Invalid colour specified in style.")

      values <- rule
      rule <- style
    } else if (type == "expression") {
      # type == "expression"
      # - style = createStyle()
      # - rule is an expression to evaluate

      # rule <- gsub(" ", "", rule)
      rule <- replaceIllegalCharacters(rule)
      rule <- gsub("!=", "&lt;&gt;", rule)
      rule <- gsub("==", "=", rule)

      if (!grepl("[A-Z]", substr(rule, 1, 2))) {
        ## formula looks like "operatorX" , attach top left cell to rule
        rule <-
          paste0(getCellRefs(data.frame(
            "x" = min(rows), "y" = min(cols)
          )), rule)
      } ## else, there is a letter in the formula and apply as is

      if (is.null(style)) {
        style <-
          createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
      }

      if (!"Style" %in% class(style)) {
        stop("If type == 'expression', style must be a Style object.")
      }

      invisible(dxfId <- wb$addDXFS(style))
    } else if (type == "duplicatedValues") {
      # type == "duplicatedValues"
      # - style is a Style object
      # - rule is ignored

      if (is.null(style)) {
        style <-
          createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
      }

      if (!"Style" %in% class(style)) {
        stop("If type == 'duplicates', style must be a Style object.")
      }

      invisible(dxfId <- wb$addDXFS(style))
      rule <- style
    } else if (type == "containsText") {
      # type == "contains"
      # - style is Style object
      # - rule is text to look for
      
      if (is.null(style)) {
        style <-
          createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
      }
      
      
      if (!"character" %in% class(rule)) {
        stop("If type == 'contains', rule must be a character vector of length 1.")
      }
      
      if (!"Style" %in% class(style)) {
        stop("If type == 'contains', style must be a Style object.")
      }
      
      invisible(dxfId <- wb$addDXFS(style))
      values <- rule
      rule <- style
    } else if (type == "notContainsText") {
      # type == "contains"
      # - style is Style object
      # - rule is text to look for
      
      if (is.null(style)) {
        style <-
          createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
      }
      
      
      if (!"character" %in% class(rule)) {
        stop("If type == 'notContains', rule must be a character vector of length 1.")
      }
      
      if (!"Style" %in% class(style)) {
        stop("If type == 'notContains', style must be a Style object.")
      }
      
      invisible(dxfId <- wb$addDXFS(style))
      values <- rule
      rule <- style
    } else if (type == "beginsWith") {
      # type == "contains"
      # - style is Style object
      # - rule is text to look for
      
      if (is.null(style)) {
        style <-
          createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
      }
      
      
      if (!"character" %in% class(rule)) {
        stop("If type == 'beginsWith', rule must be a character vector of length 1.")
      }
      
      if (!"Style" %in% class(style)) {
        stop("If type == 'beginsWith', style must be a Style object.")
      }
      
      invisible(dxfId <- wb$addDXFS(style))
      values <- rule
      rule <- style
    } else if (type == "endsWith") {
      # type == "contains"
      # - style is Style object
      # - rule is text to look for
      
      if (is.null(style)) {
        style <-
          createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
      }
      
      
      if (!"character" %in% class(rule)) {
        stop("If type == 'endsWith', rule must be a character vector of length 1.")
      }
      
      if (!"Style" %in% class(style)) {
        stop("If type == 'endsWith', style must be a Style object.")
      }
      
      invisible(dxfId <- wb$addDXFS(style))
      values <- rule
      rule <- style
    } else if (type == "between") {
      rule <- range(rule)

      if (is.null(style)) {
        style <-
          createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
      }

      if (!"Style" %in% class(style)) {
        stop("If type == 'between', style must be a Style object.")
      }

      invisible(dxfId <- wb$addDXFS(style))
    } else if (type == "topN") {
      # type == "topN"
      # - rule is ignored
      # - 'rank' and 'percent' are named params 
      
      if (is.null(style)) {
        style <-
          createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
      } 
      
      if (!"Style" %in% class(style)) {
        stop("If type == 'topN', style must be a Style object.")
      } 
      
      invisible(dxfId <- wb$addDXFS(style)) 
      
      ## Additional parameters passed by ...
      if ("percent" %in% names(params)) {
        params$percent <- as.integer(params$percent)
        if (is.na(params$percent)) {
          stop("percent must be 0/1 or TRUE/FALSE")
        }
      } 
      
      if ("rank" %in% names(params)) {
        params$rank <- as.integer(params$rank)
        if (is.na(params$rank)) {
          stop("rank must be a number")
        }
      } 
      
      invisible(dxfId <- wb$addDXFS(style))
      values <- params
      rule <- style
    } else if (type == "bottomN") {
      # type == "bottomN"
      # - rule is ignored
      # - 'rank' and 'percent' are named params 
      
      if (is.null(style)) {
        style <-
          createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
      } 
      
      if (!"Style" %in% class(style)) {
        stop("If type == 'bottomN', style must be a Style object.")
      }
      
      invisible(dxfId <- wb$addDXFS(style)) 
      
      ## Additional parameters passed by ...
      if ("percent" %in% names(params)) {
        params$percent <- as.integer(params$percent)
        if (is.na(params$percent)) {
          stop("percent must be 0/1 or TRUE/FALSE")
        }
      } 
      
      if ("rank" %in% names(params)) {
        params$rank <- as.integer(params$rank)
        if (is.na(params$rank)) {
          stop("rank must be a number")
        }
      } 
      
      invisible(dxfId <- wb$addDXFS(style))
      values <- params
      rule <- style
    }  else if (type == "containsBlanks") {
      # rule is ignored
      
      if (is.null(style)) {
        style <-
          createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
      }
      
      if (!"Style" %in% class(style)) {
        stop("If type == 'blanks', style must be a Style object.")
      }
      
      invisible(dxfId <- wb$addDXFS(style))
    } else if (type == "notContainsBlanks") {
      # rule is ignored
      
      if (is.null(style)) {
        style <-
          createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
      }
      
      if (!"Style" %in% class(style)) {
        stop("If type == 'notBlanks', style must be a Style object.")
      }
      
      invisible(dxfId <- wb$addDXFS(style))
    }



    invisible(
      wb$conditionalFormatting(
        sheet,
        startRow = min(rows),
        endRow = max(rows),
        startCol = min(cols),
        endCol = max(cols),
        dxfId = dxfId,
        formula = rule,
        type = type,
        values = values,
        params = params
      )
    )

    invisible(0)
  }

#' @name openXL
#' @title Open a Microsoft Excel file (xls/xlsx) or an openxlsx Workbook
#' @author Luca Braglia
#' @description This function tries to open a Microsoft Excel
#' (xls/xlsx) file or an openxlsx Workbook with the proper
#' application, in a portable manner.
#'
#' In Windows (c) and Mac (c), it uses system default handlers,
#' given the file type.
#'
#' In Linux it searches (via `which`) for available xls/xlsx
#' reader applications (unless `options('openxlsx.excelApp')`
#' is set to the app bin path), and if it finds anything, sets
#' `options('openxlsx.excelApp')` to the program choosen by
#' the user via a menu (if many are present, otherwise it will
#' set the only available). Currently searched for apps are
#' Libreoffice/Openoffice (`soffice` bin), Gnumeric
#' (`gnumeric`) and Calligra Sheets (`calligrasheets`).
#'
#' @param file path to the Excel (xls/xlsx) file or Workbook object.
#' @usage openXL(file=NULL)
#' @export openXL
#' @examples
#' # file example
#' example(writeData)
#' # openXL("writeDataExample.xlsx")
#'
#' # (not yet saved) Workbook example
#' wb <- createWorkbook()
#' x <- mtcars[1:6, ]
#' addWorksheet(wb, "Cars")
#' writeData(wb, "Cars", x, startCol = 2, startRow = 3, rowNames = TRUE)
#' # openXL(wb)
openXL <- function(file = NULL) {
  op <- get_set_options()
  on.exit(options(op), add = TRUE)

  if (is.null(file)) stop("A file has to be specified.")

  ## workbook handling
  if ("Workbook" %in% class(file)) {
    file <- file$saveWorkbook()
  }

  if (!file.exists(file)) stop("Non existent file or wrong path.")

  ## execution should be in background in order to not block R
  ## interpreter
  file <- normalizePath(file)
  userSystem <- Sys.info()["sysname"]


  if ("Linux" == userSystem) {
    if (is.null(app <- unlist(options("openxlsx.excelApp")))) {
      app <- chooseExcelApp()
    }
    myCommand <- paste(app, file, "&", sep = " ")
    system(command = myCommand)
  } else if ("Windows" == userSystem) {
    shell(shQuote(string = file), wait = FALSE) #nolint
  } else if ("Darwin" == userSystem) {
    myCommand <- paste0("open ", file)
    system(command = myCommand)
  } else {
    warning("Operating system not handled.")
  }
}


chooseExcelApp <- function() {
  m <- c(
    `Libreoffice/OpenOffice` = "soffice",
    `Calligra Sheets` = "calligrasheets",
    `Gnumeric` = "gnumeric"
  )

  prog <- Sys.which(m)
  names(prog) <- names(m)
  nApps <- length(availProg <- prog["" != prog])

  if (0 == nApps) {
    stop(
      "No applications (detected) available.\n",
      "Set options('openxlsx.excelApp'), instead."
    )
  } else if (1 == nApps) {
    cat("Only", names(availProg), "found; I'll use it.\n")
    unnprog <- unname(availProg)
    options(openxlsx.excelApp = unnprog)
    invisible(unnprog)
  } else if (1 < nApps) {
    if (!interactive()) {
      stop(
        "Cannot choose an Excel file opener non-interactively.\n",
        "Set options('openxlsx.excelApp'), instead."
      )
    }
    res <- menu(names(availProg), title = "Excel Apps availables")
    unnprog <- unname(availProg[res])
    if (res > 0L) options(openxlsx.excelApp = unnprog)
    invisible(unname(unnprog))
  } else {
    stop("Unexpected error.")
  }
}

#' @name write.xlsx
#' @title write data to an xlsx file
#' @description write a data.frame or list of data.frames to an xlsx file
#' @author Alexander Walker, Jordan Mark Barbone
#' @inheritParams buildWorkbook
#' @param file A file path to save the xlsx file
#' @param overwrite Overwrite existing file (Defaults to `TRUE` as with `write.table`)
#' @param ... Additional arguments passed to [buildWorkbook()]; see details
#' 
#' @inheritSection buildWorkbook Optional Parameters
#'
#' @seealso [addWorksheet()]
#' @seealso [writeData()]
#' @seealso [createStyle()] for style parameters
#' @seealso [buildWorkbook()]
#' @return A workbook object
#' @examples
#'
#' ## write to working directory
#' options("openxlsx.borderColour" = "#4F80BD") ## set default border colour
#' \dontrun{
#' write.xlsx(iris, file = "writeXLSX1.xlsx", colNames = TRUE, borders = "columns")
#' write.xlsx(iris, file = "writeXLSX2.xlsx", colNames = TRUE, borders = "surrounding")
#' }
#'
#'
#' hs <- createStyle(
#'   textDecoration = "BOLD", fontColour = "#FFFFFF", fontSize = 12,
#'   fontName = "Arial Narrow", fgFill = "#4F80BD"
#' )
#' \dontrun{
#' write.xlsx(iris,
#'   file = "writeXLSX3.xlsx",
#'   colNames = TRUE, borders = "rows", headerStyle = hs
#' )
#' }
#'
#' ## Lists elements are written to individual worksheets, using list names as sheet names if available
#' l <- list("IRIS" = iris, "MTCATS" = mtcars, matrix(runif(1000), ncol = 5))
#' \dontrun{
#' write.xlsx(l, "writeList1.xlsx", colWidths = c(NA, "auto", "auto"))
#' }
#'
#' ## different sheets can be given different parameters
#' \dontrun{
#' write.xlsx(l, "writeList2.xlsx",
#'   startCol = c(1, 2, 3), startRow = 2,
#'   asTable = c(TRUE, TRUE, FALSE), withFilter = c(TRUE, FALSE, FALSE)
#' )
#' }
#'
#' # specify column widths for multiple sheets
#' \dontrun{
#' write.xlsx(l, "writeList2.xlsx", colWidths = 20)
#' write.xlsx(l, "writeList2.xlsx", colWidths = list(100, 200, 300))
#' write.xlsx(l, "writeList2.xlsx", colWidths = list(rep(10, 5), rep(8, 11), rep(5, 5)))
#' }
#'
#' @export
write.xlsx <- function(x, file, asTable = FALSE, overwrite = TRUE, ...) {
  if ("matrix" %in% class(x)){
    x <- as.data.frame(x)
  }
  wb <- buildWorkbook(x, asTable = asTable, ...)
  saveWorkbook(wb, file = file, overwrite = overwrite)
  invisible(wb)
}

---
title: "Introduction"
author: "Alexander Walker, Philipp Schauberger"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Basic Examples


### write.xlsx

The simplest way to write to a workbook is write.xlsx().
By default, write.xlsx calls writeData.  If asTable is TRUE write.xlsx will write x as
an Excel table.

```{r include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE}
## write to working directory
library(openxlsx)
write.xlsx(iris, file = "writeXLSX1.xlsx")
write.xlsx(iris, file = "writeXLSXTable1.xlsx", asTable = TRUE)
```

### write list of data.frames to xlsx-file

```{r  include = TRUE, tidy = TRUE, eval = FALSE ,highlight = TRUE}
## write a list of data.frames to individual worksheets using list names as worksheet names
l <- list("IRIS" = iris, "MTCARS" = mtcars)
write.xlsx(l, file = "writeXLSX2.xlsx")
write.xlsx(l, file = "writeXLSXTable2.xlsx", asTable = TRUE)
```


### write.xlsx also accepts styling parameters


#### The simplest way is to set default options and set column class

```{r include = TRUE, tidy = TRUE, eval = FALSE ,highlight= TRUE}
options("openxlsx.borderColour" = "#4F80BD")
options("openxlsx.borderStyle" = "thin")
options("openxlsx.dateFormat" = "mm/dd/yyyy")
options("openxlsx.datetimeFormat" = "yyyy-mm-dd hh:mm:ss")
options("openxlsx.numFmt" = NULL) ## For default style rounding of numeric columns

df <- data.frame("Date" = Sys.Date()-0:19, "LogicalT" = TRUE,
                 "Time" = Sys.time()-0:19*60*60,
                 "Cash" = paste("$",1:20), "Cash2" = 31:50,
                 "hLink" = "https://CRAN.R-project.org/",
                 "Percentage" = seq(0, 1, length.out=20),
                 "TinyNumbers" = runif(20) / 1E9,  stringsAsFactors = FALSE)

class(df$Cash) <- "currency"
class(df$Cash2) <- "accounting"
class(df$hLink) <- "hyperlink"
class(df$Percentage) <- "percentage"
class(df$TinyNumbers) <- "scientific"

write.xlsx(df, "writeXLSX3.xlsx")
write.xlsx(df, file = "writeXLSXTable3.xlsx", asTable = TRUE)


```


## Workbook styles

### define a style for column headers

```{r include = TRUE, tidy = TRUE, eval = FALSE, highlight= TRUE}
hs <- createStyle(fontColour = "#ffffff", fgFill = "#4F80BD",
                  halign = "center", valign = "center", textDecoration = "Bold",
                  border = "TopBottomLeftRight", textRotation = 45)

write.xlsx(iris, file = "writeXLSX4.xlsx", borders = "rows", headerStyle = hs)
write.xlsx(iris, file = "writeXLSX5.xlsx", borders = "columns", headerStyle = hs)

write.xlsx(iris, "writeXLSXTable4.xlsx", asTable = TRUE,
headerStyle = createStyle(textRotation = 45))
```


### When writing a list, the stylings will apply to all list elements

```{r include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE}
l <- list("IRIS" = iris, "colClasses" = df)
write.xlsx(l, file = "writeXLSX6.xlsx", borders = "columns", headerStyle = hs)
write.xlsx(l, file = "writeXLSXTable5.xlsx", asTable = TRUE, tableStyle = "TableStyleLight2")

openXL("writeXLSX6.xlsx")
openXL("writeXLSXTable5.xlsx")
```


### write.xlsx returns the workbook object for further editing
```{r include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE}
wb <- write.xlsx(iris, "writeXLSX6.xlsx")
setColWidths(wb, sheet = 1, cols = 1:5, widths = 20)
saveWorkbook(wb, "writeXLSX6.xlsx", overwrite = TRUE)
```


## Workbook creation walk-through

### create workbook and set default border Colour and style

```{r include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE}
require(ggplot2)
wb <- createWorkbook()
options("openxlsx.borderColour" = "#4F80BD")
options("openxlsx.borderStyle" = "thin")
modifyBaseFont(wb, fontSize = 10, fontName = "Arial Narrow")
```


### Add Sheets 

```{r include = TRUE,tidy = TRUE, eval = FALSE, highlight = TRUE}
addWorksheet(wb, sheetName = "Motor Trend Car Road Tests", gridLines = FALSE)
addWorksheet(wb, sheetName = "Iris", gridLines = FALSE)
```


### write data to sheet 1

```{r include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE}
freezePane(wb, sheet = 1, firstRow = TRUE, firstCol = TRUE) ## freeze first row and column
writeDataTable(wb, sheet = 1, x = mtcars,
colNames = TRUE, rowNames = TRUE,
tableStyle = "TableStyleLight9")

setColWidths(wb, sheet = 1, cols = "A", widths = 18)
```


### write data to sheet 2

iris data.frame is added as excel table on sheet 2.

```{r include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE}
writeDataTable(wb, sheet = 2, iris, startCol = "K", startRow = 2)

qplot(data=iris, x = Sepal.Length, y= Sepal.Width, colour = Species)
insertPlot(wb, 2, xy=c("B", 16)) ## insert plot at cell B16

means <- aggregate(x = iris[,-5], by = list(iris$Species), FUN = mean)
vars <- aggregate(x = iris[,-5], by = list(iris$Species), FUN = var)
```


### add write group means

```{r include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE}
headSty <- createStyle(fgFill="#DCE6F1", halign="center", border = "TopBottomLeftRight")
writeData(wb, 2, x = "Iris dataset group means", startCol = 2, startRow = 2)
writeData(wb, 2, x = means, startCol = "B", startRow=3, borders="rows", headerStyle = headSty)
```

### add write group variances

```{r include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE}
writeData(wb, 2, x = "Iris dataset group variances", startCol = 2, startRow = 9)
writeData(wb, 2, x= vars, startCol = "B", startRow=10, borders="columns",
headerStyle = headSty)

setColWidths(wb, 2, cols=2:6, widths = 12) ## width is recycled for each col
setColWidths(wb, 2, cols=11:15, widths = 15)
```

### add style mean & variance table headers

```{r include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE}
s1 <- createStyle(fontSize=14, textDecoration=c("bold", "italic"))
addStyle(wb, 2, style = s1, rows=c(2,9), cols=c(2,2))
```

### save workbook

```{r include = TRUE, tidy = TRUE, eval = FALSE, highlight = TRUE}
saveWorkbook(wb, "basics.xlsx", overwrite = TRUE) ## save to working directory
```


## Gallery 

```{r eval = FALSE, include = TRUE}
## inspired by xtable gallery
#https://CRAN.R-project.org/package=xtable/vignettes/xtableGallery.pdf

## Create a new workbook
wb <- createWorkbook()
data(tli, package = "xtable")

## data.frame
test.n <- "data.frame"
my.df <- tli[1:10, ]
addWorksheet(wb = wb, sheetName = test.n)
writeData(wb = wb, sheet = test.n, x = my.df, borders = "n")

## matrix
test.n <- "matrix"
design.matrix <- model.matrix(~ sex * grade, data = my.df)
addWorksheet(wb = wb, sheetName = test.n)
writeData(wb = wb, sheet = test.n, x = design.matrix)

## aov
test.n <- "aov"
fm1 <- aov(tlimth ~ sex + ethnicty + grade + disadvg, data = tli)
addWorksheet(wb = wb, sheetName = test.n)
writeData(wb = wb, sheet = test.n, x = fm1)

## lm
test.n <- "lm"
fm2 <- lm(tlimth ~ sex*ethnicty, data = tli)
addWorksheet(wb = wb, sheetName = test.n)
writeData(wb = wb, sheet = test.n, x = fm2)

## anova 1 
test.n <- "anova"
my.anova <- anova(fm2)
addWorksheet(wb = wb, sheetName = test.n)
writeData(wb = wb, sheet = test.n, x = my.anova)

## anova 2
test.n <- "anova2"
fm2b <- lm(tlimth ~ ethnicty, data = tli)
my.anova2 <- anova(fm2b, fm2)
addWorksheet(wb = wb, sheetName = test.n)
writeData(wb = wb, sheet = test.n, x = my.anova2)

## glm
test.n <- "glm"
fm3 <- glm(disadvg ~ ethnicty*grade, data = tli, family = binomial())
addWorksheet(wb = wb, sheetName = test.n)
writeData(wb = wb, sheet = test.n, x = fm3)

## prcomp
test.n <- "prcomp"
pr1 <- prcomp(USArrests)
addWorksheet(wb = wb, sheetName = test.n)
writeData(wb = wb, sheet = test.n, x = pr1)

## summary.prcomp
test.n <- "summary.prcomp"
addWorksheet(wb = wb, sheetName = test.n)
writeData(wb = wb, sheet = test.n, x = summary(pr1))

## simple table
test.n <- "table"
data(airquality)
airquality$OzoneG80 <- factor(airquality$Ozone > 80,
levels = c(FALSE, TRUE),
labels = c("Oz <= 80", "Oz > 80"))
airquality$Month <- factor(airquality$Month,
levels = 5:9,
labels = month.abb[5:9])
my.table <- with(airquality, table(OzoneG80,Month) )
addWorksheet(wb = wb, sheetName = test.n)
writeData(wb = wb, sheet = test.n, x = my.table)

## survdiff 1
library(survival)
test.n <- "survdiff1"
addWorksheet(wb = wb, sheetName = test.n)
x <- survdiff(Surv(futime, fustat) ~ rx, data = ovarian)
writeData(wb = wb, sheet = test.n, x = x)

## survdiff 2
test.n <- "survdiff2"
addWorksheet(wb = wb, sheetName = test.n)
expect <- survexp(futime ~ ratetable(age=(accept.dt - birth.dt),
      sex=1,year=accept.dt,race="white"), jasa, cohort=FALSE,
      ratetable=survexp.usr)
x <- survdiff(Surv(jasa$futime, jasa$fustat) ~ offset(expect))
writeData(wb = wb, sheet = test.n, x = x)

## coxph 1
test.n <- "coxph1"
addWorksheet(wb = wb, sheetName = test.n)
bladder$rx <- factor(bladder$rx, labels = c("Pla","Thi"))
x <- coxph(Surv(stop,event) ~ rx, data = bladder)
writeData(wb = wb, sheet = test.n, x = x)

## coxph 2
test.n <- "coxph2"
addWorksheet(wb = wb, sheetName = test.n)
x <- coxph(Surv(stop,event) ~ rx + cluster(id), data = bladder)
writeData(wb = wb, sheet = test.n, x = x)

## cox.zph
test.n <- "cox.zph"
addWorksheet(wb = wb, sheetName = test.n)
x <- cox.zph(coxph(Surv(futime, fustat) ~ age + ecog.ps,  data=ovarian))
writeData(wb = wb, sheet = test.n, x = x)

## summary.coxph 1
test.n <- "summary.coxph1"
addWorksheet(wb = wb, sheetName = test.n)
x <- summary(coxph(Surv(stop,event) ~ rx, data = bladder))
writeData(wb = wb, sheet = test.n, x = x)

## summary.coxph 2
test.n <- "summary.coxph2"
addWorksheet(wb = wb, sheetName = test.n)
x <- summary(coxph(Surv(stop,event) ~ rx + cluster(id), data = bladder))
writeData(wb = wb, sheet = test.n, x = x)

## view without saving
openXL(wb)

```


## Further Examples

### Stock Price

```{r eval = FALSE, include = TRUE, tidy = TRUE, highlight= TRUE}
require(ggplot2)

wb <- createWorkbook()

## read historical prices from yahoo finance
ticker <- "CBA.AX"
csv.url <- paste0("https://query1.finance.yahoo.com/v7/finance/download/",
ticker, "?period1=1597597610&period2=1629133610&interval=1d&events=history&includeAdjustedClose= TRue")
prices <- read.csv(url(csv.url), as.is = TRUE)
prices$Date <- as.Date(prices$Date)
close <- prices$Close
prices$logReturns = c(0, log(close[2:length(close)]/close[1:(length(close)-1)]))

## Create plot of price series and add to worksheet
ggplot(data = prices, aes(as.Date(Date), as.numeric(Close))) +
geom_line(colour="royalblue2") +
labs(x = "Date", y = "Price", title = ticker) +
geom_area(fill = "royalblue1",alpha = 0.3) +
coord_cartesian(ylim=c(min(prices$Close)-1.5, max(prices$Close)+1.5)) 

## Add worksheet and write plot to sheet
addWorksheet(wb, sheetName = "CBA")
insertPlot(wb, sheet = 1, xy = c("J", 3))

## Histogram of log returns
ggplot(data = prices, aes(x = logReturns)) + geom_histogram(binwidth=0.0025) +
labs(title = "Histogram of log returns")

## currency 
class(prices$Close) <- "currency" ## styles as currency in workbook

## write historical data and  histogram of returns
writeDataTable(wb, sheet = "CBA", x = prices)
insertPlot(wb, sheet = 1, startRow=25, startCol = "J")

## Add conditional formatting to show where logReturn > 0.01 using default style
conditionalFormat(wb, sheet = 1, cols = seq_len((prices)), rows = 2:(nrow(prices)+1),
rule = "$H2 > 0.01")

## style log return col as a percentage  
logRetStyle <- createStyle(numFmt = "percentage")

addStyle(wb, 1, style = logRetStyle, rows = 2:(nrow(prices) + 1), 
cols = "H", gridExpand = TRUE)

setColWidths(wb, sheet=1, cols = c("A", "F", "G", "H"), widths = 15)

## save workbook to working directory
saveWorkbook(wb, "stockPrice.xlsx", overwrite = TRUE)
openXL("stockPrice.xlsx")
```


### Image Compression using PCA

```{r eval = FALSE, include = TRUE}
require(openxlsx)
require(jpeg)
require(ggplot2)

plotFn <- function(x, ...) {
  colvec <- grey(x)
  colmat <- array(match(colvec, unique(colvec)), dim = dim(x)[1:2])
  image(x = 0:(dim(colmat)[2]), y = 0:(dim(colmat)[1]), z = t(colmat[rev(seq_len(nrow(colmat))) , ]),
    col = unique(colvec), xlab = "", ylab = "", axes = FALSE, asp = 1,
    bty ="n", frame.plot=FALSE, ann=FALSE)
}

## Create workbook and add a worksheet, hide gridlines
wb <- createWorkbook("Einstein")
addWorksheet(wb, "Original Image", gridLines = FALSE)

A <- readJPEG(file.path(path.package("openxlsx"), "einstein.jpg"))
height <- nrow(A)
width <- ncol(A)

## write "Original Image" to cell B2
writeData(wb, 1, "Original Image", xy = c(2,2))

## write Object size to cell B3
writeData(wb, 1, sprintf("Image object size: %s bytes",
                         format(object.size(A+0)[[1]], big.mark=',')), 
          xy = c(2,3))  ## equivalent to startCol = 2, startRow = 3

## Plot image
par(mar=rep(0, 4), xpd = NA)
plotFn(A)

## insert plot currently showing in plot window
insertPlot(wb, 1, width, height, units="px", startRow= 5, startCol = 2)       

## SVD of covariance matrix
rMeans <- rowMeans(A)
rowMeans <- do.call("cbind", lapply(seq_len(ncol(A)), function(X) rMeans))
A <- A - rowMeans
E <- svd(A %*% t(A) / (ncol(A) - 1)) # SVD on covariance matrix of A
pve <- data.frame("Eigenvalues" = E$d, 
                  "PVE" = E$d/sum(E$d),
                  "Cumulative PVE" = cumsum(E$d/sum(E$d)))

## write eigenvalues to worksheet
addWorksheet(wb, "Principal Component Analysis")
hs <- createStyle(fontColour = "#ffffff", fgFill = "#4F80BD",
                  halign = "CENTER", textDecoration = "Bold",
                  border = "TopBottomLeftRight", borderColour = "#4F81BD")

writeData(wb, 2, x="Proportions of variance explained by Eigenvector" ,startRow = 2)
mergeCells(wb, sheet=2, cols=1:4, rows=2)

setColWidths(wb, 2, cols = 1:3, widths = c(14, 12, 15))
writeData(wb, 2, x=pve, startRow = 3, startCol = 1, borders="rows", headerStyle=hs)

## Plots
pve <- cbind(pve, "Ind" = seq_len(nrow(pve)))
ggplot(data = pve[1:20,], aes(x = Ind, y = 100*PVE)) +
  geom_bar(stat="identity", position = "dodge") +
  xlab("Principal Component Index") + ylab("Proportion of Variance Explained") +
  geom_line(size = 1, col = "blue") + geom_point(size = 3, col = "blue")

## Write plot to worksheet 2
insertPlot(wb, 2, width = 5, height = 4, startCol = "E", startRow = 2) 

## Plot of cumulative explained variance
ggplot(data = pve[1:50,], aes(x = Ind, y = 100*Cumulative.PVE)) +
  geom_point(size=2.5) + geom_line(size=1) + xlab("Number of PCs") +
  ylab("Cumulative Proportion of Variance Explained")
insertPlot(wb, 2, width = 5, height = 4, xy= c("M", 2)) 


## Reconstruct image using increasing number of PCs
nPCs <- c(5, 7, 12, 20, 50, 200)
startRow <- rep(c(2, 24), each = 3)
startCol <- rep(c("B", "H", "N"), 2)

## create a worksheet to save reconstructed images to
addWorksheet(wb, "Reconstructed Images", zoom = 90)

for(i in seq_len(length(nPCs))) {
  
  V <- E$v[, 1:nPCs[i]]
  imgHat <- t(V) %*% A  ## project img data on to PCs
  imgSize <- object.size(V) + object.size(imgHat) + object.size(rMeans)
  
  imgHat <- V %*% imgHat + rowMeans  ## reconstruct from PCs and add back row means
  imgHat <- round((imgHat - min(imgHat)) / (max(imgHat) - min(imgHat))*255) # scale
  plotFn(imgHat/255)
  
  ## write strings to worksheet 3
  writeData(wb, "Reconstructed Images", 
            sprintf("Number of principal components used:  %s", 
                    nPCs[[i]]), startCol[i], startRow[i])
  
  writeData(wb, "Reconstructed Images", 
            sprintf("Sum of component object sizes: %s bytes",
                    format(as.numeric(imgSize), big.mark=',')), startCol[i], startRow[i]+1)
  
  ## write reconstruced image
  insertPlot(wb, "Reconstructed Images", width, height, units="px",
             xy = c(startCol[i], startRow[i]+3))
  
}

# hide grid lines
showGridLines(wb, sheet = 3, showGridLines = FALSE)

## Make text above images BOLD
boldStyle <- createStyle(textDecoration="BOLD")

## only want to apply style to specified cells (not all combinations of rows & cols)
addStyle(wb, "Reconstructed Images", style=boldStyle, 
         rows = c(startRow, startRow+1), cols = rep(startCol, 2), 
         gridExpand = FALSE)  

## save workbook to working directory
saveWorkbook(wb, "Image dimensionality reduction.xlsx", overwrite = TRUE) 




## remove example files for cran test
if (identical(Sys.getenv("NOT_CRAN", unset = "true"), "false")) {
  file_list<-list.files(pattern="\\.xlsx",recursive = TRUE)
  file_list<-fl[!grepl("inst/extdata",file_list)&!grepl("man/",file_list)]
  
  if(length(file_list)>0) {
    rm(file_list)
  }
}
```

```{r cleanup, eval = FALSE, include = FALSE}
xlsx_files <- dir(pattern = "*.xlsx")
unlink(xlsx_files)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{getCellRefs}
\alias{getCellRefs}
\title{Return excel cell coordinates from (x,y) coordinates}
\usage{
getCellRefs(cellCoords)
}
\arguments{
\item{cellCoords}{A data.frame with two columns coordinate pairs.}
}
\value{
Excel alphanumeric cell reference
}
\description{
Return excel cell coordinates from (x,y) coordinates
}
\examples{
getCellRefs(data.frame(1, 2))
# "B1"
getCellRefs(data.frame(1:3, 2:4))
# "B1" "C2" "D3"
}
\author{
Philipp Schauberger, Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writexlsx.R
\name{write.xlsx}
\alias{write.xlsx}
\title{write data to an xlsx file}
\usage{
write.xlsx(x, file, asTable = FALSE, overwrite = TRUE, ...)
}
\arguments{
\item{x}{A data.frame or a (named) list of objects that can be handled by
\code{\link[=writeData]{writeData()}} or \code{\link[=writeDataTable]{writeDataTable()}} to write to file}

\item{file}{A file path to save the xlsx file}

\item{asTable}{If \code{TRUE} will use \code{\link[=writeDataTable]{writeDataTable()}} rather
than \code{\link[=writeData]{writeData()}} to write \code{x} to the file (default:
\code{FALSE})}

\item{overwrite}{Overwrite existing file (Defaults to \code{TRUE} as with \code{write.table})}

\item{...}{Additional arguments passed to \code{\link[=buildWorkbook]{buildWorkbook()}}; see details}
}
\value{
A workbook object
}
\description{
write a data.frame or list of data.frames to an xlsx file
}
\section{Optional Parameters}{


\strong{createWorkbook Parameters}
\describe{
\item{\strong{creator}}{ A string specifying the workbook author}
}

\strong{addWorksheet Parameters}
\describe{
\item{\strong{sheetName}}{ Name of the worksheet}
\item{\strong{gridLines}}{ A logical. If \code{FALSE}, the worksheet grid lines will be hidden.}
\item{\strong{tabColour}}{ Colour of the worksheet tab. A valid colour (belonging to colours())
or a valid hex colour beginning with "#".}
\item{\strong{zoom}}{ A numeric between 10 and 400. Worksheet zoom level as a percentage.}
}

\strong{writeData/writeDataTable Parameters}
\describe{
\item{\strong{startCol}}{ A vector specifying the starting column(s) to write df}
\item{\strong{startRow}}{ A vector specifying the starting row(s) to write df}
\item{\strong{xy}}{ An alternative to specifying startCol and startRow individually.
A vector of the form c(startCol, startRow)}
\item{\strong{colNames or col.names}}{ If \code{TRUE}, column names of x are written.}
\item{\strong{rowNames or row.names}}{ If \code{TRUE}, row names of x are written.}
\item{\strong{headerStyle}}{ Custom style to apply to column names.}
\item{\strong{borders}}{ Either "surrounding", "columns" or "rows" or NULL.  If "surrounding", a border is drawn around the
data.  If "rows", a surrounding border is drawn a border around each row. If "columns", a surrounding border is drawn with a border
between each column.  If "\code{all}" all cell borders are drawn.}
\item{\strong{borderColour}}{ Colour of cell border}
\item{\strong{borderStyle}}{ Border line style.}
\item{\strong{keepNA}}{If \code{TRUE}, NA values are converted to #N/A (or \code{na.string}, if not NULL) in Excel, else NA cells will be empty. Defaults to FALSE.}
\item{\strong{na.string}}{If not NULL, and if \code{keepNA} is \code{TRUE}, NA values are converted to this string in Excel. Defaults to NULL.}
}

\strong{freezePane Parameters}
\describe{
\item{\strong{firstActiveRow}}{Top row of active region to freeze pane.}
\item{\strong{firstActiveCol}}{Furthest left column of active region to freeze pane.}
\item{\strong{firstRow}}{If \code{TRUE}, freezes the first row (equivalent to firstActiveRow = 2)}
\item{\strong{firstCol}}{If \code{TRUE}, freezes the first column (equivalent to firstActiveCol = 2)}
}

\strong{colWidths Parameters}
\describe{
\item{\strong{colWidths}}{May be a single value for all columns (or "auto"), or a list of vectors that will be recycled for each sheet (see examples)}
}
}

\examples{

## write to working directory
options("openxlsx.borderColour" = "#4F80BD") ## set default border colour
\dontrun{
write.xlsx(iris, file = "writeXLSX1.xlsx", colNames = TRUE, borders = "columns")
write.xlsx(iris, file = "writeXLSX2.xlsx", colNames = TRUE, borders = "surrounding")
}


hs <- createStyle(
  textDecoration = "BOLD", fontColour = "#FFFFFF", fontSize = 12,
  fontName = "Arial Narrow", fgFill = "#4F80BD"
)
\dontrun{
write.xlsx(iris,
  file = "writeXLSX3.xlsx",
  colNames = TRUE, borders = "rows", headerStyle = hs
)
}

## Lists elements are written to individual worksheets, using list names as sheet names if available
l <- list("IRIS" = iris, "MTCATS" = mtcars, matrix(runif(1000), ncol = 5))
\dontrun{
write.xlsx(l, "writeList1.xlsx", colWidths = c(NA, "auto", "auto"))
}

## different sheets can be given different parameters
\dontrun{
write.xlsx(l, "writeList2.xlsx",
  startCol = c(1, 2, 3), startRow = 2,
  asTable = c(TRUE, TRUE, FALSE), withFilter = c(TRUE, FALSE, FALSE)
)
}

# specify column widths for multiple sheets
\dontrun{
write.xlsx(l, "writeList2.xlsx", colWidths = 20)
write.xlsx(l, "writeList2.xlsx", colWidths = list(100, 200, 300))
write.xlsx(l, "writeList2.xlsx", colWidths = list(rep(10, 5), rep(8, 11), rep(5, 5)))
}

}
\seealso{
\code{\link[=addWorksheet]{addWorksheet()}}

\code{\link[=writeData]{writeData()}}

\code{\link[=createStyle]{createStyle()}} for style parameters

\code{\link[=buildWorkbook]{buildWorkbook()}}
}
\author{
Alexander Walker, Jordan Mark Barbone
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{getNamedRegions}
\alias{getNamedRegions}
\title{Get named regions}
\usage{
getNamedRegions(x)
}
\arguments{
\item{x}{An xlsx file or Workbook object}
}
\description{
Return a vector of named regions in a xlsx file or
Workbook object
}
\examples{
## create named regions
wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")

## specify region
writeData(wb, sheet = 1, x = iris, startCol = 1, startRow = 1)
createNamedRegion(
  wb = wb,
  sheet = 1,
  name = "iris",
  rows = 1:(nrow(iris) + 1),
  cols = 1:ncol(iris)
)


## using writeData 'name' argument to create a named region
writeData(wb, sheet = 1, x = iris, name = "iris2", startCol = 10)
\dontrun{
out_file <- tempfile(fileext = ".xlsx")
saveWorkbook(wb, out_file, overwrite = TRUE)

## see named regions
getNamedRegions(wb) ## From Workbook object
getNamedRegions(out_file) ## From xlsx file

## read named regions
df <- read.xlsx(wb, namedRegion = "iris")
head(df)

df <- read.xlsx(out_file, namedRegion = "iris2")
head(df)
}

}
\seealso{
\code{\link[=createNamedRegion]{createNamedRegion()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{sheets}
\alias{sheets}
\title{Returns names of worksheets.}
\usage{
sheets(wb)
}
\arguments{
\item{wb}{A workbook object}
}
\value{
Name of worksheet(s) for a given index
}
\description{
DEPRECATED. Use names().
}
\details{
DEPRECATED. Use \code{\link[=names]{names()}}
}
\examples{

## Create a new workbook
wb <- createWorkbook()

## Add some worksheets
addWorksheet(wb, "Worksheet Name")
addWorksheet(wb, "This is worksheet 2")
addWorksheet(wb, "The third worksheet")

## Return names of sheets, can not be used for assignment.
names(wb)
# openXL(wb)

names(wb) <- c("A", "B", "C")
names(wb)
# openXL(wb)
}
\seealso{
\code{\link[=names]{names()}} to rename a worksheet in a Workbook
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CommentClass.R
\name{removeComment}
\alias{removeComment}
\title{Remove a comment from a cell}
\usage{
removeComment(wb, sheet, cols, rows, gridExpand = TRUE)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A vector of names or indices of worksheets}

\item{cols}{Columns to delete comments from}

\item{rows}{Rows to delete comments from}

\item{gridExpand}{If \code{TRUE}, all data in rectangle min(rows):max(rows) X min(cols):max(cols)
will be removed.}
}
\description{
Remove a cell comment from a worksheet
}
\seealso{
\code{\link[=createComment]{createComment()}}

\code{\link[=writeComment]{writeComment()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{getDateOrigin}
\alias{getDateOrigin}
\title{Get the date origin an xlsx file is using}
\usage{
getDateOrigin(xlsxFile)
}
\arguments{
\item{xlsxFile}{An xlsx or xlsm file.}
}
\value{
One of "1900-01-01" or "1904-01-01".
}
\description{
Return the date origin used internally by an xlsx or xlsm file
}
\details{
Excel stores dates as the number of days from either 1904-01-01 or 1900-01-01. This function
checks the date origin being used in an Excel file and returns is so it can be used in \code{\link[=convertToDate]{convertToDate()}}
}
\examples{

## create a file with some dates
\dontrun{
write.xlsx(as.Date("2015-01-10") - (0:4), file = "getDateOriginExample.xlsx")
m <- read.xlsx("getDateOriginExample.xlsx")

## convert to dates
do <- getDateOrigin(system.file("extdata", "readTest.xlsx", package = "openxlsx"))
convertToDate(m[[1]], do)
}
}
\seealso{
\code{\link[=convertToDate]{convertToDate()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{cloneWorksheet}
\alias{cloneWorksheet}
\title{Clone a worksheet to a workbook}
\usage{
cloneWorksheet(wb, sheetName, clonedSheet)
}
\arguments{
\item{wb}{A Workbook object to attach the new worksheet}

\item{sheetName}{A name for the new worksheet}

\item{clonedSheet}{The name of the existing worksheet to be cloned.}
}
\value{
XML tree
}
\description{
Clone a worksheet to a Workbook object
}
\examples{
## Create a new workbook
wb <- createWorkbook("Fred")

## Add 3 worksheets
addWorksheet(wb, "Sheet 1")
cloneWorksheet(wb, "Sheet 2", clonedSheet = "Sheet 1")

## Save workbook
\dontrun{
saveWorkbook(wb, "cloneWorksheetExample.xlsx", overwrite = TRUE)
}
}
\author{
Reinhold Kainhofer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conditional_formatting.R
\name{conditionalFormatting}
\alias{conditionalFormatting}
\alias{databar}
\title{Add conditional formatting to cells}
\usage{
conditionalFormatting(
  wb,
  sheet,
  cols,
  rows,
  rule = NULL,
  style = NULL,
  type = "expression",
  ...
)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{cols}{Columns to apply conditional formatting to}

\item{rows}{Rows to apply conditional formatting to}

\item{rule}{The condition under which to apply the formatting. See examples.}

\item{style}{A style to apply to those cells that satisfy the rule. Default is createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")}

\item{type}{Either 'expression', 'colourScale', 'databar', 'duplicates', 'beginsWith',
'endsWith', 'topN', 'bottomN', 'blanks', 'notBlanks', 'contains' or 'notContains' (case insensitive).}

\item{...}{See below}
}
\description{
Add conditional formatting to cells
}
\details{
See Examples.

If type == "expression"
\itemize{
\item{style is a Style object. See \code{\link[=createStyle]{createStyle()}}}
\item{rule is an expression. Valid operators are "<", "<=", ">", ">=", "==", "!=".}
}

If type == "colourScale"
\itemize{
\item{style is a vector of colours with length 2 or 3}
\item{rule can be NULL or a vector of colours of equal length to styles}
}

If type == "databar"
\itemize{
\item{style is a vector of colours with length 2 or 3}
\item{rule is a numeric vector specifying the range of the databar colours. Must be equal length to style}
\item{...
\itemize{
\item{\strong{showvalue} If FALSE the cell value is hidden. Default TRUE.}
\item{\strong{gradient} If FALSE colour gradient is removed. Default TRUE.}
\item{\strong{border} If FALSE the border around the database is hidden. Default TRUE.}
}
}
}

If type == "duplicates"
\itemize{
\item{style is a Style object. See \code{\link[=createStyle]{createStyle()}}}
\item{rule is ignored.}
}

If type == "contains"
\itemize{
\item{style is a Style object. See \code{\link[=createStyle]{createStyle()}}}
\item{rule is the text to look for within cells}
}

If type == "between"
\itemize{
\item{style is a Style object. See \code{\link[=createStyle]{createStyle()}}}
\item{rule is a numeric vector of length 2 specifying lower and upper bound (Inclusive)}
}
If type == "blanks"
\itemize{
\item{style is a Style object. See \code{\link[=createStyle]{createStyle()}}}
\item{rule is ignored.}
}

If type == "notBlanks"
\itemize{
\item{style is a Style object. See \code{\link[=createStyle]{createStyle()}}}
\item{rule is ignored.}
}

If type == "topN"
\itemize{
\item{style is a Style object. See \code{\link[=createStyle]{createStyle()}}}
\item{rule is ignored}
\item{...
\itemize{
\item{\strong{rank} numeric vector of length 1 indicating number of highest values.}
\item{\strong{percent} TRUE if you want top N percentage.}
}
}
}

If type == "bottomN"
\itemize{
\item{style is a Style object. See \code{\link[=createStyle]{createStyle()}}}
\item{rule is ignored}
\item{...
\itemize{
\item{\strong{rank} numeric vector of length 1 indicating number of lowest values.}
\item{\strong{percent} TRUE if you want bottom N percentage.}
}
}
}
}
\examples{
wb <- createWorkbook()
addWorksheet(wb, "cellIs")
addWorksheet(wb, "Moving Row")
addWorksheet(wb, "Moving Col")
addWorksheet(wb, "Dependent on")
addWorksheet(wb, "Duplicates")
addWorksheet(wb, "containsText")
addWorksheet(wb, "notcontainsText")
addWorksheet(wb, "beginsWith")
addWorksheet(wb, "endsWith")
addWorksheet(wb, "colourScale", zoom = 30)
addWorksheet(wb, "databar")
addWorksheet(wb, "between")
addWorksheet(wb, "topN")
addWorksheet(wb, "bottomN")
addWorksheet(wb, "containsBlanks")
addWorksheet(wb, "notContainsBlanks")
addWorksheet(wb, "logical operators")

negStyle <- createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
posStyle <- createStyle(fontColour = "#006100", bgFill = "#C6EFCE")

## rule applies to all each cell in range
writeData(wb, "cellIs", -5:5)
writeData(wb, "cellIs", LETTERS[1:11], startCol = 2)
conditionalFormatting(wb, "cellIs",
  cols = 1,
  rows = 1:11, rule = "!=0", style = negStyle
)
conditionalFormatting(wb, "cellIs",
  cols = 1,
  rows = 1:11, rule = "==0", style = posStyle
)

## highlight row dependent on first cell in row
writeData(wb, "Moving Row", -5:5)
writeData(wb, "Moving Row", LETTERS[1:11], startCol = 2)
conditionalFormatting(wb, "Moving Row",
  cols = 1:2,
  rows = 1:11, rule = "$A1<0", style = negStyle
)
conditionalFormatting(wb, "Moving Row",
  cols = 1:2,
  rows = 1:11, rule = "$A1>0", style = posStyle
)

## highlight column dependent on first cell in column
writeData(wb, "Moving Col", -5:5)
writeData(wb, "Moving Col", LETTERS[1:11], startCol = 2)
conditionalFormatting(wb, "Moving Col",
  cols = 1:2,
  rows = 1:11, rule = "A$1<0", style = negStyle
)
conditionalFormatting(wb, "Moving Col",
  cols = 1:2,
  rows = 1:11, rule = "A$1>0", style = posStyle
)

## highlight entire range cols X rows dependent only on cell A1
writeData(wb, "Dependent on", -5:5)
writeData(wb, "Dependent on", LETTERS[1:11], startCol = 2)
conditionalFormatting(wb, "Dependent on",
  cols = 1:2,
  rows = 1:11, rule = "$A$1<0", style = negStyle
)
conditionalFormatting(wb, "Dependent on",
  cols = 1:2,
  rows = 1:11, rule = "$A$1>0", style = posStyle
)

## highlight cells in column 1 based on value in column 2
writeData(wb, "Dependent on", data.frame(x = 1:10, y = runif(10)), startRow = 15)
conditionalFormatting(wb, "Dependent on",
  cols = 1,
  rows = 16:25, rule = "B16<0.5", style = negStyle
)
conditionalFormatting(wb, "Dependent on",
  cols = 1,
  rows = 16:25, rule = "B16>=0.5", style = posStyle
)


## highlight duplicates using default style
writeData(wb, "Duplicates", sample(LETTERS[1:15], size = 10, replace = TRUE))
conditionalFormatting(wb, "Duplicates", cols = 1, rows = 1:10, type = "duplicates")

## cells containing text
fn <- function(x) paste(sample(LETTERS, 10), collapse = "-")
writeData(wb, "containsText", sapply(1:10, fn))
conditionalFormatting(wb, "containsText", cols = 1, rows = 1:10, type = "contains", rule = "A")

## cells not containing text
fn <- function(x) paste(sample(LETTERS, 10), collapse = "-")
writeData(wb, "containsText", sapply(1:10, fn))
conditionalFormatting(wb, "notcontainsText", cols = 1, 
                     rows = 1:10, type = "notcontains", rule = "A")


## cells begins with text
fn <- function(x) paste(sample(LETTERS, 10), collapse = "-")
writeData(wb, "beginsWith", sapply(1:100, fn))
conditionalFormatting(wb, "beginsWith", cols = 1, rows = 1:100, type = "beginsWith", rule = "A")


## cells ends with text
fn <- function(x) paste(sample(LETTERS, 10), collapse = "-")
writeData(wb, "endsWith", sapply(1:100, fn))
conditionalFormatting(wb, "endsWith", cols = 1, rows = 1:100, type = "endsWith", rule = "A")

## colourscale colours cells based on cell value
df <- read.xlsx(system.file("extdata", "readTest.xlsx", package = "openxlsx"), sheet = 4)
writeData(wb, "colourScale", df, colNames = FALSE) ## write data.frame

## rule is a vector or colours of length 2 or 3 (any hex colour or any of colours())
## If rule is NULL, min and max of cells is used. Rule must be the same length as style or NULL.
conditionalFormatting(wb, "colourScale",
  cols = 1:ncol(df), rows = 1:nrow(df),
  style = c("black", "white"),
  rule = c(0, 255),
  type = "colourScale"
)

setColWidths(wb, "colourScale", cols = 1:ncol(df), widths = 1.07)
setRowHeights(wb, "colourScale", rows = 1:nrow(df), heights = 7.5)

## Databars
writeData(wb, "databar", -5:5)
conditionalFormatting(wb, "databar", cols = 1, rows = 1:11, type = "databar") ## Default colours

## Between
# Highlight cells in interval [-2, 2]
writeData(wb, "between", -5:5)
conditionalFormatting(wb, "between", cols = 1, rows = 1:11, type = "between", rule = c(-2, 2))

## Top N 
writeData(wb, "topN", data.frame(x = 1:10, y = rnorm(10)))
# Highlight top 5 values in column x
conditionalFormatting(wb, "topN", cols = 1, rows = 2:11, 
 style = posStyle, type = "topN", rank = 5)#'
# Highlight top 20 percentage in column y
conditionalFormatting(wb, "topN", cols = 2, rows = 2:11, 
 style = posStyle, type = "topN", rank = 20, percent = TRUE)

## Bottom N 
writeData(wb, "bottomN", data.frame(x = 1:10, y = rnorm(10)))
# Highlight bottom 5 values in column x
conditionalFormatting(wb, "bottomN", cols = 1, rows = 2:11, 
 style = negStyle, type = "topN", rank = 5)
# Highlight bottom 20 percentage in column y
conditionalFormatting(wb, "bottomN", cols = 2, rows = 2:11, 
 style = negStyle, type = "topN", rank = 20, percent = TRUE)
 
## cells containing blanks
sample_data <- sample(c("X", NA_character_), 10, replace = TRUE)
writeData(wb, "containsBlanks", sample_data)
conditionalFormatting(wb, "containsBlanks", cols = 1, rows = 1:10, 
type = "blanks", style = negStyle)

## cells not containing blanks
sample_data <- sample(c("X", NA_character_), 10, replace = TRUE)
writeData(wb, "notContainsBlanks", sample_data)
conditionalFormatting(wb, "notContainsBlanks", cols = 1, rows = 1:10, 
type = "notBlanks", style = posStyle)

## Logical Operators
# You can use Excels logical Operators
writeData(wb, "logical operators", 1:10)
conditionalFormatting(wb, "logical operators",
  cols = 1, rows = 1:10,
  rule = "OR($A1=1,$A1=3,$A1=5,$A1=7)"
)
\dontrun{
saveWorkbook(wb, "conditionalFormattingExample.xlsx", TRUE)
}


#########################################################################
## Databar Example

wb <- createWorkbook()
addWorksheet(wb, "databar")

## Databars
writeData(wb, "databar", -5:5, startCol = 1)
conditionalFormatting(wb, "databar", cols = 1, rows = 1:11, type = "databar") ## Defaults

writeData(wb, "databar", -5:5, startCol = 3)
conditionalFormatting(wb, "databar", cols = 3, rows = 1:11, type = "databar", border = FALSE)

writeData(wb, "databar", -5:5, startCol = 5)
conditionalFormatting(wb, "databar",
  cols = 5, rows = 1:11,
  type = "databar", style = c("#a6a6a6"), showValue = FALSE
)

writeData(wb, "databar", -5:5, startCol = 7)
conditionalFormatting(wb, "databar",
  cols = 7, rows = 1:11,
  type = "databar", style = c("#a6a6a6"), showValue = FALSE, gradient = FALSE
)

writeData(wb, "databar", -5:5, startCol = 9)
conditionalFormatting(wb, "databar",
  cols = 9, rows = 1:11,
  type = "databar", style = c("#a6a6a6", "#a6a6a6"), showValue = FALSE, gradient = FALSE
)
\dontrun{
saveWorkbook(wb, file = "databarExample.xlsx", overwrite = TRUE)
}

}
\seealso{
\code{\link[=createStyle]{createStyle()}}
}
\author{
Alexander Walker, Philipp Schauberger
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{setLastModifiedBy}
\alias{setLastModifiedBy}
\title{Set the author who modified the file last.}
\usage{
setLastModifiedBy(wb, LastModifiedBy)
}
\arguments{
\item{wb}{A workbook object}

\item{LastModifiedBy}{A string object with the name of the LastModifiedBy-User}
}
\description{
Just a wrapper of wb$changeLastModifiedBy()
}
\examples{

wb <- createWorkbook()
setLastModifiedBy(wb, "test")
}
\author{
Philipp Schauberger
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{removeFilter}
\alias{removeFilter}
\title{Remove a worksheet filter}
\usage{
removeFilter(wb, sheet)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A vector of names or indices of worksheets}
}
\description{
Removes filters from addFilter() and writeData()
}
\examples{
wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")
addWorksheet(wb, "Sheet 2")
addWorksheet(wb, "Sheet 3")

writeData(wb, 1, iris)
addFilter(wb, 1, row = 1, cols = 1:ncol(iris))

## Equivalently
writeData(wb, 2, x = iris, withFilter = TRUE)

## Similarly
writeDataTable(wb, 3, iris)

## remove filters
removeFilter(wb, 1:2) ## remove filters
removeFilter(wb, 3) ## Does not affect tables!
\dontrun{
saveWorkbook(wb, file = "removeFilterExample.xlsx", overwrite = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{addFilter}
\alias{addFilter}
\title{Add column filters}
\usage{
addFilter(wb, sheet, rows, cols)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{rows}{A row number.}

\item{cols}{columns to add filter to.}
}
\description{
Add excel column filters to a worksheet
}
\details{
adds filters to worksheet columns, same as filter parameters in writeData.
writeDataTable automatically adds filters to first row of a table.
NOTE Can only have a single filter per worksheet unless using tables.
}
\examples{
wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")
addWorksheet(wb, "Sheet 2")
addWorksheet(wb, "Sheet 3")

writeData(wb, 1, iris)
addFilter(wb, 1, row = 1, cols = 1:ncol(iris))

## Equivalently
writeData(wb, 2, x = iris, withFilter = TRUE)

## Similarly
writeDataTable(wb, 3, iris)
\dontrun{
saveWorkbook(wb, file = "addFilterExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=writeData]{writeData()}}

\code{\link[=addFilter]{addFilter()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{ungroupColumns}
\alias{ungroupColumns}
\title{Ungroup Columns}
\usage{
ungroupColumns(wb, sheet, cols)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{cols}{Indices of columns to ungroup}
}
\description{
Ungroup a selection of columns
}
\details{
If column was previously hidden, it will now be shown
}
\seealso{
\code{\link[=ungroupRows]{ungroupRows()}} To ungroup rows
}
\author{
Joshua Sturm
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{sheetVisibility}
\alias{sheetVisibility}
\alias{sheetVisibility<-}
\title{Get/set worksheet visible state}
\usage{
sheetVisibility(wb)

sheetVisibility(wb) <- value
}
\arguments{
\item{wb}{A workbook object}

\item{value}{a logical/character vector the same length as sheetVisibility(wb)}
}
\value{
Character vector of worksheet names.

Vector of "hidden", "visible", "veryHidden"
}
\description{
Get and set worksheet visible state
}
\examples{

wb <- createWorkbook()
addWorksheet(wb, sheetName = "S1", visible = FALSE)
addWorksheet(wb, sheetName = "S2", visible = TRUE)
addWorksheet(wb, sheetName = "S3", visible = FALSE)

sheetVisibility(wb)
sheetVisibility(wb)[1] <- TRUE ## show sheet 1
sheetVisibility(wb)[2] <- FALSE ## hide sheet 2
sheetVisibility(wb)[3] <- "hidden" ## hide sheet 3
sheetVisibility(wb)[3] <- "veryHidden" ## hide sheet 3 from UI
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{setHeader}
\alias{setHeader}
\title{Set header for all worksheets}
\usage{
setHeader(wb, text, position = "center")
}
\arguments{
\item{wb}{A workbook object}

\item{text}{header text. A character vector of length 1.}

\item{position}{Position of text in header. One of "left", "center" or "right"}
}
\description{
DEPRECATED
}
\examples{
\dontrun{
wb <- createWorkbook("Edgar Anderson")
addWorksheet(wb, "S1")
writeDataTable(wb, "S1", x = iris[1:30, ], xy = c("C", 5))

## set all headers
setHeader(wb, "This is a header", position = "center")
setHeader(wb, "To the left", position = "left")
setHeader(wb, "On the right", position = "right")

## set all footers
setFooter(wb, "Center Footer Here", position = "center")
setFooter(wb, "Bottom left", position = "left")
setFooter(wb, Sys.Date(), position = "right")

saveWorkbook(wb, "headerHeaderExample.xlsx", overwrite = TRUE)
}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{createNamedRegion}
\alias{createNamedRegion}
\alias{deleteNamedRegion}
\title{Create / delete a named region.}
\usage{
createNamedRegion(wb, sheet, cols, rows, name, overwrite = FALSE)

deleteNamedRegion(wb, name)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{cols}{Numeric vector specifying columns to include in region}

\item{rows}{Numeric vector specifying rows to include in region}

\item{name}{Name for region. A character vector of length 1. Note region names must be case-insensitive unique.}

\item{overwrite}{Boolean. Overwrite if exists ? Default to FALSE}
}
\description{
Create / delete a named region
}
\details{
Region is given by: min(cols):max(cols) X min(rows):max(rows)
}
\examples{
## create named regions
wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")

## specify region
writeData(wb, sheet = 1, x = iris, startCol = 1, startRow = 1)
createNamedRegion(
  wb = wb,
  sheet = 1,
  name = "iris",
  rows = 1:(nrow(iris) + 1),
  cols = 1:ncol(iris)
)


## using writeData 'name' argument
writeData(wb, sheet = 1, x = iris, name = "iris2", startCol = 10)

out_file <- tempfile(fileext = ".xlsx")
\dontrun{
saveWorkbook(wb, out_file, overwrite = TRUE)

## see named regions
getNamedRegions(wb) ## From Workbook object
getNamedRegions(out_file) ## From xlsx file

## delete one
deleteNamedRegion(wb = wb, name = "iris2")
getNamedRegions(wb)

## read named regions
df <- read.xlsx(wb, namedRegion = "iris")
head(df)

df <- read.xlsx(out_file, namedRegion = "iris2")
head(df)
}

}
\seealso{
\code{\link[=getNamedRegions]{getNamedRegions()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/openXL.R
\name{openXL}
\alias{openXL}
\title{Open a Microsoft Excel file (xls/xlsx) or an openxlsx Workbook}
\usage{
openXL(file=NULL)
}
\arguments{
\item{file}{path to the Excel (xls/xlsx) file or Workbook object.}
}
\description{
This function tries to open a Microsoft Excel
(xls/xlsx) file or an openxlsx Workbook with the proper
application, in a portable manner.

In Windows (c) and Mac (c), it uses system default handlers,
given the file type.

In Linux it searches (via \code{which}) for available xls/xlsx
reader applications (unless \code{options('openxlsx.excelApp')}
is set to the app bin path), and if it finds anything, sets
\code{options('openxlsx.excelApp')} to the program choosen by
the user via a menu (if many are present, otherwise it will
set the only available). Currently searched for apps are
Libreoffice/Openoffice (\code{soffice} bin), Gnumeric
(\code{gnumeric}) and Calligra Sheets (\code{calligrasheets}).
}
\examples{
# file example
example(writeData)
# openXL("writeDataExample.xlsx")

# (not yet saved) Workbook example
wb <- createWorkbook()
x <- mtcars[1:6, ]
addWorksheet(wb, "Cars")
writeData(wb, "Cars", x, startCol = 2, startRow = 3, rowNames = TRUE)
# openXL(wb)
}
\author{
Luca Braglia
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/openxlsx.R
\docType{data}
\name{openxlsx_options}
\alias{openxlsx_options}
\alias{op.openxlsx}
\alias{openxlsx_getOp}
\alias{openxlsx_setOp}
\title{openxlsx Options}
\format{
An object of class \code{list} of length 34.
}
\usage{
op.openxlsx

openxlsx_getOp(x, default = NULL)

openxlsx_setOp(x, value)
}
\arguments{
\item{x}{An option name (\code{"openxlsx."} prefix optional)}

\item{default}{A default value if \code{NULL}}

\item{value}{The new value for the option (optional if x is a named list)}
}
\description{
See and get the openxlsx options
}
\details{
\code{openxlsx_getOp()} retrieves the \code{"openxlsx"} options found in
\code{op.openxlsx}.  If none are set (currently \code{NULL}) retrieves the
default option from \code{op.openxlsx}.  This will also check that the
intended option is a standard option (listed in \code{op.openxlsx}) and
will provide a warning otherwise.

\code{openxlsx_setOp()} is a safer way to set an option as it will first
check that the option is a standard option (as above) before setting.
}
\examples{
openxlsx_getOp("borders")
op.openxlsx[["openxlsx.borders"]]

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readWorkbook.R
\name{read.xlsx}
\alias{read.xlsx}
\title{Read from an Excel file or Workbook object}
\usage{
read.xlsx(
  xlsxFile,
  sheet,
  startRow = 1,
  colNames = TRUE,
  rowNames = FALSE,
  detectDates = FALSE,
  skipEmptyRows = TRUE,
  skipEmptyCols = TRUE,
  rows = NULL,
  cols = NULL,
  check.names = FALSE,
  sep.names = ".",
  namedRegion = NULL,
  na.strings = "NA",
  fillMergedCells = FALSE
)
}
\arguments{
\item{xlsxFile}{An xlsx file, Workbook object or URL to xlsx file.}

\item{sheet}{The name or index of the sheet to read data from.}

\item{startRow}{first row to begin looking for data.  Empty rows at the top of a file are always skipped,
regardless of the value of startRow.}

\item{colNames}{If \code{TRUE}, the first row of data will be used as column names.}

\item{rowNames}{If \code{TRUE}, first column of data will be used as row names.}

\item{detectDates}{If \code{TRUE}, attempt to recognise dates and perform conversion.}

\item{skipEmptyRows}{If \code{TRUE}, empty rows are skipped else empty rows after the first row containing data
will return a row of NAs.}

\item{skipEmptyCols}{If \code{TRUE}, empty columns are skipped.}

\item{rows}{A numeric vector specifying which rows in the Excel file to read.
If NULL, all rows are read.}

\item{cols}{A numeric vector specifying which columns in the Excel file to read.
If NULL, all columns are read.}

\item{check.names}{logical. If TRUE then the names of the variables in the data frame
are checked to ensure that they are syntactically valid variable names}

\item{sep.names}{One character which substitutes blanks in column names. By default, "."}

\item{namedRegion}{A named region in the Workbook. If not NULL startRow, rows and cols parameters are ignored.}

\item{na.strings}{A character vector of strings which are to be interpreted as NA. Blank cells will be returned as NA.}

\item{fillMergedCells}{If TRUE, the value in a merged cell is given to all cells within the merge.}
}
\value{
data.frame
}
\description{
Read data from an Excel file or Workbook object into a data.frame
}
\details{
Formulae written using writeFormula to a Workbook object will not get picked up by read.xlsx().
This is because only the formula is written and left to be evaluated when the file is opened in Excel.
Opening, saving and closing the file with Excel will resolve this.
}
\examples{

xlsxFile <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
df1 <- read.xlsx(xlsxFile = xlsxFile, sheet = 1, skipEmptyRows = FALSE)
sapply(df1, class)

df2 <- read.xlsx(xlsxFile = xlsxFile, sheet = 3, skipEmptyRows = TRUE)
df2$Date <- convertToDate(df2$Date)
sapply(df2, class)
head(df2)

df2 <- read.xlsx(
  xlsxFile = xlsxFile, sheet = 3, skipEmptyRows = TRUE,
  detectDates = TRUE
)
sapply(df2, class)
head(df2)

wb <- loadWorkbook(system.file("extdata", "readTest.xlsx", package = "openxlsx"))
df3 <- read.xlsx(wb, sheet = 2, skipEmptyRows = FALSE, colNames = TRUE)
df4 <- read.xlsx(xlsxFile, sheet = 2, skipEmptyRows = FALSE, colNames = TRUE)
all.equal(df3, df4)

wb <- loadWorkbook(system.file("extdata", "readTest.xlsx", package = "openxlsx"))
df3 <- read.xlsx(wb,
  sheet = 2, skipEmptyRows = FALSE,
  cols = c(1, 4), rows = c(1, 3, 4)
)

## URL
##
\dontrun{
xlsxFile <- "https://github.com/awalker89/openxlsx/raw/master/inst/readTest.xlsx"
head(read.xlsx(xlsxFile))
}

}
\seealso{
\code{\link[=getNamedRegions]{getNamedRegions()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{ungroupRows}
\alias{ungroupRows}
\title{Ungroup Rows}
\usage{
ungroupRows(wb, sheet, rows)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{rows}{Indices of rows to ungroup}
}
\description{
Ungroup a selection of rows
}
\details{
If row was previously hidden, it will now be shown
}
\seealso{
\code{\link[=ungroupColumns]{ungroupColumns()}}
}
\author{
Joshua Sturm
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helperFunctions.R
\name{get_worksheet_entries}
\alias{get_worksheet_entries}
\title{Get entries from workbook worksheet}
\usage{
get_worksheet_entries(wb, sheet)
}
\arguments{
\item{wb}{workbook}

\item{sheet}{worksheet}
}
\value{
vector of strings
}
\description{
Get all entries from workbook worksheet without xml tags
}
\examples{
## Create new workbook
wb <- createWorkbook()
addWorksheet(wb, "Sheet")
sheet <- 1

## Write dummy data
writeData(wb, sheet, c("A", "BB", "CCC"), startCol = 2, startRow = 3)
writeData(wb, sheet, c(4, 5), startCol = 4, startRow = 3)

## Get text entries
get_worksheet_entries(wb, sheet)

}
\author{
David Breuer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{createWorkbook}
\alias{createWorkbook}
\title{Create a new Workbook object}
\usage{
createWorkbook(
  creator = ifelse(.Platform$OS.type == "windows", Sys.getenv("USERNAME"),
    Sys.getenv("USER")),
  title = NULL,
  subject = NULL,
  category = NULL
)
}
\arguments{
\item{creator}{Creator of the workbook (your name). Defaults to login username}

\item{title}{Workbook properties title}

\item{subject}{Workbook properties subject}

\item{category}{Workbook properties category}
}
\value{
Workbook object
}
\description{
Create a new Workbook object
}
\examples{
## Create a new workbook
wb <- createWorkbook()

## Save workbook to working directory
\dontrun{
saveWorkbook(wb, file = "createWorkbookExample.xlsx", overwrite = TRUE)
}

## Set Workbook properties
wb <- createWorkbook(
  creator = "Me",
  title = "title here",
  subject = "this & that",
  category = "something"
)
}
\seealso{
\code{\link[=loadWorkbook]{loadWorkbook()}}

\code{\link[=saveWorkbook]{saveWorkbook()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{groupColumns}
\alias{groupColumns}
\title{Group columns}
\usage{
groupColumns(wb, sheet, cols, hidden = FALSE, level = -1)
}
\arguments{
\item{wb}{A workbook object.}

\item{sheet}{A name or index of a worksheet.}

\item{cols}{Indices of cols to group. Can be either a vector of indices to
group at the same level or a (named) list of numeric vectors of
indices to create multiple groupings at once. The names of the
entries determine the grouping level. If no names are given,
the \code{level} parameter is used as default.}

\item{hidden}{Logical vector. If TRUE the grouped columns are hidden. Defaults to FALSE.}

\item{level}{Grouping level (higher value indicates multiple nestings) for the
group. A vector to assign different grouping levels to the indices.
A value of -1 indicates that the grouping level should be derived
from the existing grouping (one level added)}
}
\description{
Group a selection of columns
}
\details{
Group columns together, with the option to hide them.

NOTE: \code{\link[=setColWidths]{setColWidths()}} has a conflicting \code{hidden} parameter; changing one will update the other.
}
\seealso{
\code{\link[=ungroupColumns]{ungroupColumns()}} to ungroup columns. \code{\link[=groupRows]{groupRows()}} for grouping rows.
}
\author{
Joshua Sturm, Reinhold Kainhofer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{as_POSIXct_utc}
\alias{as_POSIXct_utc}
\title{Convert to POSIXct with timezone UTC}
\usage{
as_POSIXct_utc(x)
}
\arguments{
\item{x}{something as.POSIXct can convert}
}
\description{
Convert to POSIXct with timezone UTC
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{insertImage}
\alias{insertImage}
\title{Insert an image into a worksheet}
\usage{
insertImage(
  wb,
  sheet,
  file,
  width = 6,
  height = 3,
  startRow = 1,
  startCol = 1,
  units = "in",
  dpi = 300,
  address
)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{file}{An image file. Valid file types are: jpeg, png, bmp}

\item{width}{Width of figure.}

\item{height}{Height of figure.}

\item{startRow}{Row coordinate of upper left corner of the image}

\item{startCol}{Column coordinate of upper left corner of the image}

\item{units}{Units of width and height. Can be "in", "cm" or "px"}

\item{dpi}{Image resolution used for conversion between units.}

\item{address}{An optional character string specifying an external URL, relative or absolute path to a file, or mailto string (e.g. "mailto:example@example.com") that will be opened when the image is clicked.}
}
\description{
Insert an image into a worksheet
}
\examples{
## Create a new workbook
wb <- createWorkbook("Ayanami")

## Add some worksheets
addWorksheet(wb, "Sheet 1")
addWorksheet(wb, "Sheet 2")
addWorksheet(wb, "Sheet 3")
addWorksheet(wb, "Sheet 4")

## Insert images
img <- system.file("extdata", "einstein.jpg", package = "openxlsx")
insertImage(wb, "Sheet 1", img, startRow = 5, startCol = 3, width = 6, height = 5)
insertImage(wb, 2, img, startRow = 2, startCol = 2)
insertImage(wb, 3, img, width = 15, height = 12, startRow = 3, startCol = "G", units = "cm")
insertImage(wb, 4, img, address = "https://github.com/ycphs/openxlsx")

## Save workbook
\dontrun{
saveWorkbook(wb, "insertImageExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=insertPlot]{insertPlot()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{freezePane}
\alias{freezePane}
\title{Freeze a worksheet pane}
\usage{
freezePane(
  wb,
  sheet,
  firstActiveRow = NULL,
  firstActiveCol = NULL,
  firstRow = FALSE,
  firstCol = FALSE
)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{firstActiveRow}{Top row of active region}

\item{firstActiveCol}{Furthest left column of active region}

\item{firstRow}{If \code{TRUE}, freezes the first row (equivalent to firstActiveRow = 2)}

\item{firstCol}{If \code{TRUE}, freezes the first column (equivalent to firstActiveCol = 2)}
}
\description{
Freeze a worksheet pane
}
\examples{
## Create a new workbook
wb <- createWorkbook("Kenshin")

## Add some worksheets
addWorksheet(wb, "Sheet 1")
addWorksheet(wb, "Sheet 2")
addWorksheet(wb, "Sheet 3")
addWorksheet(wb, "Sheet 4")

## Freeze Panes
freezePane(wb, "Sheet 1", firstActiveRow = 5, firstActiveCol = 3)
freezePane(wb, "Sheet 2", firstCol = TRUE) ## shortcut to firstActiveCol = 2
freezePane(wb, 3, firstRow = TRUE) ## shortcut to firstActiveRow = 2
freezePane(wb, 4, firstActiveRow = 1, firstActiveCol = "D")

## Save workbook
\dontrun{
saveWorkbook(wb, "freezePaneExample.xlsx", overwrite = TRUE)
}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{addStyle}
\alias{addStyle}
\title{Add a style to a set of cells}
\usage{
addStyle(wb, sheet, style, rows, cols, gridExpand = FALSE, stack = FALSE)
}
\arguments{
\item{wb}{A Workbook object containing a worksheet.}

\item{sheet}{A worksheet to apply the style to.}

\item{style}{A style object returned from createStyle()}

\item{rows}{Rows to apply style to.}

\item{cols}{columns to apply style to.}

\item{gridExpand}{If \code{TRUE}, style will be applied to all combinations of rows and cols.}

\item{stack}{If \code{TRUE} the new style is merged with any existing cell styles.  If FALSE, any
existing style is replaced by the new style.}
}
\description{
Function adds a style to a specified set of cells.
}
\examples{
## See package vignette for more examples.

## Create a new workbook
wb <- createWorkbook("My name here")

## Add a worksheets
addWorksheet(wb, "Expenditure", gridLines = FALSE)

## write data to worksheet 1
writeData(wb, sheet = 1, USPersonalExpenditure, rowNames = TRUE)

## create and add a style to the column headers
headerStyle <- createStyle(
  fontSize = 14, fontColour = "#FFFFFF", halign = "center",
  fgFill = "#4F81BD", border = "TopBottom", borderColour = "#4F81BD"
)

## style for body
bodyStyle <- createStyle(border = "TopBottom", borderColour = "#4F81BD")
addStyle(wb, sheet = 1, bodyStyle, rows = 2:6, cols = 1:6, gridExpand = TRUE)
setColWidths(wb, 1, cols = 1, widths = 21) ## set column width for row names column
\dontrun{
saveWorkbook(wb, "addStyleExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=createStyle]{createStyle()}}

expand.grid
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helperFunctions.R
\name{auto_heights}
\alias{auto_heights}
\title{Compute optimal row heights}
\usage{
auto_heights(
  wb,
  sheet,
  selected,
  fontsize = NULL,
  factor = 1,
  base_height = 15,
  extra_height = 12
)
}
\arguments{
\item{wb}{workbook}

\item{sheet}{worksheet}

\item{selected}{selected rows}

\item{fontsize}{font size, optional (get base font size by default)}

\item{factor}{factor to manually adjust font width, e.g., for bold fonts,
optional}

\item{base_height}{basic row height, optional}

\item{extra_height}{additional row height per new line of text, optional}
}
\value{
list of indices of columns with fixed widths and optimal row heights
}
\description{
Compute optimal row heights for cell with fixed with and
enabled automatic row heights parameter
}
\examples{
## Create new workbook
wb <- createWorkbook()
addWorksheet(wb, "Sheet")
sheet <- 1

## Write dummy data
long_string <- "ABC ABC ABC ABC ABC ABC ABC ABC ABC ABC ABC"
writeData(wb, sheet, c("A", long_string, "CCC"), startCol = 2, startRow = 3)
writeData(wb, sheet, c(4, 5), startCol = 4, startRow = 3)

## Set column widths and get optimal row heights
setColWidths(wb, sheet, c(1,2,3,4), c(10,20,10,20))
auto_heights(wb, sheet, 1:5)

}
\author{
David Breuer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{removeTable}
\alias{removeTable}
\title{Remove an Excel table in a workbook}
\usage{
removeTable(wb, sheet, table)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{table}{Name of table to remove. See \code{\link[=getTables]{getTables()}}}
}
\value{
character vector of table names on the specified sheet
}
\description{
List Excel tables in a workbook
}
\examples{

wb <- createWorkbook()
addWorksheet(wb, sheetName = "Sheet 1")
addWorksheet(wb, sheetName = "Sheet 2")
writeDataTable(wb, sheet = "Sheet 1", x = iris, tableName = "iris")
writeDataTable(wb, sheet = 1, x = mtcars, tableName = "mtcars", startCol = 10)


removeWorksheet(wb, sheet = 1) ## delete worksheet removes table objects

writeDataTable(wb, sheet = 1, x = iris, tableName = "iris")
writeDataTable(wb, sheet = 1, x = mtcars, tableName = "mtcars", startCol = 10)

## removeTable() deletes table object and all data
getTables(wb, sheet = 1)
removeTable(wb = wb, sheet = 1, table = "iris")
writeDataTable(wb, sheet = 1, x = iris, tableName = "iris", startCol = 1)

getTables(wb, sheet = 1)
removeTable(wb = wb, sheet = 1, table = "iris")
writeDataTable(wb, sheet = 1, x = iris, tableName = "iris", startCol = 1)
\dontrun{
saveWorkbook(wb = wb, file = "removeTableExample.xlsx", overwrite = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{insertPlot}
\alias{insertPlot}
\title{Insert the current plot into a worksheet}
\usage{
insertPlot(
  wb,
  sheet,
  width = 6,
  height = 4,
  xy = NULL,
  startRow = 1,
  startCol = 1,
  fileType = "png",
  units = "in",
  dpi = 300
)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{width}{Width of figure. Defaults to 6in.}

\item{height}{Height of figure . Defaults to 4in.}

\item{xy}{Alternate way to specify startRow and startCol.  A vector of length 2 of form (startcol, startRow)}

\item{startRow}{Row coordinate of upper left corner of figure.\code{ xy[[2]]} when xy is given.}

\item{startCol}{Column coordinate of upper left corner of figure. \code{xy[[1]]} when xy is given.}

\item{fileType}{File type of image}

\item{units}{Units of width and height. Can be "in", "cm" or "px"}

\item{dpi}{Image resolution}
}
\description{
The current plot is saved to a temporary image file using dev.copy.
This file is then written to the workbook using insertImage.
}
\examples{
\dontrun{
## Create a new workbook
wb <- createWorkbook()

## Add a worksheet
addWorksheet(wb, "Sheet 1", gridLines = FALSE)

## create plot objects
require(ggplot2)
p1 <- qplot(mpg,
  data = mtcars, geom = "density",
  fill = as.factor(gear), alpha = I(.5), main = "Distribution of Gas Mileage"
)
p2 <- qplot(age, circumference,
  data = Orange, geom = c("point", "line"), colour = Tree
)

## Insert currently displayed plot to sheet 1, row 1, column 1
print(p1) # plot needs to be showing
insertPlot(wb, 1, width = 5, height = 3.5, fileType = "png", units = "in")

## Insert plot 2
print(p2)
insertPlot(wb, 1, xy = c("J", 2), width = 16, height = 10, fileType = "png", units = "cm")

## Save workbook
saveWorkbook(wb, "insertPlotExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=insertImage]{insertImage()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{getBaseFont}
\alias{getBaseFont}
\title{Return the workbook default font}
\usage{
getBaseFont(wb)
}
\arguments{
\item{wb}{A workbook object}
}
\description{
Return the workbook default font

Returns the base font used in the workbook.
}
\examples{
## create a workbook
wb <- createWorkbook()
getBaseFont(wb)

## modify base font to size 10 Arial Narrow in red
modifyBaseFont(wb, fontSize = 10, fontColour = "#FF0000", fontName = "Arial Narrow")

getBaseFont(wb)
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CommentClass.R
\name{createComment}
\alias{createComment}
\title{create a Comment object}
\usage{
createComment(
  comment,
  author = Sys.getenv("USERNAME"),
  style = NULL,
  visible = TRUE,
  width = 2,
  height = 4
)
}
\arguments{
\item{comment}{Comment text. Character vector.}

\item{author}{Author of comment. Character vector of length 1}

\item{style}{A Style object or list of style objects the same length as comment vector. See \code{\link[=createStyle]{createStyle()}}.}

\item{visible}{TRUE or FALSE. Is comment visible.}

\item{width, height}{Width and height of textbook (in number of cells);
doubles are rounded with \code{base::round()}}
}
\description{
Create a cell Comment object to pass to writeComment()
}
\examples{
wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")

c1 <- createComment(comment = "this is comment")
writeComment(wb, 1, col = "B", row = 10, comment = c1)

s1 <- createStyle(fontSize = 12, fontColour = "red", textDecoration = c("BOLD"))
s2 <- createStyle(fontSize = 9, fontColour = "black")

c2 <- createComment(comment = c("This Part Bold red\n\n", "This part black"), style = c(s1, s2))
c2

writeComment(wb, 1, col = 6, row = 3, comment = c2)
\dontrun{
saveWorkbook(wb, file = "createCommentExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=writeComment]{writeComment()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/openxlsx.R
\docType{package}
\name{openxlsx}
\alias{openxlsx-package}
\alias{openxlsx}
\title{xlsx reading, writing and editing.}
\description{
openxlsx simplifies the the process of writing and styling Excel xlsx files from R
and removes the dependency on Java.
}
\details{
The openxlsx package uses global options, most to simplify formatting.  These
are stored in the \code{op.openxlsx} object.

\describe{
\item{openxlsx.bandedCols}{FALSE}
\item{openxlsx.bandedRows}{TRUE}
\item{openxlsx.borderColour}{"black"}
\item{openxlsx.borders}{"none"}
\item{openxlsx.borderStyle}{"thin"}
\item{openxlsx.compressionLevel}{"9"}
\item{openxlsx.creator}{""}
\item{openxlsx.dateFormat}{"mm/dd/yyyy"}
\item{openxlsx.datetimeFormat}{"yyyy-mm-dd hh:mm:ss"}
\item{openxlsx.headerStyle}{NULL}
\item{openxlsx.keepNA}{FALSE}
\item{openxlsx.na.string}{NULL}
\item{openxlsx.numFmt}{NULL}
\item{openxlsx.orientation}{"portrait"}
\item{openxlsx.paperSize}{9}
\item{openxlsx.tabColour}{"TableStyleLight9"}
\item{openxlsx.tableStyle}{"TableStyleLight9"}
\item{openxlsx.withFilter}{NA Whether to write data with or without a
filter. If NA will make filters with \code{writeDataTable} and will not for
\code{writeData}}
}

See the Formatting vignette for examples.

Additional options
}
\seealso{
\itemize{
\item{\code{vignette("Introduction", package = "openxlsx")}}
\item{\code{vignette("formatting", package = "openxlsx")}}
\item{\code{\link[=writeData]{writeData()}}}
\item{\code{\link[=writeDataTable]{writeDataTable()}}}
\item{\code{\link[=write.xlsx]{write.xlsx()}}}
\item{\code{\link[=read.xlsx]{read.xlsx()}}}
\item{\code{\link[=op.openxlsx]{op.openxlsx()}}}
}
for examples
}
\author{
\strong{Maintainer}: Jan Marvin Garbuszus \email{jan.garbuszus@ruhr-uni-bochum.de} [contributor]

Authors:
\itemize{
  \item Philipp Schauberger \email{philipp@schauberger.co.at}
  \item Alexander Walker \email{Alexander.Walker1989@gmail.com}
}

Other contributors:
\itemize{
  \item Luca Braglia [contributor]
  \item Joshua Sturm [contributor]
  \item Jordan Mark Barbone \email{jmbarbone@gmail.com} (\href{https://orcid.org/0000-0001-9788-3628}{ORCID}) [contributor]
  \item David Zimmermann \email{david_j_zimmermann@hotmail.com} [contributor]
  \item Reinhold Kainhofer \email{reinhold@kainhofer.com} [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{renameWorksheet}
\alias{renameWorksheet}
\title{Rename a worksheet}
\usage{
renameWorksheet(wb, sheet, newName)
}
\arguments{
\item{wb}{A Workbook object containing a worksheet}

\item{sheet}{The name or index of the worksheet to rename}

\item{newName}{The new name of the worksheet. No longer than 31 chars.}
}
\description{
Rename a worksheet
}
\details{
DEPRECATED. Use \code{\link[=names]{names()}}
}
\examples{

## Create a new workbook
wb <- createWorkbook("CREATOR")

## Add 3 worksheets
addWorksheet(wb, "Worksheet Name")
addWorksheet(wb, "This is worksheet 2")
addWorksheet(wb, "Not the best name")

#' ## rename all worksheets
names(wb) <- c("A", "B", "C")


## Rename worksheet 1 & 3
renameWorksheet(wb, 1, "New name for sheet 1")
names(wb)[[1]] <- "New name for sheet 1"
names(wb)[[3]] <- "A better name"

## Save workbook
\dontrun{
saveWorkbook(wb, "renameWorksheetExample.xlsx", overwrite = TRUE)
}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{modifyBaseFont}
\alias{modifyBaseFont}
\title{Modify the default font}
\usage{
modifyBaseFont(wb, fontSize = 11, fontColour = "black", fontName = "Calibri")
}
\arguments{
\item{wb}{A workbook object}

\item{fontSize}{font size}

\item{fontColour}{font colour}

\item{fontName}{Name of a font}
}
\description{
Modify the default font for this workbook
}
\details{
The font name is not validated in anyway.  Excel replaces unknown font names
with Arial. Base font is black, size 11, Calibri.
}
\examples{
## create a workbook
wb <- createWorkbook()
addWorksheet(wb, "S1")
## modify base font to size 10 Arial Narrow in red
modifyBaseFont(wb, fontSize = 10, fontColour = "#FF0000", fontName = "Arial Narrow")

writeData(wb, "S1", iris)
writeDataTable(wb, "S1", x = iris, startCol = 10) ## font colour does not affect tables
\dontrun{
saveWorkbook(wb, "modifyBaseFontExample.xlsx", overwrite = TRUE)
}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{convertFromExcelRef}
\alias{convertFromExcelRef}
\title{Convert excel column name to integer index}
\usage{
convertFromExcelRef(col)
}
\arguments{
\item{col}{An excel column reference}
}
\description{
Convert excel column name to integer index e.g. "J" to 10
}
\examples{
convertFromExcelRef("DOG")
convertFromExcelRef("COW")

## numbers will be removed
convertFromExcelRef("R22")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{getTables}
\alias{getTables}
\title{List Excel tables in a workbook}
\usage{
getTables(wb, sheet)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}
}
\value{
character vector of table names on the specified sheet
}
\description{
List Excel tables in a workbook
}
\examples{

wb <- createWorkbook()
addWorksheet(wb, sheetName = "Sheet 1")
writeDataTable(wb, sheet = "Sheet 1", x = iris)
writeDataTable(wb, sheet = 1, x = mtcars, tableName = "mtcars", startCol = 10)

getTables(wb, sheet = "Sheet 1")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{worksheetOrder}
\alias{worksheetOrder}
\alias{worksheetOrder<-}
\title{Order of worksheets in xlsx file}
\usage{
worksheetOrder(wb)

worksheetOrder(wb) <- value
}
\arguments{
\item{wb}{A workbook object}

\item{value}{Vector specifying order to write worksheets to file}
}
\description{
Get/set order of worksheets in a Workbook object
}
\details{
This function does not reorder the worksheets within the workbook object, it simply
shuffles the order when writing to file.
}
\examples{
## setup a workbook with 3 worksheets
wb <- createWorkbook()
addWorksheet(wb = wb, sheetName = "Sheet 1", gridLines = FALSE)
writeDataTable(wb = wb, sheet = 1, x = iris)

addWorksheet(wb = wb, sheetName = "mtcars (Sheet 2)", gridLines = FALSE)
writeData(wb = wb, sheet = 2, x = mtcars)

addWorksheet(wb = wb, sheetName = "Sheet 3", gridLines = FALSE)
writeData(wb = wb, sheet = 3, x = Formaldehyde)

worksheetOrder(wb)
names(wb)
worksheetOrder(wb) <- c(1, 3, 2) # switch position of sheets 2 & 3
writeData(wb, 2, 'This is still the "mtcars" worksheet', startCol = 15)
worksheetOrder(wb)
names(wb) ## ordering within workbook is not changed
\dontrun{
saveWorkbook(wb, "worksheetOrderExample.xlsx", overwrite = TRUE)
}
worksheetOrder(wb) <- c(3, 2, 1)
\dontrun{
saveWorkbook(wb, "worksheetOrderExample2.xlsx", overwrite = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_workbook.R
\name{buildWorkbook}
\alias{buildWorkbook}
\title{Build Workbook}
\usage{
buildWorkbook(x, asTable = FALSE, ...)
}
\arguments{
\item{x}{A data.frame or a (named) list of objects that can be handled by
\code{\link[=writeData]{writeData()}} or \code{\link[=writeDataTable]{writeDataTable()}} to write to file}

\item{asTable}{If \code{TRUE} will use \code{\link[=writeDataTable]{writeDataTable()}} rather
than \code{\link[=writeData]{writeData()}} to write \code{x} to the file (default:
\code{FALSE})}

\item{...}{Additional arguments passed to \code{\link[=writeData]{writeData()}},
\code{\link[=writeDataTable]{writeDataTable()}}, \code{\link[=setColWidths]{setColWidths()}} (see Optional
Parameters)}
}
\value{
A Workbook object
}
\description{
Build a workbook from a data.frame or named list
}
\details{
This function can be used as shortcut to create a workbook object from a
data.frame or named list.  If names are available in the list they will be
used as the worksheet names.  The parameters in \code{...} are collected
and passed to \code{\link[=writeData]{writeData()}} or \code{\link[=writeDataTable]{writeDataTable()}} to
initially create the Workbook objects then appropriate parameters are
passed to \code{\link[=setColWidths]{setColWidths()}}.

columns of x with class Date or POSIXt are automatically
styled as dates and datetimes respectively.
}
\section{Optional Parameters}{


\strong{createWorkbook Parameters}
\describe{
\item{\strong{creator}}{ A string specifying the workbook author}
}

\strong{addWorksheet Parameters}
\describe{
\item{\strong{sheetName}}{ Name of the worksheet}
\item{\strong{gridLines}}{ A logical. If \code{FALSE}, the worksheet grid lines will be hidden.}
\item{\strong{tabColour}}{ Colour of the worksheet tab. A valid colour (belonging to colours())
or a valid hex colour beginning with "#".}
\item{\strong{zoom}}{ A numeric between 10 and 400. Worksheet zoom level as a percentage.}
}

\strong{writeData/writeDataTable Parameters}
\describe{
\item{\strong{startCol}}{ A vector specifying the starting column(s) to write df}
\item{\strong{startRow}}{ A vector specifying the starting row(s) to write df}
\item{\strong{xy}}{ An alternative to specifying startCol and startRow individually.
A vector of the form c(startCol, startRow)}
\item{\strong{colNames or col.names}}{ If \code{TRUE}, column names of x are written.}
\item{\strong{rowNames or row.names}}{ If \code{TRUE}, row names of x are written.}
\item{\strong{headerStyle}}{ Custom style to apply to column names.}
\item{\strong{borders}}{ Either "surrounding", "columns" or "rows" or NULL.  If "surrounding", a border is drawn around the
data.  If "rows", a surrounding border is drawn a border around each row. If "columns", a surrounding border is drawn with a border
between each column.  If "\code{all}" all cell borders are drawn.}
\item{\strong{borderColour}}{ Colour of cell border}
\item{\strong{borderStyle}}{ Border line style.}
\item{\strong{keepNA}}{If \code{TRUE}, NA values are converted to #N/A (or \code{na.string}, if not NULL) in Excel, else NA cells will be empty. Defaults to FALSE.}
\item{\strong{na.string}}{If not NULL, and if \code{keepNA} is \code{TRUE}, NA values are converted to this string in Excel. Defaults to NULL.}
}

\strong{freezePane Parameters}
\describe{
\item{\strong{firstActiveRow}}{Top row of active region to freeze pane.}
\item{\strong{firstActiveCol}}{Furthest left column of active region to freeze pane.}
\item{\strong{firstRow}}{If \code{TRUE}, freezes the first row (equivalent to firstActiveRow = 2)}
\item{\strong{firstCol}}{If \code{TRUE}, freezes the first column (equivalent to firstActiveCol = 2)}
}

\strong{colWidths Parameters}
\describe{
\item{\strong{colWidths}}{May be a single value for all columns (or "auto"), or a list of vectors that will be recycled for each sheet (see examples)}
}
}

\examples{
x <- data.frame(a = 1, b = 2)
wb <- buildWorkbook(x)

y <- list(a = x, b = x, c = x)
buildWorkbook(y, asTable = TRUE)
buildWorkbook(y, asTable = TRUE, tableStyle = "TableStyleLight8")

}
\seealso{
\code{\link[=write.xlsx]{write.xlsx()}}
}
\author{
Jordan Mark Barbone
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{getCreators}
\alias{getCreators}
\title{Get the names of the authors from the meta data of the file.}
\usage{
getCreators(wb)
}
\arguments{
\item{wb}{A workbook object}
}
\value{
vector of creators
}
\description{
Just a wrapper of wb$getCreators()
}
\examples{

wb <- createWorkbook()
getCreators(wb)
}
\author{
Philipp Schauberger
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{conditionalFormat}
\alias{conditionalFormat}
\title{Add conditional formatting to cells}
\usage{
conditionalFormat(
  wb,
  sheet,
  cols,
  rows,
  rule = NULL,
  style = NULL,
  type = "expression"
)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{cols}{Columns to apply conditional formatting to}

\item{rows}{Rows to apply conditional formatting to}

\item{rule}{The condition under which to apply the formatting or a vector of colours. See examples.}

\item{style}{A style to apply to those cells that satisfy the rule. A Style object returned from createStyle()}

\item{type}{Either 'expression', 'colorscale' or 'databar'. If 'expression' the formatting is determined
by a formula.  If colorScale cells are coloured based on cell value. See examples.}
}
\description{
DEPRECATED! USE \code{\link[=conditionalFormatting]{conditionalFormatting()}}
}
\details{
DEPRECATED! USE \code{\link[=conditionalFormatting]{conditionalFormatting()}}

Valid operators are "<", "<=", ">", ">=", "==", "!=". See Examples.
Default style given by: createStyle(fontColour = "#9C0006", bgFill = "#FFC7CE")
}
\seealso{
\code{\link[=createStyle]{createStyle()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{copyWorkbook}
\alias{copyWorkbook}
\title{Copy a Workbook object.}
\usage{
copyWorkbook(wb)
}
\arguments{
\item{wb}{A workbook object}
}
\value{
Workbook
}
\description{
Just a wrapper of wb$copy()
}
\examples{

wb <- createWorkbook()
wb2 <- wb ## does not create a copy
wb3 <- copyWorkbook(wb) ## wrapper for wb$copy()

addWorksheet(wb, "Sheet1") ## adds worksheet to both wb and wb2 but not wb3

names(wb)
names(wb2)
names(wb3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{groupRows}
\alias{groupRows}
\title{Group Rows}
\usage{
groupRows(wb, sheet, rows, hidden = FALSE, level = -1)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{rows}{Indices of rows to group. Can be either a vector of indices to group at the same level or a (named) list of numeric vectors of indices to create multiple groupings at once. The names of the entries determine the grouping level. If no names are given, the parameter level is used as default.}

\item{hidden}{Logical vector. If TRUE the grouped columns are hidden. Defaults to FALSE}

\item{level}{Grouping level (higher value indicates multiple nestings) for the
group. A vector to assign different grouping levels to the indices.
A value of -1 indicates that the grouping level should be derived
from the existing grouping (one level added)}
}
\description{
Group a selection of rows
}
\examples{
wb <- createWorkbook()
addWorksheet(wb, 'Sheet1')
addWorksheet(wb, 'Sheet2')

writeData(wb, "Sheet1", iris)
writeData(wb, "Sheet2", iris)

## create list of groups
# lines used for grouping (here: species)
grp <- list(
  seq(2, 51),
  seq(52, 101),
  seq(102, 151)
)
# assign group levels
names(grp) <- c("1","0","1")
groupRows(wb, "Sheet1", rows = grp)

# different grouping
names(grp) <- c("1","2","3")
groupRows(wb, "Sheet2", rows = grp)

# alternatively, one can call groupRows multiple times
addWorksheet(wb, 'Sheet3')
writeData(wb, "Sheet3", iris)
groupRows(wb, "Sheet3", 2:51, level = 1)
groupRows(wb, "Sheet3", 102:151, level = 1)

addWorksheet(wb, 'Sheet4')
writeData(wb, "Sheet4", iris)
groupRows(wb, "Sheet4", 2:51, level = 1)
groupRows(wb, "Sheet4", 52:101, level = 2)
groupRows(wb, "Sheet4", 102:151, level = 3)

# Nested grouping can also be achieved without explicitly given the levels
addWorksheet(wb, 'Sheet5')
writeData(wb, "Sheet5", iris)
groupRows(wb, "Sheet5", 2:151)
groupRows(wb, "Sheet5", 52:151)
groupRows(wb, "Sheet5", 102:151)


}
\seealso{
\code{\link[=ungroupRows]{ungroupRows()}} to ungroup rows. \code{\link[=groupColumns]{groupColumns()}} for grouping columns.
}
\author{
Joshua Sturm
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{int2col}
\alias{int2col}
\title{Convert integer to Excel column}
\usage{
int2col(x)
}
\arguments{
\item{x}{A numeric vector}
}
\description{
Converts an integer to an Excel column label.
}
\examples{
int2col(1:10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadWorkbook.R
\name{loadWorkbook}
\alias{loadWorkbook}
\title{Load an existing .xlsx file}
\usage{
loadWorkbook(file, xlsxFile = NULL, isUnzipped = FALSE, na.convert = TRUE)
}
\arguments{
\item{file}{A path to an existing .xlsx or .xlsm file}

\item{xlsxFile}{alias for file}

\item{isUnzipped}{Set to TRUE if the xlsx file is already unzipped}

\item{na.convert}{Should empty/blank cells be converted to \code{NA_character_}. Defaults to TRUE.}
}
\value{
Workbook object.
}
\description{
loadWorkbook returns a workbook object conserving styles and
formatting of the original .xlsx file.
}
\examples{
## load existing workbook from package folder
wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))
names(wb) # list worksheets
wb ## view object
## Add a worksheet
addWorksheet(wb, "A new worksheet")

## Save workbook
\dontrun{
saveWorkbook(wb, "loadExample.xlsx", overwrite = TRUE)
}

}
\seealso{
\code{\link[=removeWorksheet]{removeWorksheet()}}
}
\author{
Alexander Walker, Philipp Schauberger
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{dataValidation}
\alias{dataValidation}
\title{Add data validation to cells}
\usage{
dataValidation(
  wb,
  sheet,
  cols,
  rows,
  type,
  operator,
  value,
  allowBlank = TRUE,
  showInputMsg = TRUE,
  showErrorMsg = TRUE
)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{cols}{Contiguous columns to apply conditional formatting to}

\item{rows}{Contiguous rows to apply conditional formatting to}

\item{type}{One of 'whole', 'decimal', 'date', 'time', 'textLength', 'list' (see examples)}

\item{operator}{One of 'between', 'notBetween', 'equal',
'notEqual', 'greaterThan', 'lessThan', 'greaterThanOrEqual', 'lessThanOrEqual'}

\item{value}{a vector of length 1 or 2 depending on operator (see examples)}

\item{allowBlank}{logical}

\item{showInputMsg}{logical}

\item{showErrorMsg}{logical}
}
\description{
Add Excel data validation to cells
}
\examples{
wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")
addWorksheet(wb, "Sheet 2")

writeDataTable(wb, 1, x = iris[1:30, ])

dataValidation(wb, 1,
  col = 1:3, rows = 2:31, type = "whole",
  operator = "between", value = c(1, 9)
)

dataValidation(wb, 1,
  col = 5, rows = 2:31, type = "textLength",
  operator = "between", value = c(4, 6)
)


## Date and Time cell validation
df <- data.frame(
  "d" = as.Date("2016-01-01") + -5:5,
  "t" = as.POSIXct("2016-01-01") + -5:5 * 10000
)

writeData(wb, 2, x = df)
dataValidation(wb, 2,
  col = 1, rows = 2:12, type = "date",
  operator = "greaterThanOrEqual", value = as.Date("2016-01-01")
)

dataValidation(wb, 2,
  col = 2, rows = 2:12, type = "time",
  operator = "between", value = df$t[c(4, 8)]
)
\dontrun{
saveWorkbook(wb, "dataValidationExample.xlsx", overwrite = TRUE)
}


######################################################################
## If type == 'list'
# operator argument is ignored.

wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")
addWorksheet(wb, "Sheet 2")

writeDataTable(wb, sheet = 1, x = iris[1:30, ])
writeData(wb, sheet = 2, x = sample(iris$Sepal.Length, 10))

dataValidation(wb, 1, col = 1, rows = 2:31, type = "list", value = "'Sheet 2'!$A$1:$A$10")

# openXL(wb)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helperFunctions.R
\name{makeHyperlinkString}
\alias{makeHyperlinkString}
\title{create Excel hyperlink string}
\usage{
makeHyperlinkString(sheet, row = 1, col = 1, text = NULL, file = NULL)
}
\arguments{
\item{sheet}{Name of a worksheet}

\item{row}{integer row number for hyperlink to link to}

\item{col}{column number of letter for hyperlink to link to}

\item{text}{display text}

\item{file}{Excel file name to point to. If NULL hyperlink is internal.}
}
\description{
Wrapper to create internal hyperlink string to pass to writeFormula(). Either link to external urls or local files or straight to cells of local Excel sheets.
}
\examples{

## Writing internal hyperlinks
wb <- createWorkbook()
addWorksheet(wb, "Sheet1")
addWorksheet(wb, "Sheet2")
addWorksheet(wb, "Sheet 3")
writeData(wb, sheet = 3, x = iris)

## External Hyperlink
x <- c("https://www.google.com", "https://www.google.com.au")
names(x) <- c("google", "google Aus")
class(x) <- "hyperlink"

writeData(wb, sheet = 1, x = x, startCol = 10)


## Internal Hyperlink - create hyperlink formula manually
writeFormula(
  wb, "Sheet1",
  x = '=HYPERLINK(\"#Sheet2!B3\", "Text to Display - Link to Sheet2")',
  startCol = 3
)

## Internal - No text to display using makeHyperlinkString() function
writeFormula(
  wb, "Sheet1",
  startRow = 1,
  x = makeHyperlinkString(sheet = "Sheet 3", row = 1, col = 2)
)

## Internal - Text to display
writeFormula(
  wb, "Sheet1",
  startRow = 2,
  x = makeHyperlinkString(
    sheet = "Sheet 3", row = 1, col = 2,
    text = "Link to Sheet 3"
  )
)

## Link to file - No text to display
writeFormula(
  wb, "Sheet1",
  startRow = 4,
  x = makeHyperlinkString(
    sheet = "testing", row = 3, col = 10,
    file = system.file("extdata", "loadExample.xlsx", package = "openxlsx")
  )
)

## Link to file - Text to display
writeFormula(
  wb, "Sheet1",
  startRow = 3,
  x = makeHyperlinkString(
    sheet = "testing", row = 3, col = 10,
    file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"),
    text = "Link to File."
  )
)

## Link to external file - Text to display
writeFormula(
  wb, "Sheet1",
  startRow = 10, startCol = 1,
  x = '=HYPERLINK("[C:/Users]", "Link to an external file")'
)

## Link to internal file
x = makeHyperlinkString(text = "test.png", file = "D:/somepath/somepicture.png")
writeFormula(wb, "Sheet1", startRow = 11, startCol = 1, x = x)

\dontrun{
saveWorkbook(wb, "internalHyperlinks.xlsx", overwrite = TRUE)
}

}
\seealso{
\code{\link[=writeFormula]{writeFormula()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{pageSetup}
\alias{pageSetup}
\title{Set page margins, orientation and print scaling}
\usage{
pageSetup(
  wb,
  sheet,
  orientation = NULL,
  scale = 100,
  left = 0.7,
  right = 0.7,
  top = 0.75,
  bottom = 0.75,
  header = 0.3,
  footer = 0.3,
  fitToWidth = FALSE,
  fitToHeight = FALSE,
  paperSize = NULL,
  printTitleRows = NULL,
  printTitleCols = NULL,
  summaryRow = NULL,
  summaryCol = NULL
)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{orientation}{Page orientation. One of "portrait" or "landscape"}

\item{scale}{Print scaling. Numeric value between 10 and 400}

\item{left}{left page margin in inches}

\item{right}{right page margin in inches}

\item{top}{top page margin in inches}

\item{bottom}{bottom page margin in inches}

\item{header}{header margin in inches}

\item{footer}{footer margin in inches}

\item{fitToWidth}{If \code{TRUE}, worksheet is scaled to fit to page width on printing.}

\item{fitToHeight}{If \code{TRUE}, worksheet is scaled to fit to page height on printing.}

\item{paperSize}{See details. Default value is 9 (A4 paper).}

\item{printTitleRows}{Rows to repeat at top of page when printing. Integer vector.}

\item{printTitleCols}{Columns to repeat at left when printing. Integer vector.}

\item{summaryRow}{Location of summary rows in groupings. One of "Above" or "Below".}

\item{summaryCol}{Location of summary columns in groupings. One of "Right" or "Left".}
}
\description{
Set page margins, orientation and print scaling
}
\details{
paperSize is an integer corresponding to:
\describe{
\item{\strong{1}}{ Letter paper (8.5 in. by 11 in.)}
\item{\strong{2}}{ Letter small paper (8.5 in. by 11 in.)}
\item{\strong{3}}{ Tabloid paper (11 in. by 17 in.)}
\item{\strong{4}}{ Ledger paper (17 in. by 11 in.)}
\item{\strong{5}}{ Legal paper (8.5 in. by 14 in.)}
\item{\strong{6}}{ Statement paper (5.5 in. by 8.5 in.)}
\item{\strong{7}}{ Executive paper (7.25 in. by 10.5 in.)}
\item{\strong{8}}{ A3 paper (297 mm by 420 mm)}
\item{\strong{9}}{ A4 paper (210 mm by 297 mm)}
\item{\strong{10}}{ A4 small paper (210 mm by 297 mm)}
\item{\strong{11}}{ A5 paper (148 mm by 210 mm)}
\item{\strong{12}}{ B4 paper (250 mm by 353 mm)}
\item{\strong{13}}{ B5 paper (176 mm by 250 mm)}
\item{\strong{14}}{ Folio paper (8.5 in. by 13 in.)}
\item{\strong{15}}{ Quarto paper (215 mm by 275 mm)}
\item{\strong{16}}{ Standard paper (10 in. by 14 in.)}
\item{\strong{17}}{ Standard paper (11 in. by 17 in.)}
\item{\strong{18}}{ Note paper (8.5 in. by 11 in.)}
\item{\strong{19}}{ #9 envelope (3.875 in. by 8.875 in.)}
\item{\strong{20}}{ #10 envelope (4.125 in. by 9.5 in.)}
\item{\strong{21}}{ #11 envelope (4.5 in. by 10.375 in.)}
\item{\strong{22}}{ #12 envelope (4.75 in. by 11 in.)}
\item{\strong{23}}{ #14 envelope (5 in. by 11.5 in.)}
\item{\strong{24}}{ C paper (17 in. by 22 in.)}
\item{\strong{25}}{ D paper (22 in. by 34 in.)}
\item{\strong{26}}{ E paper (34 in. by 44 in.)}
\item{\strong{27}}{ DL envelope (110 mm by 220 mm)}
\item{\strong{28}}{ C5 envelope (162 mm by 229 mm)}
\item{\strong{29}}{ C3 envelope (324 mm by 458 mm)}
\item{\strong{30}}{ C4 envelope (229 mm by 324 mm)}
\item{\strong{31}}{ C6 envelope (114 mm by 162 mm)}
\item{\strong{32}}{ C65 envelope (114 mm by 229 mm)}
\item{\strong{33}}{ B4 envelope (250 mm by 353 mm)}
\item{\strong{34}}{ B5 envelope (176 mm by 250 mm)}
\item{\strong{35}}{ B6 envelope (176 mm by 125 mm)}
\item{\strong{36}}{ Italy envelope (110 mm by 230 mm)}
\item{\strong{37}}{ Monarch envelope (3.875 in. by 7.5 in.).}
\item{\strong{38}}{ 6 3/4 envelope (3.625 in. by 6.5 in.)}
\item{\strong{39}}{ US standard fanfold (14.875 in. by 11 in.)}
\item{\strong{40}}{ German standard fanfold (8.5 in. by 12 in.)}
\item{\strong{41}}{ German legal fanfold (8.5 in. by 13 in.)}
\item{\strong{42}}{ ISO B4 (250 mm by 353 mm)}
\item{\strong{43}}{ Japanese double postcard (200 mm by 148 mm)}
\item{\strong{44}}{ Standard paper (9 in. by 11 in.)}
\item{\strong{45}}{ Standard paper (10 in. by 11 in.)}
\item{\strong{46}}{ Standard paper (15 in. by 11 in.)}
\item{\strong{47}}{ Invite envelope (220 mm by 220 mm)}
\item{\strong{50}}{ Letter extra paper (9.275 in. by 12 in.)}
\item{\strong{51}}{ Legal extra paper (9.275 in. by 15 in.)}
\item{\strong{52}}{ Tabloid extra paper (11.69 in. by 18 in.)}
\item{\strong{53}}{ A4 extra paper (236 mm by 322 mm)}
\item{\strong{54}}{ Letter transverse paper (8.275 in. by 11 in.)}
\item{\strong{55}}{ A4 transverse paper (210 mm by 297 mm)}
\item{\strong{56}}{ Letter extra transverse paper (9.275 in. by 12 in.)}
\item{\strong{57}}{ SuperA/SuperA/A4 paper (227 mm by 356 mm)}
\item{\strong{58}}{ SuperB/SuperB/A3 paper (305 mm by 487 mm)}
\item{\strong{59}}{ Letter plus paper (8.5 in. by 12.69 in.)}
\item{\strong{60}}{ A4 plus paper (210 mm by 330 mm)}
\item{\strong{61}}{ A5 transverse paper (148 mm by 210 mm)}
\item{\strong{62}}{ JIS B5 transverse paper (182 mm by 257 mm)}
\item{\strong{63}}{ A3 extra paper (322 mm by 445 mm)}
\item{\strong{64}}{ A5 extra paper (174 mm by 235 mm)}
\item{\strong{65}}{ ISO B5 extra paper (201 mm by 276 mm)}
\item{\strong{66}}{ A2 paper (420 mm by 594 mm)}
\item{\strong{67}}{ A3 transverse paper (297 mm by 420 mm)}
\item{\strong{68}}{ A3 extra transverse paper (322 mm by 445 mm)}
}
}
\examples{
wb <- createWorkbook()
addWorksheet(wb, "S1")
addWorksheet(wb, "S2")
writeDataTable(wb, 1, x = iris[1:30, ])
writeDataTable(wb, 2, x = iris[1:30, ], xy = c("C", 5))

## landscape page scaled to 50\%
pageSetup(wb, sheet = 1, orientation = "landscape", scale = 50)

## portrait page scales to 300\% with 0.5in left and right margins
pageSetup(wb, sheet = 2, orientation = "portrait", scale = 300, left = 0.5, right = 0.5)


## print titles
addWorksheet(wb, "print_title_rows")
addWorksheet(wb, "print_title_cols")

writeData(wb, "print_title_rows", rbind(iris, iris, iris, iris))
writeData(wb, "print_title_cols", x = rbind(mtcars, mtcars, mtcars), rowNames = TRUE)

pageSetup(wb, sheet = "print_title_rows", printTitleRows = 1) ## first row
pageSetup(wb, sheet = "print_title_cols", printTitleCols = 1, printTitleRows = 1)
\dontrun{
saveWorkbook(wb, "pageSetupExample.xlsx", overwrite = TRUE)
}
}
\author{
Alexander Walker, Joshua Sturm
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writeData.R
\name{writeFormula}
\alias{writeFormula}
\title{Write a character vector as an Excel Formula}
\usage{
writeFormula(
  wb,
  sheet,
  x,
  startCol = 1,
  startRow = 1,
  array = FALSE,
  xy = NULL
)
}
\arguments{
\item{wb}{A Workbook object containing a worksheet.}

\item{sheet}{The worksheet to write to. Can be the worksheet index or name.}

\item{x}{A character vector.}

\item{startCol}{A vector specifying the starting column to write to.}

\item{startRow}{A vector specifying the starting row to write to.}

\item{array}{A bool if the function written is of type array}

\item{xy}{An alternative to specifying \code{startCol} and
\code{startRow} individually.  A vector of the form
\code{c(startCol, startRow)}.}
}
\description{
Write a a character vector containing Excel formula to a worksheet.
}
\details{
Currently only the english version of functions are supported. Please don't use the local translation.
The examples below show a small list of possible formulas:
\itemize{
\item{SUM(B2:B4)}
\item{AVERAGE(B2:B4)}
\item{MIN(B2:B4)}
\item{MAX(B2:B4)}
\item{...}

}
}
\examples{

## There are 3 ways to write a formula

wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")
writeData(wb, "Sheet 1", x = iris)

## SEE int2col() to convert int to Excel column label

## 1. -  As a character vector using writeFormula

v <- c("SUM(A2:A151)", "AVERAGE(B2:B151)") ## skip header row
writeFormula(wb, sheet = 1, x = v, startCol = 10, startRow = 2)
writeFormula(wb, 1, x = "A2 + B2", startCol = 10, startRow = 10)


## 2. - As a data.frame column with class "formula" using writeData

df <- data.frame(
  x = 1:3,
  y = 1:3,
  z = paste(paste0("A", 1:3 + 1L), paste0("B", 1:3 + 1L), sep = " + "),
  z2 = sprintf("ADDRESS(1,\%s)", 1:3),
  stringsAsFactors = FALSE
)

class(df$z) <- c(class(df$z), "formula")
class(df$z2) <- c(class(df$z2), "formula")

addWorksheet(wb, "Sheet 2")
writeData(wb, sheet = 2, x = df)



## 3. - As a vector with class "formula" using writeData

v2 <- c("SUM(A2:A4)", "AVERAGE(B2:B4)", "MEDIAN(C2:C4)")
class(v2) <- c(class(v2), "formula")

writeData(wb, sheet = 2, x = v2, startCol = 10, startRow = 2)

## Save workbook
\dontrun{
saveWorkbook(wb, "writeFormulaExample.xlsx", overwrite = TRUE)
}


## 4. - Writing internal hyperlinks

wb <- createWorkbook()
addWorksheet(wb, "Sheet1")
addWorksheet(wb, "Sheet2")
writeFormula(wb, "Sheet1", x = '=HYPERLINK("#Sheet2!B3", "Text to Display - Link to Sheet2")')

## Save workbook
\dontrun{
saveWorkbook(wb, "writeFormulaHyperlinkExample.xlsx", overwrite = TRUE)
}

}
\seealso{
\code{\link[=writeData]{writeData()}} \code{\link[=makeHyperlinkString]{makeHyperlinkString()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{saveWorkbook}
\alias{saveWorkbook}
\title{save Workbook to file}
\usage{
saveWorkbook(wb, file, overwrite = FALSE, returnValue = FALSE)
}
\arguments{
\item{wb}{A Workbook object to write to file}

\item{file}{A character string naming an xlsx file}

\item{overwrite}{If \code{TRUE}, overwrite any existing file.}

\item{returnValue}{If \code{TRUE}, returns  \code{TRUE} in case of a success, else \code{FALSE}.
If flag is \code{FALSE}, then no return value is returned.}
}
\description{
save a Workbook object to file
}
\examples{
## Create a new workbook and add a worksheet
wb <- createWorkbook("Creator of workbook")
addWorksheet(wb, sheetName = "My first worksheet")

## Save workbook to working directory
\dontrun{
saveWorkbook(wb, file = "saveWorkbookExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=createWorkbook]{createWorkbook()}}

\code{\link[=addWorksheet]{addWorksheet()}}

\code{\link[=loadWorkbook]{loadWorkbook()}}

\code{\link[=writeData]{writeData()}}

\code{\link[=writeDataTable]{writeDataTable()}}
}
\author{
Alexander Walker, Philipp Schauberger
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readWorkbook.R
\name{readWorkbook}
\alias{readWorkbook}
\title{Read from an Excel file or Workbook object}
\usage{
readWorkbook(
  xlsxFile,
  sheet = 1,
  startRow = 1,
  colNames = TRUE,
  rowNames = FALSE,
  detectDates = FALSE,
  skipEmptyRows = TRUE,
  skipEmptyCols = TRUE,
  rows = NULL,
  cols = NULL,
  check.names = FALSE,
  sep.names = ".",
  namedRegion = NULL,
  na.strings = "NA",
  fillMergedCells = FALSE
)
}
\arguments{
\item{xlsxFile}{An xlsx file, Workbook object or URL to xlsx file.}

\item{sheet}{The name or index of the sheet to read data from.}

\item{startRow}{first row to begin looking for data.  Empty rows at the top of a file are always skipped,
regardless of the value of startRow.}

\item{colNames}{If \code{TRUE}, the first row of data will be used as column names.}

\item{rowNames}{If \code{TRUE}, first column of data will be used as row names.}

\item{detectDates}{If \code{TRUE}, attempt to recognise dates and perform conversion.}

\item{skipEmptyRows}{If \code{TRUE}, empty rows are skipped else empty rows after the first row containing data
will return a row of NAs.}

\item{skipEmptyCols}{If \code{TRUE}, empty columns are skipped.}

\item{rows}{A numeric vector specifying which rows in the Excel file to read.
If NULL, all rows are read.}

\item{cols}{A numeric vector specifying which columns in the Excel file to read.
If NULL, all columns are read.}

\item{check.names}{logical. If TRUE then the names of the variables in the data frame
are checked to ensure that they are syntactically valid variable names}

\item{sep.names}{One character which substitutes blanks in column names. By default, "."}

\item{namedRegion}{A named region in the Workbook. If not NULL startRow, rows and cols parameters are ignored.}

\item{na.strings}{A character vector of strings which are to be interpreted as NA. Blank cells will be returned as NA.}

\item{fillMergedCells}{If TRUE, the value in a merged cell is given to all cells within the merge.}
}
\value{
data.frame
}
\description{
Read data from an Excel file or Workbook object into a data.frame
}
\details{
Creates a data.frame of all data in worksheet.
}
\examples{
xlsxFile <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
df1 <- readWorkbook(xlsxFile = xlsxFile, sheet = 1)

xlsxFile <- system.file("extdata", "readTest.xlsx", package = "openxlsx")
df1 <- readWorkbook(xlsxFile = xlsxFile, sheet = 1, rows = c(1, 3, 5), cols = 1:3)
}
\seealso{
\code{\link[=getNamedRegions]{getNamedRegions()}}

\code{\link[=read.xlsx]{read.xlsx()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{addWorksheet}
\alias{addWorksheet}
\title{Add a worksheet to a workbook}
\usage{
addWorksheet(
  wb,
  sheetName,
  gridLines = openxlsx_getOp("gridLines", TRUE),
  tabColour = NULL,
  zoom = 100,
  header = openxlsx_getOp("header"),
  footer = openxlsx_getOp("footer"),
  evenHeader = openxlsx_getOp("evenHeader"),
  evenFooter = openxlsx_getOp("evenFooter"),
  firstHeader = openxlsx_getOp("firstHeader"),
  firstFooter = openxlsx_getOp("firstFooter"),
  visible = TRUE,
  paperSize = openxlsx_getOp("paperSize", 9),
  orientation = openxlsx_getOp("orientation", "portrait"),
  vdpi = openxlsx_getOp("vdpi", 300),
  hdpi = openxlsx_getOp("hdpi", 300)
)
}
\arguments{
\item{wb}{A Workbook object to attach the new worksheet}

\item{sheetName}{A name for the new worksheet}

\item{gridLines}{A logical. If \code{FALSE}, the worksheet grid lines will be hidden.}

\item{tabColour}{Colour of the worksheet tab. A valid colour (belonging to colours()) or a valid hex colour beginning with "#"}

\item{zoom}{A numeric between 10 and 400. Worksheet zoom level as a percentage.}

\item{header}{document header. Character vector of length 3 corresponding to positions left, center, right. Use NA to skip a position.}

\item{footer}{document footer. Character vector of length 3 corresponding to positions left, center, right. Use NA to skip a position.}

\item{evenHeader}{document header for even pages.}

\item{evenFooter}{document footer for even pages.}

\item{firstHeader}{document header for first page only.}

\item{firstFooter}{document footer for first page only.}

\item{visible}{If FALSE, sheet is hidden else visible.}

\item{paperSize}{An integer corresponding to a paper size. See ?pageSetup for details.}

\item{orientation}{One of "portrait" or "landscape"}

\item{vdpi}{Vertical DPI. Can be set with options("openxlsx.dpi" = X) or options("openxlsx.vdpi" = X)}

\item{hdpi}{Horizontal DPI. Can be set with options("openxlsx.dpi" = X) or options("openxlsx.hdpi" = X)}
}
\value{
XML tree
}
\description{
Add a worksheet to a Workbook object
}
\details{
Headers and footers can contain special tags
\describe{
\item{\strong{&[Page]}}{ Page number}
\item{\strong{&[Pages]}}{ Number of pages}
\item{\strong{&[Date]}}{ Current date}
\item{\strong{&[Time]}}{ Current time}
\item{\strong{&[Path]}}{ File path}
\item{\strong{&[File]}}{ File name}
\item{\strong{&[Tab]}}{ Worksheet name}
}
}
\examples{
## Create a new workbook
wb <- createWorkbook("Fred")

## Add 3 worksheets
addWorksheet(wb, "Sheet 1")
addWorksheet(wb, "Sheet 2", gridLines = FALSE)
addWorksheet(wb, "Sheet 3", tabColour = "red")
addWorksheet(wb, "Sheet 4", gridLines = FALSE, tabColour = "#4F81BD")

## Headers and Footers
addWorksheet(wb, "Sheet 5",
  header = c("ODD HEAD LEFT", "ODD HEAD CENTER", "ODD HEAD RIGHT"),
  footer = c("ODD FOOT RIGHT", "ODD FOOT CENTER", "ODD FOOT RIGHT"),
  evenHeader = c("EVEN HEAD LEFT", "EVEN HEAD CENTER", "EVEN HEAD RIGHT"),
  evenFooter = c("EVEN FOOT RIGHT", "EVEN FOOT CENTER", "EVEN FOOT RIGHT"),
  firstHeader = c("TOP", "OF FIRST", "PAGE"),
  firstFooter = c("BOTTOM", "OF FIRST", "PAGE")
)

addWorksheet(wb, "Sheet 6",
  header = c("&[Date]", "ALL HEAD CENTER 2", "&[Page] / &[Pages]"),
  footer = c("&[Path]&[File]", NA, "&[Tab]"),
  firstHeader = c(NA, "Center Header of First Page", NA),
  firstFooter = c(NA, "Center Footer of First Page", NA)
)

addWorksheet(wb, "Sheet 7",
  header = c("ALL HEAD LEFT 2", "ALL HEAD CENTER 2", "ALL HEAD RIGHT 2"),
  footer = c("ALL FOOT RIGHT 2", "ALL FOOT CENTER 2", "ALL FOOT RIGHT 2")
)

addWorksheet(wb, "Sheet 8",
  firstHeader = c("FIRST ONLY L", NA, "FIRST ONLY R"),
  firstFooter = c("FIRST ONLY L", NA, "FIRST ONLY R")
)

## Need data on worksheet to see all headers and footers
writeData(wb, sheet = 5, 1:400)
writeData(wb, sheet = 6, 1:400)
writeData(wb, sheet = 7, 1:400)
writeData(wb, sheet = 8, 1:400)

## Save workbook
\dontrun{
saveWorkbook(wb, "addWorksheetExample.xlsx", overwrite = TRUE)
}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{removeCellMerge}
\alias{removeCellMerge}
\title{Create a new Workbook object}
\usage{
removeCellMerge(wb, sheet, cols, rows)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{cols}{vector of column indices}

\item{rows}{vector of row indices}
}
\description{
Unmerges any merged cells that intersect
with the region specified by, min(cols):max(cols) X min(rows):max(rows)
}
\seealso{
\code{\link[=mergeCells]{mergeCells()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{pageBreak}
\alias{pageBreak}
\title{add a page break to a worksheet}
\usage{
pageBreak(wb, sheet, i, type = "row")
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{i}{row or column number to insert page break.}

\item{type}{One of "row" or "column" for a row break or column break.}
}
\description{
insert page breaks into a worksheet
}
\examples{
wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")
writeData(wb, sheet = 1, x = iris)

pageBreak(wb, sheet = 1, i = 10, type = "row")
pageBreak(wb, sheet = 1, i = 20, type = "row")
pageBreak(wb, sheet = 1, i = 2, type = "column")
\dontrun{
saveWorkbook(wb, "pageBreakExample.xlsx", TRUE)
}
## In Excel: View tab -> Page Break Preview
}
\seealso{
\code{\link[=addWorksheet]{addWorksheet()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writeDataTable.R
\name{writeDataTable}
\alias{writeDataTable}
\title{Write to a worksheet as an Excel table}
\usage{
writeDataTable(
  wb,
  sheet,
  x,
  startCol = 1,
  startRow = 1,
  xy = NULL,
  colNames = TRUE,
  rowNames = FALSE,
  tableStyle = openxlsx_getOp("tableStyle", "TableStyleLight9"),
  tableName = NULL,
  headerStyle = openxlsx_getOp("headerStyle"),
  withFilter = openxlsx_getOp("withFilter", TRUE),
  keepNA = openxlsx_getOp("keepNA", FALSE),
  na.string = openxlsx_getOp("na.string"),
  sep = ", ",
  stack = FALSE,
  firstColumn = openxlsx_getOp("firstColumn", FALSE),
  lastColumn = openxlsx_getOp("lastColumn", FALSE),
  bandedRows = openxlsx_getOp("bandedRows", TRUE),
  bandedCols = openxlsx_getOp("bandedCols", FALSE),
  col.names,
  row.names
)
}
\arguments{
\item{wb}{A Workbook object containing a
worksheet.}

\item{sheet}{The worksheet to write to. Can be the worksheet index or name.}

\item{x}{A dataframe.}

\item{startCol}{A vector specifying the starting column to write df}

\item{startRow}{A vector specifying the starting row to write df}

\item{xy}{An alternative to specifying startCol and startRow individually.
A vector of the form c(startCol, startRow)}

\item{colNames}{If \code{TRUE}, column names of x are written.}

\item{rowNames}{If \code{TRUE}, row names of x are written.}

\item{tableStyle}{Any excel table style name or "none" (see "formatting" vignette).}

\item{tableName}{name of table in workbook. The table name must be unique.}

\item{headerStyle}{Custom style to apply to column names.}

\item{withFilter}{If \code{TRUE} or \code{NA}, columns with have filters in the first row.}

\item{keepNA}{If \code{TRUE}, NA values are converted to #N/A (or \code{na.string}, if not NULL) in Excel, else NA cells will be empty.}

\item{na.string}{If not NULL, and if \code{keepNA} is \code{TRUE}, NA values are converted to this string in Excel.}

\item{sep}{Only applies to list columns. The separator used to collapse list columns to a character vector e.g. sapply(x$list_column, paste, collapse = sep).}

\item{stack}{If \code{TRUE} the new style is merged with any existing cell styles.  If FALSE, any
existing style is replaced by the new style.
\cr\cr
\cr\strong{The below options correspond to Excel table options:}
\cr
\if{html}{\figure{tableoptions.png}{options: alt="Figure: table_options.png"}}
\if{latex}{\figure{tableoptions.pdf}{options: width=265px}}}

\item{firstColumn}{logical. If TRUE, the first column is bold}

\item{lastColumn}{logical. If TRUE, the last column is bold}

\item{bandedRows}{logical. If TRUE, rows are colour banded}

\item{bandedCols}{logical. If TRUE, the columns are colour banded}

\item{row.names, col.names}{Deprecated, please use \code{rowNames}, \code{colNames} instead}
}
\description{
Write to a worksheet and format as an Excel table
}
\details{
columns of x with class Date/POSIXt, currency, accounting,
hyperlink, percentage are automatically styled as dates, currency, accounting,
hyperlinks, percentages respectively.
}
\examples{
## see package vignettes for further examples.

#####################################################################################
## Create Workbook object and add worksheets
wb <- createWorkbook()
addWorksheet(wb, "S1")
addWorksheet(wb, "S2")
addWorksheet(wb, "S3")


#####################################################################################
## -- write data.frame as an Excel table with column filters
## -- default table style is "TableStyleMedium2"

writeDataTable(wb, "S1", x = iris)

writeDataTable(wb, "S2",
  x = mtcars, xy = c("B", 3), rowNames = TRUE,
  tableStyle = "TableStyleLight9"
)

df <- data.frame(
  "Date" = Sys.Date() - 0:19,
  "T" = TRUE, "F" = FALSE,
  "Time" = Sys.time() - 0:19 * 60 * 60,
  "Cash" = paste("$", 1:20), "Cash2" = 31:50,
  "hLink" = "https://CRAN.R-project.org/",
  "Percentage" = seq(0, 1, length.out = 20),
  "TinyNumbers" = runif(20) / 1E9, stringsAsFactors = FALSE
)

## openxlsx will apply default Excel styling for these classes
class(df$Cash) <- c(class(df$Cash), "currency")
class(df$Cash2) <- c(class(df$Cash2), "accounting")
class(df$hLink) <- "hyperlink"
class(df$Percentage) <- c(class(df$Percentage), "percentage")
class(df$TinyNumbers) <- c(class(df$TinyNumbers), "scientific")

writeDataTable(wb, "S3", x = df, startRow = 4, rowNames = TRUE, tableStyle = "TableStyleMedium9")

#####################################################################################
## Additional Header Styling and remove column filters

writeDataTable(wb,
  sheet = 1, x = iris, startCol = 7, headerStyle = createStyle(textRotation = 45),
  withFilter = FALSE
)


#####################################################################################
## Save workbook
## Open in excel without saving file: openXL(wb)
\dontrun{
saveWorkbook(wb, "writeDataTableExample.xlsx", overwrite = TRUE)
}





#####################################################################################
## Pre-defined table styles gallery

wb <- createWorkbook(paste0("tableStylesGallery.xlsx"))
addWorksheet(wb, "Style Samples")
for (i in 1:21) {
  style <- paste0("TableStyleLight", i)
  writeDataTable(wb,
    x = data.frame(style), sheet = 1,
    tableStyle = style, startRow = 1, startCol = i * 3 - 2
  )
}

for (i in 1:28) {
  style <- paste0("TableStyleMedium", i)
  writeDataTable(wb,
    x = data.frame(style), sheet = 1,
    tableStyle = style, startRow = 4, startCol = i * 3 - 2
  )
}

for (i in 1:11) {
  style <- paste0("TableStyleDark", i)
  writeDataTable(wb,
    x = data.frame(style), sheet = 1,
    tableStyle = style, startRow = 7, startCol = i * 3 - 2
  )
}

## openXL(wb)
\dontrun{
saveWorkbook(wb, file = "tableStylesGallery.xlsx", overwrite = TRUE)
}

}
\seealso{
\code{\link[=addWorksheet]{addWorksheet()}}

\code{\link[=writeData]{writeData()}}

\code{\link[=removeTable]{removeTable()}}

\code{\link[=getTables]{getTables()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{setRowHeights}
\alias{setRowHeights}
\title{Set worksheet row heights}
\usage{
setRowHeights(
  wb,
  sheet,
  rows,
  heights,
  fontsize = NULL,
  factor = 1,
  base_height = 15,
  extra_height = 12,
  wrap = TRUE
)
}
\arguments{
\item{wb}{workbook object}

\item{sheet}{name or index of a worksheet}

\item{rows}{indices of rows to set height}

\item{heights}{heights to set rows to specified in Excel column height units}

\item{fontsize}{font size, optional (get base font size by default)}

\item{factor}{factor to manually adjust font width, e.g., for bold fonts,
optional}

\item{base_height}{basic row height, optional}

\item{extra_height}{additional row height per new line of text, optional}

\item{wrap}{wrap text of entries which exceed the column width, optional}
}
\description{
Set worksheet row heights
}
\examples{
## Create a new workbook
wb <- createWorkbook()

## Add a worksheet
addWorksheet(wb, "Sheet")
sheet <- 1

## Write dummy data
writeData(wb, sheet, "fixed w/fixed h", startCol = 1, startRow = 1)
writeData(wb, sheet, "fixed w/auto h ABC ABC ABC ABC ABC ABC ABC ABC ABC ABC ABC",
 startCol = 2, startRow = 2)
writeData(wb, sheet, "variable w/fixed h", startCol = 3, startRow = 3)

## Set column widths and row heights
setColWidths(wb, sheet, cols = c(1, 2, 3, 4), widths = c(10, 20, "auto", 20))
setRowHeights(wb, sheet, rows = c(1, 2, 8, 4, 6), heights = c(30, "auto", 15, 15, 30))

## Overwrite row 1 height
setRowHeights(wb, sheet, rows = 1, heights = 40)

## Save workbook
\dontrun{
saveWorkbook(wb, "setRowHeightsExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=removeRowHeights]{removeRowHeights()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{activeSheet}
\alias{activeSheet}
\alias{activeSheet<-}
\title{Get/set active sheet of the workbook}
\usage{
activeSheet(wb)

activeSheet(wb) <- value
}
\arguments{
\item{wb}{A workbook object}

\item{value}{index of the active sheet or name of the active sheet}
}
\value{
return the active sheet of the workbook
}
\description{
Get and set active sheet of the workbook
}
\examples{

wb <- createWorkbook()
addWorksheet(wb, sheetName = "S1")
addWorksheet(wb, sheetName = "S2")
addWorksheet(wb, sheetName = "S3")

activeSheet(wb) # default value is the first sheet active
activeSheet(wb) <- 1 ## active sheet S1
activeSheet(wb)
activeSheet(wb) <- "S2" ## active sheet S2
activeSheet(wb)
}
\author{
Philipp Schauberger
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{setHeaderFooter}
\alias{setHeaderFooter}
\title{Set document headers and footers}
\usage{
setHeaderFooter(
  wb,
  sheet,
  header = NULL,
  footer = NULL,
  evenHeader = NULL,
  evenFooter = NULL,
  firstHeader = NULL,
  firstFooter = NULL
)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{header}{document header. Character vector of length 3 corresponding to positions left, center, right. Use NA to skip a position.}

\item{footer}{document footer. Character vector of length 3 corresponding to positions left, center, right. Use NA to skip a position.}

\item{evenHeader}{document header for even pages.}

\item{evenFooter}{document footer for even pages.}

\item{firstHeader}{document header for first page only.}

\item{firstFooter}{document footer for first page only.}
}
\description{
Set document headers and footers
}
\details{
Headers and footers can contain special tags
\describe{
\item{\strong{&[Page]}}{ Page number}
\item{\strong{&[Pages]}}{ Number of pages}
\item{\strong{&[Date]}}{ Current date}
\item{\strong{&[Time]}}{ Current time}
\item{\strong{&[Path]}}{ File path}
\item{\strong{&[File]}}{ File name}
\item{\strong{&[Tab]}}{ Worksheet name}
}
}
\examples{
wb <- createWorkbook()

addWorksheet(wb, "S1")
addWorksheet(wb, "S2")
addWorksheet(wb, "S3")
addWorksheet(wb, "S4")

writeData(wb, 1, 1:400)
writeData(wb, 2, 1:400)
writeData(wb, 3, 3:400)
writeData(wb, 4, 3:400)

setHeaderFooter(wb,
  sheet = "S1",
  header = c("ODD HEAD LEFT", "ODD HEAD CENTER", "ODD HEAD RIGHT"),
  footer = c("ODD FOOT RIGHT", "ODD FOOT CENTER", "ODD FOOT RIGHT"),
  evenHeader = c("EVEN HEAD LEFT", "EVEN HEAD CENTER", "EVEN HEAD RIGHT"),
  evenFooter = c("EVEN FOOT RIGHT", "EVEN FOOT CENTER", "EVEN FOOT RIGHT"),
  firstHeader = c("TOP", "OF FIRST", "PAGE"),
  firstFooter = c("BOTTOM", "OF FIRST", "PAGE")
)

setHeaderFooter(wb,
  sheet = 2,
  header = c("&[Date]", "ALL HEAD CENTER 2", "&[Page] / &[Pages]"),
  footer = c("&[Path]&[File]", NA, "&[Tab]"),
  firstHeader = c(NA, "Center Header of First Page", NA),
  firstFooter = c(NA, "Center Footer of First Page", NA)
)

setHeaderFooter(wb,
  sheet = 3,
  header = c("ALL HEAD LEFT 2", "ALL HEAD CENTER 2", "ALL HEAD RIGHT 2"),
  footer = c("ALL FOOT RIGHT 2", "ALL FOOT CENTER 2", "ALL FOOT RIGHT 2")
)

setHeaderFooter(wb,
  sheet = 4,
  firstHeader = c("FIRST ONLY L", NA, "FIRST ONLY R"),
  firstFooter = c("FIRST ONLY L", NA, "FIRST ONLY R")
)
\dontrun{
saveWorkbook(wb, "setHeaderFooterExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=addWorksheet]{addWorksheet()}} to set headers and footers when adding a worksheet
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-fontSizeLookupTables.R
\docType{data}
\name{openxlsxFontSizeLookupTable}
\alias{openxlsxFontSizeLookupTable}
\alias{openxlsxFontSizeLookupTableBold}
\title{Font Size Lookup tables}
\format{
A data.frame with column names corresponding to font names
}
\usage{
openxlsxFontSizeLookupTable

openxlsxFontSizeLookupTableBold
}
\description{
Lookup tables for font size
}
\examples{
data(openxlsxFontSizeLookupTable)
data(openxlsxFontSizeLookupTableBold)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{showGridLines}
\alias{showGridLines}
\title{Set worksheet gridlines to show or hide.}
\usage{
showGridLines(wb, sheet, showGridLines = FALSE)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{showGridLines}{A logical. If \code{FALSE}, grid lines are hidden.}
}
\description{
Set worksheet gridlines to show or hide.
}
\examples{
wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))
names(wb) ## list worksheets in workbook
showGridLines(wb, 1, showGridLines = FALSE)
showGridLines(wb, "testing", showGridLines = FALSE)
\dontrun{
saveWorkbook(wb, "showGridLinesExample.xlsx", overwrite = TRUE)
}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{removeWorksheet}
\alias{removeWorksheet}
\title{Remove a worksheet from a workbook}
\usage{
removeWorksheet(wb, sheet)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}
}
\description{
Remove a worksheet from a Workbook object

Remove a worksheet from a workbook
}
\examples{
## load a workbook
wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))

## Remove sheet 2
removeWorksheet(wb, 2)

## save the modified workbook
\dontrun{
saveWorkbook(wb, "removeWorksheetExample.xlsx", overwrite = TRUE)
}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{getStyles}
\alias{getStyles}
\title{Returns a list of all styles in the workbook}
\usage{
getStyles(wb)
}
\arguments{
\item{wb}{A workbook object}
}
\description{
Returns list of style objects in the workbook
}
\examples{
## load a workbook
wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))
getStyles(wb)[1:3]
}
\seealso{
\code{\link[=replaceStyle]{replaceStyle()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{if_null_then}
\alias{if_null_then}
\alias{\%||\%}
\title{If NULL then ...}
\usage{
x \%||\% y
}
\arguments{
\item{x}{A value to check}

\item{y}{A value to substitute if x is null}
}
\description{
Replace NULL
}
\examples{
\dontrun{
x <- NULL
x <- x \%||\% "none"
x <- x \%||\% NA
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CommentClass.R
\name{writeComment}
\alias{writeComment}
\title{write a cell comment}
\usage{
writeComment(wb, sheet, col, row, comment, xy = NULL)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A vector of names or indices of worksheets}

\item{col}{Column a column number of letter}

\item{row}{A row number.}

\item{comment}{A Comment object. See \code{\link[=createComment]{createComment()}}.}

\item{xy}{An alternative to specifying \code{col} and
\code{row} individually.  A vector of the form
\code{c(col, row)}.}
}
\description{
Write a Comment object to a worksheet
}
\examples{
wb <- createWorkbook()
addWorksheet(wb, "Sheet 1")

c1 <- createComment(comment = "this is comment")
writeComment(wb, 1, col = "B", row = 10, comment = c1)

s1 <- createStyle(fontSize = 12, fontColour = "red", textDecoration = c("BOLD"))
s2 <- createStyle(fontSize = 9, fontColour = "black")

c2 <- createComment(comment = c("This Part Bold red\n\n", "This part black"), style = c(s1, s2))
c2

writeComment(wb, 1, col = 6, row = 3, comment = c2)
\dontrun{
saveWorkbook(wb, file = "writeCommentExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=createComment]{createComment()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writeData.R
\name{writeData}
\alias{writeData}
\title{Write an object to a worksheet}
\usage{
writeData(
  wb,
  sheet,
  x,
  startCol = 1,
  startRow = 1,
  array = FALSE,
  xy = NULL,
  colNames = TRUE,
  rowNames = FALSE,
  headerStyle = openxlsx_getOp("headerStyle"),
  borders = openxlsx_getOp("borders", "none"),
  borderColour = openxlsx_getOp("borderColour", "black"),
  borderStyle = openxlsx_getOp("borderStyle", "thin"),
  withFilter = openxlsx_getOp("withFilter", FALSE),
  keepNA = openxlsx_getOp("keepNA", FALSE),
  na.string = openxlsx_getOp("na.string"),
  name = NULL,
  sep = ", ",
  col.names,
  row.names
)
}
\arguments{
\item{wb}{A Workbook object containing a worksheet.}

\item{sheet}{The worksheet to write to. Can be the worksheet index or name.}

\item{x}{Object to be written. For classes supported look at the examples.}

\item{startCol}{A vector specifying the starting column to write to.}

\item{startRow}{A vector specifying the starting row to write to.}

\item{array}{A bool if the function written is of type array}

\item{xy}{An alternative to specifying \code{startCol} and
\code{startRow} individually.  A vector of the form
\code{c(startCol, startRow)}.}

\item{colNames}{If \code{TRUE}, column names of x are written.}

\item{rowNames}{If \code{TRUE}, data.frame row names of x are written.}

\item{headerStyle}{Custom style to apply to column names.}

\item{borders}{Either "\code{none}" (default), "\code{surrounding}",
"\code{columns}", "\code{rows}" or \emph{respective abbreviations}.  If
"\code{surrounding}", a border is drawn around the data.  If "\code{rows}",
a surrounding border is drawn with a border around each row. If
"\code{columns}", a surrounding border is drawn with a border between
each column. If "\code{all}" all cell borders are drawn.}

\item{borderColour}{Colour of cell border.  A valid colour (belonging to \code{colours()} or a hex colour code, eg see \href{https://www.w3schools.com/colors/colors_picker.asp}{here}).}

\item{borderStyle}{Border line style
\describe{
\item{\strong{none}}{ no border}
\item{\strong{thin}}{ thin border}
\item{\strong{medium}}{ medium border}
\item{\strong{dashed}}{ dashed border}
\item{\strong{dotted}}{ dotted border}
\item{\strong{thick}}{ thick border}
\item{\strong{double}}{ double line border}
\item{\strong{hair}}{ hairline border}
\item{\strong{mediumDashed}}{ medium weight dashed border}
\item{\strong{dashDot}}{ dash-dot border}
\item{\strong{mediumDashDot}}{ medium weight dash-dot border}
\item{\strong{dashDotDot}}{ dash-dot-dot border}
\item{\strong{mediumDashDotDot}}{ medium weight dash-dot-dot border}
\item{\strong{slantDashDot}}{ slanted dash-dot border}
}}

\item{withFilter}{If \code{TRUE} or \code{NA}, add filters to the column name row. NOTE can only have one filter per worksheet.}

\item{keepNA}{If \code{TRUE}, NA values are converted to #N/A (or \code{na.string}, if not NULL) in Excel, else NA cells will be empty.}

\item{na.string}{If not NULL, and if \code{keepNA} is \code{TRUE}, NA values are converted to this string in Excel.}

\item{name}{If not NULL, a named region is defined.}

\item{sep}{Only applies to list columns. The separator used to collapse list columns to a character vector e.g. sapply(x$list_column, paste, collapse = sep).}

\item{row.names, col.names}{Deprecated, please use \code{rowNames}, \code{colNames} instead}
}
\value{
invisible(0)
}
\description{
Write an object to worksheet with optional styling.
}
\details{
Formulae written using writeFormula to a Workbook object will not get picked up by read.xlsx().
This is because only the formula is written and left to Excel to evaluate the formula when the file is opened in Excel.
}
\examples{

## See formatting vignette for further examples.

## Options for default styling (These are the defaults)
options("openxlsx.borderColour" = "black")
options("openxlsx.borderStyle" = "thin")
options("openxlsx.dateFormat" = "mm/dd/yyyy")
options("openxlsx.datetimeFormat" = "yyyy-mm-dd hh:mm:ss")
options("openxlsx.numFmt" = NULL)

## Change the default border colour to #4F81BD
options("openxlsx.borderColour" = "#4F81BD")


#####################################################################################
## Create Workbook object and add worksheets
wb <- createWorkbook()

## Add worksheets
addWorksheet(wb, "Cars")
addWorksheet(wb, "Formula")


x <- mtcars[1:6, ]
writeData(wb, "Cars", x, startCol = 2, startRow = 3, rowNames = TRUE)

#####################################################################################
## Bordering

writeData(wb, "Cars", x,
  rowNames = TRUE, startCol = "O", startRow = 3,
  borders = "surrounding", borderColour = "black"
) ## black border

writeData(wb, "Cars", x,
  rowNames = TRUE,
  startCol = 2, startRow = 12, borders = "columns"
)

writeData(wb, "Cars", x,
  rowNames = TRUE,
  startCol = "O", startRow = 12, borders = "rows"
)


#####################################################################################
## Header Styles

hs1 <- createStyle(
  fgFill = "#DCE6F1", halign = "CENTER", textDecoration = "italic",
  border = "Bottom"
)

writeData(wb, "Cars", x,
  colNames = TRUE, rowNames = TRUE, startCol = "B",
  startRow = 23, borders = "rows", headerStyle = hs1, borderStyle = "dashed"
)


hs2 <- createStyle(
  fontColour = "#ffffff", fgFill = "#4F80BD",
  halign = "center", valign = "center", textDecoration = "bold",
  border = "TopBottomLeftRight"
)

writeData(wb, "Cars", x,
  colNames = TRUE, rowNames = TRUE,
  startCol = "O", startRow = 23, borders = "columns", headerStyle = hs2
)




#####################################################################################
## Hyperlinks
## - vectors/columns with class 'hyperlink' are written as hyperlinks'

v <- rep("https://CRAN.R-project.org/", 4)
names(v) <- paste0("Hyperlink", 1:4) # Optional: names will be used as display text
class(v) <- "hyperlink"
writeData(wb, "Cars", x = v, xy = c("B", 32))


#####################################################################################
## Formulas
## - vectors/columns with class 'formula' are written as formulas'

df <- data.frame(
  x = 1:3, y = 1:3,
  z = paste0(paste0("A", 1:3 + 1L), paste0("B", 1:3 + 1L), sep = " + "),
  stringsAsFactors = FALSE
)

class(df$z) <- c(class(df$z), "formula")

writeData(wb, sheet = "Formula", x = df)


#####################################################################################
## Save workbook
## Open in excel without saving file: openXL(wb)
\dontrun{
saveWorkbook(wb, "writeDataExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=writeDataTable]{writeDataTable()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{sheetVisible}
\alias{sheetVisible}
\alias{sheetVisible<-}
\title{Get worksheet visible state.}
\usage{
sheetVisible(wb)

sheetVisible(wb) <- value
}
\arguments{
\item{wb}{A workbook object}

\item{value}{a logical vector the same length as sheetVisible(wb)}
}
\value{
Character vector of worksheet names.

TRUE if sheet is visible, FALSE if sheet is hidden
}
\description{
DEPRECATED - Use function 'sheetVisibility()
}
\examples{

wb <- createWorkbook()
addWorksheet(wb, sheetName = "S1", visible = FALSE)
addWorksheet(wb, sheetName = "S2", visible = TRUE)
addWorksheet(wb, sheetName = "S3", visible = FALSE)

sheetVisible(wb)
sheetVisible(wb)[1] <- TRUE ## show sheet 1
sheetVisible(wb)[2] <- FALSE ## hide sheet 2
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{removeRowHeights}
\alias{removeRowHeights}
\title{Remove custom row heights from a worksheet}
\usage{
removeRowHeights(wb, sheet, rows)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{rows}{Indices of rows to remove custom height (if any) from.}
}
\description{
Remove row heights from a worksheet
}
\examples{
## Create a new workbook
wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))

## remove any custom row heights in rows 1 to 10
removeRowHeights(wb, 1, rows = 1:10)
\dontrun{
saveWorkbook(wb, "removeRowHeightsExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=setRowHeights]{setRowHeights()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{mergeCells}
\alias{mergeCells}
\title{Merge cells within a worksheet}
\usage{
mergeCells(wb, sheet, cols, rows)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{cols}{Columns to merge}

\item{rows}{corresponding rows to merge}
}
\description{
Merge cells within a worksheet
}
\details{
As merged region must be rectangular, only min and max of cols and rows are used.
}
\examples{
## Create a new workbook
wb <- createWorkbook()

## Add a worksheet
addWorksheet(wb, "Sheet 1")
addWorksheet(wb, "Sheet 2")

## Merge cells: Row 2 column C to F (3:6)
mergeCells(wb, "Sheet 1", cols = 2, rows = 3:6)

## Merge cells:Rows 10 to 20 columns A to J (1:10)
mergeCells(wb, 1, cols = 1:10, rows = 10:20)

## Intersecting merges
mergeCells(wb, 2, cols = 1:10, rows = 1)
mergeCells(wb, 2, cols = 5:10, rows = 2)
mergeCells(wb, 2, cols = c(1, 10), rows = 12) ## equivalent to 1:10 as only min/max are used
# mergeCells(wb, 2, cols = 1, rows = c(1,10)) # Throws error because intersects existing merge

## remove merged cells
removeCellMerge(wb, 2, cols = 1, rows = 1) # removes any intersecting merges
mergeCells(wb, 2, cols = 1, rows = 1:10) # Now this works

## Save workbook
\dontrun{
saveWorkbook(wb, "mergeCellsExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=removeCellMerge]{removeCellMerge()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{convertToDate}
\alias{convertToDate}
\title{Convert from excel date number to R Date type}
\usage{
convertToDate(x, origin = "1900-01-01", ...)
}
\arguments{
\item{x}{A vector of integers}

\item{origin}{date. Default value is for Windows Excel 2010}

\item{...}{additional parameters passed to as.Date()}
}
\description{
Convert from excel date number to R Date type
}
\details{
Excel stores dates as number of days from some origin day
}
\examples{
## 2014 April 21st to 25th
convertToDate(c(41750, 41751, 41752, 41753, 41754, NA))
convertToDate(c(41750.2, 41751.99, NA, 41753))
}
\seealso{
\code{\link[=writeData]{writeData()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{temp_xlsx}
\alias{temp_xlsx}
\title{helper function to create tempory directory for testing purpose}
\usage{
temp_xlsx(name = "temp_xlsx")
}
\arguments{
\item{name}{for the temp file}
}
\description{
helper function to create tempory directory for testing purpose
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{getSheetNames}
\alias{getSheetNames}
\title{Get names of worksheets}
\usage{
getSheetNames(file)
}
\arguments{
\item{file}{An xlsx or xlsm file.}
}
\value{
Character vector of worksheet names.
}
\description{
Returns the worksheet names within an xlsx file
}
\examples{
getSheetNames(system.file("extdata", "readTest.xlsx", package = "openxlsx"))
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{replaceStyle}
\alias{replaceStyle}
\title{Replace an existing cell style}
\usage{
replaceStyle(wb, index, newStyle)
}
\arguments{
\item{wb}{A workbook object}

\item{index}{Index of style object to replace}

\item{newStyle}{A style to replace the existing style as position index}
}
\description{
Replace an existing cell style

Replace a style object
}
\examples{

## load a workbook
wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))

## create a new style and replace style 2

newStyle <- createStyle(fgFill = "#00FF00")

## replace style 2
getStyles(wb)[1:3] ## prints styles
replaceStyle(wb, 2, newStyle = newStyle)

## Save workbook
\dontrun{
saveWorkbook(wb, "replaceStyleExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=getStyles]{getStyles()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{convertToDateTime}
\alias{convertToDateTime}
\title{Convert from excel time number to R POSIXct type.}
\usage{
convertToDateTime(x, origin = "1900-01-01", ...)
}
\arguments{
\item{x}{A numeric vector}

\item{origin}{date. Default value is for Windows Excel 2010}

\item{...}{Additional parameters passed to as.POSIXct}
}
\description{
Convert from excel time number to R POSIXct type.
}
\details{
Excel stores dates as number of days from some origin date
}
\examples{
## 2014-07-01, 2014-06-30, 2014-06-29
x <- c(41821.8127314815, 41820.8127314815, NA, 41819, NaN)
convertToDateTime(x)
convertToDateTime(x, tz = "Australia/Perth")
convertToDateTime(x, tz = "UTC")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{deleteData}
\alias{deleteData}
\title{Delete cell data}
\usage{
deleteData(wb, sheet, cols, rows, gridExpand = FALSE)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{cols}{columns to delete data from.}

\item{rows}{Rows to delete data from.}

\item{gridExpand}{If \code{TRUE}, all data in rectangle min(rows):max(rows) X min(cols):max(cols)
will be removed.}
}
\description{
Delete contents and styling from a cell.
}
\examples{
## write some data
wb <- createWorkbook()
addWorksheet(wb, "Worksheet 1")
x <- data.frame(matrix(runif(200), ncol = 10))
writeData(wb, sheet = 1, x = x, startCol = 2, startRow = 3, colNames = FALSE)

## delete some data
deleteData(wb, sheet = 1, cols = 3:5, rows = 5:7, gridExpand = TRUE)
deleteData(wb, sheet = 1, cols = 7:9, rows = 5:7, gridExpand = TRUE)
deleteData(wb, sheet = 1, cols = LETTERS, rows = 18, gridExpand = TRUE)
\dontrun{
saveWorkbook(wb, "deleteDataExample.xlsx", overwrite = TRUE)
}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writeData.R
\name{as.character.formula}
\alias{as.character.formula}
\title{\code{as.character.formula()}}
\usage{
\method{as.character}{formula}(x, ...)
}
\arguments{
\item{x}{object to be coerced or tested.}

\item{...}{Not implemented}
}
\value{
\code{base::as.character.default(x)}
}
\description{
This function exists to prevent conflicts with \code{as.character.formula} methods
from other packages
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{protectWorksheet}
\alias{protectWorksheet}
\title{Protect a worksheet from modifications}
\usage{
protectWorksheet(
  wb,
  sheet,
  protect = TRUE,
  password = NULL,
  lockSelectingLockedCells = NULL,
  lockSelectingUnlockedCells = NULL,
  lockFormattingCells = NULL,
  lockFormattingColumns = NULL,
  lockFormattingRows = NULL,
  lockInsertingColumns = NULL,
  lockInsertingRows = NULL,
  lockInsertingHyperlinks = NULL,
  lockDeletingColumns = NULL,
  lockDeletingRows = NULL,
  lockSorting = NULL,
  lockAutoFilter = NULL,
  lockPivotTables = NULL,
  lockObjects = NULL,
  lockScenarios = NULL
)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{protect}{Whether to protect or unprotect the sheet (default=TRUE)}

\item{password}{(optional) password required to unprotect the worksheet}

\item{lockSelectingLockedCells}{Whether selecting locked cells is locked}

\item{lockSelectingUnlockedCells}{Whether selecting unlocked cells is locked}

\item{lockFormattingCells}{Whether formatting cells is locked}

\item{lockFormattingColumns}{Whether formatting columns is locked}

\item{lockFormattingRows}{Whether formatting rows is locked}

\item{lockInsertingColumns}{Whether inserting columns is locked}

\item{lockInsertingRows}{Whether inserting rows is locked}

\item{lockInsertingHyperlinks}{Whether inserting hyperlinks is locked}

\item{lockDeletingColumns}{Whether deleting columns is locked}

\item{lockDeletingRows}{Whether deleting rows is locked}

\item{lockSorting}{Whether sorting is locked}

\item{lockAutoFilter}{Whether auto-filter is locked}

\item{lockPivotTables}{Whether pivot tables are locked}

\item{lockObjects}{Whether objects are locked}

\item{lockScenarios}{Whether scenarios are locked}
}
\description{
Protect or unprotect a worksheet from modifications by the user in the graphical user interface. Replaces an existing protection.
}
\examples{
wb <- createWorkbook()
addWorksheet(wb, "S1")
writeDataTable(wb, 1, x = iris[1:30, ])
# Formatting cells / columns is allowed , but inserting / deleting columns is protected:
protectWorksheet(wb, "S1",
  protect = TRUE,
  lockFormattingCells = FALSE, lockFormattingColumns = FALSE,
  lockInsertingColumns = TRUE, lockDeletingColumns = TRUE
)

# Remove the protection
protectWorksheet(wb, "S1", protect = FALSE)
\dontrun{
saveWorkbook(wb, "pageSetupExample.xlsx", overwrite = TRUE)
}
}
\author{
Reinhold Kainhofer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{setColWidths}
\alias{setColWidths}
\title{Set worksheet column widths}
\usage{
setColWidths(
  wb,
  sheet,
  cols,
  widths = 8.43,
  hidden = rep(FALSE, length(cols)),
  ignoreMergedCells = FALSE
)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{cols}{Indices of cols to set width}

\item{widths}{widths to set cols to specified in Excel column width units or "auto" for automatic sizing. The widths argument is
recycled to the length of cols.}

\item{hidden}{Logical vector. If TRUE the column is hidden.}

\item{ignoreMergedCells}{Ignore any cells that have been merged with other cells in the calculation of "auto" column widths.}
}
\description{
Set worksheet column widths to specific width or "auto".
}
\details{
The global min and max column width for "auto" columns is set by (default values show):
\itemize{
\item{options("openxlsx.minWidth" = 3)}
\item{options("openxlsx.maxWidth" = 250)} ## This is the maximum width allowed in Excel
}

NOTE: The calculation of column widths can be slow for large worksheets.

NOTE: The \code{hidden} parameter may conflict with the one set in \code{groupColumns}; changing one will update the other.
}
\examples{
## Create a new workbook
wb <- createWorkbook()

## Add a worksheet
addWorksheet(wb, "Sheet 1")


## set col widths
setColWidths(wb, 1, cols = c(1, 4, 6, 7, 9), widths = c(16, 15, 12, 18, 33))

## auto columns
addWorksheet(wb, "Sheet 2")
writeData(wb, sheet = 2, x = iris)
setColWidths(wb, sheet = 2, cols = 1:5, widths = "auto")

## Save workbook
\dontrun{
saveWorkbook(wb, "setColWidthsExample.xlsx", overwrite = TRUE)
}

}
\seealso{
\code{\link[=removeColWidths]{removeColWidths()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{all.equal}
\alias{all.equal}
\alias{all.equal.Workbook}
\title{Check equality of workbooks}
\usage{
\method{all.equal}{Workbook}(target, current, ...)
}
\arguments{
\item{target}{A \code{Workbook} object}

\item{current}{A \code{Workbook} object}

\item{...}{ignored}
}
\description{
Check equality of workbooks
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{removeColWidths}
\alias{removeColWidths}
\title{Remove column widths from a worksheet}
\usage{
removeColWidths(wb, sheet, cols)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{cols}{Indices of columns to remove custom width (if any) from.}
}
\description{
Remove column widths from a worksheet
}
\examples{
## Create a new workbook
wb <- loadWorkbook(file = system.file("extdata", "loadExample.xlsx", package = "openxlsx"))

## remove column widths in columns 1 to 20
removeColWidths(wb, 1, cols = 1:20)
\dontrun{
saveWorkbook(wb, "removeColWidthsExample.xlsx", overwrite = TRUE)
}
}
\seealso{
\code{\link[=setColWidths]{setColWidths()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{createStyle}
\alias{createStyle}
\title{Create a cell style}
\usage{
createStyle(
  fontName = NULL,
  fontSize = NULL,
  fontColour = NULL,
  numFmt = openxlsx_getOp("numFmt", "GENERAL"),
  border = NULL,
  borderColour = openxlsx_getOp("borderColour", "black"),
  borderStyle = openxlsx_getOp("borderStyle", "thin"),
  bgFill = NULL,
  fgFill = NULL,
  halign = NULL,
  valign = NULL,
  textDecoration = NULL,
  wrapText = FALSE,
  textRotation = NULL,
  indent = NULL,
  locked = NULL,
  hidden = NULL
)
}
\arguments{
\item{fontName}{A name of a font. Note the font name is not validated. If fontName is NULL,
the workbook base font is used. (Defaults to Calibri)}

\item{fontSize}{Font size. A numeric greater than 0.
If fontSize is NULL, the workbook base font size is used. (Defaults to 11)}

\item{fontColour}{Colour of text in cell.  A valid hex colour beginning with "#"
or one of colours(). If fontColour is NULL, the workbook base font colours is used.
(Defaults to black)}

\item{numFmt}{Cell formatting
\itemize{
\item{\strong{GENERAL}}
\item{\strong{NUMBER}}
\item{\strong{CURRENCY}}
\item{\strong{ACCOUNTING}}
\item{\strong{DATE}}
\item{\strong{LONGDATE}}
\item{\strong{TIME}}
\item{\strong{PERCENTAGE}}
\item{\strong{FRACTION}}
\item{\strong{SCIENTIFIC}}
\item{\strong{TEXT}}
\item{\strong{COMMA}  for comma separated thousands}
\item{For date/datetime styling a combination of d, m, y and punctuation marks}
\item{For numeric rounding use "0.00" with the preferred number of decimal places}
}}

\item{border}{Cell border. A vector of "top", "bottom", "left", "right" or a single string).
\describe{
\item{\strong{"top"}}{ Top border}
\item{\strong{bottom}}{ Bottom border}
\item{\strong{left}}{ Left border}
\item{\strong{right}}{ Right border}
\item{\strong{TopBottom} or \strong{c("top", "bottom")}}{ Top and bottom border}
\item{\strong{LeftRight} or \strong{c("left", "right")}}{ Left and right border}
\item{\strong{TopLeftRight} or \strong{c("top", "left", "right")}}{ Top, Left and right border}
\item{\strong{TopBottomLeftRight} or \strong{c("top", "bottom", "left", "right")}}{ All borders}
}}

\item{borderColour}{Colour of cell border vector the same length as the number of sides specified in "border"
A valid colour (belonging to colours()) or a valid hex colour beginning with "#"}

\item{borderStyle}{Border line style vector the same length as the number of sides specified in "border"
\describe{
\item{\strong{none}}{ No Border}
\item{\strong{thin}}{ thin border}
\item{\strong{medium}}{ medium border}
\item{\strong{dashed}}{ dashed border}
\item{\strong{dotted}}{ dotted border}
\item{\strong{thick}}{ thick border}
\item{\strong{double}}{ double line border}
\item{\strong{hair}}{ Hairline border}
\item{\strong{mediumDashed}}{ medium weight dashed border}
\item{\strong{dashDot}}{ dash-dot border}
\item{\strong{mediumDashDot}}{ medium weight dash-dot border}
\item{\strong{dashDotDot}}{ dash-dot-dot border}
\item{\strong{mediumDashDotDot}}{ medium weight dash-dot-dot border}
\item{\strong{slantDashDot}}{ slanted dash-dot border}
}}

\item{bgFill}{Cell background fill colour.
A valid colour (belonging to colours()) or a valid hex colour beginning with "#".
--  \strong{Use for conditional formatting styles only.}}

\item{fgFill}{Cell foreground fill colour.
A valid colour (belonging to colours()) or a valid hex colour beginning with "#"}

\item{halign}{Horizontal alignment of cell contents
\describe{
\item{\strong{left}}{ Left horizontal align cell contents}
\item{\strong{right}}{ Right horizontal align cell contents}
\item{\strong{center}}{ Center horizontal align cell contents}
\item{\strong{justify}}{ Justify horizontal align cell contents}
}}

\item{valign}{A name
Vertical alignment of cell contents
\describe{
\item{\strong{top}}{ Top vertical align cell contents}
\item{\strong{center}}{ Center vertical align cell contents}
\item{\strong{bottom}}{ Bottom vertical align cell contents}
}}

\item{textDecoration}{Text styling.
\describe{
\item{\strong{bold}}{ Bold cell contents}
\item{\strong{strikeout}}{ Strikeout cell contents}
\item{\strong{italic}}{ Italicise cell contents}
\item{\strong{underline}}{ Underline cell contents}
\item{\strong{underline2}}{ Double underline cell contents}
\item{\strong{accounting}}{ Single accounting underline cell contents}
\item{\strong{accounting2}}{ Double accounting underline cell contents}
}}

\item{wrapText}{Logical. If \code{TRUE} cell contents will wrap to fit in column.}

\item{textRotation}{Rotation of text in degrees. 255 for vertical text.}

\item{indent}{Horizontal indentation of cell contents.}

\item{locked}{Whether cell contents are locked (if worksheet protection is turned on)}

\item{hidden}{Whether the formula of the cell contents will be hidden (if worksheet protection is turned on)}
}
\value{
A style object
}
\description{
Create a new style to apply to worksheet cells
}
\examples{
## See package vignettes for further examples

## Modify default values of border colour and border line style
options("openxlsx.borderColour" = "#4F80BD")
options("openxlsx.borderStyle" = "thin")

## Size 18 Arial, Bold, left horz. aligned, fill colour #1A33CC, all borders,
style <- createStyle(
  fontSize = 18, fontName = "Arial",
  textDecoration = "bold", halign = "left", fgFill = "#1A33CC", border = "TopBottomLeftRight"
)

## Red, size 24, Bold, italic, underline, center aligned Font, bottom border
style <- createStyle(
  fontSize = 24, fontColour = rgb(1, 0, 0),
  textDecoration = c("bold", "italic", "underline"),
  halign = "center", valign = "center", border = "Bottom"
)

# borderColour is recycled for each border or all colours can be supplied

# colour is recycled 3 times for "Top", "Bottom" & "Right" sides.
createStyle(border = "TopBottomRight", borderColour = "red")

# supply all colours
createStyle(border = "TopBottomLeft", borderColour = c("red", "yellow", "green"))
}
\seealso{
\code{\link[=addStyle]{addStyle()}}
}
\author{
Alexander Walker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{names}
\alias{names}
\alias{names.Workbook}
\alias{names<-.Workbook}
\title{get or set worksheet names}
\usage{
\method{names}{Workbook}(x)

\method{names}{Workbook}(x) <- value
}
\arguments{
\item{x}{A \code{Workbook} object}

\item{value}{a character vector the same length as wb}
}
\description{
get or set worksheet names
}
\examples{

wb <- createWorkbook()
addWorksheet(wb, "S1")
addWorksheet(wb, "S2")
addWorksheet(wb, "S3")

names(wb)
names(wb)[[2]] <- "S2a"
names(wb)
names(wb) <- paste("Sheet", 1:3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{col2int}
\alias{col2int}
\title{Convert Excel column to integer}
\usage{
col2int(x)
}
\arguments{
\item{x}{A character vector}
}
\description{
Converts an Excel column label to an integer.
}
\examples{
col2int(LETTERS)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setWindowSize.R
\name{setWindowSize}
\alias{setWindowSize}
\alias{getWindowSize}
\title{Set and Get Window Size for xlsx file}
\usage{
setWindowSize(
  wb,
  xWindow = NULL,
  yWindow = NULL,
  windowWidth = NULL,
  windowHeight = NULL
)

getWindowSize(wb)
}
\arguments{
\item{wb}{A Workbook object}

\item{xWindow}{the horizontal coordinate of the top left corner of the window}

\item{yWindow}{the vertical coordinate of the top left corner of the window}

\item{windowWidth}{the width of the window}

\item{windowHeight}{the height of the window

Set the size and position of the window when you open the xlsx file.  The units are in twips.  See
\href{https://learn.microsoft.com/en-us/dotnet/api/documentformat.openxml.spreadsheet.workbookview?view=openxml-2.8.1}{Microsoft's documentation for the xlsx standard}}
}
\description{
Set and Get Window Size for xlsx file
}
\examples{
## Create Workbook object and add worksheets
wb <- createWorkbook()
addWorksheet(wb, "S1")
getWindowSize(wb)
setWindowSize(wb, windowWidth = 10000)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{protectWorkbook}
\alias{protectWorkbook}
\title{Protect a workbook from modifications}
\usage{
protectWorkbook(
  wb,
  protect = TRUE,
  password = NULL,
  lockStructure = FALSE,
  lockWindows = FALSE,
  type = 1L
)
}
\arguments{
\item{wb}{A workbook object}

\item{protect}{Whether to protect or unprotect the sheet (default=TRUE)}

\item{password}{(optional) password required to unprotect the workbook}

\item{lockStructure}{Whether the workbook structure should be locked}

\item{lockWindows}{Whether the window position of the spreadsheet should be locked}

\item{type}{Lock type, default 1. From the xml documentation: 1 - Document is password protected. 2 - Document is recommended to be opened as read-only. 4 - Document is enforced to be opened as read-only. 8 - Document is locked for annotation.}
}
\description{
Protect or unprotect a workbook from modifications by the user in the graphical user interface. Replaces an existing protection.
}
\examples{
wb <- createWorkbook()
addWorksheet(wb, "S1")
protectWorkbook(wb, protect = TRUE, password = "Password", lockStructure = TRUE)
\dontrun{
saveWorkbook(wb, "WorkBook_Protection.xlsx", overwrite = TRUE)
}
# Remove the protection
protectWorkbook(wb, protect = FALSE)
\dontrun{
saveWorkbook(wb, "WorkBook_Protection_unprotected.xlsx", overwrite = TRUE)
}
}
\author{
Reinhold Kainhofer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{deleteDataColumn}
\alias{deleteDataColumn}
\title{Deletes a whole column from a workbook}
\usage{
deleteDataColumn(wb, sheet, col)
}
\arguments{
\item{wb}{A workbook object}

\item{sheet}{A name or index of a worksheet}

\item{col}{A column to delete}
}
\description{
Deletes the whole column from a workbook, shifting the remaining columns to the left
}
\examples{
## write some data
wb <- createWorkbook()
addWorksheet(wb, "tester")

for (i in seq(5)) {
  mat <- data.frame(x = rep(paste0(int2col(i), i), 10))
  writeData(wb, sheet = 1, startRow = 1, startCol = i, mat)
  writeFormula(wb, sheet = 1, startRow = 12, startCol = i,
               x = sprintf("=COUNTA(\%s2:\%s11)", int2col(i), int2col(i)))
}
deleteDataColumn(wb, 1, col = 3)
\dontrun{
saveWorkbook(wb, "deleteDataColumnExample.xlsx", overwrite = TRUE)
}
}
\author{
David Zimmermann
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{addCreator}
\alias{addCreator}
\title{Add another author to the meta data of the file.}
\usage{
addCreator(wb, Creator)
}
\arguments{
\item{wb}{A workbook object}

\item{Creator}{A string object with the name of the creator}
}
\description{
Just a wrapper of wb$addCreator()
}
\examples{

wb <- createWorkbook()
addCreator(wb, "test")
}
\author{
Philipp Schauberger
}
